!+ Data module for global fields needed for radar operators
!-------------------------------------------------------------------------------

MODULE data_radar

  !-------------------------------------------------------------------------------
  !
  ! Description:
  !  This module declares all namelist variables and fields needed for the
  !  radar operators
  !
  ! Current Code Owner: Yuefei Zeng, Dorit Epperlein, Ulrich Blahak
  !  phone:  +49  721 608 22805
  !  fax:    +49  721 608 ?????
  !  email:  yuefei.zeng@kit.edu
  !          dorit.epperlein@kit.edu
  !          ulrich.blahak@dwd.de
  !
  ! History:
  ! Version    Date       Name
  ! ---------- ---------- ----
  ! 4.12       2010/10/05 Daniel Leuenberger
  !  Initial version
  !
  ! Code Description:
  ! Language: Fortran 90.
  ! Software Standards: "European Standards for Writing and
  ! Documenting Exchangeable Fortran 90 Code".
  !===============================================================================
  !
  ! Declarations:
  !
  ! Modules used:
  USE data_parameters, ONLY :   &
       ireals,    & ! KIND-type parameter for real variables
       iintegers    ! KIND-type parameter for standard integer variables
  USE data_parallel,      ONLY :  &
       imp_reals, imp_integers, imp_logical, imp_character
  !==============================================================================

  IMPLICIT NONE

  !==============================================================================

  ! include statements
  INCLUDE "mpif.h"

  !===============================================================================

  ! UB>> max. length of elevation-vector and obstimes-vector:
  INTEGER(kind=iintegers), PARAMETER :: nel_max = 100_iintegers
  INTEGER(kind=iintegers), PARAMETER :: nscanstrategies_max = 5_iintegers
  INTEGER(kind=iintegers), PARAMETER :: nobstimes_max = 2000_iintegers
  INTEGER(kind=iintegers), PARAMETER :: ngpsm_max = 51_iintegers
  INTEGER(kind=iintegers), PARAMETER :: nradsta_max = 50_iintegers
  ! UB<<

  ! Parameters for reading of DWD radar data to attach a certain
  ! index to each radar data parameter:
  ! - Index 1 = vr
  ! - Index 2 = qv
  ! - Index 3 = z
  ! - Index 4 = qz
  INTEGER(kind=iintegers), PARAMETER :: ndatakind = 4_iintegers
  CHARACTER(len=1), PARAMETER        :: cndatakind = '4'        ! character form of ndatakind
  CHARACTER(len=2), PARAMETER        :: datakind(ndatakind) = (/'vr','qv','z ','qz'/)

  ! UB>> changed pointers to fixed size vectors, therefore omit problems when distributing
  !      the meta type structure amoung nodes:
  TYPE radar_meta_type
    SEQUENCE
    REAL    (KIND=ireals)             :: lambda             ! Station radar wavelength (m)
    REAL    (KIND=ireals)             :: lat                ! Station latitude  (deg)
    REAL    (KIND=ireals)             :: lon                ! Station longitude (deg)
    REAL    (KIND=ireals)             :: alt_agl_mod        ! Station height AGL model orogr.   (m)
    REAL    (KIND=ireals)             :: alt_msl            ! Station height MSL   (m)
    REAL    (KIND=ireals)             :: msl_mod            ! Model oro height MSL at station   (m)
    REAL    (KIND=ireals)             :: az_inc             ! azimuth increment  (deg)
    REAL    (KIND=ireals)             :: az_start           ! start azimut of first az bin rel. to true north  (deg)
    REAL    (KIND=ireals)             :: ra_inc             ! range increment (m)
!!$ KS>>
!!$    REAL    (KIND=ireals)             :: ra_start           ! range start (m)
!!$ KS<<
    REAL    (KIND=ireals)             :: Theta3             ! vertical 3-dB-oneway beam width in the beam coord. system  [degrees] 
                                                            !   (half width of the one-way beamfunction)
    REAL    (KIND=ireals)             :: Phi3               ! horizontal 3-dB-oneway beam width in the beam coord. system [degrees]
    REAL    (KIND=ireals)             :: dalpha             ! azimutal averaging interval for calculation of one 
                                                            !   averaged pulse (averaging over the number 
                                                            !   of statistically independent pulses)
    REAL    (KIND=ireals)             :: alpha3_eff_0       ! effective horizontal 3-dB-oneway beam width at
                                                            !   elevation = 0.0, depending on the ratio (dalpha/phi3)
                                                            !   --> will be automatically determined from a lookup table
    REAL    (KIND=ireals)             :: smth_interv_fact   ! Factor to determine the azimutal and elevational integration range
                                                            !   for the smoothing over the beam function. The ranges are computed by multiplying this
                                                            !   factor to the effective 3-dB-oneway beamwidths.
                                                            !   ( a value of 1.29 leads to the 90-%-weight-range of the beam function)
    REAL    (KIND=ireals)             :: dt_obs             ! time increment of observatio    
!KS>>
    REAL    (KIND=ireals)             :: ext_nyq            ! Extended Nyquist velocity [m/s]
    REAL    (KIND=ireals)             :: high_nyq           ! high Nyquist velocity [m/s]
    REAL    (KIND=ireals)             :: prf                ! Low PRF        [Hz]
    REAL    (KIND=ireals)             :: dualprf_ratio      ! dual PRF ratio [-]
    REAL    (KIND=ireals)             :: rngate_len         ! range gate length [m]
!KS<<
    REAL    (KIND=ireals)       :: obs_times(nobstimes_max) ! array of observation times (s) as difference to model start in seconds
    REAL    (KIND=ireals)        :: el_arr(nel_max)         ! array of nominal elevations (deg) (these are used for all computations)
    REAL    (KIND=ireals) :: el_arr_default(nel_max,nscanstrategies_max) ! array of nominal elevations (deg) (these are used for all computations)
    REAL    (KIND=ireals)        :: el_arr_obs(nel_max)     ! elevations (deg) from obs files, used for cross-checking with el_arr
    REAL    (KIND=ireals)        :: xabscsm_v(ngpsm_max)    ! array of abscissa values for vertical Gauss-Legendre quadrature [-]
    REAL    (KIND=ireals)        :: xabscsm_h(ngpsm_max)    ! array of abscissa values for horizonal Gauss-Legendre quadrature [-]
    REAL    (KIND=ireals)         :: weigsm_v(ngpsm_max)    ! array of weights for vertical Gauss-Legendre quadrature [-]
    REAL    (KIND=ireals)         :: weigsm_h(ngpsm_max)    ! array of weights for horiztontal Gauss-Legendre quadrature [-]
    INTEGER (KIND=iintegers)          :: station_id         ! Station ident code
    INTEGER (KIND=iintegers)          :: nel                ! actual number of elevations
    INTEGER (KIND=iintegers)          :: nel_default(nscanstrategies_max) ! number of elevations for each possible default scan strategy
    INTEGER (KIND=iintegers)          :: naz                ! number of nominal azimuths
    INTEGER (KIND=iintegers)          :: naz_ncdf(ndatakind)! number of azimuths in NetCDF files (can be > naz)
    INTEGER (KIND=iintegers)          :: nra                ! max. number of range bins occuring in a volume scan
    INTEGER (KIND=iintegers)          :: nobs_times         ! number of observation times
                                                            ! in seconds since model start
    INTEGER (KIND=iintegers)          :: i_nearest_mod      ! i-Index of nearest model grid point to radar station
    INTEGER (KIND=iintegers)          :: j_nearest_mod      ! j-Index of nearest model grid point to radar station
    INTEGER (KIND=iintegers)          :: ngpsm_v            ! number of vertical smoothing points for Gauss-Legendre-quadrature
    INTEGER (KIND=iintegers)          :: ngpsm_h            ! number of horizontal smoothing points for Gauss-Legendre-quadrature
    INTEGER (KIND=iintegers)          :: nrep_ncdf(ndatakind)
    INTEGER (KIND=iintegers)          :: num_gates          ! number of gates averaged
    INTEGER (KIND=iintegers)          :: num_pulses         ! number of pulses integrated
    INTEGER (KIND=iintegers)  :: obs_startrec(nobstimes_max,ndatakind) ! array of observation start records, second index for file 1=vr, 2=qv, 3=z, 4=qz
    INTEGER (KIND=iintegers)  :: obs_endrec(nobstimes_max,ndatakind)   ! array of observation end records, second index for file 1=vr, 2=qv, 3=z, 4=qz
    LOGICAL                           :: lradwind_avail     ! observations available
    LOGICAL                           :: lradwind_q_avail   ! quality flags available
    LOGICAL                           :: lzradar_avail      ! reflectivity available
    LOGICAL                           :: lzradar_q_avail    ! quality flags reflectivity available
    LOGICAL                           :: lfdbkfile_exist    ! feedback file exists
    CHARACTER (LEN=3)                 :: station_name       ! short station name
    CHARACTER (LEN=14)                :: obs_cdate(nobstimes_max)   ! date and time of observation
    CHARACTER (LEN=280)               :: ncdffile_vr      ! name of NetCDF input file for vr
    CHARACTER (LEN=280)               :: ncdffile_qv      ! name of NetCDF input file for qv
    CHARACTER (LEN=280)               :: ncdffile_z       ! name of NetCDF input file for z
    CHARACTER (LEN=280)               :: ncdffile_qz      ! name of NetCDF input file for qz
    CHARACTER (LEN=280)               :: fdbkfile         ! name of NetCDF feedback file for vr, qv, z and qz


  END TYPE radar_meta_type

  TYPE radar_data_type
    REAL    (KIND=ireals),    POINTER :: w_intp(:,:) => NULL()        ! array of horizontal interpolation 
                                                            ! weights for each observation (first dimension)
                                                            ! and each spatial direction i,j,k (second dimension) 
    REAL    (KIND=ireals),    POINTER :: w_intp_smth(:,:) => NULL()
    INTEGER (KIND=iintegers), POINTER :: ind_intp(:,:) => NULL()      ! grid indices for each observation (first dimension)
                                                            ! the second dimension contains of 
                                                            ! 1:     the continuous number of the model grid 
                                                            !        cell associated with the observation
                                                            ! 2,3,4: the observation indices in azimuthal, 
                                                            !        radial and elevation direction (m,n,o)
                                                            ! or 2 : the continuous observation index combined 
                                                            !        from radial, azimutal and elevation index
    INTEGER (KIND=iintegers), POINTER :: ind_intp_smth(:,:) => NULL()
    REAL    (KIND=ireals)   , POINTER :: radwind_mod(:) => NULL()     ! model aequivalents of radial winds on local PE
    REAL    (KIND=ireals)   , POINTER :: radwind_mod_smth(:) => NULL() 
    REAL    (KIND=ireals)   , POINTER :: z_radar_mod(:) => NULL()     ! model aequivalents of radar reflectivity on local PE
    REAL    (KIND=ireals)   , POINTER :: z_radar_mod_smth(:) => NULL()
    REAL    (KIND=ireals)   , POINTER :: z_ext_mod(:) => NULL()       ! model aequivalents of extinction of radar reflectivity
    REAL    (KIND=ireals)   , POINTER :: z_ext_mod_smth(:) => NULL()
    
    REAL    (KIND=ireals)   , POINTER :: hl_loc(:) => NULL()  
    REAL    (KIND=ireals)   , POINTER :: el_loc(:) => NULL()  
 
    INTEGER (KIND=iintegers)          :: nobs               ! number of simulated observations on local PE
    INTEGER (KIND=iintegers)          :: nobs_above_sfc     ! number of simulated observations on local PE above SFC for 4/3 earth model
    INTEGER (KIND=iintegers)          :: nsmth              ! like nobs, but for smth version
    INTEGER (KIND=iintegers)          :: nsmth_above_sfc    ! like nobs_above_sfc, but for smth version with 4/3 earth model

    ! Data vectors for observations:
    INTEGER (KIND=iintegers)          :: nobs_obs(ndatakind)  ! number of observations read from NetCDF on local PE
    INTEGER (KIND=iintegers)          :: check(4)
    REAL    (KIND=ireals)   , POINTER :: radwind_obs(:) => NULL()     ! observed radial winds on local PE
    REAL    (KIND=ireals)   , POINTER :: radwind_obs_q(:) => NULL()       ! quality index of observed radial winds
    REAL    (KIND=ireals)   , POINTER :: z_radar_obs(:) => NULL()         ! observed reflectivity
    REAL    (KIND=ireals)   , POINTER :: z_radar_obs_q(:) => NULL()       ! quality index of observed reflectivity
    INTEGER (KIND=iintegers), POINTER :: ind_intp_obs(:) => NULL()  ! continuous index constructed from ra, az, el
    INTEGER (KIND=iintegers), POINTER :: ind_intp_obs_vr(:) => NULL()
    INTEGER (KIND=iintegers), POINTER :: ind_intp_obs_qv(:) => NULL()
    INTEGER (KIND=iintegers), POINTER :: ind_intp_obs_z(:) => NULL()
    INTEGER (KIND=iintegers), POINTER :: ind_intp_obs_qz(:) => NULL()
 
  END TYPE radar_data_type

  !===============================================================================

!!!! UB>> eliminated and values taken over in radar_data_type
!!$  TYPE radar_smth_type
!!$
!!$    REAL    (KIND = ireals)               :: Theta, Phi           ! 3db range
!!$
!!$    INTEGER (KIND=iintegers)              :: ngp_v, ngp_h         ! number of smoothing points
!!$
!!$    REAL    (KIND=ireals),    POINTER :: xabsc_v(:), xabsc_h(:), weig_v(:), weig_h(:)
!!$
!!$  END TYPE radar_smth_type 


  TYPE radar_grid_type

    REAL    (KIND=ireals),    POINTER :: w_intp(:,:) => NULL()        ! array of horizontal interpolation 
                                                            ! weights for each observation (first dimension)
                                                            ! and each spatial direction i,j,k (second dimension) 

    INTEGER (KIND=iintegers), POINTER :: ind_intp(:,:) => NULL()      ! grid indices for each observation (first dimension)
                                                            ! the second dimension contains of 
                                                            ! 1:     the continuous number of the model grid 
                                                            !        cell associated with the observation
                                                            ! 2,3,4: the observation indices in azimuthal, 
    INTEGER (KIND=iintegers), POINTER :: ind_azgrd(:) => NULL()                                                      

    REAL    (KIND=ireals)   , POINTER :: hl_grd(:) => NULL()     
      
    REAL    (KIND=ireals)   , POINTER :: hl_azgrd(:) => NULL() 

    REAL    (KIND=ireals)             :: al_inc             ! arc length increment  

    INTEGER (KIND=iintegers)          :: naz_nbl            ! number of azimuths including nbl_az extra points (for lsmooth=.true.)

    INTEGER (KIND=iintegers)          :: nal                ! number of arc length

    REAL    (KIND=ireals)   , POINTER :: u_grd(:) => NULL()           ! model aequivalents of u on local PE

    REAL    (KIND=ireals)   , POINTER :: v_grd(:) => NULL()           ! model aequivalents of v on local PE

    REAL    (KIND=ireals)   , POINTER :: w_grd(:) => NULL()           ! model aequivalents of w on local PE

                                                            ! de: add variables for radar reflectivity & extinction
    REAL    (KIND=ireals)   , POINTER :: z_radar_grd(:) => NULL()     ! model aequivalents of radar reflectivity on local PE
    REAL    (KIND=ireals)   , POINTER :: z_ext_grd(:) => NULL()

    REAL    (KIND=ireals)   , POINTER :: rfridx_grd(:) => NULL()

    REAL    (KIND=ireals)             :: rfridx_sta         ! refractive index at radar station  
    REAL    (KIND=ireals)             :: rfridxgrad_sta     ! gradient of refractive index at radar station

    INTEGER (KIND=iintegers)          :: ngrd                ! total number of radar bins in structure

    REAL    (KIND=ireals)   , POINTER :: u_azgrd(:) => NULL()           ! model aequivalents of u on local PE

    REAL    (KIND=ireals)   , POINTER :: v_azgrd(:) => NULL()           ! model aequivalents of v on local PE

    REAL    (KIND=ireals)   , POINTER :: w_azgrd(:) => NULL()           ! model aequivalents of w on local PE

                                                              ! de: add variables for radar reflectivity & extinction
    REAL    (KIND=ireals)   , POINTER :: z_radar_azgrd(:) => NULL()     ! model aequivalents of radar reflectivity on local PE
    REAL    (KIND=ireals)   , POINTER :: z_ext_azgrd(:) => NULL()

    REAL    (KIND=ireals)   , POINTER :: rfridx_azgrd(:) => NULL()


  END TYPE radar_grid_type
     
CONTAINS

  !==============================================================================
  !+ Module procedure in data_radar for the generation of the derived MPI type of
  !  radar_meta_type
  !------------------------------------------------------------------------------
  SUBROUTINE def_mpi_radar_meta_type (mpi_radar_meta_typ)
    !------------------------------------------------------------------------------
    !
    ! Description:    Defines the derived MPI type for the radar_meta_type
    !                 Currently the array pointers need to be taken into
    !                 account in the definition of the MPI type (definition
    !                 of amount of memory), but they
    !                 have to be communicated seperately (I have not found a way
    !                 to include them in the communication of the derived type.
    ! Method     :    See IBM MPI book (online and locatable with Google)
    !------------------------------------------------------------------------------
    !
    ! Subroutine / Function arguments
    ! Scalar arguments with intent(out):
    INTEGER (KIND=iintegers), INTENT(OUT)                                 ::&
         mpi_radar_meta_typ
    !------------------------------------------------------------------------------
    ! Local scalars:
    INTEGER (KIND=iintegers), SAVE     ::  mpi_radar_meta_typ_
    INTEGER (KIND=iintegers)                                              ::&
         ierr, extent
    INTEGER (KIND=iintegers)                                              ::&
         offsets(5), mpitypes(5), blocklengths(5)
    LOGICAL, SAVE :: firstcall = .TRUE.

    !- End of header
    !==============================================================================

    !------------------------------------------------------------------------------
    !- Begin SUBROUTINE 
    !------------------------------------------------------------------------------

    !------------------------------------------------------------------------------
    ! Section 1: Initializations
    !------------------------------------------------------------------------------

    IF (firstcall) THEN

      ! Setup description of the 20 imp_reals variables in the structure and
      ! of the number of elevations and the number of smoothing points and weights:
      offsets(1)      = 0
      mpitypes(1)     = imp_reals
      blocklengths(1) = 20 + (2+nscanstrategies_max)*nel_max + 4*ngpsm_max + nobstimes_max

      ! Get the size of imp_reals in bytes
      CALL MPI_TYPE_EXTENT(imp_reals, extent, ierr)

      ! Setup description of the 11 imp_integers in the structure
      ! and the 2 vectors with dimension ndatakind
      ! and the obs_startrec,obs_endrec with dimension ndatakind*nobstimes_max
      offsets(2)      = offsets(1) + blocklengths(1)*extent 
      mpitypes(2)     = imp_integers
      blocklengths(2) = 11 + nscanstrategies_max + 2*ndatakind + 2*ndatakind*nobstimes_max

      !KS>>
      ! Get the size of imp_integers in bytes
      CALL MPI_TYPE_EXTENT(imp_integers, extent, ierr)

      !    ! Setup description of the 5 imp_logical in the structure
      offsets(3)      = offsets(2) + blocklengths(2)*extent
      mpitypes(3)     = imp_logical
      blocklengths(3) = 5
      !KS<<

      ! Get the size of imp_logical in bytes
      CALL MPI_TYPE_EXTENT(imp_logical, extent, ierr)

      ! Setup description of the 2 strings (len = 3 and 14) in the structure
      ! and the number of observation times
      offsets(4)      = offsets(3) + blocklengths(3)*extent
      mpitypes(4)     = imp_character
      blocklengths(4) =   1*3       & ! 3   = LEN(radar_meta%station_name)
           + nobstimes_max*14       & ! 14  = LEN(radar_meta%obs_cdate)
           + 5*280                    ! 280 = LEN(radar_meta%ncdffile_xx)

      ! Get the size of imp_character in bytes and set the upper boundary
      ! of the structure ( = memory block) explicitly by MPI_UB to prevent any 
      ! byte-number-rounding in transfers
      CALL MPI_TYPE_EXTENT(imp_character, extent, ierr)

      ! Set the upper boundary of the memory block:
      offsets(5)      = offsets(4) + blocklengths(4)*extent
      mpitypes(5)     = MPI_UB
      blocklengths(5) = 1

      ! Define structured type and commit it
      CALL MPI_TYPE_STRUCT(5, blocklengths, offsets, mpitypes, mpi_radar_meta_typ_, ierr)
      CALL MPI_TYPE_COMMIT(mpi_radar_meta_typ_,ierr)

      firstcall = .FALSE.

    END IF

    mpi_radar_meta_typ = mpi_radar_meta_typ_

  END SUBROUTINE def_mpi_radar_meta_type

  !===============================================================================

 

END MODULE data_radar

