MODULE wolken_konstanten

  IMPLICIT NONE 

  !INTERFACE OPERATOR(**)
  !   module procedure power_explog_dd     
  !   module procedure power_explog_dr     
  !END INTERFACE

TYPE PARTICLE
  CHARACTER(20)    :: name  !..Bezeichnung der Partikelklasse
  DOUBLE PRECISION :: nu    !..Breiteparameter der Verteil.
  DOUBLE PRECISION :: mu    !..Exp.-parameter der Verteil.
  DOUBLE PRECISION :: x_max !..maximale Teilchenmasse
  DOUBLE PRECISION :: x_min !..minimale Teilchenmasse
  DOUBLE PRECISION :: a_geo !..Koeff. Geometrie
  DOUBLE PRECISION :: b_geo !..Koeff. Geometrie = 1/3
  DOUBLE PRECISION :: a_vel !..Koeff. Fallgesetz
  DOUBLE PRECISION :: b_vel !..Koeff. Fallgesetz
  DOUBLE PRECISION :: a_ven !..Koeff. Ventilationsparam.
  DOUBLE PRECISION :: b_ven !..Koeff. Ventilationsparam.
  DOUBLE PRECISION :: cap   !..Koeff. Kapazitaet
END TYPE PARTICLE

! UB_20090227>>
  ! Following are two new type declarations to hold the values
  ! of equidistand lookup tables. Up to now, such lookup tables are
  ! used in the Segal-Khain parameterization of CCN-activation and
  ! for determining the wet growth diameter of graupel.

  ! Type declaration for a general 2D equidistant lookup table:
  TYPE lookupt_2D
    INTEGER :: n1  ! number of grid points in x1-direction
    INTEGER :: n2  ! number of grid points in x2-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x1  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x2  ! grid vector in x1-direction
    DOUBLE PRECISION                     :: dx1          ! dx1   (grid distance w.r.t. x1)
    DOUBLE PRECISION                     :: dx2          ! dx2   (grid distance w.r.t. x2)
    DOUBLE PRECISION                     :: odx1         ! one over dx 1
    DOUBLE PRECISION                     :: odx2         ! one over dx 2
    DOUBLE PRECISION, DIMENSION(:,:,:,:), POINTER :: ltable
  END TYPE lookupt_2D
  
  ! Type declaration for a general 4D equidistant lookup table:
  TYPE lookupt_4D
    INTEGER :: n1  ! number of grid points in x1-direction
    INTEGER :: n2  ! number of grid points in x2-direction
    INTEGER :: n3  ! number of grid points in x3-direction
    INTEGER :: n4  ! number of grid points in x4-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x1  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x2  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x3  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x4  ! grid vector in x1-direction
    DOUBLE PRECISION                     :: dx1          ! dx1   (grid distance w.r.t. x1)
    DOUBLE PRECISION                     :: dx2          ! dx2   (grid distance w.r.t. x2)
    DOUBLE PRECISION                     :: dx3          ! dx3   (grid distance w.r.t. x3)
    DOUBLE PRECISION                     :: dx4          ! dx4   (grid distance w.r.t. x4)
    DOUBLE PRECISION                     :: odx1         ! one over dx 1
    DOUBLE PRECISION                     :: odx2         ! one over dx 2
    DOUBLE PRECISION                     :: odx3         ! one over dx 3
    DOUBLE PRECISION                     :: odx4         ! one over dx 4
    DOUBLE PRECISION, DIMENSION(:,:,:,:), POINTER :: ltable
  END TYPE lookupt_4D
! UB_20090227<<

  ! .. allgemeine physikalische Konstanten und Parameter ..

  DOUBLE PRECISION, PARAMETER :: nu_l = 1.460d-5     !..Kinem. Visc. von Luft
  DOUBLE PRECISION, PARAMETER :: D_v  = 3.000d-5     !..Diffusivitaet von Wasserdampf
  DOUBLE PRECISION, PARAMETER :: K_T  = 2.500d-2     !..Waermeleitfaehigkeit     
  DOUBLE PRECISION, PARAMETER :: L_ew = 3.330d+5     !..Schmelzwaerme     
  DOUBLE PRECISION, PARAMETER :: L_ed = 2.830d+6     !..Sublimationswaerme     
  DOUBLE PRECISION, PARAMETER :: L_wd = 2.500d+6     !..Verdampfungswaerme     
  DOUBLE PRECISION, PARAMETER :: K_w  = 0.930d+0     !..Dielektrizitaetsfaktor Wasser
  DOUBLE PRECISION, PARAMETER :: K_i  = 0.176d+0     !..Dielektrizitaetsfaktor Eis
  DOUBLE PRECISION, PARAMETER :: rho0 = 1.225d+0     !..Norm-Luftdichte
  DOUBLE PRECISION, PARAMETER :: T_3  = 2.732d+2     !..Tripelpunkt Wasser
  DOUBLE PRECISION, PARAMETER :: T_f  = 2.330d+2     !..Bei T < T_f kein Fl.wasser
  DOUBLE PRECISION, PARAMETER :: R_l  = 2.870d+2     !..Gaskonstante trockener Luft 
  DOUBLE PRECISION, PARAMETER :: R_d  = 4.615d+2     !..Gaskonstante von Wasserdampf
  DOUBLE PRECISION, PARAMETER :: c_w  = 4.218d+3     !..spezifische Waerme von Wasser
  DOUBLE PRECISION, PARAMETER :: sigma_wa  = 7.1d-2  !..Oberflaechenspannung Wasser-Luft
  DOUBLE PRECISION, PARAMETER :: rho_w   = 1000.0    !..Materialdichte von Fluessigwasser
  DOUBLE PRECISION, PARAMETER :: rho_ice = 900.0     !..Materialdichte von Eis

  ! .. wolkenphysikalische Konstanten und Parameter ..

  DOUBLE PRECISION, PARAMETER :: N_sc = 0.710        !..Schmidt-Zahl (PK, S.541)
  DOUBLE PRECISION, PARAMETER :: n_f  = 0.333        !..Exponent von N_sc im Vent-koeff. (PK, S.541)
  DOUBLE PRECISION, PARAMETER :: m_f  = 0.500        !..Exponent von N_re im Vent-koeff. (PK, S.541)

  DOUBLE PRECISION, PARAMETER :: A_e  = 2.18745584d1 !..Konst. Saettigungsdamppfdruck - Eis
  DOUBLE PRECISION, PARAMETER :: A_w  = 1.72693882d1 !..Konst. Saettigungsdamppfdruck - Wasser
  DOUBLE PRECISION, PARAMETER :: B_e  = 7.66000000d0 !..Konst. Saettigungsdamppfdruck - Eis
  DOUBLE PRECISION, PARAMETER :: B_w  = 3.58600000d1 !..Konst. Saettigungsdamppfdruck - Wasser
  DOUBLE PRECISION, PARAMETER :: e_3  = 6.10780000d2 !..Saettigungsdamppfdruck bei T = T_3

  DOUBLE PRECISION, PARAMETER :: C_mult     = 3.5d8    !..Koeff. fuer Splintering
  DOUBLE PRECISION, PARAMETER :: T_mult_min = 265.0    !..Minimale Temp. Splintering
  DOUBLE PRECISION, PARAMETER :: T_mult_max = 270.0    !..Maximale Temp. Splintering
  DOUBLE PRECISION, PARAMETER :: T_mult_opt = 268.0    !..Optimale Temp. Splintering

  ! ... spezielle Parameter des KAMM2-Wolkenmoduls

  DOUBLE PRECISION, PARAMETER :: r_c     = 12.0e-6         !..mittlerer Radius (bei 1-Moment)
! UB_20100107>> 
! rho_vel auf 1 gesetzt, weil dieser Exponent nun in den
! einzelen Prozess-Routinen auf einen vom mittleren Durchmesser
! abhaengigen Wert gesetzt wird.
!!$ UB: not used anymore!!  DOUBLE PRECISION, PARAMETER :: rho_vel    = 0.5d0        !..Exponent in Dichtekorrektur
!!$ UB: not used anymore!  DOUBLE PRECISION, PARAMETER :: rho_vel    = 1.0d0        !..Exponent in Dichtekorrektur
!!$ UB: not used anymore!!  DOUBLE PRECISION, PARAMETER :: rho_vel_r  = 0.4d0        !..Exponent in Dichtekorrektur fuer Regen
!!$ UB: not used anymore!  DOUBLE PRECISION, PARAMETER :: rho_vel_r  = 1.0d0        !..Exponent in Dichtekorrektur fuer Regen
!!$ UB: not used anymore!  DOUBLE PRECISION, PARAMETER :: rho_vel_c  = 1.0d0        !..fuer Wolkentropfen
! UB_20100107<< 

! UB_20100701>> Andere Hoehenkorrektur der Fallgeschwindigkeiten:
!               (Abgeleitet aus Cd(Re)-Daten von Kugeln mit sinnvollen
!                Modifikationen fuer rauhe Partikel).
! Korrekturvorschrift: v = v0 * (rhol0/rhol)^(a + b * min(Dquer,dknick) ) fuer r,i,s,g,h
!                      v = v0 * (etal0/eta0)^b fuer c
!               (a, b wurden fuer D berechnet, nicht Dqer, aber der
!                Einfachheit halber wird nicht die volle D-Abhaengigkeit
!                in die Kollisions- und Sedimentationsintegrale mitgenommen,
!                sondern im Sinne des Mittelwertsatzes eine
!                mittlere Korrektur vor die Integrale gezogen, die nur
!                vom mittleren Durchmesser Dquer abhaengt)
  DOUBLE PRECISION, PARAMETER :: rhovelcb = 1.0d0
  DOUBLE PRECISION, PARAMETER :: rhovelra = 0.375d0
  DOUBLE PRECISION, PARAMETER :: rhovelrb = 0.025d0
  DOUBLE PRECISION, PARAMETER :: rhovelrdknick = 5.0e-3   ! mm
  DOUBLE PRECISION, PARAMETER :: rhovelia = 0.375d0
  DOUBLE PRECISION, PARAMETER :: rhovelib = 0.01666667
  DOUBLE PRECISION, PARAMETER :: rhovelidknick = 3.0e-3   ! mm
  DOUBLE PRECISION, PARAMETER :: rhovelsa = 0.375d0
  DOUBLE PRECISION, PARAMETER :: rhovelsb = 0.0125
  DOUBLE PRECISION, PARAMETER :: rhovelsdknick = 10.0e-3   ! mm
  DOUBLE PRECISION, PARAMETER :: rhovelga =0.375d0
  DOUBLE PRECISION, PARAMETER :: rhovelgb =0.025d0
  DOUBLE PRECISION, PARAMETER :: rhovelgdknick = 5.0e-3   ! mm
  DOUBLE PRECISION, PARAMETER :: rhovelha =0.375d0
  DOUBLE PRECISION, PARAMETER :: rhovelhb =0.03125d0
  DOUBLE PRECISION, PARAMETER :: rhovelhdknick = 4.0e-3   ! mm
! UB_20100701<<

  ! ... spezielle Parameter des KAMM2-Wolkenmoduls (Eisphase)

  DOUBLE PRECISION, PARAMETER :: e_ii  = 0.00              !..min. Eff.
  DOUBLE PRECISION, PARAMETER :: e_ic  = 0.80              !..max. Eff. fuer ice_cloud_riming
  DOUBLE PRECISION, PARAMETER :: e_sc  = 0.80              !..max. Eff. fuer snow_cloud_riming
  DOUBLE PRECISION, PARAMETER :: e_gc  = 1.00              !..max. Eff. fuer graupel_cloud_riming
  DOUBLE PRECISION, PARAMETER :: e_hc  = 1.00              !..max. Eff. fuer hail_cloud_riming
  DOUBLE PRECISION, PARAMETER :: e_min = 0.01              !..min. Eff. fuer gc,ic,sc
  DOUBLE PRECISION, PARAMETER :: alpha_spacefilling = 0.1  !..Raumerfuellungskoeff (max. 0.68)
  DOUBLE PRECISION, PARAMETER :: ice_s_vel  = 0.25         !..Dispersion der Fallgeschw. 
  DOUBLE PRECISION, PARAMETER :: snow_s_vel = 0.25         !..Dispersion der Fallgeschw.
  DOUBLE PRECISION, PARAMETER :: r_shedding = 500.0e-6     !..mittlerer Radius Shedding
  DOUBLE PRECISION, PARAMETER :: T_shed = 263.2

! AS_2010>> Parameters for Phillips (2008) heterogeneous ice nucleation param.
!           These parameters are the default for nuc_i_typ = 4:
  DOUBLE PRECISION :: &
    na_dust    = 162.e3,   & ! initial number density of dust [1/m³], Phillips08 value 162e3
    na_soot    =  15.e6,   & ! initial number density of soot [1/m³], Phillips08 value 15e6
    na_orga    = 177.e6,   & ! initial number density of organics [1/m3], Phillips08 value 177e6
    ni_het_max = 100.0d3,   & ! max number of IN between 1-10 per liter, i.e. 1d3-10d3 
    ni_hom_max = 5000.0d3     ! number of liquid aerosols between 100-5000 per liter

  ! Look-up table for Phillips et al. nucleation
  INTEGER, PARAMETER :: &
    ttmax  = 30,      &  ! sets limit for temperature in look-up table
    ssmax  = 60          ! sets limit for ice supersaturation in look-up table
! UB_20100421>> von integer nach double geaendert:
  DOUBLE PRECISION, PARAMETER :: &
    ttstep = 2.0d0,       &  ! increment for temperature in look-up table
    ssstep = 1.0d0           ! increment for ice supersaturation in look-up table

  !.. Lookup-table fields for Karcher and Lohmann- ice nucleation
  DOUBLE PRECISION, DIMENSION(0:100,0:100), SAVE :: &
    afrac_dust, &  ! look-up table of activated fraction of dust particles acting as ice nuclei
    afrac_soot, &  ! ... of soot particles
    afrac_orga     ! ... of organic material

  include 'phillips_nucleation.incf'
! AS_2010<<

  ! rain_freeze: der Teil des Regenspektrums kleiner als D_rainfrz_ig
  ! wird nach Gefrieren dem Eis zugeschlagen, der Teil von dort bis zu D_rainfrz_gh dem Graupel
  ! und der Rest dem Hagel.
  DOUBLE PRECISION, PARAMETER :: D_rainfrz_ig = 0.5d-3 !  rain --> ice oder graupel
  DOUBLE PRECISION, PARAMETER :: D_rainfrz_gh = 1.25d-3 ! rain --> graupel oder hail

  DOUBLE PRECISION, PARAMETER :: q_krit_ii = 1.000d-6 ! q-Schwellenwert fuer ice_selfcollection 
  DOUBLE PRECISION, PARAMETER :: D_krit_ii = 50.00d-6 ! D-Schwellenwert fuer ice_selfcollection
  DOUBLE PRECISION, PARAMETER :: D_conv_ii = 75.00d-6 ! D-Schwellenwert fuer ice_selfcollection
  DOUBLE PRECISION, PARAMETER :: q_krit_ic = 1.000d-5 ! q-Schwellenwert fuer ice_cloud_riming
  DOUBLE PRECISION, PARAMETER :: D_krit_ic = 150.0d-6 ! D-Schwellenwert fuer ice_cloud_riming
  DOUBLE PRECISION, PARAMETER :: q_krit_ir = 1.000d-5 ! q-Schwellenwert fuer ice_rain_riming
  DOUBLE PRECISION, PARAMETER :: D_krit_ir = 100.0d-6 ! D-Schwellenwert fuer ice_rain_riming
  DOUBLE PRECISION, PARAMETER :: q_krit_sc = 1.000d-5 ! q-Schwellenwert fuer snow_cloud_riming
  DOUBLE PRECISION, PARAMETER :: D_krit_sc = 150.0d-6 ! D-Schwellenwert fuer snow_cloud_riming
  DOUBLE PRECISION, PARAMETER :: q_krit_sr = 1.000d-5 ! q-Schwellenwert fuer snow_rain_riming
  DOUBLE PRECISION, PARAMETER :: D_krit_sr = 100.0d-6 ! D-Schwellenwert fuer snow_rain_riming
  DOUBLE PRECISION, PARAMETER :: q_krit_gc = 1.000d-6 ! q-Schwellenwert fuer graupel_cloud_riming
  DOUBLE PRECISION, PARAMETER :: D_krit_gc = 100.0d-6 ! D-Schwellenwert fuer graupel_cloud_riming
  DOUBLE PRECISION, PARAMETER :: q_krit_hc = 1.000d-6 ! q-Schwellenwert fuer hail_cloud_riming
  DOUBLE PRECISION, PARAMETER :: D_krit_hc = 100.0d-6 ! D-Schwellenwert fuer hail_cloud_riming
  DOUBLE PRECISION, PARAMETER :: q_krit_fr = 1.000d-6 ! q-Schwellenwert fuer rain_freeze
  DOUBLE PRECISION, PARAMETER :: q_krit_c  = 1.000d-6 ! q-Schwellenwert sonst
  DOUBLE PRECISION, PARAMETER :: q_krit    = 1.000d-9 ! q-Schwellenwert sonst
  DOUBLE PRECISION, PARAMETER :: D_conv_sg = 200.0d-6 ! D-Schwellenwert
  DOUBLE PRECISION, PARAMETER :: D_conv_ig = 200.0d-6 ! D-Schwellenwert
  DOUBLE PRECISION, PARAMETER :: x_conv    = 0.100d-9 ! minimale Graupel-/Hagelmasse riming
  DOUBLE PRECISION, PARAMETER :: D_shed_g  = 3.000d-3 ! D-Schwellenwert fuer graupel_shedding
  DOUBLE PRECISION, PARAMETER :: D_shed_h  = 5.000d-3 ! D-Schwellenwert fuer hagel_shedding
  DOUBLE PRECISION, PARAMETER :: D_krit_c  = 10.00d-6 ! D-Schwellenwert fuer cloud_collection
  DOUBLE PRECISION, PARAMETER :: D_coll_c  = 40.00d-6 ! oberer Wert fuer cloud_coll_eff
  DOUBLE PRECISION, PARAMETER :: T_nuc     = 273.2d+0 ! Temperatur ab der Eisnukleation einsetzt
  DOUBLE PRECISION, PARAMETER :: T_freeze  = 273.2d+0 ! Temperatur ab der Gefrieren einsetzt

  DOUBLE PRECISION, PARAMETER :: q_krit_aus = 1.00d-5 ! q-Schwellenwert fuer Ausgabe von D und Z 

  !..Jason's mu-Dm-relation for snow (see also Milbrandt & Yau 2005)
  DOUBLE PRECISION, PARAMETER :: snow_cmu1 = 4.5      ! a 
  DOUBLE PRECISION, PARAMETER :: snow_cmu2 = 0.5e+3   ! b 
  DOUBLE PRECISION, PARAMETER :: snow_cmu3 = 5.0e-3   ! Dnue
  DOUBLE PRECISION, PARAMETER :: snow_cmu4 = 5.5      ! a
  INTEGER,          PARAMETER :: snow_cmu5 = 1        ! exponent

  LOGICAL, PARAMETER          :: use_mu_Dm_snow_sedi = .false.
  LOGICAL, PARAMETER          :: use_mu_Dm_rain_sedi = .true.
  LOGICAL, PARAMETER          :: use_mu_Dm_rain_evap = .true.
  LOGICAL, PARAMETER          :: use_mu_orig_rain_sedi = .false.
  LOGICAL, PARAMETER          :: use_mu_orig_rain_evap = .false.

  ! Set this switch to true if intended to use exponential DSD 
  ! for evaporation. Only effective if use_mu_Dm_rain_evap = .false. and cloud_typ > 1:
  LOGICAL :: use_exponentialdsd_rain_evap = .FALSE.


  ! ub: neue Parametrisierung des Gefrierens von Regen 
  ! (Aufteilung in Eis und Hagel anstelle nur Hagel):
  LOGICAL, PARAMETER          :: use_rain_freeze_uli = .true.

  ! ub: Schalter, ob Konversion von ice und snow nur dann zu graupel 
  ! (gesteuert durch alpha_spacefilling), 
  ! wenn die riming-rate groesser als die Depositionsrate ist 
  ! (ansonsten findet dies immer statt und wird nur durch alpha_spacefilling reguliert):
  LOGICAL, PARAMETER          :: use_ice_graupel_conv_uli = .true.

  ! ub: Schalter fuer die Umwandlung von Graupel (nicht Hagel) in Schnee
  ! beim Depositionswachstum ( bei vapor_deposition_growth() ganz einfache Param.: Rate = 3 * Dep.rate;
  ! bei vapor_dep_simple() neue Parametrisierung unter Ber. der Riming-Rate von Graupel)
  LOGICAL, PARAMETER           :: graupel_to_snow = .true.
  !.. Tuning faktoren fuer graupel to snow Umwandlung:
  !.. Proportionalitaetsfaktor zwischen Umwandlungsrate und inverser integraler Prozess-Zeitskala:
  DOUBLE PRECISION, PARAMETER  :: gr_sn_conv_tune_1 = 1.5 ! wirkt bei Erhoehung verstaerkend
  !.. Faktor, um den die depos. mind. groesser sein muss als das riming, damit 
  !   die graupel to snow Umwandlung ueberhaupt einsetzt:
  DOUBLE PRECISION, PARAMETER  :: gr_sn_conv_tune_2 = 2.0 ! wirkt bei Erhoehung abschwaechend
  DOUBLE PRECISION, PARAMETER  :: D_conv_gs = 500.0d-6    ! Max. D_s fuer Schnee erzeugt aus graupel durch gr.dep.
  DOUBLE PRECISION, PARAMETER  :: D_g_max_gs = 2000.0d-6  ! Max. D_g fuer "graupel_to_snow"


  ! ... Parameter der Partikelklassen (Geometrie, Fallgeschwindigkeit, Verteilungfunktion, ...)

  ! new Graupel and Hail categories for comparison with/without hail
  ! nu and mu were chosen by comparison of size distributions from the bin-model by HUJI
  ! with generalised gamma-distribution
  ! avel, bvel derived from fit to kurves by Heymsfield and Kajikawa
  ! for hail-classes:
  !     ageo, bgeo in order to get constant density of 910, 750 and 600 kg/m**3
  !  

  TYPE(PARTICLE), PARAMETER :: graupelmedium = PARTICLE( & ! 'graupel im standard' 
       &                                 'graupelmedium' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.166666, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.10d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.300000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 7.64d+01,& !.a_vel..Koeff. Fallgesetz
       &                                 0.255200, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet  

  TYPE(PARTICLE), PARAMETER :: graupelleicht = PARTICLE( & ! 'graupel wenn hail
       &                                 'graupelleicht' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.166666, & !.mu.....Exp.-parameter der Verteil.
 !      &                                 1.000000, & !.mu.....Exp.-parameter der Verteil.
       &                                 2.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.30d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.300000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.66d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.257800, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hail900 = PARTICLE( & ! 'hail mit 920 kg/m**3'
       &                                 'hail900' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.166666, & !.mu.....Exp.-parameter der Verteil.
 !      &                                 1.000000, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-02, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.28d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 0.618d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.2141800, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hail750 = PARTICLE( & ! 'hail mit 750 kg/m**2'
       &                                 'hail750' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
!       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 0.166666, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-02, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.36d-01,& !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 0.580d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.2141500, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hail600 = PARTICLE( & ! 'hail mit 600 kg/m**3'
       &                                 'hail600' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.166666, & !.mu.....Exp.-parameter der Verteil.
 !      &                                 1.0000000, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-02, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.47d-01,& !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 0.535d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.2142000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

!----------------------------------------------------------------------------------------
! <<hn

  TYPE(PARTICLE), PARAMETER :: graupelhailA = PARTICLE( & ! 'graupel/hail Axel' 
       &                                 'graupelhailA' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.10d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.300000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.70d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.250000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet  

  TYPE(PARTICLE), PARAMETER :: graupelhail = PARTICLE( & ! 'graupel/hail Uli'
       &                                 'graupelhail' ,& !.name...Bezeichnung
       &                                 0.666666, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 2.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.10d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.300000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.70d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.250000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelhail2 = PARTICLE( & ! 'graupel/hail Uli2'
       &                                 'graupelhail2' ,& !.name...Bezeichnung
       &                                 0.333333, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 2.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.30d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.300000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.65d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.250000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelsb2006 = PARTICLE( & ! 'graupel/hail Uli2'
       &                                 'graupelsb2006' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.90d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.323000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.40d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.230000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelhail2test = PARTICLE( & ! 'graupel/hail Uli2'
       &                                 'graupelhail2test' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-03, & !.x_max..maximale Teilchenmasse
       &                                 1.00d-09, & !.x_min..minimale Teilchenmasse 
       &                                 1.50d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.323000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.33d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.165000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelhail2test2 = PARTICLE( & ! 'graupel/hail Uli2_2'
       &                                 'graupelhail2test2' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
!       &                                 1.00d-03, & !.x_max..maximale Teilchenmasse
       &                                 5.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 1.00d-09, & !.x_min..minimale Teilchenmasse 
       &                                 1.50d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.323000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.32d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.180000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelhail2test3 = PARTICLE( & ! 'graupel/hail Uli2_3'
       &                                 'graupelhail2test3' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
!       &                                 1.00d-03, & !.x_max..maximale Teilchenmasse
       &                                 5.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 1.00d-09, & !.x_min..minimale Teilchenmasse 
       &                                 1.42d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.314000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.33d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.1870000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelhail2test4 = PARTICLE( & ! 'graupel/hail Uli2_4'
       &                                 'graupelhail2test4' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
!       &                                 1.00d-03, & !.x_max..maximale Teilchenmasse
       &                                 5.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 1.00d-09, & !.x_min..minimale Teilchenmasse 
       &                                 1.42d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.314000, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 86.89371, & !.a_vel..Koeff. Fallgesetz
       &                                 0.268325, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hailstandard = PARTICLE( & ! 'hail'
       &                                 'hailstandard' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 2.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.29d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 0.56d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.212700, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hailNOPPEL = PARTICLE( & ! 'hail'
       &                                 'hailNOPPEL' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-02, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.29d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 0.56d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.212700, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hailNOPPELlarge = PARTICLE( & ! 'hail'
       &                                 'hailNOPPELlarge' ,& !.name...Bezeichnung
       &                                 2.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-02, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.29d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 0.56d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.212700, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hailULI = PARTICLE( & ! 'hail'
       &                                 'hailULI' ,& !.name...Bezeichnung
       &                                 0.666667, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 4.00d-03, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-9, & !.x_min..minimale Teilchenmasse 
       &                                 0.127567, & !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 42.31069, & !.a_vel..Koeff. Fallgesetz
       &                                 0.166667, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: hailULItest = PARTICLE( & ! 'hail'
       &                                 'hailULItest' ,& !.name...Bezeichnung
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
!       &                                 4.00d-03, & !.x_max..maximale Teilchenmasse
       &                                 5.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-9, & !.x_min..minimale Teilchenmasse 
       &                                 0.1366 , & !.a_geo..Koeff. Geometrie
       &                                 0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                                 39.3    , & !.a_vel..Koeff. Fallgesetz
       &                                 0.166667, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: cloudfastautoconv = PARTICLE( &
       &                               'cloudfastautoconv',  & !.name...Bezeichnung der Partikelklasse
       &                               0.0,      & !.nu.....Breiteparameter der Verteil.
       &                               0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                               2.60d-10, & !.x_max..maximale Teilchenmasse D=80e-6m
       &                               4.20d-15, & !.x_min..minimale Teilchenmasse D=2.e-6m
       &                               1.24d-01, & !.a_geo..Koeff. Geometrie
       &                               0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                               3.75d+05, & !.a_vel..Koeff. Fallgesetz
       &                               0.666667, & !.b_vel..Koeff. Fallgesetz
       &                               0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                               0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                               2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: cloudstandard = PARTICLE( &! HN: made variable for seeding experiments
       &                               'cloudsstandard',  & !.name...Bezeichnung der Partikelklasse
       &                               0.333333, & !.nu.....Breiteparameter der Verteil.
       &                               0.666666, & !.mu.....Exp.-parameter der Verteil.
       &                               2.60d-10, & !.x_max..maximale Teilchenmasse D=80e-6m
       &                               4.20d-15, & !.x_min..minimale Teilchenmasse D=2.e-6m
       &                               1.24d-01, & !.a_geo..Koeff. Geometrie
       &                               0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                               3.75d+05, & !.a_vel..Koeff. Fallgesetz
       &                               0.666667, & !.b_vel..Koeff. Fallgesetz
       &                               0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                               0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                               2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: cloud_nue1mue1 = PARTICLE( &
       &                               'cloud_nue1mue1',  & !.name...Bezeichnung der Partikelklasse
       &                               1.000000, & !.nu.....Breiteparameter der Verteil.
       &                               1.000000, & !.mu.....Exp.-parameter der Verteil.
       &                               2.60d-10, & !.x_max..maximale Teilchenmasse D=80e-6m
       &                               4.20d-15, & !.x_min..minimale Teilchenmasse D=2.e-6m
       &                               1.24d-01, & !.a_geo..Koeff. Geometrie
       &                               0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                               3.75d+05, & !.a_vel..Koeff. Fallgesetz
       &                               0.666667, & !.b_vel..Koeff. Fallgesetz
       &                               0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                               0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                               2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: cloudveryslowautoconv = PARTICLE( &
       &                               'cloudveryslowautoconv',  & !.name...Bezeichnung der Partikelklasse
       &                               6.000000,      & !.nu.....Breiteparameter der Verteil.
       &                               1.000000, & !.mu.....Exp.-parameter der Verteil.
       &                               2.60d-10, & !.x_max..maximale Teilchenmasse D=80e-6m
       &                               4.20d-15, & !.x_min..minimale Teilchenmasse D=2.e-6m
       &                               1.24d-01, & !.a_geo..Koeff. Geometrie
       &                               0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                               3.75d+05, & !.a_vel..Koeff. Fallgesetz
       &                               0.666667, & !.b_vel..Koeff. Fallgesetz
       &                               0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                               0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                               2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: cloudslowautoconv = PARTICLE( &
       &                               'cloudslowautoconv',  & !.name...Bezeichnung der Partikelklasse
       &                               0.333333,      & !.nu.....Breiteparameter der Verteil.
       &                               1.000000, & !.mu.....Exp.-parameter der Verteil.
       &                               2.60d-10, & !.x_max..maximale Teilchenmasse D=80e-6m
       &                               4.20d-15, & !.x_min..minimale Teilchenmasse D=2.e-6m
       &                               1.24d-01, & !.a_geo..Koeff. Geometrie
       &                               0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                               3.75d+05, & !.a_vel..Koeff. Fallgesetz
       &                               0.666667, & !.b_vel..Koeff. Fallgesetz
       &                               0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                               0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                               2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: cloudheike = PARTICLE( & ! nu und mu aus Vergl. mit bin-Modell
       &                               'cloudheike',  & !.name...Bezeichnung der Partikelklasse
       &                               6.000000, & !.nu.....Breiteparameter der Verteil.
       &                               0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                               2.60d-10, & !.x_max..maximale Teilchenmasse D=80e-6m
       &                               4.20d-15, & !.x_min..minimale Teilchenmasse D=2.e-6m
       &                               1.24d-01, & !.a_geo..Koeff. Geometrie
       &                               0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                               3.75d+05, & !.a_vel..Koeff. Fallgesetz
       &                               0.666667, & !.b_vel..Koeff. Fallgesetz
       &                               0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                               0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                               2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: iceCRY =  PARTICLE( & ! nach Andy Heymsfield (CRYSTAL-FACE)
       &                              'iceCRY', & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              5.00d-07, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-12, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              3.303633, & !.a_geo..Koeff. Geometrie
       &                              0.476191, & !.b_geo..Koeff. Geometrie = 1/2.1
       &                              2.77d+01, & !.a_vel..Koeff. Fallgesetz 
       &                              0.215790, & !.b_vel..Koeff. Fallgesetz = 0.41/1.9
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: iceCRY2test =  PARTICLE( & ! nach Andy Heymsfield (CRYSTAL-FACE)
       &                              'iceCRY2test', & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu...e..Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
! AS_2010:       &                              1.00d-07, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-06, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-12, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              0.835000, & !.a_geo..Koeff. Geometrie
       &                              0.390000, & !.b_geo..Koeff. Geometrie = 1/2.1
       &                              2.77d+01, & !.a_vel..Koeff. Fallgesetz 
       &                              0.215790, & !.b_vel..Koeff. Fallgesetz = 0.41/1.9
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

! AS_2010 >>
  TYPE(PARTICLE), PARAMETER :: iceCirrus =  PARTICLE( & ! single bullets nach Heymsfield and Iaquinta
       &                              'iceCirrus', & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu...e..Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              1.00d-06, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-12, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              32.68941, & !.a_geo..Koeff. Geometrie, m(D)=7.08e-4*D**2.08
       &                              0.480769, & !.b_geo..Koeff. Geometrie = 1/2.08
       &                              3.447d+3, & !.a_vel..Koeff. Fallgesetz v(D)=3577*D**1.31
       &                              0.629808, & !.b_vel..Koeff. Fallgesetz = 1.31/2.08
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet
! AS_2010 <<

  TYPE(PARTICLE), PARAMETER :: snowCRYSTAL =  PARTICLE( & ! nach Andy Heymsfield (CRYSTAL-FACE)
       &                              'snowCRYSTAL', & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              1.00d-07, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-12, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              3.303633, & !.a_geo..Koeff. Geometrie
       &                              0.476191, & !.b_geo..Koeff. Geometrie = 1/2.1
       &                              2.47d+02, & !.a_vel..Koeff. Fallgesetz 
       &                              0.333333, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: snowCRYSTALuli =  PARTICLE( & ! nach Andy Heymsfield (CRYSTAL-FACE)
       &                              'snowCRYSTALuli', & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.500000, & !.mu.....Exp.-parameter der Verteil.
       &                              2.00d-05, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-10, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              2.400000, & !.a_geo..Koeff. Geometrie
       &                              0.455000, & !.b_geo..Koeff. Geometrie = 1/2.1
       &                              4.200000, & !.a_vel..Koeff. Fallgesetz 
       &                              0.092000, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: snowCRYSTALuli2 =  PARTICLE( & ! nach Andy Heymsfield (CRYSTAL-FACE)
       &                              'snowCRYSTALuli2', & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.500000, & !.mu.....Exp.-parameter der Verteil.
       &                              2.00d-05, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-10, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              2.400000, & !.a_geo..Koeff. Geometrie
       &                              0.455000, & !.b_geo..Koeff. Geometrie = 1/2.1
       &                              8.800000, & !.a_vel..Koeff. Fallgesetz 
       &                              0.150000, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: snowULI =  PARTICLE( & ! nach Andy Heymsfield (CRYSTAL-FACE)
       &                              'snowULI', & !.name...Bezeichnung der Partikelklasse
       &                              0.333333, & !.nu.....Breiteparameter der Verteil.
       &                              0.5     , & !.mu.....Exp.-parameter der Verteil.
       &                              2.00d-06, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.00d-12, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              3.303633, & !.a_geo..Koeff. Geometrie
       &                              0.476191, & !.b_geo..Koeff. Geometrie = 1/2.1
       &                              3.9259,   & !.a_vel..Koeff. Fallgesetz 
       &                              0.08,     & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: iceHK87 = PARTICLE( & ! HK87 'hex plates' + 700 D**1.2
       &                             'iceHK87',    & !.name...Bezeichnung der Partikelklasse
       &                            -0.333333, & !.nu.....Breiteparameter der Verteil.
       &                             0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                             1.00d-07, & !.x_max..maximale Teilchenmasse D=13.6e-3m
       &                             1.00d-12, & !.x_min..minimale Teilchenmasse D=60.0e-6m
       &                             2.17d-01, & !.a_geo..Koeff. Geometrie
       &                             0.302115, & !.b_geo..Koeff. Geometrie = 1/3.31
       &                             3.17d+02, & !.a_vel..Koeff. Fallgesetz
       &                             0.362500, & !.b_vel..Koeff. Fallgesetz = 1.2/3.31
       &                             0.860000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                             0.280000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                             3.141593)   !.cap....Koeff. Kapazitaet = pi

  TYPE(PARTICLE), PARAMETER :: iceHK = PARTICLE( & ! HK87 'hex plates'
       &                             'iceHK',    & !.name...Bezeichnung der Partikelklasse
       &                            -0.333333,      & !.nu.....Breiteparameter der Verteil.
       &                             0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                             1.00d-07, & !.x_max..maximale Teilchenmasse D=1.e-3m
       &                             1.00d-12, & !.x_min..minimale Teilchenmasse D=60e-6m
       &                             2.17d-01, & !.a_geo..Koeff. Geometrie
       &                             0.302115, & !.b_geo..Koeff. Geometrie = 1/3.31
       &                             4.19d+01, & !.a_vel..Koeff. Fallgesetz
       &                             0.260000, & !.b_vel..Koeff. Fallgesetz
       &                             0.860000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                             0.280000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                             3.141593)   !.cap....Koeff. Kapazitaet = pi

  TYPE(PARTICLE), PARAMETER :: rainREISNER = PARTICLE( & ! Reisner98
       &                              'rainREISNER',   & !.name...Bezeichnung der Partikelklasse
       &                             -0.666666, & !.nu.....Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              3.00d-06, & !.x_max..maximale Teilchenmasse
       &                              2.60d-10, & !.x_min..minimale Teilchenmasse D=80.e-6m
       &                              1.24d-01, & !.a_geo..Koeff. Geometrie
       &                              0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                              1.59d+02, & !.a_vel..Koeff. Fallgesetz
       &                              0.266667, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: rainAS = PARTICLE( & ! 
       &                              'rainAS',   & !.name...Bezeichnung der Partikelklasse
       &                              1.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              3.00d-06, & !.x_max..maximale Teilchenmasse
       &                              2.60d-10, & !.x_min..minimale Teilchenmasse D=80.e-6m
       &                              1.24d-01, & !.a_geo..Koeff. Geometrie
       &                              0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                              1.59d+02, & !.a_vel..Koeff. Fallgesetz
       &                              0.266667, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: rainsb2006 = PARTICLE( & ! 
       &                              'rainsb2006',   & !.name...Bezeichnung der Partikelklasse
       &                             -0.666666, & !.nu.....Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              3.00d-06, & !.x_max..maximale Teilchenmasse
       &                              2.60d-10, & !.x_min..minimale Teilchenmasse D=80.e-6m
       &                              1.24d-01, & !.a_geo..Koeff. Geometrie
       &                              0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                              1.59d+02, & !.a_vel..Koeff. Fallgesetz
       &                              0.266667, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: rainULI = PARTICLE( & ! Blahak, v=v(x) gefittet 6.9.2005
       &                              'rainULI',   & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              3.00d-06, & !.x_max..maximale Teilchenmasse
       &                              2.60d-10, & !.x_min..minimale Teilchenmasse D=80.e-6m
       &                              1.24d-01, & !.a_geo..Koeff. Geometrie
       &                              0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                              114.0137, & !.a_vel..Koeff. Fallgesetz
       &                              0.234370, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: rainULI2 = PARTICLE( & ! Blahak, v=v(x) gefittet 6.9.2005
       &                              'rainULI2',   & !.name...Bezeichnung der Partikelklasse
       &                              1.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                              3.00d-06, & !.x_max..maximale Teilchenmasse
       &                              2.60d-10, & !.x_min..minimale Teilchenmasse D=80.e-6m
       &                              1.24d-01, & !.a_geo..Koeff. Geometrie
       &                              0.333333, & !.b_geo..Koeff. Geometrie = 1/3
       &                              114.0137, & !.a_vel..Koeff. Fallgesetz
       &                              0.234370, & !.b_vel..Koeff. Fallgesetz
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelHKVT = PARTICLE( & ! HK87 'lump graupel' mit mod. VT 
       &                                 'graupelHKVT',& !.name...Bezeichnung der Partikelklasse
       &                                 1.0     , & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.50d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.90d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.322580, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 0.40d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.230000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet


  TYPE(PARTICLE), PARAMETER :: graupelLH = PARTICLE( & ! LH74 'lump graupel 3' 
       &                                 'graupelLH',& !.name...Bezeichnung der Partikelklasse
       &                                 1.0     ,      & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 3.46d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.370370, & !.b_geo..Koeff. Geometrie = 1/2.7
       &                                 9.45d+00, & !.a_vel..Koeff. Fallgesetz
       &                                 0.120000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelHK = PARTICLE( & ! HK87 'cold lump graupel' 
       &                                 'graupelHK',& !.name...Bezeichnung der Partikelklasse
       &                                 1.0,      & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.77d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.311527, & !.b_geo..Koeff. Geometrie = 1/3.21
       &                                 4.64d+01, & !.a_vel..Koeff. Fallgesetz
       &                                 0.260000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelHKULI = PARTICLE( & ! HK87 'cold lump graupel' 
       &                                 'graupelHKULI',& !.name...Bezeichnung der Partikelklasse
       &                                 1.0,      & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.77d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.311527, & !.b_geo..Koeff. Geometrie = 1/3.21
       &                                 79.41,    & !.a_vel..Koeff. Fallgesetz
       &                                 0.27726,  & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: graupelHK87 = PARTICLE( & ! HK87 'lump graupel' 
       &                                 'graupelHK87',& !.name...Bezeichnung der Partikelklasse
       &                                 1.000000, & !.nu.....Breiteparameter der Verteil.
       &                                 0.333333, & !.mu.....Exp.-parameter der Verteil.
       &                                 1.00d-04, & !.x_max..maximale Teilchenmasse
       &                                 2.60d-10, & !.x_min..minimale Teilchenmasse 
       &                                 1.90d-01, & !.a_geo..Koeff. Geometrie
       &                                 0.322580, & !.b_geo..Koeff. Geometrie = 1/3.10
       &                                 1.01d+02, & !.a_vel..Koeff. Fallgesetz
       &                                 0.290000, & !.b_vel..Koeff. Fallgesetz
       &                                 0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                                 0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                                 2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: snowLHm = PARTICLE( & ! nach LHmod
       &                              'snowLHm', & !.name...Bezeichnung der Partikelklasse
       &                              0.000000, & !.nu.....Breiteparameter der Verteil.
       &                              0.5     , & !.mu.....Exp.-parameter der Verteil.
       &                              2.00d-06, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.73d-09, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              8.16d+00, & !.a_geo..Koeff. Geometrie
       &                              0.526316, & !.b_geo..Koeff. Geometrie = 1/1.9
       &                              1.50d+01, & !.a_vel..Koeff. Fallgesetz 
       &                              0.215790, & !.b_vel..Koeff. Fallgesetz = 0.41/1.9
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: snowLH = PARTICLE( & ! nach Locatelli und Hobbs
       &                              'snowLH', & !.name...Bezeichnung der Partikelklasse
       &                              0.500000, & !.nu.....Breiteparameter der Verteil.
       &                              0.5     , & !.mu.....Exp.-parameter der Verteil.
       &                              2.00d-06, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.73d-09, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              8.16d+00, & !.a_geo..Koeff. Geometrie
       &                              0.526316, & !.b_geo..Koeff. Geometrie = 1/1.9
       &                              2.77d+01, & !.a_vel..Koeff. Fallgesetz 
       &                              0.215790, & !.b_vel..Koeff. Fallgesetz = 0.41/1.9
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: snowLHULI = PARTICLE( & ! nach Locatelli und Hobbs
       &                              'snowLHULI', & !.name...Bezeichnung der Partikelklasse
       &                              0.5     , & !.nu.....Breiteparameter der Verteil.
       &                              0.5     , & !.mu.....Exp.-parameter der Verteil.
       &                              2.00d-06, & !.x_max..maximale Teilchenmasse D=???e-2m
       &                              1.73d-09, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                              8.16d+00, & !.a_geo..Koeff. Geometrie
       &                              0.526316, & !.b_geo..Koeff. Geometrie = 1/1.9
       &                              10.33,    & !.a_vel..Koeff. Fallgesetz 
       &                              0.15,     & !.b_vel..Koeff. Fallgesetz = 0.41/1.9
       &                              0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                              0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                              2.0)        !.cap....Koeff. Kapazitaet

  TYPE(PARTICLE), PARAMETER :: snowC = PARTICLE( & ! nach Cotton et. al (1986)
       &                               'snowC',& !.name...Bezeichnung der Partikelklasse
       &                               0.333333, & !.nu.....Breiteparameter der Verteil.
       &                               0.5     , & !.mu.....Exp.-parameter der Verteil.
       &                               5.00d-06, & !.x_max..maximale Teilchenmasse D=1.0e-2m
       &                               8.30d-11, & !.x_min..minimale Teilchenmasse D=200e-6m
       &                               1.33398 , & !.a_geo..Koeff. Geometrie
       &                               0.416667, & !.b_geo..Koeff. Geometrie = 1/2.4
       &                               2.876292, & !.a_vel..Koeff. Fallgesetz 
       &                               0.083333, & !.b_vel..Koeff. Fallgesetz = 0.2/2.4
       &                               0.780000, & !.a_ven..Koeff. Ventilation (PK, S.541)
       &                               0.308000, & !.b_ven..Koeff. Ventilation (PK, S.541)
       &                               2.0)        !.cap....Koeff. Kapazitaet


  TYPE(PARTICLE)                       :: cloud, rain, ice, snow, graupel, hail

  LOGICAL, PARAMETER                   :: isdebug   = .false.
  LOGICAL, PARAMETER                   :: ice_multiplication = .TRUE.


  LOGICAL, PARAMETER                   :: enhanced_melting   = .TRUE.

  ! Saturation adjustment within subroutine clouds() (.false.) or external after clouds() (.true.)
  ! (The latter requires extra care by the calling program!!!)
  LOGICAL, PARAMETER                   :: satad_nach_mikrophysik = .true.

  INTEGER                              :: ice_typ,nuc_i_typ,nuc_c_typ,cloud_typ,mu_Dm_rain_typ

  DOUBLE PRECISION, ALLOCATABLE, DIMENSION (:,:,:) :: rrho_04
  DOUBLE PRECISION, ALLOCATABLE, DIMENSION (:,:,:) :: rrho_c

  DOUBLE PRECISION :: rain_cmu0,rain_cmu1,rain_cmu2,rain_cmu3,rain_cmu4,rain_gfak
  INTEGER          :: rain_cmu5

  !====================================================
  ! Variables for namelist parameters:
  !====================================================

  DOUBLE PRECISION :: qnc_const
  DOUBLE PRECISION :: tshift_cloudfreeze
  DOUBLE PRECISION :: z_ncn_segal
  DOUBLE PRECISION :: z1oe_ncn_segal

  ! Graupel-Shedding: Das angefrorene Wasser wird bei T > T_shed, 
  ! ggf. nach enhanced-melting, wieder abgeworfen und zu Regen.
  ! Default is .false. for both parameter (see src_gscp)
  LOGICAL          :: graupel_shedding
  LOGICAL          :: hail_shedding

CONTAINS

  ! This subroutine has to be called once at the start of the model run by
  ! the main program. It properly sets the parameters for the different hydrometeor
  ! classes according to predefined parameter sets (see above).

  SUBROUTINE init_seifert( cloud_typ )
    IMPLICIT NONE
    INTEGER, INTENT(in) :: cloud_typ
    integer :: wolke_typ

    wolke_typ = MOD(cloud_typ/10,100)

    IF (cloud_typ < 2000) THEN
      ! Without hail class:
!      cloud = cloud_nue1mue1
!      cloud = cloudheike
      cloud = cloudfastautoconv
      rain  = rainULI
      ice   = iceCRY2test
      snow  = snowCRYSTALuli2
      graupel = graupelhail2test2
      ! Dummy value for hail:
      hail    = hailNOPPEL
    ELSE
      ! Including hail class:
!      cloud = cloud_nue1mue1
!      cloud = cloudheike
      cloud = cloudfastautoconv
      rain  = rainULI
      ice   = iceCRY2test
      snow  = snowCRYSTALuli2
!      graupel = graupelhail2test3
      graupel = graupelhail2test4
      hail    = hailULItest
    END IF

    !mu_Dm_rain_typ = mod(cloud_typ,10)
    mu_Dm_rain_typ = 1

    if (mu_Dm_rain_typ.eq.0) then
      !..constant mue value
      rain_cmu0 = 0.0
      rain_cmu1 = 0.0
      rain_cmu2 = 1.0
      rain_cmu3 = 1.0
      rain_cmu4 = (rain%nu+1.0d0)/rain%b_geo - 1.0d0 ! <-- this is the (constant) mue value 
      rain_cmu5 = 1
      rain_gfak = -1.0  ! In this case gamma = 1 in rain_evaporation    
    elseif (mu_Dm_rain_typ.eq.1) then
      !..Axel's mu-Dm-relation for raindrops based on 1d-bin model
      rain_cmu0 = 6.0             ! Axel's 2007 relation 
      rain_cmu1 = 30.0            ! 
      rain_cmu2 = 1.00d+3         ! 
      rain_cmu3 = 1.10d-3         ! D_eq,break
      rain_cmu4 = 1.0             ! 
      rain_cmu5 = 2               ! exponent
      rain_gfak = 1.0      
    elseif (mu_Dm_rain_typ.eq.2) then
      !..Modifikation of mu-Dm-relation for experiments with increased evaporation
      rain_cmu0 = 11.0            ! instead of 6.0      
      rain_cmu1 = 30.0            ! 
      rain_cmu2 = 1.00d+3         ! 
      rain_cmu3 = 1.10d-3         ! 
      rain_cmu4 = 4.0             ! instead of 1.0  
      rain_cmu5 = 2               ! 
      rain_gfak = 0.5             ! instead of 1.0      
    elseif (mu_Dm_rain_typ.eq.3) then
      !..Jason Milbrandts mu-Dm-relation'
      rain_cmu0 = 19.0           !
      rain_cmu1 = 19.0           ! Jason Milbrandt's mu-Dm-relation for rain
      rain_cmu2 = 0.60d+3        ! (Milbrandt&Yau 2005, JAS, Table 1)
      rain_cmu3 = 1.80d-3        !
      rain_cmu4 = 17.0           !
      rain_cmu5 = 1              !
      rain_gfak = -1.0  ! In this case gamma = 1 in rain_evaporation    
    endif
      
!<AS20080304

  END SUBROUTINE init_seifert

  DOUBLE PRECISION FUNCTION power_explog_dd(a,b)
    IMPLICIT NONE
    DOUBLE PRECISION, INTENT(in)  :: a,b
    power_explog_dd = exp(b*log(a))
  END FUNCTION power_explog_dd

  DOUBLE PRECISION FUNCTION power_explog_dr(a,b)
    IMPLICIT NONE
    DOUBLE PRECISION, INTENT(in)  :: a
    REAL, INTENT(in)  :: b
    power_explog_dr = exp(b*log(a))
  END FUNCTION power_explog_dr

  DOUBLE PRECISION FUNCTION vent_coeff_a(parti,n)
    IMPLICIT NONE

    INTEGER        :: n
    TYPE(PARTICLE) :: parti

    vent_coeff_a = parti%a_ven * gfct((parti%nu+n+parti%b_geo)/parti%mu)              &
         &                  / gfct((parti%nu+1.0)/parti%mu)                        & 
         &                * ( gfct((parti%nu+1.0)/parti%mu)                        & 
         &                  / gfct((parti%nu+2.0)/parti%mu) )**(parti%b_geo+n-1.0) 

  END FUNCTION vent_coeff_a

  DOUBLE PRECISION FUNCTION vent_coeff_b(parti,n)
    IMPLICIT NONE

    INTEGER        :: n
    TYPE(PARTICLE) :: parti

    DOUBLE PRECISION, PARAMETER :: m_f   = 0.500   ! Koeff. Ventilationskoeff. (PK, S.541)

    vent_coeff_b = parti%b_ven                                                  & 
         & * gfct((parti%nu+n+(m_f+1.0)*parti%b_geo+m_f*parti%b_vel)/parti%mu)  &
         &             / gfct((parti%nu+1.0)/parti%mu)                          & 
         &           * ( gfct((parti%nu+1.0)/parti%mu)                          &
         &             / gfct((parti%nu+2.0)/parti%mu)                          &
         &             )**((m_f+1.0)*parti%b_geo+m_f*parti%b_vel+n-1.0)

  END FUNCTION vent_coeff_b

  DOUBLE PRECISION FUNCTION moment_gamma(p,n) 
    IMPLICIT NONE

    INTEGER          :: n
    TYPE(PARTICLE)   :: p

    moment_gamma  = gfct((n+p%nu+1.0)/p%mu) / gfct((p%nu+1.0)/p%mu)        &
         &     * ( gfct((  p%nu+1.0)/p%mu) / gfct((p%nu+2.0)/p%mu) )**n
  END FUNCTION moment_gamma

  DOUBLE PRECISION FUNCTION lambda_gamma(p,x) 
    IMPLICIT NONE

    DOUBLE PRECISION :: x
    TYPE(PARTICLE)   :: p

    lambda_gamma  = ( gfct((p%nu+1.0)/p%mu) / gfct((p%nu+2.0)/p%mu) * x)**(-p%mu)
  END FUNCTION lambda_gamma

  DOUBLE PRECISION FUNCTION e_es (t_)
    !*******************************************************************************
    !                        Saettigungsdampfdruck ueber Eis                       *
    !*******************************************************************************

    DOUBLE PRECISION, INTENT (IN) :: t_

    e_es  = e_3 * EXP (A_e * (t_ - T_3) / (t_ - B_e))

  END FUNCTION e_es

  DOUBLE PRECISION FUNCTION e_ws (t_)
    !*******************************************************************************
    !                      Saettigungsdampfdruck ueber Wasser                      *
    !*******************************************************************************

    DOUBLE PRECISION, INTENT (IN) :: t_

    e_ws  = e_3 * EXP (A_w * (t_ - T_3) / (t_ - B_w))

  END FUNCTION e_ws

  FUNCTION e_ws_vec (ta,idim,jdim)

    INTEGER :: idim, jdim
    DOUBLE PRECISION               :: e_ws_vec(idim,jdim)
    DOUBLE PRECISION, INTENT (IN)  :: ta(idim,jdim)

!CDIR COLLAPSE
    e_ws_vec  = e_3 * EXP (A_w * (ta - T_3) / (ta - B_w))

  END FUNCTION e_ws_vec

  FUNCTION e_es_vec (ta,idim,jdim)

    INTEGER :: idim, jdim
    DOUBLE PRECISION               :: e_es_vec(idim,jdim)
    DOUBLE PRECISION, INTENT (IN)  :: ta(idim,jdim)

!CDIR COLLAPSE
    e_es_vec  = e_3 * EXP (A_e * (ta - T_3) / (ta - B_e))

  END FUNCTION e_es_vec

! UB_20100701>>
  DOUBLE PRECISION FUNCTION dyn_visc_air(T)
    !*******************************************************************************
    !                                                                              *
    !       Dynamic viscosity eta of dry air as function of temperature T          *
    !       (linear Fit to the Sutherland-formula by myself)                        *
    !                                                                              *
    !       Valid for T in [203.16; 313.16] K and 0.1 bar < p < 10 bar             *
    !                                                                              *
    !       Input T is in Kelvin                                                   *
    !       Output eta is in Pa s                                                  *
    !                                                                              *
    !*******************************************************************************

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: T

    dyn_visc_air = 2.92332d-6 + 5.20712d-8 * T

    RETURN
  END FUNCTION dyn_visc_air
! UB_20100701<<

  DOUBLE PRECISION FUNCTION gfct(x)
    !*******************************************************************************
    !                                                                              *
    !       Gammafunktion aus Numerical Recipes (F77)                              *
    !                                                                              *
    !*******************************************************************************
    IMPLICIT NONE

    DOUBLE PRECISION cof(6)
    DOUBLE PRECISION stp,half,one,x,xx,fpf,tmp,ser,gamma
    INTEGER j

    DATA cof,stp/76.18009173d0,-86.50532033d0,24.01409822d0,  &
         &     -1.231739516d0,.120858003d-2,-.536382d-5,2.50662827465d0/
    DATA half,one,fpf/0.5d0,1.0d0,5.5d0/

    xx  = x  - one
    tmp = xx + fpf
    tmp = (xx + half) * LOG(tmp) - tmp
    ser = one
    DO j = 1,6
      xx  = xx  + one
      ser = ser + cof(j) / xx
    ENDDO
    gamma = tmp + LOG(stp*ser)
    gamma = EXP(gamma)

    gfct = gamma
    RETURN
  END FUNCTION gfct

END MODULE wolken_konstanten
