!!$ changed number of auxiliary points for Simpson-rule from 250 to 100
!!   (n_stuetz, 20 occurences)

!------------------------------------------------------------------------------------------!
!
! Interface functions for the LM to the Karlsruhe libraries for Radar Reflectivity 
! Calculations based on Mie-Theory
!
! by Ulrich Blahak and Axel Seifert
!
! (Calculates radar reflectivity of precipitation particles using 
! full mie theorie which includes scattering by wet and melting particles)
!
! Applicable to microwave radiation, wavelength > 1 mm
!
! Contact: ulrich.blahak@imk.fzk.de
!
! Schwachpunkte: Schmelzgradparametrisierung, feste Durchmessergrenzen bei der
!                Integration des Reflektivitaetsintervalls
!
! Die Schnee-Verteilung von Paul Field ist jetzt implementiert!
! Neue Funktion(en):
!   - hydroparams_lm
!   - D_average_1M_exp_vec_n0s
!   - D_of_X_average_1M_exp_vec_n0s
!   - zradar_sn_rayl_oguchi_vec
!   - zradar_wetsn_rayl_oguchi_vec
!
! 3D-Feld mu_rain(:,:,:) fuer Gamma-Shape-Parameter mue Regen eingefuehrt anstelle
! eines festen Wertes (0.0) --- gilt nur fuer itype_gsyp < 100!
! (Allerdings fehlt der 2-Momenten-Regen itype_gscp == 5 noch!)
!
! Vektorisierte Fassung!!! (SX8/9 ...)
!
! **** Version MIT ilow, iup, jlow, jup, klow, kup, get_loc_domain() !!! ****
!
! 
! ACTIVATE ROUTINES FOR SEIFERT-BEHENG 2-MOMENT MICROPHYSICS BY COMPILING WITH
! 
!   -DTWOMOMSEIFERT 
!
!------------------------------------------------------------------------------------------!


! Modules used:

MODULE radar_mie_lm_vec

  USE data_modelconfig, ONLY :   &

       dlon,         & ! grid point distance in zonal direction (in degrees)
       dlat,         & ! grid point distance in meridional direction (in degrees)
       startlat_tot, & ! transformed latitude of the lower left grid point
                       ! of the total domain (in degrees, N>0)
       ie,           & ! number of grid points in zonal direction
       je,           & ! number of grid points in meridional direction
       ie_tot,       & ! total number of grid points in zonal direction
       je_tot,       & ! total number of grid points in meridional direction
       ke,           & ! number of grid points in vertical direction
       istartpar,    & ! start index for computations in the parallel program
       iendpar,      & ! end index for computations in the parallel program
       jstartpar,    & ! start index for computations in the parallel program
       jendpar         ! end index for computations in the parallel program


  USE data_constants, ONLY :   &
       r_earth,     &  ! mean radius of the earth
       rho_w,       &  ! density of liquid water
       rho_ice         ! density of ice          (kg/m^3)

  USE data_parameters , ONLY :   &
       ireals,     & ! kind-type parameter for "normal" integer variables
       iintegers     ! KIND-type parameters for real variables

  USE radar_mielib_vec

  USE data_fields     , ONLY :   &
       w          ,    & ! vertical wind speed (defined on half levels)  ( m/s )
       t          ,    & ! temperature                                   (  k  )
       qv         ,    & ! specific water vapor content                  (kg/kg)
       qc         ,    & ! specific cloud water content                  (kg/kg)
       qi         ,    & ! specific cloud ice   content                  (kg/kg)
       qr         ,    & ! specific rain content                         (kg/kg)
       qs         ,    & ! specific snow content                         (kg/kg)
       qg         ,    & ! specific graupel content                      (kg/kg)
       rho               ! density of moist air                          (kg/m3)

#ifdef TWOMOMSEIFERT
  USE data_fields     , ONLY :   &
       qh         ,    & ! specific graupel content                      (kg/kg)
       qnc        ,    & ! number concentration cloud droplets           ( 1/kg)
       qni        ,    & ! number concentration cloud ice                ( 1/kg)
       qnr        ,    & ! number concentration raindrops                ( 1/kg)
       qns        ,    & ! number concentration snow                     ( 1/kg)
       qng        ,    & ! number concentration graupel                  ( 1/kg)
       qnh               ! number concentration graupel                  ( 1/kg)

  USE wolken_konstanten, ONLY:   &
       particle, rain, cloud, snow, ice, graupel, hail
#endif

  USE gamma_functions_vec, ONLY:   &
       gfct, gfctvec

  USE parallel_utilities, ONLY:  distribute_field, gather_field

  USE data_parallel,      ONLY :  &
    nproc,           & ! total number of processors: nprocx * nprocy + nprocio
    num_compute,     & ! number of compute PEs: nprocx * nprocy
    my_cart_id         ! rank of this subdomain in the cartesian communicator


!===============================================================================

  IMPLICIT NONE

!===============================================================================

! Local scalars:
!---------------

  INTEGER, PARAMETER, PRIVATE :: kinddp = KIND(1.0d0)
  DOUBLE PRECISION, PARAMETER, PRIVATE :: pi = 3.14159265358979d0
  COMPLEX(kind=kinddp), PARAMETER, PRIVATE :: m_air = (1.0d0,0.0d0)

  ! .. Local domain bounds for computations (are set in the interface routines
  !    to the COSMO-model, e.g., radar_sb_mie_vec(), ...):
  INTEGER, PRIVATE :: ilow, iup, jlow, jup, klow, kup

  DOUBLE PRECISION, ALLOCATABLE, PRIVATE :: Tmax_i(:,:), Tmax_s(:,:), Tmax_g(:,:), Tmax_h(:,:)
       
#ifndef TWOMOMSEIFERT
  !.. only dummies to enable compilation:

  TYPE PARTICLE
    CHARACTER(20)    :: name  !..Bezeichnung der Partikelklasse
    DOUBLE PRECISION :: nu    !..Breiteparameter der Verteil.
    DOUBLE PRECISION :: mu    !..Exp.-parameter der Verteil.
    DOUBLE PRECISION :: x_max !..maximale Teilchenmasse
    DOUBLE PRECISION :: x_min !..minimale Teilchenmasse
    DOUBLE PRECISION :: a_geo !..Koeff. Geometrie
    DOUBLE PRECISION :: b_geo !..Koeff. Geometrie = 1/3
    DOUBLE PRECISION :: a_vel !..Koeff. Fallgesetz
    DOUBLE PRECISION :: b_vel !..Koeff. Fallgesetz
    DOUBLE PRECISION :: a_ven !..Koeff. Ventilationsparam.
    DOUBLE PRECISION :: b_ven !..Koeff. Ventilationsparam.
    DOUBLE PRECISION :: cap   !..Koeff. Kapazitaet
  END TYPE PARTICLE

  TYPE(particle) :: cloud, rain, ice, snow, graupel, hail

  DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:,:,:), PRIVATE :: &
       qh         ,    & ! specific graupel content                      (kg/kg)
       qnc        ,    & ! number concentration cloud droplets           ( 1/kg)
       qni        ,    & ! number concentration cloud ice                ( 1/kg)
       qnr        ,    & ! number concentration raindrops                ( 1/kg)
       qns        ,    & ! number concentration snow                     ( 1/kg)
       qng        ,    & ! number concentration graupel                  ( 1/kg)
       qnh               ! number concentration graupel                  ( 1/kg)
#endif

  ! Type declaration for a general 2D equidistant lookup table:
  TYPE lookupt_2D
    INTEGER :: n1  ! number of grid points in x1-direction
    INTEGER :: n2  ! number of grid points in x2-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x1  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x2  ! grid vector in x1-direction
    DOUBLE PRECISION                     :: dx1          ! dx1   (grid distance w.r.t. x1)
    DOUBLE PRECISION                     :: dx2          ! dx2   (grid distance w.r.t. x2)
    DOUBLE PRECISION                     :: odx1         ! one over dx 1
    DOUBLE PRECISION                     :: odx2         ! one over dx 2
    DOUBLE PRECISION, DIMENSION(:,:,:,:), POINTER :: ltab
  END TYPE lookupt_2D
  
  ! Type declaration for a general 3D equidistant lookup table:
  TYPE lookupt_3D
    INTEGER :: n1  ! number of grid points in x1-direction
    INTEGER :: n2  ! number of grid points in x2-direction
    INTEGER :: n3  ! number of grid points in x3-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x1  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x2  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x3  ! grid vector in x1-direction
    DOUBLE PRECISION                     :: dx1          ! dx1   (grid distance w.r.t. x1)
    DOUBLE PRECISION                     :: dx2          ! dx2   (grid distance w.r.t. x2)
    DOUBLE PRECISION                     :: dx3          ! dx3   (grid distance w.r.t. x3)
    DOUBLE PRECISION                     :: odx1         ! one over dx 1
    DOUBLE PRECISION                     :: odx2         ! one over dx 2
    DOUBLE PRECISION                     :: odx3         ! one over dx 3
    DOUBLE PRECISION, DIMENSION(:,:,:,:), POINTER :: ltab
  END TYPE lookupt_3D
  
  ! Type declaration for a general 4D equidistant lookup table:
  TYPE lookupt_4D
    INTEGER :: n1  ! number of grid points in x1-direction
    INTEGER :: n2  ! number of grid points in x2-direction
    INTEGER :: n3  ! number of grid points in x3-direction
    INTEGER :: n4  ! number of grid points in x4-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x1  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x2  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x3  ! grid vector in x1-direction
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x4  ! grid vector in x1-direction
    DOUBLE PRECISION                     :: dx1          ! dx1   (grid distance w.r.t. x1)
    DOUBLE PRECISION                     :: dx2          ! dx2   (grid distance w.r.t. x2)
    DOUBLE PRECISION                     :: dx3          ! dx3   (grid distance w.r.t. x3)
    DOUBLE PRECISION                     :: dx4          ! dx4   (grid distance w.r.t. x4)
    DOUBLE PRECISION                     :: odx1         ! one over dx 1
    DOUBLE PRECISION                     :: odx2         ! one over dx 2
    DOUBLE PRECISION                     :: odx3         ! one over dx 3
    DOUBLE PRECISION                     :: odx4         ! one over dx 4
    DOUBLE PRECISION, DIMENSION(:,:,:,:), POINTER :: ltab
  END TYPE lookupt_4D

  ! de: Type declaration for a lookup table calculating radar reflectivity dbz(T_a,q_i)
  !     and extinction ext(T_a,q_i) with q_i = q_r,q_s,q_g (rain,snow,graupel)
  TYPE dbzlookuptable
    INTEGER                                         :: nTa      ! number of lookup bins of temperature
    INTEGER                                         :: nqi      ! number of lookup bins of q_i
    INTEGER                                         :: nTm      ! number of lookup bins of Tmax (needed for melting particles)
    DOUBLE PRECISION                                :: dTa      ! increment of temperature
    DOUBLE PRECISION                                :: dqi      ! increment of q_i (logarithmic)
    DOUBLE PRECISION                                :: dTm      ! increment of Tmax
    DOUBLE PRECISION                                :: idTa     ! 1/T_a
    DOUBLE PRECISION                                :: idqi     ! 1/q_i (logarithmic)
    DOUBLE PRECISION                                :: idTm     ! 1/Tmax
    DOUBLE PRECISION                                :: Ta0      ! initial value of temperature
    DOUBLE PRECISION                                :: qi0      ! initial value of q_i (logarithmic)
    DOUBLE PRECISION                                :: Tm0      ! initial value of Tmax
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE     :: T_a      ! vector of temperature-values
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE     :: q_i      ! vector of logarithmic q_i-values
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE     :: q_i_lin  ! vector of (linear) q_i-values
    DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE     :: T_m      ! vector of Tmax
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE   :: dbz      ! value of radar reflectivity at (T_a,q_i)
    DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE   :: ext      ! value of extinction at (T_a,q_i)
    DOUBLE PRECISION, DIMENSION(:,:,:), ALLOCATABLE :: dbz_m    ! value of radar reflectivity at (T_a,q_i,T_m) for melting particles
    DOUBLE PRECISION, DIMENSION(:,:,:), ALLOCATABLE :: ext_m    ! value of extinction at (T_a,q_i,T_m) for melting particles
  END TYPE dbzlookuptable

  ! Type to hold some constants and prefactors for the Rayleigh-Oguchi-Approximation:
  TYPE rayl_oguchi_consts
    DOUBLE COMPLEX   :: m_w_0
    DOUBLE PRECISION :: K_w_0
    ! .. For LM cloud droplets:
    DOUBLE PRECISION :: D_c     ! m, Durchmesser des Partikels mit der mittleren Masse
    DOUBLE PRECISION :: nu_Dc   ! Shape-Par. nu hins. Durchmesser-Verteilung
    DOUBLE PRECISION :: mu_Dc   ! Shape-Par. mu hins. Durchmesser-Verteilun
    DOUBLE PRECISION :: nu_c    ! Shape-Par. nu hins. Massen-Verteilung
    DOUBLE PRECISION :: mu_c    ! Shape-Par. mu hins. Massen-Verteilung
    DOUBLE PRECISION :: prefactor_c
    ! .. For LM rain drops:
    DOUBLE PRECISION :: lamexprc
    DOUBLE PRECISION :: gamfak1rc
    DOUBLE PRECISION :: gamfak2rc
    ! .. For LM cloud ice:
    DOUBLE PRECISION :: D_i     ! m, Durchmesser des Partikels mit der mittleren Masse
    DOUBLE PRECISION :: nu_Di   ! Shape-Par. nu hins. Durchmesser-Verteilung
    DOUBLE PRECISION :: mu_Di   ! Shape-Par. mu hins. Durchmesser-Verteilun
    DOUBLE PRECISION :: nu_i    ! Shape-Par. nu hins. Massen-Verteilung
    DOUBLE PRECISION :: mu_i    ! Shape-Par. mu hins. Massen-Verteilung
    DOUBLE PRECISION :: prefactor_i
    DOUBLE PRECISION :: prefactor_im
    ! .. For LM snow:
    DOUBLE PRECISION :: bfaks
    DOUBLE PRECISION :: prefactor_s
    DOUBLE PRECISION :: prefactor_sm
    DOUBLE PRECISION :: gfc_bgeop1_s ! Factor gcft(bgeos+1), needed diameter of mean mass for exp. distr.
    ! .. For LM graupel:
    DOUBLE PRECISION :: bfakg
    DOUBLE PRECISION :: prefactor_g
    DOUBLE PRECISION :: prefactor_gm
    DOUBLE PRECISION :: gfc_bgeop1_g ! Factor gcft(bgeos+1), needed diameter of mean mass for exp. distr.
  END TYPE rayl_oguchi_consts

  TYPE(rayl_oguchi_consts) :: ray_const

  PRIVATE :: lookupt_2D, lookupt_3D, lookupt_4D


!===============================================================================
!===============================================================================


CONTAINS

!===============================================================================
!===============================================================================


!================================================================
!
! Domain start and end indices for computations
!
! This is to have the possibility to do computations only
! at those grid points which are absolutely necessary to
! save time.
!
! Has to be called at first before any computations are done
!  (usually at the beginning of the interface routines
!   radar_sb_mie_vec(), ...)
!
!================================================================

  SUBROUTINE get_loc_domain(zlinc_boundlines)

    IMPLICIT NONE

    LOGICAL, INTENT(in) :: zlinc_boundlines
    
    IF (zlinc_boundlines) THEN
      ! .. Full domain of this PE, i.e., including boundary lines:
      ilow = 1
      iup  = ie
      jlow = 1
      jup  = je
      klow = 1
      kup  = ke
    ELSE
      ! .. reduced domain of this PE, i.e., excluding interior
      !    boundary lines but including boudary halo at the outer
      !    model domain boundaries:
      ilow = istartpar
      iup  = iendpar
      jlow = jstartpar
      jup  = jendpar
      klow = 1
      kup  = ke
    END IF


  END SUBROUTINE get_loc_domain

!================================================================
!
! Size distribution parameters for the LM standard microphysics 
! schemes ( not the 2-moment scheme)
!
!!! itype_gscp == 5 nicht implementiert: Gamma-Verteilung fuer Regen, 
!!! 2-momenten-regen, n_rain(i,j,k), mue_rain(i,j,k)
!
!================================================================

  SUBROUTINE hydroparams_lm(ntime, itype_gscp, isnow_n0temp, &
             ageoi, &
             bgeoi, &
             ageos, &
             bgeos, &
             ageog, &
             bgeog, &
             n0_r, &
             n0_s, &
             n0_g, mu_rain, mur_is_fixed, n0r_is_fixed)

    IMPLICIT NONE
     
    INTEGER, INTENT(in) :: itype_gscp, isnow_n0temp, ntime
    REAL(kind=ireals), INTENT(out) :: &
             ageoi, &
             bgeoi, &
             ageos, &
             bgeos, &
             ageog, &
             bgeog, &
             n0_g

    REAL(kind=ireals), INTENT(out) :: n0_r(ie,je,ke), n0_s(ie,je,ke), mu_rain(ie,je,ke)
    LOGICAL, INTENT(out) :: mur_is_fixed, n0r_is_fixed  

    INTEGER(kind=iintegers)        :: i, j, k

!-----------------------------------------------------------------------------------


    !.. mass-size- relation parameters for snow and graupel:
    IF (itype_gscp == 4) THEN
      ageos = 0.038_ireals
      bgeos = 2.0_ireals
    ELSE
      ageos = 0.069_ireals
      bgeos = 2.0_ireals
    END IF
    ageoi = 130.0_ireals
    bgeoi = 3.0_ireals
    ageog = 169.6_ireals
    bgeog = 3.1_ireals

    !.. mu_rain:
    IF (itype_gscp >= 4) THEN
      mu_rain = 0.5_ireals
      mur_is_fixed = .TRUE.
!!! itype_gscp == 5 nicht implementiert: Gamma-Verteilung fuer Regen, 2-momenten-regen, n_rain(i,j,k), mue_rain(i,j,k)
    ELSE
      mu_rain = 0.0_ireals
      mur_is_fixed = .TRUE.
    END IF

    !.. N0-Parameters (except for snow):
    IF (itype_gscp >= 3) THEN
!!$      n0_r = 3.96E7_ireals
!!$ UB: moment relation of Ulbrich consistent to hydci_pp / hydci_gr (the "*0.1" is a "needed" bug):
      n0_r = 8e6 * EXP(3.2*mu_rain) * (0.01)**(-mu_rain) * 0.1
      n0r_is_fixed = .TRUE.
!!! itype_gscp == 5 nicht implementiert: Gamma-Verteilung fuer Regen, 2-momenten-regen, n_rain(i,j,k), mue_rain(i,j,k)
    ELSE
      n0_r = 8.0e6_ireals
      n0r_is_fixed = .TRUE.
    END IF
    n0_g = 4.0e6_ireals

    !.. N0-Parameter for snow:
    DO k=klow,kup
      DO j=jlow,jup
        DO i=ilow,iup
          CALL calc_n0_snow(isnow_n0temp, t(i,j,k,ntime), &
                            qs(i,j,k,ntime)*rho(i,j,k), ageos, n0_s(i,j,k))
        END DO
      END DO
    END DO
         
    RETURN
  END SUBROUTINE hydroparams_lm
  

  !-----------------------------------------------------------------------------------
  ! Subroutine to calculate n0_s as a function of temperature after Field et al. (2005)
  !
  ! Input:    T_a    Temperature in K
  !           q_s    mass density of snow in kg/m^3
  !           ageos  Parameter of mass_size-relation x=ageos*D^bgeos, x in kg, D in m
  ! Output:   n0_s   Intercept parameter of expon. snow size distrib. in m^-4
  !-----------------------------------------------------------------------------------
  
  SUBROUTINE calc_n0_snow(isnow_n0temp, T_a, q_s, ageos, n0_s)

    IMPLICIT NONE
     
    INTEGER, INTENT(in)            :: isnow_n0temp

    REAL(kind=ireals), INTENT(in)  :: T_a, q_s, ageos

    REAL(kind=ireals), INTENT(out) :: n0_s

    REAL (KIND=ireals)             :: ztc, zn0s, &
                                      alf, bet, hlp, m2s, m3s

    INTEGER(kind=iintegers)        :: nn
    
    REAL (KIND=ireals), PARAMETER  :: zn0s1 = 13.5_ireals * 5.65E5_ireals, & ! parameter in N0S(T)
                                      zn0s2 = -0.107_ireals                  ! parameter in N0S(T), Field et al

!-----------------------------------------------------------------------------------

    ! Snow: Coeffs for moment relation based on 2nd moment (Field et al. 2005)
    !        (defined as parameters because of better inlining and vectorization properties)
    REAL (KIND=ireals), DIMENSION(10), PARAMETER  ::   &
         mma = (/5.065339_ireals, -0.062659_ireals, -3.032362_ireals, 0.029469_ireals, -0.000285_ireals, &
                 0.312550_ireals,  0.000204_ireals,  0.003199_ireals, 0.000000_ireals, -0.015952_ireals /)  , &
         mmb = (/0.476221_ireals, -0.015896_ireals,  0.165977_ireals, 0.007468_ireals, -0.000141_ireals, &
                 0.060366_ireals,  0.000079_ireals,  0.000594_ireals, 0.000000_ireals, -0.003577_ireals /)

    !.. N0-Parameter for snow:
    n0_s = 1e9_ireals  ! initialization

    IF (isnow_n0temp == 1) THEN
      ! Calculate n0s using the temperature-dependent formula of Field et al. (2005)
      ztc = T_a - 273.16
      ztc = MAX(MIN(ztc,0.0_ireals),-40.0_ireals)          
      zn0s = zn0s1*EXP(zn0s2*ztc)
      zn0s = MIN(zn0s,1e9_ireals)
      zn0s = MAX(zn0s,1e6_ireals)
      n0_s = zn0s
    ELSEIF (isnow_n0temp == 2) THEN
      ! Calculate n0s using the temperature-dependent moment relations of Field et al. (2005)
      ztc = T_a - 273.16
      ztc = MAX(MIN(ztc,0.0_ireals),-40.0_ireals)
      
      nn  = 3
      hlp = mma(1)      +mma(2)*ztc      +mma(3)*nn       +mma(4)*ztc*nn+mma(5)*ztc**2 &
          + mma(6)*nn**2+mma(7)*ztc**2*nn+mma(8)*ztc*nn**2+mma(9)*ztc**3+mma(10)*nn**3
      alf = 10.0_ireals**hlp
      bet = mmb(1)      +mmb(2)*ztc      +mmb(3)*nn       +mmb(4)*ztc*nn+mmb(5)*ztc**2 &
          + mmb(6)*nn**2+mmb(7)*ztc**2*nn+mmb(8)*ztc*nn**2+mmb(9)*ztc**3+mmb(10)*nn**3
      IF (q_s >= 1e-20_ireals) THEN
!!$ UB: caution! Here is the exponent bms=2.0 hardwired! not ideal!
        m2s = q_s / ageos
        m3s = alf*EXP(bet*LOG(m2s))
        hlp = zn0s1*EXP(zn0s2*ztc)
!!$ UB: the 13.5 is actually 3^3 / gamma(3) ...
        zn0s = 13.50_ireals * m2s**4 / m3s**3
        zn0s = MAX(zn0s,0.5_ireals*hlp)
        zn0s = MIN(zn0s,1e2_ireals*hlp)
        zn0s = MIN(zn0s,1e9_ireals)
        n0_s = MAX(zn0s,8e5_ireals)
      ELSE
        n0_s = 1e9_ireals
      END IF
    ELSE
      ! Old constant n0s
      n0_s = 8.0d5
    ENDIF
    
    RETURN
  END SUBROUTINE calc_n0_snow
! de.

  !=======================================================================
  !
  ! Diverse Hilfsfunktionen:
  !
  !=======================================================================

  ! Mittlerer Durchmesser von verallg. gammaverteilten Hydrometeoren:
  FUNCTION D_average_sb (q, qn, parti) RESULT (D)

    ! gueltig fuer D = a_geo * x^b_geo

    IMPLICIT NONE

    DOUBLE PRECISION :: D
    DOUBLE PRECISION, INTENT(in) :: q, qn
    TYPE(PARTICLE), INTENT(in) :: parti
    DOUBLE PRECISION :: fakt, x

    fakt = gfct( (parti%b_geo+parti%nu+1.0)/parti%mu ) / &
         gfct( (parti%nu+1.0)/parti%mu ) * &
         (gfct( (parti%nu+1.0)/parti%mu ) / gfct( (parti%nu+2.0)/parti%mu )) ** parti%b_geo

    x = MIN(MAX(q/(qn+1d-20),parti%x_min),parti%x_max)

    D = fakt * parti%a_geo * x ** parti%b_geo

!!$    IF (q >= 1d-10 .AND. qn >= 1.0d0) THEN
!!$      D = &
!!$           parti%a_geo * gfct( (parti%b_geo+parti%nu+1.0)/parti%mu ) / &
!!$           gfct( (parti%nu+1.0)/parti%mu ) * &
!!$           (gfct( (parti%nu+1.0)/parti%mu ) / gfct( (parti%nu+2.0)/parti%mu ) * q / qn) ** parti%b_geo
!!$    ELSE
!!$      D = parti%a_geo * parti%x_min ** parti%b_geo
!!$    END IF

  END FUNCTION D_average_sb

  ! vectorized version:
  FUNCTION D_average_sb_vec (q, qn, parti, anz) RESULT (D)

    ! gueltig fuer D = a_geo * x^b_geo

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION :: D(anz)
    DOUBLE PRECISION, INTENT(in) :: q(anz), qn(anz)
    TYPE(PARTICLE), INTENT(in) :: parti
    DOUBLE PRECISION :: fakt, x(anz)
    
    fakt = gfct( (parti%b_geo+parti%nu+1.0)/parti%mu ) / &
         gfct( (parti%nu+1.0)/parti%mu ) * &
         (gfct( (parti%nu+1.0)/parti%mu ) / gfct( (parti%nu+2.0)/parti%mu )) ** parti%b_geo
    
    x = MIN(MAX(q/(qn+1d-20),parti%x_min),parti%x_max)

    D = fakt * parti%a_geo * x ** parti%b_geo

  END FUNCTION D_average_sb_vec

  FUNCTION D_average (q, qn, a_geo, b_geo, nu, mu) RESULT (D)

    ! gueltig fuer n0, nu, mu bezueglich Massen-Verteilung
    ! gueltig fuer D = a_geo * x^b_geo

    IMPLICIT NONE

    DOUBLE PRECISION :: D
    DOUBLE PRECISION, INTENT(in) :: q, qn, a_geo, b_geo, nu, mu
    DOUBLE PRECISION :: fakt

    fakt = gfct( (b_geo+nu+1.0)/mu ) / &
           gfct( (nu+1.0)/mu ) * &
           (gfct( (nu+1.0)/mu ) / gfct( (nu+2.0)/mu )) ** b_geo

    D = fakt * a_geo * MIN(MAX((q/(qn+1d-20)),1d-12),1.0d0) ** b_geo

  END FUNCTION D_average

  ! vectorized version:
  FUNCTION D_average_vec (q, qn, a_geo, b_geo, nu, mu, anz) RESULT (D)

    ! gueltig fuer n0, nu, mu bezueglich Massen-Verteilung
    ! gueltig fuer D = a_geo * x^b_geo

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in) :: q(anz), qn(anz), a_geo, b_geo, nu, mu
    DOUBLE PRECISION :: D(anz)
    DOUBLE PRECISION :: fakt

    fakt = gfct( (b_geo+nu+1.0)/mu ) / &
           gfct( (nu+1.0)/mu ) * &
           (gfct( (nu+1.0)/mu ) / gfct( (nu+2.0)/mu )) ** b_geo

    D = fakt * a_geo * MIN(MAX((q/(qn+1d-20)),1d-12),1.0d0) ** b_geo

  END FUNCTION D_average_vec

  FUNCTION D_average_1M_exp (q, a_geo, b_geo, n0) RESULT (D)

    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    DOUBLE PRECISION :: D
    DOUBLE PRECISION, INTENT(in) :: q, a_geo, b_geo, n0
    DOUBLE PRECISION :: fakt

    fakt = 1.0d0 / (a_geo * n0 * gfct(b_geo+1.0))

    D = (MAX(q,1d-12) * fakt) ** (1.0/(1.0+b_geo))

  END FUNCTION D_average_1M_exp

  FUNCTION D_average_1M_exp_cin (q, a_geo, b_geo, n0, gfc_bgeop1) RESULT (D)

    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    DOUBLE PRECISION :: D
    DOUBLE PRECISION, INTENT(in) :: q, a_geo, b_geo, n0, gfc_bgeop1
    DOUBLE PRECISION :: fakt

!!$    fakt = 1.0d0 / (a_geo * n0 * gfct(b_geo+1.0))
    fakt = 1.0d0 / (a_geo * n0 * gfc_bgeop1)
    D = (MAX(q,1d-12) * fakt ) ** (1.0/(1.0+b_geo))

  END FUNCTION D_average_1M_exp_cin

  ! vectorized version:
  FUNCTION D_average_1M_exp_vec (q, a_geo, b_geo, n0, anz) RESULT (D)

    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION :: D(anz)
    DOUBLE PRECISION, INTENT(in) :: q(anz), a_geo, b_geo, n0
    DOUBLE PRECISION :: fakt

    fakt = 1.0d0 / (a_geo * n0 * gfct(b_geo+1.0))

    D = (MAX(q,1d-12) * fakt) ** (1.0/(1.0+b_geo))

  END FUNCTION D_average_1M_exp_vec

  FUNCTION D_average_1M_exp_vec_n0s (q, a_geo, b_geo, n0, anz) RESULT (D)

    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION :: D(anz)
    DOUBLE PRECISION, INTENT(in) :: q(anz), a_geo, b_geo, n0(anz)
    DOUBLE PRECISION :: fakt(anz)

    fakt = 1.0d0 / (a_geo * n0 * gfct(b_geo+1.0))

    D = (MAX(q,1d-12) * fakt) ** (1.0/(1.0+b_geo))

  END FUNCTION D_average_1M_exp_vec_n0s

  FUNCTION D_of_X_average_1M_exp (q, a_geo, b_geo, n0) RESULT (D)

    ! Diameter of mean mass particle
    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    DOUBLE PRECISION :: D
    DOUBLE PRECISION, INTENT(in) :: q, a_geo, b_geo, n0
    DOUBLE PRECISION :: fakt, expo

    expo = 1.0d0/(b_geo*(b_geo+1.0d0))
    fakt = gfct(b_geo+1.0) ** expo / (a_geo*n0)**(b_geo*expo)
    D = MAX(q,1d-12)**(b_geo*expo) * fakt

  END FUNCTION D_of_X_average_1M_exp

  FUNCTION D_of_X_average_1M_exp_cin (q, a_geo, b_geo, n0, gfc_bgeop1) RESULT (D)

    ! Diameter of mean mass particle
    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    DOUBLE PRECISION :: D
    DOUBLE PRECISION, INTENT(in) :: q, a_geo, b_geo, n0, gfc_bgeop1
    DOUBLE PRECISION :: fakt, expo

    expo = 1.0d0/(b_geo*(b_geo+1.0d0))
!!$    fakt = gfct(b_geo+1.0) ** expo / (a_geo*n0)**(b_geo*expo)
    fakt = gfc_bgeop1 ** expo / (a_geo*n0)**(b_geo*expo)
    D = MAX(q,1d-12)**(b_geo*expo) * fakt

  END FUNCTION D_of_X_average_1M_exp_cin

  ! vectorized version:
  FUNCTION D_of_X_average_1M_exp_vec (q, a_geo, b_geo, n0, anz) RESULT (D)

    ! Diameter of mean mass particle
    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION :: D(anz)
    DOUBLE PRECISION, INTENT(in) :: q(anz), a_geo, b_geo, n0
    DOUBLE PRECISION :: fakt, expo

    expo = 1.0d0/(b_geo*(b_geo+1.0d0))
    fakt = gfct(b_geo+1.0) ** expo / (a_geo*n0)**(b_geo*expo)
    D = MAX(q,1d-12)**(b_geo*expo) * fakt

  END FUNCTION D_of_X_average_1M_exp_vec

  FUNCTION D_of_X_average_1M_exp_vec_n0s (q, a_geo, b_geo, n0, anz) RESULT (D)

    ! Diameter of mean mass particle
    ! gueltig fuer n0, nu, mu bezueglich Durchmesser-Verteilung
    ! gueltig fuer x = a_geo * D^b_geo

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION :: D(anz)
    DOUBLE PRECISION, INTENT(in) :: q(anz), a_geo, b_geo, n0(anz)
    DOUBLE PRECISION :: fakt(anz), expo

    expo = 1.0d0/(b_geo*(b_geo+1.0d0))
    fakt = gfct(b_geo+1.0) ** expo / (a_geo*n0)**(b_geo*expo)
    D = MAX(q,1d-12)**(b_geo*expo) * fakt

  END FUNCTION D_of_X_average_1M_exp_vec_n0s

  !----------------------------------------------------------------------------------------!
  ! Einfache Parametrisierung des Schmelzgrades von Eis- und Schneepartikeln
  ! als Funktion der Temperatur und des Partikeldurchmessers Dvec(i), 
  ! die aus einem einfachen parametrischen Modell besteht, dessen Parameter momentan 
  ! rein heuristisch abgeleitet worden sind und die noch der Anpassung durch ein
  ! detailliertes Schmelzmodell beduerfen!
  ! 
  !
  ! Der Schmelzgrad fuer den Ort im, jm, km wird als Gitterpunktswert zurueckgegeben und
  ! wird jeweils unter Verwendung des eingegebenen Feldes des Fluessigwassergehaltes
  ! aus der Umgebung im +/- neigh, jm +/- neigh Gitterpunkten aus der vertikalen Saeule bestimmt.
  ! 
  ! Mit der Temperatur Tmeltbegin (K) kann der Effekt des Riming-induzierten Schmelzens
  ! von Partikeln bei Temperaturen kleiner als die Temperatur Tmin (hier -1 Grad C)  modelliert werden.
  ! Ist Tmeltbegin < Tmin, so wird bei Temperaturen von Tmeltbegin < T < T_freeze ein linear mit der 
  ! Temperatur zunehmender Schmelzgrad, unabh. von der Partikelgroesse und Riming-Rate, angenommen, der
  ! bei Tmin den Wert meltdegTmin erreicht. Ist Tmeltbegin > Tmin, so wird dies nicht gemacht.
  ! Man beachte, dass dieser Effekt in der Realitaet stark von der Riming-Rate abhaengen duerfte,
  ! was hier nicht beruecksichtigt ist!
  !
  ! Die hier verwendete Parametrisierung ist ein parametrisches Modell, das
  ! eigentlich noch der Anpassung durch ein detailliertes Schmelzmodell bedarf
  ! (1D-Modell).
  !----------------------------------------------------------------------------------------!

  ! Zuerst die Masterfunktion, spaeter wrapper-funktionen fuer die einzelnen Hydrometeorarten
  ! (Fassung mit 2D-Ausgabefeld fmelt(anzt, anzD))
  ! temperatur(anzt), Dref(anzt) und Tmax(anzt) koennen z.B. die verschiedenen Werte entlang einer
  ! Datenzeile in Modell-X-Richtung enthalten.
  SUBROUTINE degree_of_melting_fun_vec(temperatur,anzt,Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anzt, anz
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt,anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz), temperatur(anzt)

    DOUBLE PRECISION, INTENT(in) :: Dref(anzt), Dexpo
    DOUBLE PRECISION, INTENT(in) :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dfullmelted(anzt,anz),tmpmeltdegTmin,Dmat(anzt,anz),tempmat(anzt,anz),Drefmat(anzt,anz),Tmaxmat(anzt,anz)
    INTEGER :: i

    fmelt = 0.0d0

    ! Die Hilfsfelder Dmat und tempmat belegen, dabei
    ! abhaengig von den Felddimensionen vektorisierbaren Code erzeugen:
    IF (anzt == 1) THEN
      DO i=1, anz
        Dmat(1,i) = Dvec(i)
        tempmat(1,i) = temperatur(1)
        Drefmat(1,i) = Dref(1)
        tmaxmat(1,i) = Tmax(1)
      END DO
    ELSE
      IF (anz == 1) THEN
        DO i=1, anzt
          Dmat(i,1) = Dvec(1)
          tempmat(i,1) = temperatur(i)
          Drefmat(i,1) = Dref(i)
          Tmaxmat(i,1) = Tmax(i)
        END DO
      ELSE
        DO i=1, anz
          Dmat(:,i) = Dvec(i)
        END DO
        DO i=1, anzt
          tempmat(i,:) = temperatur(i)
          Drefmat(i,:) = Dref(i)
          Tmaxmat(i,:) = Tmax(i)
        END DO
      END IF
    END IF
    
    ! Jetzt die eigentliche Schmelzgradparametrisierung:
    IF (Tmeltbegin < Tmin) THEN
      tmpmeltdegTmin = meltdegTmin
      WHERE (tempmat > Tmeltbegin)
        fmelt = &
             MIN(MAX(meltdegTmin/(Tmin-Tmeltbegin)*(tempmat-Tmeltbegin), 0.0d0), 1.0d0)
      ELSEWHERE
        fmelt = 0.0d0
      END WHERE
    ELSE
      tmpmeltdegTmin = 0.0d0      
    END IF
    
    Dfullmelted = 0.0d0
    WHERE (tempmat >= Tmaxmat)
      fmelt = 1.0d0
    END WHERE
    WHERE (tempmat < Tmaxmat .AND. tempmat >= Tmin .AND. Dmat >= 1d-20)
      Dfullmelted = Drefmat**Dexpo * LOG((Tmaxmat-Tmin)/(Tmaxmat-tempmat))
      fmelt = MIN((1.0d0-tmpmeltdegTmin) * Dfullmelted / Dmat**Dexpo + tmpmeltdegTmin, 1.0d0)
    END WHERE
    WHERE (tempmat < Tmaxmat .AND. tempmat >= Tmin .AND. Dmat < 1d-20)
      fmelt = 1.0d0
    END WHERE
    
    RETURN
  END SUBROUTINE degree_of_melting_fun_vec

  ! Bei dieser Variante haben temperatur(anz) und Dvec(anz) dieselbe Gestalt, d.h.,
  ! fuer jeden Gitterpunkt wird nur ein Wert von fmelt bei einem Durchmesser 
  ! berechnet:
  SUBROUTINE degree_of_melting_fun_vec_ie(temperatur,Dvec,anzt,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anzt
    DOUBLE PRECISION, intent(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anzt), temperatur(anzt)

    DOUBLE PRECISION, INTENT(in) :: Dref(anzt), Dexpo
    DOUBLE PRECISION, INTENT(in) :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dfullmelted(anzt),tmpmeltdegTmin

    fmelt = 0.0d0

    IF (Tmeltbegin < Tmin) THEN
      tmpmeltdegTmin = meltdegTmin
      WHERE (temperatur > Tmeltbegin)
        fmelt = &
             MIN(MAX(meltdegTmin/(Tmin-Tmeltbegin)*(temperatur-Tmeltbegin), 0.0d0), 1.0d0)
      ELSEWHERE
        fmelt = 0.0
      END WHERE
    ELSE
      tmpmeltdegTmin = 0.0        
    END IF

    Dfullmelted = 0.0d0
    WHERE (temperatur >= Tmax)
      fmelt = 1.0d0
    END WHERE
    WHERE(temperatur < Tmax .AND. temperatur >= Tmin .AND. Dvec >= 1d-20)
      Dfullmelted = Dref**Dexpo * LOG((Tmax-Tmin)/(Tmax-temperatur))
      fmelt = MIN((1.0d0-tmpmeltdegTmin) * Dfullmelted / Dvec**Dexpo + tmpmeltdegTmin, 1.0d0)
    END WHERE
    WHERE(temperatur < Tmax .AND. temperatur >= Tmin .AND. Dvec < 1d-20)
      fmelt = 1.0d0
    END WHERE

    RETURN
  END SUBROUTINE degree_of_melting_fun_vec_ie

  !=================================================================================
  ! Die alte Routine, die nur an einem einzigen Gitterpunkt gerechnet werden kann:
  !=================================================================================

  ! Zuerst die Masterfunktion, spaeter wrapper-funktionen fuer die einzelnen Hydrometeorarten
  SUBROUTINE degree_of_melting_fun(temperatur,Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, intent(out) :: fmelt(anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz), temperatur

    DOUBLE PRECISION, INTENT(in) :: Dref, Dexpo
    DOUBLE PRECISION, INTENT(in) :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dfullmelted,tmpmeltdegTmin

    fmelt = 0.0d0

    IF (Tmeltbegin < Tmin) THEN
      tmpmeltdegTmin = meltdegTmin
      IF (temperatur > Tmeltbegin) THEN
!!$        fmelt = &
!!$             MIN(MAX(meltdegTmin/(Tmin-Tmeltbegin)*(temperatur-Tmeltbegin), 0.0d0), 1.0d0)
        ! the max(...,1d-20) is necessary because of strange behaviour when this code is inlined ...
        fmelt = &
             MIN(meltdegTmin/MAX(Tmin-Tmeltbegin,1d-20)*(temperatur-Tmeltbegin), 1.0d0)
      ELSE
        fmelt = 0.0
      END IF
    ELSE
      tmpmeltdegTmin = 0.0        
    END IF

    IF (temperatur >= Tmax) THEN
      fmelt = 1.0d0
    ELSE IF (temperatur < Tmax .AND. temperatur >= Tmin) THEN
!!$      Dfullmelted = Dref**Dexpo * LOG((Tmax-Tmin)/(Tmax-temperatur))
      ! the max(...,1d-20) is necessary because of strange behaviour when this code is inlined ...
      Dfullmelted = Dref**Dexpo * LOG((Tmax-Tmin)/MAX(Tmax-temperatur,1d-20))
      WHERE (Dvec**Dexpo >= Dfullmelted .AND. Dvec >= 1d-20)
        fmelt = (1.0d0-tmpmeltdegTmin) * Dfullmelted / Dvec**Dexpo + tmpmeltdegTmin
      ELSEWHERE
        fmelt = 1.0d0
      END WHERE
    END IF

    RETURN
  END SUBROUTINE degree_of_melting_fun

  !=================================================================================
  ! Die Routine, die nur an einem einzigen Gitterpunkt fuer ein einziges D 
  !   gerechnet werden kann:
  !=================================================================================

  ! Zuerst die Masterfunktion, spaeter wrapper-funktionen fuer die einzelnen Hydrometeorarten
  SUBROUTINE degree_of_melting_fun_single(temperatur,D,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    IMPLICIT NONE

    DOUBLE PRECISION, intent(out) :: fmelt
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, D, temperatur

    DOUBLE PRECISION, INTENT(in) :: Dref, Dexpo
    DOUBLE PRECISION, INTENT(in) :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dfullmelted,tmpmeltdegTmin

    fmelt = 0.0d0

    IF (Tmeltbegin < Tmin) THEN
      tmpmeltdegTmin = meltdegTmin
      IF (temperatur > Tmeltbegin) THEN
!!$        fmelt = &
!!$             MIN(MAX(meltdegTmin/(Tmin-Tmeltbegin)*(temperatur-Tmeltbegin), 0.0d0), 1.0d0)
        ! the max(...,1d-20) is necessary because of strange behaviour on the NEC when this code is inlined ...
        fmelt = &
             MIN(meltdegTmin/MAX(Tmin-Tmeltbegin,1d-20)*(temperatur-Tmeltbegin), 1.0d0)
      ELSE
        fmelt = 0.0
      END IF
    ELSE
      tmpmeltdegTmin = 0.0        
    END IF

    IF (temperatur >= Tmax) THEN
      fmelt = 1.0d0
    ELSE IF (temperatur < Tmax .AND. temperatur >= Tmin) THEN
!!$      Dfullmelted = Dref**Dexpo * LOG((Tmax-Tmin)/(Tmax-temperatur))
      ! the max(...,1d-20) is necessary because of strange behaviour when this code is inlined ...
      Dfullmelted = Dref**Dexpo * LOG((Tmax-Tmin)/MAX(Tmax-temperatur,1d-20))
      IF (D**Dexpo >= Dfullmelted .AND. D >= 1d-20) THEN
        fmelt = (1.0d0-tmpmeltdegTmin) * Dfullmelted / D**Dexpo + tmpmeltdegTmin
      ELSE
        fmelt = 1.0d0
      END IF
    END IF

    RETURN
  END SUBROUTINE degree_of_melting_fun_single

  ! Bevor irgendeine der untenstehenden Funktionen zur Schmelzgradparametrisierung verwendet werden kann,
  ! muss die folgende Funktion aufgerufen werden, die einige Parameterfelder vorbereitet:
  ! (Version fuer das 2M-Schema)
  ! Vorsicht: nicht parallelisiert! Tmax-Umgebung nur innerhalb eines Prozessor-Gebietes!
  SUBROUTINE initialize_tmax_sb_vec(ntime, neigh)

    IMPLICIT NONE
    INTEGER, INTENT(in) :: ntime
    DOUBLE PRECISION, INTENT(in) :: neigh

    INTEGER :: iu, ju, i, j, k, neigh_x, neigh_y, im, jm
    DOUBLE PRECISION :: dx, dy
    DOUBLE PRECISION :: qithresh, qsthresh, qgthresh, qhthresh

    ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
    qithresh = 1d-8
    qsthresh = 1d-8
    qgthresh = 1d-8

    ALLOCATE(Tmax_i(ie,je),Tmax_s(ie,je),Tmax_g(ie,je))

    ! Initialwerte von Tmax_x: Diese Werte werden mindestens genommen:
    Tmax_i = 3.0d0 + 273.16
    Tmax_s = 3.0d0 + 273.16
    Tmax_g = 3.0d0 + 273.16

    dx = dlon * r_earth * pi / 180.0 * COS((startlat_tot+(je_tot/2)*dlat)*pi/180.0d0)
    dy = dlat * r_earth * pi / 180.0

    neigh_x = MAX(INT(neigh/dx), 0)
    neigh_y = MAX(INT(neigh/dy), 0)

    !!! neigh_x, neigh_y have to be smaller than the sub-processor grid size,
    !!! since a cross-subgrid search is not implemented so far!
    !!! Further, it is limited to 5 gridpoints to limit computational burden,
    !!! which makes also sense physically.
    IF (neigh_x > 5 .OR. neigh_x > ie) THEN
      neigh_x = MAX(MIN(MIN(neigh_x, ie ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_SB_VEC: neigh_x has been clipped to ', neigh_x, &
           ' gridpoints or ', neigh_x*dx*1e-3, ' km'
    END IF
    IF (neigh_y > 5 .OR. neigh_y > je) THEN
      neigh_y = MAX(MIN(MIN(neigh_y, je ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_SB_VEC: neigh_y has been clipped to ', neigh_y, &
           ' gridpoints or ', neigh_y*dy*1e-3, ' km'
    END IF

    ! Tmax wird aus der neigh-Umgebung des Punktes im,jm als hoechste Temperatur bestimmt,
    ! bei der noch ein nennenswerter Wassergehalt (> qthresh) der Spezies q in der Umgebung neigh vorkommt. Auf
    ! diese Weise bestimmt die Schmelzparametrisierung der Wolkenmikrophysik in 
    ! entscheidender Weise das Verhalten des hier parametrisierten Schmelzgrades.
    DO j=-neigh_x,neigh_x
      DO i=-neigh_x,neigh_x
        
        DO k=1,ke
          DO jm=1,je
            DO im=1,ie

              iu = MIN(MAX(im+i,1),ie)
              ju = MIN(MAX(jm+j,1),je)

              IF (qi(iu,ju,k,ntime) > qithresh .AND. qni(iu,ju,k,ntime) > 1.0d0) THEN
                Tmax_i(im,jm) = MAX(Tmax_i(im,jm), t(iu,ju,k,ntime))
              END IF
              IF (qs(iu,ju,k,ntime) > qsthresh .AND. qns(iu,ju,k,ntime) > 1.0d0) THEN
                Tmax_s(im,jm) = MAX(Tmax_s(im,jm), t(iu,ju,k,ntime))
              END IF
              IF (qg(iu,ju,k,ntime) > qgthresh .AND. qng(iu,ju,k,ntime) > 1.0d-3) THEN
                Tmax_g(im,jm) = MAX(Tmax_g(im,jm), t(iu,ju,k,ntime))
              END IF
            END DO
          END DO
        END DO
      END DO
    END DO

    IF (ALLOCATED(qh)) THEN

      qhthresh = 1d-8

      ALLOCATE(Tmax_h(ie,je))

      Tmax_h = 3.0d0 + 273.16

      DO j=-neigh_x,neigh_x
        DO i=-neigh_x,neigh_x
          
          DO k=1,ke
            DO jm=1,je
              DO im=1,ie
                
                iu = MIN(MAX(im+i,1),ie)
                ju = MIN(MAX(jm+j,1),je)

                IF (qh(iu,ju,k,ntime) > qhthresh .AND. qnh(iu,ju,k,ntime) > 1.0d-3) THEN
                  Tmax_h(im,jm) = MAX(Tmax_h(im,jm), t(iu,ju,k,ntime))
                END IF
              END DO
            END DO
          END DO
        END DO
      END DO

    END IF

    RETURN
  END SUBROUTINE initialize_tmax_sb_vec

  ! parallelisiert (bisher Tmax-Umgebung nur innerhalb eines Prozessor-Gebietes!):
  SUBROUTINE initialize_tmax_sb_vec_par(ntime, neigh)

    IMPLICIT NONE
    INTEGER, INTENT(in) :: ntime
    DOUBLE PRECISION, INTENT(in) :: neigh

    INTEGER :: iu, ju, i, j, k, neigh_x, neigh_y, im, jm, ierr
    DOUBLE PRECISION :: dx, dy
    DOUBLE PRECISION :: qithresh, qsthresh, qgthresh, qhthresh

    DOUBLE PRECISION, ALLOCATABLE :: Tmax_glob(:,:,:)

    ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
    qithresh = 1d-8
    qsthresh = 1d-8
    qgthresh = 1d-8

    ALLOCATE(Tmax_i(ie,je),Tmax_s(ie,je),Tmax_g(ie,je))

    ! Initialwerte von Tmax_x: Diese Werte werden mindestens genommen:
    Tmax_i = 3.0d0 + 273.16
    Tmax_s = 3.0d0 + 273.16
    Tmax_g = 3.0d0 + 273.16

    dx = dlon * r_earth * pi / 180.0 * COS((startlat_tot+(je_tot/2)*dlat)*pi/180.0d0)
    dy = dlat * r_earth * pi / 180.0

    neigh_x = MAX(INT(neigh/dx), 0)
    neigh_y = MAX(INT(neigh/dy), 0)

    !!! neigh_x, neigh_y is limited to 10 (5 on either side) gridpoints to limit computational burden,
    !!! which makes also sense physically.
    IF (neigh_x > 5) THEN
      neigh_x = 5
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_SB_VEC_PAR: neigh_x has been clipped to ', neigh_x, &
           ' gridpoints or ', neigh_x*dx*1e-3, ' km'
    END IF
    IF (neigh_y > 5) THEN
      neigh_y = 5
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_SB_VEC_PAR: neigh_y has been clipped to ', neigh_y, &
           ' gridpoints or ', neigh_y*dy*1e-3, ' km'
    END IF


    ! Tmax wird aus der neigh-Umgebung des Punktes im,jm als hoechste Temperatur bestimmt,
    ! bei der noch ein nennenswerter Wassergehalt (> qthresh) der Spezies q in der Umgebung neigh vorkommt. Auf
    ! diese Weise bestimmt die Schmelzparametrisierung der Wolkenmikrophysik in 
    ! entscheidender Weise das Verhalten des hier parametrisierten Schmelzgrades.

    ! First, compute max. Temperature where hydrometeors are present within each vertical column:
    DO k=1,ke
      DO jm=1,je
        DO im=1,ie
          
          IF (qi(im,jm,k,ntime) > qithresh .AND. qni(im,jm,k,ntime) > 1.0d0) THEN
            Tmax_i(im,jm) = MAX(Tmax_i(im,jm), t(im,jm,k,ntime))
          END IF
          IF (qs(im,jm,k,ntime) > qsthresh .AND. qns(im,jm,k,ntime) > 1.0d0) THEN
            Tmax_s(im,jm) = MAX(Tmax_s(im,jm), t(im,jm,k,ntime))
          END IF
          IF (qg(im,jm,k,ntime) > qgthresh .AND. qng(im,jm,k,ntime) > 1.0d-3) THEN
            Tmax_g(im,jm) = MAX(Tmax_g(im,jm), t(im,jm,k,ntime))
          END IF

        END DO
      END DO
    END DO
    
    IF (ALLOCATED(qh)) THEN
      
      qhthresh = 1d-8
      
      ALLOCATE(Tmax_h(ie,je))
      
      Tmax_h = 3.0d0 + 273.16
      
      DO k=1,ke
        DO jm=1,je
          DO im=1,ie
            
            IF (qh(im,jm,k,ntime) > qhthresh .AND. qnh(im,jm,k,ntime) > 1.0d-3) THEN
              Tmax_h(im,jm) = MAX(Tmax_h(im,jm), t(im,jm,k,ntime))
            END IF
            
          END DO
        END DO
      END DO
      
    END IF


    ! In case of parallel run, collect the vertical maxima on the root node,
    ! compute the maxima over the neigh_x/neigh_y-Region and redistribute
    ! the global field to the single nodes:
    IF (num_compute > 1) THEN

      ALLOCATE(Tmax_glob(ie_tot,je_tot,2))

      ! 1) Tmax for cloud ice:

      CALL gather_field(Tmax_i, ie, je, Tmax_glob(:,:,1), ie_tot, je_tot, 0, ierr)

      Tmax_glob(:,:,2) = Tmax_glob(:,:,1)

      IF (my_cart_id == 0) THEN
        
        DO j=-neigh_x,neigh_x
          DO i=-neigh_x,neigh_x
            DO jm=1,je_tot
              DO im=1,ie_tot
                
                iu = MIN(MAX(im+i,1),ie_tot)
                ju = MIN(MAX(jm+j,1),je_tot)

                Tmax_glob(im,jm,2) = MAX(Tmax_glob(iu,ju,1), Tmax_glob(im,jm,2))               
                
              END DO
            END DO
          END DO
        END DO

      END IF

      CALL distribute_field(Tmax_glob(:,:,2),ie_tot,je_tot,Tmax_i,ie,je,0,ierr)


      ! 2) Tmax for snow:


      CALL gather_field(Tmax_s, ie, je, Tmax_glob(:,:,1), ie_tot, je_tot, 0, ierr)

      Tmax_glob(:,:,2) = Tmax_glob(:,:,1)

      IF (my_cart_id == 0) THEN
        
        DO j=-neigh_x,neigh_x
          DO i=-neigh_x,neigh_x
            DO jm=1,je_tot
              DO im=1,ie_tot
                
                iu = MIN(MAX(im+i,1),ie_tot)
                ju = MIN(MAX(jm+j,1),je_tot)

                Tmax_glob(im,jm,2) = MAX(Tmax_glob(iu,ju,1), Tmax_glob(im,jm,2))               
                
              END DO
            END DO
          END DO
        END DO

      END IF  

      CALL distribute_field(Tmax_glob(:,:,2),ie_tot,je_tot,Tmax_s,ie,je,0,ierr)


      ! 3) Tmax for graupel:

      CALL gather_field(Tmax_g, ie, je, Tmax_glob(:,:,1), ie_tot, je_tot, 0, ierr)

      Tmax_glob(:,:,2) = Tmax_glob(:,:,1)

      IF (my_cart_id == 0) THEN
        
        DO j=-neigh_x,neigh_x
          DO i=-neigh_x,neigh_x
            DO jm=1,je_tot
              DO im=1,ie_tot
                
                iu = MIN(MAX(im+i,1),ie_tot)
                ju = MIN(MAX(jm+j,1),je_tot)

                Tmax_glob(im,jm,2) = MAX(Tmax_glob(iu,ju,1), Tmax_glob(im,jm,2))               
                
              END DO
            END DO
          END DO
        END DO

      END IF

      CALL distribute_field(Tmax_glob(:,:,2),ie_tot,je_tot,Tmax_g,ie,je,0,ierr)

      ! 4) Tmax for hail:

      IF (ALLOCATED(qh)) THEN

        CALL gather_field(Tmax_h, ie, je, Tmax_glob(:,:,1), ie_tot, je_tot, 0, ierr)
        
        Tmax_glob(:,:,2) = Tmax_glob(:,:,1)
        
        IF (my_cart_id == 0) THEN
          
          DO j=-neigh_x,neigh_x
            DO i=-neigh_x,neigh_x
              DO jm=1,je_tot
                DO im=1,ie_tot
                  
                  iu = MIN(MAX(im+i,1),ie_tot)
                  ju = MIN(MAX(jm+j,1),je_tot)
                  
                  Tmax_glob(im,jm,2) = MAX(Tmax_glob(iu,ju,1), Tmax_glob(im,jm,2))               
                  
                END DO
              END DO
            END DO
          END DO
          
        END IF

      END IF

      CALL distribute_field(Tmax_glob(:,:,2),ie_tot,je_tot,Tmax_h,ie,je,0,ierr)

      ! clean up:
      DEALLOCATE(Tmax_glob)

    END IF ! if (num_compute > 1)

    RETURN
  END SUBROUTINE initialize_tmax_sb_vec_par

  ! Vorsicht: nicht parallelisiert! Tmax-Umgebung nur innerhalb eines Prozessor-Gebietes!
  SUBROUTINE initialize_tmax_sb_vec2(ntime, neigh)

    IMPLICIT NONE
    INTEGER, INTENT(in) :: ntime
    DOUBLE PRECISION, INTENT(in) :: neigh

    INTEGER :: ju, i, j, k, neigh_x, neigh_y, im, jm
    DOUBLE PRECISION :: dx, dy
    DOUBLE PRECISION :: qithresh, qsthresh, qgthresh, qhthresh, thilf(1:ie), qhilf(1:ie), qnhilf(1:ie)
    LOGICAL :: flagge(1:ie)

    ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
    qithresh = 1d-8
    qsthresh = 1d-8
    qgthresh = 1d-8

    ALLOCATE(Tmax_i(ie,je),Tmax_s(ie,je),Tmax_g(ie,je))

    ! Initialwerte von Tmax_x: Diese Werte werden mindestens genommen:
    Tmax_i = 3.0d0 + 273.16
    Tmax_s = 3.0d0 + 273.16
    Tmax_g = 3.0d0 + 273.16

    dx = dlon * r_earth * pi / 180.0 * COS((startlat_tot+(je_tot/2)*dlat)*pi/180.0d0)
    dy = dlat * r_earth * pi / 180.0

    neigh_x = MAX(INT(neigh/dx), 0)
    neigh_y = MAX(INT(neigh/dy), 0)

    !!! neigh_x, neigh_y have to be smaller than the sub-processor grid size,
    !!! since a cross-subgrid search is not implemented so far!
    !!! Further, it is limited to 5 gridpoints to limit computational burden,
    !!! which makes also sense physically.
    IF (neigh_x > 5 .OR. neigh_x > iup - ilow) THEN
      neigh_x = MAX(MIN(MIN(neigh_x, iup - ilow ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_SB_VEC2: neigh_x has been clipped to ', neigh_x, &
           ' gridpoints or ', neigh_x*dx*1e-3, ' km'
    END IF
    IF (neigh_y > 5 .OR. neigh_y > jup - jlow) THEN
      neigh_y = MAX(MIN(MIN(neigh_y, jup - jlow ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_SB_VEC2: neigh_y has been clipped to ', neigh_y, &
           ' gridpoints or ', neigh_y*dy*1e-3, ' km'
    END IF

    ! Tmax wird aus der neigh-Umgebung des Punktes im,jm als hoechste Temperatur bestimmt,
    ! bei der noch ein nennenswerter Wassergehalt (> qthresh) der Spezies q in der Umgebung neigh vorkommt. Auf
    ! diese Weise bestimmt die Schmelzparametrisierung der Wolkenmikrophysik in 
    ! entscheidender Weise das Verhalten des hier parametrisierten Schmelzgrades.

    DO k=1,ke
      DO jm=jlow,jup
        DO j=-neigh_x,neigh_x
          ju = MIN(MAX(jm+j,1),je)
          DO i=-neigh_x,neigh_x
            IF ( i < 0 ) THEN
              qhilf (iup+i+1:iup)   = -999.9
              qnhilf(iup+i+1:iup)   = -999.9
              thilf (iup+i+1:iup)   = -999.9
              qhilf (ilow:iup+i)   = qi (ilow:iup+i,ju,k,ntime)
              qnhilf(ilow:iup+i)   = qni(ilow:iup+i,ju,k,ntime)
              thilf (ilow:iup+i)   = t  (ilow:iup+i,ju,k,ntime)
            ELSE IF ( i == 0 ) THEN
              qhilf (ilow:iup)   = qi (ilow:iup,ju,k,ntime)
              qnhilf(ilow:iup)   = qni(ilow:iup,ju,k,ntime)
              thilf (ilow:iup)   = t  (ilow:iup,ju,k,ntime)
            ELSE
              qhilf (ilow:ilow+i-1)   = -999.9
              qnhilf(ilow:ilow+i-1)   = -999.9
              thilf (ilow:ilow+i-1)   = -999.9
              qhilf (ilow+i:iup)       = qi (ilow+i:iup,ju,k,ntime)
              qnhilf(ilow+i:iup)       = qni(ilow+i:iup,ju,k,ntime)
              thilf (ilow+i:iup)       = t  (ilow+i:iup,ju,k,ntime)
            END IF
            
            flagge(ilow:iup) = ( qhilf(ilow:iup) > qithresh .AND. &
                                         qnhilf(ilow:iup) > 1.0d0 )
            
            DO im=ilow,iup
              IF (flagge(im)) THEN
                Tmax_i(im,jm) = MAX(Tmax_i(im,jm), thilf(im))
              END IF
            END DO

          END DO
        END DO

      END DO
    END DO

    DO k=1,ke
      DO jm=jlow,jup
        DO j=-neigh_x,neigh_x
          ju = MIN(MAX(jm+j,1),je)
          DO i=-neigh_x,neigh_x
            IF ( i < 0 ) THEN
              qhilf (iup+i+1:iup)   = -999.9
              qnhilf(iup+i+1:iup)   = -999.9
              thilf (iup+i+1:iup)   = -999.9
              qhilf (ilow:iup+i)   = qs (ilow:iup+i,ju,k,ntime)
              qnhilf(ilow:iup+i)   = qns(ilow:iup+i,ju,k,ntime)
              thilf (ilow:iup+i)   = t  (ilow:iup+i,ju,k,ntime)
            ELSE IF ( i == 0 ) THEN
              qhilf (ilow:iup)   = qs (ilow:iup,ju,k,ntime)
              qnhilf(ilow:iup)   = qns(ilow:iup,ju,k,ntime)
              thilf (ilow:iup)   = t  (ilow:iup,ju,k,ntime)
            ELSE
              qhilf (ilow:ilow+i-1)   = -999.9
              qnhilf(ilow:ilow+i-1)   = -999.9
              thilf (ilow:ilow+i-1)   = -999.9
              qhilf (ilow+i:iup)       = qs (ilow+i:iup,ju,k,ntime)
              qnhilf(ilow+i:iup)       = qns(ilow+i:iup,ju,k,ntime)
              thilf (ilow+i:iup)       = t  (ilow+i:iup,ju,k,ntime)
            END IF
            
            flagge(ilow:iup) = ( qhilf(ilow:iup) > qsthresh .AND. &
                                         qnhilf(ilow:iup) > 1.0d0 )
            
            DO im=ilow,iup
              IF (flagge(im)) THEN
                Tmax_s(im,jm) = MAX(Tmax_s(im,jm), thilf(im))
              END IF
            END DO

          END DO
        END DO

      END DO
    END DO

    DO k=1,ke
      DO jm=jlow,jup
        DO j=-neigh_x,neigh_x
          ju = MIN(MAX(jm+j,1),je)
          DO i=-neigh_x,neigh_x
            IF ( i < 0 ) THEN
              qhilf (iup+i+1:iup)   = -999.9
              qnhilf(iup+i+1:iup)   = -999.9
              thilf (iup+i+1:iup)   = -999.9
              qhilf (ilow:iup+i)   = qg (ilow:iup+i,ju,k,ntime)
              qnhilf(ilow:iup+i)   = qng(ilow:iup+i,ju,k,ntime)
              thilf (ilow:iup+i)   = t  (ilow:iup+i,ju,k,ntime)
            ELSE IF ( i == 0 ) THEN
              qhilf (ilow:iup)   = qg (ilow:iup,ju,k,ntime)
              qnhilf(ilow:iup)   = qng(ilow:iup,ju,k,ntime)
              thilf (ilow:iup)   = t  (ilow:iup,ju,k,ntime)
            ELSE
              qhilf (ilow:ilow+i-1)   = -999.9
              qnhilf(ilow:ilow+i-1)   = -999.9
              thilf (ilow:ilow+i-1)   = -999.9
              qhilf (ilow+i:iup)       = qg (ilow+i:iup,ju,k,ntime)
              qnhilf(ilow+i:iup)       = qng(ilow+i:iup,ju,k,ntime)
              thilf (ilow+i:iup)       = t  (ilow+i:iup,ju,k,ntime)
            END IF
            
            flagge(ilow:iup) = ( qhilf(ilow:iup) > qgthresh .AND. &
                 qnhilf(ilow:iup) > 1.0d-3 )
            
            DO im=ilow,iup
              IF (flagge(im)) THEN
                Tmax_g(im,jm) = MAX(Tmax_g(im,jm), thilf(im))
              END IF
            END DO

          END DO
        END DO

      END DO
    END DO


    IF (ALLOCATED(qh)) THEN

      qhthresh = 1d-8

      ALLOCATE(Tmax_h(ie,je))

      Tmax_h = 3.0d0 + 273.16

      DO k=1,ke
        DO jm=jlow,jup
          DO j=-neigh_x,neigh_x
            ju = MIN(MAX(jm+j,1),je)
            DO i=-neigh_x,neigh_x
              IF ( i < 0 ) THEN
                qhilf (iup+i+1:iup)   = -999.9
                qnhilf(iup+i+1:iup)   = -999.9
                thilf (iup+i+1:iup)   = -999.9
                qhilf (ilow:iup+i)   = qh (ilow:iup+i,ju,k,ntime)
                qnhilf(ilow:iup+i)   = qnh(ilow:iup+i,ju,k,ntime)
                thilf (ilow:iup+i)   = t  (ilow:iup+i,ju,k,ntime)
              ELSE IF ( i == 0 ) THEN
                qhilf (ilow:iup)   = qh (ilow:iup,ju,k,ntime)
                qnhilf(ilow:iup)   = qnh(ilow:iup,ju,k,ntime)
                thilf (ilow:iup)   = t  (ilow:iup,ju,k,ntime)
              ELSE
                qhilf (ilow:ilow+i-1)   = -999.9
                qnhilf(ilow:ilow+i-1)   = -999.9
                thilf (ilow:ilow+i-1)   = -999.9
                qhilf (ilow+i:iup)       = qh (ilow+i:iup,ju,k,ntime)
                qnhilf(ilow+i:iup)       = qnh(ilow+i:iup,ju,k,ntime)
                thilf (ilow+i:iup)       = t  (ilow+i:iup,ju,k,ntime)
              END IF
              
              flagge(ilow:iup) = ( qhilf(ilow:iup) > qhthresh .AND. &
                                           qnhilf(ilow:iup) > 1.0d-3 )
              
              DO im=ilow,iup
                IF (flagge(im)) THEN
                  Tmax_h(im,jm) = MAX(Tmax_h(im,jm), thilf(im))
                END IF
              END DO
              
            END DO
          END DO
          
        END DO
      END DO

    END IF

    RETURN
  END SUBROUTINE initialize_tmax_sb_vec2

  ! Version fuer die 1M-Schemata:
  SUBROUTINE initialize_tmax_vec(ntime, neigh)

    IMPLICIT NONE
    INTEGER, INTENT(in) :: ntime
    DOUBLE PRECISION, INTENT(in) :: neigh

    INTEGER :: iu, ju, i, j, k, neigh_x, neigh_y, im, jm
    DOUBLE PRECISION :: dx, dy
    DOUBLE PRECISION :: qithresh, qsthresh, qgthresh


    dx = dlon * r_earth * pi / 180.0 * COS((startlat_tot+(je_tot/2)*dlat)*pi/180.0d0)
    dy = dlat * r_earth * pi / 180.0

    neigh_x = MAX(INT(neigh/dx), 0)
    neigh_y = MAX(INT(neigh/dy), 0)

    !!! neigh_x, neigh_y have to be smaller than the sub-processor grid size,
    !!! since a cross-subgrid search is not implemented so far!
    !!! Further, it is limited to 5 gridpoints to limit computational burden,
    !!! which makes also sense physically.
    IF (neigh_x > 5 .OR. neigh_x > iup - ilow) THEN
      neigh_x = MAX(MIN(MIN(neigh_x, iup - ilow ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_VEC: neigh_x has been clipped to ', neigh_x, &
           ' gridpoints or ', neigh_x*dx*1e-3, ' km'
    END IF
    IF (neigh_y > 5 .OR. neigh_y > jup - jlow) THEN
      neigh_y = MAX(MIN(MIN(neigh_y, jup - jlow ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_VEC: neigh_y has been clipped to ', neigh_y, &
           ' gridpoints or ', neigh_y*dy*1e-3, ' km'
    END IF

    ! Tmax wird aus der neigh-Umgebung des Punktes im,jm als hoechste Temperatur bestimmt,
    ! bei der noch ein nennenswerter Wassergehalt (> qthresh) der Spezies q in der Umgebung neigh vorkommt. Auf
    ! diese Weise bestimmt die Schmelzparametrisierung der Wolkenmikrophysik in 
    ! entscheidender Weise das Verhalten des hier parametrisierten Schmelzgrades.
    
    IF (ALLOCATED(qi)) THEN

      ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
      qithresh = 1d-8
      
      ALLOCATE(Tmax_i(ie,je))
      
      ! Initialwert von Tmax_i: Dieser Wert wird mindestens genommen:
      Tmax_i = 3.0d0 + 273.16
      
      DO j=-neigh_x,neigh_x
        DO i=-neigh_x,neigh_x
          
          DO k=1,ke
            DO jm=1,je
              DO im=1,ie
                
                iu = MIN(MAX(im+i,1),ie)
                ju = MIN(MAX(jm+j,1),je)

                IF (qi(iu,ju,k,ntime) >= qithresh) THEN
                  Tmax_i(im,jm) = MAX(Tmax_i(im,jm), t(iu,ju,k,ntime))
                END IF
              END DO
            END DO
          END DO
        END DO
      END DO
    END IF

    IF (ALLOCATED(qs)) THEN

      qsthresh = 1d-8

      ALLOCATE(Tmax_s(ie,je))

      Tmax_s = 3.0d0 + 273.16 ! Dieser Wert wird mindestens genommen:

      DO j=-neigh_x,neigh_x
        DO i=-neigh_x,neigh_x
          
          DO k=1,ke
            DO jm=1,je
              DO im=1,ie
                
                iu = MIN(MAX(im+i,1),ie)
                ju = MIN(MAX(jm+j,1),je)

                IF (qs(iu,ju,k,ntime) >= qsthresh) THEN
                  Tmax_s(im,jm) = MAX(Tmax_s(im,jm), t(iu,ju,k,ntime))
                END IF
              END DO
            END DO
          END DO
        END DO
      END DO

    END IF

    IF (ALLOCATED(qg)) THEN

      qgthresh = 1d-8

      ALLOCATE(Tmax_g(ie,je))

      Tmax_g = 3.0d0 + 273.16 ! Dieser Wert wird mindestens genommen:

      DO j=-neigh_x,neigh_x
        DO i=-neigh_x,neigh_x
          
          DO k=1,ke
            DO jm=1,je
              DO im=1,ie
                
                iu = MIN(MAX(im+i,1),ie)
                ju = MIN(MAX(jm+j,1),je)

                IF (qg(iu,ju,k,ntime) >= qgthresh) THEN
                  Tmax_g(im,jm) = MAX(Tmax_g(im,jm), t(iu,ju,k,ntime))
                END IF
              END DO
            END DO
          END DO
        END DO
      END DO

    END IF

    RETURN
  END SUBROUTINE initialize_tmax_vec

  ! parallelisiert (bisher Tmax-Umgebung nur innerhalb eines Prozessor-Gebietes!):
  SUBROUTINE initialize_tmax_vec_par(ntime, neigh)

    IMPLICIT NONE
    INTEGER, INTENT(in) :: ntime
    DOUBLE PRECISION, INTENT(in) :: neigh

    INTEGER :: iu, ju, i, j, k, neigh_x, neigh_y, im, jm, ierr
    DOUBLE PRECISION :: dx, dy
    DOUBLE PRECISION :: qithresh, qsthresh, qgthresh, qhthresh

    DOUBLE PRECISION, ALLOCATABLE :: Tmax_glob(:,:,:)

    dx = dlon * r_earth * pi / 180.0 * COS((startlat_tot+(je_tot/2)*dlat)*pi/180.0d0)
    dy = dlat * r_earth * pi / 180.0

    neigh_x = MAX(INT(neigh/dx), 0)
    neigh_y = MAX(INT(neigh/dy), 0)

    !!! neigh_x, neigh_y is limited to 10 (5 on either side) gridpoints to limit computational burden,
    !!! which makes also sense physically.
    IF (neigh_x > 5) THEN
      neigh_x = 5
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_VEC_PAR: neigh_x has been clipped to ', neigh_x, &
           ' gridpoints or ', neigh_x*dx*1e-3, ' km'
    END IF
    IF (neigh_y > 5) THEN
      neigh_y = 5
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_VEC_PAR: neigh_y has been clipped to ', neigh_y, &
           ' gridpoints or ', neigh_y*dy*1e-3, ' km'
    END IF


    ! Tmax wird aus der neigh-Umgebung des Punktes im,jm als hoechste Temperatur bestimmt,
    ! bei der noch ein nennenswerter Wassergehalt (> qthresh) der Spezies q in der Umgebung neigh vorkommt. Auf
    ! diese Weise bestimmt die Schmelzparametrisierung der Wolkenmikrophysik in 
    ! entscheidender Weise das Verhalten des hier parametrisierten Schmelzgrades.

    ! First, compute max. Temperature where hydrometeors are present within each vertical column:
    IF (ALLOCATED(qi)) THEN

      ALLOCATE(Tmax_i(ie,je))

      ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
      qithresh = 1d-8

      ! Initialwerte von Tmax_x: Diese Werte werden mindestens genommen:
      Tmax_i = 3.0d0 + 273.16

      DO k=1,ke
        DO jm=1,je
          DO im=1,ie
          
            IF (qi(im,jm,k,ntime) >= qithresh) THEN
              Tmax_i(im,jm) = MAX(Tmax_i(im,jm), t(im,jm,k,ntime))
            END IF

          END DO
        END DO
      END DO

    END IF

    IF (ALLOCATED(qs)) THEN

      ALLOCATE(Tmax_s(ie,je))
      
      ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
      qsthresh = 1d-8

      ! Initialwerte von Tmax_x: Diese Werte werden mindestens genommen:
      Tmax_s = 3.0d0 + 273.16

      DO k=1,ke
        DO jm=1,je
          DO im=1,ie
          
            IF (qs(im,jm,k,ntime) >= qsthresh) THEN
              Tmax_s(im,jm) = MAX(Tmax_s(im,jm), t(im,jm,k,ntime))
            END IF

          END DO
        END DO
      END DO
      
    END IF

    IF (ALLOCATED(qg)) THEN
      
      ALLOCATE(Tmax_g(ie,je))

      ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
      qgthresh = 1d-8

      ! Initialwerte von Tmax_x: Diese Werte werden mindestens genommen:
      Tmax_g = 3.0d0 + 273.16

      DO k=1,ke
        DO jm=1,je
          DO im=1,ie
          
            IF (qg(im,jm,k,ntime) >= qgthresh) THEN
              Tmax_g(im,jm) = MAX(Tmax_g(im,jm), t(im,jm,k,ntime))
            END IF
            
          END DO
        END DO
      END DO
      
    END IF

    ! In case of parallel run, collect the vertical maxima on the root node,
    ! compute the maxima over the neigh_x/neigh_y-Region and redistribute
    ! the global field to the single nodes:
    IF (num_compute > 1) THEN

      IF (ALLOCATED(qi) .OR. ALLOCATED(qs) .OR. ALLOCATED(qg)) THEN
        ALLOCATE(Tmax_glob(ie_tot,je_tot,2))
      END IF

      ! 1) Tmax for cloud ice:
      IF (ALLOCATED(qi)) THEN

        CALL gather_field(Tmax_i, ie, je, Tmax_glob(:,:,1), ie_tot, je_tot, 0, ierr)
        
        Tmax_glob(:,:,2) = Tmax_glob(:,:,1)
        
        IF (my_cart_id == 0) THEN
          
          DO j=-neigh_x,neigh_x
            DO i=-neigh_x,neigh_x
              DO jm=1,je_tot
                DO im=1,ie_tot
                  
                  iu = MIN(MAX(im+i,1),ie_tot)
                  ju = MIN(MAX(jm+j,1),je_tot)
                  
                  Tmax_glob(im,jm,2) = MAX(Tmax_glob(iu,ju,1), Tmax_glob(im,jm,2))               
                  
                END DO
              END DO
            END DO
          END DO
          
        END IF
        
        CALL distribute_field(Tmax_glob(:,:,2),ie_tot,je_tot,Tmax_i,ie,je,0,ierr)

      END IF

      ! 2) Tmax for snow:
      IF (ALLOCATED(qs)) THEN
        
        CALL gather_field(Tmax_s, ie, je, Tmax_glob(:,:,1), ie_tot, je_tot, 0, ierr)
        
        Tmax_glob(:,:,2) = Tmax_glob(:,:,1)
        
        IF (my_cart_id == 0) THEN
          
          DO j=-neigh_x,neigh_x
            DO i=-neigh_x,neigh_x
              DO jm=1,je_tot
                DO im=1,ie_tot
                  
                  iu = MIN(MAX(im+i,1),ie_tot)
                  ju = MIN(MAX(jm+j,1),je_tot)
                  
                  Tmax_glob(im,jm,2) = MAX(Tmax_glob(iu,ju,1), Tmax_glob(im,jm,2))               
                  
                END DO
              END DO
            END DO
          END DO
          
        END IF
        
        CALL distribute_field(Tmax_glob(:,:,2),ie_tot,je_tot,Tmax_s,ie,je,0,ierr)
        
      END IF

      ! 3) Tmax for graupel:
      
      IF (ALLOCATED(qg)) THEN

        CALL gather_field(Tmax_g, ie, je, Tmax_glob(:,:,1), ie_tot, je_tot, 0, ierr)

        Tmax_glob(:,:,2) = Tmax_glob(:,:,1)
        
        IF (my_cart_id == 0) THEN
          
          DO j=-neigh_x,neigh_x
            DO i=-neigh_x,neigh_x
              DO jm=1,je_tot
                DO im=1,ie_tot
                  
                  iu = MIN(MAX(im+i,1),ie_tot)
                  ju = MIN(MAX(jm+j,1),je_tot)
                  
                  Tmax_glob(im,jm,2) = MAX(Tmax_glob(iu,ju,1), Tmax_glob(im,jm,2))               
                  
                END DO
              END DO
            END DO
          END DO
          
        END IF

        CALL distribute_field(Tmax_glob(:,:,2),ie_tot,je_tot,Tmax_g,ie,je,0,ierr)

      END IF

      ! clean up:
      IF (ALLOCATED(Tmax_glob)) DEALLOCATE(Tmax_glob)

    END IF ! if (num_compute > 1)

    RETURN
  END SUBROUTINE initialize_tmax_vec_par

  SUBROUTINE initialize_tmax_vec2(ntime, neigh)

    IMPLICIT NONE
    INTEGER, INTENT(in) :: ntime
    DOUBLE PRECISION, INTENT(in) :: neigh

    INTEGER :: ju, i, j, k, neigh_x, neigh_y, im, jm
    DOUBLE PRECISION :: dx, dy
    DOUBLE PRECISION :: qithresh, qsthresh, qgthresh, thilf(1:ie), qhilf(1:ie)
    LOGICAL :: flagge(1:ie)


    dx = dlon * r_earth * pi / 180.0 * COS((startlat_tot+(je_tot/2)*dlat)*pi/180.0d0)
    dy = dlat * r_earth * pi / 180.0

    neigh_x = MAX(INT(neigh/dx), 0)
    neigh_y = MAX(INT(neigh/dy), 0)

    !!! neigh_x, neigh_y have to be smaller than the sub-processor grid size,
    !!! since a cross-subgrid search is not implemented so far!
    !!! Further, it is limited to 5 gridpoints to limit computational burden,
    !!! which makes also sense physically.
    IF (neigh_x > 5 .OR. neigh_x > iup - ilow) THEN
      neigh_x = MAX(MIN(MIN(neigh_x, iup - ilow ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_VEC2: neigh_x has been clipped to ', neigh_x, &
           ' gridpoints or ', neigh_x*dx*1e-3, ' km'
    END IF
    IF (neigh_y > 5 .OR. neigh_y > jup - jlow) THEN
      neigh_y = MAX(MIN(MIN(neigh_y, jup - jlow ), 5), 0)
      WRITE(*,'(a,i5,a,f9.2,a)') ' *** INITIALIZE_TMAX_VEC2: neigh_y has been clipped to ', neigh_y, &
           ' gridpoints or ', neigh_y*dy*1e-3, ' km'
    END IF

    ! Tmax wird aus der neigh-Umgebung des Punktes im,jm als hoechste Temperatur bestimmt,
    ! bei der noch ein nennenswerter Wassergehalt (> qthresh) der Spezies q in der Umgebung neigh vorkommt. Auf
    ! diese Weise bestimmt die Schmelzparametrisierung der Wolkenmikrophysik in 
    ! entscheidender Weise das Verhalten des hier parametrisierten Schmelzgrades.
    
    IF (ALLOCATED(qi)) THEN

      ! Q - Thresholds fuer die Bestimmung der Tmax-Temperatur fuer die Schmelzgradparametrisierung
      qithresh = 1d-8
      
      ALLOCATE(Tmax_i(ie,je))
      
      ! Initialwert von Tmax_i: Dieser Wert wird mindestens genommen:
      Tmax_i = 3.0d0 + 273.16
      
      DO k=1,ke
        DO jm=jlow,jup
          DO j=-neigh_x,neigh_x
            ju = MIN(MAX(jm+j,1),je)
            DO i=-neigh_x,neigh_x
              IF ( i < 0 ) THEN
                qhilf (iup+i+1:iup)   = -999.9
                thilf (iup+i+1:iup)   = -999.9
                qhilf (ilow:iup+i)   = qi (ilow:iup+i,ju,k,ntime)
                thilf (ilow:iup+i)   = t  (ilow:iup+i,ju,k,ntime)
              ELSE IF ( i == 0 ) THEN
                qhilf (ilow:iup)   = qi (ilow:iup,ju,k,ntime)
                thilf (ilow:iup)   = t  (ilow:iup,ju,k,ntime)
              ELSE
                qhilf (ilow:ilow+i-1)   = -999.9
                thilf (ilow:ilow+i-1)   = -999.9
                qhilf (ilow+i:iup)       = qi (ilow+i:iup,ju,k,ntime)
                thilf (ilow+i:iup)       = t  (ilow+i:iup,ju,k,ntime)
              END IF
              
              flagge(ilow:iup) = ( qhilf(ilow:iup) >= qithresh )
              
              DO im=ilow,iup
                IF (flagge(im)) THEN
                  Tmax_i(im,jm) = MAX(Tmax_i(im,jm), thilf(im))
                END IF
              END DO
              
            END DO
          END DO
          
        END DO
      END DO

    END IF

    IF (ALLOCATED(qs)) THEN

      qsthresh = 1d-8

      ALLOCATE(Tmax_s(ie,je))

      Tmax_s = 3.0d0 + 273.16 ! Dieser Wert wird mindestens genommen:

      DO k=1,ke
        DO jm=jlow,jup
          DO j=-neigh_x,neigh_x
            ju = MIN(MAX(jm+j,1),je)
            DO i=-neigh_x,neigh_x
              IF ( i < 0 ) THEN
                qhilf (iup+i+1:iup)   = -999.9
                thilf (iup+i+1:iup)   = -999.9
                qhilf (ilow:iup+i)   = qs (ilow:iup+i,ju,k,ntime)
                thilf (ilow:iup+i)   = t  (ilow:iup+i,ju,k,ntime)
              ELSE IF ( i == 0 ) THEN
                qhilf (ilow:iup)   = qs (ilow:iup,ju,k,ntime)
                thilf (ilow:iup)   = t  (ilow:iup,ju,k,ntime)
              ELSE
                qhilf (ilow:ilow+i-1)   = -999.9
                thilf (ilow:ilow+i-1)   = -999.9
                qhilf (ilow+i:iup)       = qs (ilow+i:iup,ju,k,ntime)
                thilf (ilow+i:iup)       = t  (ilow+i:iup,ju,k,ntime)
              END IF
              
              flagge(ilow:iup) = ( qhilf(ilow:iup) >= qsthresh )
              
              DO im=ilow,iup
                IF (flagge(im)) THEN
                  Tmax_s(im,jm) = MAX(Tmax_s(im,jm), thilf(im))
                END IF
              END DO
              
            END DO
          END DO
          
        END DO
      END DO

    END IF

    IF (ALLOCATED(qg)) THEN

      qgthresh = 1d-8

      ALLOCATE(Tmax_g(ie,je))

      Tmax_g = 3.0d0 + 273.16 ! Dieser Wert wird mindestens genommen:

      DO k=1,ke
        DO jm=jlow,jup
          DO j=-neigh_x,neigh_x
            ju = MIN(MAX(jm+j,1),je)
            DO i=-neigh_x,neigh_x
              IF ( i < 0 ) THEN
                qhilf (iup+i+1:iup)   = -999.9
                thilf (iup+i+1:iup)   = -999.9
                qhilf (ilow:iup+i)   = qg (ilow:iup+i,ju,k,ntime)
                thilf (ilow:iup+i)   = t  (ilow:iup+i,ju,k,ntime)
              ELSE IF ( i == 0 ) THEN
                qhilf (ilow:iup)   = qg (ilow:iup,ju,k,ntime)
                thilf (ilow:iup)   = t  (ilow:iup,ju,k,ntime)
              ELSE
                qhilf (ilow:ilow+i-1)   = -999.9
                thilf (ilow:ilow+i-1)   = -999.9
                qhilf (ilow+i:iup)       = qg (ilow+i:iup,ju,k,ntime)
                thilf (ilow+i:iup)       = t  (ilow+i:iup,ju,k,ntime)
              END IF
              
              flagge(ilow:iup) = ( qhilf(ilow:iup) >= qgthresh )
              
              DO im=ilow,iup
                IF (flagge(im)) THEN
                  Tmax_g(im,jm) = MAX(Tmax_g(im,jm), thilf(im))
                END IF
              END DO
              
            END DO
          END DO
          
        END DO
      END DO

    END IF

    RETURN
  END SUBROUTINE initialize_tmax_vec2


  ! Aufraeumen:
  SUBROUTINE finalize_tmax()

    IMPLICIT NONE

    IF (ALLOCATED(Tmax_i)) DEALLOCATE(Tmax_i)
    IF (ALLOCATED(Tmax_s)) DEALLOCATE(Tmax_s)
    IF (ALLOCATED(Tmax_g)) DEALLOCATE(Tmax_g)
    IF (ALLOCATED(Tmax_h)) DEALLOCATE(Tmax_h)

    RETURN
  END SUBROUTINE finalize_tmax

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_ha_sb_vec(ims,ime,jm,km,ntime,anzt,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anz, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt, anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN( hail%a_geo * MIN(MAX(qh(ims:ime,jm,km,ntime)/&
         (qnh(ims:ime,jm,km,ntime)+1d-20),hail%x_min),hail%x_max)**hail%b_geo, 6d-3)
    Dexpo = 0.8d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_h(i,jm), 293.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec(tvec(ims:ime),ime-ims+1,Dvec,anz,Dref(ims:ime),&
         Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime,:))

    RETURN
  END SUBROUTINE degree_of_melting_ha_sb_vec

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_gr_sb_vec(ims,ime,jm,km,ntime,anzt,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anz, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt, anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN( graupel%a_geo * MIN(MAX(qg(ims:ime,jm,km,ntime)/&
         (qng(ims:ime,jm,km,ntime)+1d-20),graupel%x_min),graupel%x_max)**graupel%b_geo, 6d-3)
    Dexpo = 0.6d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_g(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec(tvec(ims:ime),ime-ims+1,Dvec,anz,Dref(ims:ime),&
         Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime,:))

    RETURN
  END SUBROUTINE degree_of_melting_gr_sb_vec

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_sn_sb_vec(ims,ime,jm,km,ntime,anzt,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anz, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt, anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN( snow%a_geo * MIN(MAX(qs(ims:ime,jm,km,ntime)/&
         (qns(ims:ime,jm,km,ntime)+1d-20),snow%x_min),snow%x_max)**snow%b_geo, 6d-3)
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_s(i,jm), 276.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec(tvec(ims:ime),ime-ims+1,Dvec,anz,Dref(ims:ime),&
         Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime,:))

    RETURN
  END SUBROUTINE degree_of_melting_sn_sb_vec

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_ice_sb_vec(ims,ime,jm,km,ntime,anzt,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anz, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt, anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN( ice%a_geo * MIN(MAX(qi(ims:ime,jm,km,ntime)/&
         (qni(ims:ime,jm,km,ntime)+1d-20),ice%x_min),ice%x_max)**ice%b_geo, 3d-3)
    Dexpo = 1.0d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_i(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec(tvec(ims:ime),ime-ims+1,Dvec,anz,Dref(ims:ime),&
         Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime,:))

    RETURN
  END SUBROUTINE degree_of_melting_ice_sb_vec

  ! Version fuer die 1M-Schemata:
  SUBROUTINE degree_of_melting_gr_vec(ims,ime,jm,km,ntime,anzt,Dvec,anz,&
       Tmeltbegin,meltdegTmin,n0_g,a_geo_g,b_geo_g,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anz, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt, anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz), n0_g, a_geo_g, b_geo_g

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_1M_exp (qg(im,jm,km,ntime), a_geo_g, b_geo_g, n0_g)
!    Dref(ims:ime) = MIN(D_of_X_average_1M_exp_vec (qg(ims:ime,jm,km,ntime), a_geo_g, b_geo_g, n0_g, ime-ims+1), 6d-3)
     Dref(ims:ime) = MIN(D_of_X_average_1M_exp_vec (qg(ims:ime,jm,km,ntime)*rho(ims:ime,jm,km), &
          a_geo_g, b_geo_g, n0_g, ime-ims+1), 6d-3)
    Dexpo = 0.6d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_g(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec(tvec(ims:ime),ime-ims+1,Dvec,anz,Dref(ims:ime),&
         Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime,:))

    RETURN
  END SUBROUTINE degree_of_melting_gr_vec

  ! Version fuer die 1M-Schemata:
  SUBROUTINE degree_of_melting_sn_vec(ims,ime,jm,km,ntime,anzt,Dvec,anz,&
       Tmeltbegin,meltdegTmin,n0_s,a_geo_s,b_geo_s,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anz, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt, anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, n0_s, Dvec(anz), a_geo_s, b_geo_s

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN(D_of_X_average_1M_exp_vec (qs(ims:ime,jm,km,ntime)*rho(ims:ime,jm,km), &
         a_geo_s, b_geo_s, n0_s, ims-ime+1), 6d-3)
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_s(i,jm), 276.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec(tvec(ims:ime),ime-ims+1,Dvec,anz,Dref(ims:ime),&
         Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime,:))

    RETURN
  END SUBROUTINE degree_of_melting_sn_vec

  ! Version fuer die 1M-Schemata:
  SUBROUTINE degree_of_melting_ice_vec(ims,ime,jm,km,ntime,anzt,Dvec,anz,Tmeltbegin,meltdegTmin,Dref_in,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anz, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt, anz)
    DOUBLE PRECISION, INTENT(in) :: Dref_in, Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf 0.1 mm gesetzt. --- nein, muss extern vorgegeben werden!
    ! Dref = 1d-4
    Dref = Dref_in
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_i(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec(tvec(ims:ime),ime-ims+1,Dvec,anz,Dref(ims:ime),&
         Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime,:))

    RETURN
  END SUBROUTINE degree_of_melting_ice_vec

  !=============================================================================================
  ! Die Routinen fuer einen Wert je Gitterpunkt (1 Durchmesser je Gitterpunkt)
  !=============================================================================================

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_ha_sb_vec_ie(ims,ime,jm,km,ntime,Dvec,anzt,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anzt)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN( hail%a_geo * MIN(MAX(qh(ims:ime,jm,km,ntime)/&
         (qnh(ims:ime,jm,km,ntime)+1d-20),hail%x_min),hail%x_max)**hail%b_geo, 6d-3)
    Dexpo = 0.8d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_h(i,jm), 293.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec_ie(tvec(ims:ime),Dvec(ims:ime),ime-ims+1,&
         Dref(ims:ime),Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime))

    RETURN
  END SUBROUTINE degree_of_melting_ha_sb_vec_ie

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_gr_sb_vec_ie(ims,ime,jm,km,ntime,Dvec,anzt,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anzt)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN( graupel%a_geo * MIN(MAX(qg(ims:ime,jm,km,ntime)/&
         (qng(ims:ime,jm,km,ntime)+1d-20),graupel%x_min),graupel%x_max)**graupel%b_geo, 6d-3)
    Dexpo = 0.6d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_g(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec_ie(tvec(ims:ime),Dvec(ims:ime),ime-ims+1,&
         Dref(ims:ime),Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime))

    RETURN
  END SUBROUTINE degree_of_melting_gr_sb_vec_ie

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_sn_sb_vec_ie(ims,ime,jm,km,ntime,Dvec,anzt,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anzt)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN( snow%a_geo * MIN(MAX(qs(ims:ime,jm,km,ntime)/&
         (qns(ims:ime,jm,km,ntime)+1d-20),snow%x_min),snow%x_max)**snow%b_geo, 6d-3)
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_s(i,jm), 276.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec_ie(tvec(ims:ime),Dvec(ims:ime),ime-ims+1,&
         Dref(ims:ime),Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime))

    RETURN
  END SUBROUTINE degree_of_melting_sn_sb_vec_ie

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_ice_sb_vec_ie(ims,ime,jm,km,ntime,Dvec,anzt,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anzt)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_sb(qi(im,jm,km,ntime), &
!         qi(im,jm,km,ntime)/(MIN(MAX(qi(im,jm,km,ntime)/(qni(im,jm,km,ntime)+1d-20),ice%x_min),ice%x_max)), ice)
    Dref(ims:ime) = MIN( ice%a_geo * MIN(MAX(qi(ims:ime,jm,km,ntime)/&
         (qni(ims:ime,jm,km,ntime)+1d-20),ice%x_min),ice%x_max)**ice%b_geo, 3d-3)
    Dexpo = 1.0d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_i(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec_ie(tvec(ims:ime),Dvec(ims:ime),ime-ims+1,&
         Dref(ims:ime),Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime))

    RETURN
  END SUBROUTINE degree_of_melting_ice_sb_vec_ie

  ! Version fuer die 1M-Schemata:
  SUBROUTINE degree_of_melting_gr_vec_ie(ims,ime,jm,km,ntime,Dvec,anzt,&
       Tmeltbegin,meltdegTmin,n0_g,a_geo_g,b_geo_g,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anzt), n0_g, a_geo_g, b_geo_g

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_1M_exp (qg(im,jm,km,ntime), a_geo_g, b_geo_g, n0_g)
    Dref(ims:ime) = MIN(D_of_X_average_1M_exp_vec (qg(ims:ime,jm,km,ntime)*rho(ims:ime,jm,km), &
         a_geo_g, b_geo_g, n0_g, ime-ims+1), 6d-3)
    Dexpo = 0.6d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_g(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec_ie(tvec(ims:ime),Dvec(ims:ime),ime-ims+1,&
         Dref(ims:ime),Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime))

    RETURN
  END SUBROUTINE degree_of_melting_gr_vec_ie

  ! Version fuer die 1M-Schemata:
  SUBROUTINE degree_of_melting_sn_vec_ie(ims,ime,jm,km,ntime,Dvec,anzt,&
       Tmeltbegin,meltdegTmin,n0_s,a_geo_s,b_geo_s,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, n0_s(anzt), Dvec(anzt), a_geo_s, b_geo_s

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref(ims:ime) = MIN(D_of_X_average_1M_exp_vec_n0s (qs(ims:ime,jm,km,ntime)*rho(ims:ime,jm,km), &
         a_geo_s, b_geo_s, n0_s(ims:ime), ime-ims+1), 6d-3)
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_s(i,jm), 276.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec_ie(tvec(ims:ime),Dvec(ims:ime),ime-ims+1,&
         Dref(ims:ime),Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime))

    RETURN
  END SUBROUTINE degree_of_melting_sn_vec_ie

  ! Version fuer die 1M-Schemata:
  SUBROUTINE degree_of_melting_ice_vec_ie(ims,ime,jm,km,ntime,&
       Dvec,anzt,Tmeltbegin,meltdegTmin,Dref_in,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: jm, km, ntime, anzt
    INTEGER, INTENT(in) :: ims, ime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anzt)
    DOUBLE PRECISION, INTENT(in) :: Dref_in, Tmeltbegin, meltdegTmin, Dvec(anzt)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax(anzt)     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref(anzt), Dexpo, tvec(anzt)
    INTEGER :: i


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf 0.1 mm gesetzt. --- nein, muss extern vorgegeben werden!
    ! Dref = 1d-4
    Dref = Dref_in
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    DO i=ims, ime
      Tmax(i) = MAX(Tmax_i(i,jm), 278.16d0)
    END DO

    tvec(ims:ime) = t(ims:ime,jm,km,ntime)
    fmelt = 0.0d0
    CALL degree_of_melting_fun_vec_ie(tvec(ims:ime),Dvec(ims:ime),ime-ims+1,&
         Dref(ims:ime),Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax(ims:ime),fmelt(ims:ime))

    RETURN
  END SUBROUTINE degree_of_melting_ice_vec_ie

  !=============================================================================================
  ! Die alten Routinen, die lediglich an einem Gitterpunkt gerechnet werden koennen
  !=============================================================================================

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_ha_sb(im,jm,km,ntime,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime, anz
    DOUBLE PRECISION, INTENT(out) :: fmelt(anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref, Dexpo

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_sb(qh(im,jm,km,ntime), &
!         qh(im,jm,km,ntime)/(MIN(MAX(qh(im,jm,km,ntime)/(qnh(im,jm,km,ntime)+1d-20),hail%x_min),hail%x_max)), hail)
    Dref = MIN( hail%a_geo * MIN(MAX(qh(im,jm,km,ntime)/(qnh(im,jm,km,ntime)+1d-20),&
         hail%x_min),hail%x_max)**hail%b_geo, 6d-3)
    Dexpo = 0.8d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_h(im,jm), 293.16d0)

    CALL degree_of_melting_fun(t(im,jm,km,ntime),Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_ha_sb

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_gr_sb(im,jm,km,ntime,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime, anz
    DOUBLE PRECISION, INTENT(out) :: fmelt(anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref,  Dexpo

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_sb(qg(im,jm,km,ntime), &
!         qg(im,jm,km,ntime)/(MIN(MAX(qg(im,jm,km,ntime)/(qng(im,jm,km,ntime)+1d-20),graupel%x_min),graupel%x_max)), graupel)
    Dref = MIN( graupel%a_geo * MIN(MAX(qg(im,jm,km,ntime)/(qng(im,jm,km,ntime)+1d-20),&
         graupel%x_min),graupel%x_max)**graupel%b_geo, 6d-3)
    Dexpo = 0.6d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_g(im,jm), 278.16d0)

    CALL degree_of_melting_fun(t(im,jm,km,ntime),Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_gr_sb

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_sn_sb(im,jm,km,ntime,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime, anz
    DOUBLE PRECISION, INTENT(out) :: fmelt(anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref, Dexpo


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_sb(qs(im,jm,km,ntime), &
!         qs(im,jm,km,ntime)/(MIN(MAX(qs(im,jm,km,ntime)/(qns(im,jm,km,ntime)+1d-20),snow%x_min),snow%x_max)), snow)
    Dref = MIN( snow%a_geo * MIN(MAX(qs(im,jm,km,ntime)/(qns(im,jm,km,ntime)+1d-20),&
         snow%x_min),snow%x_max)**snow%b_geo, 6d-3)
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_s(im,jm), 276.16d0)

    CALL degree_of_melting_fun(t(im,jm,km,ntime),Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_sn_sb

  ! Version fuer das 2M-Schema:
  SUBROUTINE degree_of_melting_ice_sb(im,jm,km,ntime,Dvec,anz,Tmeltbegin,meltdegTmin,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime, anz
    DOUBLE PRECISION, INTENT(out) :: fmelt(anz)
    DOUBLE PRECISION, INTENT(in) :: Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref, Dexpo

    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_sb(qi(im,jm,km,ntime), &
!         qi(im,jm,km,ntime)/(MIN(MAX(qi(im,jm,km,ntime)/(qni(im,jm,km,ntime)+1d-20),ice%x_min),ice%x_max)), ice)
    Dref = MIN( ice%a_geo * MIN(MAX(qi(im,jm,km,ntime)/(qni(im,jm,km,ntime)+1d-20),&
         ice%x_min),ice%x_max)**ice%b_geo, 3d-3)
    Dexpo = 1.0d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_i(im,jm), 278.16d0)

    CALL degree_of_melting_fun(t(im,jm,km,ntime),Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_ice_sb

  ! Version fuer die 1M-Schemata:
  ! de: add q_g, T_a and Tmax_look for lookup table, remove km, ntime
  SUBROUTINE degree_of_melting_gr(im,jm,q_g,T_a,Dvec,anz,Tmeltbegin,meltdegTmin,n0_g,a_geo_g,b_geo_g,fmelt,Tmax_look)

    IMPLICIT NONE

    ! de: remove km, ntime
    INTEGER, INTENT(in) :: im, jm, anz !, km, ntime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anz)
    ! de: add q_g and T_a; add optional parameter Tmax_look
    DOUBLE PRECISION, INTENT(in) :: q_g, T_a, Tmeltbegin, meltdegTmin, Dvec(anz), n0_g, a_geo_g, b_geo_g
    DOUBLE PRECISION, INTENT(in), OPTIONAL :: Tmax_look

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref, Dexpo


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_1M_exp (qg(im,jm,km,ntime), a_geo_g, b_geo_g, n0_g)
    ! de: add *rho(im,jm,km) -> later: replace qg*rho with q_g
    !Dref = MIN(D_of_X_average_1M_exp (qg(im,jm,km,ntime)*rho(im,jm,km), a_geo_g, b_geo_g, n0_g), 6d-3)
    Dref = MIN(D_of_X_average_1M_exp (q_g, a_geo_g, b_geo_g, n0_g), 6d-3)
    Dexpo = 0.6d0

    Tmin = 0.0d0 + 273.16d0
    ! de: add if-enquiry 
    IF (PRESENT(Tmax_look)) THEN
      ! value fron lookup table
      Tmax = Tmax_look
    ELSE
      ! value from model area
      ! de: changed to 278.16d0
      Tmax = MAX(Tmax_g(im,jm), 278.16d0)
    END IF

    ! de: changed t to T_a
    !CALL degree_of_melting_fun(t(im,jm,km,ntime),Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)
    CALL degree_of_melting_fun(T_a,Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_gr

  SUBROUTINE degree_of_melting_gr_single(im,jm,q_g,T_a,D,Tmeltbegin,meltdegTmin,n0_g,a_geo_g,b_geo_g,gfc_bgeop1_g,fmelt)

    IMPLICIT NONE

    ! de: remove km, ntime
    INTEGER, INTENT(in) :: im, jm
    DOUBLE PRECISION, INTENT(out) :: fmelt
    ! de: add q_g and T_a; add optional parameter Tmax_look
    DOUBLE PRECISION, INTENT(in) :: q_g, T_a, Tmeltbegin, meltdegTmin, D, n0_g, a_geo_g, b_geo_g, gfc_bgeop1_g

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref, Dexpo


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref = MIN(D_of_X_average_1M_exp_cin (q_g, a_geo_g, b_geo_g, n0_g, gfc_bgeop1_g), 6d-3)
    Dexpo = 0.6d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_g(im,jm), 278.16d0)

    CALL degree_of_melting_fun_single(T_a,D,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_gr_single

  ! Version fuer die 1M-Schemata:
  ! de: add q_s, T_a and Tmax_look for lookup table, remove km, ntime
  SUBROUTINE degree_of_melting_sn(im,jm,q_s,T_a,Dvec,anz,Tmeltbegin,meltdegTmin,n0_s,a_geo_s,b_geo_s,fmelt,Tmax_look)

    IMPLICIT NONE

    ! de: remove km, ntime
    INTEGER, INTENT(in) :: im, jm, anz !, km, ntime
    DOUBLE PRECISION, INTENT(out) :: fmelt(anz)
    ! de: add q_s and T_a; add optional parameter Tmax_look
    DOUBLE PRECISION, INTENT(in) :: q_s, T_a, Tmeltbegin, meltdegTmin, n0_s, Dvec(anz), a_geo_s, b_geo_s
    DOUBLE PRECISION, INTENT(in), OPTIONAL :: Tmax_look

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref, Dexpo


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
!    Dref = D_average_1M_exp (qs(im,jm,km,ntime), a_geo_s, b_geo_s, n0_s)
    ! de: add *rho(im,jm,km) -> later: replace qs*rho with q_s
    !Dref = MIN(D_of_X_average_1M_exp (qs(im,jm,km,ntime)*rho(im,jm,km), a_geo_s, b_geo_s, n0_s), 6d-3)
    Dref = MIN(D_of_X_average_1M_exp (q_s, a_geo_s, b_geo_s, n0_s), 6d-3)
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    ! de: add if-enquiry 
    IF (PRESENT(Tmax_look)) THEN
      ! value fron lookup table
      Tmax = Tmax_look
    ELSE
      ! value from model area
      ! de: changed to 276.16d0
      !Tmax = MAX(Tmax_s(im,jm), 283.16d0)
      Tmax = MAX(Tmax_s(im,jm), 276.16d0)
    END IF

    ! de: changed t to T_a
    !CALL degree_of_melting_fun(t(im,jm,km,ntime),Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)
    CALL degree_of_melting_fun(T_a,Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)


    RETURN
  END SUBROUTINE degree_of_melting_sn

  ! Version fuer die 1M-Schemata nur fuer 1 Durchmesser:
  SUBROUTINE degree_of_melting_sn_single(im,jm,q_s,T_a,D,Tmeltbegin,meltdegTmin,n0_s,a_geo_s,b_geo_s,gfc_bgeop1_s,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm
    DOUBLE PRECISION, INTENT(out) :: fmelt
    DOUBLE PRECISION, INTENT(in) :: q_s, T_a, Tmeltbegin, meltdegTmin, n0_s, D, a_geo_s, b_geo_s, gfc_bgeop1_s

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dref, Dexpo


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf den mittleren Durchmesser des 
    ! Spektrums gesetzt.
    Dref = MIN(D_of_X_average_1M_exp_cin (q_s, a_geo_s, b_geo_s, n0_s, gfc_bgeop1_s), 6d-3)
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_s(im,jm), 276.16d0)

    CALL degree_of_melting_fun_single(T_a,D,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)


    RETURN
  END SUBROUTINE degree_of_melting_sn_single

  ! Version fuer die 1M-Schemata:
  SUBROUTINE degree_of_melting_ice(im,jm,km,ntime,Dvec,anz,Tmeltbegin,meltdegTmin,Dref,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime, anz
    DOUBLE PRECISION, INTENT(out) :: fmelt(anz)
    DOUBLE PRECISION, INTENT(in) :: Dref, Tmeltbegin, meltdegTmin, Dvec(anz)

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dexpo


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf 0.1 mm gesetzt.
    ! Dref = 1d-4
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_i(im,jm), 278.16d0)

    CALL degree_of_melting_fun(t(im,jm,km,ntime),Dvec,anz,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_ice

  ! Version fuer die 1M-Schemata nur fuer 1 Durchmesser:
  SUBROUTINE degree_of_melting_ice_single(im,jm,T_a,D,Tmeltbegin,meltdegTmin,Dref,fmelt)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm
    DOUBLE PRECISION, INTENT(out) :: fmelt
    DOUBLE PRECISION, INTENT(in) :: T_a, Dref, Tmeltbegin, meltdegTmin, D

    DOUBLE PRECISION :: &
         Tmin, &  ! Ab dieser Temperatur schmelzen die kleinsten Partikel
         Tmax     ! Ab dieser Temperatur sind alle geschmolzen
    DOUBLE PRECISION :: &
         Dexpo


    ! Dref ist der Durchmesser derjenigen Partikel, die bei Erreichen des 1/e-ten Teiles von (T-Tmin)/(Tmax-Tmin)
    ! gerade vollstaendig geschmolzen sind. Dref wird auf 0.1 mm gesetzt.
    ! Dref = 1d-4
    Dexpo = 0.5d0

    Tmin = 0.0d0 + 273.16d0
    Tmax = MAX(Tmax_i(im,jm), 278.16d0)

    CALL degree_of_melting_fun_single(T_a,D,Dref,Dexpo,Tmeltbegin,meltdegTmin,Tmin,Tmax,fmelt)

    RETURN
  END SUBROUTINE degree_of_melting_ice_single



!==========================================================================================
!
! Berechnung der Reflektivitaet fuer die verschiedenen Hydrometeorarten.
!
!==========================================================================================


  !----------------------------------------------------------------------------------------!
  !  Calculate effective radar reflectivity of dry graupel/hail/snow/cloud ice 
  !  using rayleigh-approximation                                                          !
  !  as a function of mass density and number density                                      !
  !  assuming a gamma distribution --- Version for the 2M-Scheme                           !
  !  Similar inconsistency as in zradar_graupel_mie_sb: 
  !  There is no particle larger than a certain Dmax,
  !  but an infinite diameter range for L (and N) is assumed in the microphysics
  !  parameterization and also in the calculation of radar reflectivity.
  !----------------------------------------------------------------------------------------!


  FUNCTION zradar_gr_rayl_oguchi_sb_vec(L_g,N_g,m_i,lambda,parti,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_g(anz), N_g(anz), lambda
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz)
    TYPE(particle), INTENT(in) :: parti
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0
    DOUBLE PRECISION :: K_i(anz), Z_g, K_w_0
    DOUBLE PRECISION, SAVE :: zg
    TYPE(particle), SAVE :: parti_speicher
    INTEGER, SAVE :: firstcall = 0

    ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
    IF (firstcall /= 1 .OR. parti_speicher%nu /= parti%nu .OR. parti_speicher%mu /= parti%mu) THEN
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      zg =  gfct((2+parti%nu+1.0)/parti%mu)        &
           & / gfct((parti%nu+1.0)/parti%mu)        &
           & * ( gfct((parti%nu+1.0)/parti%mu)      &
           &   / gfct((parti%nu+2.0)/parti%mu) )**2 &
           & * (6.0/(pi*rho_ice))**2 / K_w_0
      firstcall = 1
      parti_speicher = parti
    END IF
    
    Z = 0.0d0
    DO i=is, ie
      IF (L_g(i) > 0.0d0 .AND. N_g(i) > 0.0d0) THEN
        K_i(i)  = (ABS( (m_i(i)**2-1.0)/(m_i(i)**2+2.0) ))**2
        Z(i) = K_i(i) * zg * L_g(i)**2 / N_g(i) * 1d18
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_gr_rayl_oguchi_sb_vec

  ! Version fuer die 1-Momenten-Schemen unter Annahme einer Exponentialverteilung
  ! bezueglich des Partikeldurchmessers: (Hier: x = a_geo * D^b_geo)
  ! Nur sinnvoll fuer Schnee und Graupel, nicht fuer Wolkeneis!!
  FUNCTION zradar_gr_rayl_oguchi(L_g,n_0,kioki0xshapefact,b_fak) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_g, n_0, kioki0xshapefact, b_fak
    DOUBLE PRECISION :: Z

    IF (L_g > 0.0d0) THEN
      Z = kioki0xshapefact * L_g**b_fak *  1d18
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_gr_rayl_oguchi

  FUNCTION zradar_gr_rayl_oguchi_vec(L_g,m_i,lambda,n_0,a_geo,b_geo,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_g(anz), lambda, a_geo, b_geo, n_0
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz)
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0
    DOUBLE PRECISION :: K_w_0, K_i(anz)
    DOUBLE PRECISION, SAVE :: zg, b_fak, as, bs, n0s
    INTEGER, SAVE :: firstcall = 0

    IF (firstcall /= 1 .OR. n_0 /= n0s .OR. a_geo /= as .OR. b_geo /= bs) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      b_fak = (2*b_geo+1.)/(b_geo+1.)
      zg = a_geo**2 * n_0 * gfct(2*b_geo+1.0d0) * &
           (a_geo*n_0*gfct(b_geo+1.))**(-b_fak) * (6.0/(pi*rho_ice))**2 / K_w_0
      n0s = n_0
      as = a_geo
      bs = b_geo
      firstcall = 1
    END IF

    Z = 0.0d0
    DO i=is, ie
      IF (L_g(i) > 0.0d0) THEN
        K_i(i)  = (ABS( (m_i(i)**2-1.0)/(m_i(i)**2+2.0) ))**2
        Z(i) = zg * K_i(i) * L_g(i)**b_fak *  1d18
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_gr_rayl_oguchi_vec

  ! Vektorisierte Version mit n_0 als Vektor anstelle eines Skalars wie bei zradar_gr_rayl_oguchi_vec:
  FUNCTION zradar_sn_rayl_oguchi(L_g,n_0,kioki0xshapefact,b_fak) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_g, n_0, kioki0xshapefact, b_fak
    DOUBLE PRECISION :: Z

    IF (L_g > 0.0d0) THEN
      Z = kioki0xshapefact * n_0**(1.0d0-b_fak) * L_g**b_fak *  1d18
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_sn_rayl_oguchi

  FUNCTION zradar_sn_rayl_oguchi_vec(L_g,m_i,lambda,n_0,a_geo,b_geo,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_g(anz), lambda, a_geo, b_geo, n_0(anz)
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz)
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0
    DOUBLE PRECISION :: K_w_0, K_i(anz)
    DOUBLE PRECISION, SAVE :: zg, b_fak, as, bs
    INTEGER, SAVE :: firstcall = 0

    IF (firstcall /= 1 .OR. a_geo /= as .OR. b_geo /= bs) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      b_fak = (2*b_geo+1.)/(b_geo+1.)
      zg = a_geo**2 * gfct(2*b_geo+1.0d0) * &
           (a_geo*gfct(b_geo+1.))**(-b_fak) * (6.0/(pi*rho_ice))**2 / K_w_0
      as = a_geo
      bs = b_geo
      firstcall = 1
    END IF

    Z = 0.0d0
    DO i=is, ie
      IF (L_g(i) > 0.0d0) THEN
        K_i(i)  = (ABS( (m_i(i)**2-1.0)/(m_i(i)**2+2.0) ))**2
        Z(i) = zg * n_0(i)**(1.0d0-b_fak) * K_i(i) * L_g(i)**b_fak *  1d18
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_sn_rayl_oguchi_vec

  ! Version fuer Wolkeneis (Hier: x = a_geo * D^b_geo, 
  ! nu_D, mu_D bezueglich Verteilungsfunktion im Durchmesser)
  FUNCTION zradar_ice_rayleigh_oguchi(L_i,D_i,a_geo,b_geo,kioki0xshapefact) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_i, a_geo, b_geo, D_i, kioki0xshapefact
    DOUBLE PRECISION :: Z

    DOUBLE PRECISION :: x_i

    Z = 0.0d0
    IF (L_i > 0.0d0) THEN

      ! D_i muss der Durchmesser des Partikels mit der mittleren Masse sein, daraus die mittlere Masse:
      x_i = a_geo * D_i ** b_geo
      
      Z = kioki0xshapefact * L_i * x_i * 1d18
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_ice_rayleigh_oguchi

  FUNCTION zradar_ice_rayleigh_oguchi_vec(L_i,m_i,lambda,D_i,nu_D,mu_D,a_geo,b_geo,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_i(anz), lambda, a_geo, b_geo, D_i, nu_D, mu_D
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz)
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0
    DOUBLE PRECISION :: K_w_0, K_i(anz), x_i(anz)
    DOUBLE PRECISION, SAVE :: zg, as, bs, xcs, nu, mu, nuds, muds
    INTEGER, SAVE :: firstcall = 0

    IF (firstcall /= 1 .OR. D_i /= xcs .OR. a_geo /= as .OR. b_geo /= bs .OR. muds /= mu_D .OR. nuds /= nu_D) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      nu = (nu_D + 1.0d0) / b_geo - 1.0d0
      mu = mu_D / b_geo
      zg =  gfct((2+nu+1.0)/mu)        &
           & / gfct((nu+1.0)/mu)        &
           & * ( gfct((nu+1.0)/mu)      &
           &   / gfct((nu+2.0)/mu) )**2 &
           & * (6.0/(pi*rho_ice))**2 / K_w_0
      xcs = D_i
      as = a_geo
      bs = b_geo
      muds = mu_D
      nuds = nu_D
      firstcall = 1
    END IF

    Z = 0.0d0
    DO i=is, ie
      IF (L_i(i) > 0.0d0) THEN
        K_i(i)  = (ABS( (m_i(i)**2-1.0)/(m_i(i)**2+2.0) ))**2

        ! D_i muss der Durchmesser des Partikels mit der mittleren Masse sein, daraus die mittlere Masse:
        x_i(i) = a_geo * D_i ** b_geo

        Z(i) = K_i(i) * zg * L_i(i) * x_i(i) * 1d18
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_ice_rayleigh_oguchi_vec

  ! Rayleigh-Naeherung mit beliebiger Brechungsindexformulierung fuer trockenen Graupel/Hagel/Schnee/Wolkeneis
  ! fuer das 2M-Schema:
  FUNCTION zradar_gr_rayl_sb_vec(L_g,N_g,m_i,lambda,Dmin,Dmax,parti,&
       mixingrulestring, matrixstring, inclusionstring) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: L_g,N_g,lambda
    COMPLEX(kind=kinddp), INTENT(in) :: m_i
    TYPE(particle), INTENT(in) :: parti
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    DOUBLE PRECISION :: Z

!!$    INTEGER, PARAMETER :: n_stuetz = 250 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),x(n_stuetz+1),dD,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6 .AND. N_g >= 1.0d-3) THEN

!      Dmin Untere Integrationsgrenze des Durchmessers in m
!      Dmax Obere Integrationsgrenze des Durchmessers in m
      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i)   = Dmin + (i-1) * dD
        m_iv(i)  = m_i
      ENDDO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = parti%mu / parti%b_geo
      nu_g = (parti%nu + 1.0) / parti%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((parti%nu+1.0)/parti%mu) / gfct((parti%nu+2.0)/parti%mu) * L_g / N_g)**(-parti%mu)
      n_0 = parti%mu * N_g * lam**((parti%nu+1.0)/parti%mu) / gfct((parti%nu+1.0)/parti%mu)
      lam = lam / (parti%a_geo**(mu_g))
      n_0 = n_0 / (parti%b_geo * parti%a_geo**((parti%nu+1.0)/parti%b_geo))

      x = (D_g/parti%a_geo) ** (1.0/parti%b_geo)
      CALL RAYLEIGH_DRYGR_VEC(x,parti%a_geo,parti%b_geo,m_iv,lambda,CBACK, &
           mixingrulestring, matrixstring, inclusionstring, n_stuetz+1)
      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_gr_rayl_sb_vec

  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  FUNCTION zradar_gr_rayl_vec(L_g,m_i,lambda,n_0,a_geo,b_geo,Dmin,Dmax,&
       mixingrulestring, matrixstring, inclusionstring) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: L_g,lambda,n_0,a_geo,b_geo
    COMPLEX(kind=kinddp), INTENT(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    DOUBLE PRECISION :: Z

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),x(n_stuetz+1),dD,a_geo_m,b_geo_m,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6) THEN

!      Dmin = 0.1d-3 ! Untere Integrationsgrenze des Durchmessers in m
!      Dmax = 6.0d-3 ! Obere Integrationsgrenze des Durchmessers in m
      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i)   = Dmin + (i-1) * dD
        m_iv(i)  = m_i
      ENDDO

      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_g) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0 / b_geo
      a_geo_m = (1.0/a_geo) ** (b_geo_m)

      x = a_geo * D_g ** b_geo
      CALL RAYLEIGH_DRYGR_VEC(x,a_geo_m,b_geo_m,m_iv,lambda,CBACK, &
           mixingrulestring, matrixstring, inclusionstring, n_stuetz+1)
      f_d = n_0 * EXP(-lam * D_g)
      eta = SUM(f_d * CBACK * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_gr_rayl_vec


  !----------------------------------------------------------------------------------------!
  !  Calculate effective radar reflectivity of melting graupel/snow/cloud ice using rayleigh-approximation           !
  !  as a function of mass density and number density                                      !
  !  assuming a gamma distribution --- Version for the 2M-Scheme                           !
  !  All particles obey the same degree of melting.
  !
  !  Effective refractive Index of a melting particle (air-ice-water-mixture) is described by the
  !  Oguchi-Formula with form factor 2.0 (equals Debye's Formula for a homogeneous mixture).
  !  In case all the particles have the same melting fraction 
  !  f_melt_mean = m_water / (m_water + m_ice),
  !  this formula enables a nice analytic solution of the reflectivity integral.
  ! 
  !  Similar inconsistency as in zradar_graupel_mie_sb: 
  !  There is no particle larger than a certain Dmax,
  !  but an infinite diameter range for L (and N) is assumed in the microphysics
  !  parameterization and also in the calculation of radar reflectivity.
  !----------------------------------------------------------------------------------------!

  FUNCTION zradar_wetgr_rayl_oguchi_sb_vec(L_g,N_g,m_i,m_w,lambda,f_melt_mean,parti,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_g(anz), N_g(anz), lambda, f_melt_mean(anz)
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz), m_w(anz)
    TYPE(particle), INTENT(in) :: parti
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0, K_i(anz), K_w(anz)
    DOUBLE PRECISION :: K_w_0, fm(anz)
    DOUBLE PRECISION, SAVE :: zg
    TYPE(particle), SAVE :: parti_speicher
    INTEGER, SAVE :: firstcall = 0

    IF (firstcall /= 1 .OR. parti_speicher%nu /= parti%nu .OR. &
         parti_speicher%mu /= parti%mu .OR. parti_speicher%name /= parti%name) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0d0)/(m_w_0**2+2.0) ))**2
      zg =  gfct((2d0+parti%nu+1.0d0)/parti%mu)        &
           & / gfct((parti%nu+1.0d0)/parti%mu)        &
           & * ( gfct((parti%nu+1.0d0)/parti%mu)      &
           &   / gfct((parti%nu+2.0d0)/parti%mu) )**2 &
           & * (6.0d0/pi)**2 / K_w_0
      firstcall = 1
      parti_speicher = parti
    END IF

    fm = MAX(MIN(f_melt_mean,1.0d0),0.0d0)

    Z = 0.0d0
    DO i=is, ie
      IF (L_g(i) > 0.0d0 .AND. N_g(i) > 0.0d0) THEN
        K_i(i)  = (m_i(i)**2-1.0)/(m_i(i)**2+2.0)
        K_w(i)  = (m_w(i)**2-1.0)/(m_w(i)**2+2.0)
        
        Z(i) = ( (ABS(K_i(i))*(1.0-fm(i))/rho_ice)**2 + &
             2.0*(DBLE(K_i(i))*DBLE(K_w(i))+AIMAG(K_i(i))*AIMAG(K_w(i)))*fm(i)*(1.0-fm(i))/(rho_ice*rho_w) + &
             (ABS(K_w(i))*fm(i)/rho_w)**2 ) * &
             zg * L_g(i)**2 / N_g(i) * 1d18
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_wetgr_rayl_oguchi_sb_vec

  ! Version fuer die 1-Momenten-Schemen unter Annahme einer Exponentialverteilung
  ! bezueglich des Partikeldurchmessers: x = a_geo * D^b_geo
  ! Nur sinnvoll fuer Schnee und Graupel, nicht fuer Wolkeneis!!
  FUNCTION zradar_wetgr_rayl_oguchi(L_g,m_i,m_w,f_melt_mean,n_0,shapefactokw0,b_fak) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_g, f_melt_mean, n_0, shapefactokw0, b_fak
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    DOUBLE PRECISION :: Z

    COMPLEX(kind=kinddp) :: K_w, K_i
    DOUBLE PRECISION :: fm

    fm = MAX(MIN(f_melt_mean,1.0d0),0.0d0)

    IF (L_g > 0.0d0) THEN
      K_i  = (m_i**2-1.0)/(m_i**2+2.0)
      K_w  = (m_w**2-1.0)/(m_w**2+2.0)

      Z = ( (ABS(K_i)*(1.0-fm)/rho_ice)**2 + &
           2.0*(DBLE(K_i)*DBLE(K_w)+AIMAG(K_i)*AIMAG(K_w))*fm*(1.0-fm)/(rho_ice*rho_w) + &
           (ABS(K_w)*fm/rho_w)**2 ) * &
           shapefactokw0 * L_g**b_fak * 1d18
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_wetgr_rayl_oguchi

  FUNCTION zradar_wetgr_rayl_oguchi_vec(L_g,m_i,m_w,lambda,f_melt_mean,n_0,a_geo,b_geo,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_g(anz), lambda, f_melt_mean(anz), n_0, a_geo, b_geo
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz), m_w(anz)
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0, K_w(anz), K_i(anz)
    DOUBLE PRECISION :: K_w_0, fm(anz)
    DOUBLE PRECISION, SAVE :: zg, b_fak, as, bs, n0s
    INTEGER, SAVE :: firstcall = 0

    IF (firstcall /= 1 .OR. n_0 /= n0s .OR. a_geo /= as .OR. b_geo /= bs) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      b_fak = (2*b_geo+1.)/(b_geo+1.)
      zg = a_geo**2 * n_0 * gfct(2*b_geo+1.0d0) * &
           (a_geo*n_0*gfct(b_geo+1.))**(-b_fak) * (6.0/pi)**2 / K_w_0
      n0s = n_0
      as = a_geo
      bs = b_geo
      firstcall = 1
    END IF

    fm = MAX(MIN(f_melt_mean,1.0d0),0.0d0)

    Z = 0.0d0
    DO i=is, ie
      IF (L_g(i) > 0.0d0) THEN
        K_i(i)  = (m_i(i)**2-1.0)/(m_i(i)**2+2.0)
        K_w(i)  = (m_w(i)**2-1.0)/(m_w(i)**2+2.0)

        Z(i) = ( (ABS(K_i(i))*(1.0-fm(i))/rho_ice)**2 + &
             2.0*(DBLE(K_i(i))*DBLE(K_w(i))+AIMAG(K_i(i))*AIMAG(K_w(i)))*fm(i)*(1.0-fm(i))/(rho_ice*rho_w) + &
             (ABS(K_w(i))*fm(i)/rho_w)**2 ) * &
             zg * L_g(i)**b_fak * 1d18
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_wetgr_rayl_oguchi_vec

  ! Vektorisierte Variante mit n_0 als Vektor anstelle eines Skalars wie bei zradar_wetgr_rayl_oguchi_vec:
  FUNCTION zradar_wetsn_rayl_oguchi(L_g,m_i,m_w,f_melt_mean,n_0,shapefactokw0,b_fak) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_g, f_melt_mean, n_0, shapefactokw0, b_fak
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    DOUBLE PRECISION :: Z

    COMPLEX(kind=kinddp) :: K_w, K_i
    DOUBLE PRECISION :: fm

    fm = MAX(MIN(f_melt_mean,1.0d0),0.0d0)

    IF (L_g > 0.0d0) THEN
      K_i  = (m_i**2-1.0)/(m_i**2+2.0)
      K_w  = (m_w**2-1.0)/(m_w**2+2.0)
      
      Z = ( (ABS(K_i)*(1.0-fm)/rho_ice)**2 + &
           2.0*(DBLE(K_i)*DBLE(K_w)+AIMAG(K_i)*AIMAG(K_w))*fm*(1.0-fm)/(rho_ice*rho_w) + &
           (ABS(K_w)*fm/rho_w)**2 ) * &
           shapefactokw0 * n_0**(1.0d0-b_fak) * L_g**b_fak * 1d18
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_wetsn_rayl_oguchi

  FUNCTION zradar_wetsn_rayl_oguchi_vec(L_g,m_i,m_w,lambda,f_melt_mean,n_0,a_geo,b_geo,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_g(anz), lambda, f_melt_mean(anz), n_0(anz), a_geo, b_geo
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz), m_w(anz)
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0, K_w(anz), K_i(anz)
    DOUBLE PRECISION :: K_w_0, fm(anz)
    DOUBLE PRECISION, SAVE :: zg, b_fak, as, bs
    INTEGER, SAVE :: firstcall = 0

    IF (firstcall /= 1 .OR. a_geo /= as .OR. b_geo /= bs) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      b_fak = (2*b_geo+1.)/(b_geo+1.)
      zg = a_geo**2 * gfct(2*b_geo+1.0d0) * &
           (a_geo*gfct(b_geo+1.))**(-b_fak) * (6.0/pi)**2 / K_w_0
      as = a_geo
      bs = b_geo
      firstcall = 1
    END IF

    fm = MAX(MIN(f_melt_mean,1.0d0),0.0d0)

    Z = 0.0d0
    DO i=is, ie
      IF (L_g(i) > 0.0d0) THEN
        K_i(i)  = (m_i(i)**2-1.0)/(m_i(i)**2+2.0)
        K_w(i)  = (m_w(i)**2-1.0)/(m_w(i)**2+2.0)

        Z(i) = ( (ABS(K_i(i))*(1.0-fm(i))/rho_ice)**2 + &
             2.0*(DBLE(K_i(i))*DBLE(K_w(i))+AIMAG(K_i(i))*AIMAG(K_w(i)))*fm(i)*(1.0-fm(i))/(rho_ice*rho_w) + &
             (ABS(K_w(i))*fm(i)/rho_w)**2 ) * &
             zg * n_0(i)**(1.0d0-b_fak) * L_g(i)**b_fak * 1d18
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_wetsn_rayl_oguchi_vec

  ! Version fuer Wolkeneis fuer die 1-Momenten-Schemen unter Annahme einer gen. Gammaverteilung (nu_D, mu_D)
  ! bezueglich des Partikeldurchmessers. Hier: x = a_geo * D^b_geo
  ! Nur sinnvoll fuer Wolkeneis!!
  FUNCTION zradar_wetice_rayl_oguchi(L_i,m_i,m_w,f_melt_mean,D_i,a_geo,b_geo,shapefactokw0) RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_i, D_i, f_melt_mean, a_geo, b_geo, shapefactokw0
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    DOUBLE PRECISION :: Z

    COMPLEX(kind=kinddp) :: K_w, K_i
    DOUBLE PRECISION :: fm, x_i

    fm = MAX(MIN(f_melt_mean,1.0d0),0.0d0)

    IF (L_i > 0.0d0) THEN
      K_i  = (m_i**2-1.0)/(m_i**2+2.0)
      K_w  = (m_w**2-1.0)/(m_w**2+2.0)

      ! D_i muss der Durchmesser des Partikels mit der mittleren Masse sein, daraus die mittlere Masse:
      x_i = a_geo * D_i ** b_geo

      Z = ( (ABS(K_i)*(1.0-fm)/rho_ice)**2 + &
           2.0*(DBLE(K_i)*DBLE(K_w)+AIMAG(K_i)*AIMAG(K_w))*fm*(1.0-fm)/(rho_ice*rho_w) + &
           (ABS(K_w)*fm/rho_w)**2 ) * &
           shapefactokw0 * L_i * x_i * 1d18       
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_wetice_rayl_oguchi

  FUNCTION zradar_wetice_rayl_oguchi_vec(L_i,m_i,m_w,lambda,f_melt_mean,D_i,nu_D,mu_D,a_geo,b_geo,anz,is,ie) RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    DOUBLE PRECISION, INTENT(in) :: L_i(anz), lambda, f_melt_mean(anz), D_i, mu_D, nu_D, a_geo, b_geo
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz), m_w(anz)
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: K_w(anz), K_i(anz), m_w_0
    DOUBLE PRECISION :: K_w_0, fm(anz), x_i(anz)
    DOUBLE PRECISION, SAVE :: zg, xcs, as, bs, nu, mu, nuds, muds
    INTEGER, SAVE :: firstcall = 0

    IF (firstcall /= 1 .OR. D_i /= xcs .OR. a_geo /= as .OR. b_geo /= bs .OR. mu_D /= muds .OR. nu_D /= nuds) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      nu = (nu_D + 1.0d0) / b_geo - 1.0d0
      mu = mu_D / b_geo
      zg =  gfct((2.0d0+nu+1.0d0)/mu)        &
           & / gfct((nu+1.0d0)/mu)        &
           & * ( gfct((nu+1.0d0)/mu)      &
           &   / gfct((nu+2.0d0)/mu) )**2 &
           & * (6.0d0/pi)**2 / K_w_0
      xcs = D_i
      as = a_geo
      bs = b_geo
      muds = mu_D
      nuds = nu_D
      firstcall = 1
    END IF

    fm = MAX(MIN(f_melt_mean,1.0d0),0.0d0)

    Z = 0.0d0
    DO i=is, ie
      IF (L_i(i) > 0.0d0) THEN
        K_i(i)  = (m_i(i)**2-1.0)/(m_i(i)**2+2.0)
        K_w(i)  = (m_w(i)**2-1.0)/(m_w(i)**2+2.0)

        ! D_i muss der Durchmesser des Partikels mit der mittleren Masse sein, daraus die mittlere Masse:
        x_i(i) = a_geo * D_i ** b_geo

        Z(i) = ( (ABS(K_i(i))*(1.0-fm(i))/rho_ice)**2 + &
             2.0*(DBLE(K_i(i))*DBLE(K_w(i))+AIMAG(K_i(i))*AIMAG(K_w(i)))*fm(i)*(1.0-fm(i))/(rho_ice*rho_w) + &
             (ABS(K_w(i))*fm(i)/rho_w)**2 ) * &
             zg * L_i(i) * x_i(i) * 1d18       
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_wetice_rayl_oguchi_vec

  !----------------------------------------------------------------------------------------!
  ! 
  !  Calculate effective radar reflectivity of melting hail/graupel/snow/cloud ice using rayleigh-approximation           !
  !  as a function of mass density and number density                                      !
  !  assuming a gamma distribution --- Version for the 2M-Scheme                           !
  !  Here, the particles all have different degree of melting according to their
  !  size and temperature.
  !
  !  Effective refractive Index of a melting particle (air-ice-water-mixture) is described by
  !  a selectable parameterization.
  ! 
  !----------------------------------------------------------------------------------------!

  ! Version fuer schmelzenden Graupel beim 2M-Schema:
  FUNCTION zradar_wetgr_rayl_sb_vec(im,jm,km,ntime,L_g,N_g,m_i,m_w,&
       Tmeltbegin,meltdegTmin,lambda,Dmin,Dmax,parti,name,&
       mixingrulestring, matrixstring, inclusionstring, hoststring, hostmatrixstring, hostinclusionstring) &
       RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime
    DOUBLE PRECISION, INTENT(in) :: L_g,N_g,lambda,Tmeltbegin,meltdegTmin
    COMPLEX(kind=kinddp) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: name, mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring
    TYPE(particle), INTENT(in) :: parti
    DOUBLE PRECISION :: Z

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: D_g(n_stuetz+1),CBACK(n_stuetz+1),x(n_stuetz+1),dD,fmelt(n_stuetz+1),meltratio_outside

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
        ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF
    
    IF (L_g >= 1.d-6 .AND. N_g >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = parti%mu / parti%b_geo
      nu_g = (parti%nu + 1.0) / parti%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((parti%nu+1.0)/parti%mu) / gfct((parti%nu+2.0)/parti%mu) * L_g / N_g)**(-parti%mu)
      n_0 = parti%mu * N_g * lam**((parti%nu+1.0)/parti%mu) / gfct((parti%nu+1.0)/parti%mu)
      lam = lam / (parti%a_geo**(mu_g))
      n_0 = n_0 / (parti%b_geo * parti%a_geo**((parti%nu+1.0)/parti%b_geo))

        
      ! Schmelzgradparametrisierung:
      SELECT CASE (TRIM(ADJUSTL(name)))
      CASE('ice')
        CALL degree_of_melting_ice_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)
        meltratio_outside = 1.0d0
      CASE('snow')
        CALL degree_of_melting_sn_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)
        meltratio_outside = 0.7d0
      CASE('graupel')
        CALL degree_of_melting_gr_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)
        meltratio_outside = 0.7d0
      CASE('hail')
        CALL degree_of_melting_ha_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)
        meltratio_outside = 1.0d0
      CASE default
        WRITE (*,*) 'ZRADAR_WETGR_RAYL_SB: Ungueltige Partikelart ', name
        Z = 0.0d0
        RETURN
      END SELECT

      x = (D_g/parti%a_geo) ** (1.0/parti%b_geo)
      CALL RAYLEIGH_SOAK_WETGR_VEC(x,parti%a_geo,parti%b_geo,fmelt,meltratio_outside,m_wv,m_iv,lambda,CBACK,&
           mixingrulestring, matrixstring, inclusionstring, &
           hoststring, hostmatrixstring, hostinclusionstring, n_stuetz+1)
      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
    END IF

    RETURN

  END FUNCTION zradar_wetgr_rayl_sb_vec

  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  ! de: add T_a for calculating degree_of_melting_sn
  FUNCTION zradar_wetgr_rayl_vec(im,jm,km,ntime,L_g,T_a,m_i,m_w,&
       Tmeltbegin,meltdegTmin,lambda,n_0,a_geo,b_geo,name,Dmin,Dmax,&
       mixingrulestring, matrixstring, inclusionstring, hoststring, hostmatrixstring, hostinclusionstring) &
       RESULT(Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime
    ! de: add T_a
    DOUBLE PRECISION, INTENT(in)     :: L_g,T_a,lambda,n_0,a_geo,b_geo,Tmeltbegin,meltdegTmin
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: name, mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring
    DOUBLE PRECISION :: Z

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: D_g(n_stuetz+1),CBACK(n_stuetz+1),x(n_stuetz+1),dD,&
         meltratio_outside,fmelt(n_stuetz+1),a_geo_m,b_geo_m

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6) THEN

!      Dmin = 0.1d-3 ! Untere Integrationsgrenze des Durchmessers in m
!      Dmax = 6.0d-3 ! Obere Integrationsgrenze des Durchmessers in m
      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! Schmelzgradparametrisierung:
      SELECT CASE (TRIM(ADJUSTL(name)))
      CASE('ice')
        CALL degree_of_melting_ice(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,100.0d-6,fmelt)
        meltratio_outside = 1.0d0
      CASE('snow')
        ! de: add L_g and T_a for lookup table (remark: in case('snow') L_g=L_s !); remove km, ntime
        CALL degree_of_melting_sn(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt)
        meltratio_outside = 0.7d0
      CASE('graupel')
        ! de: add L_g and T_a for lookup table; remove km, ntime
        CALL degree_of_melting_gr(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt)
        meltratio_outside = 0.7d0
      CASE default
        WRITE (*,*) 'ZRADAR_WETGR_RAYL: Ungueltige Partikelart ', name
        Z = 0.0d0
        RETURN
      END SELECT

      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_g) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0/b_geo
      a_geo_m = (1.0/a_geo)**b_geo_m

      x = a_geo * D_g ** b_geo
      CALL RAYLEIGH_SOAK_WETGR_VEC(x,a_geo_m,b_geo_m,fmelt,meltratio_outside,m_wv,m_iv,lambda,CBACK,&
             mixingrulestring, matrixstring, inclusionstring, &
             hoststring, hostmatrixstring, hostinclusionstring, n_stuetz+1)
      f_d = n_0 * EXP(-lam * D_g)
      eta = SUM(f_d * CBACK * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_wetgr_rayl_vec

  !----------------------------------------------------------------------------------------!
  !  Calculate radar reflectivity of dry graupel/hail using full mie theory                     !
  !  as a function of mass density and number density                                      !
  !  assuming a gamma distribution (in case of 2-moment-scheme of Seifert and Beheng) or an
  !  exponential size distribution (in case of LM 1-moment-scheme)
  !
  !  The integration of the reflectivity integral is done in diameter-space
  !  by means of Simpson's rule. Diameter space is chosen
  !  since the distribution of equally spaced sampling points is better
  !  suited for the integrand function compared to mass space, in that more sampling
  !  points are within the "small drop" end of the size distribution and less sampling
  !  points at the "large drop" end.
  !
  !  The procedure contains a minor inconsistency:
  !  To deduce the parameters of the DSD with respect to diameter, 
  !  it is assumed in accordance with the Microphysics parameterizations
  !  that L (and N in CASE of 2M) are representative
  !  for a DSD spanning over an infinite diameter range (to infinity).
  !  In contrast, when integrating over the spectrum, the integral is taken
  !  from Dmin to Dmax, neglecting the part to infinity.
  !  In fact, the latter is consistent with the physical argument, that no drops larger
  !  than a Dmax can exist for longer time because of spontaneous breakup, whereas it is
  !  felt that the assumption of an "infinite" diameter range for L (and N) in
  !  microphysics parameterizations is rather problematic.
  !
  !----------------------------------------------------------------------------------------!

  ! de: replace function with subroutine, add output parameter for extinction zeta,
  !     analogue for all reflectivity functions calculated with mie theory

  ! Fuer trockenen Hagel:
  SUBROUTINE zradar_hail_mie_sb_vec(L_h ,N_h ,m_i ,lambda ,Dmin ,Dmax ,Z ,zeta , &
                                    mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: L_h,N_h,lambda
    COMPLEX(kind=kinddp), INTENT(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_h >= 1.d-6 .AND. N_h >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i)   = Dmin + (i-1) * dD
        m_iv(i)  = m_i
      ENDDO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = hail%mu / hail%b_geo
      nu_g = (hail%nu + 1.0) / hail%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((hail%nu+1.0)/hail%mu) / gfct((hail%nu+2.0)/hail%mu) * L_h / N_h)**(-hail%mu)
      n_0 = hail%mu * N_h * lam**((hail%nu+1.0)/hail%mu) / gfct((hail%nu+1.0)/hail%mu)
      lam = lam / (hail%a_geo**(mu_g))
      n_0 = n_0 / (hail%b_geo * hail%a_geo**((hail%nu+1.0)/hail%b_geo))

      x = (D_g/hail%a_geo) ** (1.0/hail%b_geo)
      CALL MIE_DRYGR_VEC(x,hail%a_geo,hail%b_geo,m_iv,lambda,CBACK,CEXT,&
             mixingrulestring, matrixstring, inclusionstring, n_stuetz+1)

      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_hail_mie_sb_vec

  ! Fuer trockenen Graupel:
  SUBROUTINE zradar_graupel_mie_sb_vec(L_g,N_g,m_i,lambda,Dmin,Dmax,Z,zeta,&
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: L_g,N_g,lambda
    COMPLEX(kind=kinddp), INTENT(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6 .AND. N_g >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i)   = Dmin + (i-1) * dD
        m_iv(i)  = m_i
      enddo

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = graupel%mu / graupel%b_geo
      nu_g = (graupel%nu + 1.0) / graupel%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+2.0)/graupel%mu) * L_g / N_g)**(-graupel%mu)
      n_0 = graupel%mu * N_g * lam**((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+1.0)/graupel%mu)
      lam = lam / (graupel%a_geo**(mu_g))
      n_0 = n_0 / (graupel%b_geo * graupel%a_geo**((graupel%nu+1.0)/graupel%b_geo))

      x = (D_g/graupel%a_geo) ** (1.0/graupel%b_geo)
      CALL MIE_DRYGR_VEC(x,graupel%a_geo,graupel%b_geo,m_iv,lambda,CBACK,CEXT, &
           mixingrulestring, matrixstring, inclusionstring, n_stuetz+1)

      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_graupel_mie_sb_vec

  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  ! de: replace m_i with T_a, needed for lookup table
  SUBROUTINE zradar_graupel_mie_vec(L_g,T_a,lambda,n_0,a_geo,b_geo,Dmin,Dmax,Z,zeta,&
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    ! de: add T_a
    DOUBLE PRECISION, INTENT(in)     :: L_g,T_a,lambda,n_0,a_geo,b_geo
    ! de:
    !COMPLEX(kind=kinddp), INTENT(in)  :: m_i
    CHARACTER(len=*), INTENT(in)  :: mixingrulestring, matrixstring, inclusionstring
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    ! de: add m_i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_i, m_w_0
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,a_geo_m,b_geo_m,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0 = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6) THEN

      ! de: add calculation of m_i(T_a)
      m_i = m_complex_ice_maetzler(lambda,MIN(MAX(T_a-273.16d0,-80.0d0),0.0d0))
      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i)  = Dmin + (i-1) * dD
        m_iv(i) = m_i
      END DO

      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_g) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0 / b_geo
      a_geo_m = (1.0/a_geo) ** (b_geo_m)

      x = a_geo * D_g ** b_geo
      CALL MIE_DRYGR_vec(x,a_geo_m,b_geo_m,m_iv,lambda,CBACK,CEXT, &
           mixingrulestring, matrixstring, inclusionstring, n_stuetz+1)

      f_d = n_0 * EXP(-lam * D_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_graupel_mie_vec

  ! Entsprechende Routinen fuer Schnee (Zweischalenmodell, radienverh = r_innen / r_gesamt):
  SUBROUTINE zradar_snow_mie_sb_vec(L_s,N_s,m_i,radienverh,lambda,Dmin,Dmax,Z,zeta,&
       mixingrulestring_core, matrixstring_core, inclusionstring_core, &
       mixingrulestring_shell, matrixstring_shell, inclusionstring_shell)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: L_s,N_s,lambda,radienverh
    COMPLEX(kind=kinddp), INTENT(in)  :: m_i
    CHARACTER(len=*), INTENT(in)  :: mixingrulestring_core, matrixstring_core, inclusionstring_core, &
         mixingrulestring_shell, matrixstring_shell, inclusionstring_shell
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_s,nu_s
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,D_s(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_s >= 1.d-6 .AND. N_s >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_s(i)   = Dmin + (i-1) * dD
        m_iv(i)  = m_i
      ENDDO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_s = snow%mu / snow%b_geo
      nu_s = (snow%nu + 1.0) / snow%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((snow%nu+1.0)/snow%mu) / gfct((snow%nu+2.0)/snow%mu) * L_s / N_s)**(-snow%mu)
      n_0 = snow%mu * N_s * lam**((snow%nu+1.0)/snow%mu) / gfct((snow%nu+1.0)/snow%mu)
      lam = lam / (snow%a_geo**(mu_s))
      n_0 = n_0 / (snow%b_geo * snow%a_geo**((snow%nu+1.0)/snow%b_geo))

      x = (D_s/snow%a_geo) ** (1.0/snow%b_geo)
      CALL MIE_DRYSNOW_TWOSPH_VEC(x,snow%a_geo,snow%b_geo,m_iv,radienverh,lambda,CBACK,CEXT, &
             mixingrulestring_core, matrixstring_core, inclusionstring_core, &
             mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, n_stuetz+1)

      f_d = n_0 * D_s**nu_s * EXP(-lam * D_s**mu_s)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_snow_mie_sb_vec


  ! Entsprechende Routinen fuer Schnee (Zweischalenmodell, radienverh = r_innen / r_gesamt):
  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  ! de: replace m_i with T_a, needed for lookup table
  SUBROUTINE zradar_snow_mie_vec(L_s,T_a,radienverh,lambda,n_0,a_geo,b_geo,Dmin,Dmax,Z,zeta,&
       mixingrulestring_core, matrixstring_core, inclusionstring_core, &
       mixingrulestring_shell, matrixstring_shell, inclusionstring_shell)

    IMPLICIT NONE

    ! de: add T_a
    DOUBLE PRECISION, INTENT(in)     :: L_s,T_a,lambda,radienverh,n_0,a_geo,b_geo
    ! de:
    !COMPLEX(kind=kinddp), INTENT(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_core, matrixstring_core, inclusionstring_core, &
                                    mixingrulestring_shell, matrixstring_shell, inclusionstring_shell
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    ! de: add m_i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_i, m_w_0
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,a_geo_m,b_geo_m,D_s(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_s >= 1.d-6) THEN

      ! de: add calculation of m_i(T_a)
      m_i = m_complex_ice_maetzler(lambda,MIN(MAX(T_a-273.16d0,-80.0d0),0.0d0))
      dD  = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_s(i)   = Dmin + (i-1) * dD
        m_iv(i) = m_i
      END DO


      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_s) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0 / b_geo
      a_geo_m = (1.0/a_geo) ** b_geo_m

      x = a_geo * D_s ** b_geo
      CALL MIE_DRYSNOW_TWOSPH_VEC(x,a_geo_m,b_geo_m,m_iv,radienverh,lambda,CBACK,CEXT, &
           mixingrulestring_core, matrixstring_core, inclusionstring_core, &
           mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, n_stuetz+1)

      f_d = n_0 * EXP(-lam * D_s)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_snow_mie_vec

  ! Version fuer schmelzenden Graupel beim 2M-Schema, soaked version:
  SUBROUTINE zradar_wetgr_mie_sb_vec(im,jm,km,ntime,L_g,N_g,m_i,m_w,&
       Tmeltbegin,meltdegTmin,lambda,Dmin,Dmax,Z,zeta,&
       mixingrulestring, matrixstring, inclusionstring, hoststring, hostmatrixstring, hostinclusionstring)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime
    DOUBLE PRECISION, INTENT(in) :: L_g,N_g,lambda,Tmeltbegin,meltdegTmin
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            D_g(n_stuetz+1),dD,fmelt(n_stuetz+1),meltratio_outside

    INTEGER, SAVE        :: firstcall = 0
    DOUBLE PRECISION, SAVE :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6 .AND. N_g >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = graupel%mu / graupel%b_geo
      nu_g = (graupel%nu + 1.0) / graupel%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+2.0)/graupel%mu) * L_g / N_g)**(-graupel%mu)
      n_0 = graupel%mu * N_g * lam**((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+1.0)/graupel%mu)
      lam = lam / (graupel%a_geo**(mu_g))
      n_0 = n_0 / (graupel%b_geo * graupel%a_geo**((graupel%nu+1.0)/graupel%b_geo))

      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! Schmelzgradparametrisierung:
      CALL degree_of_melting_gr_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)

      ! Verhaeltnis der an der Oberflaeche wegschmelzenden Masse zur gesamten geschmolzenen Masse.
      ! Der komplementaere Anteil schmilzt im Inneren des Partikels. Dieses Verhaeltnis nimmt vom hier
      ! vorzugebenden Wert mit zunehmendem Schmelzgrad linear auf den Wert 1.0 zu.
      meltratio_outside = 0.7d0

      x = (D_g/graupel%a_geo) ** (1.0/graupel%b_geo)
      ! Alternativen: MIE_WATERSPH_WETGR_VEC(), MIE_SOAK_TWOSPH_WETGR_VEC(), 
      ! MIE_MEAN_WETGR_VEC(), MIE_WETSNOW_TWOSPH_VEC()
      CALL MIE_SOAK_WETGR_VEC(x,graupel%a_geo,graupel%b_geo,fmelt,meltratio_outside,m_wv,m_iv,lambda,CBACK,CEXT,&
             mixingrulestring, matrixstring, inclusionstring, &
             hoststring, hostmatrixstring, hostinclusionstring, n_stuetz+1)

      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wetgr_mie_sb_vec

  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  ! de: replace m_i,m_w with T_a, add optional parameter Tmax -> needed for lookup table
  !     remove km, ntime
  SUBROUTINE zradar_wetgr_mie_vec(im,jm,L_g,T_a,Tmeltbegin,meltdegTmin,lambda,n_0,a_geo,b_geo,Dmin,Dmax,Z,zeta,&
       mixingrulestring, matrixstring, inclusionstring, hoststring, hostmatrixstring, hostinclusionstring,Tmax)

    IMPLICIT NONE

    ! de: remove km, ntime
    INTEGER, INTENT(in) :: im, jm !, km, ntime
    ! de: add T_a and Tmax for lookup table
    DOUBLE PRECISION, INTENT(in)    :: L_g,T_a,lambda,n_0,a_geo,b_geo,Tmeltbegin,meltdegTmin
    DOUBLE PRECISION, INTENT(in), OPTIONAL :: Tmax
    ! de:
    !COMPLEX(kind=kinddp) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    ! de: add m_w, m_i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0, m_w, m_i
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,meltratio_outside,fmelt(n_stuetz+1),&
                            a_geo_m,b_geo_m,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6) THEN

      ! de: add calculation of m_w(T_a) and m_i(T_a)
      m_w = m_complex_water_ray(lambda,MIN(MAX(T_a-273.16d0,0.0d0),30.0d0))
      m_i = m_complex_ice_maetzler(lambda,MIN(MAX(T_a-273.16d0,-80.0d0),0.0d0))

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! Schmelzgradparametrisierung:
      ! de: add if-enquiry and add L_g and T_a (and Tmax) for lookup table; remove km, ntime
      IF (PRESENT(Tmax)) THEN
        ! de: with additional optional parameter Tmax for calculationg lookup table
        CALL degree_of_melting_gr(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt,Tmax)
      ELSE
        CALL degree_of_melting_gr(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt)
      END IF

      ! Verhaeltnis der an der Oberflaeche wegschmelzenden Masse zur gesamten geschmolzenen Masse.
      ! Der komplementaere Anteil schmilzt im Inneren des Partikels. Dieses Verhaeltnis nimmt vom hier
      ! vorzugebenden Wert mit zunehmendem Schmelzgrad linear auf den Wert 1.0 zu.
      meltratio_outside = 0.7d0

      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_g) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0/b_geo
      a_geo_m = (1.0/a_geo)**b_geo_m

      x = a_geo * D_g ** b_geo
        ! Alternativen: MIE_WATERSPH_WETGR_VEC(), MIE_SOAK_TWOSPH_WETGR_VEC(), MIE_MEAN_WETGR_VEC(), MIE_WETSNOW_TWOSPH_VEC()
      CALL MIE_SOAK_WETGR_VEC(x,a_geo_m,b_geo_m,fmelt,meltratio_outside,m_wv,m_iv,lambda,CBACK,CEXT,&
           mixingrulestring, matrixstring, inclusionstring, &
           hoststring, hostmatrixstring, hostinclusionstring, n_stuetz+1)

      f_d = n_0 * EXP(-lam * D_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wetgr_mie_vec

  ! Version fuer schmelzenden Graupel beim 2M-Schema, soaked twosphere version:
  SUBROUTINE zradar_wetgr_twosph_mie_sb_vec(im,jm,km,ntime,L_g,N_g,m_i,m_w,&
       Tmeltbegin,meltdegTmin,lambda,Dmin,Dmax,Z,zeta,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
       mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime
    DOUBLE PRECISION, INTENT(in) :: L_g,N_g,lambda,Tmeltbegin,meltdegTmin
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            D_g(n_stuetz+1),dD,fmelt(n_stuetz+1)

    INTEGER, SAVE        :: firstcall = 0
    DOUBLE PRECISION, SAVE :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6 .AND. N_g >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = graupel%mu / graupel%b_geo
      nu_g = (graupel%nu + 1.0) / graupel%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+2.0)/graupel%mu) * L_g / N_g)**(-graupel%mu)
      n_0 = graupel%mu * N_g * lam**((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+1.0)/graupel%mu)
      lam = lam / (graupel%a_geo**(mu_g))
      n_0 = n_0 / (graupel%b_geo * graupel%a_geo**((graupel%nu+1.0)/graupel%b_geo))

      ! Schmelzgradparametrisierung:
      CALL degree_of_melting_gr_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)

      x = (D_g/graupel%a_geo) ** (1.0/graupel%b_geo)
      ! Alternativen: MIE_WATERSPH_WETGR_VEC(), MIE_SOAK_WETGR_VEC(), 
      ! MIE_MEAN_WETGR_VEC(), MIE_WETSNOW_TWOSPH_VEC()
      CALL MIE_SOAK_TWOSPH_WETGR_BH_VEC(x,graupel%a_geo,graupel%b_geo,fmelt,m_wv,m_iv,lambda,CBACK,CEXT,&
             mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
             mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, n_stuetz+1)

      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wetgr_twosph_mie_sb_vec

  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  ! de: replace m_i,m_w with T_a, add optional parameter Tmax -> needed for lookup table
  !     remove km, ntime
  SUBROUTINE zradar_wetgr_twosph_mie_vec(im,jm,L_g,T_a,&
       Tmeltbegin,meltdegTmin,lambda,n_0,a_geo,b_geo,Dmin,Dmax,Z,zeta,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
       mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater,Tmax)

    IMPLICIT NONE

    ! de: remove km, ntime
    INTEGER, INTENT(in) :: im, jm !, km, ntime
    ! de: add T_a and Tmax for lookup table
    DOUBLE PRECISION, INTENT(in)     :: L_g,T_a,lambda,n_0,a_geo,b_geo,Tmeltbegin,meltdegTmin
    DOUBLE PRECISION, INTENT(in), OPTIONAL :: Tmax
    ! de:
    !COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    ! de: add m_w, m_i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0, m_w, m_i
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,fmelt(n_stuetz+1),a_geo_m,b_geo_m,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6) THEN

      ! de: add calculation of m_w(T_a) and m_i(T_a)
      m_w = m_complex_water_ray(lambda,MIN(MAX(T_a-273.16d0,0.0d0),30.0d0))
      m_i = m_complex_ice_maetzler(lambda,MIN(MAX(T_a-273.16d0,-80.0d0),0.0d0))

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! Schmelzgradparametrisierung:
      ! de: add if-enquiry and add L_g and T_a (and Tmax) for lookup table; remove km, ntime
      IF (PRESENT(Tmax)) THEN
        ! de: with additional optional parameter Tmax for calculationg lookup table
        CALL degree_of_melting_gr(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt,Tmax)
      ELSE
        CALL degree_of_melting_gr(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt)
      END IF

      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_g) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0/b_geo
      a_geo_m = (1.0/a_geo)**b_geo_m

      x = a_geo * D_g ** b_geo
      ! Alternativen: MIE_WATERSPH_WETGR_VEC(), MIE_SOAK_WETGR_VEC(), 
      ! MIE_MEAN_WETGR_VEC(), MIE_WETSNOW_TWOSPH_VEC()
      CALL MIE_SOAK_TWOSPH_WETGR_BH_VEC(x,a_geo_m,b_geo_m,fmelt,m_wv,m_iv,lambda,CBACK,CEXT,&
           mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
           mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, n_stuetz+1)

      f_d = n_0 * EXP(-lam * D_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wetgr_twosph_mie_vec

  ! Version fuer schmelzenden Graupel beim 2M-Schema, watersphere version:
  SUBROUTINE zradar_wetgr_wsph_mie_sb_ve(im,jm,km,ntime,L_g,N_g,m_i,m_w,&
       Tmeltbegin,meltdegTmin,lambda,Dmin,Dmax,Z,zeta,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime
    DOUBLE PRECISION, INTENT(in) :: L_g,N_g,lambda,Tmeltbegin,meltdegTmin
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            D_g(n_stuetz+1),dD,fmelt(n_stuetz+1)

    INTEGER, SAVE        :: firstcall = 0
    DOUBLE PRECISION, SAVE :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_g >= 1.d-6 .AND. N_g >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = graupel%mu / graupel%b_geo
      nu_g = (graupel%nu + 1.0) / graupel%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+2.0)/graupel%mu) * L_g / N_g)**(-graupel%mu)
      n_0 = graupel%mu * N_g * lam**((graupel%nu+1.0)/graupel%mu) / gfct((graupel%nu+1.0)/graupel%mu)
      lam = lam / (graupel%a_geo**(mu_g))
      n_0 = n_0 / (graupel%b_geo * graupel%a_geo**((graupel%nu+1.0)/graupel%b_geo))

      ! Schmelzgradparametrisierung:
      CALL degree_of_melting_gr_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)

      x = (D_g/graupel%a_geo) ** (1.0/graupel%b_geo)
      ! Alternativen: MIE_SOAK_TWOSPH_WETGR_VEC(), MIE_SOAK_WETGR_VEC(), 
      ! MIE_MEAN_WETGR_VEC(), MIE_WETSNOW_TWOSPH_VEC()
      CALL MIE_WATERSPH_WETGR_BH_VEC(x,graupel%a_geo,graupel%b_geo,fmelt,m_wv,m_iv,lambda,CBACK,CEXT,&
           mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, n_stuetz+1)

      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wetgr_wsph_mie_sb_ve

  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  ! de: replace m_i,m_w with T_a, add optional parameter Tmax -> needed for lookup table
  !     remove km, ntime
  SUBROUTINE zradar_wetgr_watersph_mie_vec(im,jm,L_g,T_a,&
       Tmeltbegin,meltdegTmin,lambda,n_0,a_geo,b_geo,Dmin,Dmax,Z,zeta,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair,Tmax)

    IMPLICIT NONE

    ! de: remove km, ntime
    INTEGER, INTENT(in) :: im, jm !, km, ntime
    ! de: add T_a and Tmax for lookup table
    DOUBLE PRECISION, INTENT(in)     :: L_g,T_a,lambda,n_0,a_geo,b_geo,Tmeltbegin,meltdegTmin
    DOUBLE PRECISION, INTENT(in), OPTIONAL :: Tmax
    ! de:
    !COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    ! de: add m_w, m_i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0, m_w, m_i
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            dD,fmelt(n_stuetz+1),a_geo_m,b_geo_m,D_g(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF
    
    IF (L_g >= 1.d-6) THEN

      ! de: add calculation of m_w(T_a) and m_i(T_a)
      m_w = m_complex_water_ray(lambda,MIN(MAX(T_a-273.16d0,0.0d0),30.0d0))
      m_i = m_complex_ice_maetzler(lambda,MIN(MAX(T_a-273.16d0,-80.0d0),0.0d0))

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! Schmelzgradparametrisierung:
      ! de: add if-enquiry and add L_g and T_a (and Tmax) for lookup table; remove km, ntime
      IF (PRESENT(Tmax)) THEN
        ! de: with additional optional parameter Tmax for calculationg lookup table
        CALL degree_of_melting_gr(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt,Tmax)
      ELSE
        CALL degree_of_melting_gr(im,jm,L_g,T_a,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt)
      END IF
      
      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_g) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0/b_geo
      a_geo_m = (1.0/a_geo)**b_geo_m

      x = a_geo * D_g ** b_geo
      ! Alternativen: MIE_WATERSPH_WETGR_VEC(), MIE_SOAK_WETGR_VEC(), 
      ! MIE_MEAN_WETGR_VEC(), MIE_WETSNOW_TWOSPH_VEC()
      CALL MIE_WATERSPH_WETGR_BH_VEC(x,a_geo_m,b_geo_m,fmelt,m_wv,m_iv,lambda,CBACK,CEXT,&
           mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, n_stuetz+1)

      f_d = n_0 * EXP(-lam * D_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wetgr_watersph_mie_vec

  ! Version fuer schmelzenden Hagel fuer das 2M-Schema:
  SUBROUTINE zradar_wethail_mie_sb_vec(im,jm,km,ntime,L_h,N_h,m_i,m_w,&
       Tmeltbegin,meltdegTmin,lambda,Dmin,Dmax,Z,zeta,&
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime
    DOUBLE PRECISION , INTENT(in)    :: L_h,N_h,lambda,Tmeltbegin,meltdegTmin
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_w_0, m_iv(n_stuetz+1), m_wv(n_stuetz+1)
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            D_g(n_stuetz+1),dD,fwater_shell(n_stuetz+1),fmelt(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF
    
    IF (L_h >= 1.d-6 .AND. N_h >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = hail%mu / hail%b_geo
      nu_g = (hail%nu + 1.0) / hail%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((hail%nu+1.0)/hail%mu) / gfct((hail%nu+2.0)/hail%mu) * L_h / N_h)**(-hail%mu)
      n_0 = hail%mu * N_h * lam**((hail%nu+1.0)/hail%mu) / gfct((hail%nu+1.0)/hail%mu)
      lam = lam / (hail%a_geo**(mu_g))
      n_0 = n_0 / (hail%b_geo * hail%a_geo**((hail%nu+1.0)/hail%b_geo))

      ! Schmelzgradparametrisierung:
      CALL degree_of_melting_ha_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)
      fwater_shell = 0.5d0

      x = (D_g/hail%a_geo) ** (1.0/hail%b_geo)
      ! Alternativen: MIE_SPONGY_WETHAIL(), MIE_WATERSPH_WETHAIL(), MIE_SOAK_WETGR()
      CALL MIE_SPONGY_WETHAIL_BH_VEC(x,fmelt,fwater_shell,m_iv,m_wv,lambda,CBACK,CEXT, &
           mixingrulestring, matrixstring, inclusionstring, n_stuetz+1)

      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18
      
    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wethail_mie_sb_vec

  ! Entsprechende Routinen fuer Schnee (Zweischalenmodell, radienverh = r_innen / r_gesamt):
  ! D = a_geo * x ^ b_geo
  SUBROUTINE zradar_wetsnow_mie_sb_vec(im,jm,km,ntime,L_s,N_s,m_i,m_w,radienverh,&
       Tmeltbegin,meltdegTmin,lambda,Dmin,Dmax,Z,zeta,&
       mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
       mixingrulestring_core, matrixstring_core, inclusionstring_core, &
       hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
       hoststring_core, hostmatrixstring_core, hostinclusionstring_core)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: im, jm, km, ntime
    DOUBLE PRECISION, INTENT(in)     :: L_s,N_s,lambda, radienverh, Tmeltbegin, meltdegTmin
    COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: &
         mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
         mixingrulestring_core, matrixstring_core, inclusionstring_core
    CHARACTER(len=*), INTENT(in) :: &
         hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
         hoststring_core, hostmatrixstring_core, hostinclusionstring_core
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_w_0, m_iv(n_stuetz+1), m_wv(n_stuetz+1)
    DOUBLE PRECISION     :: mu_g,nu_g
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            D_g(n_stuetz+1),dD,fmelt(n_stuetz+1),meltratio_outside

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_s >= 1.d-6 .AND. N_s >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_g(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! n_0, lam, mu, nu auf Durchmesserverteilung umrechnen:
      mu_g = snow%mu / snow%b_geo
      nu_g = (snow%nu + 1.0) / snow%b_geo - 1.0
      ! lam, n_0 bezgl. Massenverteilung:
      lam = ( gfct((snow%nu+1.0)/snow%mu) / gfct((snow%nu+2.0)/snow%mu) * L_s / N_s)**(-snow%mu)
      n_0 = snow%mu * N_s * lam**((snow%nu+1.0)/snow%mu) / gfct((snow%nu+1.0)/snow%mu)
      lam = lam / (snow%a_geo**(mu_g))
      n_0 = n_0 / (snow%b_geo * snow%a_geo**((snow%nu+1.0)/snow%b_geo))


      ! Schmelzgradparametrisierung:
      CALL degree_of_melting_sn_sb(im,jm,km,ntime,D_g,n_stuetz+1,Tmeltbegin,meltdegTmin,fmelt)

      ! Verhaeltnis der an der Oberflaeche der inneren und aeusseren Kugelschale 
      ! wegschmelzenden Masse zur gesamten geschmolzenen Masse der jeweiligen Schale.
      ! Fuer beide Kugelschalen gleich.
      ! Der komplementaere Anteil schmilzt im Inneren des Partikels. Dieses Verhaeltnis nimmt vom hier
      ! vorzugebenden Wert mit zunehmendem Schmelzgrad linear auf den Wert 1.0 zu.
      meltratio_outside = 0.7d0

      x = (D_g/snow%a_geo) ** (1.0/snow%b_geo)
      ! Alternativen: MIE_WATERSPH_WETGR_VEC(), MIE_SOAK_TWOSPH_WETGR_VEC(), 
      ! MIE_SOAK_WETGR_VEC(), MIE_MEAN_WETGR_VEC()
      CALL MIE_WETSNOW_TWOSPH_VEC(x,snow%a_geo,snow%b_geo,fmelt,meltratio_outside,m_wv,m_iv,&
           lambda,radienverh,CBACK,CEXT,&
           mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
           mixingrulestring_core, matrixstring_core, inclusionstring_core, &
           hoststring_shell, &
           hostmatrixstring_shell, &
           hostinclusionstring_shell, &
           hoststring_core, &
           hostmatrixstring_core, &
           hostinclusionstring_core, n_stuetz+1)

      f_d = n_0 * D_g**nu_g * EXP(-lam * D_g**mu_g)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CBACK * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF


    RETURN
  END SUBROUTINE zradar_wetsnow_mie_sb_vec

  ! Version fuer die 1M-Schemata: Annahme einer Exponentialverteilung in D.
  ! x = a_geo * D ^ b_geo
  ! de: replace m_i,m_w with T_a, add optional parameter Tmax -> needed for lookup table
  !     remove km, ntime
  SUBROUTINE zradar_wetsnow_mie_vec(im,jm,L_s,T_a, &
       radienverh,Tmeltbegin,meltdegTmin,lambda,n_0,a_geo,b_geo,Dmin,Dmax,Z,zeta,&
       mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
       mixingrulestring_core, matrixstring_core, inclusionstring_core, &
       hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
       hoststring_core, hostmatrixstring_core, hostinclusionstring_core,Tmax)

    IMPLICIT NONE

    ! de: remove km, ntime
    INTEGER, INTENT(in) :: im, jm !, km, ntime
    ! de: add T_a and Tmax for lookup table
    DOUBLE PRECISION, INTENT(in)     :: L_s,T_a,lambda,radienverh,n_0,a_geo,b_geo,Tmeltbegin,meltdegTmin
    DOUBLE PRECISION, INTENT(in), OPTIONAL :: Tmax
    ! de: 
    !COMPLEX(kind=kinddp), INTENT(in) :: m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_core, matrixstring_core, inclusionstring_core, &
         mixingrulestring_shell, matrixstring_shell, inclusionstring_shell
    CHARACTER(len=*), INTENT(in) :: &
         hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
         hoststring_core, hostmatrixstring_core, hostinclusionstring_core
    DOUBLE PRECISION, INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    ! de: add m_w, m_i
    COMPLEX(kind=kinddp) :: m_iv(n_stuetz+1), m_wv(n_stuetz+1), m_w_0, m_w, m_i
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),x(n_stuetz+1),&
                            D_s(n_stuetz+1),dD,fmelt(n_stuetz+1),meltratio_outside,a_geo_m,b_geo_m

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w_0

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_s >= 1.d-6) THEN

      ! de: add calculation of m_w(T_a) and m_i(T_a)
      m_w = m_complex_water_ray(lambda,MIN(MAX(T_a-273.16d0,0.0d0),30.0d0))
      m_i = m_complex_ice_maetzler(lambda,MIN(MAX(T_a-273.16d0,-80.0d0),0.0d0))

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_s(i) = Dmin + (i-1) * dD
        m_iv(i) = m_i
        m_wv(i) = m_w
      END DO

      ! Schmelzgradparametrisierung:
      ! de: add if-enquiry and add L_s and T_a (and Tmax) for lookup table
      IF (PRESENT(Tmax)) THEN
        ! de: with additional optional parameter Tmax for calculating lookup table; remove km, ntime
        CALL degree_of_melting_sn(im,jm,L_s,T_a,D_s,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt,Tmax)
      ELSE
        CALL degree_of_melting_sn(im,jm,L_s,T_a,D_s,n_stuetz+1,Tmeltbegin,meltdegTmin,n_0,a_geo,b_geo,fmelt)
      END IF

      ! Verhaeltnis der an der Oberflaeche der inneren und aeusseren Kugelschale 
      ! wegschmelzenden Masse zur gesamten geschmolzenen Masse der jeweiligen Schale.
      ! Fuer beide Kugelschalen gleich.
      ! Der komplementaere Anteil schmilzt im Inneren des Partikels. Dieses Verhaeltnis nimmt vom hier
      ! vorzugebenden Wert mit zunehmendem Schmelzgrad linear auf den Wert 1.0 zu.
      meltratio_outside = 0.7d0

      lam = (n_0*a_geo*gfct(b_geo+1.0)/L_s) ** (1.0/(1.0+b_geo))
      b_geo_m = 1.0/b_geo
      a_geo_m = (1.0/a_geo)**b_geo_m

      x = a_geo * D_s ** b_geo
      ! Alternativen: MIE_WATERSPH_WETGR_VEC(), MIE_SOAK_TWOSPH_WETGR_VEC(), MIE_SOAK_WETGR_VEC(), MIE_MEAN_WETGR_VEC()
      CALL MIE_WETSNOW_TWOSPH_VEC(x,a_geo_m,b_geo_m,fmelt,meltratio_outside,m_wv,m_iv,&
           lambda,radienverh,CBACK,CEXT,&
           mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
           mixingrulestring_core, matrixstring_core, inclusionstring_core, &
           hoststring_shell, &
           hostmatrixstring_shell, &
           hostinclusionstring_shell, &
           hoststring_core, &
           hostmatrixstring_core, &
           hostinclusionstring_core, n_stuetz+1)

      f_d = n_0 * EXP(-lam * D_s)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w_0) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_wetsnow_mie_vec


  !----------------------------------------------------------------------------------------!
  !  Calculate radar reflectivity of rain/cloud droplets using full mie theory                            !
  !  as a function of mass density and number density                                      !
  !  assuming a gamma mass distribution (in case of 2-moment-scheme of Seifert and Beheng) or an
  !  exponential size distribution (in case of LM 1-moment-scheme)
  !
  !  The integration of the reflectivity integral is done in diameter-space
  !  by means of Simpson's rule. Diameter space is chosen
  !  since the distribution of equally spaced sampling points is better
  !  suited for the integrand function compared to mass space, in that more sampling
  !  points are within the "small drop" end of the size distribution and less sampling
  !  points at the "large drop" end, compared to mass space.
  !
  !  The procedure contains a minor inconsistency:
  !  To deduce the parameters of the DSD with respect to diameter, 
  !  it is assumed in accordance with the Microphysics parameterizations
  !  that L (and N in CASE of 2M) are representative
  !  for a DSD spanning over an infinite diameter range (to infinity).
  !  In contrast, when integrating over the spectrum, the integral is taken
  !  from Dmin to Dmax, neglecting the part to infinity.
  !  In fact, the latter is consistent with the physical argument, that no drops larger
  !  than a certain Dmax can exist for longer time because of spontaneous breakup, whereas it is
  !  felt that the assumption of an "infinite" diameter range for L (and N) in
  !  microphysics parameterizations is rather problematic. However, reflectivity
  !  is slightly underestimated, since the mass fraction above Dmax is not redistributed
  !  to smaller D, as it should be. This would be accomplished if DSD-Parameters 
  !  N0 and lambda would be derived by partial moments (=incomplete gamma functions) 
  !  rather than "normal"
  !  gamma functions. However, that would require an iterative Procedure since the
  !  bounds in the incomplete gamma functions themselves depend on lambda; this is omitted here.
  !----------------------------------------------------------------------------------------!

  !----------------------------------------------------------------------------------------!
  ! 1) Version for the 2-moment scheme of Seifert and Beheng:
  !----------------------------------------------------------------------------------------!

!!! Vektorisierung ueber die Stuetzstellen der numer. Integration.
!!! Andere moegliche Strategie: Vektorisierung ueber raeumlichen Index i,
!!! lookup table fuer den Rueckstreuquerschnitt als Funktion von D und T.
!!! Dann muesste m_w durch T als input-Parameter ersetzt werden, lookup table
!!! generation im ersten Aufruf dieser Routine.
  SUBROUTINE zradar_rain_mie_sb_vec(L_r,N_r,m_w,lambda,parti,Dmin,Dmax,Z,zeta)

    IMPLICIT NONE

    DOUBLE PRECISION,INTENT(in)     :: L_r,N_r,lambda
    COMPLEX(kind=kinddp),INTENT(in) :: m_w
    TYPE(particle),INTENT(in) :: parti
    DOUBLE PRECISION,INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    COMPLEX(kind=kinddp) :: m_w_0, m_wv(n_stuetz+1)
    DOUBLE PRECISION     :: mu_r,nu_r
    DOUBLE PRECISION     :: n_0,lam,f_d(n_stuetz+1),eta,Dmin,Dmax,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),C_sca(n_stuetz+1),x(n_stuetz+1),dD,D_w(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION, SAVE :: simpson(n_stuetz+1), K_w

    IF (firstcall /= 1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF

    IF (L_r >= 1.d-7 .AND. N_r >= 1.0d-3) THEN

      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_w(i)   = Dmin + (i-1) * dD
        m_wv(i)  = m_w
      END DO

      ! parti muss entweder rain oder cloud sein:
      mu_r = parti%mu
      nu_r = parti%nu

      lam = ( gfct((nu_r+1.0)/mu_r) / gfct((nu_r+2.0)/mu_r) * L_r / N_r)**(-mu_r)
      n_0 = mu_r * N_r * lam**((nu_r+1.0)/mu_r) / gfct((nu_r+1.0)/mu_r)

      CALL SPHERE_SCATTER_BH_VEC(D_w*0.5d0,m_wv,DBLE(m_air),lambda,CBACK,CEXT,C_sca,n_stuetz+1)

      x   = (D_w/parti%a_geo)**(1.0d0/parti%b_geo)
      f_d = n_0 * x**nu_r * EXP(-lam * x**mu_r) * pi * 0.5 * rho_w * D_w**2
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_rain_mie_sb_vec

  !----------------------------------------------------------------------------------------!
  ! 2) Version for the 1-moment schemes of LM:
  !----------------------------------------------------------------------------------------!

! de: replace m_w with T_a, needed for lookup table
  SUBROUTINE zradar_rain_mie_vec(L_r,T_a,lambda,Dmin,Dmax,Z,zeta,n_0,mu)

    IMPLICIT NONE

    ! de: add T_a
    DOUBLE PRECISION, INTENT(in)     :: L_r,T_a,lambda,n_0,Dmin,Dmax,mu
    ! de:
    !COMPLEX(kind=kinddp), INTENT(in) :: m_w
    DOUBLE PRECISION,INTENT(out) :: Z, zeta

    INTEGER, PARAMETER :: n_stuetz = 100 ! muss gerade sein wg. Simpson-Regel!!
    INTEGER    :: i
    ! de: add m_w
    COMPLEX(kind=kinddp) :: m_w_0, m_w, m_wv(n_stuetz+1)
    DOUBLE PRECISION     :: lam,f_d(n_stuetz+1),eta,basis(3)
    DOUBLE PRECISION     :: CBACK(n_stuetz+1),CEXT(n_stuetz+1),C_sca(n_stuetz+1),x(n_stuetz+1),dD,D_w(n_stuetz+1)

    INTEGER, SAVE      :: firstcall = 0
    DOUBLE PRECISION,  SAVE      :: simpson(n_stuetz+1), K_w

    IF (firstcall.NE.1) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      ! Koeffizientenvektor fuer die Simpson-Regel:
      basis = (/1.d0/3.d0, 4.d0/3.d0, 1.d0/3.d0/)
      simpson = 0.0d0
      DO i=1,n_stuetz-1,2
        simpson(i:i+2) = simpson(i:i+2) + basis
      END DO
      firstcall = 1
    ENDIF
      
    IF (L_r >= 1d-7) THEN

      ! de: add calculation of m_w(T_a)
      m_w = m_complex_water_ray(DBLE(lambda), MIN(MAX(T_a-273.16d0,0.0d0),30.0d0))
      dD   = (Dmax-Dmin)/n_stuetz
      DO i=1,n_stuetz+1
        D_w(i) = Dmin + (i-1) * dD
        m_wv(i) = m_w
      END DO

      lam = (pi*rho_w*n_0*gfct(mu+4.0d0)/(6.0d0*L_r))**(1.0d0/(mu+4.0d0))

      CALL SPHERE_SCATTER_BH_VEC(D_w*0.5d0,m_wv,DBLE(m_air),lambda,CBACK,CEXT,C_sca,n_stuetz+1)

      f_d = n_0 * D_w**mu * EXP(-lam * D_w)
      eta = SUM(f_d * CBACK * simpson * dD)
      zeta = SUM(f_d * CEXT * simpson * dD)

      Z = lambda**4 / (pi**5 * K_w) * eta * 1d18

    ELSE
      Z = 0.0d0
      zeta = 0.0d0
    END IF

    RETURN
  END SUBROUTINE zradar_rain_mie_vec
  
  !----------------------------------------------------------------------------------------!
  !  Calculate effective radar reflectivity of rain/cloud droplets using Rayleigh-         !
  !  Approximation as a function of mass density and number density                        !
  !  assuming a gamma distribution --- Version for the 2M-Scheme                           !
  !  Similar inconsistency as above: There is no rain drop larger than a certain Dmax,     !
  !  but an infinite diameter range for L (and N) is assumed in the microphysics           !
  !  parameterization and also in the calculation of radar reflectivity.                   !
  !  This may leed to larger values of reflectivity in case of long wavelengths (Rayleigh- !
  !  Approximation valid) compared to Mie-Calculations.                                    !
  !----------------------------------------------------------------------------------------!

  FUNCTION zradar_rain_rayleigh_sb_vec(L_r,N_r,m_w,lambda,parti,anz,is,ie) &
       RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz)
    DOUBLE PRECISION, INTENT(in) :: L_r(anz),N_r(anz),lambda
    TYPE(particle), INTENT(in) :: parti
    DOUBLE PRECISION :: Z(anz)

    INTEGER :: i
    COMPLEX(kind=kinddp) :: m_w_0
    TYPE(particle), SAVE :: parti_speicher
    DOUBLE PRECISION :: K_w(anz), K_w_0
    DOUBLE PRECISION, SAVE :: zr
    INTEGER, SAVE :: firstcall = 0

    ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
    IF (firstcall /= 1 .OR. parti_speicher%nu /= parti%nu .OR. parti_speicher%mu /= parti%mu) THEN
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      zr = gfct((2.0d0+parti%nu+1.0d0)/parti%mu)        &
           & / gfct((parti%nu+1.0d0)/parti%mu)        &
           & * ( gfct((parti%nu+1.0d0)/parti%mu)      &
           &   / gfct((parti%nu+2.0d0)/parti%mu) )**2 &
           & * (6.0d0/(pi*rho_w))**2 / K_w_0
      firstcall = 1
      parti_speicher = parti
    END IF

    Z = 0.0d0
    DO i=is,ie
      IF (L_r(i) >= 1d-7 .AND. N_r(i) >= 1.0d-3) THEN
        K_w(i)    = (ABS( (m_w(i)**2-1.0)/(m_w(i)**2+2.0) ))**2 
        Z(i) = zr * L_r(i)**2 / N_r(i) * 1d18 * K_w(i)
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_rain_rayleigh_sb_vec

  !----------------------------------------------------------------------------------------!
  !  Calculate effective radar reflectivity of rain using rayleigh-approximation           !
  !  as a function of mass density                                                         !
  !  assuming an exponential distribution --- Version for the 1M-Schemes                   !
  !----------------------------------------------------------------------------------------!

  FUNCTION zradar_rain_rayleigh(L_r,n_0,mu,lamexpc,gamfak1c,gamfak2c) &
       RESULT(Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_r,n_0,mu
    DOUBLE PRECISION, INTENT(in) :: lamexpc, gamfak1c, gamfak2c
    DOUBLE PRECISION :: Z

    DOUBLE PRECISION :: lam

    IF (L_r >= 1d-7) THEN
      lam  = (pi*rho_w*n_0/L_r*gamfak1c)**lamexpc
      Z = n_0 * gamfak2c * lam**(-mu-7.0d0) * 1d18
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_rain_rayleigh

  FUNCTION zradar_rain_rayleigh_vec(L_r,m_w,lambda,n_0,mu,anz,is,ie,mu_is_fixed) &
       RESULT(Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz)
    DOUBLE PRECISION, INTENT(in) :: L_r(anz),lambda,n_0(anz),mu(anz)
    DOUBLE PRECISION :: Z(anz)
    LOGICAL          :: mu_is_fixed  ! indicates that all values in "mu" are equal and constant over time
    INTEGER :: i
    DOUBLE PRECISION :: K_w, lam
    DOUBLE PRECISION, SAVE :: lamexpc, gamfak1c, gamfak2c
    DOUBLE PRECISION :: lamexp(anz), gamfak1(anz), gamfak2(anz)
    COMPLEX(kind=kinddp) :: m_w_0
    DOUBLE PRECISION, SAVE :: K_w_0
    INTEGER, SAVE :: firstcall = 0

    ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
    IF (firstcall /= 1) THEN
      m_w_0  = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2
      IF (mu_is_fixed) THEN
        IF (mu(1) /= 0.0d0) THEN
          lamexpc = 1.0 / (mu(1)+4.0d0)
          gamfak1c = gfct(mu(1)+4.0d0) / 6.0d0
          gamfak2c = gfct(mu(1)+7.0d0)
        ELSE
          lamexpc = 0.25d0
          gamfak1c = 1.0d0
          gamfak2c = 720.0d0
        END IF
      END IF
      firstcall = 1
    END IF

    IF (.NOT. mu_is_fixed) THEN

      IF (ANY(mu /= 0.0d0)) THEN
        lamexp = 1.0 / (mu+4.0d0)
        gamfak1 = gfctvec(mu+4.0d0, anz) / 6.0d0
        gamfak2 = gfctvec(mu+7.0d0, anz)
      ELSE
        lamexp = 0.25d0
        gamfak1 = 1.0d0
        gamfak2 = 720.0d0
      END IF

!CDIR NODEP
!CDIR ON_ADB(L_r,m_w,gamfak1,gamfak2,lamexp,n_0,mu)
      Z = 0.0d0
      DO i=is,ie
        IF (L_r(i) >= 1d-7) THEN
          K_w  = (ABS( (m_w(i)**2-1.0)/(m_w(i)**2+2.0) ))**2 
          lam  = (pi*rho_w*n_0(i)/L_r(i)*gamfak1(i))**lamexp(i)
          Z(i) = n_0(i) * gamfak2(i) * lam**(-mu(i)-7.0d0) * 1d18 * K_w / K_w_0
        ELSE
          Z(i) = 0.0d0
        END IF
      END DO

    ELSE

      Z = 0.0d0
!!$!CDIR NODEP
!!$!CDIR ON_ADB(L_r,n_0,mu)
      DO i=is,ie
        IF (L_r(i) >= 1d-7) THEN
          K_w  = (ABS( (m_w(i)**2-1.0)/(m_w(i)**2+2.0) ))**2 
          lam  = (pi*rho_w*n_0(i)/L_r(i)*gamfak1c)**lamexpc
          Z(i) = n_0(i) * gamfak2c * lam**(-mu(i)-7.0d0) * 1d18 * K_w / K_w_0
        ELSE
          Z(i) = 0.0d0
        END IF
      END DO

    END IF

    RETURN
  END FUNCTION zradar_rain_rayleigh_vec

  !----------------------------------------------------------------------------------------!
  !  Calculate effective radar reflectivity of cloud drops using rayleigh-approximation    !
  !  as a function of mass density                                                         !
  !  assuming an exponential distribution --- Version for the 1M-Schemes                   !
  !
  !  In case of LM 1M-Schemes, cloud droplets cannot be treated with the same routine
  !  as raindrops, since for cloud droplets, no size distribution is considered.
  !  Assumptions about the distribution have to be made here, in contrast to
  !  the rain case. We assume a generalized gamma distribution with
  !  the parameters mue and nue specified below.
  !
  !----------------------------------------------------------------------------------------!

  FUNCTION zradar_cloud_rayleigh(L_c,kwokw0xshapefact,D_c) &
       RESULT (Z)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: L_c,kwokw0xshapefact,D_c
    DOUBLE PRECISION :: Z

    DOUBLE PRECISION :: x_c

    IF (L_c >= 1d-7) THEN

        ! Hier muessen nun Annahmen ueber die Wolkentropfenverteilung
        ! gemacht werden. Wir nehmen eine gamma-Verteilung mit
        ! den (vorzugebenden) Parametern mu, nu und Durchmesser D_c an, 
        ! der dem Teilchen mit der mittleren Masse x_c entspricht.
        ! Somit ist die Reflektivitaet direkt proportional zum Fluessigwassergehalt.
        ! mu und nu sind bezueglich der Massen-Verteilungsfunktion zu nehmen:

!        D_c = 15d-6
      x_c = pi / 6.0 * rho_w * D_c**3
      Z = kwokw0xshapefact * L_c * x_c * 1d18
    ELSE
      Z = 0.0d0
    END IF

    RETURN
  END FUNCTION zradar_cloud_rayleigh

  FUNCTION zradar_cloud_rayleigh_vec(L_c,m_w,lambda,D_c,nu_D,mu_D,anz,is,ie) &
       RESULT (Z)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz)
    DOUBLE PRECISION, INTENT(in) :: L_c(anz),lambda,D_c,mu_D,nu_D
    DOUBLE PRECISION :: Z(anz)

    COMPLEX(kind=kinddp) :: m_w_0
    DOUBLE PRECISION :: K_w(anz), K_w_0, x_c(anz)
    DOUBLE PRECISION, SAVE :: zc, nu_c, mu_c, nuds, muds
    INTEGER, SAVE :: firstcall = 0
    INTEGER :: i

    ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
    IF (firstcall /= 1 .OR. nuds /= nu_D .OR. muds /= mu_D) THEN
      m_w_0 = m_complex_water_ray(lambda, 0.0d0)
      K_w_0  = (ABS( (m_w_0**2-1.0)/(m_w_0**2+2.0) ))**2 
      nu_c = (nu_D+1.0d0)/3.0d0 - 1.0d0
      mu_c = mu_D/3.0d0
      zc = gfct((2.0+nu_c+1.0)/mu_c)        &
           & / gfct((nu_c+1.0)/mu_c)        &
           & * ( gfct((nu_c+1.0)/mu_c)      &
           &   / gfct((nu_c+2.0)/mu_c) )**2 &
           & * (6.0/(pi*rho_w))**2 / K_w_0
      muds = mu_D
      nuds = nu_D
      firstcall = 1
    END IF

    Z = 0.0d0
    DO i=is,ie
      IF (L_c(i) >= 1d-7) THEN
        K_w(i)    = (ABS( (m_w(i)**2-1.0)/(m_w(i)**2+2.0) ))**2

        ! Hier muessen nun Annahmen ueber die Wolkentropfenverteilung
        ! gemacht werden. Wir nehmen eine gamma-Verteilung mit
        ! den (vorzugebenden) Parametern mu, nu und Durchmesser D_c an, 
        ! der dem Teilchen mit der mittleren Masse x_c entspricht.
        ! Somit ist die Reflektivitaet direkt proportional zum Fluessigwassergehalt.
        ! mu und nu sind bezueglich der Massen-Verteilungsfunktion zu nehmen:

!        D_c = 15d-6
        x_c(i) = pi / 6.0 * rho_w * D_c**3
        Z(i) = zc * L_c(i) * x_c(i) * 1d18 * K_w(i)
      ELSE
        Z(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zradar_cloud_rayleigh_vec

  ! de: calculation of extinction of radar reflectivity of cloud drops
  !     using equation described in Doviak/Zrnic (page 43)
  FUNCTION zext_cloud_rayleigh_vec(L_c,m_w,lambda,anz,is,ie) &
       RESULT (zeta)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, is, ie
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz)
    DOUBLE PRECISION, INTENT(in) :: L_c(anz),lambda
    DOUBLE PRECISION :: zeta(anz)
    DOUBLE PRECISION :: Im_K_w(anz)
    INTEGER :: i

    zeta = 0.0d0
    DO i=is,ie
      IF (L_c(i) >= 1d-7) THEN
        Im_K_w(i)  = AIMAG(-(m_w(i)**2-1.0)/(m_w(i)**2+2.0))
        zeta(i) = 6.0 * pi * L_c(i) / (rho_w * lambda) * Im_K_w(i)        
      ELSE
        zeta(i) = 0.0d0
      END IF
    END DO

    RETURN
  END FUNCTION zext_cloud_rayleigh_vec


!===================================================================================================
! de: begin lookup tables for radar reflectivity (subroutines for creating as well as interpolation)

  !*******************************************************************************
  !
  ! Create Lookup-table vectors for radar reflectivity and its extinction 
  ! for dry hydrometeors using Mie theory and 1-moment-scheme as a function
  ! of temperature (T_a) and specific content (q_i) of rain water (q_i=q_r),
  ! dry snow (q_i=q_s) and dry graupel (q_i=q_g)
  !
  !*******************************************************************************
  
  ! Saving and restoring lookup table to/from file prepared but not active due
  ! to missing MPI-exchange of table! Up to now, table is written from every processor
  ! and read from every processor, which is generally not a good idea!
  ! I'm not sure if I want to have MPI routines in this entire module ...

  ! NOTE: only for mu_rain and n0_r = const.! Otherwise a 3D/4D table is needed!
  SUBROUTINE zradar_rain_lookupcreate(look_Z_rain,nTa,nqr,lambda_radar,Dmin,Dmax,n0_r,mu_rain) !!$ ,savepath,unitnr)

    IMPLICIT NONE

    TYPE(dbzlookuptable), INTENT(out) :: look_Z_rain
    INTEGER, INTENT(in)               :: nTa, nqr  !!$, unitnr
    DOUBLE PRECISION, INTENT(in)      :: lambda_radar,Dmin, Dmax, n0_r, mu_rain
!!$    CHARACTER(len=*), INTENT(in) :: savepath

    DOUBLE PRECISION :: Z, zeta, magicnr
    INTEGER          :: i,j  !!$ ,ierr
!!$    CHARACTER(len=250) :: savefile
!!$    LOGICAL :: fileexist

    ! one more to include initial value Ta0 and qr0
    look_Z_rain%nTa = nTa+1
    look_Z_rain%nqi = nqr+1

    ! Allocate Memory for the table vectors
    ALLOCATE(look_Z_rain%T_a(look_Z_rain%nTa))
    ALLOCATE(look_Z_rain%q_i(look_Z_rain%nqi))
    ALLOCATE(look_Z_rain%q_i_lin(look_Z_rain%nqi))
    ALLOCATE(look_Z_rain%dbz(look_Z_rain%nqi,look_Z_rain%nTa))
    ALLOCATE(look_Z_rain%ext(look_Z_rain%nqi,look_Z_rain%nTa))


!!$    ! Put together the filename for the binary table storage file,
!!$    !   and, if it already exists, read table from file, otherwise
!!$    !   create table and write it to file:
!!$    magicnr = (nTa+nqr)*lambda_radar*(Dmin+Dmax)*n0_r
!!$    magicnr = (magicnr/10**CEILING(LOG10(ABS(magicnr)))+mu_rain)*1000.0
!!$    savefile(:) = ' '
!!$    IF (LEN_TRIM(savepath) > 0) THEN
!!$      WRITE(savefile, '(a,"/radar_mietab_rain_",f0.6,".bin")') TRIM(savepath), magicnr
!!$    ELSE
!!$      WRITE(savefile, '("radar_mietab_rain_",f0.6,".bin")') magicnr
!!$    END IF
!!$    INQUIRE(file=TRIM(savefile), exist=fileexist)
!!$
!!$    IF (.NOT. fileexist) THEN
!!$
!!$      ! NOTE: if you change something with the table vectors below,
!!$      !       please delete all old savefiles floating around on your system!

      ! intial and final value for lookup vector of temperature in K
      look_Z_rain%Ta0 = -40.0d0 + 273.16d0
      look_Z_rain%T_a(look_Z_rain%nTa) = 40.0d0 + 273.16d0
      ! intial and final value for lookup vector of specific rain water content
      ! in logarithmic space
      look_Z_rain%qi0 = -7.0d0
      look_Z_rain%q_i(look_Z_rain%nqi) = -1.0d0

      ! calculate increment
      look_Z_rain%dTa = (look_Z_rain%T_a(look_Z_rain%nTa) - look_Z_rain%Ta0)/nTa
      look_Z_rain%dqi = (look_Z_rain%q_i(look_Z_rain%nqi) - look_Z_rain%qi0)/nqr
      ! invert increment
      look_Z_rain%idTa = 1.0d0 / look_Z_rain%dTa
      look_Z_rain%idqi = 1.0d0 / look_Z_rain%dqi

      ! create lookup table vectors
      DO i=1, look_Z_rain%nqi
        look_Z_rain%q_i(i) = look_Z_rain%qi0 + (i-1)*look_Z_rain%dqi
        ! convert q_r in linear space since it is needed linearly in zradar_rain_mie_vec
        look_Z_rain%q_i_lin(i) = 10.0d0**look_Z_rain%q_i(i)
        DO j=1, look_Z_rain%nTa
          look_Z_rain%T_a(j) = look_Z_rain%Ta0 + (j-1)*look_Z_rain%dTa

          CALL  zradar_rain_mie_vec(look_Z_rain%q_i_lin(i),look_Z_rain%T_a(j),lambda_radar,Dmin,Dmax,Z,zeta,n0_r,mu_rain)
          look_Z_rain%dbz(i,j) = Z
          look_Z_rain%ext(i,j) = zeta

          ! convert reflectivity from linear space to logarithmic space in dBZ
          IF (look_Z_rain%dbz(i,j) >= 1d-9) THEN
            look_Z_rain%dbz(i,j) = 10.0d0 * LOG10(look_Z_rain%dbz(i,j))
          ELSE
            look_Z_rain%dbz(i,j) = -99.99
          END IF

        END DO
      END DO

!!$      OPEN(unitnr, file=TRIM(savefile), status='new', FORMAT='unformatted', iostat=ierr)
!!$      IF (ierr /= 0) THEN
!!$        WRITE (*,*) 'ERROR zradar_rain_lookupcreate: binary lookup-table file ' &
!!$             // TRIM(savefile)// ' could not be created!'
!!$        STOP
!!$      END IF
!!$
!!$      WRITE(unitnr) look_Z_rain%nTa
!!$      WRITE(unitnr) look_Z_rain%nqi
!!$      WRITE(unitnr) look_Z_rain%Ta0
!!$      WRITE(unitnr) look_Z_rain%qi0
!!$      WRITE(unitnr) look_Z_rain%dTa
!!$      WRITE(unitnr) look_Z_rain%dqi
!!$      WRITE(unitnr) look_Z_rain%idTa
!!$      WRITE(unitnr) look_Z_rain%idqi
!!$      WRITE(unitnr) look_Z_rain%T_a
!!$      WRITE(unitnr) look_Z_rain%q_i
!!$      WRITE(unitnr) look_Z_rain%q_i_lin
!!$      WRITE(unitnr) look_Z_rain%dbz
!!$      WRITE(unitnr) look_Z_rain%ext
!!$
!!$      CLOSE(unitnr)
!!$
!!$    ELSE
!!$
!!$      OPEN(unitnr, file=TRIM(savefile), status='old', FORMAT='unformatted', iostat=ierr)
!!$      IF (ierr /= 0) THEN
!!$        WRITE (*,*) 'ERROR zradar_rain_lookupcreate: binary lookup-table file ' &
!!$             // TRIM(savefile)// ' could not be opened!'
!!$        STOP
!!$      END IF
!!$
!!$      READ(unitnr) look_Z_rain%nTa
!!$      READ(unitnr) look_Z_rain%nqi
!!$      READ(unitnr) look_Z_rain%Ta0
!!$      READ(unitnr) look_Z_rain%qi0
!!$      READ(unitnr) look_Z_rain%dTa
!!$      READ(unitnr) look_Z_rain%dqi
!!$      READ(unitnr) look_Z_rain%idTa
!!$      READ(unitnr) look_Z_rain%idqi
!!$      READ(unitnr) look_Z_rain%T_a
!!$      READ(unitnr) look_Z_rain%q_i
!!$      READ(unitnr) look_Z_rain%q_i_lin
!!$      READ(unitnr) look_Z_rain%dbz
!!$      READ(unitnr) look_Z_rain%ext
!!$
!!$      CLOSE(unitnr)
!!$
!!$    END IF


    RETURN
  END SUBROUTINE zradar_rain_lookupcreate

  SUBROUTINE zradar_snow_lookupcreate(look_Z_snow,isnow_n0temp,nTa,nqs,lambda_radar,Dmin,Dmax,a_geo_s,b_geo_s, &
                                      Tmeltbegin,mixingrulestring_shell,matrixstring_shell,inclusionstring_shell, &
                                      mixingrulestring_core,matrixstring_core,inclusionstring_core)
    IMPLICIT NONE

    TYPE(dbzlookuptable), INTENT(out) :: look_Z_snow
    INTEGER, INTENT(in)               :: isnow_n0temp, nTa, nqs
    DOUBLE PRECISION, INTENT(in)      :: Tmeltbegin,lambda_radar,Dmin, Dmax, a_geo_s, b_geo_s
    CHARACTER(len=*), INTENT(in)      :: mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
                                         mixingrulestring_core, matrixstring_core, inclusionstring_core

    DOUBLE PRECISION                  :: n0_s, Z, zeta
    INTEGER                           :: i,j

    ! one more to include initial value Ta0 and qs0
    look_Z_snow%nTa = nTa+1
    look_Z_snow%nqi = nqs+1

    ! Allocate Memory for the table vectors
    ALLOCATE(look_Z_snow%T_a(look_Z_snow%nTa))
    ALLOCATE(look_Z_snow%q_i(look_Z_snow%nqi))
    ALLOCATE(look_Z_snow%q_i_lin(look_Z_snow%nqi))
    ALLOCATE(look_Z_snow%dbz(look_Z_snow%nqi,look_Z_snow%nTa))
    ALLOCATE(look_Z_snow%ext(look_Z_snow%nqi,look_Z_snow%nTa))

    ! intial and final value for lookup vector of temperature in K
    look_Z_snow%Ta0 = -80.0d0 + 273.16d0
!     look_Z_snow%T_a(look_Z_snow%nTa) = 0.0d0 + 273.16d0
    look_Z_snow%T_a(look_Z_snow%nTa) = Tmeltbegin
    ! intial and final value for lookup vector of specific snow content
    ! in logarithmic space
    look_Z_snow%qi0 = -6.0d0
    look_Z_snow%q_i(look_Z_snow%nqi) = -2.0d0
    
    ! calculate increment
    look_Z_snow%dTa = (look_Z_snow%T_a(look_Z_snow%nTa) - look_Z_snow%Ta0)/nTa
    look_Z_snow%dqi = (look_Z_snow%q_i(look_Z_snow%nqi) - look_Z_snow%qi0)/nqs
    ! invert increment
    look_Z_snow%idTa = 1.0d0 / look_Z_snow%dTa
    look_Z_snow%idqi = 1.0d0 / look_Z_snow%dqi

    ! create lookup table vectors
    DO i=1, look_Z_snow%nqi
      look_Z_snow%q_i(i) = look_Z_snow%qi0 + (i-1)*look_Z_snow%dqi
      ! convert q_s in linear space since it is needed linearly in zradar_snow_mie_vec
      look_Z_snow%q_i_lin(i) = 10.0d0**look_Z_snow%q_i(i)
      DO j=1, look_Z_snow%nTa
        look_Z_snow%T_a(j) = look_Z_snow%Ta0 + (j-1)*look_Z_snow%dTa
        ! calculate temperature dependent value of n0_s
        CALL calc_n0_snow(isnow_n0temp, look_Z_snow%T_a(j), look_Z_snow%q_i_lin(i), a_geo_s, n0_s)
        
        CALL zradar_snow_mie_vec(look_Z_snow%q_i_lin(i),look_Z_snow%T_a(j),0.5d0,lambda_radar,n0_s,a_geo_s,b_geo_s,&
                                 Dmin,Dmax,Z,zeta,mixingrulestring_shell,matrixstring_shell,inclusionstring_shell,&
                                 mixingrulestring_core,matrixstring_core,inclusionstring_core)
        look_Z_snow%dbz(i,j) = Z
        look_Z_snow%ext(i,j) = zeta

        ! convert reflectivity from linear space to logarithmic space in dBZ
        IF (look_Z_snow%dbz(i,j) >= 1d-9) THEN
          look_Z_snow%dbz(i,j) = 10.0d0 * LOG10(look_Z_snow%dbz(i,j))
        ELSE
          look_Z_snow%dbz(i,j) = -99.99
        END IF

      END DO
    END DO

    RETURN
  END SUBROUTINE zradar_snow_lookupcreate

  SUBROUTINE zradar_graupel_lookupcreate(look_Z_graupel,nTa,nqg,lambda_radar,Dmin,Dmax,n0_g,a_geo_g,b_geo_g,&
                                         Tmeltbegin,mixingrulestring,matrixstring,inclusionstring)
    IMPLICIT NONE

    TYPE(dbzlookuptable), INTENT(out) :: look_Z_graupel
    INTEGER, INTENT(in)               :: nTa, nqg
    DOUBLE PRECISION, INTENT(in)      :: Tmeltbegin,lambda_radar,Dmin, Dmax, n0_g, a_geo_g, b_geo_g
    CHARACTER(len=*), INTENT(in)      :: mixingrulestring, matrixstring, inclusionstring

    DOUBLE PRECISION :: Z, zeta
    INTEGER          :: i,j

    ! one more to include initial value Ta0 and qg0
    look_Z_graupel%nTa = nTa+1
    look_Z_graupel%nqi = nqg+1

    ! Allocate Memory for the table vectors
    ALLOCATE(look_Z_graupel%T_a(look_Z_graupel%nTa))
    ALLOCATE(look_Z_graupel%q_i(look_Z_graupel%nqi))
    ALLOCATE(look_Z_graupel%q_i_lin(look_Z_graupel%nqi))
    ALLOCATE(look_Z_graupel%dbz(look_Z_graupel%nqi,look_Z_graupel%nTa))
    ALLOCATE(look_Z_graupel%ext(look_Z_graupel%nqi,look_Z_graupel%nTa))

    ! intial and final value for lookup vector of temperature in K
    look_Z_graupel%Ta0 = -80.0d0 + 273.16d0
!     look_Z_graupel%T_a(look_Z_graupel%nTa) = -10.0d0 + 273.16d0
    look_Z_graupel%T_a(look_Z_graupel%nTa) = Tmeltbegin
    ! intial and final value for lookup vector of specific graupel content
    ! in logarithmic space
    look_Z_graupel%qi0 = -6.0d0
    look_Z_graupel%q_i(look_Z_graupel%nqi) = -1.0d0
    
    ! calculate increment
    look_Z_graupel%dTa = (look_Z_graupel%T_a(look_Z_graupel%nTa) - look_Z_graupel%Ta0)/nTa
    look_Z_graupel%dqi = (look_Z_graupel%q_i(look_Z_graupel%nqi) - look_Z_graupel%qi0)/nqg
    ! invert increment
    look_Z_graupel%idTa = 1.0d0 / look_Z_graupel%dTa
    look_Z_graupel%idqi = 1.0d0 / look_Z_graupel%dqi

    ! create lookup table vectors
    DO i=1, look_Z_graupel%nqi
      look_Z_graupel%q_i(i) = look_Z_graupel%qi0 + (i-1)*look_Z_graupel%dqi
      ! convert q_g in linear space since it is needed linearly in zradar_graupel_mie_vec
      look_Z_graupel%q_i_lin(i) = 10.0d0**look_Z_graupel%q_i(i)
      DO j=1, look_Z_graupel%nTa
        look_Z_graupel%T_a(j) = look_Z_graupel%Ta0 + (j-1)*look_Z_graupel%dTa
        
        CALL  zradar_graupel_mie_vec(look_Z_graupel%q_i_lin(i),look_Z_graupel%T_a(j),lambda_radar,n0_g,&
                                     a_geo_g,b_geo_g,Dmin,Dmax,Z,zeta,mixingrulestring,matrixstring,inclusionstring)
        look_Z_graupel%dbz(i,j) = Z
        look_Z_graupel%ext(i,j) = zeta

        ! convert reflectivity from linear space to logarithmic space in dBZ
        IF (look_Z_graupel%dbz(i,j) >= 1d-9) THEN
          look_Z_graupel%dbz(i,j) = 10.0d0 * LOG10(look_Z_graupel%dbz(i,j))
        ELSE
          look_Z_graupel%dbz(i,j) = -99.99
        END IF

      END DO
    END DO

    RETURN
  END SUBROUTINE zradar_graupel_lookupcreate

  !*******************************************************************************
  !
  ! Create Lookup-table vectors for radar reflectivity and its extinction for 
  ! melting hydrometeors using Mie theory and 1-moment-scheme as a function of
  ! temperature (T_a), specific content of snow (q_i=q_s) or graupel (q_i=q_g)
  ! and maximum temperature (Tmax) where all particles are assumed as totally melted.
  !
  !*******************************************************************************
  
    SUBROUTINE zradar_meltsnow_lookupcreate(look_Z_meltsnow,isnow_n0temp,nTa,nqs,nTm, &
                                            Tmeltbegin,meltdegTmin,lambda_radar,Dmin,Dmax,a_geo_s,b_geo_s, &
                                            mixingrulestring_shell,matrixstring_shell,inclusionstring_shell, &
                                            mixingrulestring_core,matrixstring_core,inclusionstring_core, &
                                            hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
                                            hoststring_core, hostmatrixstring_core, hostinclusionstring_core)
    IMPLICIT NONE

    TYPE(dbzlookuptable), INTENT(out) :: look_Z_meltsnow
    INTEGER, INTENT(in)               :: isnow_n0temp, nTa, nqs, nTm
    DOUBLE PRECISION, INTENT(in)      :: Tmeltbegin,meltdegTmin,lambda_radar,Dmin, Dmax, a_geo_s, b_geo_s
    CHARACTER(len=*), INTENT(in)      :: mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
                                         mixingrulestring_core, matrixstring_core, inclusionstring_core, &
                                         hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
                                         hoststring_core, hostmatrixstring_core, hostinclusionstring_core

    DOUBLE PRECISION                  :: n0_s, Z, zeta
    INTEGER                           :: i,j,k,idummy,jdummy

    ! one more to include initial value Ta0 and qs0 and Tm0
    look_Z_meltsnow%nTa = nTa+1
    look_Z_meltsnow%nqi = nqs+1
    look_Z_meltsnow%nTm = nTm+1

    ! Allocate Memory for the table vectors
    ALLOCATE(look_Z_meltsnow%T_a(look_Z_meltsnow%nTa))
    ALLOCATE(look_Z_meltsnow%q_i(look_Z_meltsnow%nqi))
    ALLOCATE(look_Z_meltsnow%q_i_lin(look_Z_meltsnow%nqi))
    ALLOCATE(look_Z_meltsnow%T_m(look_Z_meltsnow%nTm))
    ALLOCATE(look_Z_meltsnow%dbz_m(look_Z_meltsnow%nqi,look_Z_meltsnow%nTa,look_Z_meltsnow%nTm))
    ALLOCATE(look_Z_meltsnow%ext_m(look_Z_meltsnow%nqi,look_Z_meltsnow%nTa,look_Z_meltsnow%nTm))

    ! intial and final value for lookup vector of temperature in K
!     look_Z_meltsnow%Ta0 = 0.0d0 + 273.16d0
    look_Z_meltsnow%Ta0 = Tmeltbegin
    look_Z_meltsnow%T_a(look_Z_meltsnow%nTa) = 10.0d0 + 273.16d0
    ! intial and final value for lookup vector of specific snow content in logarithmic space
    look_Z_meltsnow%qi0 = -6.0d0
    look_Z_meltsnow%q_i(look_Z_meltsnow%nqi) = -2.0d0
    ! intial and final value for lookup vector of Tmax in K
    look_Z_meltsnow%Tm0 = 3.0d0 + 273.16d0
    look_Z_meltsnow%T_m(look_Z_meltsnow%nTm) = 10.0d0 + 273.16d0
    
    ! calculate increment
    look_Z_meltsnow%dTa = (look_Z_meltsnow%T_a(look_Z_meltsnow%nTa) - look_Z_meltsnow%Ta0)/nTa
    look_Z_meltsnow%dqi = (look_Z_meltsnow%q_i(look_Z_meltsnow%nqi) - look_Z_meltsnow%qi0)/nqs
    look_Z_meltsnow%dTm = (look_Z_meltsnow%T_m(look_Z_meltsnow%nTm) - look_Z_meltsnow%Tm0)/nTm
    ! invert increment
    look_Z_meltsnow%idTa = 1.0d0 / look_Z_meltsnow%dTa
    look_Z_meltsnow%idqi = 1.0d0 / look_Z_meltsnow%dqi
    look_Z_meltsnow%idTm = 1.0d0 / look_Z_meltsnow%dTm

    ! create lookup table vectors
    idummy = 1  ! only dummies, will not further be used in zradar_wetsnow_mie_vec()
    jdummy = 1
    DO i=1, look_Z_meltsnow%nqi
      look_Z_meltsnow%q_i(i) = look_Z_meltsnow%qi0 + (i-1)*look_Z_meltsnow%dqi
      ! convert q_s in linear space since it is needed linearly in zradar_wetsnow_mie_vec
      look_Z_meltsnow%q_i_lin(i) = 10.0d0**look_Z_meltsnow%q_i(i)
      DO j=1, look_Z_meltsnow%nTa
        look_Z_meltsnow%T_a(j) = look_Z_meltsnow%Ta0 + (j-1)*look_Z_meltsnow%dTa
        ! calculate temperature dependent value of n0_s
        CALL calc_n0_snow(isnow_n0temp, look_Z_meltsnow%T_a(j), look_Z_meltsnow%q_i_lin(i), a_geo_s, n0_s)
        DO k=1, look_Z_meltsnow%nTm
          look_Z_meltsnow%T_m(k) = look_Z_meltsnow%Tm0 + (k-1)*look_Z_meltsnow%dTm
          ! 0.5d0 = radienverh, Tmeltbegin = 273.16d0, meltdegTmin = 0.0d0
          CALL zradar_wetsnow_mie_vec(idummy,jdummy,look_Z_meltsnow%q_i_lin(i),look_Z_meltsnow%T_a(j),0.5d0,&
                                      Tmeltbegin,meltdegTmin,lambda_radar,n0_s,a_geo_s,b_geo_s,Dmin,Dmax,Z,zeta,&
                                      mixingrulestring_shell,matrixstring_shell,inclusionstring_shell,&
                                      mixingrulestring_core,matrixstring_core,inclusionstring_core,&
                                      hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell,&
                                      hoststring_core, hostmatrixstring_core, hostinclusionstring_core,&
                                      look_Z_meltsnow%T_m(k))
          look_Z_meltsnow%dbz_m(i,j,k) = Z
          look_Z_meltsnow%ext_m(i,j,k) = zeta

          ! convert reflectivity from linear space to logarithmic space in dBZ
          IF (look_Z_meltsnow%dbz_m(i,j,k) >= 1d-9) THEN
            look_Z_meltsnow%dbz_m(i,j,k) = 10.0d0 * LOG10(look_Z_meltsnow%dbz_m(i,j,k))
          ELSE
            look_Z_meltsnow%dbz_m(i,j,k) = -99.99
          END IF

        END DO
      END DO
    END DO

    RETURN
  END SUBROUTINE zradar_meltsnow_lookupcreate
  
  SUBROUTINE zradar_meltgraupel_lookupcreate(look_Z_meltgraupel,igraupel_type,nTa,nqg,nTm, &
                                             Tmeltbegin,meltdegTmin,lambda_radar,Dmin,Dmax,n0_g,a_geo_g,b_geo_g, &
                                             mixingrulestring,matrixstring,inclusionstring, &
                                             mixingrulestring_shell,matrixstring_shell,inclusionstring_shell, &
                                             hoststring, hostmatrixstring, hostinclusionstring)
    IMPLICIT NONE

    TYPE(dbzlookuptable), INTENT(out) :: look_Z_meltgraupel
    INTEGER, INTENT(in)               :: igraupel_type, nTa, nqg, nTm
    DOUBLE PRECISION, INTENT(in)      :: Tmeltbegin,meltdegTmin,lambda_radar,Dmin, Dmax, n0_g, a_geo_g, b_geo_g
    CHARACTER(len=*), INTENT(in)      :: mixingrulestring, matrixstring, inclusionstring, &
                                         mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
                                         hoststring, hostmatrixstring, hostinclusionstring

    DOUBLE PRECISION                  :: Z, zeta
    INTEGER                           :: i,j,k,idummy,jdummy

    ! one more to include initial value Ta0 and qg0 and Tm0
    look_Z_meltgraupel%nTa = nTa+1
    look_Z_meltgraupel%nqi = nqg+1
    look_Z_meltgraupel%nTm = nTm+1

    ! Allocate Memory for the table vectors
    ALLOCATE(look_Z_meltgraupel%T_a(look_Z_meltgraupel%nTa))
    ALLOCATE(look_Z_meltgraupel%q_i(look_Z_meltgraupel%nqi))
    ALLOCATE(look_Z_meltgraupel%q_i_lin(look_Z_meltgraupel%nqi))
    ALLOCATE(look_Z_meltgraupel%T_m(look_Z_meltgraupel%nTm))
    ALLOCATE(look_Z_meltgraupel%dbz_m(look_Z_meltgraupel%nqi,look_Z_meltgraupel%nTa,look_Z_meltgraupel%nTm))
    ALLOCATE(look_Z_meltgraupel%ext_m(look_Z_meltgraupel%nqi,look_Z_meltgraupel%nTa,look_Z_meltgraupel%nTm))
    
    ! intial and final value for lookup vector of temperature in K
!     look_Z_meltgraupel%Ta0 = -10.0d0 + 273.16d0
    look_Z_meltgraupel%Ta0 = Tmeltbegin
    look_Z_meltgraupel%T_a(look_Z_meltgraupel%nTa) = 15.0d0 + 273.16d0
    ! intial and final value for lookup vector of specific snow content in logarithmic space
    look_Z_meltgraupel%qi0 = -6.0d0
    look_Z_meltgraupel%q_i(look_Z_meltgraupel%nqi) = -1.5d0
    ! intial and final value for lookup vector of Tmax in K
    look_Z_meltgraupel%Tm0 = 3.0d0 + 273.16d0
    look_Z_meltgraupel%T_m(look_Z_meltgraupel%nTm) = 15.0d0 + 273.16d0
    
    ! calculate increment
    look_Z_meltgraupel%dTa = (look_Z_meltgraupel%T_a(look_Z_meltgraupel%nTa) - look_Z_meltgraupel%Ta0)/nTa
    look_Z_meltgraupel%dqi = (look_Z_meltgraupel%q_i(look_Z_meltgraupel%nqi) - look_Z_meltgraupel%qi0)/nqg
    look_Z_meltgraupel%dTm = (look_Z_meltgraupel%T_m(look_Z_meltgraupel%nTm) - look_Z_meltgraupel%Tm0)/nTm
    ! invert increment
    look_Z_meltgraupel%idTa = 1.0d0 / look_Z_meltgraupel%dTa
    look_Z_meltgraupel%idqi = 1.0d0 / look_Z_meltgraupel%dqi
    look_Z_meltgraupel%idTm = 1.0d0 / look_Z_meltgraupel%dTm

    ! create lookup table vectors
    idummy = 1  ! only dummies, will not further be used in zradar_wetgr_*_mie_vec()
    jdummy = 1
    DO i=1, look_Z_meltgraupel%nqi
      look_Z_meltgraupel%q_i(i) = look_Z_meltgraupel%qi0 + (i-1)*look_Z_meltgraupel%dqi
      ! convert q_g in linear space since it is needed linearly in zradar_wetgr*_mie_vec
      look_Z_meltgraupel%q_i_lin(i) = 10.0d0**look_Z_meltgraupel%q_i(i)
      DO j=1, look_Z_meltgraupel%nTa
        look_Z_meltgraupel%T_a(j) = look_Z_meltgraupel%Ta0 + (j-1)*look_Z_meltgraupel%dTa
        DO k=1, look_Z_meltgraupel%nTm
          look_Z_meltgraupel%T_m(k) = look_Z_meltgraupel%Tm0 + (k-1)*look_Z_meltgraupel%dTm
          ! Tmeltbegin = 263.16d0, meltdegTmin = 0.2d0
          SELECT CASE (igraupel_type)
            CASE(1)
              CALL zradar_wetgr_mie_vec(idummy,jdummy,look_Z_meltgraupel%q_i_lin(i),look_Z_meltgraupel%T_a(j),&
                                        Tmeltbegin,meltdegTmin,lambda_radar,&
                                        n0_g,a_geo_g,b_geo_g,Dmin,Dmax,Z,zeta,&
                                        mixingrulestring, matrixstring, inclusionstring, &
                                        hoststring, hostmatrixstring, hostinclusionstring,&
                                        look_Z_meltgraupel%T_m(k))
            CASE(2)
              CALL zradar_wetgr_twosph_mie_vec(idummy,jdummy,look_Z_meltgraupel%q_i_lin(i),look_Z_meltgraupel%T_a(j),&
                                               Tmeltbegin,meltdegTmin,lambda_radar,&
                                               n0_g,a_geo_g,b_geo_g,Dmin,Dmax,Z,zeta,&
                                               mixingrulestring, matrixstring, inclusionstring, &
                                               mixingrulestring_shell, matrixstring_shell, inclusionstring_shell,&
                                               look_Z_meltgraupel%T_m(k))
            CASE(3)
              CALL zradar_wetgr_watersph_mie_vec(idummy,jdummy,look_Z_meltgraupel%q_i_lin(i),look_Z_meltgraupel%T_a(j),&
                                                 Tmeltbegin,meltdegTmin,lambda_radar,&
                                                 n0_g,a_geo_g,b_geo_g,Dmin,Dmax,Z,zeta,&
                                                 mixingrulestring, matrixstring, inclusionstring,&
                                                 look_Z_meltgraupel%T_m(k))
          END SELECT
          look_Z_meltgraupel%dbz_m(i,j,k) = Z
          look_Z_meltgraupel%ext_m(i,j,k) = zeta

          ! convert reflectivity from linear space to logarithmic space in dBZ
          IF (look_Z_meltgraupel%dbz_m(i,j,k) >= 1d-9) THEN
            look_Z_meltgraupel%dbz_m(i,j,k) = 10.0d0 * LOG10(look_Z_meltgraupel%dbz_m(i,j,k))
          ELSE
            look_Z_meltgraupel%dbz_m(i,j,k) = -99.99
          END IF
        END DO
      END DO
    END DO

    RETURN
  END SUBROUTINE zradar_meltgraupel_lookupcreate

!===================================================================================================

  !*******************************************************************************
  !
  ! Retrieve values of radar reflectivity and its extinction
  ! for not melting hydrometeors (rain, dry snow, dry graupel,...)
  ! by bilinear interpolation from a lookup table as function of
  ! T_a and q_i(=q_r,q_s,q_g) for which the lookup table has been 
  ! created from the output of the subroutine zradar_*_lookupcreate
  ! (* = rain, snow, graupel)
  !
  !*******************************************************************************
  
  SUBROUTINE zradar_biinterpolate_lookup(look_Z_type,T_a,q_i,Z_radar_int,Z_ext_int)
    IMPLICIT NONE

    TYPE(dbzlookuptable), intent(in) :: look_Z_type
    DOUBLE PRECISION, INTENT(in)     :: T_a, q_i

    DOUBLE PRECISION, INTENT(out)    :: Z_radar_int, Z_ext_int

    INTEGER              :: iu, io, ju, jo
    DOUBLE PRECISION     :: Ta_loc, qi_loc

    IF (q_i >= look_Z_type%qi0) THEN

      ! Truncate q_i-values and T_a-values to the range of the table:
      qi_loc = MAX(MIN(q_i, look_Z_type%q_i(look_Z_type%nqi)), look_Z_type%qi0)
      Ta_loc = MAX(MIN(T_a, look_Z_type%T_a(look_Z_type%nTa)), look_Z_type%Ta0)

      ! calculate indices of the neighbouring regular q_i-values in the table:
      iu = MAX(MIN(FLOOR((qi_loc - look_Z_type%qi0) * look_Z_type%idqi)+1, look_Z_type%nqi-1),1)
      io = iu + 1

      ! calculate indices of the neighbouring regular T_a-values in the table:
      ju = MAX(MIN(FLOOR((Ta_loc - look_Z_type%Ta0) * look_Z_type%idTa)+1, look_Z_type%nTa-1),1)
      jo = ju + 1

      ! bilinearily interpolation of...
      ! ... a) reflectivity
      Z_radar_int = look_Z_type%dbz(iu,ju) * look_Z_type%idqi * look_Z_type%idTa &
           * (look_Z_type%q_i(io) - qi_loc) * (look_Z_type%T_a(jo) - Ta_loc) &
           + look_Z_type%dbz(io,ju) * look_Z_type%idqi * look_Z_type%idTa &
           * (qi_loc - look_Z_type%q_i(iu)) * (look_Z_type%T_a(jo) - Ta_loc) &
           + look_Z_type%dbz(iu,jo) * look_Z_type%idqi * look_Z_type%idTa &
           * (look_Z_type%q_i(io) - qi_loc) * (Ta_loc - look_Z_type%T_a(ju)) &
           + look_Z_type%dbz(io,jo) * look_Z_type%idqi * look_Z_type%idTa &
           * (qi_loc - look_Z_type%q_i(iu)) * (Ta_loc - look_Z_type%T_a(ju))

      ! ... b) extinction
      Z_ext_int   = look_Z_type%ext(iu,ju) * look_Z_type%idqi * look_Z_type%idTa &
           * (look_Z_type%q_i(io) - qi_loc) * (look_Z_type%T_a(jo) - Ta_loc) &
           + look_Z_type%ext(io,ju) * look_Z_type%idqi * look_Z_type%idTa &
           * (qi_loc - look_Z_type%q_i(iu)) * (look_Z_type%T_a(jo) - Ta_loc) &
           + look_Z_type%ext(iu,jo) * look_Z_type%idqi * look_Z_type%idTa &
           * (look_Z_type%q_i(io) - qi_loc) * (Ta_loc - look_Z_type%T_a(ju)) &
           + look_Z_type%ext(io,jo) * look_Z_type%idqi * look_Z_type%idTa &
           * (qi_loc - look_Z_type%q_i(iu)) * (Ta_loc - look_Z_type%T_a(ju))

    ELSE

      Z_radar_int = -99.99
      Z_ext_int   = 0.0

    END IF


    RETURN
  END SUBROUTINE zradar_biinterpolate_lookup

  !*******************************************************************************
  !
  ! Retrieve values of radar reflectivity and its extinction
  ! for melting hydrometeors (snow, graupel,...)
  ! by trilinear interpolation from a lookup table as function of
  ! T_a and q_i(=q_r,q_s,q_g) and T_m for which the lookup table has been 
  ! created from the output of the subroutine zradar_*_lookupcreate
  ! (* = wetsnow, wetgraupel)
  !
  !*******************************************************************************
  
  SUBROUTINE zradar_triinterpolate_lookup(look_Z_type,T_a,q_i,T_m,Z_radar_int,Z_ext_int)
    IMPLICIT NONE

    TYPE(dbzlookuptable), intent(in) :: look_Z_type
    DOUBLE PRECISION, INTENT(in)     :: T_a, q_i, T_m

    DOUBLE PRECISION, INTENT(out)    :: Z_radar_int, Z_ext_int

    INTEGER              :: iu, io, ju, jo, ku, ko
    DOUBLE PRECISION     :: Ta_loc, qi_loc, Tm_loc
    DOUBLE PRECISION     :: Ta_du, qi_du, Tm_du
    DOUBLE PRECISION     :: Ta_do, qi_do, Tm_do

    IF (q_i >= look_Z_type%qi0) THEN

      ! Trunkcate given values of q_i, T_a and T_m to the range of the table:
      qi_loc = MAX(MIN(q_i, look_Z_type%q_i(look_Z_type%nqi)), look_Z_type%qi0)
      Ta_loc = MAX(MIN(T_a, look_Z_type%T_a(look_Z_type%nTa)), look_Z_type%Ta0)
      Tm_loc = MAX(MIN(T_m, look_Z_type%T_m(look_Z_type%nTm)), look_Z_type%Tm0)

      ! calculate indices of the neighbouring regular q_i-values in the table:
      iu = MAX(MIN(FLOOR((qi_loc - look_Z_type%qi0) * look_Z_type%idqi)+1, look_Z_type%nqi-1),1)
      io = iu + 1

      ! calculate indices of the neighbouring regular T_a-values in the table:
      ju = MAX(MIN(FLOOR((Ta_loc - look_Z_type%Ta0) * look_Z_type%idTa)+1, look_Z_type%nTa-1),1)
      jo = ju + 1

      ! calculate indices of the neighbouring regular T_m-values in the table:
      ku = MAX(MIN(FLOOR((Tm_loc - look_Z_type%Tm0) * look_Z_type%idTm)+1, look_Z_type%nTm-1),1)
      ko = ku + 1

      ! calculate difference to the lower coordinate
      qi_du = (qi_loc - look_Z_type%q_i(iu))*look_Z_type%idqi
      Ta_du = (Ta_loc - look_Z_type%T_a(ju))*look_Z_type%idTa
      Tm_du = (Tm_loc - look_Z_type%T_m(ku))*look_Z_type%idTm

      ! calculate difference to the upper coordinate
      qi_do = (look_Z_type%q_i(io) - qi_loc)*look_Z_type%idqi
      Ta_do = (look_Z_type%T_a(jo) - Ta_loc)*look_Z_type%idTa
      Tm_do = (look_Z_type%T_m(ko) - Tm_loc)*look_Z_type%idTm 

      ! trilinearily interpolation of...
      ! ... a) reflectivity
      Z_radar_int = ((look_Z_type%dbz_m(iu,ju,ku)*qi_do + look_Z_type%dbz_m(io,ju,ku)*qi_du)*Ta_do + &
           (look_Z_type%dbz_m(iu,jo,ku)*qi_do + look_Z_type%dbz_m(io,jo,ku)*qi_du)*Ta_du)*Tm_do + &
           ((look_Z_type%dbz_m(iu,ju,ko)*qi_do + look_Z_type%dbz_m(io,ju,ko)*qi_du)*Ta_do + &
           (look_Z_type%dbz_m(iu,jo,ko)*qi_do + look_Z_type%dbz_m(io,jo,ko)*qi_du)*Ta_du)*Tm_du

      ! ... b) extinction
      Z_ext_int   = ((look_Z_type%ext_m(iu,ju,ku)*qi_do + look_Z_type%ext_m(io,ju,ku)*qi_du)*Ta_do + &
           (look_Z_type%ext_m(iu,jo,ku)*qi_do + look_Z_type%ext_m(io,jo,ku)*qi_du)*Ta_du)*Tm_do + &
           ((look_Z_type%ext_m(iu,ju,ko)*qi_do + look_Z_type%ext_m(io,ju,ko)*qi_du)*Ta_do + &
           (look_Z_type%ext_m(iu,jo,ko)*qi_do + look_Z_type%ext_m(io,jo,ko)*qi_du)*Ta_du)*Tm_du 

    ELSE

      Z_radar_int = -99.99
      Z_ext_int   = 0.0

    END IF

    RETURN
  END SUBROUTINE zradar_triinterpolate_lookup

! de. end lookup tables for radar reflectivity (subroutines for creating as well as interpolation)
!===================================================================================================

  SUBROUTINE decode_controlstring_3C(ctype_relrefind, &
       mixingrulestring, matrixstring, inclusionstring, &
       hoststring, hostmatrixstring, hostinclusionstring, &
       ctype_default)

    IMPLICIT NONE

    CHARACTER(len=6), INTENT(in) :: ctype_relrefind, ctype_default
    CHARACTER(len=*), INTENT(out) :: &
         mixingrulestring, matrixstring, inclusionstring, &
         hoststring, hostmatrixstring, hostinclusionstring

    INTEGER(kind=iintegers) :: i
    CHARACTER(len=6) :: buf(2)

    buf(1:2) = REPEAT(' ',6)
    buf(1) = ctype_default(1:6)
    buf(2) = ctype_relrefind(1:6)

    ! Ausgabestrings initialisieren:
    mixingrulestring    = REPEAT(' ',LEN(mixingrulestring))
    matrixstring        = REPEAT(' ',LEN(matrixstring))
    inclusionstring     = REPEAT(' ',LEN(inclusionstring))
    hoststring          = REPEAT(' ',LEN(hoststring))
    hostmatrixstring    = REPEAT(' ',LEN(hostmatrixstring))
    hostinclusionstring = REPEAT(' ',LEN(hostinclusionstring))

    ! Im ersten Durchlauf defaults setzen, im 2. Durchlauf die eigentlichen Strings:
    DO i=1,2

      SELECT CASE (buf(i)(1:1))
      CASE('m','M')
        mixingrulestring = 'maxwellgarnett'
      CASE('o','O','w','W')
        mixingrulestring = 'oguchi'
      CASE('b','B')
        mixingrulestring = 'bruggemann'
      CASE default
        WRITE (*,*) '  decode_controlstrings_3C: Ungueltiges ctype_relrefind, 1. Zeichen: ', TRIM(buf(i))
      END SELECT

      SELECT CASE (buf(i)(2:2))
      CASE('n','N')
        hoststring = 'none'
      CASE('a','A')
        hoststring = 'air'
      CASE('i','I')
        hoststring = 'ice'
      CASE('w','W')
        hoststring = 'water'
      CASE default
        WRITE (*,*) '  decode_controlstrings_3C: Ungueltiges ctype_relrefind, 2. Zeichen: ', TRIM(buf(i))
      END SELECT

      SELECT CASE (buf(i)(3:3))
      CASE('a','A')
        matrixstring = 'air'
      CASE('i','I')
        matrixstring = 'ice'
      CASE('w','W')
        matrixstring = 'water'
      CASE default
        WRITE (*,*) '  decode_controlstrings_3C: Ungueltiges ctype_relrefind, 3. Zeichen: ', TRIM(buf(i))
      END SELECT

      SELECT CASE (buf(i)(4:4))
      CASE('s','S')
        inclusionstring = 'spheroidal'
      CASE('k','K')
        inclusionstring = 'spherical'
      CASE default
        WRITE (*,*) '  decode_controlstrings_3C: Ungueltiges ctype_relrefind, 4. Zeichen: ', TRIM(buf(i))
      END SELECT

      SELECT CASE (buf(i)(5:5))
      CASE('a','A')
        hostmatrixstring = 'air'
      CASE('i','I')
        hostmatrixstring = 'ice'
      CASE('w','W')
        hostmatrixstring = 'water'
      CASE('m','M')
        hostmatrixstring = 'icewater'
      CASE('s','S')
        hostmatrixstring = 'airice'
      CASE('r','R')
        hostmatrixstring = 'airwater'
      CASE default
        WRITE (*,*) '  decode_controlstrings_3C: Ungueltiges ctype_relrefind, 5. Zeichen: ', TRIM(buf(i))
      END SELECT

      SELECT CASE (buf(i)(6:6))
      CASE('s','S')
        hostinclusionstring = 'spheroidal'
      CASE('k','K')
        hostinclusionstring = 'spherical'
      CASE default
        WRITE (*,*) '  decode_controlstrings_3C: Ungueltiges ctype_relrefind, 6. Zeichen: ', TRIM(buf(i))
      END SELECT

    END DO

    RETURN
  END SUBROUTINE decode_controlstring_3C

  SUBROUTINE decode_controlstring_2C(ctype_relrefind, &
       mixingrulestring, matrixstring, inclusionstring, &
       ctype_default)

    IMPLICIT NONE

    CHARACTER(len=3), INTENT(in) :: ctype_relrefind, ctype_default
    CHARACTER(len=*), INTENT(out) :: &
         mixingrulestring, matrixstring, inclusionstring

    INTEGER(kind=iintegers) :: i
    CHARACTER(len=3) :: buf(2)

    buf(1:2) = REPEAT(' ',3)
    buf(1) = ctype_default(1:3)
    buf(2) = ctype_relrefind(1:3)

    ! Ausgabestrings initialisieren:
    mixingrulestring    = REPEAT(' ',LEN(mixingrulestring))
    matrixstring        = REPEAT(' ',LEN(matrixstring))
    inclusionstring     = REPEAT(' ',LEN(inclusionstring))

    ! Im ersten Durchlauf defaults setzen, im 2. Durchlauf die eigentlichen Strings:
    DO i=1,2

      SELECT CASE (buf(i)(1:1))
      CASE('m','M')
        mixingrulestring = 'maxwellgarnett'
      CASE('o','O','w','W')
        mixingrulestring = 'oguchi'
      CASE('b','B')
        mixingrulestring = 'bruggemann'
      CASE default
        WRITE (*,*) '  decode_controlstrings_2C: Ungueltiges ctype_relrefind, 1. Zeichen: ', TRIM(buf(i))
      END SELECT

      SELECT CASE (buf(i)(2:2))
      CASE('a','A')
        matrixstring = 'air'
      CASE('i','I')
        matrixstring = 'ice'
      CASE('w','W')
        matrixstring = 'water'
      CASE default
        WRITE (*,*) '  decode_controlstrings_2C: Ungueltiges ctype_relrefind, 2. Zeichen: ', TRIM(buf(i))
      END SELECT

      SELECT CASE (buf(i)(3:3))
      CASE('s','S')
        inclusionstring = 'spheroidal'
      CASE('k','K')
        inclusionstring = 'spherical'
      CASE default
        WRITE (*,*) '  decode_controlstrings_2C: Ungueltiges ctype_relrefind, 3. Zeichen: ', TRIM(buf(i))
      END SELECT

    END DO

    RETURN
  END SUBROUTINE decode_controlstring_2C


  !=======================================================================
  !
  ! Interface-Funktionen zur Berechnung der Radarreflektivitaet nach Mie,
  ! die in src_output.f90 aufgerufen werden. Diese Routinen benutzen
  ! wiederum viele Unterprogramme, die im weiteren Verlauf des vorliegenden
  ! Moduls definiert werden.
  ! qx in kg/kg, qnx in 1/kg !!!
  !=======================================================================

#ifdef TWOMOMSEIFERT
  SUBROUTINE radar_sb_mie_vec(ntime, klv850, myproc, lambda_radar, igraupel_type, &
       ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel, ctype_dryhail, ctype_wethail, &
       Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, Tmeltbegin_h, meltdegTmin_h, &
       z_radar, z_radar_850, z_radar_cmax, z_ext, linc_boundlines )

    ! ctype_drysnow: 6 Zeichen
    ! ctype_wetsnow: 12 Zeichen
    ! ctype_drygraupel: 3 Zeichen
    ! ctype_wetgraupel: 6 Zeichen
    ! ctype_dryhail: 3 Zeichen (Eis-Luft-Mischung)
    ! ctype_wethail: 3 Zeichen (Wasser-Eis-Mischung der Schale, Kern wird als Eiskugel angenommen; Ausgangspunkt des Schmelzens
    !  ist ebenfalls eine Eiskugel, im Gegensatz zum trockenen Hagel, wo auch Lufteinschluesse beruecksichtigt werden.)

    INTEGER(KIND=iintegers), INTENT(IN)       :: ntime, igraupel_type

    INTEGER(KIND=iintegers), INTENT(IN)       :: klv850, myproc

    REAL(KIND=ireals)      , INTENT(IN)       :: lambda_radar, &
         Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, Tmeltbegin_h, meltdegTmin_h

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar(ie,je,ke), z_ext(ie,je,ke)

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar_850(ie,je),   &
         z_radar_cmax(ie,je)

    CHARACTER(len=*), INTENT(in) :: &
         ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel, ctype_dryhail, ctype_wethail

    LOGICAL, INTENT(in), OPTIONAL             :: linc_boundlines

    ! Local Variables
    !----------------

    DOUBLE PRECISION, PARAMETER :: eps=1d-15, q_krit_radar=1d-7
    DOUBLE PRECISION ::  z_local(ie,je,ke), z_ext_local(ie,je,ke)

    INTEGER :: i, j, k, ku, ko
    
    DOUBLE PRECISION, DIMENSION(ie) :: T_a, fmelt_mean, Dtmp, &
         q_r, n_r, x_r,                  &
         q_c, n_c, x_c,                  &
         q_g, n_g, x_g,                  &
         q_s, n_s, x_s,                  &
         q_i, n_i, x_i,                  &
         q_h, n_h, x_h, hlp
    DOUBLE PRECISION :: Dmin, Dmax, Z, zeta, &
         neigh, meltdegTmin, Tmeltbegin

    COMPLEX(kind=kinddp), DIMENSION(ie) :: m_i, m_w

    CHARACTER(len=20) :: &
         mixingrulestring_s_shell, matrixstring_s_shell, inclusionstring_s_shell, &
         mixingrulestring_s_core, matrixstring_s_core, inclusionstring_s_core, &
         hoststring_s_shell, hostmatrixstring_s_shell, hostinclusionstring_s_shell, &
         hoststring_s_core, hostmatrixstring_s_core, hostinclusionstring_s_core, &
         mixingrulestring_s_shell_dry, matrixstring_s_shell_dry, inclusionstring_s_shell_dry, &
         mixingrulestring_s_core_dry, matrixstring_s_core_dry, inclusionstring_s_core_dry, &
         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
         mixingrulestring_g_shell, matrixstring_g_shell, inclusionstring_g_shell, &
         hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry, &
         mixingrulestring_h, matrixstring_h, inclusionstring_h, &
         mixingrulestring_h_dry, matrixstring_h_dry, inclusionstring_h_dry


    ! .. Set ilow, iup, jlow, jup, klow, kup for computations in this sub-domain
    !    depending on the optional switch "linc_boundlines":
    IF ( PRESENT(linc_boundlines) ) THEN
      CALL get_loc_domain( zlinc_boundlines = linc_boundlines )
    ELSE
      ! .. Default: no computations on interior boundlines to save time
      CALL get_loc_domain( zlinc_boundlines = .FALSE.)
    END IF
    
    ! Nachbarschaft bei der Bestimmung des Tmax-Parameters fuer die Schmelzgradparametrisierung:
    neigh = 5000.0d0 ! m
    CALL initialize_tmax_sb_vec_par(ntime, neigh)

    ! Steuerstrings fuer den effektiven Brechungsindex generieren:
    CALL decode_controlstring_2C(ctype_drysnow(1:3), &
         mixingrulestring_s_core_dry, matrixstring_s_core_dry, inclusionstring_s_core_dry, &
         'mas')
    CALL decode_controlstring_2C(ctype_drysnow(4:6), &
         mixingrulestring_s_shell_dry, matrixstring_s_shell_dry, inclusionstring_s_shell_dry, &
         'mas')
    CALL decode_controlstring_3C(ctype_wetsnow(1:6), &
         mixingrulestring_s_core, matrixstring_s_core, inclusionstring_s_core, &
         hoststring_s_core, hostmatrixstring_s_core, hostinclusionstring_s_core, &
         'mawsas')
    CALL decode_controlstring_3C(ctype_wetsnow(7:12), &
         mixingrulestring_s_shell, matrixstring_s_shell, inclusionstring_s_shell, &
         hoststring_s_shell, hostmatrixstring_s_shell, hostinclusionstring_s_shell, &
         'mawsas')
    CALL decode_controlstring_2C(ctype_drygraupel(1:3), &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry, &
         'mas')
    SELECT CASE (igraupel_type)
    CASE(1)
      CALL decode_controlstring_3C(ctype_wetgraupel(1:6), &
           mixingrulestring_g, matrixstring_g, inclusionstring_g, &
           hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
           'mawsas')
    CASE(2)
      CALL decode_controlstring_2C(ctype_wetgraupel(1:3), &
           mixingrulestring_g, matrixstring_g, inclusionstring_g, &
           'mis')
      CALL decode_controlstring_2C(ctype_wetgraupel(4:6), &
           mixingrulestring_g_shell, matrixstring_g_shell, inclusionstring_g_shell, &
           'mis')
    CASE(3)
      CALL decode_controlstring_2C(ctype_wetgraupel(1:3), &
           mixingrulestring_g, matrixstring_g, inclusionstring_g, &
           'mis')    
    END SELECT
    CALL decode_controlstring_2C(ctype_dryhail(1:3), &
         mixingrulestring_h_dry, matrixstring_h_dry, inclusionstring_h_dry, &
         'mis')
    CALL decode_controlstring_2C(ctype_wethail(1:3), &
         mixingrulestring_h, matrixstring_h, inclusionstring_h, &
         'mws')
   
    z_local = 0.0d0
    z_ext_local = 0.0d0

    IF (PRESENT(z_radar) .OR. PRESENT(z_radar_cmax)) THEN
      ku = klow
      ko = kup
    ELSE
      IF (PRESENT(z_radar_850)) THEN
        ku = klv850
        ko = klv850
      ELSE
        ! Mache einfach gar nix!
        ku = 2
        ko = 1
      END IF
    END IF

    DO k= ku, ko
      PRINT *, 'Progress: k=', k-ku+1, ' of ',ko-ku+1
      DO j = jlow, jup

        T_a = 273.16d0
        DO i = ilow, iup
          T_a(i) = t(i,j,k,ntime)
          hlp(i) = rho(i,j,k)
        END DO

        ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
        m_w = (1.0d0,0.0d0)
        m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
        ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
        ! Eis aber trotzdem am Gefrierpunkt verharren.
        m_i = (1.0d0,0.0d0)
        m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)

        DO i = ilow, iup
          q_r(i) = qr(i,j,k,ntime) * hlp(i)
          n_r(i) = qnr(i,j,k,ntime) * hlp(i)
          q_c(i) = qc(i,j,k,ntime) * hlp(i)
          n_c(i) = qnc(i,j,k,ntime) * hlp(i)
          q_g(i) = qg(i,j,k,ntime) * hlp(i)
          n_g(i) = qng(i,j,k,ntime) * hlp(i)
          q_s(i) = qs(i,j,k,ntime) * hlp(i)
          n_s(i) = qns(i,j,k,ntime) * hlp(i)
          q_i(i) = qi(i,j,k,ntime) * hlp(i)
          n_i(i) = qni(i,j,k,ntime) * hlp(i)
          x_r(i) = MIN( MAX(q_r(i)/(n_r(i)+eps),rain%x_min),rain%x_max )
          n_r(i) = q_r(i) / x_r(i)
          x_g(i) = MIN( MAX(q_g(i)/(n_g(i)+eps),graupel%x_min),graupel%x_max )
          n_g(i) = q_g(i) / x_g(i)
          x_s(i) = MIN( MAX(q_s(i)/(n_s(i)+eps),snow%x_min),snow%x_max )
          n_s(i) = q_s(i) / x_s(i)
          x_i(i) = MIN( MAX(q_i(i)/(n_i(i)+eps),ice%x_min),ice%x_max )
          n_i(i) = q_i(i) / x_i(i)
          x_c(i) = MIN( MAX(q_c(i)/(n_c(i)+eps),cloud%x_min),cloud%x_max )
          n_c(i) = q_c(i) / x_c(i)
          IF (q_s(i) < q_krit_radar) q_s(i) = 0.0d0
          IF (q_r(i) < q_krit_radar) q_r(i) = 0.0d0
          IF (q_g(i) < q_krit_radar) q_g(i) = 0.0d0
          IF (q_i(i) < q_krit_radar) q_i(i) = 0.0d0
          IF (q_c(i) < q_krit_radar) q_c(i) = 0.0d0
        END DO

        ! Reflektivitaet von Wolkentropfen nach Rayleigh-Approximation:        
        IF (MAXVAL(q_c(ilow:iup)) >= q_krit_radar) & 
             z_local(:,j,k) = z_local(:,j,k) + &
             zradar_rain_rayleigh_sb_vec(q_c,n_c,m_w,lambda_radar,cloud,ie,ilow,iup)

        ! Reflektivitaet von Regen nach Mie (Integration bzgl. Durchmesser):
        Dmin=0.5d-4 ! Untegrenze Integrationsintervall der Reflektivitaet: 0.05 mm
        Dmax=10.0d-3 ! Obergrenze: 8.0 mm
        DO i=ilow, iup
          !!! Vektorisierung ueber die Stuetzstellenanzahl der numer. Integraion. 
          !!! Auch andere Strategien sind denkbar, z.B. Vektorisierung ueber Index i,
          !!! Stuetzstellen in aeussere Schleife, QBACk mittels table lookup als
          !!! Funktion von D und T.
          IF (q_r(i) >= q_krit_radar) THEN
               CALL zradar_rain_mie_sb_vec(q_r(i),n_r(i),m_w(i),lambda_radar,rain,Dmin,Dmax,Z,zeta)
               z_local(i,j,k) = z_local(i,j,k) + Z
               z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
          END IF
        END DO


        ! Reflektivitaet von Wolkeneis nach Rayleigh:
        Tmeltbegin = 273.16d0
        meltdegTmin = 0.0d0

        IF (MAXVAL(q_i(ilow:iup)) >= q_krit_radar) THEN
          IF (MAXVAL(T_a(ilow:iup)) > Tmeltbegin) THEN
            fmelt_mean = 0.0d0
            Dtmp(ilow:iup) = D_average_sb_vec(q_i(ilow:iup), &
                 n_i(ilow:iup), ice, iup-ilow+1) * 2.0d0
            CALL degree_of_melting_ice_sb_vec_ie(ilow,iup,j,k,ntime,&
                 Dtmp,ie,Tmeltbegin,meltdegTmin,fmelt_mean)
            z_local(:,j,k) = z_local(:,j,k) + &
                 zradar_wetgr_rayl_oguchi_sb_vec(q_i,n_i,m_i,m_w,lambda_radar,fmelt_mean,ice,ie,ilow,iup)
          ELSE
            z_local(:,j,k) = z_local(:,j,k) + &
                 zradar_gr_rayl_oguchi_sb_vec(q_i,n_i,m_i,lambda_radar,ice,ie,ilow,iup)           
          END IF
        END IF

        ! Reflektivitaet von Schnee nach Mie (Integration bzgl. Durchmesser):
!        Tmeltbegin = 273.16d0
!        meltdegTmin = 0.0d0
        Tmeltbegin = Tmeltbegin_s
        meltdegTmin = meltdegTmin_s
        DO i=ilow, iup
          IF (T_a(i) > Tmeltbegin) THEN
            Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
            Dmax=30.0d-3 ! Obergrenze: 20.0 mm
            IF (q_s(i) >= q_krit_radar) THEN
                 CALL zradar_wetsnow_mie_sb_vec(i,j,k,ntime,q_s(i),n_s(i),m_i(i),m_w(i),0.5d0,&
                      Tmeltbegin,meltdegTmin,lambda_radar,Dmin,Dmax,Z,zeta,&
                      mixingrulestring_s_shell, matrixstring_s_shell, inclusionstring_s_shell, &
                      mixingrulestring_s_core, matrixstring_s_core, inclusionstring_s_core, &
                      hoststring_s_shell, hostmatrixstring_s_shell, hostinclusionstring_s_shell, &
                      hoststring_s_core, hostmatrixstring_s_core, hostinclusionstring_s_core)
                 z_local(i,j,k) = z_local(i,j,k) + Z
                 z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
            END IF
          ELSE
            Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
            Dmax=30.0d-3 ! Obergrenze: 20.0 mm
            IF (q_s(i) >= q_krit_radar) THEN
                 CALL zradar_snow_mie_sb_vec(q_s(i),n_s(i),m_i(i),0.5d0,lambda_radar,Dmin,Dmax,Z,zeta, &
                      mixingrulestring_s_shell_dry, matrixstring_s_shell_dry, inclusionstring_s_shell_dry, &
                      mixingrulestring_s_core_dry, matrixstring_s_core_dry, inclusionstring_s_core_dry)
                 z_local(i,j,k) = z_local(i,j,k) + Z
                 z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
            END IF
          END IF
        END DO


        ! Reflektivitaet von Graupel nach Mie (Integration bzgl. Durchmesser):
!        Tmeltbegin = 263.16d0
!        meltdegTmin = 0.2d0
        Tmeltbegin = Tmeltbegin_g
        meltdegTmin = meltdegTmin_g
        DO i=ilow, iup
          IF (q_g(i) >= q_krit_radar) THEN
            IF (T_a(i) > Tmeltbegin) THEN
              Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
              Dmax=30.0d-3 ! Obergrenze: 15.0 mm
              SELECT CASE (igraupel_type)
              CASE(1)
                CALL zradar_wetgr_mie_sb_vec(i,j,k,ntime,q_g(i),n_g(i),m_i(i),m_w(i),Tmeltbegin,meltdegTmin,&
                     lambda_radar,Dmin,Dmax,Z,zeta,&
                     mixingrulestring_g, matrixstring_g, inclusionstring_g, &
                     hoststring_g, hostmatrixstring_g, hostinclusionstring_g)
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              CASE(2)
                CALL zradar_wetgr_twosph_mie_sb_vec(i,j,k,ntime,q_g(i),n_g(i),m_i(i),m_w(i),&
                     Tmeltbegin,meltdegTmin,lambda_radar,Dmin,Dmax,Z,zeta,&
                     mixingrulestring_g, matrixstring_g, inclusionstring_g, &
                     mixingrulestring_g_shell, matrixstring_g_shell, inclusionstring_g_shell)
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              CASE(3)
                CALL zradar_wetgr_wsph_mie_sb_ve(i,j,k,ntime,q_g(i),n_g(i),m_i(i),m_w(i),&
                     Tmeltbegin,meltdegTmin,lambda_radar,Dmin,Dmax,Z,zeta,&
                     mixingrulestring_g, matrixstring_g, inclusionstring_g)
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              END SELECT
            ELSE
              Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.3 mm
              Dmax=30.0d-3 ! Obergrenze: 15.0 mm
              CALL zradar_graupel_mie_sb_vec(q_g(i),n_g(i),m_i(i),lambda_radar,Dmin,Dmax,Z,zeta,&
                   mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry)
              z_local(i,j,k) = z_local(i,j,k) + Z
              z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
            END IF
          END IF
        END DO

        ! Reflektivitaet von Hagel nach Mie (Integration bzgl. Durchmesser):
        IF (ALLOCATED(qh)) THEN

          DO i=ilow, iup
            q_h(i) = qh(i,j,k,ntime) * hlp(i)
            n_h(i) = qnh(i,j,k,ntime) * hlp(i)
            x_h(i) = MIN( MAX(q_h(i)/(n_h(i)+eps),hail%x_min),hail%x_max )
            n_h(i) = q_h(i) / x_h(i)
            IF (q_h(i) < q_krit_radar) q_h(i) = 0.0d0
          END DO

!          Tmeltbegin = 263.16d0
!          meltdegTmin = 0.2d0
          Tmeltbegin = Tmeltbegin_h
          meltdegTmin = meltdegTmin_h

          DO i=ilow, iup
            IF (q_h(i) >= q_krit_radar) THEN
              IF (T_a(i) > Tmeltbegin) THEN
                Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
                Dmax=30.0d-3 ! Obergrenze: 15.0 mm
                ! Spongy Hail --- Zweischalenmodell mit Eiskugel und Haut aus Eis-Wasser mit Volumenverh. Wasser/Gesamt = 0.7
                CALL zradar_wethail_mie_sb_vec(i,j,k,ntime,q_h(i),n_h(i),m_i(i),m_w(i),Tmeltbegin,meltdegTmin,&
                     lambda_radar,Dmin,Dmax,Z,zeta,&
                     mixingrulestring_h, matrixstring_h, inclusionstring_h)
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              ELSE
                Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.3 mm
                Dmax=30.0d-3 ! Obergrenze: 15.0 mm
                CALL zradar_hail_mie_sb_vec(q_h(i),n_h(i),m_i(i),lambda_radar,Dmin,Dmax,Z,zeta,&
                     mixingrulestring_h_dry, matrixstring_h_dry, inclusionstring_h_dry)
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              END IF
            END IF

          END DO

        END IF  ! IF (allocated(qh))

      END DO
    END DO

    WHERE (z_local >= 1d-9)
      z_local = 10.0d0 * LOG10(z_local)
    ELSEWHERE
      z_local = -99.99
    END WHERE
    WRITE (6,'(2(A,F10.3))') "  MAX dBZ = ",MAXVAL(z_local),"  MIN dBZ = ",MINVAL(z_local)

    IF ( PRESENT(z_radar) ) THEN
      z_radar = z_local
    ENDIF
    IF ( PRESENT(z_ext) ) THEN
      z_ext = z_ext_local
    ENDIF

    IF ( PRESENT(z_radar_850) ) THEN
      DO j = jlow, jup
        DO i = ilow, iup
          z_radar_850(i,j) = z_local(i,j,klv850)
        ENDDO
      ENDDO
    ENDIF

    IF ( PRESENT(z_radar_cmax) ) THEN
      z_radar_cmax = MAXVAL(z_local,3)
    ENDIF

    CALL finalize_tmax()

    RETURN
  END SUBROUTINE radar_sb_mie_vec
#endif

  ! Version fuer die 1M-Schemata:
  SUBROUTINE radar_lm_mie_vec(ntime, klv850, myproc, lambda_radar, itype_gscp, isnow_n0temp, igraupel_type, &
       ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel, &
       Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, llookup, &
       z_radar, z_radar_850, z_radar_cmax, z_ext, linc_boundlines )

    ! ctype_drysnow: 3 Zeichen
    ! ctype_wetsnow: 6 Zeichen
    ! ctype_drygraupel: 3 Zeichen
    ! ctype_wetgraupel: 6 Zeichen

    ! COSMO specific:
    USE environment, ONLY:  get_free_unit, release_unit

    IMPLICIT NONE

    INTEGER(KIND=iintegers), INTENT(IN)       :: ntime, igraupel_type, itype_gscp, isnow_n0temp

    INTEGER(KIND=iintegers), INTENT(IN)       :: klv850, myproc

    REAL(KIND=ireals)      , INTENT(IN)       :: lambda_radar, &
         Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar(ie,je,ke), z_ext(ie,je,ke)

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar_850(ie,je), z_radar_cmax(ie,je)

    CHARACTER(len=*), INTENT(in) :: &
         ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel

    LOGICAL, INTENT(in)                       :: llookup
    LOGICAL, INTENT(in), OPTIONAL             :: linc_boundlines

    ! Local Variables
    !----------------

    DOUBLE PRECISION, PARAMETER :: q_krit_radar=1d-7
    DOUBLE PRECISION ::  z_local(ie,je,ke), z_ext_local(ie,je,ke)
    ! de: add nTa, nqr, nqs, nqg, nTm number of bins in lookup table vector
    INTEGER :: i, j, k, ku, ko, nTa, nqr, nqs, nqg, nTm
    ! de: add logarithmic value for q_r,q_s,q_g ; add T_m 
    DOUBLE PRECISION                :: T_a, T_m,         &
                                       fmelt_mean, Dtmp, &
                                       q_r,q_r_log,      &
                                       q_c,              &
                                       q_g,q_g_log,      &
                                       q_s,q_s_log,      &
                                       q_i,              &
                                       hlp

    DOUBLE PRECISION :: Dmin_r, Dmax_r, &
                        Dmin_s, Dmax_s, &
                        Dmin_g, Dmax_g

    DOUBLE PRECISION :: Z, zeta, nu_c, mu_c, D_c, a_geo, b_geo, neigh, &
         n0_r(ie,je,ke), n0_s(ie,je,ke), n0_g, mu_rain(ie,je,ke), &
         ageoi, bgeoi, ageos, bgeos, ageog, bgeog, meltdegTmin_i, Tmeltbegin_i, K_w, K_i

    COMPLEX(kind=kinddp), DIMENSION(ie,je,ke) :: m_i, m_w

    CHARACTER(len=20) :: &
         mixingrulestring_s_shell, matrixstring_s_shell, inclusionstring_s_shell, &
         mixingrulestring_s_core, matrixstring_s_core, inclusionstring_s_core, &
         hoststring_s_shell, hostmatrixstring_s_shell, hostinclusionstring_s_shell, &
         hoststring_s_core, hostmatrixstring_s_core, hostinclusionstring_s_core, &
         mixingrulestring_s_shell_dry, matrixstring_s_shell_dry, inclusionstring_s_shell_dry, &
         mixingrulestring_s_core_dry, matrixstring_s_core_dry, inclusionstring_s_core_dry, &
         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
         mixingrulestring_g_shell, matrixstring_g_shell, inclusionstring_g_shell, &
         hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry

    LOGICAL :: mur_is_fixed, n0r_is_fixed

    LOGICAL :: lalloc_qi, lalloc_qs, lalloc_qg

    TYPE(dbzlookuptable), SAVE :: look_Z_rain
    TYPE(dbzlookuptable), SAVE :: look_Z_snow
    TYPE(dbzlookuptable), SAVE :: look_Z_graupel
    TYPE(dbzlookuptable), SAVE :: look_Z_meltsnow
    TYPE(dbzlookuptable), SAVE :: look_Z_meltgraupel
    
    INTEGER, SAVE              :: firstcall_rain = 0
    INTEGER, SAVE              :: firstcall_snow = 0
    INTEGER, SAVE              :: firstcall_graupel = 0
    INTEGER, SAVE              :: firstcall_meltsnow = 0
    INTEGER, SAVE              :: firstcall_meltgraupel = 0
    ! de.

    ! .. Set ilow, iup, jlow, jup, klow, kup for computations in this sub-domain
    !    depending on the optional switch "linc_boundlines":
    IF ( PRESENT(linc_boundlines) ) THEN
      CALL get_loc_domain( zlinc_boundlines = linc_boundlines )
    ELSE
      ! .. Default: no computations on interior boundlines to save time
      CALL get_loc_domain( zlinc_boundlines = .FALSE.)
    END IF
    
    ! Parameter der Hydrometeorarten und deren Verteilungsfunktionen:
    CALL hydroparams_lm(ntime, itype_gscp, isnow_n0temp, &
         ageoi, &
         bgeoi, &
         ageos, &
         bgeos, &
         ageog, &
         bgeog, &
         n0_r, &
         n0_s, &
         n0_g, mu_rain, mur_is_fixed, n0r_is_fixed)


    ! Nachbarschaft bei der Bestimmung des Tmax-Parameters fuer die Schmelzgradparametrisierung:
    ! (domain including boundlines)
    neigh = 5000.0d0 ! m
    CALL initialize_tmax_vec_par(ntime, neigh)

    ! Steuerstrings fuer den effektiven Brechungsindex generieren:
    CALL decode_controlstring_2C(ctype_drysnow(1:3), &
         mixingrulestring_s_core_dry, matrixstring_s_core_dry, inclusionstring_s_core_dry, &
         'mas')
    CALL decode_controlstring_2C(ctype_drysnow(4:6), &
         mixingrulestring_s_shell_dry, matrixstring_s_shell_dry, inclusionstring_s_shell_dry, &
         'mas')
    CALL decode_controlstring_3C(ctype_wetsnow(1:6), &
         mixingrulestring_s_core, matrixstring_s_core, inclusionstring_s_core, &
         hoststring_s_core, hostmatrixstring_s_core, hostinclusionstring_s_core, &
         'mawsas')
    CALL decode_controlstring_3C(ctype_wetsnow(7:12), &
         mixingrulestring_s_shell, matrixstring_s_shell, inclusionstring_s_shell, &
         hoststring_s_shell, hostmatrixstring_s_shell, hostinclusionstring_s_shell, &
         'mawsas')
    CALL decode_controlstring_2C(ctype_drygraupel(1:3), &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry, &
         'mas')
    SELECT CASE (igraupel_type)
    CASE(1)
      CALL decode_controlstring_3C(ctype_wetgraupel(1:6), &
           mixingrulestring_g, matrixstring_g, inclusionstring_g, &
           hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
           'mawsas')
    CASE(2)
      CALL decode_controlstring_2C(ctype_wetgraupel(1:3), &
           mixingrulestring_g, matrixstring_g, inclusionstring_g, &
           'mis')
      CALL decode_controlstring_2C(ctype_wetgraupel(4:6), &
           mixingrulestring_g_shell, matrixstring_g_shell, inclusionstring_g_shell, &
           'mis')
    CASE(3)
      CALL decode_controlstring_2C(ctype_wetgraupel(1:3), &
           mixingrulestring_g, matrixstring_g, inclusionstring_g, &
           'mis')
    END SELECT


    z_local = 0.0d0
    z_ext_local = 0.0d0

    IF (PRESENT(z_radar) .OR. PRESENT(z_radar_cmax)) THEN
      ku = klow
      ko = kup
    ELSE
      IF (PRESENT(z_radar_850)) THEN
        ku = klv850
        ko = klv850
      ELSE
        ! Mache einfach gar nix!
        ku = 2
        ko = 1
      END IF
    END IF


    lalloc_qi = ALLOCATED(qi)
    lalloc_qs = ALLOCATED(qs)
    lalloc_qg = ALLOCATED(qg)


    ! .. Initializations for PSD-parameters for cloud drops and cloud ice,
    !    since there are no such assumptions in the COSMO microphysics:

    ! Cloud drops: generalized gamma (original COSMO assumes no explicit DSD)
    ray_const%D_c = 20d-6  ! Durchmesser des Teilchens mit der mittleren Masse in m
    ray_const%nu_Dc = 3.0d0       ! nu bezgl. Durchmesserverteilung
    ray_const%mu_Dc = 3.0d0       ! mu bezgl. Durchmesserverteilung

    ! Cloud ice: generalized gamma (original COSMO is however monodisperse)
    ray_const%D_i = 100d-6 ! m, Durchmesser des Partikels mit der mittleren Masse
    ray_const%nu_Di = (-0.3333333d0 + 1.0d0) * bgeoi - 1.0d0   ! entspricht -0.33333 bezgl. Massenverteilung
    ray_const%mu_Di = 0.3333333d0 * bgeoi   ! entspricht  0.33333 bezgl. Massenverteilung
    Tmeltbegin_i = 273.16d0
    meltdegTmin_i = 0.0d0

    ! On its first call or if otherwise necessary, this computes 
    ! some constant prefactors needed for cloud drops and cloud ice
    ! (Rayleigh-Oguchi) below
    ! and stores it in global struct "ray_const":
    ! (The values for rain, snow and graupel are not used below)
    CALL init_radar_lm_ray_oguchi (  &
         .TRUE., .FALSE., .TRUE., .FALSE., .FALSE.,  &
         DBLE(lambda_radar), mur_is_fixed, mu_rain(1,1,1), &
         ageoi, bgeoi, ageos, bgeos, ageog, bgeog, n0_g)


    ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
    m_w = RESHAPE(m_complex_water_ray_vec(DBLE(lambda_radar),&
         MIN(MAX(RESHAPE(t(:,:,:,ntime),(/ie*je*ke/))-273.16d0,0.0d0),30.0d0),ie*je*ke), &
         (/ie,je,ke/) )
    ! Brechungsindex von Eis nach C. Maetzler, gueltig fuer -80 < T < 0 Grad C
    m_i = RESHAPE(m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
         MIN(MAX(RESHAPE(t(:,:,:,ntime),(/ie*je*ke/))-273.16d0,-80.0d0),0.0d0), ie*je*ke), &
         (/ie,je,ke/) )


    ! .. Reflectivity of cloud droplets and cloud ice with simple Rayleigh-Oguchi-Theory:
    DO k = ku, ko
      DO j = jlow, jup
        DO i = ilow, iup

          T_a = t(i,j,k,ntime)
          hlp = rho(i,j,k)

          q_c = qc(i,j,k,ntime) * hlp
          IF (q_c < q_krit_radar) q_c = 0.0d0
          
          K_w    = (ABS( (m_w(i,j,k)**2-1.0)/(m_w(i,j,k)**2+2.0) ))**2
          
          z_local(i,j,k) = z_local(i,j,k) + &
               zradar_cloud_rayleigh(q_c,K_w*ray_const%prefactor_c,ray_const%D_c)

          IF (lalloc_qi) THEN

            q_i = qi(i,j,k,ntime) * hlp
            IF (q_i < q_krit_radar) q_i = 0.0d0

            K_i    = (ABS( (m_i(i,j,k)**2-1.0)/(m_i(i,j,k)**2+2.0) ))**2

            IF (T_a > Tmeltbegin_i) THEN
              fmelt_mean = 0.0d0
              Dtmp = ray_const%D_i * 2.0d0 ! m, Durchmesser des Teilchens, das zur Bestimmung des mittl. Schmelzgrades 
              ! herangezogen wird.
              
              CALL degree_of_melting_ice_single(i,j,T_a,&
                   Dtmp,Tmeltbegin_i,meltdegTmin_i,100d-6,fmelt_mean)
              
              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_wetice_rayl_oguchi(q_i, m_w(i,j,k), m_i(i,j,k), &
                   fmelt_mean, ray_const%D_i, ageoi, bgeoi, ray_const%prefactor_im)
            ELSE
              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_ice_rayleigh_oguchi(q_i, &
                   ray_const%D_i, ageoi, bgeoi, K_i*ray_const%prefactor_i)           
            END IF
            
          END IF

        END DO
      END DO
    END DO


    !==================================================================================
    !
    ! .. Reflectivity of rain, snow and graupel with full Mie-Theory:
    !    (Integration with respect to particle diameter)
    !
    !==================================================================================

    ! .. Integration limits for the integrals of cross sections over the PSDs:

    Dmin_r =  0.5d-4 ! Rain: 0.05mm - 6.0 mm
    Dmax_r = 10.0d-3
            
    Dmin_s =  0.5d-4 ! Snow: 0.05mm - 30.0 mm
    Dmax_s = 30.0d-3
            
    Dmin_g =  0.5d-4 ! Graupel: 0.05mm - 30.0 mm
    Dmax_g = 30.0d-3


    ! .. Creating lookup tables and calculating reflectivity:

    IF (llookup) THEN

      ! de: create lookup table for rain (only first call)
      IF (firstcall_rain /= 1) THEN
!        IF (MAXVAL(qr(:,:,:,ntime)*rho) >= q_krit_radar) THEN
        IF (mur_is_fixed .AND. n0r_is_fixed) THEN
          firstcall_rain = 1 
          nTa = 20
          nqr = 30
          CALL zradar_rain_lookupcreate(look_Z_rain,nTa,nqr,DBLE(lambda_radar), &
                                        Dmin_r,Dmax_r,n0_r(ilow,jlow,ku),mu_rain(ilow,jlow,ku))
        ELSE
          WRITE (*,*) 'ERROR radar_lm_mie_vec: use of rain lookup tables only '// &
                             'implemented for const. mu_rain and n0_r! Stop!'
          STOP
        END IF
!        END IF
      END IF
          
      ! de: create lookup table for melting snow (only first call)
      IF (firstcall_meltsnow /= 1) THEN 
!        IF (MAXVAL(qs(:,:,:,ntime)*rho, mask=(t(:,:,:,ntime) > Tmeltbegin_s)) >= q_krit_radar) THEN
          firstcall_meltsnow = 1 
          nTa = 50
          nqs = 15
          nTm = 14
          CALL zradar_meltsnow_lookupcreate(look_Z_meltsnow,isnow_n0temp,nTa,nqs,nTm,&
               DBLE(Tmeltbegin_s),DBLE(meltdegTmin_s),DBLE(lambda_radar),Dmin_s,Dmax_s,ageos,bgeos,&
               mixingrulestring_s_shell,matrixstring_s_shell,inclusionstring_s_shell, &
               mixingrulestring_s_core,matrixstring_s_core,inclusionstring_s_core, &
               hoststring_s_shell, hostmatrixstring_s_shell, hostinclusionstring_s_shell, &
               hoststring_s_core, hostmatrixstring_s_core, hostinclusionstring_s_core)
!        END IF
      END IF
                
      ! de: create lookup table for snow (only first call)
      IF (firstcall_snow /= 1) THEN 
!        IF (MAXVAL(qs(:,:,:,ntime)*rho, mask=(t(:,:,:,ntime) <= Tmeltbegin_s)) >= q_krit_radar) THEN
          firstcall_snow = 1 
          nTa = 40
          nqs = 15
          CALL zradar_snow_lookupcreate(look_Z_snow,isnow_n0temp,nTa,nqs,DBLE(lambda_radar),Dmin_s,Dmax_s,ageos,bgeos,&
               DBLE(Tmeltbegin_s),mixingrulestring_s_shell_dry,matrixstring_s_shell_dry,inclusionstring_s_shell_dry, &
               mixingrulestring_s_core_dry,matrixstring_s_core_dry,inclusionstring_s_core_dry)
!        END IF
      END IF
    
      ! de: create lookup table for melting graupel (only first call)
      IF (firstcall_meltgraupel /= 1) THEN 
!        IF (MAXVAL(qg(:,:,:,ntime)*rho, mask=(t(:,:,:,ntime) > Tmeltbegin_g)) >= q_krit_radar) THEN
          firstcall_meltgraupel = 1 
          nTa = 50
          nqg = 30
          nTm = 24
          CALL zradar_meltgraupel_lookupcreate(look_Z_meltgraupel,igraupel_type,nTa,nqg,nTm,&
               DBLE(Tmeltbegin_g),DBLE(meltdegTmin_g),DBLE(lambda_radar),Dmin_g,Dmax_g,n0_g,ageog,bgeog,&
               mixingrulestring_g, matrixstring_g, inclusionstring_g, &
               mixingrulestring_g_shell, matrixstring_g_shell, inclusionstring_g_shell, &
               hoststring_g, hostmatrixstring_g, hostinclusionstring_g)
!        END IF
      END IF
      
      ! de: create lookup table for graupel (only first call)
      IF (firstcall_graupel /= 1) THEN 
!        IF (MAXVAL(qg(:,:,:,ntime)*rho, mask=(t(:,:,:,ntime) <= Tmeltbegin_g)) < q_krit_radar) THEN
          firstcall_graupel = 1 
          nTa = 20
          nqg = 30
          CALL zradar_graupel_lookupcreate(look_Z_graupel,nTa,nqg,DBLE(lambda_radar),Dmin_g,Dmax_g,n0_g,ageog,bgeog,&
               Tmeltbegin_g,mixingrulestring_g_dry,matrixstring_g_dry,inclusionstring_g_dry)
!        END IF
      END IF
    
      DO k = ku, ko
        DO j = jlow, jup
          DO i = ilow, iup

            T_a = t(i,j,k,ntime)
            hlp = rho(i,j,k)

            q_r = qr(i,j,k,ntime) * hlp

            IF (q_r >= q_krit_radar) THEN
              ! de: add option to interpolate z_local via lookup table
              q_r_log = LOG10(q_r)
              CALL zradar_biinterpolate_lookup(look_Z_rain,T_a,q_r_log,Z,zeta)
              ! convert reflectivity from logarithmic space to linear space
              IF (Z >= -90.0) THEN
                Z = 10.0**(0.1*Z)
              ELSE
                Z = 0.0
              END IF
              z_local(i,j,k) = z_local(i,j,k) + Z
              z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
            END IF

            IF (lalloc_qs) THEN

              q_s = qs(i,j,k,ntime) * hlp

              IF (q_s >= q_krit_radar) THEN
                IF (T_a > Tmeltbegin_s) THEN          
                  q_s_log = LOG10(q_s)
                  T_m = MAX(Tmax_s(i,j), 276.16d0)
                  CALL zradar_triinterpolate_lookup(look_Z_meltsnow,T_a,q_s_log,T_m,Z,zeta)
                ELSE
                  q_s_log = LOG10(q_s)
                  CALL zradar_biinterpolate_lookup(look_Z_snow,T_a,q_s_log,Z,zeta)
                END IF
                ! convert reflectivity from logarithmic space to linear space
                IF (Z >= -90.0) THEN
                  Z = 10.0**(0.1*Z)
                ELSE
                  Z = 0.0
                END IF
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              END IF

            END IF

            IF (lalloc_qg) THEN

              q_g = qg(i,j,k,ntime) * hlp

              IF (q_g >= q_krit_radar) THEN
                IF (T_a > Tmeltbegin_g) THEN
                  ! de: add option to interpolate z_local via lookup table
                  q_g_log = LOG10(q_g)
                  T_m = MAX(Tmax_g(i,j), 278.16d0)
                  CALL zradar_triinterpolate_lookup(look_Z_meltgraupel,T_a,q_g_log,T_m,Z,zeta)
                ELSE
                  q_g_log = LOG10(q_g)
                  CALL zradar_biinterpolate_lookup(look_Z_graupel,T_a,q_g_log,Z,zeta)
                END IF
                ! convert reflectivity from logarithmic space to linear space
                IF (Z >= -90.0) THEN
                  Z = 10.0**(0.1*Z)
                ELSE
                  Z = 0.0
                END IF
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              END IF

            END IF

          END DO
        END DO
      END DO

    ELSE  ! .NOT. llookup

      DO k = ku, ko
        DO j = jlow, jup
          DO i = ilow, iup
          
            T_a = t(i,j,k,ntime)
            hlp = rho(i,j,k)
          
            q_r = qr(i,j,k,ntime) * hlp
                    
            IF (q_r >= q_krit_radar) THEN
              ! de: replace m_w with T_a
              CALL zradar_rain_mie_vec(q_r,T_a,DBLE(lambda_radar),Dmin_r,Dmax_r,Z,zeta,n0_r(i,j,k),mu_rain(i,j,k))
              z_local(i,j,k) = z_local(i,j,k) + Z
              z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
            END IF  ! q_r(i) >= q_krit_radar

    
            ! Reflektivitaet von Schnee nach Mie (Integration bzgl. Durchmesser):
            IF (lalloc_qs) THEN
              
              q_s = qs(i,j,k,ntime) * hlp
              
              IF (q_s >= q_krit_radar) THEN
                IF (T_a > Tmeltbegin_s) THEN          
                  ! de: replace m_i(and m_w with T_a; remove k, ntime
                  CALL zradar_wetsnow_mie_vec(i,j,q_s,T_a,0.5d0,DBLE(Tmeltbegin_s),DBLE(meltdegTmin_s),&
                       DBLE(lambda_radar),n0_s(i,j,k),ageos,bgeos,Dmin_s,Dmax_s,Z,zeta, &
                       mixingrulestring_s_shell, matrixstring_s_shell, inclusionstring_s_shell, &
                       mixingrulestring_s_core, matrixstring_s_core, inclusionstring_s_core, &
                       hoststring_s_shell, hostmatrixstring_s_shell, hostinclusionstring_s_shell, &
                       hoststring_s_core, hostmatrixstring_s_core, hostinclusionstring_s_core)
                ELSE
                  ! de: replace m_i with T_a
                  CALL zradar_snow_mie_vec(q_s,T_a,0.5d0,DBLE(lambda_radar),n0_s(i,j,k),ageos,bgeos,Dmin_s,Dmax_s,Z,zeta, &
                       mixingrulestring_s_shell_dry, matrixstring_s_shell_dry, inclusionstring_s_shell_dry, &
                       mixingrulestring_s_core_dry, matrixstring_s_core_dry, inclusionstring_s_core_dry)
                END IF  ! T_a > Tmeltbegin_s
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
              END IF  ! q_s >= q_krit_radar)

            END IF   ! lalloc_qs

            ! Reflektivitaet von Graupel nach Mie (Integration bzgl. Durchmesser):
            IF (lalloc_qg) THEN
              
              q_g = qg(i,j,k,ntime) * hlp
              
              IF (q_g >= q_krit_radar) THEN
                
                IF (T_a > Tmeltbegin_g) THEN
                  
                  SELECT CASE (igraupel_type)
                  CASE(1)
                    ! de: replace m_i and m_w with T_a; remove k, ntime
                    CALL zradar_wetgr_mie_vec(i,j,q_g,T_a,DBLE(Tmeltbegin_g),DBLE(meltdegTmin_g),&
                         DBLE(lambda_radar),n0_g,ageog,bgeog,Dmin_g,Dmax_g,Z,zeta,&
                         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
                         hoststring_g, hostmatrixstring_g, hostinclusionstring_g)
                  CASE(2)
                    ! de: replace m_i and m_w with T_a; remove k, ntime
                    CALL zradar_wetgr_twosph_mie_vec(i,j,q_g,T_a,&
                         DBLE(Tmeltbegin_g),DBLE(meltdegTmin_g),DBLE(lambda_radar),n0_g,ageog,bgeog,Dmin_g,Dmax_g,Z,zeta,&
                         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
                         mixingrulestring_g_shell, matrixstring_g_shell, inclusionstring_g_shell)
                  CASE(3)
                    ! de: replace m_i and m_w with T_a; remove k, ntime
                    CALL zradar_wetgr_watersph_mie_vec(i,j,q_g,T_a,&
                         DBLE(Tmeltbegin_g),DBLE(meltdegTmin_g),DBLE(lambda_radar),n0_g,ageog,bgeog,Dmin_g,Dmax_g,Z,zeta,&
                         mixingrulestring_g, matrixstring_g, inclusionstring_g)
                  END SELECT
                ELSE
                  ! de: add option to interpolate z_local via lookup table
                  ! de: replace m_i with T_a
                  CALL zradar_graupel_mie_vec(q_g,T_a,DBLE(lambda_radar),n0_g,ageog,bgeog,Dmin_g,Dmax_g,Z,zeta,&
                       mixingrulestring_g_dry,matrixstring_g_dry,inclusionstring_g_dry)
                END IF  ! T_a(i) > Tmeltbegin_g
                
                z_local(i,j,k) = z_local(i,j,k) + Z
                z_ext_local(i,j,k) = z_ext_local(i,j,k) + zeta
                
              END IF  !q_g >= q_krit_radar

            END IF  ! lalloc_qg
            
          END DO  ! i
        END DO  ! j
      END DO ! k
      
    END IF  ! .NOT. llookup


    WHERE (z_local >= 1d-9)
      z_local = 10.0d0 * LOG10(z_local)
    ELSEWHERE
      z_local = -99.99
    END WHERE

    WRITE (6,'(2(A,F10.3))') "  MAX dBZ = ",MAXVAL(z_local),"  MIN dBZ = ",MINVAL(z_local)

    IF ( PRESENT(z_radar) ) THEN
      z_radar = z_local
     ENDIF
    IF ( PRESENT(z_ext) ) THEN
      z_ext = z_ext_local
    ENDIF

    IF ( PRESENT(z_radar_850) ) THEN
      DO j = jlow, jup
        DO i = ilow, iup
          z_radar_850(i,j) = z_local(i,j,klv850)
        ENDDO
      ENDDO
    ENDIF

    IF ( PRESENT(z_radar_cmax) ) THEN
      z_radar_cmax = MAXVAL(z_local,3)
    ENDIF

    CALL finalize_tmax()

    RETURN
  END SUBROUTINE radar_lm_mie_vec


#ifdef TWOMOMSEIFERT
  ! Rayleigh-Naeherung, 2M-Schema:
  SUBROUTINE radar_sb_rayleigh_vec(ntime, klv850, myproc, lambda_radar, &
       ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel, ctype_dryhail, ctype_wethail, &
       Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, Tmeltbegin_h, meltdegTmin_h, &
       z_radar, z_radar_850, z_radar_cmax, linc_boundlines )

    ! ctype_drysnow: 3 Zeichen
    ! ctype_wetsnow: 6 Zeichen
    ! ctype_drygraupel: 3 Zeichen
    ! ctype_wetgraupel: 6 Zeichen
    ! ctype_dryhail: 3 Zeichen
    ! ctype_wethail: 6 Zeichen

    INTEGER(KIND=iintegers), INTENT(IN)       :: ntime

    INTEGER(KIND=iintegers), INTENT(IN)       :: klv850, myproc

    REAL(KIND=ireals)      , INTENT(IN)       :: lambda_radar, Tmeltbegin_s, &
         meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, Tmeltbegin_h, meltdegTmin_h

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar(ie,je,ke)

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar_850(ie,je),   &
         z_radar_cmax(ie,je)

    CHARACTER(len=*), INTENT(in) :: &
         ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel, ctype_dryhail, ctype_wethail

    LOGICAL, INTENT(in), OPTIONAL             :: linc_boundlines

    ! Local Variables
    !----------------

    DOUBLE PRECISION, PARAMETER :: eps=1d-15, q_krit_radar=1d-7
    DOUBLE PRECISION ::  z_local(ie,je,ke)
    INTEGER :: i, j, k, ku, ko
    DOUBLE PRECISION, DIMENSION(ie) :: T_a, fmelt_mean, Dtmp, &
         q_r, n_r, x_r,                  &
         q_c, n_c, x_c,                  &
         q_g, n_g, x_g,                  &
         q_s, n_s, x_s,                  &
         q_i, n_i, x_i,                  &
         q_h, n_h, x_h, hlp
    DOUBLE PRECISION :: Dmin, Dmax, neigh, meltdegTmin, Tmeltbegin
    COMPLEX(kind=kinddp), DIMENSION(ie) :: m_i, m_w

    CHARACTER(len=20) :: &
         mixingrulestring_s, matrixstring_s, inclusionstring_s, &
         hoststring_s, hostmatrixstring_s, hostinclusionstring_s, &
         mixingrulestring_s_dry, matrixstring_s_dry, inclusionstring_s_dry, &
         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
         hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry, &
         mixingrulestring_h, matrixstring_h, inclusionstring_h, &
         hoststring_h, hostmatrixstring_h, hostinclusionstring_h, &
         mixingrulestring_h_dry, matrixstring_h_dry, inclusionstring_h_dry


    ! .. Set ilow, iup, jlow, jup, klow, kup for computations in this sub-domain
    !    depending on the optional switch "linc_boundlines":
    IF ( PRESENT(linc_boundlines) ) THEN
      CALL get_loc_domain( zlinc_boundlines = linc_boundlines )
    ELSE
      ! .. Default: no computations on interior boundlines to save time
      CALL get_loc_domain( zlinc_boundlines = .FALSE.)
    END IF
    
    ! Nachbarschaft bei der Bestimmung des Tmax-Parameters fuer die Schmelzgradparametrisierung:
    neigh = 5000.0d0 ! m
    CALL initialize_tmax_sb_vec_par(ntime, neigh)

    ! Steuerstrings fuer den effektiven Brechungsindex generieren:
    CALL decode_controlstring_2C(ctype_drysnow(1:3), &
         mixingrulestring_s_dry, matrixstring_s_dry, inclusionstring_s_dry, &
         'mas')
    CALL decode_controlstring_3C(ctype_wetsnow(1:6), &
         mixingrulestring_s, matrixstring_s, inclusionstring_s, &
         hoststring_s, hostmatrixstring_s, hostinclusionstring_s, &
         'mawsas')
    CALL decode_controlstring_2C(ctype_drygraupel(1:3), &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry, &
         'mas')
    CALL decode_controlstring_3C(ctype_wetgraupel(1:6), &
         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
         hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
         'mawsms')
    CALL decode_controlstring_2C(ctype_dryhail(1:3), &
         mixingrulestring_h_dry, matrixstring_h_dry, inclusionstring_h_dry, &
         'mis')
    CALL decode_controlstring_3C(ctype_wethail(1:6), &
         mixingrulestring_h, matrixstring_h, inclusionstring_h, &
         hoststring_h, hostmatrixstring_h, hostinclusionstring_h, &
         'mawsms')
    
    z_local = 0.0d0
    IF (PRESENT(z_radar) .OR. PRESENT(z_radar_cmax)) THEN
      ku = klow
      ko = kup
    ELSE
      IF (PRESENT(z_radar_850)) THEN
        ku = klv850
        ko = klv850
      ELSE
        ! Mache einfach gar nix!
        ku = 2
        ko = 1
      END IF
    END IF

    DO k= ku, ko
      DO j = jlow, jup

        T_a = 273.16d0
        DO i = ilow, iup
          T_a(i) = t(i,j,k,ntime)
          hlp(i) = rho(i,j,k)
        END DO

        ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
        m_w = (1.0d0,0.0d0)
!        m_w(ilow:iup) = m_complex_water_ray_vec(DBLE(lambda_radar),&
!             MIN(MAX(T_a(ilow:iup)-273.16d0,0.0d0),30.0d0), iup-ilow+1)
        m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
        ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
        ! Eis aber trotzdem am Gefrierpunkt verharren.
        m_i = (1.0d0,0.0d0)
!        m_i(ilow:iup) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
!             MIN(MAX(T_a(ilow:iup)-273.16d0,-80.0d0),0.0d0), iup-ilow+1)
        m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)

        DO i = ilow, iup
          q_r(i) = qr(i,j,k,ntime) * hlp(i)
          n_r(i) = qnr(i,j,k,ntime) * hlp(i)
          q_c(i) = qc(i,j,k,ntime) * hlp(i)
          n_c(i) = qnc(i,j,k,ntime) * hlp(i)
          q_i(i) = qi(i,j,k,ntime) * hlp(i)
          n_i(i) = qni(i,j,k,ntime) * hlp(i)
          x_r(i) = MIN( MAX(q_r(i)/(n_r(i)+eps),rain%x_min),rain%x_max )
          n_r(i) = q_r(i) / x_r(i)
          x_i(i) = MIN( MAX(q_i(i)/(n_i(i)+eps),ice%x_min),ice%x_max )
          n_i(i) = q_i(i) / x_i(i)
          x_c(i) = MIN( MAX(q_c(i)/(n_c(i)+eps),cloud%x_min),cloud%x_max )
          n_c(i) = q_c(i) / x_c(i)
          IF (q_r(i) < q_krit_radar) q_r(i) = 0.0d0
          IF (q_i(i) < q_krit_radar) q_i(i) = 0.0d0
          IF (q_c(i) < q_krit_radar) q_c(i) = 0.0d0
        END DO


        ! Reflektivitaet von Wolkentropfen nach Rayleigh-Approximation:
        IF (MAXVAL(q_c(ilow:iup)) >= q_krit_radar) &
             z_local(:,j,k) = z_local(:,j,k) + &
             zradar_rain_rayleigh_sb_vec(q_c,n_c,m_w,lambda_radar,cloud,ie,ilow,iup)

        ! Reflektivitaet von Regen nach Rayleigh-Approximation (Integration bzgl. Durchmesser):
        IF (MAXVAL(q_r(ilow:iup)) >= q_krit_radar) &
             z_local(:,j,k) = z_local(:,j,k) + &
             zradar_rain_rayleigh_sb_vec(q_r,n_r,m_w,lambda_radar,rain,ie,ilow,iup)

        ! Reflektivitaet von Wolkeneis nach Rayleigh:
        Tmeltbegin = 273.16d0
        meltdegTmin = 0.0d0
        IF (MAXVAL(q_i(ilow:iup)) >= q_krit_radar) THEN
          IF (MAXVAL(T_a(ilow:iup)) > Tmeltbegin) THEN
            fmelt_mean = 0.0d0
            Dtmp(ilow:iup) = D_average_sb_vec(q_i(ilow:iup), &
                 n_i(ilow:iup), ice, iup-ilow+1) * 2.0d0
            CALL degree_of_melting_ice_sb_vec_ie(ilow,iup,j,k,ntime,&
                 Dtmp,ie,Tmeltbegin,meltdegTmin,fmelt_mean)
            z_local(:,j,k) = z_local(:,j,k) + &
                 zradar_wetgr_rayl_oguchi_sb_vec(q_i,n_i,m_i,m_w,lambda_radar,fmelt_mean,ice,ie,ilow,iup)
          ELSE
            z_local(:,j,k) = z_local(:,j,k) + &
                 zradar_gr_rayl_oguchi_sb_vec(q_i,n_i,m_i,lambda_radar,ice,ie,ilow,iup)           
          END IF
        END IF

      END DO
    END DO

    ! Reflektivitaet von Schnee nach Mie (Integration bzgl. Durchmesser):
    DO k= ku, ko
      DO j = jlow, jup

        T_a = 273.16d0
        DO i = ilow, iup
          T_a(i) = t(i,j,k,ntime)
          hlp(i) = rho(i,j,k)
        END DO

        DO i = ilow, iup
          q_s(i) = qs(i,j,k,ntime) * hlp(i)
          n_s(i) = qns(i,j,k,ntime) * hlp(i)
          x_s(i) = MIN( MAX(q_s(i)/(n_s(i)+eps),snow%x_min),snow%x_max )
          n_s(i) = q_s(i) / x_s(i)
          IF (q_s(i) < q_krit_radar) q_s(i) = 0.0d0
        END DO

        IF (MAXVAL(q_s(ilow:iup)) >= q_krit_radar) THEN

          ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
          m_w = (1.0d0,0.0d0)
          m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
               MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
          ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
          ! Eis aber trotzdem am Gefrierpunkt verharren.
          m_i = (1.0d0,0.0d0)
          m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
               MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)
          
          
          !        Tmeltbegin = 273.16d0
          !        meltdegTmin = 0.0d0
          Tmeltbegin = Tmeltbegin_s
          meltdegTmin = meltdegTmin_s
          DO i = ilow, iup
            IF (T_a(i) > Tmeltbegin) THEN          
              Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
              Dmax=30.0d-3 ! Obergrenze: 20.0 mm
              IF (q_s(i) >= q_krit_radar) &
                   z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_wetgr_rayl_sb_vec(i,j,k,ntime,q_s(i),n_s(i),m_i(i),m_w(i),Tmeltbegin,meltdegTmin,&
                   lambda_radar,Dmin,Dmax,snow,'snow',&
                   mixingrulestring_s, matrixstring_s, inclusionstring_s, &
                   hoststring_s, hostmatrixstring_s, hostinclusionstring_s)
            ELSE
              Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
              Dmax=30.0d-3 ! Obergrenze: 20.0 mm
              IF (q_s(i) >= q_krit_radar) &
                   z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_gr_rayl_sb_vec(q_s(i),n_s(i),m_i(i),lambda_radar,Dmin,Dmax,snow,&
                   mixingrulestring_s_dry, matrixstring_s_dry, inclusionstring_s_dry)
            END IF
          END DO
        END IF
      END DO
    END DO

    ! Reflektivitaet von Graupel nach Mie (Integration bzgl. Durchmesser):
    DO k= ku, ko
      DO j = jlow, jup

        T_a = 273.16d0
        DO i = ilow, iup
          T_a(i) = t(i,j,k,ntime)
          hlp(i) = rho(i,j,k)
        END DO

        DO i = ilow, iup
          q_g(i) = qg(i,j,k,ntime) * hlp(i)
          n_g(i) = qng(i,j,k,ntime) * hlp(i)
          x_g(i) = MIN( MAX(q_g(i)/(n_g(i)+eps),graupel%x_min),graupel%x_max )
          n_g(i) = q_g(i) / x_g(i)
          IF (q_g(i) < q_krit_radar) q_g(i) = 0.0d0
        END DO

        IF (MAXVAL(q_g(ilow:iup)) >= q_krit_radar) THEN

          ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
          m_w = (1.0d0,0.0d0)
          m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
               MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
          ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
          ! Eis aber trotzdem am Gefrierpunkt verharren.
          m_i = (1.0d0,0.0d0)
          m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
               MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)
          
          
          !        Tmeltbegin = 263.16d0
          !        meltdegTmin = 0.2d0
          Tmeltbegin = Tmeltbegin_g
          meltdegTmin = meltdegTmin_g
          DO i = ilow, iup
            IF (T_a(i) > Tmeltbegin) THEN          
              Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
              Dmax=30.0d-3 ! Obergrenze: 15.0 mm
              IF (q_g(i) >= q_krit_radar) &
                   z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_wetgr_rayl_sb_vec(i,j,k,ntime,q_g(i),n_g(i),m_i(i),m_w(i),Tmeltbegin,meltdegTmin,&
                   lambda_radar,Dmin,Dmax,graupel,'graupel',&
                   mixingrulestring_g, matrixstring_g, inclusionstring_g, &
                   hoststring_g, hostmatrixstring_g, hostinclusionstring_g)
            ELSE
              Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
              Dmax=30.0d-3 ! Obergrenze: 15.0 mm
              IF (q_g(i) >= q_krit_radar) &
                   z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_gr_rayl_sb_vec(q_g(i),n_g(i),m_i(i),lambda_radar,Dmin,Dmax,graupel,&
                   mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry)
            END IF
          END DO
        END IF
      END DO
    END DO
    
    ! Reflektivitaet von Hagel nach Mie (Integration bzgl. Durchmesser):
    IF (ALLOCATED(qh)) THEN
      DO k= ku, ko
        DO j = jlow, jup
          
          T_a = 273.16d0
          DO i = ilow, iup
            T_a(i) = t(i,j,k,ntime)
            hlp(i) = rho(i,j,k)
          END DO
          
          DO i = ilow, iup
            q_h(i) = qh(i,j,k,ntime) * hlp(i)
            n_h(i) = qnh(i,j,k,ntime) * hlp(i)
            x_h(i) = MIN( MAX(q_h(i)/(n_h(i)+eps),hail%x_min),hail%x_max )
            n_h(i) = q_h(i) / x_h(i)
            IF (q_h(i) < q_krit_radar) q_h(i) = 0.0d0
          END DO

          IF (MAXVAL(q_h(ilow:iup)) >= q_krit_radar) THEN

            ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
            m_w = (1.0d0,0.0d0)
            m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
                 MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
            ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
            ! Eis aber trotzdem am Gefrierpunkt verharren.
            m_i = (1.0d0,0.0d0)
            m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
                 MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)
            
            !          Tmeltbegin = 263.16d0
            !          meltdegTmin = 0.2d0
            Tmeltbegin = Tmeltbegin_h
            meltdegTmin = meltdegTmin_h
            DO i = ilow, iup
              IF (T_a(i) > Tmeltbegin) THEN          
                Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
                Dmax=30.0d-3 ! Obergrenze: 15.0 mm
                ! Spongy Hail --- Zweischalenmodell mit Eiskugel und Haut aus Eis-Wasser mit Volumenverh. Wasser/Gesamt = 0.7
                IF (q_h(i) >= q_krit_radar) &
                     z_local(i,j,k) = z_local(i,j,k) + &
                     zradar_wetgr_rayl_sb_vec(i,j,k,ntime,q_h(i),n_h(i),m_i(i),m_w(i),Tmeltbegin,meltdegTmin,&
                     lambda_radar,Dmin,Dmax,hail,'hail',&
                     mixingrulestring_h, matrixstring_h, inclusionstring_h, &
                     hoststring_h, hostmatrixstring_h, hostinclusionstring_h)
              ELSE
                Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
                Dmax=30.0d-3 ! Obergrenze: 15.0 mm
                IF (q_h(i) >= q_krit_radar) &
                     z_local(i,j,k) = z_local(i,j,k) + &
                     zradar_gr_rayl_sb_vec(q_h(i),n_h(i),m_i(i),lambda_radar,Dmin,Dmax,hail,&
                     mixingrulestring_h_dry, matrixstring_h_dry, inclusionstring_h_dry)
              END IF
            END DO
          END IF
        END DO
      END DO
    END IF

    WHERE (z_local >= 1d-9)
      z_local = 10.0d0 * LOG10(z_local)
    ELSEWHERE
      z_local = -99.99
    END WHERE
    WRITE (6,'(2(A,F10.3))') "  MAX dBZ = ",MAXVAL(z_local),"  MIN dBZ = ",MINVAL(z_local)

    IF ( PRESENT(z_radar) ) THEN
      z_radar = z_local
    ENDIF

    IF ( PRESENT(z_radar_850) ) THEN
      DO j = jlow, jup
        DO i = ilow, iup
          z_radar_850(i,j) = z_local(i,j,klv850)
        ENDDO
      ENDDO
    ENDIF

    IF ( PRESENT(z_radar_cmax) ) THEN
      z_radar_cmax = MAXVAL(z_local,3)
    ENDIF

    CALL finalize_tmax()

    RETURN
  END SUBROUTINE radar_sb_rayleigh_vec
#endif

  ! Version fuer die 1M-Schemata:
  SUBROUTINE radar_lm_rayleigh_vec(ntime, klv850, myproc, lambda_radar, &
       itype_gscp, isnow_n0temp, ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel, &
       Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, &
       z_radar, z_radar_850, z_radar_cmax, linc_boundlines )

    ! ctype_drysnow: 3 Zeichen
    ! ctype_wetsnow: 6 Zeichen
    ! ctype_drygraupel: 3 Zeichen
    ! ctype_wetgraupel: 6 Zeichen

    INTEGER(KIND=iintegers), INTENT(IN)       :: ntime, itype_gscp, isnow_n0temp

    INTEGER(KIND=iintegers), INTENT(IN)       :: klv850, myproc

    REAL(KIND=ireals)      , INTENT(IN)       :: lambda_radar, &
         Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar(ie,je,ke)

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar_850(ie,je),   &
         z_radar_cmax(ie,je)

    CHARACTER(len=*), INTENT(in) :: &
         ctype_drysnow, ctype_wetsnow, ctype_drygraupel, ctype_wetgraupel

    LOGICAL, INTENT(in), OPTIONAL             :: linc_boundlines

    ! Local Variables
    !----------------

    DOUBLE PRECISION, PARAMETER :: q_krit_radar=1d-7
    DOUBLE PRECISION ::  z_local(ie,je,ke)
    INTEGER :: i, j, k, ku, ko
    DOUBLE PRECISION, DIMENSION(ie) :: T_a, fmelt_mean, Dtmp, &
         q_r,                  &
         q_c,                  &
         q_g,                  &
         q_s,                  &
         q_i,                  &
         hlp   
    DOUBLE PRECISION :: Dmin, Dmax, neigh, &
         D_c, nu_c, mu_c, a_geo, b_geo, &
         n0_r(ie,je,ke), n0_s(ie,je,ke), n0_g, ageoi, bgeoi, ageos, bgeos, ageog, bgeog, meltdegTmin, Tmeltbegin, &
         mu_rain(ie,je,ke)
    COMPLEX(kind=kinddp), DIMENSION(ie) :: m_i, m_w

    CHARACTER(len=20) :: &
         mixingrulestring_s, matrixstring_s, inclusionstring_s, &
         hoststring_s, hostmatrixstring_s, hostinclusionstring_s, &
         mixingrulestring_s_dry, matrixstring_s_dry, inclusionstring_s_dry, &
         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
         hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry

    LOGICAL :: mur_is_fixed, n0r_is_fixed


    ! .. Set ilow, iup, jlow, jup, klow, kup for computations in this sub-domain
    !    depending on the optional switch "linc_boundlines":
    IF ( PRESENT(linc_boundlines) ) THEN
      CALL get_loc_domain( zlinc_boundlines = linc_boundlines )
    ELSE
      ! .. Default: no computations on interior boundlines to save time
      CALL get_loc_domain( zlinc_boundlines = .FALSE.)
    END IF
    
    ! Parameter der Hydrometeorarten und deren Verteilungsfunktionen:
    CALL hydroparams_lm(ntime, itype_gscp, isnow_n0temp, &
         ageoi, &
         bgeoi, &
         ageos, &
         bgeos, &
         ageog, &
         bgeog, &
         n0_r, &
         n0_s, &
         n0_g, mu_rain, mur_is_fixed, n0r_is_fixed)

    ! Nachbarschaft bei der Bestimmung des Tmax-Parameters fuer die Schmelzgradparametrisierung:
    neigh = 5000.0d0 ! m
    CALL initialize_tmax_vec_par(ntime, neigh)

    ! Steuerstrings fuer den effektiven Brechungsindex generieren:
    CALL decode_controlstring_2C(ctype_drysnow(1:3), &
         mixingrulestring_s_dry, matrixstring_s_dry, inclusionstring_s_dry, &
         'mas')
    CALL decode_controlstring_3C(ctype_wetsnow(1:6), &
         mixingrulestring_s, matrixstring_s, inclusionstring_s, &
         hoststring_s, hostmatrixstring_s, hostinclusionstring_s, &
         'mawsas')
    CALL decode_controlstring_2C(ctype_drygraupel(1:3), &
         mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry, &
         'mas')
    CALL decode_controlstring_3C(ctype_wetgraupel(1:6), &
         mixingrulestring_g, matrixstring_g, inclusionstring_g, &
         hoststring_g, hostmatrixstring_g, hostinclusionstring_g, &
         'mawsas')

    z_local = 0.0d0
    IF (PRESENT(z_radar) .OR. PRESENT(z_radar_cmax)) THEN
      ku = klow
      ko = kup
    ELSE
      IF (PRESENT(z_radar_850)) THEN
        ku = klv850
        ko = klv850
      ELSE
        ! Mache einfach gar nix!
        ku = 2
        ko = 1
      END IF
    END IF
    DO k= ku, ko
      DO j = jlow, jup

        T_a = 273.16d0
        DO i = ilow, iup
          T_a(i) = t(i,j,k,ntime)
          hlp(i) = rho(i,j,k)
        END DO

        DO i = ilow, iup
          q_r(i) = qr(i,j,k,ntime) * hlp(i)
          q_c(i) = qc(i,j,k,ntime) * hlp(i)
          IF (q_r(i) < q_krit_radar) q_r(i) = 0.0d0
          IF (q_c(i) < q_krit_radar) q_c(i) = 0.0d0
        END DO

        ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
        m_w = (1.0d0,0.0d0)
        m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
        ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
        ! Eis aber trotzdem am Gefrierpunkt verharren.
        m_i = (1.0d0,0.0d0)
        m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)
          
        IF (MAXVAL(q_c(ilow:iup)) >= q_krit_radar .OR. MAXVAL(q_r(ilow:iup)) >= q_krit_radar) THEN

          ! Reflektivitaet von Wolkentropfen nach Rayleigh-Approximation:
          !   Parameter der Wolkentropfenverteilung (Gen. Gamma):
          D_c = 20d-6  ! Durchmesser des Teilchens mit der mittleren Masse in m
          nu_c = 3.0d0    ! nu bezgl. Durchmesserverteilung
          mu_c = 3.0d0    ! mu bezgl. Durchmesserverteilung
          z_local(:,j,k) = z_local(:,j,k) + &
               zradar_cloud_rayleigh_vec(q_c,m_w,lambda_radar,D_c,nu_c,mu_c,ie,ilow,iup)
          
          ! Reflektivitaet von Regen nach Rayleigh:
          z_local(:,j,k) = z_local(:,j,k) + &
               zradar_rain_rayleigh_vec(q_r,m_w,lambda_radar,&
               n0_r(:,j,k),mu_rain(:,j,k),ie,ilow,iup,mur_is_fixed)
        END IF

        ! Reflektivitaet von Wolkeneis nach Rayleigh:
        IF (ALLOCATED(qi)) THEN

          DO i = ilow, iup
            q_i(i) = qi(i,j,k,ntime) * hlp(i)
            IF (q_i(i) < q_krit_radar) q_i(i) = 0.0d0
          END DO

          IF (MAXVAL(q_i(ilow:iup)) >= q_krit_radar) THEN

!            b_geo = 1.0 / 0.302115
!            a_geo = (1.0 / 2.17d-01)**(b_geo)
            b_geo = bgeoi
            a_geo = ageoi
            D_c = 100d-6 ! m, Durchmesser des Partikels mit der mittleren Masse
            nu_c = (-0.3333333d0 + 1.0d0) * b_geo - 1.0d0   ! entspricht -0.33333 bezgl. Massenverteilung
            mu_c = 0.3333333d0 * b_geo   ! entspricht  0.33333 bezgl. Massenverteilung
            Tmeltbegin = 273.16d0
            meltdegTmin = 0.0d0

            IF (MAXVAL(T_a(ilow:iup)) > Tmeltbegin) THEN
              fmelt_mean = 0.0d0
              Dtmp = 100d-6  * 2.0 ! m, Durchmesser des Teilchens, das zur Bestimmung des mittl. Schmelzgrades 
              ! herangezogen wird.
              CALL degree_of_melting_ice_vec_ie(ilow,iup,j,k,ntime,&
                   Dtmp,ie,Tmeltbegin,meltdegTmin,100.0d-6,fmelt_mean)
              z_local(:,j,k) = z_local(:,j,k) + &
                   zradar_wetice_rayl_oguchi_vec(q_i,m_i,m_w,lambda_radar,&
                   fmelt_mean, D_c, nu_c, mu_c, a_geo, b_geo, ie, ilow, iup)
            ELSE
              z_local(:,j,k) = z_local(:,j,k) + &
                   zradar_ice_rayleigh_oguchi_vec(q_i,m_i,lambda_radar, &
                   D_c, nu_c, mu_c, a_geo, b_geo, ie, ilow, iup)           
            END IF
          END IF
        END IF
      END DO
    END DO

    ! Reflektivitaet von Schnee nach Mie (Integration bzgl. Durchmesser):
    IF (ALLOCATED(qs)) THEN
      DO k= ku, ko
        DO j = jlow, jup
          
          T_a = 273.16d0
          DO i = ilow, iup
            T_a(i) = t(i,j,k,ntime)
            hlp(i) = rho(i,j,k)
          END DO
          
          DO i = ilow, iup
            q_s(i) = qs(i,j,k,ntime) * hlp(i)
            IF (q_s(i) < q_krit_radar) q_s(i) = 0.0d0
          END DO
          
          IF (MAXVAL(q_s(ilow:iup)) >= q_krit_radar) THEN
            
            ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
            m_w = (1.0d0,0.0d0)
            m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
                 MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
            ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
            ! Eis aber trotzdem am Gefrierpunkt verharren.
            m_i = (1.0d0,0.0d0)
            m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
                 MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)
            
            !          Tmeltbegin = 273.16d0
            !          meltdegTmin = 0.0d0
            Tmeltbegin = Tmeltbegin_s
            meltdegTmin = meltdegTmin_s
            
            DO i = ilow, iup
              IF (q_s(i) >= q_krit_radar) THEN 
                IF (T_a(i) > Tmeltbegin) THEN          
                  Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
                  Dmax=30.0d-3 ! Obergrenze: 20.0 mm
                  ! de: add T_a
                  z_local(i,j,k) = z_local(i,j,k) + &
                       zradar_wetgr_rayl_vec(i,j,k,ntime,q_s(i),T_a(i),m_i(i),m_w(i),Tmeltbegin,meltdegTmin,&
                       lambda_radar,n0_s(i,j,k),ageos,bgeos,'snow',Dmin,Dmax,&
                       mixingrulestring_s, matrixstring_s, inclusionstring_s, &
                       hoststring_s, hostmatrixstring_s, hostinclusionstring_s)
                ELSE
                  Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
                  Dmax=30.0d-3 ! Obergrenze: 20.0 mm
                  z_local(i,j,k) = z_local(i,j,k) + &
                       zradar_gr_rayl_vec(q_s(i),m_i(i),lambda_radar,n0_s(i,j,k),ageos,bgeos,Dmin,Dmax,&
                       mixingrulestring_s_dry, matrixstring_s_dry, inclusionstring_s_dry)
                END IF
              END IF
            END DO
          END IF
        END DO
      END DO
    END IF
    
    ! Reflektivitaet von Graupel nach Rayleigh (Integration bzgl. Durchmesser):
    IF (ALLOCATED(qg)) THEN
      DO k= ku, ko
        DO j = jlow, jup

          T_a = 273.16d0
          DO i = ilow, iup
            T_a(i) = t(i,j,k,ntime)
            hlp(i) = rho(i,j,k)
          END DO

          DO i = ilow, iup
            q_g(i) = qg(i,j,k,ntime) * hlp(i)
            IF (q_g(i) < q_krit_radar) q_g(i) = 0.0d0
          END DO

          IF (MAXVAL(q_g(ilow:iup)) >= q_krit_radar) THEN

            ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
            m_w = (1.0d0,0.0d0)
            m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
                 MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
            ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
            ! Eis aber trotzdem am Gefrierpunkt verharren.
            m_i = (1.0d0,0.0d0)
            m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
                 MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)
            
            !          Tmeltbegin = 263.16d0
            !          meltdegTmin = 0.2d0
            Tmeltbegin = Tmeltbegin_g
            meltdegTmin = meltdegTmin_g
            DO i = ilow, iup
              IF (q_g(i) >= q_krit_radar) THEN
                IF (T_a(i) > Tmeltbegin) THEN          
                  Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
                  Dmax=30.0d-3 ! Obergrenze: 15.0 mm
                  ! de: add T_a
                  z_local(i,j,k) = z_local(i,j,k) + &
                       zradar_wetgr_rayl_vec(i,j,k,ntime,q_g(i),T_a(i),m_i(i),m_w(i),Tmeltbegin,meltdegTmin,&
                       lambda_radar,n0_g,ageog,bgeog,'graupel',Dmin,Dmax,&
                       mixingrulestring_g, matrixstring_g, inclusionstring_g, &
                       hoststring_g, hostmatrixstring_g, hostinclusionstring_g)
                ELSE
                  Dmin=0.5d-4 ! Untegrenze des Integrationsintervalls der Reflektivitaet: 0.05 mm
                  Dmax=30.0d-3 ! Obergrenze: 15.0 mm
                  z_local(i,j,k) = z_local(i,j,k) + &
                       zradar_gr_rayl_vec(q_g(i),m_i(i),lambda_radar,n0_g,ageog,bgeog,Dmin,Dmax,&
                       mixingrulestring_g_dry, matrixstring_g_dry, inclusionstring_g_dry)
                END IF
              END IF
            END DO
          END IF
        END DO
      END DO
    END IF

    WHERE (z_local >= 1d-9)
      z_local = 10.0d0 * LOG10(z_local)
    ELSEWHERE
      z_local = -99.99
    END WHERE

    WRITE (6,'(2(A,F10.3))') "  MAX dBZ = ",MAXVAL(z_local),"  MIN dBZ = ",MINVAL(z_local)

    IF ( PRESENT(z_radar) ) THEN
      z_radar = z_local
    ENDIF

    IF ( PRESENT(z_radar_850) ) THEN
      DO j = jlow, jup
        DO i = ilow, iup
          z_radar_850(i,j) = z_local(i,j,klv850)
        ENDDO
      ENDDO
    ENDIF

    IF ( PRESENT(z_radar_cmax) ) THEN
      z_radar_cmax = MAXVAL(z_local,3)
    ENDIF

    CALL finalize_tmax()

    RETURN
  END SUBROUTINE radar_lm_rayleigh_vec

#ifdef TWOMOMSEIFERT
  ! Rayleigh-Naeherung, 2M-Schema, Oguchi-Brechungsindexformulierung. Sehr schnell, aber ungenau.
  ! Die Radarwellenlaenge wird nur fuer die Berechnung des K_W_0^2 benutzt.
  SUBROUTINE radar_sb_rayleigh_oguchi_vec(ntime, klv850, myproc, lambda_radar, &
       Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, Tmeltbegin_h, meltdegTmin_h, &
       z_radar, z_radar_850, z_radar_cmax, linc_boundlines )

    INTEGER(KIND=iintegers), INTENT(IN)       :: ntime

    INTEGER(KIND=iintegers), INTENT(IN)       :: klv850, myproc

    REAL(KIND=ireals)      , INTENT(IN)       :: lambda_radar, &
         Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, Tmeltbegin_h, meltdegTmin_h

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar(ie,je,ke)

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar_850(ie,je),   &
         z_radar_cmax(ie,je)

    LOGICAL, INTENT(in), OPTIONAL             :: linc_boundlines

    ! Local Variables
    !----------------

    DOUBLE PRECISION, PARAMETER :: eps=1d-15, q_krit_radar=1d-7
    DOUBLE PRECISION ::  z_local(ie,je,ke)
    INTEGER :: i, j, k, ku, ko
    DOUBLE PRECISION, DIMENSION(ie) :: T_a, fmelt_mean, Dtmp, &
         q_r, n_r, x_r,                  &
         q_c, n_c, x_c,                  &
         q_g, n_g, x_g,                  &
         q_s, n_s, x_s,                  &
         q_i, n_i, x_i,                  &
         q_h, n_h, x_h, hlp
    DOUBLE PRECISION :: neigh, meltdegTmin, Tmeltbegin
    COMPLEX(kind=kinddp), DIMENSION(ie) :: m_i, m_w


    ! .. Set ilow, iup, jlow, jup, klow, kup for computations in this sub-domain
    !    depending on the optional switch "linc_boundlines":
    IF ( PRESENT(linc_boundlines) ) THEN
      CALL get_loc_domain( zlinc_boundlines = linc_boundlines )
    ELSE
      ! .. Default: no computations on interior boundlines to save time
      CALL get_loc_domain( zlinc_boundlines = .FALSE.)
    END IF
    
    ! Nachbarschaft bei der Bestimmung des Tmax-Parameters fuer die Schmelzgradparametrisierung:
    neigh = 5000.0d0 ! m
    CALL initialize_tmax_sb_vec_par(ntime, neigh)

    z_local = 0.0d0
    IF (PRESENT(z_radar) .OR. PRESENT(z_radar_cmax)) THEN
      ku = klow
      ko = kup
    ELSE
      IF (PRESENT(z_radar_850)) THEN
        ku = klv850
        ko = klv850
      ELSE
        ! Mache einfach gar nix!
        ku = 2
        ko = 1
      END IF
    END IF
    DO k= ku, ko
      DO j = jlow, jup

        T_a = 273.16d0
        DO i = ilow, iup
          T_a(i) = t(i,j,k,ntime)
          hlp(i) = rho(i,j,k)
        END DO

        m_w = (1.0d0,0.0d0)
        m_w(1:ie) = m_complex_water_ray_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,0.0d0),30.0d0), ie)
        ! Brechungsindex von Eis nach Maetzler (1998) --- Bei T > 0 Grad C soll 
        ! Eis aber trotzdem am Gefrierpunkt verharren.
        m_i = (1.0d0,0.0d0)
        m_i(1:ie) = m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
             MIN(MAX(T_a(1:ie)-273.16d0,-80.0d0),0.0d0), ie)

        DO i = ilow, iup
          q_r(i) = qr(i,j,k,ntime) * hlp(i)
          n_r(i) = qnr(i,j,k,ntime) * hlp(i)
          q_c(i) = qc(i,j,k,ntime) * hlp(i)
          n_c(i) = qnc(i,j,k,ntime) * hlp(i)
          q_g(i) = qg(i,j,k,ntime) * hlp(i)
          n_g(i) = qng(i,j,k,ntime) * hlp(i)
          q_s(i) = qs(i,j,k,ntime) * hlp(i)
          n_s(i) = qns(i,j,k,ntime) * hlp(i)
          q_i(i) = qi(i,j,k,ntime) * hlp(i)
          n_i(i) = qni(i,j,k,ntime) * hlp(i)
          x_r(i) = MIN( MAX(q_r(i)/(n_r(i)+eps),rain%x_min),rain%x_max )
          n_r(i) = q_r(i) / x_r(i)
          x_g(i) = MIN( MAX(q_g(i)/(n_g(i)+eps),graupel%x_min),graupel%x_max )
          n_g(i) = q_g(i) / x_g(i)
          x_s(i) = MIN( MAX(q_s(i)/(n_s(i)+eps),snow%x_min),snow%x_max )
          n_s(i) = q_s(i) / x_s(i)
          x_i(i) = MIN( MAX(q_i(i)/(n_i(i)+eps),ice%x_min),ice%x_max )
          n_i(i) = q_i(i) / x_i(i)
          x_c(i) = MIN( MAX(q_c(i)/(n_c(i)+eps),cloud%x_min),cloud%x_max )
          n_c(i) = q_c(i) / x_c(i)
          IF (q_s(i) < q_krit_radar) q_s(i) = 0.0d0
          IF (q_r(i) < q_krit_radar) q_r(i) = 0.0d0
          IF (q_g(i) < q_krit_radar) q_g(i) = 0.0d0
          IF (q_i(i) < q_krit_radar) q_i(i) = 0.0d0
          IF (q_c(i) < q_krit_radar) q_c(i) = 0.0d0
        END DO

        ! Reflektivitaet von Wolkentropfen nach Rayleigh-Approximation:
        z_local(:,j,k) = z_local(:,j,k) + &
             zradar_rain_rayleigh_sb_vec(q_c,n_c,m_w,lambda_radar,cloud,ie,ilow,iup)

        ! Reflektivitaet von Regen nach Rayleigh:
        z_local(:,j,k) = z_local(:,j,k) + &
             zradar_rain_rayleigh_sb_vec(q_r,n_r,m_w,lambda_radar,rain,ie,ilow,iup)

        ! Reflektivitaet von Wolkeneis nach Rayleigh:
        Tmeltbegin = 273.16d0
        meltdegTmin = 0.0d0
        IF (MAXVAL(T_a(ilow:iup)) > Tmeltbegin) THEN
          fmelt_mean = 0.0d0
          Dtmp(ilow:iup) = D_average_sb_vec(q_i(ilow:iup), &
               n_i(ilow:iup), ice, iup-ilow+1) * 2.0d0
          CALL degree_of_melting_ice_sb_vec_ie(ilow,iup,j,k,ntime,&
               Dtmp,ie,Tmeltbegin,meltdegTmin,fmelt_mean)
          z_local(:,j,k) = z_local(:,j,k) + &
                 zradar_wetgr_rayl_oguchi_sb_vec(q_i,n_i,m_i,m_w,&
                 lambda_radar,fmelt_mean,ice,ie,ilow,iup)
        ELSE
          z_local(:,j,k) = z_local(:,j,k) + &
               zradar_gr_rayl_oguchi_sb_vec(q_i,n_i,m_i,lambda_radar,ice,ie,ilow,iup)           
        END IF

        ! Reflektivitaet von Schnee nach Rayleigh:
!        Tmeltbegin = 273.16d0
!        meltdegTmin = 0.0d0
        Tmeltbegin = Tmeltbegin_s
        meltdegTmin = meltdegTmin_s

        IF (MAXVAL(T_a(ilow:iup)) > Tmeltbegin) THEN          
          fmelt_mean = 0.0d0
          Dtmp(ilow:iup) = D_average_sb_vec(q_s(ilow:iup), &
               n_s(ilow:iup), snow, iup-ilow+1) * 2.0d0
          CALL degree_of_melting_sn_sb_vec_ie(ilow,iup,j,k,ntime,&
               Dtmp,ie,Tmeltbegin,meltdegTmin,fmelt_mean)
          z_local(:,j,k) = z_local(:,j,k) + &
               zradar_wetgr_rayl_oguchi_sb_vec(q_s,n_s,m_i,m_w,&
               lambda_radar,fmelt_mean,snow,ie,ilow,iup)
        ELSE
          z_local(:,j,k) = z_local(:,j,k) + &
               zradar_gr_rayl_oguchi_sb_vec(q_s,n_s,m_i,lambda_radar,snow,ie,ilow,iup)
        END IF

        ! Reflektivitaet von Graupel nach Rayleigh:
!        Tmeltbegin = 263.16d0
!        meltdegTmin = 0.2d0
        Tmeltbegin = Tmeltbegin_g
        meltdegTmin = meltdegTmin_g
      
        IF (MAXVAL(T_a(ilow:iup)) > Tmeltbegin) THEN          
          fmelt_mean = 0.0d0
          Dtmp(ilow:iup) = D_average_sb_vec(q_g(ilow:iup), &
               n_g(ilow:iup), graupel, iup-ilow+1) * 2.0d0
          CALL degree_of_melting_gr_sb_vec_ie(ilow,iup,j,k,ntime,&
               Dtmp,ie,Tmeltbegin,meltdegTmin,fmelt_mean)
          z_local(:,j,k) = z_local(:,j,k) + &
               zradar_wetgr_rayl_oguchi_sb_vec(q_g,n_g,m_i,m_w,&
               lambda_radar,fmelt_mean,graupel,ie,ilow,iup)
        ELSE
          z_local(:,j,k) = z_local(:,j,k) + &
               zradar_gr_rayl_oguchi_sb_vec(q_g,n_g,m_i,lambda_radar,graupel,ie,ilow,iup)
        END IF

        ! Reflektivitaet von Hagel:
        IF (ALLOCATED(qh)) THEN
          DO i = ilow, iup
            q_h(i) = qh(i,j,k,ntime) * hlp(i)
            n_h(i) = qnh(i,j,k,ntime) * hlp(i)
            x_h(i) = MIN( MAX(q_h(i)/(n_h(i)+eps),hail%x_min),hail%x_max )
            n_h(i) = q_h(i) / x_h(i)
            IF (q_h(i) < q_krit_radar) q_h(i) = 0.0d0
          END DO

!          Tmeltbegin = 263.16d0
!          meltdegTmin = 0.2d0
          Tmeltbegin = Tmeltbegin_h
          meltdegTmin = meltdegTmin_h
          IF (MAXVAL(T_a(ilow:iup)) > Tmeltbegin) THEN          
            fmelt_mean = 0.0d0
            ! Spongy Hail --- Eis-Wasser-Mischung:
            Dtmp(ilow:iup) = D_average_sb_vec(q_h(ilow:iup), &
                 n_h(ilow:iup), hail, iup-ilow+1) * 2.0d0
            CALL degree_of_melting_ha_sb_vec_ie(ilow,iup,j,k,&
                 ntime,Dtmp,ie,Tmeltbegin,meltdegTmin,fmelt_mean)
            z_local(:,j,k) = z_local(:,j,k) + &
                 zradar_wetgr_rayl_oguchi_sb_vec(q_h,n_h,m_i,m_w,lambda_radar,fmelt_mean,hail,ie,ilow,iup)
          ELSE
            z_local(:,j,k) = z_local(:,j,k) + &
                 zradar_gr_rayl_oguchi_sb_vec(q_h,n_h,m_i,lambda_radar,hail,ie,ilow,iup)
          END IF
          
        END IF

      END DO
    END DO

    WHERE (z_local >= 1d-9)
      z_local = 10.0d0 * LOG10(z_local)
    ELSEWHERE
      z_local = -99.99
    END WHERE

    WRITE (6,'(2(A,F10.3))') "  MAX dBZ = ",MAXVAL(z_local),"  MIN dBZ = ",MINVAL(z_local)

    IF ( PRESENT(z_radar) ) THEN
      z_radar = z_local
    ENDIF

    IF ( PRESENT(z_radar_850) ) THEN
      DO j = jlow, jup
        DO i = ilow, iup
          z_radar_850(i,j) = z_local(i,j,klv850)
        ENDDO
      ENDDO
    ENDIF

    IF ( PRESENT(z_radar_cmax) ) THEN
      z_radar_cmax = MAXVAL(z_local,3)
    ENDIF

    CALL finalize_tmax()

    RETURN
  END SUBROUTINE radar_sb_rayleigh_oguchi_vec
#endif



  ! Version fuer die 1M-Schemata:
  SUBROUTINE radar_lm_rayleigh_oguchi_vec(ntime, klv850, myproc, lambda_radar, &
       itype_gscp, isnow_n0temp, Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g, &
       z_radar, z_radar_850, z_radar_cmax, linc_boundlines )

    INTEGER(KIND=iintegers), INTENT(IN)       :: ntime, itype_gscp, isnow_n0temp

    INTEGER(KIND=iintegers), INTENT(IN)       :: klv850, myproc

    REAL(KIND=ireals)      , INTENT(IN)       :: lambda_radar, &
         Tmeltbegin_s, meltdegTmin_s, Tmeltbegin_g, meltdegTmin_g

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar(ie,je,ke)

    REAL(KIND=ireals), INTENT(OUT), OPTIONAL  :: z_radar_850(ie,je),   &
         z_radar_cmax(ie,je)

    LOGICAL, INTENT(in), OPTIONAL             :: linc_boundlines

    ! Local Variables
    !----------------

    DOUBLE PRECISION, PARAMETER :: q_krit_radar=1d-7
    DOUBLE PRECISION ::  z_local(ie,je,ke)
    INTEGER :: i, j, k, ku, ko
    DOUBLE PRECISION :: T_a, fmelt_mean, Dtmp, &
         q_r,                  &
         q_c,                  &
         q_g,                  &
         q_s,                  &
         q_i,                  &
         hlp 
    DOUBLE PRECISION :: neigh, K_w, K_i, &
         n0_r(ie,je,ke), n0_s(ie,je,ke), n0_g, &
         ageoi, bgeoi, ageos, bgeos,  ageog, bgeog, meltdegTmin_i, Tmeltbegin_i, mu_rain(ie,je,ke)
    COMPLEX(kind=kinddp) :: m_i(ie,je,ke), m_w(ie,je,ke)

    LOGICAL :: mur_is_fixed, n0r_is_fixed

    DOUBLE PRECISION :: lamexpr(ie,je,ke), gamfak1r(ie,je,ke), gamfak2r(ie,je,ke)


    LOGICAL :: lalloc_qi, lalloc_qs, lalloc_qg

    ! .. Set ilow, iup, jlow, jup, klow, kup for computations in this sub-domain
    !    depending on the optional switch "linc_boundlines":
    IF ( PRESENT(linc_boundlines) ) THEN
      CALL get_loc_domain( zlinc_boundlines = linc_boundlines )
    ELSE
      ! .. Default: no computations on interior boundlines to save time
      CALL get_loc_domain( zlinc_boundlines = .FALSE.)
    END IF
    
    ! Parameter der Hydrometeorarten und deren Verteilungsfunktionen:
    CALL hydroparams_lm(ntime, itype_gscp, isnow_n0temp, &
         ageoi, &
         bgeoi, &
         ageos, &
         bgeos, &
         ageog, &
         bgeog, &
         n0_r, &
         n0_s, &
         n0_g, mu_rain, mur_is_fixed, n0r_is_fixed)

    ! Nachbarschaft bei der Bestimmung des Tmax-Parameters fuer die Schmelzgradparametrisierung:
    neigh = 5000.0d0 ! m
    CALL initialize_tmax_vec_par(ntime, neigh)

    z_local = 0.0d0
    IF (PRESENT(z_radar) .OR. PRESENT(z_radar_cmax)) THEN
      ku = klow
      ko = kup
    ELSE
      IF (PRESENT(z_radar_850)) THEN
        ku = klv850
        ko = klv850
      ELSE
        ! Mache einfach gar nix!
        ku = 2
        ko = 1
      END IF
    END IF

    lalloc_qi = ALLOCATED(qi)
    lalloc_qs = ALLOCATED(qs)
    lalloc_qg = ALLOCATED(qg)

    ! .. Initializations for PSD-parameters for cloud drops and cloud ice,
    !    since there are no such assumptions in the COSMO microphysics:

    ! Cloud drops: generalized gamma (original COSMO assumes no explicit DSD)
    ray_const%D_c = 20d-6  ! Durchmesser des Teilchens mit der mittleren Masse in m
    ray_const%nu_Dc = 3.0d0       ! nu bezgl. Durchmesserverteilung
    ray_const%mu_Dc = 3.0d0       ! mu bezgl. Durchmesserverteilung

    ! Cloud ice: generalized gamma (original COSMO is however monodisperse)
    ray_const%D_i = 100d-6 ! m, Durchmesser des Partikels mit der mittleren Masse
    ray_const%nu_Di = (-0.3333333d0 + 1.0d0) * bgeoi - 1.0d0   ! entspricht -0.33333 bezgl. Massenverteilung
    ray_const%mu_Di = 0.3333333d0 * bgeoi   ! entspricht  0.33333 bezgl. Massenverteilung
    Tmeltbegin_i = 273.16d0
    meltdegTmin_i = 0.0d0

    ! On its first call or if otherwise necessary, this computes 
    ! some constant prefactors needed below and stores it on global struct "ray_const":
    CALL init_radar_lm_ray_oguchi (                        &
         .TRUE., .TRUE., lalloc_qi, lalloc_qs, lalloc_qg,  &
         DBLE(lambda_radar), mur_is_fixed, mu_rain(1,1,1), &
         ageoi, bgeoi, ageos, bgeos, ageog, bgeog, n0_g)

    ! If mu of rain DSD is variable in space and time, the
    ! actual computation of some prefactors has to be done every timestep:
    IF (.NOT. mur_is_fixed) THEN

      DO k= ku, ko
        DO j = jlow, jup
          IF (ANY(mu_rain(:,j,k) /= 0.0d0)) THEN
            lamexpr(:,j,k) = 1.0 / (mu_rain(:,j,k)+4.0d0)
            gamfak1r(:,j,k) = gfctvec(mu_rain(:,j,k)+4.0d0, ie) / 6.0d0
            gamfak2r(:,j,k) = gfctvec(mu_rain(:,j,k)+7.0d0, ie) / ray_const%K_w_0
          ELSE
            lamexpr(:,j,k) = 0.25d0
            gamfak1r(:,j,k) = 1.0d0
            gamfak2r(:,j,k) = 720.0d0 / ray_const%K_w_0
          END IF
        END DO
      END DO
      
    END IF

    ! Brechungsindex von Wasser nach Ray (1972), nur gueltig fuer -10 < T < 30 Grad C
    m_w = RESHAPE(m_complex_water_ray_vec(DBLE(lambda_radar),&
         MIN(MAX(RESHAPE(t(:,:,:,ntime),(/ie*je*ke/))-273.16d0,0.0d0),30.0d0),ie*je*ke), &
         (/ie,je,ke/) )
    ! Brechungsindex von Eis nach C. Maetzler, gueltig fuer -80 < T < 0 Grad C
    m_i = RESHAPE(m_complex_ice_maetzler_vec(DBLE(lambda_radar),&
         MIN(MAX(RESHAPE(t(:,:,:,ntime),(/ie*je*ke/))-273.16d0,-80.0d0),0.0d0), ie*je*ke), &
         (/ie,je,ke/) )

    DO k= ku, ko
      DO j = jlow, jup
        DO i = ilow, iup

          T_a = t(i,j,k,ntime)
          hlp = rho(i,j,k)

          K_w    = (ABS( (m_w(i,j,k)**2-1.0)/(m_w(i,j,k)**2+2.0) ))**2
          K_i    = (ABS( (m_i(i,j,k)**2-1.0)/(m_i(i,j,k)**2+2.0) ))**2

          q_r = qr(i,j,k,ntime) * hlp
          q_c = qc(i,j,k,ntime) * hlp
          IF (q_r < q_krit_radar) q_r = 0.0d0
          IF (q_c < q_krit_radar) q_c = 0.0d0


          z_local(i,j,k) = z_local(i,j,k) + &
             zradar_cloud_rayleigh(q_c,K_w*ray_const%prefactor_c,ray_const%D_c)

          ! Reflektivitaet von Regen nach Rayleigh:
          IF (mur_is_fixed) THEN
            z_local(i,j,k) = z_local(i,j,k) + &
                 zradar_rain_rayleigh(q_r, n0_r(i,j,k), mu_rain(i,j,k),   &
                 ray_const%lamexprc, ray_const%gamfak1rc, K_w*ray_const%gamfak2rc)
          ELSE
            z_local(i,j,k) = z_local(i,j,k) + &
                 zradar_rain_rayleigh(q_r, n0_r(i,j,k), mu_rain(i,j,k),   &
                 lamexpr(i,j,k), gamfak1r(i,j,k), K_w*gamfak2r(i,j,k))

          END IF

          IF (lalloc_qi) THEN

            q_i = qi(i,j,k,ntime) * hlp
            IF (q_i < q_krit_radar) q_i = 0.0d0

            IF (T_a > Tmeltbegin_i) THEN
              fmelt_mean = 0.0d0
              Dtmp = ray_const%D_i * 2.0d0 ! m, Durchmesser des Teilchens, das zur Bestimmung des mittl. Schmelzgrades 
              ! herangezogen wird.

              CALL degree_of_melting_ice_single(i,j,T_a,&
                   Dtmp,Tmeltbegin_i,meltdegTmin_i,100d-6,fmelt_mean)
              
              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_wetice_rayl_oguchi(q_i, m_w(i,j,k), m_i(i,j,k), &
                   fmelt_mean, ray_const%D_i, ageoi, bgeoi, ray_const%prefactor_im)
            ELSE
              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_ice_rayleigh_oguchi(q_i, &
                   ray_const%D_i, ageoi, bgeoi, K_i*ray_const%prefactor_i)           
            END IF

          END IF


          IF (lalloc_qs) THEN

            q_s = qs(i,j,k,ntime) * hlp
            IF (q_s < q_krit_radar) q_s = 0.0d0

            IF (T_a > Tmeltbegin_s) THEN          
              fmelt_mean = 0.0d0
              ! m, Durchmesser des Teilchens, das zur Bestimmung des mittl. Schmelzgrades 
              ! herangezogen wird:
              Dtmp = D_average_1M_exp_cin(q_s,&
                 ageos,bgeos,n0_s(i,j,k),ray_const%gfc_bgeop1_s) * 2.0d0
              CALL degree_of_melting_sn_single(i,j,q_s,T_a, &
                   Dtmp,DBLE(Tmeltbegin_s),DBLE(meltdegTmin_s),n0_s(i,j,k),ageos,bgeos,ray_const%gfc_bgeop1_s, &
                   fmelt_mean)

              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_wetsn_rayl_oguchi(q_s, m_i(i,j,k), m_w(i,j,k), &
                   fmelt_mean, n0_s(i,j,k), ray_const%prefactor_sm, ray_const%bfaks)
            ELSE
              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_sn_rayl_oguchi(q_s, n0_s(i,j,k), K_i*ray_const%prefactor_s, ray_const%bfaks)
            END IF

          END IF


          IF (lalloc_qg) THEN

            q_g = qg(i,j,k,ntime) * hlp
            IF (q_g < q_krit_radar) q_g = 0.0d0

            IF (T_a > Tmeltbegin_g) THEN
              fmelt_mean = 0.0d0
              ! m, Durchmesser des Teilchens, das zur Bestimmung des mittl. Schmelzgrades 
              ! herangezogen wird:
              Dtmp = D_average_1M_exp_cin(q_g,&
                 ageog,bgeog,n0_g, ray_const%gfc_bgeop1_g) * 2.0d0
              CALL degree_of_melting_gr_single(i,j,q_g,T_a,&
                   Dtmp,DBLE(Tmeltbegin_g),DBLE(meltdegTmin_g),n0_g,ageog,bgeog,ray_const%gfc_bgeop1_g, &
                   fmelt_mean)

              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_wetgr_rayl_oguchi(q_g, m_i(i,j,k), m_w(i,j,k),&
                   fmelt_mean, n0_g, ray_const%prefactor_gm, ray_const%bfakg)
            ELSE
              z_local(i,j,k) = z_local(i,j,k) + &
                   zradar_gr_rayl_oguchi(q_g, n0_g, K_i*ray_const%prefactor_g, ray_const%bfakg)
            END IF

          END IF


        END DO
      END DO
    END DO


    WHERE (z_local >= 1d-9)
      z_local = 10.0d0 * LOG10(z_local)
    ELSEWHERE
      z_local = -99.99
    END WHERE

    WRITE (6,'(2(A,F10.3))') "  MAX dBZ = ",MAXVAL(z_local),"  MIN dBZ = ",MINVAL(z_local)

    IF ( PRESENT(z_radar) ) THEN
      z_radar = z_local
    ENDIF

    IF ( PRESENT(z_radar_850) ) THEN
      DO j = jlow, jup
        DO i = ilow, iup
          z_radar_850(i,j) = z_local(i,j,klv850)
        ENDDO
      ENDDO
    ENDIF

    IF ( PRESENT(z_radar_cmax) ) THEN
      z_radar_cmax = MAXVAL(z_local,3)
    ENDIF

    CALL finalize_tmax()

    RETURN
  END SUBROUTINE radar_lm_rayleigh_oguchi_vec


  ! Initialization of prefactors for reflectivity calculations in the global struct "ray_const":
  SUBROUTINE init_radar_lm_ray_oguchi (ldo_qc, ldo_qr, ldo_qi, ldo_qs, ldo_qg, &
       lambda_radar, mur_is_fixed, mu_rain_c, &
       ageoi, bgeoi, ageos, bgeos, ageog, bgeog, n0_g)

    IMPLICIT NONE

    ! .. Input/output variables:
    LOGICAL, INTENT(in) :: ldo_qc, ldo_qr, ldo_qi, ldo_qs, ldo_qg, mur_is_fixed
    DOUBLE PRECISION, INTENT(in) :: lambda_radar, mu_rain_c, &
         ageoi, bgeoi, ageos, bgeos, ageog, bgeog, n0_g

    ! .. Local variables:
    INTEGER, SAVE :: firstcall   = 0, &
                     firstcall_c = 0, &
                     firstcall_r = 0, &
                     firstcall_i = 0, &
                     firstcall_s = 0, &
                     firstcall_g = 0

    DOUBLE PRECISION, SAVE :: lambda_radars=0.0d0
    DOUBLE PRECISION, SAVE :: nu_Dcs=0.0d0, mu_Dcs=0.0d0
    DOUBLE PRECISION, SAVE :: nu_Dis=0.0d0, mu_Dis=0.0d0, ageois=0.0d0, bgeois=0.0d0
    DOUBLE PRECISION, SAVE :: ageoss=0.0d0, bgeoss=0.0d0
    DOUBLE PRECISION, SAVE :: ageogs=0.0d0, bgeogs=0.0d0, n0gs=0.0d0

    LOGICAL :: kw0_has_changed


    ! 0) Value for K_w_0 depending on radar wavelength:
    !

    IF (firstcall /= 1 .OR. lambda_radar /= lambda_radars) THEN
      ! Der Standardwert fuer K_w, der meist in der Radarsoftware verwendet wird
      ray_const%m_w_0 = m_complex_water_ray(lambda_radar, 0.0d0)
      ray_const%K_w_0  = (ABS( (ray_const%m_w_0**2-1.0)/(ray_const%m_w_0**2+2.0) ))**2
      lambda_radars = lambda_radar
      firstcall = 1
      kw0_has_changed = .TRUE.
    ELSE
      kw0_has_changed = .FALSE.
    END IF

    ! 1) Constant (pre)factors for Rayleigh-Oguchi-Approx. of cloud droplets:
    !

    IF (ldo_qc) THEN
      IF (firstcall_c /= 1 .OR. nu_Dcs /= ray_const%nu_Dc .OR. mu_Dcs /= ray_const%mu_Dc .OR. kw0_has_changed) THEN
        ray_const%nu_c = (ray_const%nu_Dc+1.0d0)/3.0d0 - 1.0d0
        ray_const%mu_c = ray_const%mu_Dc/3.0d0
        ray_const%prefactor_c = gfct((2.0+ray_const%nu_c+1.0)/ray_const%mu_c)        &
             & / gfct((ray_const%nu_c+1.0)/ray_const%mu_c)        &
             & * ( gfct((ray_const%nu_c+1.0)/ray_const%mu_c)      &
             &   / gfct((ray_const%nu_c+2.0)/ray_const%mu_c) )**2 &
             & * (6.0/(pi*rho_w))**2 / ray_const%K_w_0
        
        mu_Dcs = ray_const%mu_Dc
        nu_Dcs = ray_const%nu_Dc
        firstcall_c = 1
      END IF
    END IF

    ! 2) Constant (pre)factors for Rayleigh-Oguchi-Approx. of rain drops (ordinary gamma):
    !

    IF (ldo_qr) THEN
      IF (firstcall_r /= 1 .OR. kw0_has_changed) THEN
        IF (mur_is_fixed) THEN
          IF (mu_rain_c /= 0.0d0) THEN
            ray_const%lamexprc = 1.0 / (mu_rain_c+4.0d0)
            ray_const%gamfak1rc = gfct(mu_rain_c+4.0d0) / 6.0d0
            ray_const%gamfak2rc = gfct(mu_rain_c+7.0d0) / ray_const%K_w_0
          ELSE
            ray_const%lamexprc = 0.25d0
            ray_const%gamfak1rc = 1.0d0
            ray_const%gamfak2rc = 720.0d0 / ray_const%K_w_0
          END IF
        END IF
        firstcall_r = 1
      END IF
    END IF

    ! 3) Constant (pre)factors for Rayleigh-Oguchi-Approx. of cloud ice (general. gamma):
    !

    IF (ldo_qi) THEN
      IF (firstcall_i /= 1 .OR. ageoi /= ageois .OR. bgeoi /= bgeois .OR. &
           ray_const%mu_Di /= mu_Dis .OR. ray_const%nu_Di /= nu_Dis .OR. kw0_has_changed) THEN
        ray_const%nu_i = (ray_const%nu_Di + 1.0d0) / bgeoi - 1.0d0
        ray_const%mu_i = ray_const%mu_Di / bgeoi
        ray_const%prefactor_im =  gfct((2.0d0+ray_const%nu_i+1.0d0)/ray_const%mu_i)        &
             & / gfct((ray_const%nu_i+1.0d0)/ray_const%mu_i)        &
             & * ( gfct((ray_const%nu_i+1.0d0)/ray_const%mu_i)      &
             &   / gfct((ray_const%nu_i+2.0d0)/ray_const%mu_i) )**2 &
             & * (6.0d0/pi)**2 / ray_const%K_w_0
        ray_const%prefactor_i = ray_const%prefactor_im / (rho_ice**2)
        ageois = ageoi
        bgeois = bgeoi
        mu_Dis = ray_const%mu_Di
        nu_Dis = ray_const%nu_Di
        firstcall_i = 1
      END IF
    END IF

    ! 4) Constant (pre)factors for Rayleigh-Oguchi-Approx. of snow
    !    (expon. DSD WITH variable N0 (dep. on T and QS)):
    !

    IF (ldo_qs) THEN
      IF (firstcall_s /= 1 .OR. ageos /= ageoss .OR. bgeos /= bgeoss .OR. kw0_has_changed) THEN
        ray_const%bfaks = (2*bgeos+1.)/(bgeos+1.)
        ray_const%prefactor_sm = ageos**2 * gfct(2*bgeos+1.0d0) * &
             (ageos*gfct(bgeos+1.))**(-ray_const%bfaks) * (6.0/pi)**2 / ray_const%K_w_0
        ray_const%prefactor_s = ray_const%prefactor_sm / (rho_ice**2)
        ray_const%gfc_bgeop1_s = gfct(bgeos+1.0)
        ageoss = ageos
        bgeoss = bgeos
        firstcall_s = 1
      END IF
    END IF

    ! 5) Constant (pre)factors for Rayleigh-Oguchi-Approx. of graupel
    !    (expon. DSD with fixed N0):
    !
    IF (ldo_qg) THEN
      IF (firstcall_g /= 1 .OR. n0_g /= n0gs .OR. ageog /= ageogs .OR. bgeog /= bgeogs .OR. kw0_has_changed) THEN
        ray_const%bfakg = (2*bgeog+1.)/(bgeog+1.)
        ray_const%prefactor_gm = ageog**2 * n0_g * gfct(2*bgeog+1.0d0) * &
             (ageog*n0_g*gfct(bgeog+1.))**(-ray_const%bfakg) * (6.0/pi)**2 / ray_const%K_w_0
        ray_const%prefactor_g = ray_const%prefactor_gm / (rho_ice**2)
        ray_const%gfc_bgeop1_g = gfct(bgeog+1.0)
        n0gs = n0_g
        ageogs = ageog
        bgeogs = bgeog
        firstcall_g = 1
      END IF
    END IF

    RETURN
  END SUBROUTINE init_radar_lm_ray_oguchi

END MODULE radar_mie_lm_vec
