PROGRAM lm_refl_offline

  !===========================================================================================
  !
  ! DESCRIPTION:
  ! Serial program for offline reflectivity calculations and for the 
  ! COSMO radar forward operator. It is actually a wrapper around
  ! the standard Fortran90 routines for the COSMO-model radar forward operator.
  ! To exploit the full parallel computing capability of the forward operator,
  ! there is the original online coupled version of the operator within the COSMO-model.
  !
  ! METHOD:
  ! Mimicking of a serial COSMO environment (global fields, global switches and parameters),
  ! so that the standard radar modules from the COSMO-code can be applied
  ! without changing their code.
  ! 
  ! The main program is in the file
  !
  !   src/cosmo_refl_offline.f90
  !
  ! This uses some standard routines from the COSMO-model (also in src/):
  ! 
  !   data_constants.f90
  !   data_fields.f90
  !   data_io.f90
  !   data_modelconfig.f90
  !   data_nudge_all.f90
  !   data_parallel.f90
  !   data_parameters.f90
  !   data_runcontrol.f90
  !   dummy_mpi.f90
  !   environment.f90
  !   mo_fdbk_cosmo.f90
  !   mo_fdbk.f90
  !   mo_fdbk_tables.f90
  !   mo_kind.f90
  !   mo_netcdf_param.f90
  !   mo_t_netcdf_file.f90
  !   mo_t_table.f90
  !   mpif.h
  !   parallel_utilities.f90
  !   pp_utilities.f90
  !   time_utilities.f90
  !   utilities.f90
  !
  ! For the radar operator itself, the following COSMO radar modules are necessary,
  ! which are currently still under development and not part of the official model code:
  !
  ! - radar_mielib_vec.f90
  ! - radar_mie_lm_vec.f90
  ! - src_radar.f90 (only SR calc_dbz_vec)
  !
  ! NOTE FOR PRODUCING SO-CALLED "FEEDBACK FILES" FOR KENDA LETKF DATA ASSIMILATION AT DWD:
  ! For writing radar netcdf feedback files, compile with -DNETCDF and -DNUDGING.
  !
  ! NOTE FOR 2-MOMENT MICROPHYSICS (SEIFERT & BEHENG):
  ! For the 2-moment microphysics, compile with -DTWOMOMSEIFERT and -DTWOMOM_SB !
  ! In this case, the source code of this scheme is needed, i.e.
  !
  !   src_seifert.F90
  !   phillips_nucleation.incf
  !
  ! or whatever the name is now. From this code, the Makefile only takes out the
  ! module wolken_konstanten, nothing else.
  !
  !
  ! DISTRIBUTION:
  ! The software distribution is organized in the following way:
  !   drwxr-xr-x 4 ublahak fe1 4096 24. Apr 16:45 IDEAL       --> directory with exemplary namelists and input/output data sets
  !   -rw-r--r-- 1 ublahak fe1 5333 24. Apr 16:30 Makefile    --> the Makefile
  !   -rw-r--r-- 1 ublahak fe1 2635  4. Jul 11:59 README      --> THIS README
  !   drwxr-xr-x 2 ublahak fe1 8192 24. Apr 17:18 src         --> directory of source code files
  !
  ! ALGORITHM:
  ! 1) read namelist input from required file INPUT_DBZSIM
  ! 2) read in the required model fields in ASCII format: u,v,w,T,p,qv,qX (qh and qnX in case of 2-moment scheme)
  ! 3) set some global variables to appropriate values to mimick a serial COSMO environment
  ! 4) compute grid point reflectivities using SR calc_dbz_vec from src_radar.f90
  ! 5) output of the dbz-data in an ASCII-file
  ! 6) run the radar forward operator (independent reflectivity calculations, output of polar radar files
  !      and/or feedback files)
  !
  ! COMPILING AND RUNNING:
  ! 1) make sure you have appropriate versions of radar_mielib_vec.f90, radar_mie_lm_vec.f90
  !    and src_radar.f90 (and src_seifert.F90 in case of 2-moment scheme) in subdir. src/
  ! 2) Edit Makefile (choose compiler and options, there are templates for ifort and gfortran version >= 4.3) and type "make"
  ! 3) edit ./INPUT_DBZSIM (Namelist for grid point reflectivity values) and 
  !         ./INPUT_IO (Namelist for radar simulator)
  !         (See examples in subdirectory IDEAL/)
  ! 4) run ./cosmo_refl_offline.exe
  !
  !
  ! REQUIREMENT:
  ! - Recent fortran compiler which supports allocatable arrays as components of a derived type. 
  !   Normally a Fortran2003 compiler should do this, but it was for example not the case with gfortran 4.1!
  !   I use ifort 12 and gfortran >= 4.5.
  !
  ! --
  !
  ! CURRENT CODE OWNER:
  ! Ulrich Blahak, Deutscher Wetterdienst
  ! Email: ulrich.blahak@dwd.de
  ! Tel:   +49 69 8062 2393
  !
  ! VERSION HISTORY:
  ! - 24.4.2013: initial release (Ulrich Blahak)
  ! - 24.7.2013: Added organize_radar(), the radar forward operator (Ulrich Blahak)
  !
  !
  !===========================================================================================

  USE data_modelconfig, ONLY :   &
       
       dt,           & ! long time-step
       pollon,       & ! longitude of the rotated north pole (in degrees, E>0)
       pollat,       & ! latitude of the rotated north pole (in degrees, N>0)
       polgam,       & ! angle between the north poles of the systems  (???????)
       dlon,         & ! grid point distance in zonal direction (in degrees)
       dlat,         & ! grid point distance in meridional direction (in degrees)
       startlon_tot, & ! transformed latitude of the lower left grid point
                                ! of the total domain (in degrees, N>0)
       startlat_tot, & ! transformed latitude of the lower left grid point
                                ! of the total domain (in degrees, N>0)
       startlon    , & ! transformed latitude of the lower left grid point
                                ! of the local subdomain (in degrees, N>0)
       startlat    , & ! transformed latitude of the lower left grid point
                                ! of the local subdomain (in degrees, N>0)
       degrad,       & ! factor for transforming degree to rad
       raddeg,       & ! factor for transforming rad to degree
       ie,           & ! number of grid points in zonal direction
       je,           & ! number of grid points in meridional direction
       ke,           & ! number of grid points in vertical direction
       ie_tot,       & ! total number of grid points in meridional direction
       je_tot,       & ! total number of grid points in meridional direction
       ke_tot,       & ! total number of grid points in vertical direction
       istartpar,    & ! start index for computations in the parallel program
       iendpar,      & ! end index for computations in the parallel program
       jstartpar,    & ! start index for computations in the parallel program
       jendpar,      & ! end index for computations in the parallel program
       klv850

  USE data_runcontrol , ONLY :   &
       nvers,        & ! version number of experiment for documentation
       nuspecif,     & ! output unit for protocolling the namelist input
       yuspecif,     & ! file name
       nstop,        & ! last time step of the forecast
       ntstep,       & ! actual time step
       nnow,         & ! corresponds to ntstep
       nnew,         & ! corresponds to ntstep + 1
       itype_gscp,   & ! type of grid-scale precipitation physics
       lgsp,         & ! switch to include grid-scale precipitation
       ltime,        & ! switch on timing of the model run, YUTIMING
!!$       leps,         & ! switch ensemble mode on/off
       iepsmem         ! ID of the member in the ensemble (ID >= 0)

  !------------------------------------------------------------------------------

  USE data_parallel,      ONLY :  &
       num_compute,     & ! number of compute PEs
       nboundlines,     & ! number of boundary lines of the domain for which
                                ! no forecast is computed = overlapping boundary
                                ! lines of the subdomains
       my_cart_id,      & ! rank of this subdomain in the cartesian communicator
       my_world_id,     & ! rank of this subdomain in the global communicator
       isubpos,         & ! positions of the subdomains in the total domain. Given
                                ! are the i- and the j-indices of the lower left and the
                                ! upper right grid point in the order
                                ! i_ll, j_ll, i_ur, j_ur.
                                ! Only the interior of the domains are considered, not
                                ! the boundary lines.
       icomm_cart,      & ! communicator for the virtual cartesian topology
       icomm_world,     & ! communicator for all PEs
       imp_reals,       & ! determines the correct REAL type used in the model for MPI
       imp_integers,    & ! determines the correct INTEGER type used in the model for MPI
       imp_character,   & ! determines the correct CHARACTER type used in the model for MPI
       imp_logical,     & ! determines the correct LOGICAL type used in the model for MPI
       intbuf,          & ! Buffers for distributing Namelists
       logbuf,          & ! Buffers for distributing Namelists
       realbuf,         & ! Buffers for distributing Namelists
       charbuf
  !------------------------------------------------------------------------------

  USE data_io,            ONLY :  &
       root,            & ! pointer to the root of gribout namelists
       ydate_ini,       & ! start of the forecast yyyymmddhh (year,month,day,hour)
       yncglob_institution, & ! originating center name
       yncglob_source,  & ! program name and version
       dbzcalc_params, pp_nl


  !------------------------------------------------------------------------------

  USE data_constants

  USE data_runcontrol, ONLY :   ntstep
       
  USE data_parameters , ONLY :   &
       ireals,     & ! kind-type parameter for "normal" integer variables
       iintegers     ! KIND-type parameters for real variables

  USE data_parallel

  USE wolken_konstanten, ONLY:   &
       particle, rain, cloud, snow, ice, graupel, hail, rho_w_seifert => rho_w, rho_ice_seifert => rho_ice, &
       init_seifert


  USE data_fields     , ONLY :   &
       dbz        ,    & ! radar reflectivity                            (dBZ)
       u          ,    & ! zonal wind speed                              ( m/s )
       v          ,    & ! meridional wind speed                         ( m/s )
       w          ,    & ! vertical wind speed (defined on half levels)  ( m/s )
       t          ,    & ! temperature                                   (  k  )
       p          ,    & ! pressure                                      ( Pa  )
       p0         ,    & ! reference pressure                            ( Pa  )
       pp         ,    & ! pressure deviation from reference p0          ( Pa  )
       qv         ,    & ! specific water vapor content                  (kg/kg)
       qc         ,    & ! specific cloud water content                  (kg/kg)
       qi         ,    & ! specific cloud ice   content                  (kg/kg)
       qr         ,    & ! specific rain content                         (kg/kg)
       qs         ,    & ! specific snow content                         (kg/kg)
       qg         ,    & ! specific graupel content                      (kg/kg)
       qh         ,    & ! specific graupel content                      (kg/kg)
       qnc        ,    & ! number concentration cloud droplets           ( 1/kg)
       qni        ,    & ! number concentration cloud ice                ( 1/kg)
       qnr        ,    & ! number concentration raindrops                ( 1/kg)
       qns        ,    & ! number concentration snow                     ( 1/kg)
       qng        ,    & ! number concentration graupel                  ( 1/kg)
       qnh        ,    & ! number concentration graupel                  ( 1/kg)
       rho        ,    & ! density of moist air                          (kg/m3)
       hhl        ,    & ! geometrical height of half levels             ( m   )
       rlat       ,    & ! geographical latitude                         ( rad )
       rlon              ! geographical longitude                        ( rad )


  USE environment, ONLY        : get_free_unit, release_unit, model_abort

  USE src_radar, ONLY          : calc_dbz_vec, dbz_namlst_d, organize_radar

  USE utilities, ONLY          : rlarot2rla, phirot2phi


  IMPLICIT NONE

  INTEGER(KIND=iintegers) :: anz_tl

  INTEGER(KIND=iintegers) :: fehler, i, j, k, ioerr
  CHARACTER(len=250) :: namelistdatei
  CHARACTER(len=14) :: model_starttime
  CHARACTER(len=10) :: dbz_suffix
  CHARACTER(len=8) :: forecast_time
  LOGICAL          :: ldebug_radsim
  LOGICAL          :: llookup_mie      ! only for itype_gscp == 4!

  CHARACTER(len=3) :: levtyp
  CHARACTER(len=250) :: inputdir, outputdir

  TYPE(dbzcalc_params) :: dbz_meta

  LOGICAL :: linput_q_densities, gzipflag

  REAL(kind=ireals) :: lonrot, latrot

  INTEGER(kind=iintegers) :: dd_vvz, hh_vvz, mm_vvz, ss_vvz

  TYPE(pp_nl), TARGET :: outpar

  !===========================================================================================
  ! Section 0: Preparations:
  !===========================================================================================

  ! Define PI, DEGRAD and RADDEG:
  pi = 4.0_ireals * ATAN(1.0_ireals)
  degrad = pi / 180.0_ireals
  raddeg = 1.0_ireals / degrad

  !===========================================================================================
  ! Section 1: Read the namelist file
  !===========================================================================================

  ! Define some fixed parameters:
  !  - Name of namelist input file:
  namelistdatei(:) = ' '
  namelistdatei = 'INPUT_DBZSIM'

  ! Read namelist 'GRID_IN':
  CALL input_gridin (namelistdatei)

  ! Read namelist 'REFL_OFFLINE':
  CALL input_dbzsim (namelistdatei)

  CALL initializations

  !===========================================================================================
  ! Section 2: Read the input data (T, p, qv, qX, qnX) and allocate global 4D-fields
  !
  !===========================================================================================

  CALL allocate_and_read

  IF (itype_gscp >= 100) THEN
    CALL init_seifert( itype_gscp )
    WRITE (*,*) 'Cloud type: ', cloud%name
    WRITE (*,*) 'Rain type: ', rain%name
    WRITE (*,*) 'Ice type: ', ice%name
    WRITE (*,*) 'Snow type: ', snow%name
    WRITE (*,*) 'Graupel type: ', graupel%name
    IF (itype_gscp >= 2000) WRITE (*,*) 'Hail type: ', hail%name
  END IF

  ! Determination of the k-indices for those full levels which correspond
  ! to about 850 hPa (first height above): 
  IF (p(1,1,1,nnew) > 850.0e2_ireals) THEN
    klv850 = 1 
  ELSE 
    klv850 = ke
  END IF

  DO k=1, ke-1
    IF (p(1,1,k,nnew) <= 850.0e2_ireals .AND. p(1,1,k+1,nnew) > 850.0e2_ireals) klv850 = k
  END DO


  !===========================================================================================
  !
  ! Section 3:  Here comes the reflectivity calculation (SR taken from src_radar.f90)
  !
  !    NOTE: THIS REFLECTIVITY IS INDEPENDENT OF THE RADARS SIMULATED BY THE FORWARD OPERATOR BELOW.
  !
  !===========================================================================================

  ALLOCATE(dbz(ie,je,ke))
  dbz = -999.99_ireals


  ! 3) Here comes the reflectivity calculation (SR from src_radar.f90):
  call calc_dbz_vec(nnew, dbz_meta, z_radar = dbz)

  !===========================================================================================
  !
  ! Section 4: Write reflectivity data to simple ASCII file:
  !
  !===========================================================================================

  IF (TRIM(levtyp) /= 'eta') THEN
    ! Take into account that k-index runs from top to bottom!
    CALL schreibe_feld(TRIM(outputdir), TRIM(dbz_suffix), TRIM(levtyp), dbz(:,:,ke:1:-1), ie_tot, je_tot, ke_tot, &
         '# reflectivity [dBZ]', gzipflag, fehler)
  ELSE
    CALL schreibe_feld(TRIM(outputdir), TRIM(dbz_suffix), TRIM(levtyp), dbz, ie_tot, je_tot, ke_tot, &
         '# reflectivity [dBZ]', gzipflag, fehler)
  END IF

  !===========================================================================================
  !
  ! Section 4: The radar forward operator (writes his own ASCII or feedback files):
  !
  !    NOTE: THE REFLECTIVITY IS RECOMPUTED WITHIN THE OPERATOR FOR EACH SINGLE RADAR INDEPENDENTLY.
  !
  !===========================================================================================

  ! 4) Here comes the radar forward operator (SR from src_radar.f90):
  CALL organize_radar

  !===========================================================================================
  !
  ! Section 5: Some cleanup
  !
  !===========================================================================================

  CALL finalizations

  !==================================================================================================
  !==================================================================================================


CONTAINS

  !==================================================================================================
  !==================================================================================================

  SUBROUTINE input_gridin(namelistdatei)

    IMPLICIT NONE

    CHARACTER(len=*), INTENT(in) :: namelistdatei

    CHARACTER(len=250) :: errmsg
    CHARACTER(len=3)   :: levtyp_d
    REAL (KIND=ireals) :: dlon_d, dlat_d, startlon_tot_d, startlat_tot_d
    REAL (KIND=ireals) :: pollon_d, pollat_d, dt_d
    INTEGER(KIND=iintegers) :: ie_tot_d, je_tot_d, ke_tot_d, nboundlines_d

    NAMELIST /grid_in/ &
         levtyp, dlon, dlat, startlon_tot, startlat_tot, pollon, pollat, &
         ie_tot, je_tot, ke_tot, nboundlines, dt

    levtyp_d               = 'eta'
    dlon_d                 = 0.008992893_ireals
    dlat_d                 = 0.008992893_ireals
    startlon_tot_d         = -0.5_ireals
    startlat_tot_d         = -0.5_ireals
    pollon_d               = -180.0_ireals
    pollat_d               = 90.0_ireals
    ie_tot_d               = 100
    je_tot_d               = 100
    ke_tot_d               = 50
    nboundlines_d          = 3
    dt_d                   = 6.0

    levtyp               = levtyp_d
    dlon                 = dlon_d 
    dlat                 = dlat_d
    startlon_tot         = startlon_tot_d
    startlat_tot         = startlat_tot_d
    pollon               = pollon_d
    pollat               = pollat_d
    ie_tot               = ie_tot_d
    je_tot               = je_tot_d
    ke_tot               = ke_tot_d
    nboundlines          = nboundlines_d
    dt                   = dt_d

    errmsg(:) = ' '
    OPEN(10, FILE=TRIM(namelistdatei), FORM='FORMATTED', STATUS='OLD', IOSTAT=fehler, IOMSG=errmsg)
    IF(fehler /= 0) THEN
      WRITE(*,*) 'Error while opening namelist file ', TRIM(namelistdatei)
      WRITE(*,*) TRIM(errmsg)
      STOP
    ENDIF

    READ(10, grid_in, IOSTAT=fehler, IOMSG=errmsg)
    IF(fehler /= 0) THEN
      IF (fehler > 0) THEN
        WRITE(*,*) 'Error while reading namelist file ', TRIM(namelistdatei)
        WRITE(*,*) TRIM(errmsg)
      ELSE IF (fehler < 0) THEN
        WRITE(*,*) ' ERROR    *** Premature EOF encountered while reading namelist file ', TRIM(namelistdatei)
        WRITE(*,*) TRIM(errmsg)  
      ENDIF
      STOP
    ENDIF

    CLOSE(10)


  END SUBROUTINE input_gridin


  SUBROUTINE input_dbzsim(namelistdatei)

    IMPLICIT NONE

    CHARACTER(len=*), INTENT(in) :: namelistdatei

    CHARACTER(len=250) :: errmsg
    CHARACTER(len=14) :: model_starttime_d
    CHARACTER(len=10) :: dbz_suffix_d
    CHARACTER(len=8) :: forecast_time_d
    LOGICAL          :: ldebug_radsim_d
    LOGICAL          :: llookup_mie_d      ! only for itype_gscp == 4!
    CHARACTER(len=250) :: inputdir_d, outputdir_d
    TYPE(dbzcalc_params) :: dbz_meta_d
    INTEGER(KIND=iintegers) :: itype_gscp_d
    REAL (KIND=ireals)      :: dlon_d, dlat_d, startlat_tot_d
    LOGICAL :: linput_q_densities_d, gzipflag_d

    NAMELIST /refl_offline/ &
         itype_gscp, dbz_meta, linput_q_densities, &
         model_starttime, forecast_time, dbz_suffix, gzipflag, ldebug_radsim, llookup_mie, &
         inputdir, outputdir

    ! 1) Read namelist file:
    !  a) default-values:
    dbz_meta_d = dbz_namlst_d
    itype_gscp_d = 3
    linput_q_densities_d     = .FALSE.
    gzipflag_d             = .FALSE.
    dbz_suffix_d             = 'dbz-xxx  '
    ldebug_radsim_d        = .FALSE.
    model_starttime_d(:)   = ' '
    model_starttime_d      = '2003032100'
    forecast_time_d(:)     = ' '
    forecast_time_d        = '00000000'
    inputdir_d(:) = ' '
    outputdir_d(:) = ' '

    !  b) initialize namelist parameters with defaults:
    dbz_meta = dbz_meta_d
    itype_gscp        = itype_gscp_d
    linput_q_densities  = linput_q_densities_d
    gzipflag             = gzipflag_d
    dbz_suffix             = dbz_suffix_d
    ldebug_radsim = ldebug_radsim_d
    model_starttime = model_starttime_d
    forecast_time = forecast_time_d
    inputdir = inputdir_d
    outputdir = outputdir_d

    errmsg(:) = ' '
    OPEN(10, FILE=TRIM(namelistdatei), FORM='FORMATTED', STATUS='OLD', IOSTAT=fehler, IOMSG=errmsg)
    IF(fehler /= 0) THEN
      WRITE(*,*) 'Error while opening namelist file ', TRIM(namelistdatei)
      WRITE(*,*) TRIM(errmsg)
      STOP
    ENDIF

    READ(10, refl_offline, IOSTAT=fehler, IOMSG=errmsg)
    IF(fehler /= 0) THEN
      IF (fehler > 0) THEN
        WRITE(*,*) 'Error while reading namelist file ', TRIM(namelistdatei)
        WRITE(*,*) TRIM(errmsg)
      ELSE IF (fehler < 0) THEN
        WRITE(*,*) ' ERROR    *** Premature EOF encountered while reading namelist file ', TRIM(namelistdatei)
        WRITE(*,*) TRIM(errmsg)  
      ENDIF
      STOP
    ENDIF

    CLOSE(10)

    IF (TRIM(levtyp) /= 'z' .AND. TRIM(levtyp) /= 'p' .AND. TRIM(levtyp) /= 'eta') THEN
      WRITE (*,*) 'Unknown type of vertical levels. z, p or eta required.'
      STOP
    END IF

    IF (TRIM(dbz_suffix) == 'dbz') THEN
      WRITE (*,*) 'Don''t name it ''dbz''! It might overwrite the original LM-dbz-file!'
      STOP
    END IF

    IF (dbz_meta%itype_refl < 1 .OR. dbz_meta%itype_refl > 4) THEN
      WRITE (*,*) '  Unknown type of radar reflectivity calculation: ',dbz_meta%itype_refl
      STOP
    END IF

    IF (llookup_mie .AND. itype_gscp /= 4) THEN
      WRITE (*,*) '  llookup_mie = .true. only possible for itype_gscp = 4, but actual itype_gscp = ',itype_gscp
      STOP
    END IF

  END SUBROUTINE input_dbzsim

  !==================================================================================================
  !==================================================================================================

  SUBROUTINE initializations

    IMPLICIT NONE

    nuspecif = 123
    yuspecif = 'YUSPECIF'
    OPEN (nuspecif, file='YUSPECIF', status='replace', iostat=ioerr)
    IF (ioerr /= 0) THEN
      fehler = 1
      WRITE (*,*) 'Error opening file YUSPECIF !'
      STOP
    END IF

    lgsp = .TRUE.
    ltime = .FALSE.
    iepsmem = -1

    nvers = 0

    ! This program only for serial compiles, therefore set my_cart_id = 0:
    my_cart_id = 0
    my_world_id = 0
    num_compute = 1

    ALLOCATE(isubpos(0:0,1:4))
    isubpos(0,1) = nboundlines + 1
    isubpos(0,2) = nboundlines + 1
    isubpos(0,3) = ie_tot - nboundlines
    isubpos(0,4) = je_tot - nboundlines

    icomm_cart = -1
    icomm_world = -1          
    imp_reals = -1
    imp_integers = -1
    imp_character = -1
    imp_logical = -1

    ALLOCATE(intbuf(10));  intbuf = 0
    ALLOCATE(logbuf(10));  logbuf = .FALSE.
    ALLOCATE(realbuf(10)); realbuf = 0
    ALLOCATE(charbuf(10)); charbuf(:)(:) = ' '

    ! Set the number of the current timestep:
    READ (forecast_time,'(4i2)'), dd_vvz, hh_vvz, mm_vvz, ss_vvz
    ntstep = NINT( ( ss_vvz + 60.0*(mm_vvz + 60.0*(hh_vvz + 24.0*dd_vvz) ) )  / dt )
    nstop  = ntstep  ! ntstop has to be = ntstep
    nnow = 1         ! nnew has to be 1
    nnew = 1         ! nnew has to be 1
    anz_tl = 1       ! 4. dimension of the fields (time dimension) has to be 1
    startlon = startlon_tot
    startlat = startlat_tot

    NULLIFY(root)
    root => outpar
    root%ydir(:)           = ' '
    root%ydir              = outputdir
    ydate_ini(:)           = ' '
    IF (LEN_TRIM(model_starttime) == 10) THEN
      ydate_ini              = model_starttime // '0000'
    ELSE IF (LEN_TRIM(model_starttime) == 14) THEN
      ydate_ini              = model_starttime
    ELSE
      WRITE (*,*) 'Wrong format of namelist parameter model_starttime: 10 or 14 digit date format YYYYMMDDhh(mmss) required!'
      STOP
    END IF
    yncglob_institution(:) = ' '
    yncglob_institution    = '-'
    yncglob_source(:)      = ' '
    yncglob_source         = '-'

    ! Set loop dimensions properly:
    ie = ie_tot
    je = je_tot
    ke = ke_tot
    istartpar = 1
    iendpar = ie_tot
    jstartpar = 1
    jendpar = je_tot

    ! compute rlat, rlon:
    ALLOCATE (rlon(ie_tot,je_tot))
    ALLOCATE (rlat(ie_tot,je_tot))
    polgam = 0.0_ireals            ! Has to be 0.0 so that the geographical system is really the geographical system

    DO j = 1, je_tot
      DO i = 1, ie_tot
        lonrot = startlon_tot + (i-1) * dlon
        latrot = startlat_tot + (j-1) * dlat
        rlon(i,j) = rlarot2rla ( latrot, lonrot, pollat, pollon, polgam ) * degrad
        rlat(i,j) = phirot2phi ( latrot, lonrot, pollat, pollon, polgam ) * degrad
      END DO
    END DO
    
  END SUBROUTINE initializations
  

  SUBROUTINE allocate_and_read

    IMPLICIT NONE

    ALLOCATE (hhl(ie_tot,je_tot,ke_tot+1))
    CALL leselm(TRIM(inputdir), 'hhl', TRIM(levtyp), hhl(:,:,:), fehler, dim3=ke_tot+1)
    IF (fehler /= 0) THEN
      WRITE (*,*) 'Error reading HHL file!'
      STOP
    END IF

    ALLOCATE (u(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 'u', TRIM(levtyp), u(:,:,:,nnew), fehler)
    IF (fehler /= 0) THEN
      WRITE (*,*) 'Error reading U file!'
      STOP
    END IF

    ALLOCATE (v(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 'v', TRIM(levtyp), v(:,:,:,nnew), fehler)
    IF (fehler /= 0) THEN
      WRITE (*,*) 'Error reading V file!'
      STOP
    END IF

    ALLOCATE (w(ie_tot,je_tot,ke_tot+1,anz_tl))
    CALL leselm(TRIM(inputdir), 'w', TRIM(levtyp), w(:,:,:,nnew), fehler, dim3=ke_tot+1)
    IF (fehler /= 0) THEN
      WRITE (*,*) 'Error reading W file!'
      STOP
    END IF

    ALLOCATE (t(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 't', TRIM(levtyp), t(:,:,:,nnew), fehler)
    IF (fehler /= 0) THEN
      WRITE (*,*) 'Error reading temperature file!'
      STOP
    END IF

    ALLOCATE (p(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 'p', TRIM(levtyp), p(:,:,:,nnew), fehler)
    IF (fehler /= 0) THEN
      WRITE (*,*) 'Error reading pressure file!'
      STOP
    END IF
    ! After reading pressure, it is split arbitrarily into p0 and pp simply by
    ! setting p0 = 1000 hPa. The radar code only uses the combination of both,
    ! so this is justified!
    ALLOCATE (p0(ie_tot,je_tot,ke_tot))
    p0 = 1e5_ireals
    ALLOCATE (pp(ie_tot,je_tot,ke_tot,anz_tl))
    pp(:,:,:,nnew) = p(:,:,:,nnew) - p0

    ALLOCATE (qv(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 'qv', TRIM(levtyp), qv(:,:,:,nnew), fehler)
    IF (fehler /= 0) THEN
      WRITE (*,*) 'Error reading qv file!'
      qv = 0.0
    END IF

    ALLOCATE (qc(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 'qc', TRIM(levtyp), qc(:,:,:,nnew), fehler)
    IF (fehler /= 0) STOP
    ALLOCATE (qr(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 'qr', TRIM(levtyp), qr(:,:,:,nnew), fehler)
    IF (fehler /= 0) STOP
    ALLOCATE (qi(ie_tot,je_tot,ke_tot,anz_tl))
    CALL leselm(TRIM(inputdir), 'qi', TRIM(levtyp), qi(:,:,:,nnew), fehler)
    IF (fehler /= 0) STOP

    ALLOCATE(qs(ie_tot,je_tot,ke_tot,anz_tl))
    IF (itype_gscp >= 3) THEN
      CALL leselm(TRIM(inputdir), 'qs', TRIM(levtyp), qs(:,:,:,nnew), fehler)
      IF (fehler /= 0) STOP
    ELSE
      qs = 0.0
    END IF

    ALLOCATE(qg(ie_tot,je_tot,ke_tot,anz_tl))
    IF (itype_gscp >= 4) THEN
      CALL leselm(TRIM(inputdir), 'qg', TRIM(levtyp), qg(:,:,:,nnew), fehler)
      IF (fehler /= 0) STOP
    ELSE
      qg = 0.0
    END IF

    IF (itype_gscp >= 100) THEN
      ALLOCATE(qh(ie_tot,je_tot,ke_tot,anz_tl))
      CALL leselm(TRIM(inputdir), 'qh', TRIM(levtyp), qh(:,:,:,nnew), fehler)
      IF (fehler /= 0) THEN
        qh = 0.0
      END IF
      ALLOCATE(qnc(ie_tot,je_tot,ke_tot,anz_tl))
      CALL leselm(TRIM(inputdir), 'qnc', TRIM(levtyp), qnc(:,:,:,nnew), fehler)
      IF (fehler /= 0) STOP
      ALLOCATE(qnr(ie_tot,je_tot,ke_tot,anz_tl))
      CALL leselm(TRIM(inputdir), 'qnr', TRIM(levtyp), qnr(:,:,:,nnew), fehler)
      IF (fehler /= 0) STOP
      ALLOCATE(qni(ie_tot,je_tot,ke_tot,anz_tl))
      CALL leselm(TRIM(inputdir), 'qni', TRIM(levtyp), qni(:,:,:,nnew), fehler)
      IF (fehler /= 0) STOP
      ALLOCATE(qns(ie_tot,je_tot,ke_tot,anz_tl))
      CALL leselm(TRIM(inputdir), 'qns', TRIM(levtyp), qns(:,:,:,nnew), fehler)
      IF (fehler /= 0) STOP
      ALLOCATE(qng(ie_tot,je_tot,ke_tot,anz_tl))
      CALL leselm(TRIM(inputdir), 'qng', TRIM(levtyp), qng(:,:,:,nnew), fehler)
      IF (fehler /= 0) STOP
      ALLOCATE(qnh(ie_tot,je_tot,ke_tot,anz_tl))
      CALL leselm(TRIM(inputdir), 'qnh', TRIM(levtyp), qnh(:,:,:,nnew), fehler)
      IF (fehler /= 0) THEN
        qnh = 0.0
      END IF

    END IF

    ! 2a) Re-sort the k-index of the data if needed.
    !     For the radar routines, it has to be sorted from top to bottom,
    !     like the model level (eta-level) fields in COSMO.

    IF (TRIM(levtyp) /= 'eta') THEN
      u(:,:,:,nnew) = u(:,:,ke_tot:1:-1,nnew)
      v(:,:,:,nnew) = v(:,:,ke_tot:1:-1,nnew)
      w(:,:,:,nnew) = w(:,:,ke_tot+1:1:-1,nnew)
      t(:,:,:,nnew) = t(:,:,ke_tot:1:-1,nnew)
      p(:,:,:,nnew) = p(:,:,ke_tot:1:-1,nnew)
      p0(:,:,:) = p0(:,:,ke_tot:1:-1)
      pp(:,:,:,nnew) = pp(:,:,ke_tot:1:-1,nnew)
      qv(:,:,:,nnew) = qv(:,:,ke_tot:1:-1,nnew)
      qc(:,:,:,nnew) = qc(:,:,ke_tot:1:-1,nnew)
      qr(:,:,:,nnew) = qr(:,:,ke_tot:1:-1,nnew)
      qi(:,:,:,nnew) = qi(:,:,ke_tot:1:-1,nnew)
      qs(:,:,:,nnew) = qs(:,:,ke_tot:1:-1,nnew)
      qg(:,:,:,nnew) = qg(:,:,ke_tot:1:-1,nnew)
      IF (itype_gscp >= 100) THEN
        qh(:,:,:,nnew) = qh(:,:,ke_tot:1:-1,nnew)
        qnc(:,:,:,nnew) = qnc(:,:,ke_tot:1:-1,nnew)
        qnr(:,:,:,nnew) = qnr(:,:,ke_tot:1:-1,nnew)
        qni(:,:,:,nnew) = qni(:,:,ke_tot:1:-1,nnew)
        qns(:,:,:,nnew) = qns(:,:,ke_tot:1:-1,nnew)
        qng(:,:,:,nnew) = qng(:,:,ke_tot:1:-1,nnew)
        qnh(:,:,:,nnew) = qnh(:,:,ke_tot:1:-1,nnew)
      END IF
    END IF

    ! 3) Calculate total density:

    ALLOCATE(rho(ie_tot, je_tot, ke_tot))
    rho = 0.0

    IF (linput_q_densities) THEN

      ! qx und qnx wurden als Dichten eingelesen. Diese auf massenspez. Groessen umrechnen:
      ! Zuerst: Dichte berechnen (LM-Naeherung umgeschrieben auf qx = Dichten):

      IF (itype_gscp < 100) THEN
        rho = (p(:,:,:,nnew) / ( r_d*t(:,:,:,nnew) ) + &
             qc(:,:,:,nnew) + qr(:,:,:,nnew) + qi(:,:,:,nnew) + qs(:,:,:,nnew) + qg(:,:,:,nnew)) / &
             (1.0 + rvd_m_o*qv(:,:,:,nnew))
      ELSE
        rho = (p(:,:,:,nnew) / ( r_d*t(:,:,:,nnew) ) + &
             qc(:,:,:,nnew) + qr(:,:,:,nnew) + qi(:,:,:,nnew) + qs(:,:,:,nnew) + qg(:,:,:,nnew) + qh(:,:,:,nnew)) / &
             (1.0 + rvd_m_o*qv(:,:,:,nnew))
      END IF

      ! Dann umrechnen:
      qc(:,:,:,nnew) = qc(:,:,:,nnew) / rho
      qr(:,:,:,nnew) = qr(:,:,:,nnew) / rho
      qi(:,:,:,nnew) = qi(:,:,:,nnew) / rho
      qs(:,:,:,nnew) = qs(:,:,:,nnew) / rho
      qg(:,:,:,nnew) = qg(:,:,:,nnew) / rho
      IF (itype_gscp >= 100) THEN
        qh(:,:,:,nnew) = qh(:,:,:,nnew) / rho
        qnc(:,:,:,nnew) = qnc(:,:,:,nnew) / rho
        qnr(:,:,:,nnew) = qnr(:,:,:,nnew) / rho
        qni(:,:,:,nnew) = qni(:,:,:,nnew) / rho
        qns(:,:,:,nnew) = qns(:,:,:,nnew) / rho
        qng(:,:,:,nnew) = qng(:,:,:,nnew) / rho
        qnh(:,:,:,nnew) = qnh(:,:,:,nnew) / rho
      END IF

    ELSE

      IF (itype_gscp < 100) THEN
        rho = p(:,:,:,nnew) / ( r_d*t(:,:,:,nnew)*(1.0 + rvd_m_o*qv(:,:,:,nnew) - &
             qc(:,:,:,nnew) - qr(:,:,:,nnew) - qi(:,:,:,nnew) - qs(:,:,:,nnew) - qg(:,:,:,nnew)) )

      ELSE
        rho = p(:,:,:,nnew) / ( r_d*t(:,:,:,nnew)*(1.0 + rvd_m_o*qv(:,:,:,nnew) - &
             qc(:,:,:,nnew) - qr(:,:,:,nnew) - qi(:,:,:,nnew) - qs(:,:,:,nnew) - qg(:,:,:,nnew) - qh(:,:,:,nnew)) )
      END IF

    END IF

  END SUBROUTINE allocate_and_read


  SUBROUTINE finalizations
    
    CLOSE (nuspecif)

    IF (ALLOCATED(u)) DEALLOCATE(u)
    IF (ALLOCATED(v)) DEALLOCATE(v)
    IF (ALLOCATED(w)) DEALLOCATE(w)
    IF (ALLOCATED(t)) DEALLOCATE(t)
    IF (ALLOCATED(p)) DEALLOCATE(p)
    IF (ALLOCATED(p0)) DEALLOCATE(p0)
    IF (ALLOCATED(pp)) DEALLOCATE(pp)
    IF (ALLOCATED(qv)) DEALLOCATE(qv)
    IF (ALLOCATED(qc)) DEALLOCATE(qc)
    IF (ALLOCATED(qnc)) DEALLOCATE(qnc)
    IF (ALLOCATED(qr)) DEALLOCATE(qr)
    IF (ALLOCATED(qnr)) DEALLOCATE(qnr)
    IF (ALLOCATED(qi)) DEALLOCATE(qi)
    IF (ALLOCATED(qni)) DEALLOCATE(qni)
    IF (ALLOCATED(qs)) DEALLOCATE(qs)
    IF (ALLOCATED(qns)) DEALLOCATE(qns)
    IF (ALLOCATED(qg)) DEALLOCATE(qg)
    IF (ALLOCATED(qng)) DEALLOCATE(qng)
    IF (ALLOCATED(qh)) DEALLOCATE(qh)
    IF (ALLOCATED(qnh)) DEALLOCATE(qnh)
    IF (ALLOCATED(dbz)) DEALLOCATE(dbz)


  END SUBROUTINE finalizations

  !==================================================================================================
  !==================================================================================================

  SUBROUTINE leselm (indir, cvar, levtyp, var, fehler, dim3) 
    IMPLICIT NONE
    CHARACTER(*), INTENT(in) :: indir, cvar, levtyp
    INTEGER, INTENT(in), OPTIONAL :: dim3
    INTEGER, INTENT(out) :: fehler
    INTEGER :: ie_in, je_in, ke_in
    INTEGER :: iunit, i, j, k, ke_loc, error
    CHARACTER(len=250) :: dateiname, header
    LOGICAL :: fileexist, is_gzipped
    CHARACTER(len=250) :: tmpdir
    REAL(kind=ireals)  :: var(:,:,:)

    fehler = 0
    iunit = 20
    is_gzipped = .FALSE.
    
    ! Make the third dimension of the output field optional because
    ! of possible ke and ke+1 field dims:
    IF (PRESENT(dim3)) THEN
      ke_loc = dim3
    ELSE
      ke_loc = ke_tot
    END IF

    tmpdir(:) = ' '
    IF (LEN_TRIM(indir) > 0) THEN
      tmpdir = TRIM(ADJUSTL(indir))//'/'
    END IF

    dateiname(:) = ' '
    dateiname = TRIM(tmpdir)//TRIM(cvar)//'_'//TRIM(model_starttime)//'_'//TRIM(forecast_time)//'_'//TRIM(levtyp)//'.dat'
    INQUIRE(file=TRIM(ADJUSTL(dateiname)), exist=fileexist)
    IF (fileexist) THEN
      OPEN(iunit, file=TRIM(ADJUSTL(dateiname)), status='old', iostat=error)
      IF (error /= 0) THEN
        fehler = 1
        WRITE (*,*) 'Error opening ', TRIM(dateiname), ' !'
        RETURN
      END IF
    ELSE
      dateiname(:) = ' '
      dateiname = TRIM(tmpdir)//TRIM(cvar)//'_'//TRIM(model_starttime)//'_'//TRIM(forecast_time)//'_'//TRIM(levtyp)//'.dat.gz'
      INQUIRE(file=TRIM(ADJUSTL(dateiname)), exist=fileexist)
      IF (fileexist) THEN
        is_gzipped = .TRUE.
        CALL system('gunzip '//trim(dateiname))
        dateiname(:) = ' '
        dateiname = TRIM(tmpdir)//TRIM(cvar)//'_'//TRIM(model_starttime)//'_'//TRIM(forecast_time)//'_'//TRIM(levtyp)//'.dat'
        OPEN(iunit, file=TRIM(ADJUSTL(dateiname)), status='old', iostat=error)
        IF (error /= 0) THEN
          fehler = 1
          WRITE (*,*) 'Error opening ', TRIM(ADJUSTL(dateiname)), ' !'
          RETURN
        END IF
      ELSE
        fehler = 1
        WRITE (*,*) 'File ', TRIM(ADJUSTL(dateiname)), ' does not exist!'
        RETURN
      END IF

    END IF

    ! 2 Zeilen Header lesen:
    header = REPEAT(' ', LEN(header))
    READ (iunit, '(A)') header
    WRITE (*,*) TRIM(header)
    READ (iunit, *) ie_in, je_in, ke_in

    IF ( ie_in /= ie_tot .OR. &
         je_in /= je_tot .OR. &
         ke_in /= ke_loc        ) THEN
      fehler = 2
      WRITE (*,*) 'Wrong data dimension in ', TRIM(ADJUSTL(dateiname)), ': ', &
           ie_in, je_in, ke_in, ' instead of ', ie_tot, je_tot, ke_loc
      CLOSE(iunit)
      RETURN
    END IF

    DO k=1,ke_in
      DO j=1,je_in
        DO i=1,ie_in
          READ(iunit,*) var(i,j,k)
        END DO
      END DO
    END DO

    CLOSE(iunit)

    IF (is_gzipped) THEN
      CALL system('gzip -f '//TRIM(ADJUSTL(dateiname)))
    END IF

    RETURN
  END SUBROUTINE leselm

  SUBROUTINE schreibe_feld(outdir, cvar, levtyp, feld, ie, je, ke, header, gzipflag, fehler)

    IMPLICIT NONE

    CHARACTER(*), INTENT(in) :: cvar, levtyp, outdir
    INTEGER(kind=iintegers), INTENT(in) :: ie, je, ke
    CHARACTER(len=*), INTENT(in) :: header
    REAL(kind=ireals), INTENT(in) :: feld(ie,je,ke)
    INTEGER(kind=iintegers), INTENT(out) :: fehler

    INTEGER(kind=iintegers) :: iunit, i, j, k, error
    CHARACTER(len=120) :: dateiname
    CHARACTER(len=20) :: cie, cje, cke
    LOGICAL, INTENT(in) :: gzipflag
    CHARACTER(len=250) :: tmpdir

    fehler = 0
    iunit = 30

    tmpdir(:) = ' '
    IF (LEN_TRIM(outdir) > 0) THEN
      tmpdir = TRIM(ADJUSTL(outdir))//'/'
    END IF

    dateiname = REPEAT(' ', len(dateiname))
    dateiname = TRIM(tmpdir)//TRIM(cvar)//'_'//TRIM(model_starttime)//'_'//TRIM(forecast_time)//'_'//TRIM(levtyp)//'.dat'
    OPEN(iunit, file=TRIM(ADJUSTL(dateiname)), status='replace', iostat=error)
    IF (error /= 0) THEN
      fehler = 1
      WRITE (*,*) 'Error opening ', TRIM(ADJUSTL(dateiname)), ' !'
      RETURN
    END IF

    cie = REPEAT(' ', LEN(cie))
    cje = REPEAT(' ', LEN(cje))
    cke = REPEAT(' ', LEN(cke))
    WRITE (cie, '(i20)') ie
    WRITE (cje, '(i20)') je
    WRITE (cke, '(i20)') ke
    WRITE (iunit,'(a)') header
    WRITE (iunit,'(a,x,a,x,a)') TRIM(ADJUSTL(cie)), TRIM(ADJUSTL(cje)), TRIM(ADJUSTL(cke))
    DO k=1,ke
      DO j=1,je
        DO i=1,ie
          IF (ABS(feld(i,j,k)) < 1.0d-30) THEN
            WRITE (iunit, '(i1)') 0
          ELSE
            WRITE (iunit, '(es13.5)') feld(i,j,k)
          END IF
        END DO
      END DO
    END DO


    CLOSE(iunit)

    IF (gzipflag) THEN
      CALL system('gzip -f '//TRIM(ADJUSTL(dateiname)))
    END IF


    RETURN
  END SUBROUTINE schreibe_feld


END PROGRAM lm_refl_offline
