!!$ Add to COMFLG1 for compiling:
!!$  -pi rexp=m_complex_bruggemann_vec,m_complex_maxwellgarnett_vec --> does not help much ...
!!$     maybe inlining is more beneficial on another machine ...

!!$ get_m_mix_vec and get_m_mix_nested_vec seem to be very slow because of the SELECT CASE 's

!------------------------------------------------------------------------------------------!
!
! Karlsruhe Libraries for Radar Reflectivity Calculations based on Mie-Theory
!
! by Uli Blahak, in an early stage based on a module by Axel Seifert
!
! (Calculates radar reflectivity of precipitation particles using 
! full mie theorie which includes scattering by wet and melting particles)
!
! Applicable to microwave radiation, wavelength > 1 mm
!
! Contact: ulrich.blahak@imk.fzk.de
!
!------------------------------------------------------------------------------------------!
!
! Applicable to microwave radiation, wavelength > 1 mm
!
! This version is the first shot on vectorizing the code.
!
!=====================================================================
!
! Das folgende Modul radar_sphbessel_vec stellt die
! sphaerischen Besselfunktionen 1. und 2. Gattung mit komplexem 
! Argument und ganzzahliger Ordnung und deren Ableitungen bereit.
!
!=====================================================================

MODULE radar_sphbessel_vec

  IMPLICIT NONE

  INTEGER, PARAMETER :: kinddp = KIND(1.0d0)

  PRIVATE MSTA1VEC, MSTA2VEC, ENVJ, ENVJVEC

CONTAINS

!=====================================================================
!
! Diverse Hilfsfunktionen fuer die Berechnung der sphaerischen 
! Besselfunktionen in CSPHJY() weiter unten.
!
!=====================================================================

  FUNCTION ENVJVEC(N,X,anz) RESULT (Y)
    INTEGER :: anz
    INTEGER :: N(anz)
    DOUBLE PRECISION :: X(anz)
    DOUBLE PRECISION :: Y(anz)
    Y = 0.5D0*LOG10(6.28D0*N)-N*LOG10(1.36D0*X/N)
    RETURN
  END FUNCTION ENVJVEC

  FUNCTION ENVJ(N,X) RESULT (Y)
    INTEGER :: N
    DOUBLE PRECISION :: X
    DOUBLE PRECISION :: Y
    Y = 0.5D0*LOG10(6.28D0*N)-N*LOG10(1.36D0*X/N)
    RETURN
  END FUNCTION ENVJ

  FUNCTION MSTA1VEC(X,MP,anz) RESULT (Y)

    !       ===================================================
    !       Purpose: Determine the starting point for backward  
    !                recurrence such that the magnitude of    
    !                Jn(x) at that point is about 10^(-MP)
    !       Input :  x     --- Argument of Jn(x)
    !                MP    --- Value of magnitude
    !       Output:  MSTA1 --- Starting point   
    !       ===================================================

    IMPLICIT NONE

    INTEGER :: anz
    INTEGER :: MP
    DOUBLE PRECISION :: X(anz)
    DOUBLE PRECISION :: Y(anz)
    DOUBLE PRECISION :: A0(anz), F(anz), F0(anz), F1(anz)
    INTEGER :: N0(anz), N1(anz), IT, NN(anz), K

    A0 = ABS(X)
    N0 = INT(1.1*A0) + 1
    F0 = ENVJVEC(N0,A0,anz) - MP
    N1 = N0 + 5
    F1 = ENVJVEC(N1,A0,anz) - MP
    NN = N1-(N1-N0)/(1.0D0-F0/F1)                  
    F  = ENVJVEC(NN,A0,anz)-MP
    DO IT=1,20
      DO k=1,anz
        IF(ABS(NN(k)-N1(k)) >= 1) THEN
          N0(k)=N1(k)
          F0(k)=F1(k)
          N1(k)=NN(k)
          F1(k)=F(k)
          NN(k)=N1(k)-(N1(k)-N0(k))/(1.0D0-F0(k)/F1(k))                  
          F(k) =ENVJ(NN(k),A0(k))-MP
        END IF
      END DO
    END DO
    Y=NN

    RETURN
  END FUNCTION MSTA1VEC

  FUNCTION MSTA2VEC(X,N,MP,anz) RESULT (Y)

    !       ===================================================
    !       Purpose: Determine the starting point for backward
    !                recurrence such that all Jn(x) has MP
    !                significant digits
    !       Input :  x  --- Argument of Jn(x)
    !                n  --- Order of Jn(x)
    !                MP --- Significant digit
    !       Output:  MSTA2 --- Starting point
    !       ===================================================

    IMPLICIT NONE

    INTEGER :: anz
    INTEGER :: N, MP
    DOUBLE PRECISION :: X(anz)
    DOUBLE PRECISION :: Y(anz)

    INTEGER :: N0(anz), N1(anz), IT, NN(anz), K
    DOUBLE PRECISION :: EJN, OBJ(anz), A0(anz), HMP, F(anz), F0(anz), F1(anz)

    A0 = ABS(X)
    HMP = 0.5D0*MP
    DO K=1,anz
      EJN = ENVJ(N,A0(K))
      IF (EJN < HMP) THEN
        OBJ(K) = MP
        N0(K) = INT(1.1*A0(K))
      ELSE
        OBJ(K) = HMP + EJN
        N0(K) = N
      END IF
    END DO
    F0 = ENVJVEC(N0,A0,anz)-OBJ
    N1 = N0 + 5
    F1 = ENVJVEC(N1,A0,anz) - OBJ
    NN = N1-(N1-N0)/(1.0D0-F0/F1)
    F = ENVJVEC(NN,A0,anz) - OBJ
    DO IT=1,20
      DO k=1,anz
        IF (ABS(NN(k)-N1(k)) >= 1) THEN
          N0(k)=N1(k)
          F0(k)=F1(k)
          N1(k)=NN(k)
          F1(k)=F(k)
          NN(k) = N1(k)-(N1(k)-N0(k))/(1.0D0-F0(k)/F1(k))
          F(k) = ENVJ(NN(k),A0(k)) - OBJ(k)
        END IF
      END DO
    END DO
    Y = NN + 10

    RETURN
  END FUNCTION MSTA2VEC

  SUBROUTINE CSPHJYVEC(N,Z,NM,CSJ,CDJ,CSY,CDY,anz)
!*********************************************************************
!
! Vectorized version!
!
!*********************************************************************
    !       ==========================================================
    !       Purpose: Compute spherical Bessel functions jn(z) & yn(z)
    !                and their derivatives for a complex argument up
    !                to order N.
    !       Input :  z --- Complex argument
    !                n --- Order of jn(z) & yn(z) ( n = 0,1,2,... )
    !
    !       Output:  Vectors CSJ(0:N), CDJ(0:N),CSY(0:N),CDY(0:N):
    !                CSJ(n) --- jn(z)
    !                CDJ(n) --- jn'(z)
    !                CSY(n) --- yn(z)
    !                CDY(n) --- yn'(z)
    !                NM --- Highest order computed
    !                --> Higher orders are virtually 0.0 and need not
    !                be computed!!!
    !       Routines called:
    !                MSTA1 and MSTA2 for computing the starting
    !                point for backward recurrence
    !
    !       From: "Computation of Special Functions",  Shanjie Zhang, Jianming Jin, 1996, Wiley,
    !              ISBN: 0-471-11963-6
    !
    !       ==========================================================

    IMPLICIT NONE

    INTEGER :: anz
    INTEGER :: N, NM(anz)
    COMPLEX(kind=kinddp) :: CSJ(anz,0:N),CDJ(anz,0:N),CSY(anz,0:N),CDY(anz,0:N), Z(anz)

    COMPLEX(kind=kinddp) :: CSA(anz), CSB(anz), CF0(anz), CF1(anz), CF(anz), CS(anz)

    INTEGER :: K, M(anz), J, Mhelp(anz)

    DOUBLE PRECISION :: A0(anz)

    A0=ABS(Z)
    NM=N

    ! Preset values for abs(A0) < 1d-60, which will
    ! be overwritten later in case of abs(A0) >= 1d-60.
    CSJ(:,0:N)=0.0D0
    CDJ(:,0:N)=0.0D0
    CSY(:,0:N)=-1.0D+300
    CDY(:,0:N)=1.0D+300
    CSJ(:,0)=(1.0D0,0.0D0)
    CDJ(:,1)=(0.333333333333333D0,0.0D0)

    DO j=1,anz
      IF (A0(j) >= 1.0D-60) THEN
        CSJ(j,0)=SIN(Z(j))/Z(j)
        CSJ(j,1)=(CSJ(j,0)-COS(Z(j)))/Z(j)
      END IF
    END DO

    IF (N >= 2) THEN
      CSA=CSJ(:,0)
      CSB=CSJ(:,1)
      M=MSTA1VEC(A0,200,anz)
      WHERE (M < N)
        NM=M
      END WHERE
      Mhelp=MSTA2VEC(A0,N,15,anz)
      WHERE (M >= N)
        M = Mhelp
      END WHERE

      CF0=0.0D0
      CF1=1.0D-100
      DO K=MAXVAL(M),0,-1
        DO j=1,anz
          IF (A0(j) >= 1.0D-60 .AND. K <= M(j)) THEN
            CF(j)=(2.0D0*K+3.0D0)*CF1(j)/Z(j)-CF0(j)
            CF0(j)=CF1(j)
            CF1(j)=CF(j)
          END IF
          IF (A0(j) >= 1.0D-60 .AND. K <= M(j) .AND. K <= NM(j)) THEN
            CSJ(j,K)=CF(j)
          END IF
        END DO
      END DO
      WHERE (ABS(CSA) > ABS(CSB))
        CS=CSA/CF
      ELSEWHERE
        CS=CSB/CF0
      END WHERE
      DO K=0,MAXVAL(NM)
        DO j=1,anz
          IF (A0(j) >= 1.0D-60 .AND. K <= NM(j)) THEN
            CSJ(j,K)=CS(j)*CSJ(j,K)
          END IF
        END DO
      END DO
    ENDIF


    DO j=1,anz
      IF (A0(j) >= 1.0D-60) THEN
        CDJ(j,0)=(COS(Z(j))-SIN(Z(j))/Z(j))/Z(j)
      END IF
    END DO
    DO K=1,MAXVAL(NM)
      DO j=1,anz
        IF (A0(j) >= 1.0D-60 .AND. K <= NM(j)) THEN
          CDJ(j,K)=CSJ(j,K-1)-(K+1.0D0)*CSJ(j,K)/Z(j)
        END IF
      END DO
    END DO
    DO j=1,anz
      IF (A0(j) >= 1.0D-60) THEN
        CSY(j,0)=-COS(Z(j))/Z(j)
        CSY(j,1)=(CSY(j,0)-SIN(Z(j)))/Z(j)
        CDY(j,0)=(SIN(Z(j))+COS(Z(j))/Z(j))/Z(j)
        CDY(j,1)=(2.0D0*CDY(j,0)-COS(Z(j)))/Z(j)
      END IF
    END DO

    DO K=2,MAXVAL(NM)
      DO j=1,anz      
        IF (A0(j) >= 1.0D-60 .AND. K <= NM(j) .AND. ABS(CSJ(j,K-1)) > ABS(CSJ(j,K-2))) THEN
          CSY(j,K)=(CSJ(j,K)*CSY(j,K-1)-1.0D0/(Z(j)*Z(j)))/CSJ(j,K-1)
        ELSE IF (A0(j) >= 1.0D-60 .AND. K <= NM(j)) THEN
          CSY(j,K)=(CSJ(j,K)*CSY(j,K-2)-(2.0D0*K-1.0D0)/Z(j)**3)/CSJ(j,K-2)
        ENDIF
      END DO
    END DO
    DO K=2,MAXVAL(NM)
      DO j=1,anz      
        IF (A0(j) >= 1.0D-60 .AND. K <= NM(j)) THEN
          CDY(j,K)=CSY(j,K-1)-(K+1.0D0)*CSY(j,K)/Z(j)
        END IF
      END DO
    END DO

    RETURN
  END SUBROUTINE CSPHJYVEC

  INTEGER FUNCTION MSTA1(X,MP)

    !       ===================================================
    !       Purpose: Determine the starting point for backward  
    !                recurrence such that the magnitude of    
    !                Jn(x) at that point is about 10^(-MP)
    !       Input :  x     --- Argument of Jn(x)
    !                MP    --- Value of magnitude
    !       Output:  MSTA1 --- Starting point   
    !       ===================================================

    IMPLICIT NONE

    DOUBLE PRECISION :: X
    INTEGER :: MP
    DOUBLE PRECISION :: A0, F, F0, F1
    INTEGER :: N0, N1, IT, NN

    A0 = ABS(X)
    N0 = INT(1.1*A0) + 1
    F0 = ENVJ(N0,A0) - MP
    N1 = N0 + 5
    F1 = ENVJ(N1,A0) - MP
    DO IT=1,20             
      NN=N1-(N1-N0)/(1.0D0-F0/F1)                  
      F=ENVJ(NN,A0)-MP
      IF(ABS(NN-N1) < 1) exit
      N0=N1
      F0=F1
      N1=NN
      F1=F
    END DO
    MSTA1=NN

    RETURN
  END FUNCTION MSTA1

  INTEGER FUNCTION MSTA2(X,N,MP)

    !       ===================================================
    !       Purpose: Determine the starting point for backward
    !                recurrence such that all Jn(x) has MP
    !                significant digits
    !       Input :  x  --- Argument of Jn(x)
    !                n  --- Order of Jn(x)
    !                MP --- Significant digit
    !       Output:  MSTA2 --- Starting point
    !       ===================================================

    IMPLICIT NONE

    INTEGER :: N, MP
    DOUBLE PRECISION :: X

    INTEGER :: N0, N1, IT, NN
    DOUBLE PRECISION :: EJN, OBJ, A0, HMP, F, F0, F1

    A0 = ABS(X)
    HMP = 0.5D0*MP
    EJN = ENVJ(N,A0)
    IF (EJN < HMP) THEN
      OBJ = MP
      N0 = INT(1.1*A0)
    ELSE
      OBJ = HMP + EJN
      N0 = N
    ENDIF
    F0 = ENVJ(N0,A0)-OBJ
    N1 = N0 + 5
    F1 = ENVJ(N1,A0) - OBJ
    DO IT=1,20
      NN = N1-(N1-N0)/(1.0D0-F0/F1)
      F = ENVJ(NN,A0) - OBJ
      IF (ABS(NN-N1) < 1) EXIT
      N0=N1
      F0=F1
      N1=NN
      F1=F
    END DO
    MSTA2 = NN + 10

    RETURN
  END FUNCTION MSTA2

  SUBROUTINE CSPHJY(N,Z,NM,CSJ,CDJ,CSY,CDY)

    !       ==========================================================
    !       Purpose: Compute spherical Bessel functions jn(z) & yn(z)
    !                and their derivatives for a complex argument up
    !                to order N.
    !       Input :  z --- Complex argument
    !                n --- Order of jn(z) & yn(z) ( n = 0,1,2,... )
    !
    !       Output:  Vectors CSJ(0:N), CDJ(0:N),CSY(0:N),CDY(0:N):
    !                CSJ(n) --- jn(z)
    !                CDJ(n) --- jn'(z)
    !                CSY(n) --- yn(z)
    !                CDY(n) --- yn'(z)
    !                NM --- Highest order computed
    !                --> Higher orders are virtually 0.0 and need not
    !                be computed!!!
    !       Routines called:
    !                MSTA1 and MSTA2 for computing the starting
    !                point for backward recurrence
    !
    !       From: "Computation of Special Functions",  Shanjie Zhang, Jianming Jin, 1996, Wiley,
    !              ISBN: 0-471-11963-6
    !
    !       ==========================================================

    IMPLICIT NONE

    INTEGER :: N, NM
    COMPLEX(kind=kinddp) :: CSJ(0:N),CDJ(0:N),CSY(0:N),CDY(0:N), Z

    COMPLEX(kind=kinddp) :: CSA, CSB, CF0, CF1, CF, CS

    INTEGER :: K, M

    DOUBLE PRECISION :: A0

    A0=ABS(Z)
    NM=N
    IF (A0 < 1.0D-60) THEN
      DO  K=0,N
        CSJ(K)=0.0D0
        CDJ(K)=0.0D0
        CSY(K)=-1.0D+300
        CDY(K)=1.0D+300
      END DO
      CSJ(0)=(1.0D0,0.0D0)
      CDJ(1)=(0.333333333333333D0,0.0D0)
      RETURN
    ENDIF

    CSJ(0)=SIN(Z)/Z
    CSJ(1)=(CSJ(0)-COS(Z))/Z
    IF (N >= 2) THEN
      CSA=CSJ(0)
      CSB=CSJ(1)
      M=MSTA1(A0,200)
      IF (M < N) THEN
        NM=M
      ELSE
        M=MSTA2(A0,N,15)
      ENDIF
      CF0=0.0D0
      CF1=1.0D-100
      DO K=M,0,-1
        CF=(2.0D0*K+3.0D0)*CF1/Z-CF0
        IF (K <= NM) CSJ(K)=CF
        CF0=CF1
        CF1=CF
      END DO
      IF (ABS(CSA) > ABS(CSB)) THEN
        CS=CSA/CF
      ELSE 
        CS=CSB/CF0
      END IF
      DO K=0,NM
        CSJ(K)=CS*CSJ(K)
      END DO
    ENDIF
    CDJ(0)=(COS(Z)-SIN(Z)/Z)/Z
    DO K=1,NM
      CDJ(K)=CSJ(K-1)-(K+1.0D0)*CSJ(K)/Z
    END DO
    CSY(0)=-COS(Z)/Z
    CSY(1)=(CSY(0)-SIN(Z))/Z
    CDY(0)=(SIN(Z)+COS(Z)/Z)/Z
    CDY(1)=(2.0D0*CDY(0)-COS(Z))/Z
    DO K=2,NM
      IF (ABS(CSJ(K-1)) > ABS(CSJ(K-2))) THEN
        CSY(K)=(CSJ(K)*CSY(K-1)-1.0D0/(Z*Z))/CSJ(K-1)
      ELSE
        CSY(K)=(CSJ(K)*CSY(K-2)-(2.0D0*K-1.0D0)/Z**3)/CSJ(K-2)
      ENDIF
    END DO
    DO K=2,NM
      CDY(K)=CSY(K-1)-(K+1.0D0)*CSY(K)/Z
    END DO

    RETURN
  END SUBROUTINE CSPHJY

END MODULE radar_sphbessel_vec


MODULE gamma_functions_vec

  IMPLICIT NONE

!  INTEGER, PARAMETER                     :: nlookup   = 2000    ! Internal number of bins (low res part)
!  INTEGER, PARAMETER                     :: nlookuphr = 10000   ! Internal number of bins (high res part)

  ! dummy of internal number of bins (high res part) in case the high resolution part is not really needed:
!  INTEGER, PARAMETER                     :: nlookuphr_dummy = 10 

  ! Type to hold the lookup table for the incomplete gamma functions.
  ! The table is divided into a low resolution part, which spans the
  ! whole range of x-values up to the 99.5 % x-value, and a high resolution part for the
  ! smallest 1 % of these x-values, where the incomplete gamma function may increase
  ! very rapidly and nonlinearily, depending on paramter a.
  ! For some applications (e.g., Newtons Method in subroutine 
  ! graupel_hail_conv_wetgrowth_Dg_gamlook() ), this rapid change requires a much higher
  ! accuracy of the table lookup as compared to be achievable with the low resolution table.
  TYPE gamlookuptable
    ! Number of bins in the tables:
    INTEGER                              :: n           ! Internal number of bins (low res part)
    INTEGER                              :: nhr         ! Internal number of bins (high res part)
    DOUBLE PRECISION                     :: a           ! a-parameter
    DOUBLE PRECISION, DIMENSION(:), POINTER :: x        ! vector of x-parameters (limit of integration) - 
                                                        ! always starts at 0 and has equidistant dx (low resolution part)
    DOUBLE PRECISION, DIMENSION(:), POINTER :: xhr      ! vector of x-parameters (limit of integration) - 
                                                        ! always starts at 0 and has equidistant dxhr (high resolution part) 
    DOUBLE PRECISION                     :: dx          ! dx   (low resolution part)
    DOUBLE PRECISION                     :: dxhr        ! dxhr (high resolution part) 
    DOUBLE PRECISION                     :: odx         ! one over dx 
    DOUBLE PRECISION                     :: odxhr       ! one over dxhr 
    DOUBLE PRECISION, DIMENSION(:), POINTER :: igf      ! value of the inc. gamma function at (a,x) (low res)
    DOUBLE PRECISION, DIMENSION(:), POINTER :: igfhr    ! value of the inc. gamma function at (a,x) (high res)
  END TYPE gamlookuptable

  PRIVATE :: gcf, gser

CONTAINS

  FUNCTION gammlnvec(x,anz) RESULT (y)
    !*******************************************************************************
    !                                                                              *
    !       Log(Gammafunktion) aus Numerical Recipes (F77)                         *
    !       (original)                                                             *
    !*******************************************************************************
    IMPLICIT NONE
    
    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in) :: x(anz)
    DOUBLE PRECISION             :: y(anz)
    DOUBLE PRECISION, SAVE :: cof(6), stp
    DOUBLE PRECISION :: xx(anz),tmp(anz),ser(anz)
    INTEGER :: j, k
    DATA cof /76.18009172947146d0,-86.50532032941677d0, &
         24.01409824083091d0,-1.231739572450155d0,.1208650973866179d-2, &
         -.5395239384953d-5/
    DATA stp /2.5066282746310005d0/

    xx  = x
    tmp = xx + 5.5d0
    tmp = (xx + 0.5d0) * LOG(tmp) - tmp
    ser = 1.000000000190015d0
    DO j = 1,6
      DO k=1, anz
        xx(k)  = xx(k)  + 1.0d0
        ser(k) = ser(k) + cof(j) / xx(k)
      ENDDO
    ENDDO
    y = tmp + LOG(stp*ser/x)
    RETURN
  END FUNCTION gammlnvec

  DOUBLE PRECISION FUNCTION gammln(x)
    !*******************************************************************************
    !                                                                              *
    !       Log(Gammafunktion) aus Numerical Recipes (F77)                         *
    !       (original)                                                             *
    !*******************************************************************************
    IMPLICIT NONE
      
    DOUBLE PRECISION, INTENT(in) :: x
    DOUBLE PRECISION, SAVE :: cof(6), stp
    DOUBLE PRECISION :: xx,tmp,ser
    INTEGER :: j
    DATA cof /76.18009172947146d0,-86.50532032941677d0, &
         24.01409824083091d0,-1.231739572450155d0,.1208650973866179d-2, &
         -.5395239384953d-5/
    DATA stp /2.5066282746310005d0/

    xx  = x
    tmp = xx + 5.5d0
    tmp = (xx + 0.5d0) * LOG(tmp) - tmp
    ser = 1.000000000190015d0
    DO j = 1,6
       xx  = xx  + 1.0d0
       ser = ser + cof(j) / xx
    ENDDO
    gammln = tmp + LOG(stp*ser/x)
    RETURN
  END FUNCTION gammln

  FUNCTION gfctvec(x, anz) RESULT (y)
    !*******************************************************************************
    !                                                                              *
    !       Gammafunktion aus Numerical Recipes (F77)                              *
    !       (original)                                                             *
    !*******************************************************************************
    IMPLICIT NONE
    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)  :: x(anz)
    DOUBLE PRECISION              :: y(anz)

    y = EXP(gammlnvec(x,anz))
    RETURN
  END FUNCTION gfctvec

  DOUBLE PRECISION FUNCTION gfct(x)
    !*******************************************************************************
    !                                                                              *
    !       Gammafunktion aus Numerical Recipes (F77)                              *
    !       (original)                                                             *
    !*******************************************************************************
    IMPLICIT NONE
    DOUBLE PRECISION, INTENT(in) :: x

    gfct = EXP(gammln(x))
    RETURN
  END FUNCTION gfct

  FUNCTION gfct2vec(x, anz) RESULT(y)
    !*******************************************************************************
    !                                                                              *
    !       Gammafunktion aus Numerical Recipes (F77)                              *
    !       (etwas umformuliert, aber dieselben Ergebnisse wie obige Originalfunktion)
    !*******************************************************************************
    IMPLICIT NONE
      
    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in) :: x(anz)
    DOUBLE PRECISION             :: y(anz)
    DOUBLE PRECISION, SAVE :: cof(6), stp, half, one, fpf
    DOUBLE PRECISION :: xx(anz),tmp(anz),ser(anz)
    INTEGER j, k
      
    DATA cof,stp/76.18009173d0,-86.50532033d0,24.01409822d0,  &
          &     -1.231739516d0,.120858003d-2,-.536382d-5,2.50662827465d0/
    DATA half,one,fpf/0.5d0,1.0d0,5.5d0/
      
    xx  = x  - one
    tmp = xx + fpf
    tmp = (xx + half) * LOG(tmp) - tmp
    ser = one
    DO j = 1,6
      DO k=1,anz
        xx(k)  = xx(k)  + one
        ser(k) = ser(k) + cof(j) / xx(k)
      ENDDO
    ENDDO
    y = EXP(tmp + LOG(stp*ser))

    RETURN
  END FUNCTION gfct2vec
  
  DOUBLE PRECISION FUNCTION gfct2(x)
    !*******************************************************************************
    !                                                                              *
    !       Gammafunktion aus Numerical Recipes (F77)                              *
    !       (etwas umformuliert, aber dieselben Ergebnisse wie obige Originalfunktion)
    !*******************************************************************************
    IMPLICIT NONE
      
    DOUBLE PRECISION, INTENT(in) :: x
    DOUBLE PRECISION, SAVE :: cof(6), stp, half, one, fpf
    DOUBLE PRECISION :: xx,tmp,ser,gamma
    INTEGER j
      
    DATA cof,stp/76.18009173d0,-86.50532033d0,24.01409822d0,  &
          &     -1.231739516d0,.120858003d-2,-.536382d-5,2.50662827465d0/
    DATA half,one,fpf/0.5d0,1.0d0,5.5d0/
      
    xx  = x  - one
    tmp = xx + fpf
    tmp = (xx + half) * LOG(tmp) - tmp
    ser = one
    DO j = 1,6
       xx  = xx  + one
       ser = ser + cof(j) / xx
    ENDDO
    gamma = tmp + LOG(stp*ser)
    gamma = EXP(gamma)

    gfct2 = gamma
    RETURN
  END FUNCTION gfct2

  !*******************************************************************************
  !                                                                              *
  !       Unvollstaendige Gammafunktion aus Numerical Recipes (F77)              *
  !                                                                              *
  !*******************************************************************************

  !*******************************************************************************
  ! 1) diverse Hilfsfunktionen:

  SUBROUTINE gcf(gammcf,a,x,gln)

    IMPLICIT NONE

    INTEGER, parameter :: ITMAX = 100
    DOUBLE PRECISION, PARAMETER :: EPS = 3.d-7, FPMIN = 1.d-30
    DOUBLE PRECISION, INTENT(in) :: a, x
    DOUBLE PRECISION, INTENT(out) :: gammcf, gln

    INTEGER :: i
    DOUBLE PRECISION :: an,b,c,d,del,h

    gln=gammln(a)
    b=x+1.-a
    c=1./FPMIN
    d=1./b
    h=d
    DO i=1,ITMAX
      an=-i*(i-a)
      b=b+2.0d0
      d=an*d+b
      if (abs(d).lt.FPMIN) d=FPMIN
      c=b+an/c
      if (abs(c).lt.FPMIN) c=FPMIN
      d=1./d
      del=d*c
      h=h*del
      IF (ABS(del-1.).LT.EPS) EXIT
    END DO

    IF (ABS(del-1.).GE.EPS) THEN
      WRITE (*,*) 'ERROR in GCF: a too large, ITMAX too small (MODULE gamma_functions, radar_mielib.f90)'
      gammcf = 0.0d0
      RETURN
    END IF

    gammcf=exp(-x+a*log(x)-gln)*h

    return
  END SUBROUTINE gcf

  SUBROUTINE gser(gamser,a,x,gln)

    IMPLICIT NONE

    INTEGER, PARAMETER :: ITMAX = 100
    DOUBLE PRECISION, PARAMETER :: EPS=3.d-7
    DOUBLE PRECISION, INTENT(in) :: a, x
    DOUBLE PRECISION, INTENT(out) :: gamser, gln
      
    INTEGER :: n
    DOUBLE PRECISION :: ap,del,sum
    
    gln=gammln(a)
    IF (x.LE.0.) THEN
      IF (x.LT.0.) THEN
        WRITE (*,*) 'ERROR in GSER: x < 0 (MODULE gamma_functions, radar_mielib.f90)'
      END IF
      gamser=0.0d0
      RETURN
    ENDIF

    ap=a
    sum=1./a
    del=sum
    DO n=1,ITMAX
      ap=ap+1.
      del=del*x/ap
      sum=sum+del
      IF (ABS(del).LT.ABS(sum)*EPS) EXIT
    END DO

    IF (ABS(del).GE.ABS(sum)*EPS) THEN
      WRITE (*,*) 'ERROR in GSER: a too large, ITMAX too small (MODULE gamma_functions, radar_mielib.f90)'
      gamser = 0.0d0
      RETURN
    END IF

    gamser = sum*EXP(-x+a*LOG(x)-gln)

    RETURN
  END SUBROUTINE gser

  DOUBLE PRECISION FUNCTION gammp(a,x,gln)
  
    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: a, x
    DOUBLE PRECISION, INTENT(out) :: gln
   
    DOUBLE PRECISION :: gammcf, gamser
    
    IF (x.LT.0.0d0 .OR. a .LE. 0.0d0) THEN
      WRITE(*,*) 'ERROR in GAMMP: bad arguments (MODULE gamma_functions, radar_mielib.f90)'
      gammp = 0.0d0
      RETURN
    END IF

    IF (x .LT. a+1.)THEN
      CALL gser(gamser,a,x,gln)
      gammp = gamser
    ELSE
      CALL gcf(gammcf,a,x,gln)
      gammp = 1.0d0 - gammcf
    ENDIF
    RETURN
  END FUNCTION gammp

  DOUBLE PRECISION FUNCTION gammq(a,x,gln)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: a, x
    DOUBLE PRECISION, INTENT(out) :: gln
    DOUBLE PRECISION :: gammcf, gamser

    IF (x.LT.0.0d0 .OR. a .LE. 0.0d0) THEN
      WRITE(*,*) 'ERROR in GAMMQ: bad arguments (MODULE gamma_functions, radar_mielib.f90)'
      gammq = 0.0d0
      RETURN
    END IF

    IF (x.LT.a+1.) THEN
      CALL gser(gamser,a,x,gln)
      gammq = 1.0d0 - gamser
    ELSE
      CALL gcf(gammcf,a,x,gln)
      gammq = gammcf
    ENDIF
    RETURN
  END FUNCTION gammq

  ! Ende diverse Hilfsfunktionen
  !*******************************************************************************

  !*******************************************************************************
  !
  ! Eigentliche obere unvollstaendige Gamma-Funktion, hier direkt
  ! das Integral
  !              int(x)(oo) exp(-t) t^(a-1) dt
  !
  !*******************************************************************************

  DOUBLE PRECISION FUNCTION incgfct_upper(a,x)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: a, x
    DOUBLE PRECISION :: gam, gln
    
    gam = gammq(a,x,gln)
    incgfct_upper = EXP(gln) * gam

  END FUNCTION incgfct_upper

  !*******************************************************************************
  !
  ! Eigentliche untere unvollstaendige Gamma-Funktion, hier direkt
  ! das Integral
  !              int(0)(x) exp(-t) t^(a-1) dt
  !
  !*******************************************************************************

  DOUBLE PRECISION FUNCTION incgfct_lower(a,x)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: a, x
    DOUBLE PRECISION :: gam, gln
    
    gam = gammp(a,x,gln)
    incgfct_lower = EXP(gln) * gam

  END FUNCTION incgfct_lower

  !*******************************************************************************
  !
  ! Eigentliche unvollstaendige Gamma-Funktion, hier direkt
  ! das Integral
  !              int(x1)(x2) exp(-t) t^(a-1) dt
  !
  !*******************************************************************************

  DOUBLE PRECISION FUNCTION incgfct(a,x1,x2)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: a, x1, x2
    
    incgfct = incgfct_lower(a,x2) - incgfct_lower(a,x1)

  END FUNCTION incgfct

  !*******************************************************************************
  !
  ! Create Lookup-table vectors for the lower incomplete gamma function,
  !              int(0)(x) exp(-t) t^(a-1) dt
  ! as function of x at constant a. 
  ! The table runs from x=0 to the 99.5 % - value of the normalized 
  ! incomplete gamma function. This 99.5 % - value has been fitted
  ! with high accuracy as function of a in the range a in [0;20], but can
  ! safely be applied also to higher values of a. (Fit created with the
  ! matlab-program "gamma_unvoll_lower_lookup.m" by Ulrich Blahak, 2008/11/13).
  !
  ! The last value in the table corresponds to x = infinity, so that
  ! during the reconstruction of incgfct-values from the table,
  ! the x-value can safely be truncated at the maximum table x-value.
  !
  !*******************************************************************************

  SUBROUTINE incgfct_lower_lookupcreate(a,ltable,nl,nlhr)
    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: a  ! value of a
    TYPE(gamlookuptable), intent(inout) :: ltable
    INTEGER, INTENT(in) :: nl, nlhr

    INTEGER :: i, err

    DOUBLE PRECISION, PARAMETER ::   &
           c1 =  36.629433904824623d0, &
           c2 = -0.119475603955226d0,  &
           c3 =  0.339332937820052d0,  &
           c4 =  1.156369000458310d0

    ! Store parameters in the structure ltable:
    ltable%a = a
    ltable%n = nl
    ltable%nhr = nlhr

    ! Allocate Memory for the table vectors:
    NULLIFY(ltable%x)
    NULLIFY(ltable%xhr)
    NULLIFY(ltable%igf)
    NULLIFY(ltable%igfhr)

    ALLOCATE(ltable%x(nl), STAT=err)
    IF (err /= 0) THEN
      WRITE (*,*) 'INCGFCT_LOWER_LOOKUPCREATE: Allocation error x'
      STOP
    END IF
    ALLOCATE(ltable%xhr(nlhr), STAT=err)
    IF (err /= 0) THEN
      WRITE (*,*) 'INCGFCT_LOWER_LOOKUPCREATE: Allocation error xhr'
      STOP
    END IF
    ALLOCATE(ltable%igf(nl), STAT=err)
    IF (err /= 0) THEN
      WRITE (*,*) 'INCGFCT_LOWER_LOOKUPCREATE: Allocation error igf'
      STOP
    END IF
    ALLOCATE(ltable%igfhr(nlhr), STAT=err)
    IF (err /= 0) THEN
      WRITE (*,*) 'INCGFCT_LOWER_LOOKUPCREATE: Allocation error igfhr'
      STOP
    END IF

    !==================================================================
    ! low resolution part of the table:
    !==================================================================

    ! maximum x-value of the lookup table (99.5-%-value):
    ltable%x(ltable%n-1) = c1 * ( 1.0d0 - EXP(c2*a**c3) ) + c4*a

    ! create lookup table vectors:
    ltable%dx = ltable%x(ltable%n-1) / (ltable%n-2.0d0)
    ltable%odx = 1.0d0 / ltable%dx
!!! This loop does not vectorize because of incgfct_lower():
    DO i = 1, ltable%n - 1
      ltable%x(i) = (i-1) * ltable%dx
      ltable%igf(i) = incgfct_lower(a,ltable%x(i))
    END DO
       
    ! The last value is for x = infinity:
    ltable%x(ltable%n) = (ltable%n-1) * ltable%dx
    ltable%igf(ltable%n) = gfct2(a)
    
    !==================================================================
    ! high resolution part of the table (lowest 1 % of the X-values):
    !==================================================================

    ! create lookup table vectors:
    ltable%dxhr = ltable%x(NINT(0.01*(ltable%n-1))) / (ltable%nhr-1.0d0)
    ltable%odxhr = 1.0d0 / ltable%dxhr
!!! This loop does not vectorize because of incgfct_lower():
    DO i = 1, ltable%nhr
      ltable%xhr(i) = (i-1) * ltable%dxhr
      ltable%igfhr(i) = incgfct_lower(a,ltable%xhr(i))
    END DO

    RETURN
  END SUBROUTINE incgfct_lower_lookupcreate

  !*******************************************************************************
  !
  ! Retrieve values from a lookup table of the lower incomplete gamma function,
  ! as function of x at a constant a, for which the lookup table has been 
  ! created. Before the below function can be applied, the table itself
  ! has to be created by function incgfct_lower_lookupcreate(a,ltable,nl,nlhr) above !!!
  !
  ! The last value in the table has to correspond to x = infinity, so that
  ! during the reconstruction of incgfct-values from the table,
  ! the x-value can safely be truncated at the maximum table x-value:
  !
  ! ltable%igf( ltable%x(ltable%n),...) = gfct(a)
  !
  ! Profiling with ifort on a Linux-PC shows, that table lookup for the
  ! incompl. gamma-Funktion is faster by a factor of about 15 compared
  ! to the original function without optimization (-O0). Using optimization
  ! could change this ratio (we encoutered up to 300 depending on function inlining).
  !
  ! Concerning the accuracy, comparisons show that the results of table lookup
  ! are accurate to within better than 0.1 % or even much less, except for
  ! very small values of X, for which the absolute values are however very
  ! close to 0. For X -> infinity (X > 99.5 % - value), accuracy may be 
  ! somewhat reduced up to about 0.5 % ,
  ! because the table is truncated at the 99.5 % value (second-last value)
  ! and the last value is set to the ordinary gamma function.
  !
  !*******************************************************************************

  DOUBLE PRECISION FUNCTION incgfct_lower_lookup(x, ltable)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: x  ! value of x for table lookup
    TYPE(gamlookuptable), INTENT(in) :: ltable

    INTEGER :: iu, io
    DOUBLE PRECISION :: xt

    ! Trunkcate x to the range of the table:
    xt = MAX(MIN(x, ltable%x(ltable%n)), 0.0d0)

    ! calculate indices of the neighbouring regular x-values
    ! in the table:
    iu = MIN(FLOOR(xt * ltable%odx) + 1, ltable%n-1)
    io = iu + 1

    ! interpolate linearily:
    incgfct_lower_lookup = ltable%igf(iu) + &
         (ltable%igf(io) - ltable%igf(iu)) * ltable%odx * (xt-ltable%x(iu))

    RETURN
  END FUNCTION incgfct_lower_lookup


  !*******************************************************************************
  !
  ! Retrieve values of the upper incomplete gamma function 
  ! from a lookup table of the lower incomplete gamma function,
  ! as function of x at a constant a, for which the lookup table has been 
  ! created. Before the below function can be applied, the table itself
  ! has to be created by function incgfct_lower_lookupcreate(a,ltable,nl,nlhr) above !!!
  !
  ! The last value in the table has to correspond to x = infinity 
  ! (the ordinary gamma function of a), so that
  ! during the reconstruction of incgfct-values from the table,
  ! the x-value can safely be truncated at the maximum table x-value:
  !
  ! ltable%igf( ltable%x(ltable%n),...) = gfct(a)
  !
  !*******************************************************************************

  DOUBLE PRECISION FUNCTION incgfct_upper_lookup(x, ltable)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: x  ! value of x for table lookup
    TYPE(gamlookuptable), INTENT(in) :: ltable

    INTEGER :: iu, io
    DOUBLE PRECISION :: xt

    ! Trunkcate x to the range of the table:
    xt = MAX(MIN(x, ltable%x(ltable%n)), 0.0d0)

    ! calculate indices of the neighbouring regular x-values
    ! in the table:
    iu = MIN(FLOOR(xt * ltable%odx) + 1, ltable%n-1)
    io = iu + 1

    ! interpolate lower inc. gamma function linearily and subtract from 
    ! the ordinary gamma function to get the upper
    ! incomplete gamma function:
    incgfct_upper_lookup = ltable%igf(ltable%n) - ltable%igf(iu) -  &
         (ltable%igf(io) - ltable%igf(iu)) * ltable%odx * (xt-ltable%x(iu))

    ! Aufgrund von Rundungsfehlern (Differenz von 2 fast gleichen Zahlen) kann es beim table lookup passieren,
    ! dass incgfct_upper_lookup(x, ltable) kleiner 0 wird, wenn eigentlich nahezu 0.0 herauskommen muesste.
    ! Dies kommt vor allem dann vor, wenn x sehr gross ist.
    ! Deswegen Begrenzung:

    incgfct_upper_lookup = MAX(incgfct_upper_lookup, 0.0d0)

    RETURN
  END FUNCTION incgfct_upper_lookup


END MODULE gamma_functions_vec




!*********************************************************************************************

  !----------------------------------------------------------------------------------------!
  !  Routinen zur Berechnung von Q_back, Q_ext und Q_sca von verschiedenen
  !  einzelnen Hydrometeoren nach der Mie-Theorie und alle dazu benoetigten
  !  Hilfsmittel wie: komplexer Brechungsindex von Mischmaterialien nach verschiedenen
  !  Ansaetzen, komplexer Brechungsindex von Eis und Wasser nach verschiedenen
  !  Parametrisierungen.
  !----------------------------------------------------------------------------------------!


MODULE radar_mielib_vec

  USE radar_sphbessel_vec, ONLY: CSPHJYVEC, CSPHJY 

  USE data_constants, ONLY :   &
       rho_w,       &  ! density of liquid water
       rho_ice         ! density of ice          (kg/m^3)

  IMPLICIT NONE

  INTEGER, PARAMETER, PRIVATE :: kinddp = KIND(1.0d0)
  DOUBLE PRECISION,     PARAMETER, PRIVATE :: pi = 3.141592653589793238462643383d0
!  COMPLEX(kind=kinddp), PARAMETER, PRIVATE :: m_air = CMPLX(1.0d0,0.0d0,kind=kinddp)
  COMPLEX(kind=kinddp), PARAMETER, PRIVATE :: m_air = (1.0d0,0.0d0)

CONTAINS

  !----------------------------------------------------------------------------------------!
  ! 1) Routinen zur Berechnung des Komplexen Brechungsindex fuer Wasser und Eis
  !    als Funktion von Wellenlaenge und Temperatur nach verschiedenen 
  !    Parametrisierungen aus der Literatur.
  !----------------------------------------------------------------------------------------!

  !----------------------------------------------------------------------------------------!
  ! 1a) Komplexer Brechungsindex fr Flssigwasser nach Ray (1972)
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_water_ray_vec(lambda,T,anz)
    IMPLICIT NONE

    ! Komplexer Brechungsindex von Wasser als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.001,1.0] m; T aus [-10.0,30.0] Grad C
    ! nach Ray (1972)

    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_water_ray_vec(anz)
    DOUBLE PRECISION     :: T(anz), lambda
    DOUBLE PRECISION     :: epsinf,epss,alpha,lambdas,nenner,epsr,epsi

    COMPLEX(kind=kinddp), PARAMETER :: i  = (0d0,1d0)
    INTEGER :: k

    DO k=1,anz
      epsinf  = 5.27137d0 + 0.02164740d0 * T(k)**1  &
           &              - 0.00131198d0 * T(k)**2
      epss    = 78.54d+0 * (1.0 - 4.579d-3 * (T(k) - 25.0)**1  &
           &                    + 1.190d-5 * (T(k) - 25.0)**2  &
           &                    - 2.800d-8 * (T(k) - 25.0)**3)
      alpha   = -16.8129d0/(T(k)+273.16) + 0.0609265d0
      lambdas = 0.00033836d0 * EXP(2513.98d0/(T(k)+273.16)) * 1e-2
      
      nenner  = 1.d0 + 2.d0*(lambdas/lambda)**(1d0-alpha)*SIN(alpha*PI*0.5) & 
           &         +      (lambdas/lambda)**(2d0-2d0*alpha)
      
      epsr   = epsinf + ((epss-epsinf) * ((lambdas/lambda)**(1d0-alpha)    &
           &                           * SIN(alpha*PI*0.5)+1d0)) / nenner
      epsi   =          ((epss-epsinf) * ((lambdas/lambda)**(1d0-alpha)    &
           &                           * COS(alpha*PI*0.5)+0d0)) / nenner  &
           & + lambda*1.25664/1.88496
      m_complex_water_ray_vec(k) = SQRT(CMPLX(epsr,-epsi,kind=kinddp))
    END DO

    RETURN
  END FUNCTION m_complex_water_ray_vec

  FUNCTION m_complex_water_ray(lambda,T)
    IMPLICIT NONE

    ! Komplexer Brechungsindex von Wasser als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.001,1.0] m; T aus [-10.0,30.0] Grad C
    ! nach Ray (1972)

    COMPLEX(kind=kinddp) :: m_complex_water_ray
    DOUBLE PRECISION     :: T, lambda
    DOUBLE PRECISION     :: epsinf,epss,alpha,lambdas,nenner,epsr,epsi

    COMPLEX(kind=kinddp), PARAMETER :: i  = (0d0,1d0)

    epsinf  = 5.27137d0 + 0.02164740d0 * T**1  &
         &              - 0.00131198d0 * T**2
    epss    = 78.54d+0 * (1.0 - 4.579d-3 * (T - 25.0)**1  &
         &                    + 1.190d-5 * (T - 25.0)**2  &
         &                    - 2.800d-8 * (T - 25.0)**3)
    alpha   = -16.8129d0/(T+273.16) + 0.0609265d0
    lambdas = 0.00033836d0 * exp(2513.98d0/(T+273.16)) * 1e-2

    nenner  = 1.d0 + 2.d0*(lambdas/lambda)**(1d0-alpha)*sin(alpha*PI*0.5) & 
         &         +      (lambdas/lambda)**(2d0-2d0*alpha)

    epsr   = epsinf + ((epss-epsinf) * ((lambdas/lambda)**(1d0-alpha)    &
         &                           * sin(alpha*PI*0.5)+1d0)) / nenner
    epsi   =          ((epss-epsinf) * ((lambdas/lambda)**(1d0-alpha)    &
         &                           * cos(alpha*PI*0.5)+0d0)) / nenner  &
         & + lambda*1.25664/1.88496

    m_complex_water_ray = SQRT(CMPLX(epsr,-epsi,kind=kinddp))

    RETURN
  END FUNCTION m_complex_water_ray

  !----------------------------------------------------------------------------------------!
  ! 1b) Komplexer Brechungsindex fr Eis nach Ray (1972)
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_ice_ray_vec(lambda,T,anz)
    IMPLICIT NONE

    ! Komplexer Brechungsindex von Eis als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.001,10**7] m; T aus [-20.0,0.0] Grad C
    ! nach Ray (1972)

    !  ;;; Vorsicht: Erzeugt falsche Werte im Vergleich zur Abb. in Ray
    !  ;;; 1972  -- muss an den Formeln im Paper liegen!
    !  ;;; mit 1e-3 bei lambdas scheints zu stimmen ...

    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_ice_ray_vec(anz)
    DOUBLE PRECISION     :: T(anz),lambda
    DOUBLE PRECISION     :: epsinf,epss,alpha,lambdas,sigma,nenner,epsr,epsi
    INTEGER :: k

    DO k=1,anz
      epsinf  = 3.168d0
      epss    = 203.168d0 + 2.5d0*T(k) + 0.15d0*T(k)**2
      alpha   = 0.288d0 + 0.00520d0 * T(k)**1 &
           &            + 0.00023d0 * T(k)**2
      lambdas = 9.990288d-4 * EXP(13200.0d0/(T(k)+273.16)/1.9869d0) * 1e-3
      sigma   = 1.26d0 * EXP(-12500.0d0/(T(k)+273.16)/1.9869d0)
      
      nenner = 1.d0 + 2.d0*(lambdas/lambda)**(1d0-alpha)*SIN(alpha*PI*0.5)  &
           & + (lambdas/lambda)**(2d0-2d0*alpha)
      
      epsr   = epsinf + ((epss-epsinf) * (1d0+(lambdas/lambda)**(1d0-alpha) &
           &                           * SIN(alpha*PI*0.5))) / nenner
      epsi   =          ((epss-epsinf) * (lambdas/lambda)**(1d0-alpha) &
           &                           * COS(alpha*PI*0.5)) / nenner &
           & +   lambda*sigma/1.88496d9
      m_complex_ice_ray_vec(k) = SQRT(CMPLX(epsr,-epsi,kind=kinddp))
    END DO


  END FUNCTION m_complex_ice_ray_vec

  FUNCTION m_complex_ice_ray(lambda,T)
    IMPLICIT NONE

    ! Komplexer Brechungsindex von Eis als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.001,10**7] m; T aus [-20.0,0.0] Grad C
    ! nach Ray (1972)

    !  ;;; Vorsicht: Erzeugt falsche Werte im Vergleich zur Abb. in Ray
    !  ;;; 1972  -- muss an den Formeln im Paper liegen!
    !  ;;; mit 1e-3 bei lambdas scheints zu stimmen ...

    COMPLEX(kind=kinddp) :: m_complex_ice_ray
    DOUBLE PRECISION     :: T,lambda
    DOUBLE PRECISION     :: epsinf,epss,alpha,lambdas,sigma,nenner,epsr,epsi

    epsinf  = 3.168d0
    epss    = 203.168d0 + 2.5d0*T + 0.15d0*T**2
    alpha   = 0.288d0 + 0.00520d0 * T**1 &
         &            + 0.00023d0 * T**2
    lambdas = 9.990288d-4 * exp(13200.0d0/(T+273.16)/1.9869d0) * 1e-3
    sigma   = 1.26d0 * exp(-12500.0d0/(T+273.16)/1.9869d0)

    nenner = 1.d0 + 2.d0*(lambdas/lambda)**(1d0-alpha)*sin(alpha*PI*0.5)  &
         & + (lambdas/lambda)**(2d0-2d0*alpha)

    epsr   = epsinf + ((epss-epsinf) * (1d0+(lambdas/lambda)**(1d0-alpha) &
         &                           * sin(alpha*PI*0.5))) / nenner
    epsi   =          ((epss-epsinf) * (lambdas/lambda)**(1d0-alpha) &
         &                           * cos(alpha*PI*0.5)) / nenner &
         & +   lambda*sigma/1.88496d9

    m_complex_ice_ray = SQRT(CMPLX(epsr,-epsi,kind=kinddp))

  END FUNCTION m_complex_ice_ray

  !----------------------------------------------------------------------------------------!
  ! 1c) Komplexer Brechungsindex fr Flssigwasser nach Liebe (1991)
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_water_liebe_vec(lambda,T,anz)

    ! Komplexer Brechungsindex von Wasser als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.0003,0.3] m; T aus [-3.0,30.0] Grad C
    ! nach Liebe (1991) (aus den Matlab-Programmen von Prof. Maetzler, Juni 2003)

    IMPLICIT NONE

    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_water_liebe_vec(anz)
    DOUBLE PRECISION     :: T(anz),lambda
    DOUBLE PRECISION     :: c,TK,fGHz,TETA,e0,e1,f1,e2,f2
    INTEGER :: k

    c = 2.99d8 ! m/s, Lichtgeschw. im Vakuum
    fGHz = c / lambda * 1d-9

    DO k=1,anz
      TK = T(k) + 273.16 ! Kelvin
      TETA = 1.0 - 300./TK
      e0 = 77.66 - 103.3*TETA
      e1 = 0.0671*e0
      f1 = 20.2 + 146.4*TETA + 316*TETA**2
      e2 = 3.52 + 7.52*TETA
      ! version of Liebe MPM 1993 uses: e2=3.52
      f2 = 39.8*f1
      ! Liebe verwendet die Konvention mit positivem Imaginaerteil; auf
      ! negativen umstellen.
      ! Die komplexe Wurzelfunktion liefert diejenige Wurzel mit
      ! positivem Realteil (Hauptzweig), was in diesem Falle richtig ist. 
      m_complex_water_liebe_vec(k) = SQRT(CONJG(e2 + (e1-e2)/CMPLX(1.0,-fGHz/f2,kind=kinddp) &
           + (e0-e1)/CMPLX(1.0,-fGHz/f1,kind=kinddp)))
    END DO

  END FUNCTION m_complex_water_liebe_vec

  FUNCTION m_complex_water_liebe(lambda,T)

    ! Komplexer Brechungsindex von Wasser als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.0003,0.3] m; T aus [-3.0,30.0] Grad C
    ! nach Liebe (1991) (aus den Matlab-Programmen von Prof. Maetzler, Juni 2003)

    IMPLICIT NONE

    COMPLEX(kind=kinddp) :: m_complex_water_liebe
    DOUBLE PRECISION     :: T,lambda
    DOUBLE PRECISION     :: c,TK,fGHz,TETA,e0,e1,f1,e2,f2


    c = 2.99d8 ! m/s, Lichtgeschw. im Vakuum
    TK = T + 273.16 ! Kelvin
    fGHz = c / lambda * 1d-9

    TETA = 1.0 - 300./TK
    e0 = 77.66 - 103.3*TETA
    e1 = 0.0671*e0
    f1 = 20.2 + 146.4*TETA + 316*TETA**2
    e2 = 3.52 + 7.52*TETA
    ! version of Liebe MPM 1993 uses: e2=3.52
    f2 = 39.8*f1
    ! Liebe verwendet die Konvention mit positivem Imaginaerteil; auf
    ! negativen umstellen.
    ! Die komplexe Wurzelfunktion liefert diejenige Wurzel mit
    ! positivem Realteil (Hauptzweig), was in diesem Falle richtig ist. 
    m_complex_water_liebe = SQRT(CONJG(e2 + (e1-e2)/CMPLX(1.0,-fGHz/f2,kind=kinddp) + &
         (e0-e1)/CMPLX(1.0,-fGHz/f1,kind=kinddp)))

  END FUNCTION m_complex_water_liebe

  !----------------------------------------------------------------------------------------!
  ! 1d) Komplexer Brechungsindex fr Eis nach Maetzler (1998)
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_ice_maetzler_vec(lambda,T,anz)
    IMPLICIT NONE

    ! Komplexer Brechungsindex von Eis als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.0001,30] m; T aus [-250.0,0.0] Grad C
    ! nach Maetzler (1998)

    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_ice_maetzler_vec(anz)
    DOUBLE PRECISION     :: T(anz),lambda
    DOUBLE PRECISION     :: f,c,TK,B1,B2,b,deltabeta,betam,beta,theta,alfa
    INTEGER :: k

!!! Originalkommentar aus der Matlab-Routine von Prof. Maetzler:
!!! ============================================================
    ! Function for calculating the relative permittivity of pure ice in the
    ! microwave region, according to C. Maetzler, "Microwave properties of ice
    ! and snow", in B. Schmitt et al. (eds.) Solar System Ices, Astrophys.
    ! and Space Sci. Library, Vol. 227, Kluwer Academic Publishers,
    ! Dordrecht, pp. 241-257 (1998). Input:
    ! TK = temperature (K), range 20 to 273.15
    ! f = frequency in GHz, range 0.01 to 3000

    c = 2.99d8 ! m/s, Lichtgeschw. im Vakuum
    f = c / lambda * 1d-9

    DO k= 1,anz
      TK = T(k) + 273.16 ! Kelvin
      B1 = 0.0207
      B2 = 1.16d-11
      b = 335.0d0
      deltabeta = EXP(-10.02 + 0.0364*(TK-273.16))
      betam = (B1/TK) * ( EXP(b/TK) / ((EXP(b/TK)-1)**2) ) + B2*f**2
      beta = betam + deltabeta
      theta = 300. / TK - 1.
      alfa = (0.00504d0 + 0.0062d0*theta) * EXP(-22.1d0*theta)
      m_complex_ice_maetzler_vec(k) = 3.1884 + 9.1e-4*(TK-273.16)
      m_complex_ice_maetzler_vec(k) = m_complex_ice_maetzler_vec(k) + &
           CMPLX(0.0d0, (alfa/f + beta*f), kind=kinddp)
      ! Maetzler verwendet die Konvention mit positivem Imaginaerteil; auf
      ! negativen umstellen.
      ! Die komplexe Wurzelfunktion liefert diejenige Wurzel mit
      ! positivem Realteil (Hauptzweig), was in diesem Falle richtig ist. 
      m_complex_ice_maetzler_vec(k) = SQRT(CONJG(m_complex_ice_maetzler_vec(k)))
    END DO

  END FUNCTION m_complex_ice_maetzler_vec

  FUNCTION m_complex_ice_maetzler(lambda,T)
    IMPLICIT NONE

    ! Komplexer Brechungsindex von Eis als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [0.0001,30] m; T aus [-250.0,0.0] Grad C
    ! nach Maetzler (1998)

    COMPLEX(kind=kinddp) :: m_complex_ice_maetzler
    DOUBLE PRECISION     :: T,lambda
    DOUBLE PRECISION     :: f,c,TK,B1,B2,b,deltabeta,betam,beta,theta,alfa


!!! Originalkommentar aus der Matlab-Routine von Prof. Maetzler:
!!! ============================================================
    ! Function for calculating the relative permittivity of pure ice in the
    ! microwave region, according to C. Maetzler, "Microwave properties of ice
    ! and snow", in B. Schmitt et al. (eds.) Solar System Ices, Astrophys.
    ! and Space Sci. Library, Vol. 227, Kluwer Academic Publishers,
    ! Dordrecht, pp. 241-257 (1998). Input:
    ! TK = temperature (K), range 20 to 273.15
    ! f = frequency in GHz, range 0.01 to 3000

    c = 2.99d8 ! m/s, Lichtgeschw. im Vakuum
    TK = T + 273.16 ! Kelvin
    f = c / lambda * 1d-9

    B1 = 0.0207
    B2 = 1.16d-11
    b = 335.0d0
    deltabeta = EXP(-10.02 + 0.0364*(TK-273.16))
    betam = (B1/TK) * ( EXP(b/TK) / ((EXP(b/TK)-1)**2) ) + B2*f**2
    beta = betam + deltabeta
    theta = 300. / TK - 1.
    alfa = (0.00504d0 + 0.0062d0*theta) * EXP(-22.1d0*theta)
    m_complex_ice_maetzler = 3.1884 + 9.1e-4*(TK-273.16)
    m_complex_ice_maetzler = m_complex_ice_maetzler + CMPLX(0.0d0, (alfa/f + beta*f), kind=kinddp)
    ! Maetzler verwendet die Konvention mit positivem Imaginaerteil; auf
    ! negativen umstellen.
    ! Die komplexe Wurzelfunktion liefert diejenige Wurzel mit
    ! positivem Realteil (Hauptzweig), was in diesem Falle richtig ist. 
    m_complex_ice_maetzler = SQRT(CONJG(m_complex_ice_maetzler))

  END FUNCTION m_complex_ice_maetzler

  !----------------------------------------------------------------------------------------!
  ! 1e) Komplexer Brechungsindex fr Eis nach Warren (1983)
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_ice_warren_vec(lambda,T,anz)

    ! Komplexer Brechungsindex von Eis als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [45*1e-9,8.6] m; T aus [-60.0,0.0] Grad C
    ! nach Warren (1983) - seine Fortran-routine wird benutzt
    ! Brechungsindex m_i wird in der Konvention mit negativem 
    ! Imaginaerteil zurueckgegeben.

    IMPLICIT NONE

    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_ice_warren_vec(anz)
    DOUBLE PRECISION     :: T(anz),lambda
    DOUBLE PRECISION     :: m_r(anz), m_c(anz), absind(anz), abscof(anz)

    CALL REFICEVEC(1,lambda*1000.0d0,T+273.16,m_r,m_c,absind,abscof,anz)

    m_complex_ice_warren_vec = CMPLX(m_r,-m_c,kind=kinddp)

  CONTAINS

    SUBROUTINE REFICEVEC(IUNIT,XLAM,T,RN,CN,ABSIND,ABSCOF,anz)

      IMPLICIT NONE
      ! Arguments:
      INTEGER :: anz
      INTEGER :: IUNIT
      DOUBLE PRECISION :: ABSCOF(anz),ABSIND(anz),CN(anz),XLAM,RN(anz),T(anz)
      ! Parameters:
      INTEGER :: I,LT1(anz),LT2(anz),NWL,NWLT
      PARAMETER(NWL=468,NWLT=62)
      ! Local variables:
      DOUBLE PRECISION :: &
           ALAM,CUTICE,T1,T2,TK(anz),WLMAX,WLMIN, &
           X,X1,X2,Y,Y1,Y2,YLO,YHI

      DOUBLE PRECISION :: &
           TABIM(NWL),TABIMT(NWLT,4),TABRE(NWL),TABRET(NWLT,4),TEMREF(4),&
           WL(NWL),WLT(NWLT)

      INTEGER :: k, I1(1)
      INTEGER :: Ihlp(NWL), IhlpT(NWLT)

      !     DEFINES WAVELENGTH DEPENDENT COMPLEX INDEX OF REFRACTION FOR ICE.
      !     ALLOWABLE WAVELENGTH RANGE EXTENDS FROM 0.045 MICRONS TO 8.6 METER
      !     TEMPERATURE DEPENDENCE ONLY CONSIDERED BEYOND 167 MICRONS.

      !     INTERPOLATION IS DONE     RN  VS. LOG(XLAM)
      !                               RN  VS.        T
      !                           LOG(CN) VS. LOG(XLAM)
      !                           LOG(CN) VS.        T

      !     STEPHEN G. WARREN - 1983
      !     DEPT. OF ATMOSPHERIC SCIENCES
      !     UNIVERSITY OF WASHINGTON
      !     SEATTLE, WA  98195

      !     BASED ON

      !        WARREN,S.G.,1984.
      !        OPTICAL CONSTANTS OF ICE FROM THE ULTRAVIOLET TO THE MICROWAVE.
      !        APPLIED OPTICS,23,1206-1225

      !     INPUT PARAMETERS

      !     IUNIT = 0 FOR WAVELENGTH SPECIFIED IN MICRONS
      !           = 1 FOR WAVELENGTH SPECIFIED IN MILLIMETERS
      !           = 2 FOR WAVELENGTH SPECIFIED IN CENTIMETERS
      !           = 3 FOR WAVELENGTH SPECIFIED IN INVERSE CENTIMETERS ( WAVE N
      !     XLAM = WAVELENGTH ( MICRONS OR MM OR CM OR CM**-1 )
      !     T = TEMPERATURE ( DEGREES KELVIN )

      !     OUTPUT PARAMETERS

      !     RN = REAL PORTION ( SCATTERING )
      !     CN = COMPLEX PORTION ( ABSORPTION )
      !     ABSIND = ABSORPTIVE INDEX ( CN/RN )
      !     ABSCOF = ABORPTION COEFFICIENT ( 4*PI*CN/XLAM )

      !      DIMENSION WL(NWL),WLT(NWLT)
      !      DIMENSION TABRE(NWL),TABRET(NWLT,4),TABIM(NWL),TABIMT(NWLT,4)
      !      DIMENSION TEMREF(4)

      !     REFERENCE TEMPERATURES ARE -1.0,-5.0,-20.0, AND -60.0 DEG CENTIGRA

      TEMREF = (/272.16,268.16,253.16,213.16/)

      WLMIN = 0.045
      WLMAX = 8.6E6
      CUTICE = 167.0

      WL(1:114) = (/ &
           0.4430E-01,0.4510E-01,0.4590E-01,0.4680E-01,0.4770E-01,0.4860E-01,&
           0.4960E-01,0.5060E-01,0.5170E-01,0.5280E-01,0.5390E-01,0.5510E-01,&
           0.5640E-01,0.5770E-01,0.5900E-01,0.6050E-01,0.6200E-01,0.6360E-01,&
           0.6530E-01,0.6700E-01,0.6890E-01,0.7080E-01,0.7290E-01,0.7380E-01,&
           0.7510E-01,0.7750E-01,0.8000E-01,0.8270E-01,0.8550E-01,0.8860E-01,&
           0.9180E-01,0.9300E-01,0.9540E-01,0.9920E-01,0.1033E+00,0.1078E+00,&
           0.1100E+00,0.1127E+00,0.1140E+00,0.1181E+00,0.1210E+00,0.1240E+00,&
           0.1272E+00,0.1295E+00,0.1305E+00,0.1319E+00,0.1333E+00,0.1348E+00,&
           0.1362E+00,0.1370E+00,0.1378E+00,0.1387E+00,0.1393E+00,0.1409E+00,&
           0.1425E+00,0.1435E+00,0.1442E+00,0.1450E+00,0.1459E+00,0.1468E+00,&
           0.1476E+00,0.1480E+00,0.1485E+00,0.1494E+00,0.1512E+00,0.1531E+00,&
           0.1540E+00,0.1550E+00,0.1569E+00,0.1580E+00,0.1589E+00,0.1610E+00,&
           0.1625E+00,0.1648E+00,0.1669E+00,0.1692E+00,0.1713E+00,0.1737E+00,&
           0.1757E+00,0.1779E+00,0.1802E+00,0.1809E+00,0.1821E+00,0.1833E+00,&
           0.1843E+00,0.1850E+00,0.1860E+00,0.1870E+00,0.1880E+00,0.1890E+00,&
           0.1900E+00,0.1910E+00,0.1930E+00,0.1950E+00,0.2100E+00,0.2500E+00,&
           0.3000E+00,0.3500E+00,0.4000E+00,0.4100E+00,0.4200E+00,0.4300E+00,&
           0.4400E+00,0.4500E+00,0.4600E+00,0.4700E+00,0.4800E+00,0.4900E+00,&
           0.5000E+00,0.5100E+00,0.5200E+00,0.5300E+00,0.5400E+00,0.5500E+00/)
      WL(115:228) = (/ &
           0.5600E+00,0.5700E+00,0.5800E+00,0.5900E+00,0.6000E+00,0.6100E+00,&
           0.6200E+00,0.6300E+00,0.6400E+00,0.6500E+00,0.6600E+00,0.6700E+00,&
           0.6800E+00,0.6900E+00,0.7000E+00,0.7100E+00,0.7200E+00,0.7300E+00,&
           0.7400E+00,0.7500E+00,0.7600E+00,0.7700E+00,0.7800E+00,0.7900E+00,&
           0.8000E+00,0.8100E+00,0.8200E+00,0.8300E+00,0.8400E+00,0.8500E+00,&
           0.8600E+00,0.8700E+00,0.8800E+00,0.8900E+00,0.9000E+00,0.9100E+00,&
           0.9200E+00,0.9300E+00,0.9400E+00,0.9500E+00,0.9600E+00,0.9700E+00,&
           0.9800E+00,0.9900E+00,0.1000E+01,0.1010E+01,0.1020E+01,0.1030E+01,&
           0.1040E+01,0.1050E+01,0.1060E+01,0.1070E+01,0.1080E+01,0.1090E+01,&
           0.1100E+01,0.1110E+01,0.1120E+01,0.1130E+01,0.1140E+01,0.1150E+01,&
           0.1160E+01,0.1170E+01,0.1180E+01,0.1190E+01,0.1200E+01,0.1210E+01,&
           0.1220E+01,0.1230E+01,0.1240E+01,0.1250E+01,0.1260E+01,0.1270E+01,&
           0.1280E+01,0.1290E+01,0.1300E+01,0.1310E+01,0.1320E+01,0.1330E+01,&
           0.1340E+01,0.1350E+01,0.1360E+01,0.1370E+01,0.1380E+01,0.1390E+01,&
           0.1400E+01,0.1410E+01,0.1420E+01,0.1430E+01,0.1440E+01,0.1449E+01,&
           0.1460E+01,0.1471E+01,0.1481E+01,0.1493E+01,0.1504E+01,0.1515E+01,&
           0.1527E+01,0.1538E+01,0.1563E+01,0.1587E+01,0.1613E+01,0.1650E+01,&
           0.1680E+01,0.1700E+01,0.1730E+01,0.1760E+01,0.1800E+01,0.1830E+01,&
           0.1840E+01,0.1850E+01,0.1855E+01,0.1860E+01,0.1870E+01,0.1890E+01/)
      WL(229:342) = (/ &
           0.1905E+01,0.1923E+01,0.1942E+01,0.1961E+01,0.1980E+01,0.2000E+01,&
           0.2020E+01,0.2041E+01,0.2062E+01,0.2083E+01,0.2105E+01,0.2130E+01,&
           0.2150E+01,0.2170E+01,0.2190E+01,0.2220E+01,0.2240E+01,0.2245E+01,&
           0.2250E+01,0.2260E+01,0.2270E+01,0.2290E+01,0.2310E+01,0.2330E+01,&
           0.2350E+01,0.2370E+01,0.2390E+01,0.2410E+01,0.2430E+01,0.2460E+01,&
           0.2500E+01,0.2520E+01,0.2550E+01,0.2565E+01,0.2580E+01,0.2590E+01,&
           0.2600E+01,0.2620E+01,0.2675E+01,0.2725E+01,0.2778E+01,0.2817E+01,&
           0.2833E+01,0.2849E+01,0.2865E+01,0.2882E+01,0.2899E+01,0.2915E+01,&
           0.2933E+01,0.2950E+01,0.2967E+01,0.2985E+01,0.3003E+01,0.3021E+01,&
           0.3040E+01,0.3058E+01,0.3077E+01,0.3096E+01,0.3115E+01,0.3135E+01,&
           0.3155E+01,0.3175E+01,0.3195E+01,0.3215E+01,0.3236E+01,0.3257E+01,&
           0.3279E+01,0.3300E+01,0.3322E+01,0.3345E+01,0.3367E+01,0.3390E+01,&
           0.3413E+01,0.3436E+01,0.3460E+01,0.3484E+01,0.3509E+01,0.3534E+01,&
           0.3559E+01,0.3624E+01,0.3732E+01,0.3775E+01,0.3847E+01,0.3969E+01,&
           0.4099E+01,0.4239E+01,0.4348E+01,0.4387E+01,0.4444E+01,0.4505E+01,&
           0.4547E+01,0.4560E+01,0.4580E+01,0.4719E+01,0.4904E+01,0.5000E+01,&
           0.5100E+01,0.5200E+01,0.5263E+01,0.5400E+01,0.5556E+01,0.5714E+01,&
           0.5747E+01,0.5780E+01,0.5814E+01,0.5848E+01,0.5882E+01,0.6061E+01,&
           0.6135E+01,0.6250E+01,0.6289E+01,0.6329E+01,0.6369E+01,0.6410E+01/)
      WL(343:456) = (/ &
           0.6452E+01,0.6494E+01,0.6579E+01,0.6667E+01,0.6757E+01,0.6897E+01,&
           0.7042E+01,0.7143E+01,0.7246E+01,0.7353E+01,0.7463E+01,0.7576E+01,&
           0.7692E+01,0.7812E+01,0.7937E+01,0.8065E+01,0.8197E+01,0.8333E+01,&
           0.8475E+01,0.8696E+01,0.8929E+01,0.9091E+01,0.9259E+01,0.9524E+01,&
           0.9804E+01,0.1000E+02,0.1020E+02,0.1031E+02,0.1042E+02,0.1053E+02,&
           0.1064E+02,0.1075E+02,0.1087E+02,0.1100E+02,0.1111E+02,0.1136E+02,&
           0.1163E+02,0.1190E+02,0.1220E+02,0.1250E+02,0.1282E+02,0.1299E+02,&
           0.1316E+02,0.1333E+02,0.1351E+02,0.1370E+02,0.1389E+02,0.1408E+02,&
           0.1429E+02,0.1471E+02,0.1515E+02,0.1538E+02,0.1563E+02,0.1613E+02,&
           0.1639E+02,0.1667E+02,0.1695E+02,0.1724E+02,0.1818E+02,0.1887E+02,&
           0.1923E+02,0.1961E+02,0.2000E+02,0.2041E+02,0.2083E+02,0.2222E+02,&
           0.2260E+02,0.2305E+02,0.2360E+02,0.2460E+02,0.2500E+02,0.2600E+02,&
           0.2857E+02,0.3100E+02,0.3333E+02,0.3448E+02,0.3564E+02,0.3700E+02,&
           0.3824E+02,0.3960E+02,0.4114E+02,0.4276E+02,0.4358E+02,0.4458E+02,&
           0.4550E+02,0.4615E+02,0.4671E+02,0.4736E+02,0.4800E+02,0.4878E+02,&
           0.5003E+02,0.5128E+02,0.5275E+02,0.5350E+02,0.5424E+02,0.5500E+02,&
           0.5574E+02,0.5640E+02,0.5700E+02,0.5746E+02,0.5840E+02,0.5929E+02,&
           0.6000E+02,0.6100E+02,0.6125E+02,0.6250E+02,0.6378E+02,0.6467E+02,&
           0.6558E+02,0.6655E+02,0.6760E+02,0.6900E+02,0.7053E+02,0.7300E+02/)
      WL(457:468) = (/ &
           0.7500E+02,0.7629E+02,0.8000E+02,0.8297E+02,0.8500E+02,0.8680E+02,&
           0.9080E+02,0.9517E+02,0.1000E+03,0.1200E+03,0.1500E+03,0.1670E+03/)
      WLT = (/ &
           0.1670E+03,0.1778E+03,0.1884E+03,&
           0.1995E+03,0.2113E+03,0.2239E+03,0.2371E+03,0.2512E+03,0.2661E+03,&
           0.2818E+03,0.2985E+03,0.3162E+03,0.3548E+03,0.3981E+03,0.4467E+03,&
           0.5012E+03,0.5623E+03,0.6310E+03,0.7943E+03,0.1000E+04,0.1259E+04,&
           0.2500E+04,0.5000E+04,0.1000E+05,0.2000E+05,0.3200E+05,0.3500E+05,&
           0.4000E+05,0.4500E+05,0.5000E+05,0.6000E+05,0.7000E+05,0.9000E+05,&
           0.1110E+06,0.1200E+06,0.1300E+06,0.1400E+06,0.1500E+06,0.1600E+06,&
           0.1700E+06,0.1800E+06,0.2000E+06,0.2500E+06,0.2900E+06,0.3200E+06,&
           0.3500E+06,0.3800E+06,0.4000E+06,0.4500E+06,0.5000E+06,0.6000E+06,&
           0.6400E+06,0.6800E+06,0.7200E+06,0.7600E+06,0.8000E+06,0.8400E+06,&
           0.9000E+06,0.1000E+07,0.2000E+07,0.5000E+07,0.8600E+07/)
      TABRE(1:114) = (/ &
           0.83441,   0.83676,   0.83729,   0.83771,   0.83827,   0.84038,&
           0.84719,   0.85522,   0.86047,   0.86248,   0.86157,   0.86093,&
           0.86419,   0.86916,   0.87764,   0.89296,   0.91041,   0.93089,&
           0.95373,   0.98188,   1.02334,   1.06735,   1.11197,   1.13134,&
           1.15747,   1.20045,   1.23840,   1.27325,   1.32157,   1.38958,&
           1.41644,   1.40906,   1.40063,   1.40169,   1.40934,   1.40221,&
           1.39240,   1.38424,   1.38075,   1.38186,   1.39634,   1.40918,&
           1.40256,   1.38013,   1.36303,   1.34144,   1.32377,   1.30605,&
           1.29054,   1.28890,   1.28931,   1.30190,   1.32025,   1.36302,&
           1.41872,   1.45834,   1.49028,   1.52128,   1.55376,   1.57782,&
           1.59636,   1.60652,   1.61172,   1.61919,   1.62522,   1.63404,&
           1.63689,   1.63833,   1.63720,   1.63233,   1.62222,   1.58269,&
           1.55635,   1.52453,   1.50320,   1.48498,   1.47226,   1.45991,&
           1.45115,   1.44272,   1.43498,   1.43280,   1.42924,   1.42602,&
           1.42323,   1.42143,   1.41897,   1.41660,   1.41434,   1.41216,&
           1.41006,   1.40805,   1.40423,   1.40067,   1.38004,   1.35085,&
           1.33394,   1.32492,   1.31940,   1.31854,   1.31775,   1.31702,&
           1.31633,   1.31569,   1.31509,   1.31452,   1.31399,   1.31349,&
           1.31302,   1.31257,   1.31215,   1.31175,   1.31136,   1.31099/)
      TABRE(115:228) = (/ &
           1.31064,   1.31031,   1.30999,   1.30968,   1.30938,   1.30909,&
           1.30882,   1.30855,   1.30829,   1.30804,   1.30780,   1.30756,&
           1.30733,   1.30710,   1.30688,   1.30667,   1.30646,   1.30625,&
           1.30605,   1.30585,   1.30566,   1.30547,   1.30528,   1.30509,&
           1.30491,   1.30473,   1.30455,   1.30437,   1.30419,   1.30402,&
           1.30385,   1.30367,   1.30350,   1.30333,   1.30316,   1.30299,&
           1.30283,   1.30266,   1.30249,   1.30232,   1.30216,   1.30199,&
           1.30182,   1.30166,   1.30149,   1.30132,   1.30116,   1.30099,&
           1.30082,   1.30065,   1.30048,   1.30031,   1.30014,   1.29997,&
           1.29979,   1.29962,   1.29945,   1.29927,   1.29909,   1.29891,&
           1.29873,   1.29855,   1.29837,   1.29818,   1.29800,   1.29781,&
           1.29762,   1.29743,   1.29724,   1.29705,   1.29686,   1.29666,&
           1.29646,   1.29626,   1.29605,   1.29584,   1.29563,   1.29542,&
           1.29521,   1.29499,   1.29476,   1.29453,   1.29430,   1.29406,&
           1.29381,   1.29355,   1.29327,   1.29299,   1.29272,   1.29252,&
           1.29228,   1.29205,   1.29186,   1.29167,   1.29150,   1.29130,&
           1.29106,   1.29083,   1.29025,   1.28962,   1.28891,   1.28784,&
           1.28689,   1.28623,   1.28521,   1.28413,   1.28261,   1.28137,&
           1.28093,   1.28047,   1.28022,   1.27998,   1.27948,   1.27849/)
      TABRE(229:342) = (/ &
           1.27774,   1.27691,   1.27610,   1.27535,   1.27471,   1.27404,&
           1.27329,   1.27240,   1.27139,   1.27029,   1.26901,   1.26736,&
           1.26591,   1.26441,   1.26284,   1.26036,   1.25860,   1.25815,&
           1.25768,   1.25675,   1.25579,   1.25383,   1.25179,   1.24967,&
           1.24745,   1.24512,   1.24266,   1.24004,   1.23725,   1.23270,&
           1.22583,   1.22198,   1.21548,   1.21184,   1.20790,   1.20507,&
           1.20209,   1.19566,   1.17411,   1.14734,   1.10766,   1.06739,&
           1.04762,   1.02650,   1.00357,   0.98197,   0.96503,   0.95962,&
           0.97269,   0.99172,   1.00668,   1.02186,   1.04270,   1.07597,&
           1.12954,   1.21267,   1.32509,   1.42599,   1.49656,   1.55095,&
           1.59988,   1.63631,   1.65024,   1.64278,   1.62691,   1.61284,&
           1.59245,   1.57329,   1.55770,   1.54129,   1.52654,   1.51139,&
           1.49725,   1.48453,   1.47209,   1.46125,   1.45132,   1.44215,&
           1.43366,   1.41553,   1.39417,   1.38732,   1.37735,   1.36448,&
           1.35414,   1.34456,   1.33882,   1.33807,   1.33847,   1.34053,&
           1.34287,   1.34418,   1.34634,   1.34422,   1.33453,   1.32897,&
           1.32333,   1.31800,   1.31432,   1.30623,   1.29722,   1.28898,&
           1.28730,   1.28603,   1.28509,   1.28535,   1.28813,   1.30156,&
           1.30901,   1.31720,   1.31893,   1.32039,   1.32201,   1.32239/)
      TABRE(343:456) = (/ &
           1.32149,   1.32036,   1.31814,   1.31705,   1.31807,   1.31953,&
           1.31933,   1.31896,   1.31909,   1.31796,   1.31631,   1.31542,&
           1.31540,   1.31552,   1.31455,   1.31193,   1.30677,   1.29934,&
           1.29253,   1.28389,   1.27401,   1.26724,   1.25990,   1.24510,&
           1.22241,   1.19913,   1.17150,   1.15528,   1.13700,   1.11808,&
           1.10134,   1.09083,   1.08734,   1.09254,   1.10654,   1.14779,&
           1.20202,   1.25825,   1.32305,   1.38574,   1.44478,   1.47170,&
           1.49619,   1.51652,   1.53328,   1.54900,   1.56276,   1.57317,&
           1.58028,   1.57918,   1.56672,   1.55869,   1.55081,   1.53807,&
           1.53296,   1.53220,   1.53340,   1.53289,   1.51705,   1.50097,&
           1.49681,   1.49928,   1.50153,   1.49856,   1.49053,   1.46070,&
           1.45182,   1.44223,   1.43158,   1.41385,   1.40676,   1.38955,&
           1.34894,   1.31039,   1.26420,   1.23656,   1.21663,   1.20233,&
           1.19640,   1.19969,   1.20860,   1.22173,   1.24166,   1.28175,&
           1.32784,   1.38657,   1.46486,   1.55323,   1.60379,   1.61877,&
           1.62963,   1.65712,   1.69810,   1.72065,   1.74865,   1.76736,&
           1.76476,   1.75011,   1.72327,   1.68490,   1.62398,   1.59596,&
           1.58514,   1.59917,   1.61405,   1.66625,   1.70663,   1.73713,&
           1.76860,   1.80343,   1.83296,   1.85682,   1.87411,   1.89110/)
      TABRE(457:468) = (/ &
           1.89918,   1.90432,   1.90329,   1.88744,   1.87499,   1.86702,&
           1.85361,   1.84250,   1.83225,   1.81914,   1.82268,   1.82961/)
      TABRET(1:NWLT,1) = (/ &
           1.82961,   1.83258,   1.83149,&
           1.82748,   1.82224,   1.81718,   1.81204,   1.80704,   1.80250,&
           1.79834,   1.79482,   1.79214,   1.78843,   1.78601,   1.78434,&
           1.78322,   1.78248,   1.78201,   1.78170,   1.78160,   1.78190,&
           1.78300,   1.78430,   1.78520,   1.78620,   1.78660,   1.78680,&
           1.78690,   1.78700,   1.78700,   1.78710,   1.78710,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78800/)
      TABRET(1:NWLT,2) = (/ &
           1.82961,   1.83258,   1.83149,   1.82748,&
           1.82224,   1.81718,   1.81204,   1.80704,   1.80250,   1.79834,&
           1.79482,   1.79214,   1.78843,   1.78601,   1.78434,   1.78322,&
           1.78248,   1.78201,   1.78170,   1.78160,   1.78190,   1.78300,&
           1.78430,   1.78520,   1.78610,   1.78630,   1.78640,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78720/)
      TABRET(1:NWLT,3) = (/ &
           1.82961,   1.83258,   1.83149,   1.82748,   1.82224,&
           1.81718,   1.81204,   1.80704,   1.80250,   1.79834,   1.79482,&
           1.79214,   1.78843,   1.78601,   1.78434,   1.78322,   1.78248,&
           1.78201,   1.78160,   1.78140,   1.78160,   1.78220,   1.78310,&
           1.78380,   1.78390,   1.78400,   1.78400,   1.78400,   1.78400,&
           1.78400,   1.78390,   1.78380,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78400,   1.78450/)
      TABRET(1:NWLT,4) = (/ &
           1.82961,   1.83258,   1.83149,   1.82748,   1.82224,   1.81718,&
           1.81204,   1.80704,   1.80250,   1.79834,   1.79482,   1.79214,&
           1.78843,   1.78601,   1.78434,   1.78322,   1.78248,   1.78201,&
           1.78150,   1.78070,   1.78010,   1.77890,   1.77790,   1.77730,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77800/)
      TABIM(1:114)  = (/ &
           0.1640E+00,0.1730E+00,0.1830E+00,0.1950E+00,0.2080E+00,0.2230E+00,&
           0.2400E+00,0.2500E+00,0.2590E+00,0.2680E+00,0.2790E+00,0.2970E+00,&
           0.3190E+00,0.3400E+00,0.3660E+00,0.3920E+00,0.4160E+00,0.4400E+00,&
           0.4640E+00,0.4920E+00,0.5170E+00,0.5280E+00,0.5330E+00,0.5340E+00,&
           0.5310E+00,0.5240E+00,0.5100E+00,0.5000E+00,0.4990E+00,0.4680E+00,&
           0.3800E+00,0.3600E+00,0.3390E+00,0.3180E+00,0.2910E+00,0.2510E+00,&
           0.2440E+00,0.2390E+00,0.2390E+00,0.2440E+00,0.2470E+00,0.2240E+00,&
           0.1950E+00,0.1740E+00,0.1720E+00,0.1800E+00,0.1940E+00,0.2130E+00,&
           0.2430E+00,0.2710E+00,0.2890E+00,0.3340E+00,0.3440E+00,0.3820E+00,&
           0.4010E+00,0.4065E+00,0.4050E+00,0.3890E+00,0.3770E+00,0.3450E+00,&
           0.3320E+00,0.3150E+00,0.2980E+00,0.2740E+00,0.2280E+00,0.1980E+00,&
           0.1720E+00,0.1560E+00,0.1100E+00,0.8300E-01,0.5800E-01,0.2200E-01,&
           0.1000E-01,0.3000E-02,0.1000E-02,0.3000E-03,0.1000E-03,0.3000E-04,&
           0.1000E-04,0.3000E-05,0.1000E-05,0.7000E-06,0.4000E-06,0.2000E-06,&
           0.1000E-06,0.6377E-07,0.3750E-07,0.2800E-07,0.2400E-07,0.2200E-07,&
           0.1900E-07,0.1750E-07,0.1640E-07,0.1590E-07,0.1325E-07,0.8623E-08,&
           0.5504E-08,0.3765E-08,0.2710E-08,0.2510E-08,0.2260E-08,0.2080E-08,&
           0.1910E-08,0.1540E-08,0.1530E-08,0.1550E-08,0.1640E-08,0.1780E-08,&
           0.1910E-08,0.2140E-08,0.2260E-08,0.2540E-08,0.2930E-08,0.3110E-08/)
      TABIM(115:228)  = (/ &
           0.3290E-08,0.3520E-08,0.4040E-08,0.4880E-08,0.5730E-08,0.6890E-08,&
           0.8580E-08,0.1040E-07,0.1220E-07,0.1430E-07,0.1660E-07,0.1890E-07,&
           0.2090E-07,0.2400E-07,0.2900E-07,0.3440E-07,0.4030E-07,0.4300E-07,&
           0.4920E-07,0.5870E-07,0.7080E-07,0.8580E-07,0.1020E-06,0.1180E-06,&
           0.1340E-06,0.1400E-06,0.1430E-06,0.1450E-06,0.1510E-06,0.1830E-06,&
           0.2150E-06,0.2650E-06,0.3350E-06,0.3920E-06,0.4200E-06,0.4440E-06,&
           0.4740E-06,0.5110E-06,0.5530E-06,0.6020E-06,0.7550E-06,0.9260E-06,&
           0.1120E-05,0.1330E-05,0.1620E-05,0.2000E-05,0.2250E-05,0.2330E-05,&
           0.2330E-05,0.2170E-05,0.1960E-05,0.1810E-05,0.1740E-05,0.1730E-05,&
           0.1700E-05,0.1760E-05,0.1820E-05,0.2040E-05,0.2250E-05,0.2290E-05,&
           0.3040E-05,0.3840E-05,0.4770E-05,0.5760E-05,0.6710E-05,0.8660E-05,&
           0.1020E-04,0.1130E-04,0.1220E-04,0.1290E-04,0.1320E-04,0.1350E-04,&
           0.1330E-04,0.1320E-04,0.1320E-04,0.1310E-04,0.1320E-04,0.1320E-04,&
           0.1340E-04,0.1390E-04,0.1420E-04,0.1480E-04,0.1580E-04,0.1740E-04,&
           0.1980E-04,0.2500E-04,0.5400E-04,0.1040E-03,0.2030E-03,0.2708E-03,&
           0.3511E-03,0.4299E-03,0.5181E-03,0.5855E-03,0.5899E-03,0.5635E-03,&
           0.5480E-03,0.5266E-03,0.4394E-03,0.3701E-03,0.3372E-03,0.2410E-03,&
           0.1890E-03,0.1660E-03,0.1450E-03,0.1280E-03,0.1030E-03,0.8600E-04,&
           0.8220E-04,0.8030E-04,0.8500E-04,0.9900E-04,0.1500E-03,0.2950E-03/)
      TABIM(229:342)  = (/ &
           0.4687E-03,0.7615E-03,0.1010E-02,0.1313E-02,0.1539E-02,0.1588E-02,&
           0.1540E-02,0.1412E-02,0.1244E-02,0.1068E-02,0.8414E-03,0.5650E-03,&
           0.4320E-03,0.3500E-03,0.2870E-03,0.2210E-03,0.2030E-03,0.2010E-03,&
           0.2030E-03,0.2140E-03,0.2320E-03,0.2890E-03,0.3810E-03,0.4620E-03,&
           0.5480E-03,0.6180E-03,0.6800E-03,0.7300E-03,0.7820E-03,0.8480E-03,&
           0.9250E-03,0.9200E-03,0.8920E-03,0.8700E-03,0.8900E-03,0.9300E-03,&
           0.1010E-02,0.1350E-02,0.3420E-02,0.7920E-02,0.2000E-01,0.3800E-01,&
           0.5200E-01,0.6800E-01,0.9230E-01,0.1270E+00,0.1690E+00,0.2210E+00,&
           0.2760E+00,0.3120E+00,0.3470E+00,0.3880E+00,0.4380E+00,0.4930E+00,&
           0.5540E+00,0.6120E+00,0.6250E+00,0.5930E+00,0.5390E+00,0.4910E+00,&
           0.4380E+00,0.3720E+00,0.3000E+00,0.2380E+00,0.1930E+00,0.1580E+00,&
           0.1210E+00,0.1030E+00,0.8360E-01,0.6680E-01,0.5400E-01,0.4220E-01,&
           0.3420E-01,0.2740E-01,0.2200E-01,0.1860E-01,0.1520E-01,0.1260E-01,&
           0.1060E-01,0.8020E-02,0.6850E-02,0.6600E-02,0.6960E-02,0.9160E-02,&
           0.1110E-01,0.1450E-01,0.2000E-01,0.2300E-01,0.2600E-01,0.2900E-01,&
           0.2930E-01,0.3000E-01,0.2850E-01,0.1730E-01,0.1290E-01,0.1200E-01,&
           0.1250E-01,0.1340E-01,0.1400E-01,0.1750E-01,0.2400E-01,0.3500E-01,&
           0.3800E-01,0.4200E-01,0.4600E-01,0.5200E-01,0.5700E-01,0.6900E-01,&
           0.7000E-01,0.6700E-01,0.6500E-01,0.6400E-01,0.6200E-01,0.5900E-01/)
      TABIM(343:456)  = (/ &
           0.5700E-01,0.5600E-01,0.5500E-01,0.5700E-01,0.5800E-01,0.5700E-01,&
           0.5500E-01,0.5500E-01,0.5400E-01,0.5200E-01,0.5200E-01,0.5200E-01,&
           0.5200E-01,0.5000E-01,0.4700E-01,0.4300E-01,0.3900E-01,0.3700E-01,&
           0.3900E-01,0.4000E-01,0.4200E-01,0.4400E-01,0.4500E-01,0.4600E-01,&
           0.4700E-01,0.5100E-01,0.6500E-01,0.7500E-01,0.8800E-01,0.1080E+00,&
           0.1340E+00,0.1680E+00,0.2040E+00,0.2480E+00,0.2800E+00,0.3410E+00,&
           0.3790E+00,0.4090E+00,0.4220E+00,0.4220E+00,0.4030E+00,0.3890E+00,&
           0.3740E+00,0.3540E+00,0.3350E+00,0.3150E+00,0.2940E+00,0.2710E+00,&
           0.2460E+00,0.1980E+00,0.1640E+00,0.1520E+00,0.1420E+00,0.1280E+00,&
           0.1250E+00,0.1230E+00,0.1160E+00,0.1070E+00,0.7900E-01,0.7200E-01,&
           0.7600E-01,0.7500E-01,0.6700E-01,0.5500E-01,0.4500E-01,0.2900E-01,&
           0.2750E-01,0.2700E-01,0.2730E-01,0.2890E-01,0.3000E-01,0.3400E-01,&
           0.5300E-01,0.7550E-01,0.1060E+00,0.1350E+00,0.1761E+00,0.2229E+00,&
           0.2746E+00,0.3280E+00,0.3906E+00,0.4642E+00,0.5247E+00,0.5731E+00,&
           0.6362E+00,0.6839E+00,0.7091E+00,0.6790E+00,0.6250E+00,0.5654E+00,&
           0.5433E+00,0.5292E+00,0.5070E+00,0.4883E+00,0.4707E+00,0.4203E+00,&
           0.3771E+00,0.3376E+00,0.3056E+00,0.2835E+00,0.3170E+00,0.3517E+00,&
           0.3902E+00,0.4509E+00,0.4671E+00,0.4779E+00,0.4890E+00,0.4899E+00,&
           0.4873E+00,0.4766E+00,0.4508E+00,0.4193E+00,0.3880E+00,0.3433E+00/)
      TABIM(457:468)  = (/ &
           0.3118E+00,0.2935E+00,0.2350E+00,0.1981E+00,0.1865E+00,0.1771E+00,&
           0.1620E+00,0.1490E+00,0.1390E+00,0.1200E+00,0.9620E-01,0.8300E-01/)


      TABIMT(1:NWLT,1)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,&
           0.4560E-01,0.3790E-01,0.3140E-01,0.2620E-01,0.2240E-01,0.1960E-01,&
           0.1760E-01,0.1665E-01,0.1620E-01,0.1550E-01,0.1470E-01,0.1390E-01,&
           0.1320E-01,0.1250E-01,0.1180E-01,0.1060E-01,0.9540E-02,0.8560E-02,&
           0.6210E-02,0.4490E-02,0.3240E-02,0.2340E-02,0.1880E-02,0.1740E-02,&
           0.1500E-02,0.1320E-02,0.1160E-02,0.8800E-03,0.6950E-03,0.4640E-03,&
           0.3400E-03,0.3110E-03,0.2940E-03,0.2790E-03,0.2700E-03,0.2640E-03,&
           0.2580E-03,0.2520E-03,0.2490E-03,0.2540E-03,0.2640E-03,0.2740E-03,&
           0.2890E-03,0.3050E-03,0.3150E-03,0.3460E-03,0.3820E-03,0.4620E-03,&
           0.5000E-03,0.5500E-03,0.5950E-03,0.6470E-03,0.6920E-03,0.7420E-03,&
           0.8200E-03,0.9700E-03,0.1950E-02,0.5780E-02,0.9700E-02/)
      TABIMT(1:NWLT,2)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,0.4560E-01,&
           0.3790E-01,0.3140E-01,0.2620E-01,0.2240E-01,0.1960E-01,0.1760E-01,&
           0.1665E-01,0.1600E-01,0.1500E-01,0.1400E-01,0.1310E-01,0.1230E-01,&
           0.1150E-01,0.1080E-01,0.9460E-02,0.8290E-02,0.7270E-02,0.4910E-02,&
           0.3300E-02,0.2220E-02,0.1490E-02,0.1140E-02,0.1060E-02,0.9480E-03,&
           0.8500E-03,0.7660E-03,0.6300E-03,0.5200E-03,0.3840E-03,0.2960E-03,&
           0.2700E-03,0.2520E-03,0.2440E-03,0.2360E-03,0.2300E-03,0.2280E-03,&
           0.2250E-03,0.2200E-03,0.2160E-03,0.2170E-03,0.2200E-03,0.2250E-03,&
           0.2320E-03,0.2390E-03,0.2600E-03,0.2860E-03,0.3560E-03,0.3830E-03,&
           0.4150E-03,0.4450E-03,0.4760E-03,0.5080E-03,0.5400E-03,0.5860E-03,&
           0.6780E-03,0.1280E-02,0.3550E-02,0.5600E-02/)
      TABIMT(1:NWLT,3)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,0.4560E-01,0.3790E-01,&
           0.3140E-01,0.2620E-01,0.2190E-01,0.1880E-01,0.1660E-01,0.1540E-01,&
           0.1470E-01,0.1350E-01,0.1250E-01,0.1150E-01,0.1060E-01,0.9770E-02,&
           0.9010E-02,0.7660E-02,0.6520E-02,0.5540E-02,0.3420E-02,0.2100E-02,&
           0.1290E-02,0.7930E-03,0.5700E-03,0.5350E-03,0.4820E-03,0.4380E-03,&
           0.4080E-03,0.3500E-03,0.3200E-03,0.2550E-03,0.2120E-03,0.2000E-03,&
           0.1860E-03,0.1750E-03,0.1660E-03,0.1560E-03,0.1490E-03,0.1440E-03,&
           0.1350E-03,0.1210E-03,0.1160E-03,0.1160E-03,0.1170E-03,0.1200E-03,&
           0.1230E-03,0.1320E-03,0.1440E-03,0.1680E-03,0.1800E-03,0.1900E-03,&
           0.2090E-03,0.2160E-03,0.2290E-03,0.2400E-03,0.2600E-03,0.2920E-03,&
           0.6100E-03,0.1020E-02,0.1810E-02/)
      TABIMT(1:NWLT,4)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,0.4450E-01,0.3550E-01,0.2910E-01,&
           0.2440E-01,0.1970E-01,0.1670E-01,0.1400E-01,0.1235E-01,0.1080E-01,&
           0.8900E-02,0.7340E-02,0.6400E-02,0.5600E-02,0.5000E-02,0.4520E-02,&
           0.3680E-02,0.2990E-02,0.2490E-02,0.1550E-02,0.9610E-03,0.5950E-03,&
           0.3690E-03,0.2670E-03,0.2510E-03,0.2290E-03,0.2110E-03,0.1960E-03,&
           0.1730E-03,0.1550E-03,0.1310E-03,0.1130E-03,0.1060E-03,0.9900E-04,&
           0.9300E-04,0.8730E-04,0.8300E-04,0.7870E-04,0.7500E-04,0.6830E-04,&
           0.5600E-04,0.4960E-04,0.4550E-04,0.4210E-04,0.3910E-04,0.3760E-04,&
           0.3400E-04,0.3100E-04,0.2640E-04,0.2510E-04,0.2430E-04,0.2390E-04,&
           0.2370E-04,0.2380E-04,0.2400E-04,0.2460E-04,0.2660E-04,0.4450E-04,&
           0.8700E-04,0.1320E-03/)

      !     ZERO PARAMETERS

      RN=0.0d0
      CN=0.0d0
      ABSIND=0.0d0
      ABSCOF=0.0d0

      !     CONVERT WAVELENGTH TO MICRONS

      ALAM=XLAM
      IF (IUNIT.EQ.1) ALAM=1000*ALAM
      IF (IUNIT.EQ.2) ALAM=10000*ALAM
      IF (IUNIT.EQ.3) ALAM=10000*(1.0/ALAM)

      IF (ALAM.GE.WLMIN.AND.ALAM.LE.WLMAX) THEN
        IF (ALAM.LE.CUTICE) THEN
          !     REGION FROM 0.045 MICRONS TO 167.0 MICRONS - NO TEMPERATURE DEPEND
          !        DO I=2,NWL
          !          IF (ALAM.LT.WL(I)) EXIT
          !        END DO
          ! alternatively: vectorized search for interpolation interval:
          Ihlp = 0
          DO k=2,NWL
            IF (ALAM.LE.WL(k).AND.ALAM.GE.WL(k-1)) THEN
              Ihlp(k) = 1
            END IF
          END DO
          I1 = MAXLOC(Ihlp)
          I = I1(1)

          X1=LOG(WL(I-1))
          X2=LOG(WL(I))
          Y1=TABRE(I-1)
          Y2=TABRE(I)
          X=LOG(ALAM)
          Y=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
          RN=Y
          Y1=LOG(ABS(TABIM(I-1)))
          Y2=LOG(ABS(TABIM(I)))
          Y=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
          CN=EXP(Y)
        ELSE
          !     REGION FROM 167.0 MICRONS TO 8.6 METERS - TEMPERATURE DEPENDENCE     
          TK=T
          WHERE (TK.GT.TEMREF(1)) 
            TK=TEMREF(1)
          END WHERE
          WHERE (TK.LT.TEMREF(4)) 
            TK=TEMREF(4)
          END WHERE

          LT1 = 2
          DO I=2,4
            DO k=1,anz
              IF(TK(k).GE.TEMREF(I).AND.TK(k).LE.TEMREF(I-1)) THEN
                LT1(k) = I
              END IF
            END DO
          END DO
          LT2=LT1-1

          !        DO I=2,NWLT
          !          IF (ALAM.LT.WLT(I)) EXIT
          !        END DO
          ! alternatively: vectorized search for interpolation interval:
          IhlpT = 0
          DO k=2,NWLT
            IF (ALAM.LE.WLT(k).AND.ALAM.GE.WLT(k-1)) THEN
              IhlpT(k) = 1
            END IF
          END DO
          I1 = MAXLOC(IhlpT)
          I = I1(1)

          X1=LOG(WLT(I-1))
          X2=LOG(WLT(I))

          X=LOG(ALAM)
          DO k=1,anz
            T1=TEMREF(LT1(k))
            T2=TEMREF(LT2(k))
            Y1=TABRET(I-1,LT1(k))
            Y2=TABRET(I,LT1(k))
            YLO=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
            Y1=TABRET(I-1,LT2(k))
            Y2=TABRET(I,LT2(k))
            YHI=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
            Y=((TK(k)-T1)*(YHI-YLO)/(T2-T1))+YLO
            RN(k)=Y
            Y1=LOG(ABS(TABIMT(I-1,LT1(k))))
            Y2=LOG(ABS(TABIMT(I,LT1(k))))
            YLO=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
            Y1=LOG(ABS(TABIMT(I-1,LT2(k))))
            Y2=LOG(ABS(TABIMT(I,LT2(k))))
            YHI=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
            Y=((TK(k)-T1)*(YHI-YLO)/(T2-T1))+YLO
            CN(k)=EXP(Y)
          END DO
        END IF

        !     ABSORPTIVE QUANITIES
        ABSIND=CN/RN
        ABSCOF=4.0*PI*CN/ALAM

      END IF

      RETURN

    END SUBROUTINE REFICEVEC

  END FUNCTION m_complex_ice_warren_vec

  FUNCTION m_complex_ice_warren(lambda,T)

    ! Komplexer Brechungsindex von Eis als Funktion der Temperatur
    ! [Grad C] und der Wellenlaenge [m]; Gueltig fuer 
    ! lambda aus [45*1e-9,8.6] m; T aus [-60.0,0.0] Grad C
    ! nach Warren (1983) - seine Fortran-routine wird benutzt
    ! Brechungsindex m_i wird in der Konvention mit negativem 
    ! Imaginaerteil zurueckgegeben.

    IMPLICIT NONE

    COMPLEX(kind=kinddp) :: m_complex_ice_warren
    DOUBLE PRECISION     :: T,lambda,TT
    DOUBLE PRECISION     :: m_r, m_c, absind, abscof

    TT = T + 273.16
    CALL REFICE(1,lambda*1000.0d0,TT,m_r,m_c,absind,abscof)

    m_complex_ice_warren = CMPLX(m_r,-m_c,kind=kinddp)

  CONTAINS

    SUBROUTINE REFICE(IUNIT,XLAM,T,RN,CN,ABSIND,ABSCOF)

      IMPLICIT NONE
      ! Arguments:
      INTEGER :: IUNIT
      DOUBLE PRECISION :: ABSCOF,ABSIND,CN,XLAM,RN,T
      ! Parameters:
      INTEGER :: I,LT1,LT2,NWL,NWLT
      PARAMETER(NWL=468,NWLT=62)
      ! Local variables:
      DOUBLE PRECISION :: &
           ALAM,CUTICE,T1,T2,TK,WLMAX,WLMIN, &
           X,X1,X2,Y,Y1,Y2,YLO,YHI

      DOUBLE PRECISION :: &
           TABIM(NWL),TABIMT(NWLT,4),TABRE(NWL),TABRET(NWLT,4),TEMREF(4),&
           WL(NWL),WLT(NWLT)

      !     DEFINES WAVELENGTH DEPENDENT COMPLEX INDEX OF REFRACTION FOR ICE.
      !     ALLOWABLE WAVELENGTH RANGE EXTENDS FROM 0.045 MICRONS TO 8.6 METER
      !     TEMPERATURE DEPENDENCE ONLY CONSIDERED BEYOND 167 MICRONS.

      !     INTERPOLATION IS DONE     RN  VS. LOG(XLAM)
      !                               RN  VS.        T
      !                           LOG(CN) VS. LOG(XLAM)
      !                           LOG(CN) VS.        T

      !     STEPHEN G. WARREN - 1983
      !     DEPT. OF ATMOSPHERIC SCIENCES
      !     UNIVERSITY OF WASHINGTON
      !     SEATTLE, WA  98195

      !     BASED ON

      !        WARREN,S.G.,1984.
      !        OPTICAL CONSTANTS OF ICE FROM THE ULTRAVIOLET TO THE MICROWAVE.
      !        APPLIED OPTICS,23,1206-1225

      !     INPUT PARAMETERS

      !     IUNIT = 0 FOR WAVELENGTH SPECIFIED IN MICRONS
      !           = 1 FOR WAVELENGTH SPECIFIED IN MILLIMETERS
      !           = 2 FOR WAVELENGTH SPECIFIED IN CENTIMETERS
      !           = 3 FOR WAVELENGTH SPECIFIED IN INVERSE CENTIMETERS ( WAVE N
      !     XLAM = WAVELENGTH ( MICRONS OR MM OR CM OR CM**-1 )
      !     T = TEMPERATURE ( DEGREES KELVIN )

      !     OUTPUT PARAMETERS

      !     RN = REAL PORTION ( SCATTERING )
      !     CN = COMPLEX PORTION ( ABSORPTION )
      !     ABSIND = ABSORPTIVE INDEX ( CN/RN )
      !     ABSCOF = ABORPTION COEFFICIENT ( 4*PI*CN/XLAM )

      !      DIMENSION WL(NWL),WLT(NWLT)
      !      DIMENSION TABRE(NWL),TABRET(NWLT,4),TABIM(NWL),TABIMT(NWLT,4)
      !      DIMENSION TEMREF(4)

      !     REFERENCE TEMPERATURES ARE -1.0,-5.0,-20.0, AND -60.0 DEG CENTIGRA

      TEMREF = (/272.16,268.16,253.16,213.16/)

      WLMIN = 0.045
      WLMAX = 8.6E6
      CUTICE = 167.0

      WL(1:114) = (/ &
           0.4430E-01,0.4510E-01,0.4590E-01,0.4680E-01,0.4770E-01,0.4860E-01,&
           0.4960E-01,0.5060E-01,0.5170E-01,0.5280E-01,0.5390E-01,0.5510E-01,&
           0.5640E-01,0.5770E-01,0.5900E-01,0.6050E-01,0.6200E-01,0.6360E-01,&
           0.6530E-01,0.6700E-01,0.6890E-01,0.7080E-01,0.7290E-01,0.7380E-01,&
           0.7510E-01,0.7750E-01,0.8000E-01,0.8270E-01,0.8550E-01,0.8860E-01,&
           0.9180E-01,0.9300E-01,0.9540E-01,0.9920E-01,0.1033E+00,0.1078E+00,&
           0.1100E+00,0.1127E+00,0.1140E+00,0.1181E+00,0.1210E+00,0.1240E+00,&
           0.1272E+00,0.1295E+00,0.1305E+00,0.1319E+00,0.1333E+00,0.1348E+00,&
           0.1362E+00,0.1370E+00,0.1378E+00,0.1387E+00,0.1393E+00,0.1409E+00,&
           0.1425E+00,0.1435E+00,0.1442E+00,0.1450E+00,0.1459E+00,0.1468E+00,&
           0.1476E+00,0.1480E+00,0.1485E+00,0.1494E+00,0.1512E+00,0.1531E+00,&
           0.1540E+00,0.1550E+00,0.1569E+00,0.1580E+00,0.1589E+00,0.1610E+00,&
           0.1625E+00,0.1648E+00,0.1669E+00,0.1692E+00,0.1713E+00,0.1737E+00,&
           0.1757E+00,0.1779E+00,0.1802E+00,0.1809E+00,0.1821E+00,0.1833E+00,&
           0.1843E+00,0.1850E+00,0.1860E+00,0.1870E+00,0.1880E+00,0.1890E+00,&
           0.1900E+00,0.1910E+00,0.1930E+00,0.1950E+00,0.2100E+00,0.2500E+00,&
           0.3000E+00,0.3500E+00,0.4000E+00,0.4100E+00,0.4200E+00,0.4300E+00,&
           0.4400E+00,0.4500E+00,0.4600E+00,0.4700E+00,0.4800E+00,0.4900E+00,&
           0.5000E+00,0.5100E+00,0.5200E+00,0.5300E+00,0.5400E+00,0.5500E+00/)
      WL(115:228) = (/ &
           0.5600E+00,0.5700E+00,0.5800E+00,0.5900E+00,0.6000E+00,0.6100E+00,&
           0.6200E+00,0.6300E+00,0.6400E+00,0.6500E+00,0.6600E+00,0.6700E+00,&
           0.6800E+00,0.6900E+00,0.7000E+00,0.7100E+00,0.7200E+00,0.7300E+00,&
           0.7400E+00,0.7500E+00,0.7600E+00,0.7700E+00,0.7800E+00,0.7900E+00,&
           0.8000E+00,0.8100E+00,0.8200E+00,0.8300E+00,0.8400E+00,0.8500E+00,&
           0.8600E+00,0.8700E+00,0.8800E+00,0.8900E+00,0.9000E+00,0.9100E+00,&
           0.9200E+00,0.9300E+00,0.9400E+00,0.9500E+00,0.9600E+00,0.9700E+00,&
           0.9800E+00,0.9900E+00,0.1000E+01,0.1010E+01,0.1020E+01,0.1030E+01,&
           0.1040E+01,0.1050E+01,0.1060E+01,0.1070E+01,0.1080E+01,0.1090E+01,&
           0.1100E+01,0.1110E+01,0.1120E+01,0.1130E+01,0.1140E+01,0.1150E+01,&
           0.1160E+01,0.1170E+01,0.1180E+01,0.1190E+01,0.1200E+01,0.1210E+01,&
           0.1220E+01,0.1230E+01,0.1240E+01,0.1250E+01,0.1260E+01,0.1270E+01,&
           0.1280E+01,0.1290E+01,0.1300E+01,0.1310E+01,0.1320E+01,0.1330E+01,&
           0.1340E+01,0.1350E+01,0.1360E+01,0.1370E+01,0.1380E+01,0.1390E+01,&
           0.1400E+01,0.1410E+01,0.1420E+01,0.1430E+01,0.1440E+01,0.1449E+01,&
           0.1460E+01,0.1471E+01,0.1481E+01,0.1493E+01,0.1504E+01,0.1515E+01,&
           0.1527E+01,0.1538E+01,0.1563E+01,0.1587E+01,0.1613E+01,0.1650E+01,&
           0.1680E+01,0.1700E+01,0.1730E+01,0.1760E+01,0.1800E+01,0.1830E+01,&
           0.1840E+01,0.1850E+01,0.1855E+01,0.1860E+01,0.1870E+01,0.1890E+01/)
      WL(229:342) = (/ &
           0.1905E+01,0.1923E+01,0.1942E+01,0.1961E+01,0.1980E+01,0.2000E+01,&
           0.2020E+01,0.2041E+01,0.2062E+01,0.2083E+01,0.2105E+01,0.2130E+01,&
           0.2150E+01,0.2170E+01,0.2190E+01,0.2220E+01,0.2240E+01,0.2245E+01,&
           0.2250E+01,0.2260E+01,0.2270E+01,0.2290E+01,0.2310E+01,0.2330E+01,&
           0.2350E+01,0.2370E+01,0.2390E+01,0.2410E+01,0.2430E+01,0.2460E+01,&
           0.2500E+01,0.2520E+01,0.2550E+01,0.2565E+01,0.2580E+01,0.2590E+01,&
           0.2600E+01,0.2620E+01,0.2675E+01,0.2725E+01,0.2778E+01,0.2817E+01,&
           0.2833E+01,0.2849E+01,0.2865E+01,0.2882E+01,0.2899E+01,0.2915E+01,&
           0.2933E+01,0.2950E+01,0.2967E+01,0.2985E+01,0.3003E+01,0.3021E+01,&
           0.3040E+01,0.3058E+01,0.3077E+01,0.3096E+01,0.3115E+01,0.3135E+01,&
           0.3155E+01,0.3175E+01,0.3195E+01,0.3215E+01,0.3236E+01,0.3257E+01,&
           0.3279E+01,0.3300E+01,0.3322E+01,0.3345E+01,0.3367E+01,0.3390E+01,&
           0.3413E+01,0.3436E+01,0.3460E+01,0.3484E+01,0.3509E+01,0.3534E+01,&
           0.3559E+01,0.3624E+01,0.3732E+01,0.3775E+01,0.3847E+01,0.3969E+01,&
           0.4099E+01,0.4239E+01,0.4348E+01,0.4387E+01,0.4444E+01,0.4505E+01,&
           0.4547E+01,0.4560E+01,0.4580E+01,0.4719E+01,0.4904E+01,0.5000E+01,&
           0.5100E+01,0.5200E+01,0.5263E+01,0.5400E+01,0.5556E+01,0.5714E+01,&
           0.5747E+01,0.5780E+01,0.5814E+01,0.5848E+01,0.5882E+01,0.6061E+01,&
           0.6135E+01,0.6250E+01,0.6289E+01,0.6329E+01,0.6369E+01,0.6410E+01/)
      WL(343:456) = (/ &
           0.6452E+01,0.6494E+01,0.6579E+01,0.6667E+01,0.6757E+01,0.6897E+01,&
           0.7042E+01,0.7143E+01,0.7246E+01,0.7353E+01,0.7463E+01,0.7576E+01,&
           0.7692E+01,0.7812E+01,0.7937E+01,0.8065E+01,0.8197E+01,0.8333E+01,&
           0.8475E+01,0.8696E+01,0.8929E+01,0.9091E+01,0.9259E+01,0.9524E+01,&
           0.9804E+01,0.1000E+02,0.1020E+02,0.1031E+02,0.1042E+02,0.1053E+02,&
           0.1064E+02,0.1075E+02,0.1087E+02,0.1100E+02,0.1111E+02,0.1136E+02,&
           0.1163E+02,0.1190E+02,0.1220E+02,0.1250E+02,0.1282E+02,0.1299E+02,&
           0.1316E+02,0.1333E+02,0.1351E+02,0.1370E+02,0.1389E+02,0.1408E+02,&
           0.1429E+02,0.1471E+02,0.1515E+02,0.1538E+02,0.1563E+02,0.1613E+02,&
           0.1639E+02,0.1667E+02,0.1695E+02,0.1724E+02,0.1818E+02,0.1887E+02,&
           0.1923E+02,0.1961E+02,0.2000E+02,0.2041E+02,0.2083E+02,0.2222E+02,&
           0.2260E+02,0.2305E+02,0.2360E+02,0.2460E+02,0.2500E+02,0.2600E+02,&
           0.2857E+02,0.3100E+02,0.3333E+02,0.3448E+02,0.3564E+02,0.3700E+02,&
           0.3824E+02,0.3960E+02,0.4114E+02,0.4276E+02,0.4358E+02,0.4458E+02,&
           0.4550E+02,0.4615E+02,0.4671E+02,0.4736E+02,0.4800E+02,0.4878E+02,&
           0.5003E+02,0.5128E+02,0.5275E+02,0.5350E+02,0.5424E+02,0.5500E+02,&
           0.5574E+02,0.5640E+02,0.5700E+02,0.5746E+02,0.5840E+02,0.5929E+02,&
           0.6000E+02,0.6100E+02,0.6125E+02,0.6250E+02,0.6378E+02,0.6467E+02,&
           0.6558E+02,0.6655E+02,0.6760E+02,0.6900E+02,0.7053E+02,0.7300E+02/)
      WL(457:468) = (/ &
           0.7500E+02,0.7629E+02,0.8000E+02,0.8297E+02,0.8500E+02,0.8680E+02,&
           0.9080E+02,0.9517E+02,0.1000E+03,0.1200E+03,0.1500E+03,0.1670E+03/)
      WLT = (/ &
           0.1670E+03,0.1778E+03,0.1884E+03,&
           0.1995E+03,0.2113E+03,0.2239E+03,0.2371E+03,0.2512E+03,0.2661E+03,&
           0.2818E+03,0.2985E+03,0.3162E+03,0.3548E+03,0.3981E+03,0.4467E+03,&
           0.5012E+03,0.5623E+03,0.6310E+03,0.7943E+03,0.1000E+04,0.1259E+04,&
           0.2500E+04,0.5000E+04,0.1000E+05,0.2000E+05,0.3200E+05,0.3500E+05,&
           0.4000E+05,0.4500E+05,0.5000E+05,0.6000E+05,0.7000E+05,0.9000E+05,&
           0.1110E+06,0.1200E+06,0.1300E+06,0.1400E+06,0.1500E+06,0.1600E+06,&
           0.1700E+06,0.1800E+06,0.2000E+06,0.2500E+06,0.2900E+06,0.3200E+06,&
           0.3500E+06,0.3800E+06,0.4000E+06,0.4500E+06,0.5000E+06,0.6000E+06,&
           0.6400E+06,0.6800E+06,0.7200E+06,0.7600E+06,0.8000E+06,0.8400E+06,&
           0.9000E+06,0.1000E+07,0.2000E+07,0.5000E+07,0.8600E+07/)
      TABRE(1:114) = (/ &
           0.83441,   0.83676,   0.83729,   0.83771,   0.83827,   0.84038,&
           0.84719,   0.85522,   0.86047,   0.86248,   0.86157,   0.86093,&
           0.86419,   0.86916,   0.87764,   0.89296,   0.91041,   0.93089,&
           0.95373,   0.98188,   1.02334,   1.06735,   1.11197,   1.13134,&
           1.15747,   1.20045,   1.23840,   1.27325,   1.32157,   1.38958,&
           1.41644,   1.40906,   1.40063,   1.40169,   1.40934,   1.40221,&
           1.39240,   1.38424,   1.38075,   1.38186,   1.39634,   1.40918,&
           1.40256,   1.38013,   1.36303,   1.34144,   1.32377,   1.30605,&
           1.29054,   1.28890,   1.28931,   1.30190,   1.32025,   1.36302,&
           1.41872,   1.45834,   1.49028,   1.52128,   1.55376,   1.57782,&
           1.59636,   1.60652,   1.61172,   1.61919,   1.62522,   1.63404,&
           1.63689,   1.63833,   1.63720,   1.63233,   1.62222,   1.58269,&
           1.55635,   1.52453,   1.50320,   1.48498,   1.47226,   1.45991,&
           1.45115,   1.44272,   1.43498,   1.43280,   1.42924,   1.42602,&
           1.42323,   1.42143,   1.41897,   1.41660,   1.41434,   1.41216,&
           1.41006,   1.40805,   1.40423,   1.40067,   1.38004,   1.35085,&
           1.33394,   1.32492,   1.31940,   1.31854,   1.31775,   1.31702,&
           1.31633,   1.31569,   1.31509,   1.31452,   1.31399,   1.31349,&
           1.31302,   1.31257,   1.31215,   1.31175,   1.31136,   1.31099/)
      TABRE(115:228) = (/ &
           1.31064,   1.31031,   1.30999,   1.30968,   1.30938,   1.30909,&
           1.30882,   1.30855,   1.30829,   1.30804,   1.30780,   1.30756,&
           1.30733,   1.30710,   1.30688,   1.30667,   1.30646,   1.30625,&
           1.30605,   1.30585,   1.30566,   1.30547,   1.30528,   1.30509,&
           1.30491,   1.30473,   1.30455,   1.30437,   1.30419,   1.30402,&
           1.30385,   1.30367,   1.30350,   1.30333,   1.30316,   1.30299,&
           1.30283,   1.30266,   1.30249,   1.30232,   1.30216,   1.30199,&
           1.30182,   1.30166,   1.30149,   1.30132,   1.30116,   1.30099,&
           1.30082,   1.30065,   1.30048,   1.30031,   1.30014,   1.29997,&
           1.29979,   1.29962,   1.29945,   1.29927,   1.29909,   1.29891,&
           1.29873,   1.29855,   1.29837,   1.29818,   1.29800,   1.29781,&
           1.29762,   1.29743,   1.29724,   1.29705,   1.29686,   1.29666,&
           1.29646,   1.29626,   1.29605,   1.29584,   1.29563,   1.29542,&
           1.29521,   1.29499,   1.29476,   1.29453,   1.29430,   1.29406,&
           1.29381,   1.29355,   1.29327,   1.29299,   1.29272,   1.29252,&
           1.29228,   1.29205,   1.29186,   1.29167,   1.29150,   1.29130,&
           1.29106,   1.29083,   1.29025,   1.28962,   1.28891,   1.28784,&
           1.28689,   1.28623,   1.28521,   1.28413,   1.28261,   1.28137,&
           1.28093,   1.28047,   1.28022,   1.27998,   1.27948,   1.27849/)
      TABRE(229:342) = (/ &
           1.27774,   1.27691,   1.27610,   1.27535,   1.27471,   1.27404,&
           1.27329,   1.27240,   1.27139,   1.27029,   1.26901,   1.26736,&
           1.26591,   1.26441,   1.26284,   1.26036,   1.25860,   1.25815,&
           1.25768,   1.25675,   1.25579,   1.25383,   1.25179,   1.24967,&
           1.24745,   1.24512,   1.24266,   1.24004,   1.23725,   1.23270,&
           1.22583,   1.22198,   1.21548,   1.21184,   1.20790,   1.20507,&
           1.20209,   1.19566,   1.17411,   1.14734,   1.10766,   1.06739,&
           1.04762,   1.02650,   1.00357,   0.98197,   0.96503,   0.95962,&
           0.97269,   0.99172,   1.00668,   1.02186,   1.04270,   1.07597,&
           1.12954,   1.21267,   1.32509,   1.42599,   1.49656,   1.55095,&
           1.59988,   1.63631,   1.65024,   1.64278,   1.62691,   1.61284,&
           1.59245,   1.57329,   1.55770,   1.54129,   1.52654,   1.51139,&
           1.49725,   1.48453,   1.47209,   1.46125,   1.45132,   1.44215,&
           1.43366,   1.41553,   1.39417,   1.38732,   1.37735,   1.36448,&
           1.35414,   1.34456,   1.33882,   1.33807,   1.33847,   1.34053,&
           1.34287,   1.34418,   1.34634,   1.34422,   1.33453,   1.32897,&
           1.32333,   1.31800,   1.31432,   1.30623,   1.29722,   1.28898,&
           1.28730,   1.28603,   1.28509,   1.28535,   1.28813,   1.30156,&
           1.30901,   1.31720,   1.31893,   1.32039,   1.32201,   1.32239/)
      TABRE(343:456) = (/ &
           1.32149,   1.32036,   1.31814,   1.31705,   1.31807,   1.31953,&
           1.31933,   1.31896,   1.31909,   1.31796,   1.31631,   1.31542,&
           1.31540,   1.31552,   1.31455,   1.31193,   1.30677,   1.29934,&
           1.29253,   1.28389,   1.27401,   1.26724,   1.25990,   1.24510,&
           1.22241,   1.19913,   1.17150,   1.15528,   1.13700,   1.11808,&
           1.10134,   1.09083,   1.08734,   1.09254,   1.10654,   1.14779,&
           1.20202,   1.25825,   1.32305,   1.38574,   1.44478,   1.47170,&
           1.49619,   1.51652,   1.53328,   1.54900,   1.56276,   1.57317,&
           1.58028,   1.57918,   1.56672,   1.55869,   1.55081,   1.53807,&
           1.53296,   1.53220,   1.53340,   1.53289,   1.51705,   1.50097,&
           1.49681,   1.49928,   1.50153,   1.49856,   1.49053,   1.46070,&
           1.45182,   1.44223,   1.43158,   1.41385,   1.40676,   1.38955,&
           1.34894,   1.31039,   1.26420,   1.23656,   1.21663,   1.20233,&
           1.19640,   1.19969,   1.20860,   1.22173,   1.24166,   1.28175,&
           1.32784,   1.38657,   1.46486,   1.55323,   1.60379,   1.61877,&
           1.62963,   1.65712,   1.69810,   1.72065,   1.74865,   1.76736,&
           1.76476,   1.75011,   1.72327,   1.68490,   1.62398,   1.59596,&
           1.58514,   1.59917,   1.61405,   1.66625,   1.70663,   1.73713,&
           1.76860,   1.80343,   1.83296,   1.85682,   1.87411,   1.89110/)
      TABRE(457:468) = (/ &
           1.89918,   1.90432,   1.90329,   1.88744,   1.87499,   1.86702,&
           1.85361,   1.84250,   1.83225,   1.81914,   1.82268,   1.82961/)
      TABRET(1:NWLT,1) = (/ &
           1.82961,   1.83258,   1.83149,&
           1.82748,   1.82224,   1.81718,   1.81204,   1.80704,   1.80250,&
           1.79834,   1.79482,   1.79214,   1.78843,   1.78601,   1.78434,&
           1.78322,   1.78248,   1.78201,   1.78170,   1.78160,   1.78190,&
           1.78300,   1.78430,   1.78520,   1.78620,   1.78660,   1.78680,&
           1.78690,   1.78700,   1.78700,   1.78710,   1.78710,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78720,   1.78720,&
           1.78720,   1.78720,   1.78720,   1.78720,   1.78800/)
      TABRET(1:NWLT,2) = (/ &
           1.82961,   1.83258,   1.83149,   1.82748,&
           1.82224,   1.81718,   1.81204,   1.80704,   1.80250,   1.79834,&
           1.79482,   1.79214,   1.78843,   1.78601,   1.78434,   1.78322,&
           1.78248,   1.78201,   1.78170,   1.78160,   1.78190,   1.78300,&
           1.78430,   1.78520,   1.78610,   1.78630,   1.78640,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78650,   1.78650,   1.78650,&
           1.78650,   1.78650,   1.78650,   1.78720/)
      TABRET(1:NWLT,3) = (/ &
           1.82961,   1.83258,   1.83149,   1.82748,   1.82224,&
           1.81718,   1.81204,   1.80704,   1.80250,   1.79834,   1.79482,&
           1.79214,   1.78843,   1.78601,   1.78434,   1.78322,   1.78248,&
           1.78201,   1.78160,   1.78140,   1.78160,   1.78220,   1.78310,&
           1.78380,   1.78390,   1.78400,   1.78400,   1.78400,   1.78400,&
           1.78400,   1.78390,   1.78380,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78370,   1.78370,   1.78370,   1.78370,   1.78370,&
           1.78370,   1.78400,   1.78450/)
      TABRET(1:NWLT,4) = (/ &
           1.82961,   1.83258,   1.83149,   1.82748,   1.82224,   1.81718,&
           1.81204,   1.80704,   1.80250,   1.79834,   1.79482,   1.79214,&
           1.78843,   1.78601,   1.78434,   1.78322,   1.78248,   1.78201,&
           1.78150,   1.78070,   1.78010,   1.77890,   1.77790,   1.77730,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77720,   1.77720,   1.77720,   1.77720,   1.77720,&
           1.77720,   1.77800/)
      TABIM(1:114)  = (/ &
           0.1640E+00,0.1730E+00,0.1830E+00,0.1950E+00,0.2080E+00,0.2230E+00,&
           0.2400E+00,0.2500E+00,0.2590E+00,0.2680E+00,0.2790E+00,0.2970E+00,&
           0.3190E+00,0.3400E+00,0.3660E+00,0.3920E+00,0.4160E+00,0.4400E+00,&
           0.4640E+00,0.4920E+00,0.5170E+00,0.5280E+00,0.5330E+00,0.5340E+00,&
           0.5310E+00,0.5240E+00,0.5100E+00,0.5000E+00,0.4990E+00,0.4680E+00,&
           0.3800E+00,0.3600E+00,0.3390E+00,0.3180E+00,0.2910E+00,0.2510E+00,&
           0.2440E+00,0.2390E+00,0.2390E+00,0.2440E+00,0.2470E+00,0.2240E+00,&
           0.1950E+00,0.1740E+00,0.1720E+00,0.1800E+00,0.1940E+00,0.2130E+00,&
           0.2430E+00,0.2710E+00,0.2890E+00,0.3340E+00,0.3440E+00,0.3820E+00,&
           0.4010E+00,0.4065E+00,0.4050E+00,0.3890E+00,0.3770E+00,0.3450E+00,&
           0.3320E+00,0.3150E+00,0.2980E+00,0.2740E+00,0.2280E+00,0.1980E+00,&
           0.1720E+00,0.1560E+00,0.1100E+00,0.8300E-01,0.5800E-01,0.2200E-01,&
           0.1000E-01,0.3000E-02,0.1000E-02,0.3000E-03,0.1000E-03,0.3000E-04,&
           0.1000E-04,0.3000E-05,0.1000E-05,0.7000E-06,0.4000E-06,0.2000E-06,&
           0.1000E-06,0.6377E-07,0.3750E-07,0.2800E-07,0.2400E-07,0.2200E-07,&
           0.1900E-07,0.1750E-07,0.1640E-07,0.1590E-07,0.1325E-07,0.8623E-08,&
           0.5504E-08,0.3765E-08,0.2710E-08,0.2510E-08,0.2260E-08,0.2080E-08,&
           0.1910E-08,0.1540E-08,0.1530E-08,0.1550E-08,0.1640E-08,0.1780E-08,&
           0.1910E-08,0.2140E-08,0.2260E-08,0.2540E-08,0.2930E-08,0.3110E-08/)
      TABIM(115:228)  = (/ &
           0.3290E-08,0.3520E-08,0.4040E-08,0.4880E-08,0.5730E-08,0.6890E-08,&
           0.8580E-08,0.1040E-07,0.1220E-07,0.1430E-07,0.1660E-07,0.1890E-07,&
           0.2090E-07,0.2400E-07,0.2900E-07,0.3440E-07,0.4030E-07,0.4300E-07,&
           0.4920E-07,0.5870E-07,0.7080E-07,0.8580E-07,0.1020E-06,0.1180E-06,&
           0.1340E-06,0.1400E-06,0.1430E-06,0.1450E-06,0.1510E-06,0.1830E-06,&
           0.2150E-06,0.2650E-06,0.3350E-06,0.3920E-06,0.4200E-06,0.4440E-06,&
           0.4740E-06,0.5110E-06,0.5530E-06,0.6020E-06,0.7550E-06,0.9260E-06,&
           0.1120E-05,0.1330E-05,0.1620E-05,0.2000E-05,0.2250E-05,0.2330E-05,&
           0.2330E-05,0.2170E-05,0.1960E-05,0.1810E-05,0.1740E-05,0.1730E-05,&
           0.1700E-05,0.1760E-05,0.1820E-05,0.2040E-05,0.2250E-05,0.2290E-05,&
           0.3040E-05,0.3840E-05,0.4770E-05,0.5760E-05,0.6710E-05,0.8660E-05,&
           0.1020E-04,0.1130E-04,0.1220E-04,0.1290E-04,0.1320E-04,0.1350E-04,&
           0.1330E-04,0.1320E-04,0.1320E-04,0.1310E-04,0.1320E-04,0.1320E-04,&
           0.1340E-04,0.1390E-04,0.1420E-04,0.1480E-04,0.1580E-04,0.1740E-04,&
           0.1980E-04,0.2500E-04,0.5400E-04,0.1040E-03,0.2030E-03,0.2708E-03,&
           0.3511E-03,0.4299E-03,0.5181E-03,0.5855E-03,0.5899E-03,0.5635E-03,&
           0.5480E-03,0.5266E-03,0.4394E-03,0.3701E-03,0.3372E-03,0.2410E-03,&
           0.1890E-03,0.1660E-03,0.1450E-03,0.1280E-03,0.1030E-03,0.8600E-04,&
           0.8220E-04,0.8030E-04,0.8500E-04,0.9900E-04,0.1500E-03,0.2950E-03/)
      TABIM(229:342)  = (/ &
           0.4687E-03,0.7615E-03,0.1010E-02,0.1313E-02,0.1539E-02,0.1588E-02,&
           0.1540E-02,0.1412E-02,0.1244E-02,0.1068E-02,0.8414E-03,0.5650E-03,&
           0.4320E-03,0.3500E-03,0.2870E-03,0.2210E-03,0.2030E-03,0.2010E-03,&
           0.2030E-03,0.2140E-03,0.2320E-03,0.2890E-03,0.3810E-03,0.4620E-03,&
           0.5480E-03,0.6180E-03,0.6800E-03,0.7300E-03,0.7820E-03,0.8480E-03,&
           0.9250E-03,0.9200E-03,0.8920E-03,0.8700E-03,0.8900E-03,0.9300E-03,&
           0.1010E-02,0.1350E-02,0.3420E-02,0.7920E-02,0.2000E-01,0.3800E-01,&
           0.5200E-01,0.6800E-01,0.9230E-01,0.1270E+00,0.1690E+00,0.2210E+00,&
           0.2760E+00,0.3120E+00,0.3470E+00,0.3880E+00,0.4380E+00,0.4930E+00,&
           0.5540E+00,0.6120E+00,0.6250E+00,0.5930E+00,0.5390E+00,0.4910E+00,&
           0.4380E+00,0.3720E+00,0.3000E+00,0.2380E+00,0.1930E+00,0.1580E+00,&
           0.1210E+00,0.1030E+00,0.8360E-01,0.6680E-01,0.5400E-01,0.4220E-01,&
           0.3420E-01,0.2740E-01,0.2200E-01,0.1860E-01,0.1520E-01,0.1260E-01,&
           0.1060E-01,0.8020E-02,0.6850E-02,0.6600E-02,0.6960E-02,0.9160E-02,&
           0.1110E-01,0.1450E-01,0.2000E-01,0.2300E-01,0.2600E-01,0.2900E-01,&
           0.2930E-01,0.3000E-01,0.2850E-01,0.1730E-01,0.1290E-01,0.1200E-01,&
           0.1250E-01,0.1340E-01,0.1400E-01,0.1750E-01,0.2400E-01,0.3500E-01,&
           0.3800E-01,0.4200E-01,0.4600E-01,0.5200E-01,0.5700E-01,0.6900E-01,&
           0.7000E-01,0.6700E-01,0.6500E-01,0.6400E-01,0.6200E-01,0.5900E-01/)
      TABIM(343:456)  = (/ &
           0.5700E-01,0.5600E-01,0.5500E-01,0.5700E-01,0.5800E-01,0.5700E-01,&
           0.5500E-01,0.5500E-01,0.5400E-01,0.5200E-01,0.5200E-01,0.5200E-01,&
           0.5200E-01,0.5000E-01,0.4700E-01,0.4300E-01,0.3900E-01,0.3700E-01,&
           0.3900E-01,0.4000E-01,0.4200E-01,0.4400E-01,0.4500E-01,0.4600E-01,&
           0.4700E-01,0.5100E-01,0.6500E-01,0.7500E-01,0.8800E-01,0.1080E+00,&
           0.1340E+00,0.1680E+00,0.2040E+00,0.2480E+00,0.2800E+00,0.3410E+00,&
           0.3790E+00,0.4090E+00,0.4220E+00,0.4220E+00,0.4030E+00,0.3890E+00,&
           0.3740E+00,0.3540E+00,0.3350E+00,0.3150E+00,0.2940E+00,0.2710E+00,&
           0.2460E+00,0.1980E+00,0.1640E+00,0.1520E+00,0.1420E+00,0.1280E+00,&
           0.1250E+00,0.1230E+00,0.1160E+00,0.1070E+00,0.7900E-01,0.7200E-01,&
           0.7600E-01,0.7500E-01,0.6700E-01,0.5500E-01,0.4500E-01,0.2900E-01,&
           0.2750E-01,0.2700E-01,0.2730E-01,0.2890E-01,0.3000E-01,0.3400E-01,&
           0.5300E-01,0.7550E-01,0.1060E+00,0.1350E+00,0.1761E+00,0.2229E+00,&
           0.2746E+00,0.3280E+00,0.3906E+00,0.4642E+00,0.5247E+00,0.5731E+00,&
           0.6362E+00,0.6839E+00,0.7091E+00,0.6790E+00,0.6250E+00,0.5654E+00,&
           0.5433E+00,0.5292E+00,0.5070E+00,0.4883E+00,0.4707E+00,0.4203E+00,&
           0.3771E+00,0.3376E+00,0.3056E+00,0.2835E+00,0.3170E+00,0.3517E+00,&
           0.3902E+00,0.4509E+00,0.4671E+00,0.4779E+00,0.4890E+00,0.4899E+00,&
           0.4873E+00,0.4766E+00,0.4508E+00,0.4193E+00,0.3880E+00,0.3433E+00/)
      TABIM(457:468)  = (/ &
           0.3118E+00,0.2935E+00,0.2350E+00,0.1981E+00,0.1865E+00,0.1771E+00,&
           0.1620E+00,0.1490E+00,0.1390E+00,0.1200E+00,0.9620E-01,0.8300E-01/)


      TABIMT(1:NWLT,1)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,&
           0.4560E-01,0.3790E-01,0.3140E-01,0.2620E-01,0.2240E-01,0.1960E-01,&
           0.1760E-01,0.1665E-01,0.1620E-01,0.1550E-01,0.1470E-01,0.1390E-01,&
           0.1320E-01,0.1250E-01,0.1180E-01,0.1060E-01,0.9540E-02,0.8560E-02,&
           0.6210E-02,0.4490E-02,0.3240E-02,0.2340E-02,0.1880E-02,0.1740E-02,&
           0.1500E-02,0.1320E-02,0.1160E-02,0.8800E-03,0.6950E-03,0.4640E-03,&
           0.3400E-03,0.3110E-03,0.2940E-03,0.2790E-03,0.2700E-03,0.2640E-03,&
           0.2580E-03,0.2520E-03,0.2490E-03,0.2540E-03,0.2640E-03,0.2740E-03,&
           0.2890E-03,0.3050E-03,0.3150E-03,0.3460E-03,0.3820E-03,0.4620E-03,&
           0.5000E-03,0.5500E-03,0.5950E-03,0.6470E-03,0.6920E-03,0.7420E-03,&
           0.8200E-03,0.9700E-03,0.1950E-02,0.5780E-02,0.9700E-02/)
      TABIMT(1:NWLT,2)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,0.4560E-01,&
           0.3790E-01,0.3140E-01,0.2620E-01,0.2240E-01,0.1960E-01,0.1760E-01,&
           0.1665E-01,0.1600E-01,0.1500E-01,0.1400E-01,0.1310E-01,0.1230E-01,&
           0.1150E-01,0.1080E-01,0.9460E-02,0.8290E-02,0.7270E-02,0.4910E-02,&
           0.3300E-02,0.2220E-02,0.1490E-02,0.1140E-02,0.1060E-02,0.9480E-03,&
           0.8500E-03,0.7660E-03,0.6300E-03,0.5200E-03,0.3840E-03,0.2960E-03,&
           0.2700E-03,0.2520E-03,0.2440E-03,0.2360E-03,0.2300E-03,0.2280E-03,&
           0.2250E-03,0.2200E-03,0.2160E-03,0.2170E-03,0.2200E-03,0.2250E-03,&
           0.2320E-03,0.2390E-03,0.2600E-03,0.2860E-03,0.3560E-03,0.3830E-03,&
           0.4150E-03,0.4450E-03,0.4760E-03,0.5080E-03,0.5400E-03,0.5860E-03,&
           0.6780E-03,0.1280E-02,0.3550E-02,0.5600E-02/)
      TABIMT(1:NWLT,3)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,0.4560E-01,0.3790E-01,&
           0.3140E-01,0.2620E-01,0.2190E-01,0.1880E-01,0.1660E-01,0.1540E-01,&
           0.1470E-01,0.1350E-01,0.1250E-01,0.1150E-01,0.1060E-01,0.9770E-02,&
           0.9010E-02,0.7660E-02,0.6520E-02,0.5540E-02,0.3420E-02,0.2100E-02,&
           0.1290E-02,0.7930E-03,0.5700E-03,0.5350E-03,0.4820E-03,0.4380E-03,&
           0.4080E-03,0.3500E-03,0.3200E-03,0.2550E-03,0.2120E-03,0.2000E-03,&
           0.1860E-03,0.1750E-03,0.1660E-03,0.1560E-03,0.1490E-03,0.1440E-03,&
           0.1350E-03,0.1210E-03,0.1160E-03,0.1160E-03,0.1170E-03,0.1200E-03,&
           0.1230E-03,0.1320E-03,0.1440E-03,0.1680E-03,0.1800E-03,0.1900E-03,&
           0.2090E-03,0.2160E-03,0.2290E-03,0.2400E-03,0.2600E-03,0.2920E-03,&
           0.6100E-03,0.1020E-02,0.1810E-02/)
      TABIMT(1:NWLT,4)  = (/ &
           0.8300E-01,0.6900E-01,0.5700E-01,0.4450E-01,0.3550E-01,0.2910E-01,&
           0.2440E-01,0.1970E-01,0.1670E-01,0.1400E-01,0.1235E-01,0.1080E-01,&
           0.8900E-02,0.7340E-02,0.6400E-02,0.5600E-02,0.5000E-02,0.4520E-02,&
           0.3680E-02,0.2990E-02,0.2490E-02,0.1550E-02,0.9610E-03,0.5950E-03,&
           0.3690E-03,0.2670E-03,0.2510E-03,0.2290E-03,0.2110E-03,0.1960E-03,&
           0.1730E-03,0.1550E-03,0.1310E-03,0.1130E-03,0.1060E-03,0.9900E-04,&
           0.9300E-04,0.8730E-04,0.8300E-04,0.7870E-04,0.7500E-04,0.6830E-04,&
           0.5600E-04,0.4960E-04,0.4550E-04,0.4210E-04,0.3910E-04,0.3760E-04,&
           0.3400E-04,0.3100E-04,0.2640E-04,0.2510E-04,0.2430E-04,0.2390E-04,&
           0.2370E-04,0.2380E-04,0.2400E-04,0.2460E-04,0.2660E-04,0.4450E-04,&
           0.8700E-04,0.1320E-03/)

      !     ZERO PARAMETERS

      RN=0.0
      CN=0.0
      ABSIND=0.0
      ABSCOF=0.0

      !     CONVERT WAVELENGTH TO MICRONS

      ALAM=XLAM
      IF (IUNIT.EQ.1) ALAM=1000*ALAM
      IF (IUNIT.EQ.2) ALAM=10000*ALAM
      IF (IUNIT.EQ.3) ALAM=10000*(1.0/ALAM)
      IF (ALAM.LT.WLMIN.OR.ALAM.GT.WLMAX) RETURN
      IF (ALAM.LE.CUTICE) THEN
        !     REGION FROM 0.045 MICRONS TO 167.0 MICRONS - NO TEMPERATURE DEPEND
        DO I=2,NWL
          IF (ALAM.LT.WL(I)) EXIT
        END DO
        X1=LOG(WL(I-1))
        X2=LOG(WL(I))
        Y1=TABRE(I-1)
        Y2=TABRE(I)
        X=LOG(ALAM)
        Y=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
        RN=Y
        Y1=LOG(ABS(TABIM(I-1)))
        Y2=LOG(ABS(TABIM(I)))
        Y=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
        CN=EXP(Y)
      ELSE
        !     REGION FROM 167.0 MICRONS TO 8.6 METERS - TEMPERATURE DEPENDENCE

        TK=T
        IF (TK.GT.TEMREF(1)) TK=TEMREF(1)
        IF (TK.LT.TEMREF(4)) TK=TEMREF(4)
        DO I=2,4
          IF(TK.GE.TEMREF(I)) EXIT
        END DO
        LT1=I
        LT2=I-1
        DO I=2,NWLT
          IF (ALAM.LE.WLT(I)) EXIT
        END DO
        X1=LOG(WLT(I-1))
        X2=LOG(WLT(I))
        Y1=TABRET(I-1,LT1)
        Y2=TABRET(I,LT1)
        X=LOG(ALAM)
        YLO=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
        Y1=TABRET(I-1,LT2)
        Y2=TABRET(I,LT2)
        YHI=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
        T1=TEMREF(LT1)
        T2=TEMREF(LT2)
        Y=((TK-T1)*(YHI-YLO)/(T2-T1))+YLO
        RN=Y
        Y1=LOG(ABS(TABIMT(I-1,LT1)))
        Y2=LOG(ABS(TABIMT(I,LT1)))
        YLO=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
        Y1=LOG(ABS(TABIMT(I-1,LT2)))
        Y2=LOG(ABS(TABIMT(I,LT2)))
        YHI=((X-X1)*(Y2-Y1)/(X2-X1))+Y1
        Y=((TK-T1)*(YHI-YLO)/(T2-T1))+YLO
        CN=EXP(Y)
      END IF

      !     ABSORPTIVE QUANITIES
      ABSIND=CN/RN
      ABSCOF=4.0*PI*CN/ALAM
      RETURN

    END SUBROUTINE REFICE

  END FUNCTION m_complex_ice_warren


!*********************************************************************************************

  !----------------------------------------------------------------------------------------!
  ! 2) Routinen zum effektiven Brechungsindex von Mischmaterialien nach verschiedenen
  !    Ansaetzen (es werden jeweils immer 3 Mischungspartner beruecksichtigt).
  !----------------------------------------------------------------------------------------!

  !----------------------------------------------------------------------------------------!
  ! 2a) Homogene Mischung 3er Substanzen nach Oguchi (1983), 
  !     erwaehnt in Wiener (1912), zurueckgehend auf Mosotti, Clausius, Poisson
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_oguchi_vec(vol1, vol2, vol3, m1, m2, m3, anz, fehler)

    ! vol1, vol2, vol3 sind die Volumenanteile der drei Substanzen mit dem
    ! komplexen Brechungsindex m1, m2, m3 an der homogenen Mischung (am Gesamtvolumen).
    ! Hier wird vol1 bzw. m1 als Matrix und die anderen beiden als
    ! kugelfoermige mikroskopische Einschluesse (formfakt = 2.0) behandelt.

    ! Formel ist symetrisch in der Wahl, welcher Stoff die Matrix und
    ! welche beiden die Einschluesse sind.

    IMPLICIT NONE
    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_oguchi_vec(anz)
    COMPLEX(kind=kinddp) :: m1(anz), m2(anz), m3(anz)
    DOUBLE PRECISION :: vol1(anz), vol2(anz), vol3(anz)
    COMPLEX(kind=kinddp) :: m1t(anz), m2t(anz), m3t(anz)
    DOUBLE PRECISION :: formfakt
    INTEGER, INTENT(out) :: fehler

    fehler = 0

    IF (ANY(ABS(vol1+vol2+vol3-1.0d0) > 1d-6)) THEN
      WRITE (*,*) 'M_COMPLEX_OGUCHI_VEC: Summe der rel. Partialvolumen ergibt nicht 1! Abbruch!'
      m_complex_oguchi_vec = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    formfakt = 2.0d0  ! Einschluesse kugelfoermig

    m1t = m1**2
    m2t = m2**2
    m3t = m3**2

    m_complex_oguchi_vec = SQRT((m1t*vol1+(m1t+formfakt)/(m2t+formfakt)*m2t*vol2+(m1t+formfakt)/&
         (m3t+formfakt)*m3t*vol3) / &
         (vol1+(m1t+formfakt)/(m2t+formfakt)*vol2+(m1t+formfakt)/(m3t+formfakt)*vol3))
    ! Die komplexe Wurzelfunktion liefert diejenige Wurzel mit
    ! positivem Realteil, was in diesem Falle richtig ist. Das
    ! Vorzeichen des Imagin"arteils bleibt dann in der Konvention der
    ! Brechungsindices m1, m2 und m3. 
    RETURN

  END FUNCTION m_complex_oguchi_vec

  FUNCTION m_complex_oguchi(vol1, vol2, vol3, m1, m2, m3, fehler)

    ! vol1, vol2, vol3 sind die Volumenanteile der drei Substanzen mit dem
    ! komplexen Brechungsindex m1, m2, m3 an der homogenen Mischung (am Gesamtvolumen).
    ! Hier wird vol1 bzw. m1 als Matrix und die anderen beiden als
    ! kugelfoermige mikroskopische Einschluesse (formfakt = 2.0) behandelt.

    ! Formel ist symetrisch in der Wahl, welcher Stoff die Matrix und
    ! welche beiden die Einschluesse sind.

    IMPLICIT NONE
    COMPLEX(kind=kinddp) :: m_complex_oguchi
    COMPLEX(kind=kinddp) :: m1, m2, m3
    DOUBLE PRECISION :: vol1, vol2, vol3
    COMPLEX(kind=kinddp) :: m1t, m2t, m3t
    DOUBLE PRECISION :: formfakt
    INTEGER, INTENT(out) :: fehler

    fehler = 0

    IF (ABS(vol1+vol2+vol3-1.0d0) > 1d-6) THEN
      WRITE (*,*) 'M_COMPLEX_OGUCHI: Summe der rel. Partialvolumen ergibt nicht 1! Abbruch!'
      m_complex_oguchi = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    formfakt = 2.0d0  ! Einschluesse kugelfoermig

    m1t = m1**2
    m2t = m2**2
    m3t = m3**2

    m_complex_oguchi = SQRT((m1t*vol1+(m1t+formfakt)/(m2t+formfakt)*m2t*vol2+(m1t+formfakt)/(m3t+formfakt)*m3t*vol3) / &
         (vol1+(m1t+formfakt)/(m2t+formfakt)*vol2+(m1t+formfakt)/(m3t+formfakt)*vol3))
    ! Die komplexe Wurzelfunktion liefert diejenige Wurzel mit
    ! positivem Realteil, was in diesem Falle richtig ist. Das
    ! Vorzeichen des Imagin"arteils bleibt dann in der Konvention der
    ! Brechungsindices m1, m2 und m3. 
    RETURN

  END FUNCTION m_complex_oguchi

  !----------------------------------------------------------------------------------------!
  ! 2b) Homogene Mischung 3er Substanzen nach Maxwell-Garnett (1904), 
  !     entnommen Bohren und Huffman (1983)
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_maxwellgarnett_vec(vol1, vol2, vol3, m1, m2, m3, inclusionstring, anz, fehler)

    ! vol1, vol2, vol3 sind die Volumenanteile der drei Substanzen mit dem
    ! komplexen Brechungsindex m1, m2, m3 an der homogenen Mischung (am Gesamtvolumen).
    ! Hier wird vol1 bzw. m1 als Matrix und die anderen beiden als
    ! kugelfoermige bzw. ellipsoide mikroskopische Einschluesse behandelt.

    ! Der Stoff vol1 ist die Matrix, vol2 und vol3 sind die
    ! Einschluesse.

    ! Der inclusionstring legt fest, wie die inclusion zu behandeln
    ! ist: 'shperical' oder 'spheroidal'

    ! Diese Formel ist nicht symetrisch bezueglich der Wahl der
    ! Matrix. Ist jedoch die Matrix festgelegt, koennen die anderen
    ! beiden Stoffe (kugelf"ormige Einschluesse) beliebig vertauscht werden.

    ! Bohren, Huffman 1983; Bohren 1982

    ! Ergebnisse eines Profilings mit gprof: Die Auswertung des komplexen Logarithmus ist bei
    ! weitem die aufwendigste Funktion im untenstehenden Code. Fast 2/3 der gesamten fuer
    ! die Funktion benoetigten Rechenzeit geht hierfuer drauf. Danach kommt die 
    ! komplexe Wurzelfunktion. Die select-Anweisung spielt nur eine sehr untergeordnete
    ! Rolle.

    IMPLICIT NONE
    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_maxwellgarnett_vec(anz)
    COMPLEX(kind=kinddp) :: m1(anz), m2(anz), m3(anz)
    DOUBLE PRECISION :: vol1(anz), vol2(anz), vol3(anz)
    CHARACTER(len=*) :: inclusionstring

    COMPLEX(kind=kinddp) :: beta2(anz), beta3(anz), m1t(anz), m2t(anz), m3t(anz)
    INTEGER, INTENT(out) :: fehler
    CHARACTER(len=20) :: csi

    fehler = 0

    IF (ANY(ABS(vol1+vol2+vol3-1.0d0) > 1d-6)) THEN
      WRITE (*,*) 'M_COMPLEX_MAXWELLGARNETT_VEC: Summe der rel. Partialvolumen ergibt nicht 1! Abbruch!'
      m_complex_maxwellgarnett_vec = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    m1t = m1**2
    m2t = m2**2
    m3t = m3**2

    csi(:) = ' '
    csi = inclusionstring(1:LEN_TRIM(inclusionstring))
    csi = ADJUSTL(csi)

!!$ here, "TRIM()" obstructs inlining, therefore changed to something which does inline:
!!$    SELECT CASE (TRIM(ADJUSTL(inclusionstring)))
    SELECT CASE (csi(1:10))
!!$    CASE ('spherical','SPHERICAL','Spherical')
    CASE ('spherical ','SPHERICAL ','Spherical ')
      beta2 = 3.0d0*m1t/(m2t+2.0d0*m1t)
      beta3 = 3.0d0*m1t/(m3t+2.0d0*m1t)
    CASE ('spheroidal','SPHEROIDAL','Spheroidal')
      ! Die LOG-Funktion liefert den Hauptzweig
      ! der Komplexen Funktion, wie gefordert:
      beta2 = 2.0d0*m1t/(m2t-m1t) * (m2t/(m2t-m1t)*LOG(m2t/m1t)-1.0d0)
      beta3 = 2.0d0*m1t/(m3t-m1t) * (m3t/(m3t-m1t)*LOG(m3t/m1t)-1.0d0)
    CASE DEFAULT
      WRITE (*,*) 'M_COMPLEX_MAXWELLGARNETT_VEC: Unbekannter inclusionstring: ', inclusionstring
      m_complex_maxwellgarnett_vec = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END SELECT
    
    m_complex_maxwellgarnett_vec = SQRT(((1.0d0-vol2-vol3)*m1t + vol2*beta2*m2t + vol3*beta3*m3t) / &
         (1.0d0-vol2-vol3+vol2*beta2+vol3*beta3))

    ! Die komplexe Wurzelfunktion liefert diejenige der beiden Wurzeln mit
    ! positivem Realteil, was in diesem Falle richtig ist. Das
    ! Vorzeichen des Imagin"arteils bleibt dann in der Konvention der
    ! Brechungsindices m1, m2 und m3. 
    RETURN
  END FUNCTION m_complex_maxwellgarnett_vec

  FUNCTION m_complex_maxwellgarnett(vol1, vol2, vol3, m1, m2, m3, inclusionstring, fehler)

    ! vol1, vol2, vol3 sind die Volumenanteile der drei Substanzen mit dem
    ! komplexen Brechungsindex m1, m2, m3 an der homogenen Mischung (am Gesamtvolumen).
    ! Hier wird vol1 bzw. m1 als Matrix und die anderen beiden als
    ! kugelfoermige bzw. ellipsoide mikroskopische Einschluesse behandelt.

    ! Der Stoff vol1 ist die Matrix, vol2 und vol3 sind die
    ! Einschluesse.

    ! Der inclusionstring legt fest, wie die inclusion zu behandeln
    ! ist: 'shperical' oder 'spheroidal'

    ! Diese Formel ist nicht symetrisch bezueglich der Wahl der
    ! Matrix. Ist jedoch die Matrix festgelegt, koennen die anderen
    ! beiden Stoffe (kugelf"ormige Einschluesse) beliebig vertauscht werden.

    ! Bohren, Huffman 1983; Bohren 1982

    ! Ergebnisse eines Profilings mit gprof: Die Auswertung des komplexen Logarithmus ist bei
    ! weitem die aufwendigste Funktion im untenstehenden Code. Fast 2/3 der gesamten fuer
    ! die Funktion benoetigten Rechenzeit geht hierfuer drauf. Danach kommt die 
    ! komplexe Wurzelfunktion. Die select-Anweisung spielt nur eine sehr untergeordnete
    ! Rolle.

    IMPLICIT NONE

    COMPLEX(kind=kinddp) :: m_complex_maxwellgarnett
    COMPLEX(kind=kinddp) :: m1, m2, m3
    DOUBLE PRECISION :: vol1, vol2, vol3
    CHARACTER(len=*) :: inclusionstring

    COMPLEX(kind=kinddp) :: beta2, beta3, m1t, m2t, m3t
    INTEGER, INTENT(out) :: fehler

    fehler = 0

    IF (ABS(vol1+vol2+vol3-1.0d0) > 1d-6) THEN
      WRITE (*,*) 'M_COMPLEX_MAXWELLGARNETT: Summe der rel. Partialvolumen ergibt nicht 1! Abbruch!'
      m_complex_maxwellgarnett = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    m1t = m1**2
    m2t = m2**2
    m3t = m3**2

    SELECT CASE (TRIM(ADJUSTL(inclusionstring)))
    CASE ('spherical','SPHERICAL','Spherical')
      beta2 = 3.0d0*m1t/(m2t+2.0d0*m1t)
      beta3 = 3.0d0*m1t/(m3t+2.0d0*m1t)
    CASE ('spheroidal','SPHEROIDAL','Spheroidal')
      ! Die LOG-Funktion liefert den Hauptzweig
      ! der Komplexen Funktion, wie gefordert:
      beta2 = 2.0d0*m1t/(m2t-m1t) * (m2t/(m2t-m1t)*LOG(m2t/m1t)-1.0d0)
      beta3 = 2.0d0*m1t/(m3t-m1t) * (m3t/(m3t-m1t)*LOG(m3t/m1t)-1.0d0)
    CASE DEFAULT
      WRITE (*,*) 'M_COMPLEX_MAXWELLGARNETT: Unbekannter inclusionstring: ', inclusionstring
      m_complex_maxwellgarnett = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END SELECT

    m_complex_maxwellgarnett = SQRT(((1.0d0-vol2-vol3)*m1t + vol2*beta2*m2t + vol3*beta3*m3t) / &
         (1.0d0-vol2-vol3+vol2*beta2+vol3*beta3))

    ! Die komplexe Wurzelfunktion liefert diejenige der beiden Wurzeln mit
    ! positivem Realteil, was in diesem Falle richtig ist. Das
    ! Vorzeichen des Imagin"arteils bleibt dann in der Konvention der
    ! Brechungsindices m1, m2 und m3. 
    RETURN
  END FUNCTION m_complex_maxwellgarnett

  !----------------------------------------------------------------------------------------!
  ! 2c) Homogene Mischung 3er Substanzen nach Bruggemann
  !----------------------------------------------------------------------------------------!

  FUNCTION m_complex_bruggemann_vec (volair, volice, volwater, mair, mice, mwater, anz, fehler)

    ! volwater, volice, volair sind die Volumenanteile von Wasser, Eis
    ! und Luft an der homogenen Mischung (am Gesamtvolumen).

    IMPLICIT NONE
    INTEGER :: anz
    COMPLEX(kind=kinddp) :: m_complex_bruggemann_vec(anz)
    DOUBLE PRECISION, INTENT(in) :: volair(anz), volice(anz), volwater(anz)
    COMPLEX(kind=kinddp), INTENT(in) :: mair(anz), mice(anz), mwater(anz)
    INTEGER, intent(out) :: fehler
    COMPLEX(kind=kinddp) :: m1t(anz), m2t(anz), m3t(anz), koeff(anz,4), roots(anz,3), sroots(anz,3), &
         p(anz), q(anz), u(anz), v(anz), diskri(anz), y(anz,3), epsilon1, epsilon2

    INTEGER :: i, j, count(anz)

    fehler = 0

    IF (ANY(ABS(volair+volice+volwater-1.0d0) > 1d-6)) THEN
      WRITE (*,*) 'M_COMPLEX_BRUGGEMANN_VEC: Summe der rel. Partialvolumen ergibt nicht 1! Abbruch!'
      m_complex_bruggemann_vec = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    m1t = mair**2
    m2t = mice**2
    m3t = mwater**2

    ! die m's sind jetzt die epsilons! 

    ! Die Bruggemann-Formel fuer N=3 auf ein Polynom 3. Grades
    ! umschreiben, von dem dann die 0-Stellen gesucht werden:
    ! hier kommen die Koeffizienten:

    koeff(:,1) = m1t*m2t*m3t
    koeff(:,2) = m1t*m2t*(2.*volair+2.*volice-volwater) + &
         m1t*m3t*(2.*volair-volice+2.*volwater) + &
         m2t*m3t*(-volair+2.*volice+2.*volwater)
    koeff(:,3) = m1t*(4.*volair-2.*volice-2.*volwater) + &
         m2t*(-2.*volair+4.*volice-2.*volwater) + &
         m3t*(-2.*volair-2.*volice+4.*volwater)
    koeff(:,4) = (-4.0d0,0.0d0)

    ! Laguerre-Verfahren zur iterativen Nullstellenbestimmung 
    ! (in dieser vektorisierten Fassung leider noch nicht implementiert...):
!    call zrootsvec(koeff,3,anz,roots,.true.)

    !=============================================================
    ! alternativ: Cardanische Formel (exakte Loesungen):

    ! Normalform des Polynoms:
    koeff(:,1) = koeff(:,1) / koeff(:,4)
    koeff(:,2) = koeff(:,2) / koeff(:,4)
    koeff(:,3) = koeff(:,3) / koeff(:,4)
    koeff(:,4) = koeff(:,4) / koeff(:,4)

    ! reduzierte Form:
    p = (3.0*koeff(:,2)-koeff(:,3)**2) / 3.0d0
    q = 2.0d0*koeff(:,3)**3/27.0d0 - koeff(:,2)*koeff(:,3)/3.0d0 + koeff(:,1)

    ! Cardanische Formel fuer Polynom 3. Grades:
    diskri = (p/3.0d0)**3 + (q/2.0d0)**2
    u = (-q/2.0d0+SQRT(diskri))**(1.d0/3.d0)
    v = -p/(3.0d0*u)

    epsilon1 = CMPLX(-0.5d0,SQRT(3.0d0)*0.5d0,kind=kinddp)
    epsilon2 = CMPLX(-0.5d0,-SQRT(3.0d0)*0.5d0,kind=kinddp)
    y(:,1) = u + v
    y(:,2) = epsilon1*u + epsilon2*v
    y(:,3) = epsilon2*u + epsilon1*v
  
    roots(:,1) = y(:,1) - koeff(:,3)/3.0d0
    roots(:,2) = y(:,2) - koeff(:,3)/3.0d0
    roots(:,3) = y(:,3) - koeff(:,3)/3.0d0
    !===============================================================

    sroots = SQRT(roots)

    ! herumspielen mit einem IDL-testdriver-Programm (darstellen der
    ! berechneten 3 sroots in der komplexen Zahlenebene) hat
    ! ergeben, dass die physikalisch sinnvolle Loesung einen Realteil
    ! hat, der zwischen 1.0 (Luft) und dem Wert fuer reines Wasser
    ! liegt. Der Imaginaerteil ist dann kleiner als 0.0 und kleiner als der Wert fuer
    ! Wasser. Mit diesen Bedingungen kann die richtige Wurzel
    ! absepariert werden, weil die anderen 2 nie in diesem Bereich liegen.

    ! Aus numerischen Gruenden nicht genau 1.0, sondern 0.98!
    count = 0
    DO i=1,3
      DO j=1,anz
        IF (REAL(sroots(j,i)) >= 0.98 .AND. AIMAG(sroots(j,i)) < 0.0) THEN 
          count(j) = count(j) + 1
          m_complex_bruggemann_vec(j) = sroots(j,i)
        END IF
      END DO
    END DO
    IF (ANY(COUNT == 0)) THEN
      WRITE (*,*) 'M_COMPLEX_BRUGGEMANN_VEC: Keine passende Loesung gefunden! Abbruch!'
      m_complex_bruggemann_vec = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF
    IF (ANY(COUNT > 1)) THEN
      WRITE (*,*) 'M_COMPLEX_BRUGGEMANN_VEC: Mehrere passende Loesungen gefunden! Da is was faul! Abbruch!'
      m_complex_bruggemann_vec = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    RETURN

  END FUNCTION m_complex_bruggemann_vec

  FUNCTION m_complex_bruggemann (volair, volice, volwater, mair, mice, mwater, fehler)

    ! volwater, volice, volair sind die Volumenanteile von Wasser, Eis
    ! und Luft an der homogenen Mischung (am Gesamtvolumen).

    IMPLICIT NONE
    COMPLEX(kind=kinddp) :: m_complex_bruggemann
    DOUBLE PRECISION, INTENT(in) :: volair, volice, volwater
    COMPLEX(kind=kinddp), INTENT(in) :: mair, mice, mwater
    INTEGER, intent(out) :: fehler
    COMPLEX(kind=kinddp) :: m1t, m2t, m3t, koeff(4), roots(3), sroots(3), mm, check, &
         p, q, u, v, diskri, y(3), epsilon1, epsilon2

    INTEGER :: i, count

    fehler = 0

    IF (ABS(volair+volice+volwater-1.0d0) > 1d-6) THEN
      WRITE (*,*) 'M_COMPLEX_BRUGGEMANN: Summe der rel. Partialvolumen ergibt nicht 1! Abbruch!'
      m_complex_bruggemann = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    m1t = mair**2
    m2t = mice**2
    m3t = mwater**2

    ! die m's sind jetzt die epsilons! 

    ! Die Bruggemann-Formel fuer N=3 auf ein Polynom 3. Grades
    ! umschreiben, von dem dann die 0-Stellen gesucht werden:
    ! hier kommen die Koeffizienten:

    koeff(1) = m1t*m2t*m3t
    koeff(2) = m1t*m2t*(2.*volair+2.*volice-volwater) + &
         m1t*m3t*(2.*volair-volice+2.*volwater) + &
         m2t*m3t*(-volair+2.*volice+2.*volwater)
    koeff(3) = m1t*(4.*volair-2.*volice-2.*volwater) + &
         m2t*(-2.*volair+4.*volice-2.*volwater) + &
         m3t*(-2.*volair-2.*volice+4.*volwater)
    koeff(4) = (-4.0d0,0.0d0)

    ! Laguerre-Verfahren zur iterativen Nullstellenbestimmung:
!    call zroots(koeff,3,roots,.true.)

    !=============================================================
    ! alternativ: Cardanische Formel (exakte Loesungen):

    ! Normalform des Polynoms:
    koeff = koeff / koeff(4)

    ! reduzierte Form:
    p = (3.0*koeff(2)-koeff(3)**2) / 3.0d0
    q = 2.0d0*koeff(3)**3/27.0d0 - koeff(2)*koeff(3)/3.0d0 + koeff(1)

    ! Cardanische Formel fuer Polynom 3. Grades:
    diskri = (p/3.0d0)**3 + (q/2.0d0)**2
    u = (-q/2.0d0+SQRT(diskri))**(1.d0/3.d0)
    v = -p/(3.0d0*u)

    epsilon1 = CMPLX(-0.5d0,SQRT(3.0d0)*0.5d0,kind=kinddp)
    epsilon2 = CMPLX(-0.5d0,-SQRT(3.0d0)*0.5d0,kind=kinddp)
    y(1) = u + v
    y(2) = epsilon1*u + epsilon2*v
    y(3) = epsilon2*u + epsilon1*v
  
    roots = y - koeff(3)/3.0d0
    !===============================================================

    sroots = SQRT(roots)

    ! herumspielen mit einem IDL-testdriver-Programm (darstellen der
    ! berechneten 3 sroots in der komplexen Zahlenebene) hat
    ! ergeben, dass die physikalisch sinnvolle Loesung einen Realteil
    ! hat, der zwischen 1.0 (Luft) und dem Wert fuer reines Wasser
    ! liegt. Der Imaginaerteil ist dann kleiner als 0.0 und kleiner als der Wert fuer
    ! Wasser. Mit diesen Bedingungen kann die richtige Wurzel
    ! absepariert werden, weil die anderen 2 nie in diesem Bereich liegen.

    ! Aus numerischen Gruenden nicht genau 1.0, sondern 0.98!
    count = 0
    DO i=1,3
      IF (REAL(sroots(i)) >= 0.98 .AND. AIMAG(sroots(i)) < 0.0) THEN 
        count = count + 1
        m_complex_bruggemann = sroots(i)
        mm = roots(i)
      END IF
    END DO
    IF (count == 0) THEN
      WRITE (*,*) 'M_COMPLEX_BRUGGEMANN: Keine passende Loesung gefunden! Abbruch!'
      m_complex_bruggemann = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF
    IF (count > 1) THEN
      WRITE (*,*) 'M_COMPLEX_BRUGGEMANN: Mehrere passende Loesungen gefunden! Da is was faul! Abbruch!'
      m_complex_bruggemann = CMPLX(-999.99d0,-999.99d0,kind=kinddp)
      fehler = 1
      RETURN
    END IF

    ! Einsetzen der Loesung in die Bruggemann-Formel muss 0 ergeben: (nur bei Anwendung des Laguerre-Verfahrens pruefen!)
!    check = volair*(m1t-mm)/(m1t+2d0*mm)+ &
!         volice*(m2t-mm)/(m2t+2d0*mm)+ &
!         volwater*(m3t-mm)/(m3t+2d0*mm)
!    IF (ABS(check) > 1d-2) THEN
!      WRITE (*,'(a,1x,e12.5)') 'M_COMPLEX_BRUGGEMANN: Die Loesung fuer m_eff hat ein Residuum von ', ABS(check)
!    END IF

    RETURN

  CONTAINS

!!! 2 Routinen fuer das Laguerre-Verfahren zur Bestimmung der Wurzeln eines komplexen Polynoms,
!!! siehe Numerical Recipes in Fortran 77, p. 365 ff.

    SUBROUTINE laguer(a,m,x,its)
      IMPLICIT NONE
      INTEGER :: m,its,MAXIT,MR,MT
      DOUBLE PRECISION :: EPSS
      COMPLEX(kind=kinddp) :: a(m+1),x
      PARAMETER (EPSS=2.d-7,MR=8,MT=10,MAXIT=MT*MR)
      INTEGER :: iter,j
      DOUBLE PRECISION :: abx,abp,abm,err,frac(MR)
      COMPLEX(kind=kinddp) :: dx,x1,b,d,f,g,h,sq,gp,gm,g2
      SAVE frac
      frac = (/.5,.25,.75,.13,.38,.62,.88,1./)
      DO iter=1,MAXIT
        its=iter
        b=a(m+1)
        err=abs(b)
        d=CMPLX(0.,0.,kind=kinddp)
        f=cmplx(0.,0.,kind=kinddp)
        abx=abs(x)
        DO j=m,1,-1
          f=x*f+d
          d=x*d+b
          b=x*b+a(j)
          err=abs(b)+abx*err
        END DO
        err=EPSS*err
        if(abs(b).le.err) then
          return
        else
          g=d/b
          g2=g*g
          h=g2-2.*f/b
          sq=sqrt((m-1)*(m*h-g2))
          gp=g+sq
          gm=g-sq
          abp=abs(gp)
          abm=abs(gm)
          if(abp.lt.abm) gp=gm
          if (max(abp,abm).gt.0.) then
            dx=m/gp
          else
            dx=EXP(CMPLX(LOG(1.+abx),dble(iter),kind=kinddp))
          endif
        endif
        x1=x-dx
        if(x.eq.x1)return
        if (mod(iter,MT).ne.0) then
          x=x1
        else
          x=x-dx*frac(iter/MT)
        endif
      END DO
      WRITE (*,*) 'M_COMPLEX_BRUGGEMANN: too many iterations in laguer (Laguerre-Method)'
      STOP
      RETURN
    END SUBROUTINE laguer

    SUBROUTINE zroots(a,m,roots,polish)
      INTEGER :: m,MAXM
      DOUBLE PRECISION :: EPS
      COMPLEX(kind=kinddp) :: a(m+1),roots(m)
      LOGICAL :: polish
      PARAMETER (EPS=1.e-6,MAXM=101)
      !    USES laguer
      INTEGER :: i,j,jj,its
      COMPLEX(kind=kinddp) :: ad(MAXM),x,b,c
      DO j=1,m+1
        ad(j)=a(j)
      END DO
      DO j=m,1,-1
        x=CMPLX(0.,0.,kind=kinddp)
        CALL laguer(ad,j,x,its)
        IF(ABS(AIMAG(x)).LE.2.*EPS**2*ABS(REAL(x))) x=CMPLX(DBLE(x),0.d0,kind=kinddp)
        roots(j)=x
        b=ad(j+1)
        DO jj=j,1,-1
          c=ad(jj)
          ad(jj)=b
          b=x*b+c
        END DO
      END DO
      if (polish) then
        DO j=1,m
          call laguer(a,m,roots(j),its)
        END DO
      endif
      DO j=2,m
        x=roots(j)
        DO i=j-1,1,-1
          if(real(roots(i)).le.real(x)) exit
          roots(i+1)=roots(i)
        END DO
        !        i=0
        roots(i+1)=x
      END DO
      RETURN
    END SUBROUTINE zroots

  END FUNCTION m_complex_bruggemann

  !----------------------------------------------------------------------------------------!
  ! 2d) Interfaces zu den Brechungsindexroutinen, gesteuert durch Schluesselworstrings:
  !----------------------------------------------------------------------------------------!

  ! Fuer Eis-Wasser-Luft-Mischung in einem Durchgang:
  FUNCTION get_m_mix_vec(m_a, m_i, m_w, volair, volice, volwater, &
       mixingrulestring, matrixstring, inclusionstring, anz, fehler) RESULT(get_m_mix)

    IMPLICIT NONE

    INTEGER, INTENT(in)  :: anz
    DOUBLE PRECISION, INTENT(in) :: volice(anz), volair(anz), volwater(anz)
    COMPLEX(kind=kinddp), INTENT(in) :: m_a(anz), m_i(anz), m_w(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    INTEGER, INTENT(out) :: fehler
    COMPLEX(kind=kinddp) :: get_m_mix(anz)

    fehler = 0
    get_m_mix = CMPLX(1.0d0,0.0d0,kind=kinddp)

    SELECT CASE (TRIM(ADJUSTL(mixingrulestring)))

    CASE ('maxwellgarnett','MAXWELLGARNETT','Maxwellgarnett')
      ! Als erstes Argument kommt der Volumenanteil der Matrix, dann die anderen beiden Konstituenten
      ! (Luft kommt in der Mischung nicht vor und wird mit volair = 0.0 belegt)
      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE ('ice','ICE','Ice')
!CDIR NEXPAND
        get_m_mix = m_complex_maxwellgarnett_vec(volice, volair, volwater, m_i, m_a, m_w, inclusionstring, anz, fehler)
      CASE ('water','WATER','Water')
!CDIR NEXPAND
        get_m_mix = m_complex_maxwellgarnett_vec(volwater, volair, volice, m_w, m_a, m_i, inclusionstring, anz, fehler)
      CASE ('air','AIR','Air')
!CDIR NEXPAND
        get_m_mix = m_complex_maxwellgarnett_vec(volair, volwater, volice, m_a, m_w, m_i, inclusionstring, anz, fehler)
      CASE DEFAULT
        WRITE (*,*) 'GET_M_MIX_VEC: Unbekannter matrixstring: ', matrixstring
        fehler = 1
      END SELECT
    CASE ('bruggemann','BRUGGEMANN','Bruggemann')
!CDIR NEXPAND
      get_m_mix = m_complex_bruggemann_vec(volair, volice, volwater, m_a, m_i, m_w, anz, fehler)
    CASE ('oguchi','OGUCHI','Oguchi')
!CDIR NEXPAND
      get_m_mix = m_complex_oguchi_vec(volair, volice, volwater, m_a, m_i, m_w, anz, fehler)        
    CASE default
      WRITE (*,*) 'GET_M_MIX_VEC: Unbekannter mixingrulestring: ', mixingrulestring
      fehler = 2
    END SELECT
    IF (fehler /= 0) THEN
      WRITE (*,*) 'GET_M_MIX_VEC: Fehler beim eff. Brechungsindex aufgetreten!'
      STOP
      RETURN
    END IF

    RETURN
  END FUNCTION get_m_mix_vec

  COMPLEX(kind=kinddp) FUNCTION get_m_mix(m_a, m_i, m_w, volair, volice, volwater, &
       mixingrulestring, matrixstring, inclusionstring, fehler)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: volice, volair, volwater
    COMPLEX(kind=kinddp), INTENT(in) :: m_a, m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    INTEGER, INTENT(out) :: fehler

    fehler = 0
    get_m_mix = CMPLX(1.0d0,0.0d0,kind=kinddp)

    SELECT CASE (TRIM(ADJUSTL(mixingrulestring)))

    CASE ('maxwellgarnett','MAXWELLGARNETT','Maxwellgarnett')
      ! Als erstes Argument kommt der Volumenanteil der Matrix, dann die anderen beiden Konstituenten
      ! (Luft kommt in der Mischung nicht vor und wird mit volair = 0.0 belegt)
      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE ('ice','ICE','Ice')
        get_m_mix = m_complex_maxwellgarnett(volice, volair, volwater, m_i, m_a, m_w, inclusionstring, fehler)
      CASE ('water','WATER','Water')
        get_m_mix = m_complex_maxwellgarnett(volwater, volair, volice, m_w, m_a, m_i, inclusionstring, fehler)
      CASE ('air','AIR','Air')
        get_m_mix = m_complex_maxwellgarnett(volair, volwater, volice, m_a, m_w, m_i, inclusionstring, fehler)
      CASE DEFAULT
        WRITE (*,*) 'GET_M_MIX: Unbekannter matrixstring: ', matrixstring
        fehler = 1
      END SELECT
    CASE ('bruggemann','BRUGGEMANN','Bruggemann')
      get_m_mix = m_complex_bruggemann(volair, volice, volwater, m_a, m_i, m_w, fehler)
    CASE ('oguchi','OGUCHI','Oguchi')
      get_m_mix = m_complex_oguchi(volair, volice, volwater, m_a, m_i, m_w, fehler)        
    CASE default
      WRITE (*,*) 'GET_M_MIX: Unbekannter mixingrulestring: ', mixingrulestring
      fehler = 2
    END SELECT
    IF (fehler /= 0) THEN
      WRITE (*,*) 'GET_M_MIX: Fehler beim eff. Brechungsindex aufgetreten!'
      STOP
      RETURN
    END IF

    RETURN
  END FUNCTION get_m_mix

  ! Ineinandergeschachtelt: ( (m1 + m2) + m3), wobei m1,m2,m3 jeweils Luft, Eis oder Wasser sein koennen!
  FUNCTION get_m_mix_nested_vec(m_a, m_i, m_w, volair, volice, volwater, &
       mixingrulestring, hoststring, matrixstring, inclusionstring, hostmatrixstring, &
       hostinclusionstring, anz, kumfehler) &
       RESULT (get_m_mix_nested)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in) :: volice(anz), volair(anz), volwater(anz)
    COMPLEX(kind=kinddp), INTENT(in) :: m_a(anz), m_i(anz), m_w(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, hoststring, matrixstring, &
         inclusionstring, hostmatrixstring, hostinclusionstring
    INTEGER, INTENT(out) :: kumfehler
    COMPLEX(kind=kinddp) :: get_m_mix_nested(anz)

    DOUBLE PRECISION :: vol1(anz), vol2(anz), nullv(anz)
    COMPLEX(kind=kinddp) :: mtmp(anz)
    INTEGER :: fehler

    kumfehler = 0
    get_m_mix_nested = CMPLX(1.0d0,0.0d0,kind=kinddp)
    nullv = 0.0d0

    SELECT CASE (TRIM(ADJUSTL(hoststring)))

    CASE('air','AIR','Air')

      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE('air','AIR','Air')
        WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger matrixstring: ', matrixstring
        kumfehler = kumfehler + 1
      CASE default
        vol1 = volice / MAX(volice+volwater,1d-10)
        vol2 = 1.0d0 - vol1
        mtmp = get_m_mix_vec(m_a, m_i, m_w, nullv, vol1, vol2, &
             mixingrulestring, matrixstring, inclusionstring, anz, fehler)
        kumfehler = kumfehler + fehler

        SELECT CASE (TRIM(ADJUSTL(hostmatrixstring)))
        CASE('air','AIR','Air')
          get_m_mix_nested = get_m_mix_vec(m_a, mtmp, 2.0*m_a, volair, (1.0d0-volair), nullv, &
               mixingrulestring, hostmatrixstring, hostinclusionstring, anz, fehler)
          kumfehler = kumfehler + fehler
        CASE('icewater','ICEWATER','Icewater')
          get_m_mix_nested = get_m_mix_vec(m_a, mtmp, 2.0*m_a, volair, (1.0d0-volair), nullv, &
               mixingrulestring, 'ice', hostinclusionstring, anz, fehler)
          kumfehler = kumfehler + fehler          
        CASE default
          WRITE (*,*) 'GET_M_MIX_NESTED_VEC: Ungueltiger hostmatrixstring: ', hostmatrixstring
          kumfehler = kumfehler + 1
        END SELECT
      END SELECT

    CASE('ice','ICE','Ice')

      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE('ice','ICE','Ice')
        WRITE (*,*) 'GET_M_MIX_NESTED_VEC: Ungueltiger matrixstring: ', matrixstring
        kumfehler = kumfehler + 1
      CASE default
        vol1 = volair / MAX(volair+volwater,1d-10)
        vol2 = 1.0d0 - vol1
        mtmp = get_m_mix_vec(m_a, m_i, m_w, vol1, nullv, vol2, &
             mixingrulestring, matrixstring, inclusionstring, anz, fehler)
        kumfehler = kumfehler + fehler

        SELECT CASE (TRIM(ADJUSTL(hostmatrixstring)))
        CASE('ice','ICE','Ice')
          get_m_mix_nested = get_m_mix_vec(mtmp, m_i, 2.0*m_a, (1.0d0-volice), volice, nullv, &
               mixingrulestring, hostmatrixstring, hostinclusionstring, anz, fehler)
          kumfehler = kumfehler + fehler
        CASE('airwater','AIRWATER','Airwater')
          get_m_mix_nested = get_m_mix_vec(mtmp, m_i, 2.0*m_a, (1.0d0-volice), volice, nullv, &
               mixingrulestring, 'air', hostinclusionstring, anz, fehler)
          kumfehler = kumfehler + fehler          
        CASE default
          WRITE (*,*) 'GET_M_MIX_NESTED_VEC: Ungueltiger hostmatrixstring: ', hostmatrixstring
          kumfehler = kumfehler + 1
        END SELECT
      END SELECT

    CASE('water','WATER','Water')

      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE('water','WATER','Water')
        WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger matrixstring: ', matrixstring
        kumfehler = kumfehler + 1
      CASE default
        vol1 = volair / MAX(volice+volair,1d-10)
        vol2 = 1.0d0 - vol1
        mtmp = get_m_mix_vec(m_a, m_i, m_w, vol1, vol2, nullv, &
             mixingrulestring, matrixstring, inclusionstring, anz, fehler)
        kumfehler = kumfehler + fehler

        SELECT CASE (TRIM(ADJUSTL(hostmatrixstring)))
        CASE('water','WATER','Water')
          get_m_mix_nested = get_m_mix_vec(2.0d0*m_a, mtmp, m_w, nullv, (1.0d0-volwater), volwater, &
               mixingrulestring, hostmatrixstring, hostinclusionstring, anz, fehler)
          kumfehler = kumfehler + fehler
        CASE('airice','AIRICE','Airice')
          get_m_mix_nested = get_m_mix_vec(2.0d0*m_a, mtmp, m_w, nullv, (1.0d0-volwater), volwater, &
               mixingrulestring, 'ice', hostinclusionstring, anz, fehler)
          kumfehler = kumfehler + fehler          
        CASE default
          WRITE (*,*) 'GET_M_MIX_NESTED_VEC: Ungueltiger hostmatrixstring: ', hostmatrixstring
          kumfehler = kumfehler + 1
        END SELECT
      END SELECT

    CASE ('none','NONE','None')

      get_m_mix_nested = get_m_mix_vec(m_a, m_i, m_w, volair, volice, volwater, &
           mixingrulestring, matrixstring, inclusionstring, anz, fehler)
      kumfehler = kumfehler + fehler

    CASE default
      WRITE (*,*) 'GET_M_MIX_NESTED_VEC: Unbekannter hoststring: ', mixingrulestring
      kumfehler = kumfehler + 1
    END SELECT


    IF (kumfehler /= 0) THEN
      WRITE (*,*) 'GET_M_MIX_NESTED_VEC: Fehler beim eff. Brechungsindex aufgetreten!'
      get_m_mix_nested = CMPLX(1.0d0,0.0d0,kind=kinddp)
      STOP
      RETURN
    END IF

    RETURN
  END FUNCTION get_m_mix_nested_vec

  COMPLEX(kind=kinddp) FUNCTION get_m_mix_nested(m_a, m_i, m_w, volair, volice, volwater, &
       mixingrulestring, hoststring, matrixstring, inclusionstring, hostmatrixstring, hostinclusionstring, kumfehler)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: volice, volair, volwater
    COMPLEX(kind=kinddp), INTENT(in) :: m_a, m_i, m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, hoststring, matrixstring, &
         inclusionstring, hostmatrixstring, hostinclusionstring
    INTEGER, INTENT(out) :: kumfehler

    DOUBLE PRECISION :: vol1, vol2
    COMPLEX(kind=kinddp) :: mtmp
    INTEGER :: fehler

    kumfehler = 0
    get_m_mix_nested = CMPLX(1.0d0,0.0d0,kind=kinddp)

    SELECT CASE (TRIM(ADJUSTL(hoststring)))

    CASE('air','AIR','Air')

      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE('air','AIR','Air')
        WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger matrixstring: ', matrixstring
        kumfehler = kumfehler + 1
      CASE default
        vol1 = volice / MAX(volice+volwater,1d-10)
        vol2 = 1.0d0 - vol1
        mtmp = get_m_mix(m_a, m_i, m_w, 0.0d0, vol1, vol2, &
             mixingrulestring, matrixstring, inclusionstring, fehler)
        kumfehler = kumfehler + fehler

        SELECT CASE (TRIM(ADJUSTL(hostmatrixstring)))
        CASE('air','AIR','Air')
          get_m_mix_nested = get_m_mix(m_a, mtmp, 2.0*m_a, volair, (1.0d0-volair), 0.0d0, &
               mixingrulestring, hostmatrixstring, hostinclusionstring, fehler)
          kumfehler = kumfehler + fehler
        CASE('icewater','ICEWATER','Icewater')
          get_m_mix_nested = get_m_mix(m_a, mtmp, 2.0*m_a, volair, (1.0d0-volair), 0.0d0, &
               mixingrulestring, 'ice', hostinclusionstring, fehler)
          kumfehler = kumfehler + fehler          
        CASE default
          WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger hostmatrixstring: ', hostmatrixstring
          kumfehler = kumfehler + 1
        END SELECT
      END SELECT

    CASE('ice','ICE','Ice')

      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE('ice','ICE','Ice')
        WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger matrixstring: ', matrixstring
        kumfehler = kumfehler + 1
      CASE default
        vol1 = volair / MAX(volair+volwater,1d-10)
        vol2 = 1.0d0 - vol1
        mtmp = get_m_mix(m_a, m_i, m_w, vol1, 0.0d0, vol2, &
             mixingrulestring, matrixstring, inclusionstring, fehler)
        kumfehler = kumfehler + fehler

        SELECT CASE (TRIM(ADJUSTL(hostmatrixstring)))
        CASE('ice','ICE','Ice')
          get_m_mix_nested = get_m_mix(mtmp, m_i, 2.0*m_a, (1.0d0-volice), volice, 0.0d0, &
               mixingrulestring, hostmatrixstring, hostinclusionstring, fehler)
          kumfehler = kumfehler + fehler
        CASE('airwater','AIRWATER','Airwater')
          get_m_mix_nested = get_m_mix(mtmp, m_i, 2.0*m_a, (1.0d0-volice), volice, 0.0d0, &
               mixingrulestring, 'air', hostinclusionstring, fehler)
          kumfehler = kumfehler + fehler          
        CASE default
          WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger hostmatrixstring: ', hostmatrixstring
          kumfehler = kumfehler + 1
        END SELECT
      END SELECT

    CASE('water','WATER','Water')

      SELECT CASE (TRIM(ADJUSTL(matrixstring)))
      CASE('water','WATER','Water')
        WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger matrixstring: ', matrixstring
        kumfehler = kumfehler + 1
      CASE default
        vol1 = volair / MAX(volice+volair,1d-10)
        vol2 = 1.0d0 - vol1
        mtmp = get_m_mix(m_a, m_i, m_w, vol1, vol2, 0.0d0, &
             mixingrulestring, matrixstring, inclusionstring, fehler)
        kumfehler = kumfehler + fehler

        SELECT CASE (TRIM(ADJUSTL(hostmatrixstring)))
        CASE('water','WATER','Water')
          get_m_mix_nested = get_m_mix(2.0d0*m_a, mtmp, m_w, 0.0d0, (1.0d0-volwater), volwater, &
               mixingrulestring, hostmatrixstring, hostinclusionstring, fehler)
          kumfehler = kumfehler + fehler
        CASE('airice','AIRICE','Airice')
          get_m_mix_nested = get_m_mix(2.0d0*m_a, mtmp, m_w, 0.0d0, (1.0d0-volwater), volwater, &
               mixingrulestring, 'ice', hostinclusionstring, fehler)
          kumfehler = kumfehler + fehler          
        CASE default
          WRITE (*,*) 'GET_M_MIX_NESTED: Ungueltiger hostmatrixstring: ', hostmatrixstring
          kumfehler = kumfehler + 1
        END SELECT
      END SELECT

    CASE ('none','NONE','None')

      get_m_mix_nested = get_m_mix(m_a, m_i, m_w, volair, volice, volwater, &
           mixingrulestring, matrixstring, inclusionstring, fehler)
      kumfehler = kumfehler + fehler

    CASE default
      WRITE (*,*) 'GET_M_MIX_NESTED: Unbekannter hoststring: ', mixingrulestring
      kumfehler = kumfehler + 1
    END SELECT


    IF (kumfehler /= 0) THEN
      WRITE (*,*) 'GET_M_MIX_NESTED: Fehler beim eff. Brechungsindex aufgetreten!'
      get_m_mix_nested = CMPLX(1.0d0,0.0d0,kind=kinddp)
      STOP
      RETURN
    END IF

    RETURN
  END FUNCTION get_m_mix_nested

!*********************************************************************************************

  !----------------------------------------------------------------------------------------!
  ! 3) Routinen zur Berechnung der Streufunktionen nach Mie
  !----------------------------------------------------------------------------------------!

  !----------------------------------------------------------------------------------------!
  ! 3a) Rekursionsformeln fuer Qext, Qback und Qseit nach 
  !     Bohren und Huffman (1983) fuer einschalige Kugeln
  !----------------------------------------------------------------------------------------!

  SUBROUTINE BHMIESEITVEC(X,REFREL,QBACK,QEXT,QSCA,anz)

    ! Es werden nur Streuwinkel von 180 Grad und 0 Grad (Extinktion!) berechnet.
    ! Erweiterungen auf bel. Streurichtung aber sehr einfach moeglich!!!

    IMPLICIT NONE

    INTEGER, PARAMETER :: kinddp = KIND(0.0d0), NANG = 2
    
    INTEGER :: anz
    DOUBLE PRECISION :: GSCA(anz),QBACK(anz),QEXT(anz),QSCA(anz),X(anz)
    COMPLEX(kind=kinddp) :: REFREL(anz)
    COMPLEX(kind=kinddp) :: S1(anz,NANG),S2(anz,NANG)

    INTEGER :: J,N,NSTOP(anz),NMX(anz),NN,K
    DOUBLE PRECISION :: APSI(anz),APSI1(anz),CHI(anz),CHI0(anz),CHI1(anz),FN,PII,&
         XSTOP(anz),YMOD(anz)
    DOUBLE PRECISION ::  AMU(NANG),PI(anz,NANG),PI0(anz,NANG),PI1(anz,NANG),TAU(anz,NANG),THETA(NANG)
    DOUBLE PRECISION :: PSI0(anz),PSI1(anz),PSI(anz),DN,degrad
    COMPLEX(kind=kinddp) :: AN(anz),AN1(anz),BN(anz),BN1(anz),XI(anz),XI1(anz),Y(anz)
    COMPLEX(kind=kinddp), ALLOCATABLE :: D(:,:)
!===========================================================================
!
! Singularity for X = 0.0! The calling program has to take care of this!
!
!===========================================================================
    !***********************************************************************
    ! Subroutine BHMIE is the Bohren-Huffman Mie scattering subroutine
    !    to calculate scattering and absorption by a homogenous isotropic
    !    sphere.
    ! Given:
    !    X = 2*pi*a/lambda
    !    REFREL = (complex refr. index of sphere)/(real index of medium)
    !    NANG = number of angles between 0 and 90 degrees
    !           (will calculate 2*NANG-1 directions from 0 to 180 deg.)
    !           if called with NANG<2, will set NANG=2 and will compute
    !           scattering for theta=0,90,180.
    ! Returns:
    !    S1(1 - 2*NANG-1) = -i*f_22 (incid. E perp. to scatt. plane,
    !                                scatt. E perp. to scatt. plane)
    !    S2(1 - 2*NANG-1) = -i*f_11 (incid. E parr. to scatt. plane,
    !                               scatt. E parr. to scatt. plane)
    !   QEXT = C_ext/pi*a**2 = efficiency factor for extinction
    !   QSCA = C_sca/pi*a**2 = efficiency factor for scattering
    !   QBACK = (dC_sca/domega)/pi*a**2
    !         = backscattering efficiency
    !   GSCA = <cos(theta)> for scattering
    !
    !Original program taken from Bohren and Huffman (1983), Appendix A
    !Modified by B.T.Draine, Princeton Univ. Obs., 90/10/26
    !in order to compute <cos(theta)>
    !91/05/07 (BTD): Modified to allow NANG=1
    !91/08/15 (BTD): Corrected error (failure to initialize P)
    !91/08/15 (BTD): Modified to enhance vectorizability.
    !91/08/15 (BTD): Modified to make NANG=2 if called with NANG=1
    !91/08/15 (BTD): Changed definition of QBACK.
    !92/01/08 (BTD): Note that this version has been superceded by
    !                fully double precision version = bhmie.f which,
    !                unfortunately, is not standard f77.
    !                However, retain this in case standard f77 version
    !                is required for porting to some other system.
    !2005/11/07 U. Blahak: transferred to F90 and full double precision,
    !                major modifications
    !***********************************************************************

    !*** Obtain pi:
    PII=4.d0*ATAN(1.d0)
    degrad = PII / 180.0d0
    Y=X*REFREL
    YMOD=ABS(Y)
    !
    !*** Series expansion terminated after NSTOP terms
    !   Logarithmic derivatives calculated from NMX on down
    XSTOP=FLOOR(X+4.d0*X**0.3333+2.0)+1
    !*** Original code:
    !     NMX=AMAX1(XSTOP,YMOD)+15
    !     NSTOP=XSTOP
    !*** Experimental code:
    NMX=MAX(XSTOP,YMOD)+15
    NSTOP=XSTOP
    ALLOCATE(D(anz,MAXVAL(NMX)))

    !*** Require NANG.GE.1 in order to calculate scattering intensities
!!! Hier werden die Streuwinkel theta (zenitwinkel zur Vorwaertsstreurichtung)
!!! gesetzt:
    THETA(1) = 0.0d0 * degrad
!!! THETA(...) = ... ! THETA prinzipiell erweiterbar auf beliebige Winkel. NANG entsprechend erhoehen!!!
    THETA(2) = 180.0d0 * degrad

    DO j=1,NANG
      AMU(J)=COS(THETA(J))
    END DO
    DO J=1,NANG
      DO k=1,anz
        PI0(k,J)=0.d0
        PI1(k,J)=1.d0
      END DO
    END DO
    NN=NANG
    DO J=1,NANG
      DO k=1,anz
        S1(k,J)=CMPLX(0.d0,0.d0,kind=kinddp)
        S2(k,J)=CMPLX(0.d0,0.d0,kind=kinddp)
      END DO
    END DO

! Maybe simply set NMX to the scalar value maxval(NMX)?

    !*** Logarithmic derivative D(J) calculated by downward recurrence
    !   beginning with initial value (0.,0.) at J=NMX
!    DO k=1,anz
!      D(k,NMX(k))=CMPLX(0.d0,0.d0,kind=kinddp)
!    END DO
    D = CMPLX(0.0d0, 0.0d0, kind=kinddp)
    NN=MAXVAL(NMX)-1
    DO N=1,NN
      DO k=1,anz
        IF (NMX(k)-N >= 1) THEN
          DN=NMX(k)-N+1
          D(k,NMX(k)-N)=(DN/Y(k))-(1.d0/(D(k,NMX(k)-N+1)+DN/Y(k)))
        END IF
      END DO
    END DO

    !*** Riccati-Bessel functions with real argument X
    !   calculated by upward recurrence

    PSI0=COS(X)
    PSI1=SIN(X)
    CHI0=-SIN(X)
    CHI1=COS(X)
    !APSI0 never used, so this line removed from program:
    !     APSI0=PSI0
    APSI1=PSI1
    !XI0 never used, so this line removed from program:
    !     In the next two lines, the minus-sign is changed to switch to convention of REFREL with negative imaginary part
    !     XI0=CMPLX(APSI0,-CHI0)
    !      XI1=CMPLX(APSI1,-CHI1)
    XI1=CMPLX(APSI1,CHI1,kind=kinddp)
    QSCA=0.d0
    GSCA=0.d0
    AN = (0.0d0, 0.0d0)
    BN = (0.0d0, 0.0d0)
    DO N=1,MAXVAL(NSTOP)
      DN=N
      FN=(2.d0*DN+1.d0)/(DN*(DN+1.d0))
      DO k=1,anz
        IF (N <= NSTOP(k)) THEN
          PSI(k)=(2.d0*DN-1.d0)*PSI1(k)/X(k)-PSI0(k)
          APSI(k)=PSI(k)
          CHI(k)=(2.d0*DN-1.d0)*CHI1(k)/X(k)-CHI0(k)

      !     In the next line, the minus-sign is changed to switch to convention of REFREL with negative imaginary part
      !          XI(k)=CMPLX(APSI(k),-CHI(k))
          XI(k)=CMPLX(APSI(k),CHI(k))

      !*** Store previous values of AN and BN for use
      !   in computation of g=<cos(theta)>
          AN1(k)=AN(k)
          BN1(k)=BN(k)


      !*** Compute AN and BN:
          AN(k)=(D(k,N)/REFREL(k)+DN/X(k))*APSI(k)-APSI1(k)
          AN(k)=AN(k)/((D(k,N)/REFREL(k)+DN/X(k))*XI(k)-XI1(k))
          BN(k)=(REFREL(k)*D(k,N)+DN/X(k))*APSI(k)-APSI1(k)
          BN(k)=BN(k)/((REFREL(k)*D(k,N)+DN/X(k))*XI(k)-XI1(k))

      !*** Augment sums for Qsca and g=<cos(theta)>
          QSCA(k)=QSCA(k)+(2.*DN+1.)*(ABS(AN(k))**2+ABS(BN(k))**2)
          GSCA(k)=GSCA(k)+((2.*DN+1.)/(DN*(DN+1.))) * &
               (DBLE(AN(k))*DBLE(BN(k))+AIMAG(AN(k))*AIMAG(BN(k)))
        END IF
      END DO

      IF(N.GT.1)THEN
        DO k=1,anz
          IF (N <= NSTOP(k)) THEN
            GSCA(k)=GSCA(k)+((DN-1.)*(DN+1.)/DN)* &
                 (DBLE(AN1(k))*DBLE(AN(k))+AIMAG(AN1(k))*AIMAG(AN(k))+ &
                 DBLE(BN1(k))*DBLE(BN(k))+AIMAG(BN1(k))*AIMAG(BN(k)))
          END IF
        END DO
      ENDIF

      !*** Now calculate scattering intensity pattern
      DO J=1,NANG
        DO k=1,anz
          IF (N <= NSTOP(k)) THEN
            PI(k,J)=PI1(k,J)
            TAU(k,J)=DN*AMU(J)*PI(k,J)-(DN+1.d0)*PI0(k,J)
            S1(k,J)=S1(k,J)+FN*(AN(k)*PI(k,J)+BN(k)*TAU(k,J))
            S2(k,J)=S2(k,J)+FN*(AN(k)*TAU(k,J)+BN(k)*PI(k,J))

      !*** Compute pi_n for next value of n
      !   For each angle J, compute pi_n+1
      !   from PI = pi_n , PI0 = pi_n-1
            PI1(k,J)=((2.*DN+1.)*AMU(J)*PI(k,J)-(DN+1.)*PI0(k,J))/DN
            PI0(k,J)=PI(k,J)

          END IF
        END DO
      END DO


      DO k=1,anz
        IF (N <= NSTOP(k)) THEN
          PSI0(k)=PSI1(k)
          PSI1(k)=PSI(k)
          APSI1(k)=PSI1(k)
          CHI0(k)=CHI1(k)
          CHI1(k)=CHI(k)

          !     In the next line, the minus-sign is changed to switch to convention of REFREL with negative imaginary part
          !          XI1=CMPLX(APSI1,-CHI1)
          XI1(k)=CMPLX(APSI1(k),CHI1(k),kind=kinddp)
        END IF
      END DO

    END DO

    !*** Have summed sufficient terms.
    !   Now compute QSCA,QEXT,QBACK,and GSCA (efficiency factors)
    GSCA=2.*GSCA/QSCA
    QSCA=(2.d0/(X*X))*QSCA
    QEXT=(4.d0/(X*X))*DBLE(S1(:,1))
    ! QSEIT ohne Ber. der Polarisation, es wird vorausgesetzt, dass das Messgeraet alle Polarisationsrichtungen empfangen kann.
    !  QSEIT=(4.d0/(X*X))*(ABS(S1(2:NANG-1))**2*SIN(PHI(2:NANG-1)*degrad)**2 + ABS(S1(NANG(2:NANG-1)))**2*COS(PHI(2:NANG-1)*degrad)**2)
    QBACK=(4.d0/(X*X))*(ABS(S1(:,NANG)))**2

    DEALLOCATE(D)

    RETURN
  END SUBROUTINE BHMIESEITVEC

  SUBROUTINE BHMIESEIT(X,REFREL,QBACK,QEXT,QSCA)

    ! Es werden nur Streuwinkel von 180 Grad und 0 Grad (Extinktion!) berechnet.
    ! Erweiterungen auf bel. Streurichtung aber sehr einfach moeglich!!!
    ! Siehe unten, Variablen theta() und NANG erweitern.

    IMPLICIT NONE

    INTEGER, PARAMETER :: kinddp = KIND(0.0d0), NANG = 2

    DOUBLE PRECISION :: GSCA,QBACK,QEXT,QSCA,X
    COMPLEX(kind=kinddp) :: REFREL
    COMPLEX(kind=kinddp) :: S1(NANG),S2(NANG)

    INTEGER :: J,N,NSTOP,NMX,NN
    DOUBLE PRECISION :: APSI,APSI1,CHI,CHI0,CHI1,FN,PII,&
         XSTOP,YMOD
    DOUBLE PRECISION ::  AMU(NANG),PI(NANG),PI0(NANG),PI1(NANG),TAU(NANG),THETA(NANG)
    DOUBLE PRECISION :: PSI0,PSI1,PSI,DN,degrad
    COMPLEX(kind=kinddp) :: AN,AN1,BN,BN1,XI,XI1,Y
    COMPLEX(kind=kinddp), ALLOCATABLE :: D(:)
!===========================================================================
!
! Singularity for X = 0.0! The calling program has to take care of this!
!
!===========================================================================
    !***********************************************************************
    ! Subroutine BHMIE is the Bohren-Huffman Mie scattering subroutine
    !    to calculate scattering and absorption by a homogenous isotropic
    !    sphere.
    ! Given:
    !    X = 2*pi*a/lambda
    !    REFREL = (complex refr. index of sphere)/(real index of medium)
    !    NANG = number of angles between 0 and 90 degrees
    !           (will calculate 2*NANG-1 directions from 0 to 180 deg.)
    !           if called with NANG<2, will set NANG=2 and will compute
    !           scattering for theta=0,90,180.
    ! Returns:
    !    S1(1 - 2*NANG-1) = -i*f_22 (incid. E perp. to scatt. plane,
    !                                scatt. E perp. to scatt. plane)
    !    S2(1 - 2*NANG-1) = -i*f_11 (incid. E parr. to scatt. plane,
    !                               scatt. E parr. to scatt. plane)
    !   QEXT = C_ext/pi*a**2 = efficiency factor for extinction
    !   QSCA = C_sca/pi*a**2 = efficiency factor for scattering
    !   QBACK = (dC_sca/domega)/pi*a**2
    !         = backscattering efficiency
    !   GSCA = <cos(theta)> for scattering
    !
    !Original program taken from Bohren and Huffman (1983), Appendix A
    !Modified by B.T.Draine, Princeton Univ. Obs., 90/10/26
    !in order to compute <cos(theta)>
    !91/05/07 (BTD): Modified to allow NANG=1
    !91/08/15 (BTD): Corrected error (failure to initialize P)
    !91/08/15 (BTD): Modified to enhance vectorizability.
    !91/08/15 (BTD): Modified to make NANG=2 if called with NANG=1
    !91/08/15 (BTD): Changed definition of QBACK.
    !92/01/08 (BTD): Note that this version has been superceded by
    !                fully double precision version = bhmie.f which,
    !                unfortunately, is not standard f77.
    !                However, retain this in case standard f77 version
    !                is required for porting to some other system.
    !2005/11/07 U. Blahak: transferred to F90 and full double precision,
    !                major modifications
    !***********************************************************************

    !*** Obtain pi:
    PII=4.d0*ATAN(1.d0)
    degrad = PII / 180.0d0
    Y=X*REFREL
    YMOD=ABS(Y)
    !
    !*** Series expansion terminated after NSTOP terms
    !   Logarithmic derivatives calculated from NMX on down
    XSTOP=FLOOR(X+4.d0*X**0.3333+2.0)+1
    !*** Original code:
    !     NMX=AMAX1(XSTOP,YMOD)+15
    !     NSTOP=XSTOP
    !*** Experimental code:
    NMX=MAX(XSTOP,YMOD)+15
    NSTOP=XSTOP
    ALLOCATE(D(NMX))

    !*** Require NANG.GE.1 in order to calculate scattering intensities
!!! Hier werden die Streuwinkel theta (zenitwinkel zur Vorwaertsstreurichtung)
!!! gesetzt:
    THETA(1) = 0.0d0 * degrad
!!! THETA(...) = ... ! THETA prinzipiell erweiterbar auf beliebige Winkel. NANG entsprechend erhoehen!!!
    THETA(2) = 180.0d0 * degrad
    DO j=1,NANG
      AMU(J)=COS(THETA(J))
    END DO
    DO J=1,NANG
      PI0(J)=0.d0
      PI1(J)=1.d0
    END DO
    NN=NANG
    DO J=1,NANG
      S1(J)=CMPLX(0.d0,0.d0,kind=kinddp)
      S2(J)=CMPLX(0.d0,0.d0,kind=kinddp)
    END DO

    !*** Logarithmic derivative D(J) calculated by downward recurrence
    !   beginning with initial value (0.,0.) at J=NMX

    D(NMX)=CMPLX(0.d0,0.d0,kind=kinddp)
    NN=NMX-1
    DO N=1,NN
      DN=NMX-N+1
      D(NMX-N)=(DN/Y)-(1.d0/(D(NMX-N+1)+DN/Y))
    END DO

    !*** Riccati-Bessel functions with real argument X
    !   calculated by upward recurrence

    PSI0=COS(X)
    PSI1=SIN(X)
    CHI0=-SIN(X)
    CHI1=COS(X)
    !APSI0 never used, so this line removed from program:
    !     APSI0=PSI0
    APSI1=PSI1
    !XI0 never used, so this line removed from program:
    !     In the next two lines, the minus-sign is changed to switch to convention of REFREL with negative imaginary part
    !     XI0=CMPLX(APSI0,-CHI0)
    !      XI1=CMPLX(APSI1,-CHI1)
    XI1=CMPLX(APSI1,CHI1,kind=kinddp)
    QSCA=0.d0
    GSCA=0.d0
    DO N=1,NSTOP
      DN=N
      FN=(2.d0*DN+1.d0)/(DN*(DN+1.d0))
      PSI=(2.d0*DN-1.d0)*PSI1/X-PSI0
      APSI=PSI
      CHI=(2.d0*DN-1.d0)*CHI1/X-CHI0

      !     In the next line, the minus-sign is changed to switch to convention of REFREL with negative imaginary part
      !          XI=CMPLX(APSI,-CHI)
      XI=CMPLX(APSI,CHI)

      !*** Store previous values of AN and BN for use
      !   in computation of g=<cos(theta)>
      IF(N.GT.1)THEN
        AN1=AN
        BN1=BN
      ENDIF

      !*** Compute AN and BN:
      AN=(D(N)/REFREL+DN/X)*APSI-APSI1
      AN=AN/((D(N)/REFREL+DN/X)*XI-XI1)
      BN=(REFREL*D(N)+DN/X)*APSI-APSI1
      BN=BN/((REFREL*D(N)+DN/X)*XI-XI1)

      !*** Augment sums for Qsca and g=<cos(theta)>
      QSCA=QSCA+(2.*DN+1.)*(ABS(AN)**2+ABS(BN)**2)
      GSCA=GSCA+((2.*DN+1.)/(DN*(DN+1.))) * &
           (DBLE(AN)*DBLE(BN)+AIMAG(AN)*AIMAG(BN))
      IF(N.GT.1)THEN
        GSCA=GSCA+((DN-1.)*(DN+1.)/DN)* &
             (DBLE(AN1)*DBLE(AN)+AIMAG(AN1)*AIMAG(AN)+ &
             DBLE(BN1)*DBLE(BN)+AIMAG(BN1)*AIMAG(BN))
      ENDIF

      !*** Now calculate scattering intensity pattern
      DO J=1,NANG
        PI(J)=PI1(J)
        TAU(J)=DN*AMU(J)*PI(J)-(DN+1.d0)*PI0(J)
        S1(J)=S1(J)+FN*(AN*PI(J)+BN*TAU(J))
        S2(J)=S2(J)+FN*(AN*TAU(J)+BN*PI(J))
      END DO

      PSI0=PSI1
      PSI1=PSI
      APSI1=PSI1
      CHI0=CHI1
      CHI1=CHI

      !     In the next line, the minus-sign is changed to switch to convention of REFREL with negative imaginary part
      !          XI1=CMPLX(APSI1,-CHI1)
      XI1=CMPLX(APSI1,CHI1,kind=kinddp)

      !*** Compute pi_n for next value of n
      !   For each angle J, compute pi_n+1
      !   from PI = pi_n , PI0 = pi_n-1
      DO J=1,NANG
        PI1(J)=((2.*DN+1.)*AMU(J)*PI(J)-(DN+1.)*PI0(J))/DN
        PI0(J)=PI(J)
      END DO

    END DO

    !*** Have summed sufficient terms.
    !   Now compute QSCA,QEXT,QBACK,and GSCA (efficiency factors)
    GSCA=2.*GSCA/QSCA
    QSCA=(2.d0/(X*X))*QSCA
    QEXT=(4.d0/(X*X))*DBLE(S1(1))
    ! QSEIT ohne Ber. der Polarisation, es wird vorausgesetzt, dass das Messgeraet alle Polarisationsrichtungen empfangen kann.
    !  QSEIT=(4.d0/(X*X))*(ABS(S1(2:NANG-1))**2*SIN(PHI(2:NANG-1)*degrad)**2 + ABS(S1(NANG(2:NANG-1)))**2*COS(PHI(2:NANG-1)*degrad)**2)
    QBACK=(4.d0/(X*X))*(ABS(S1(NANG)))**2

    DEALLOCATE(D)

    RETURN
  END SUBROUTINE BHMIESEIT



  !----------------------------------------------------------------------------------------!
  ! 3b) Rekursionsformeln nach Bohren und Huffman (1983) fuer zweischalige Kugeln
  !----------------------------------------------------------------------------------------!

  SUBROUTINE BHCOATBACKVEC(X,Y,RRFRL1,RRFRL2,QBACK,QEXT,QSCA,anz)

    IMPLICIT NONE

    INTEGER, PARAMETER :: kinddp = KIND(0.0d0)

    INTEGER :: anz
    COMPLEX(kind=kinddp) :: RRFRL1(anz),RRFRL2(anz)

    INTEGER :: IFLAG(anz),N,NSTOP(anz),K
    DOUBLE PRECISION :: CHI0Y(anz),CHI1Y(anz),CHIY(anz),DEL,PSI0Y(anz),&
         PSI1Y(anz),PSIY(anz),QEXT(anz),QBACK(anz),QSCA(anz),X(anz),Y(anz),YSTOP(anz),RN
    COMPLEX(kind=kinddp) :: AMESS1,AMESS2,AMESS3,AMESS4,AN(anz),ANCAP, &
         BN(anz),BNCAP,BRACK(anz), &
         CHI0X2(anz),CHI0Y2(anz),CHI1X2(anz),CHI1Y2(anz),&
         CHIX2(anz),CHIPX2(anz),CHIPY2(anz),CHIY2(anz),CRACK(anz), &
         D0X1(anz),D0X2(anz),D0Y2(anz),D1X1(anz),D1X2(anz),D1Y2(anz),DNBAR,GNBAR,II, &
         REFREL(anz),RFREL1(anz),RFREL2(anz), &
         XBACK(anz),XI0Y(anz),XI1Y(anz),XIY(anz), &
         X1(anz),X2(anz),Y2(anz)
!===========================================================================
!
! - Singularity for X = 0.0, Y = 0.0. The calling program has to deal with this.
!
! - Use of optimization may slightly change the results, i.e., ifort -O2
!   or all optimization options for sxf90.
!
!===========================================================================
    !***********************************************************************
    !
    ! Subroutine BHCOAT calculates Q_ext, Q_sca, Q_back for coated sphere.
    ! All bessel functions computed by upward recurrence.
    ! Input:
    !        X = 2*PI*RCORE*REFMED/WAVEL
    !        Y = 2*PI*RMANT*REFMED/WAVEL
    !        RFREL1 = REFCOR/REFMED
    !        RFREL2 = REFMAN/REFMED 
    ! where  REFCOR = complex refr.index of core)
    !        REFMAN = complex refr.index of mantle)
    !        REFMED = real refr.index of medium)
    !        RCORE = radius of core
    !        RMANT = radius of mantle
    !        WAVEL = wavelength of light in ambient medium
    !
    ! Routine BHCOAT is taken from Bohren & Huffman (1983)
    ! Obtained from C.L.Joseph
    !
    ! History:
    ! 92/11/24 (BTD) Explicit declaration of all variables
    ! 2005/11/07 U.Blahak completely rewritten in F90
    !***********************************************************************

    DEL = 1d-8
    II = CMPLX(0.D0,1.D0,kind=kinddp)
    RFREL1=RRFRL1
    RFREL2=RRFRL2
    !         -----------------------------------------------------------
    !              del is the inner sphere convergence criterion
    !         -----------------------------------------------------------
    x1 = rfrel1*x
    x2 = rfrel2*x
    y2 = rfrel2*y
    ystop = y + 4.*y**0.3333 + 2.0
    refrel = rfrel2/rfrel1
    nstop = ystop
    !         -----------------------------------------------------------
    !              series terminated after nstop terms
    !         -----------------------------------------------------------
    d0x1 = COS(x1)/SIN(x1)
    d0x2 = COS(x2)/SIN(x2)
    d0y2 = COS(y2)/SIN(y2)
    psi0y = COS(y)
    psi1y = SIN(y)
    chi0y = -SIN(y)
    chi1y = COS(y)
    !     The minus-sign in the next 2 lines was changed to + to switch to the refractive index convention with negative imaginary part
    !     xi0y = psi0y-II*chi0y
    !     xi1y = psi1y-II*chi1y
    xi0y = psi0y+II*chi0y
    xi1y = psi1y+II*chi1y
    chi0y2 = -SIN(y2)
    chi1y2 = COS(y2)
    chi0x2 = -SIN(x2)
    chi1x2 = COS(x2)
    qsca = 0.0
    qext = 0.0
    xback = CMPLX(0.0d0,0.0d0,kind=kinddp)

    iflag = 0

    DO n=1,MAXVAL(nstop)

      rn = n

      DO k=1,anz
        IF (n <= nstop(k)) THEN
          psiy(k) = (2.0*rn-1.)*psi1y(k)/y(k) - psi0y(k)
          chiy(k) = (2.0*rn-1.)*chi1y(k)/y(k) - chi0y(k)
          xiy(k) = psiy(k)+II*chiy(k)
          d1y2(k) = 1.0/(rn/y2(k)-d0y2(k)) - rn/y2(k)
        END IF
      END DO
 
      DO k=1,anz
        IF (n <= nstop(k)) THEN
          IF (iflag(k) .EQ. 0) THEN
            d1x1(k) = 1.0/(rn/x1(k)-d0x1(k)) - rn/x1(k)
            d1x2(k) = 1.0/(rn/x2(k)-d0x2(k)) - rn/x2(k)
            chix2(k) = (2.0*rn - 1.0)*chi1x2(k)/x2(k) - chi0x2(k)
            chiy2(k) = (2.0*rn - 1.0)*chi1y2(k)/y2(k) - chi0y2(k)
            chipx2(k) = chi1x2(k) - rn*chix2(k)/x2(k)
            chipy2(k) = chi1y2(k) - rn*chiy2(k)/y2(k)
            ancap = refrel(k)*d1x1(k) - d1x2(k)
            ancap = ancap/(refrel(k)*d1x1(k)*chix2(k) - chipx2(k))
            ancap = ancap/(chix2(k)*d1x2(k) - chipx2(k))
            brack(k) = ancap*(chiy2(k)*d1y2(k) - chipy2(k))
            bncap = refrel(k)*d1x2(k) - d1x1(k)
            bncap = bncap/(refrel(k)*chipx2(k) - d1x1(k)*chix2(k))
            bncap = bncap/(chix2(k)*d1x2(k) - chipx2(k))
            crack(k) = bncap*(chiy2(k)*d1y2(k) - chipy2(k))
            amess1 = brack(k)*chipy2(k)
            amess2 = brack(k)*chiy2(k)
            amess3 = crack(k)*chipy2(k)
            amess4 = crack(k)*chiy2(k)
            IF (ABS(amess1) .LE. del*ABS(d1y2(k)) .AND. &
                 ABS(amess2) .LE. del .AND. &
                 ABS(amess3) .LE. del*ABS(d1y2(k)) .AND. &
                 ABS(amess4) .LE. del) THEN
              brack(k) = CMPLX(0.d0,0.d0,kind=kinddp)
              crack(k) = CMPLX(0.d0,0.d0,kind=kinddp)
              iflag(k) = 1
            END IF
          END IF
          dnbar = d1y2(k) - brack(k)*chipy2(k)
          dnbar = dnbar/(1.0-brack(k)*chiy2(k))
          gnbar = d1y2(k) - crack(k)*chipy2(k)
          gnbar = gnbar/(1.0-crack(k)*chiy2(k))
          an(k) = (dnbar/rfrel2(k) + rn/y(k))*psiy(k) - psi1y(k)
          an(k) = an(k)/((dnbar/rfrel2(k)+rn/y(k))*xiy(k)-xi1y(k))
          bn(k) = (rfrel2(k)*gnbar + rn/y(k))*psiy(k) - psi1y(k)
          bn(k) = bn(k)/((rfrel2(k)*gnbar+rn/y(k))*xiy(k)-xi1y(k))
          qsca(k) = qsca(k) + 0.5*(2.0*rn+1.0)*(ABS(an(k))*ABS(an(k))+ABS(bn(k))*ABS(bn(k)))
          xback(k) = xback(k) + 0.5*(2.0*rn+1.0)*(-1.)**n*(an(k)-bn(k))
          qext(k) = qext(k) + 0.5*(2.0*rn + 1.0)*DBLE(an(k)+bn(k))
          psi0y(k) = psi1y(k)
          psi1y(k) = psiy(k)
          chi0y(k) = chi1y(k)
          chi1y(k) = chiy(k)
          xi1y(k) = psi1y(k)+II*chi1y(k)
          chi0x2(k) = chi1x2(k)
          chi1x2(k) = chix2(k)
          chi0y2(k) = chi1y2(k)
          chi1y2(k) = chiy2(k)
          d0x1(k) = d1x1(k)
          d0x2(k) = d1x2(k)
          d0y2(k) = d1y2(k)
        END IF
      END DO
    END DO

    QSCA = (4.0/(y*y))*QSCA
    QEXT = (4.0/(y*y))*QEXT
    qback = (4.0/(y*y))*(ABS(XBACK))**2
    RETURN
  END SUBROUTINE BHCOATBACKVEC

  SUBROUTINE BHCOATBACK(X,Y,RRFRL1,RRFRL2,QBACK,QEXT,QSCA)

    IMPLICIT NONE

    INTEGER, PARAMETER :: kinddp = KIND(0.0d0)

    COMPLEX(kind=kinddp) :: RRFRL1,RRFRL2

    INTEGER :: IFLAG,N,NSTOP
    DOUBLE PRECISION :: CHI0Y,CHI1Y,CHIY,DEL,PSI0Y,PSI1Y,PSIY,QEXT,QBACK,RN,QSCA,X,Y,YSTOP
    COMPLEX(kind=kinddp) :: AMESS1,AMESS2,AMESS3,AMESS4,AN,ANCAP, &
         BN,BNCAP,BRACK, &
         CHI0X2,CHI0Y2,CHI1X2,CHI1Y2,CHIX2,CHIPX2,CHIPY2,CHIY2,CRACK, &
         D0X1,D0X2,D0Y2,D1X1,D1X2,D1Y2,DNBAR,GNBAR,II, &
         REFREL,RFREL1,RFREL2, &
         XBACK,XI0Y,XI1Y,XIY, &
         X1,X2,Y2
    !***********************************************************************
    !
    ! Subroutine BHCOAT calculates Q_ext, Q_sca, Q_back for coated sphere.
    ! All bessel functions computed by upward recurrence.
    ! Input:
    !        X = 2*PI*RCORE*REFMED/WAVEL
    !        Y = 2*PI*RMANT*REFMED/WAVEL
    !        RFREL1 = REFCOR/REFMED
    !        RFREL2 = REFMAN/REFMED 
    ! where  REFCOR = complex refr.index of core)
    !        REFMAN = complex refr.index of mantle)
    !        REFMED = real refr.index of medium)
    !        RCORE = radius of core
    !        RMANT = radius of mantle
    !        WAVEL = wavelength of light in ambient medium
    !
    ! Routine BHCOAT is taken from Bohren & Huffman (1983)
    ! Obtained from C.L.Joseph
    !
    ! History:
    ! 92/11/24 (BTD) Explicit declaration of all variables
    ! 2005/11/07 U.Blahak completely rewritten in F90
    !***********************************************************************

    DEL = 1d-8
    II = CMPLX(0.D0,1.D0,kind=kinddp)
    RFREL1=RRFRL1
    RFREL2=RRFRL2
    !         -----------------------------------------------------------
    !              del is the inner sphere convergence criterion
    !         -----------------------------------------------------------
    x1 = rfrel1*x
    x2 = rfrel2*x
    y2 = rfrel2*y
    ystop = y + 4.*y**0.3333 + 2.0
    refrel = rfrel2/rfrel1
    nstop = ystop
    !         -----------------------------------------------------------
    !              series terminated after nstop terms
    !         -----------------------------------------------------------
    d0x1 = COS(x1)/SIN(x1)
    d0x2 = COS(x2)/SIN(x2)
    d0y2 = COS(y2)/SIN(y2)
    psi0y = COS(y)
    psi1y = SIN(y)
    chi0y = -SIN(y)
    chi1y = COS(y)
    !     The minus-sign in the next 2 lines was changed to + to switch to the refractive index convention with negative imaginary part
    !     xi0y = psi0y-II*chi0y
    !     xi1y = psi1y-II*chi1y
    xi0y = psi0y+II*chi0y
    xi1y = psi1y+II*chi1y
    chi0y2 = -SIN(y2)
    chi1y2 = COS(y2)
    chi0x2 = -SIN(x2)
    chi1x2 = COS(x2)
    qsca = 0.0
    qext = 0.0
    xback = CMPLX(0.0d0,0.0d0,kind=kinddp)

    iflag = 0
    DO n=1,nstop

      rn = n
      psiy = (2.0*rn-1.)*psi1y/y - psi0y
      chiy = (2.0*rn-1.)*chi1y/y - chi0y
      xiy = psiy+II*chiy
      d1y2 = 1.0/(rn/y2-d0y2) - rn/y2
      IF (iflag .EQ. 0) THEN
        d1x1 = 1.0/(rn/x1-d0x1) - rn/x1
        d1x2 = 1.0/(rn/x2-d0x2) - rn/x2
        chix2 = (2.0*rn - 1.0)*chi1x2/x2 - chi0x2
        chiy2 = (2.0*rn - 1.0)*chi1y2/y2 - chi0y2
        chipx2 = chi1x2 - rn*chix2/x2
        chipy2 = chi1y2 - rn*chiy2/y2
        ancap = refrel*d1x1 - d1x2
        ancap = ancap/(refrel*d1x1*chix2 - chipx2)
        ancap = ancap/(chix2*d1x2 - chipx2)
        brack = ancap*(chiy2*d1y2 - chipy2)
        bncap = refrel*d1x2 - d1x1
        bncap = bncap/(refrel*chipx2 - d1x1*chix2)
        bncap = bncap/(chix2*d1x2 - chipx2)
        crack = bncap*(chiy2*d1y2 - chipy2)
        amess1 = brack*chipy2
        amess2 = brack*chiy2
        amess3 = crack*chipy2
        amess4 = crack*chiy2
        IF (ABS(amess1) .LE. del*ABS(d1y2) .AND. &
             ABS(amess2) .LE. del .AND. &
             ABS(amess3) .LE. del*ABS(d1y2) .AND. &
             ABS(amess4) .LE. del) THEN
          brack = CMPLX(0.d0,0.d0,kind=kinddp)
          crack = CMPLX(0.d0,0.d0,kind=kinddp)
          iflag = 1
        END IF
      END IF
      dnbar = d1y2 - brack*chipy2
      dnbar = dnbar/(1.0-brack*chiy2)
      gnbar = d1y2 - crack*chipy2
      gnbar = gnbar/(1.0-crack*chiy2)
      an = (dnbar/rfrel2 + rn/y)*psiy - psi1y
      an = an/((dnbar/rfrel2+rn/y)*xiy-xi1y)
      bn = (rfrel2*gnbar + rn/y)*psiy - psi1y
      bn = bn/((rfrel2*gnbar+rn/y)*xiy-xi1y)
      qsca = qsca + 0.5*(2.0*rn+1.0)*(ABS(an)*ABS(an)+ABS(bn)*ABS(bn))
      xback = xback + 0.5*(2.0*rn+1.0)*(-1.)**n*(an-bn)
      qext = qext + 0.5*(2.0*rn + 1.0)*DBLE(an+bn)
      psi0y = psi1y
      psi1y = psiy
      chi0y = chi1y
      chi1y = chiy
      xi1y = psi1y+II*chi1y
      chi0x2 = chi1x2
      chi1x2 = chix2
      chi0y2 = chi1y2
      chi1y2 = chiy2
      d0x1 = d1x1
      d0x2 = d1x2
      d0y2 = d1y2
    END DO

    QSCA = (4.0/(y*y))*QSCA
    QEXT = (4.0/(y*y))*QEXT
    qback = (4.0/(y*y))*(ABS(XBACK))**2
    RETURN
  END SUBROUTINE BHCOATBACK

  !----------------------------------------------------------------------------------------!
  ! 3c) Ansteuern von BHMIESEIT()
  !----------------------------------------------------------------------------------------!

  SUBROUTINE SPHERE_SCATTER_BH_VEC(r,m,m_medium,lambda,C_back,C_ext,C_sca, anz)
    IMPLICIT NONE

    !..Berechnung des Rueckstreuquerschnitts nach Bohren and Huffman (1983)

    !..Eingabevariablen
    INTEGER :: anz
    DOUBLE PRECISION     :: r(anz),lambda,m_medium   ! in [m] angeben, dann C_back in [m^2]
    COMPLEX(kind=kinddp) :: m(anz)

    !..Ausgabevariablen
    DOUBLE PRECISION  :: C_back(anz), C_ext(anz), C_sca(anz)
    DOUBLE PRECISION  :: Q_back(anz), Q_ext(anz), Q_sca(anz)

    !..Lokale Variablen
    DOUBLE PRECISION     :: alf(anz)

        !..Dimensionslose Radius
    alf = 2.0*pi / lambda * r * m_medium

      !..Berechung der Streueffizienz Q_back:
    CALL BHMIESEITVEC(MAX(alf,1d-12),m/m_medium,Q_back,Q_ext,Q_sca,anz)

    ! Rueckstreuquerschnitt in der Einheit [Laenge^2]
    WHERE (alf >= 1d-12)
      C_back = pi * r**2 *  Q_back !!! blahak, 4.3.2003
      C_ext = pi * r**2 *  Q_ext !!! blahak, 4.3.2003
      C_sca = pi * r**2 *  Q_sca !!! blahak, 4.3.2003
    ELSEWHERE
      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0
    END WHERE

  END SUBROUTINE SPHERE_SCATTER_BH_VEC

  SUBROUTINE SPHERE_SCATTER_BH(r,m,m_medium,lambda,C_back,C_ext,C_sca)
    IMPLICIT NONE

    !..Berechnung des Rueckstreuquerschnitts nach Bohren and Huffman (1983)

    !..Eingabevariablen
    DOUBLE PRECISION     :: r,lambda,m_medium   ! in [m] angeben, dann C_back in [m^2]
    COMPLEX(kind=kinddp) :: m

    !..Ausgabevariablen
    DOUBLE PRECISION  :: C_back, C_ext, C_sca
    DOUBLE PRECISION  :: Q_back, Q_ext, Q_sca

    !..Lokale Variablen
    DOUBLE PRECISION     :: alf

        !..Dimensionslose Radius
    alf = 2.0*pi / lambda * r * m_medium

    IF (alf >= 1d-12) THEN
      !..Berechung der Streueffizienz Q_back:
      CALL BHMIESEIT(alf,m/m_medium,Q_back,Q_ext,Q_sca)

      ! Rueckstreuquerschnitt in der Einheit [Laenge^2]
      C_back = pi * r**2 *  Q_back !!! blahak, 4.3.2003
      C_ext = pi * r**2 *  Q_ext !!! blahak, 4.3.2003
      C_sca = pi * r**2 *  Q_sca !!! blahak, 4.3.2003

    ELSE

      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0

    END IF

  END SUBROUTINE SPHERE_SCATTER_BH

  !----------------------------------------------------------------------------------------!
  ! 3d) Ansteuern von BHCOATBACKVEC()
  !----------------------------------------------------------------------------------------!

  SUBROUTINE COATEDSPHERE_SCATTER_BH_VEC(r,r_shell,m_core,m_shell,m_medium,lambda,C_back,C_ext,C_sca,anz)
    IMPLICIT NONE

    !..Berechnung des Rueckstreuquerschnitts, des Wirkungsquerschnitts 
    !  und des effektiven Dielektrizitaetsfaktors nach Bohren and Huffman (1983)

    !..Eingabevariablen
    INTEGER :: anz
    DOUBLE PRECISION     :: r(anz),r_shell(anz),lambda
    COMPLEX(kind=kinddp) :: m_core(anz),m_shell(anz)
    DOUBLE PRECISION    :: m_medium

    !..Ausgabevariablen
    DOUBLE PRECISION     :: Q_sca(anz),C_sca(anz),Q_ext(anz),C_ext(anz),Q_back(anz),C_back(anz)
    DOUBLE PRECISION     :: Q_scahlp(anz),Q_exthlp(anz),Q_backhlp(anz)

    !..Lokale Variablen
    DOUBLE PRECISION            :: alf(anz),nue(anz),alfhlp(anz),nuehlp(anz)

    !..Dimensionslose Radien (vgl. Kerker, S. 192, Gl. (5.1.25)
    alf = MAX(2.0*pi / lambda * (r-r_shell) * m_medium , 1d-16)
    nue = 2.0*pi / lambda * r * m_medium

    alfhlp = MAX(alf,1d-7)
    nuehlp = MAX(MAX(nue,2d-7),alf)

    CALL BHCOATBACKVEC(alfhlp,nuehlp,m_core/m_medium,m_shell/m_medium,Q_back,Q_ext,Q_sca,anz)

    IF (ANY(alf < 1d-8 .AND. nue >= 1d-7)) THEN
      CALL BHMIESEITVEC(nuehlp,m_shell/m_medium,Q_backhlp,Q_exthlp,Q_scahlp,anz)  
      WHERE(alf < 1d-8 .AND. nue >= 1d-7)
        Q_back = Q_backhlp
        Q_ext  = Q_exthlp
        Q_sca  = Q_scahlp
      END WHERE
    END IF

    WHERE (nue >= 1d-7 .AND. nue >= alf)
      C_back =  pi * r**2 * Q_back
      C_ext =  pi * r**2 * Q_ext
      C_sca =  pi * r**2 * Q_sca
    ELSEWHERE
      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0
    END WHERE

    RETURN
  END SUBROUTINE COATEDSPHERE_SCATTER_BH_VEC

  SUBROUTINE COATEDSPHERE_SCATTER_BH(r,r_shell,m_core,m_shell,m_medium,lambda,C_back,C_ext,C_sca)
    IMPLICIT NONE

    !..Berechnung des Rueckstreuquerschnitts, des Wirkungsquerschnitts 
    !  und des effektiven Dielektrizitaetsfaktors nach Bohren and Huffman (1983)

    !..Eingabevariablen
    DOUBLE PRECISION     :: r,r_shell,lambda
    COMPLEX(kind=kinddp) :: m_core,m_shell
    DOUBLE PRECISION    :: m_medium

    !..Ausgabevariablen
    DOUBLE PRECISION     :: Q_sca,C_sca,Q_ext,C_ext,Q_back,C_back

    !..Lokale Variablen
    DOUBLE PRECISION            :: alf,nue

    !..Dimensionslose Radien (vgl. Kerker, S. 192, Gl. (5.1.25)
    alf = MAX(2.0*pi / lambda * (r-r_shell) * m_medium , 1d-14)
    nue = 2.0*pi / lambda * r * m_medium

    IF (nue >= 2d-7 .AND. nue >= alf) THEN

      !..Berechung der div. Querschnitte:
      IF (alf >= 1d-7) THEN
        CALL BHCOATBACK(alf,nue,m_core/m_medium,m_shell/m_medium,Q_back,Q_ext,Q_sca)
      ELSE
        CALL BHMIESEIT(nue,m_shell/m_medium,Q_back,Q_ext,Q_sca)
      END IF
      
      C_back =  pi * r**2 * Q_back
      C_ext =  pi * r**2 * Q_ext
      C_sca =  pi * r**2 * Q_sca

    ELSE

      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0

    END IF

    RETURN
  END SUBROUTINE COATEDSPHERE_SCATTER_BH


!----------------------------------------------------------------------------------------!
!
! Alternativ fuer BHCOATBACK:
! Mie-Theorie fuer zweischalige Kugel nach Kerker (1969)
! (etwas langsamer, aber nachvollziehbar)
!
!----------------------------------------------------------------------------------------!

!==================================================================================
!
! Mie-Koeffizienten an und bn der Ordnung n einer Kugel mit
! Brechungsindex m1 (=m_kugel/m_medium), die mit einer Kugelschale 
! des Materials mit Brechungsindex m2 (=m_schale/m_medium)
! ueberzogen ist, als Funktion des Mie-Parameters der Kugel x1
! (=2*pi*r1/lambda_medium = 2*pi*r1*m_medium/lambda) und des Mie-Parameters
! von Kugel+Schale x2 (=2*pi*(r1+r2)/lambda_medium) -- beide reell,
! d.h. das aeussere Medium absorbiert nicht: nach Kerker (1969): The scattering 
! of light, S. 189 ff.
! Das ganze wird in double-Arithmetik gerechnet!
!
!     ---> hat fuer x1 = 0.0 eine Singularitaet!
!
!==================================================================================

  ! Vektorisierte Form: Es werden die Streukoeffizienten an und bn von der
  ! Ordnung 0 bis zur Ordnung NMAX zurueckgegeben, an(1:NMAX), bn(1:NMAX)
  !
  ! Davon die nochmals ueber verschiedene x-elemente vektorisierte Fassung:
  SUBROUTINE COATEDSPHERE_SCATTERCOEFFS_VEC(anz,NMAX,m1,m2,x1,x2,an,bn,maske)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz, NMAX
    COMPLEX(kind=kinddp), INTENT(in) :: m1(anz), m2(anz)
    DOUBLE PRECISION, INTENT(in) :: x1(anz), x2(anz)
    LOGICAL, OPTIONAL, INTENT(in) :: maske(anz)

    COMPLEX(kind=kinddp), INTENT(out) :: an(anz,1:NMAX), bn(anz,1:NMAX)
    COMPLEX(kind=kinddp), PARAMETER :: imag = (0.0d0,1.0d0), czero = (0.0d0,0.0d0)

    COMPLEX(kind=kinddp) :: y1(anz), y2(anz), z2(anz)
    COMPLEX(kind=kinddp), DIMENSION(1:anz,0:NMAX) :: sphj_y1, sphdj_y1, sphy_y1, sphdy_y1, &
         sphj_y2, sphdj_y2, sphy_y2, sphdy_y2, &
         sphj_z2, sphdj_z2, sphy_z2, sphdy_z2, &
         sphj_x2, sphdj_x2, sphy_x2, sphdy_x2
    COMPLEX(kind=kinddp) :: psiy1, psiy2, psiz2, &
         dpsiy1, dpsiy2, dpsiz2, &
         xsiy2, xsiz2, dxsiy2, dxsiz2, zetax2, dzetax2
    COMPLEX(kind=kinddp) :: det1, det2, det3, det4, det5, det6, nenner
    DOUBLE PRECISION     :: psix2, dpsix2
    INTEGER :: NM(anz), i, k
    LOGICAL :: maskeloc(anz)

    IF (ANY(x1 > x2)) THEN
      WRITE (*,*) 'COATEDSPHERE_SCATTERCOEFFS_VEC: Fehler: Innerer Kugelradius > Aeusserer Radius!'
    ENDIF

    IF (.NOT.PRESENT(maske)) THEN
      maskeloc = .TRUE.
    ELSE
      maskeloc = maske
    END IF

    y1 = m1 * x1
    y2 = m2 * x1
    z2 = m2 * x2

    ! Zuerst die weiter unten benoetigten Bessel-Funktionen und deren Ableitungen
    ! bis zur Ordnung NMAX berechnen (backward recurrence):
    CALL CSPHJYVEC(NMAX,y1,NM,sphj_y1,sphdj_y1,sphy_y1,sphdy_y1,anz)
    CALL CSPHJYVEC(NMAX,y2,NM,sphj_y2,sphdj_y2,sphy_y2,sphdy_y2,anz)
    CALL CSPHJYVEC(NMAX,z2,NM,sphj_z2,sphdj_z2,sphy_z2,sphdy_z2,anz)
    CALL CSPHJYVEC(NMAX,CMPLX(x2,0.0d0,kind=kinddp),NM,sphj_x2,sphdj_x2,sphy_x2,sphdy_x2,anz)

    an = czero
    bn = czero
    DO i=1,NMAX
      DO k=1,anz
        IF (maskeloc(k)) THEN
          IF (x1(k) > 0.0d0) THEN
            ! Die Riccati-Bessel-Funktionen der benoetigten Argumente:

            psiy1 = y1(k) * sphj_y1(k,i)
            dpsiy1 = sphj_y1(k,i) + y1(k) * sphdj_y1(k,i)

            psiy2 = y2(k) * sphj_y2(k,i)
            dpsiy2 = sphj_y2(k,i) + y2(k) * sphdj_y2(k,i)

            psiz2 = z2(k) * sphj_z2(k,i)
            dpsiz2 = sphj_z2(k,i) + z2(k) * sphdj_z2(k,i)

            psix2 = x2(k) * DBLE(sphj_x2(k,i))
            dpsix2 = DBLE(sphj_x2(k,i)) + x2(k) * DBLE(sphdj_x2(k,i))

            xsiy2 = -y2(k) * sphy_y2(k,i)
            dxsiy2 = -sphy_y2(k,i) - y2(k) * sphdy_y2(k,i)

            xsiz2 = -z2(k) * sphy_z2(k,i)
            dxsiz2 = -sphy_z2(k,i) - z2(k) * sphdy_z2(k,i)

            zetax2 = x2(k) * (DBLE(sphj_x2(k,i)) - imag * DBLE(sphy_x2(k,i)))
            dzetax2 = x2(k) * (DBLE(sphdj_x2(k,i)) - imag * DBLE(sphdy_x2(k,i))) + &
                 (DBLE(sphj_x2(k,i)) - imag * DBLE(sphy_x2(k,i)))

            ! Einige vorbereitende Terme auf die Streukoeffizienten:

            det1 = m1(k)*dpsiy2*psiy1 - m2(k)*psiy2*dpsiy1
            det2 = dxsiz2*psix2 - m2(k)*xsiz2*dpsix2
            det3 = m1(k)*dxsiy2*psiy1 - m2(k)*xsiy2*dpsiy1
            det4 = dpsiz2*psix2 - m2(k)*psiz2*dpsix2
            det5 = dxsiz2*zetax2 - m2(k)*xsiz2*dzetax2
            det6 = dpsiz2*zetax2 - m2(k)*psiz2*dzetax2

            nenner = det1*det5-det3*det6

            IF (nenner /= czero) THEN
              an(k,i) = (det1*det2-det3*det4)/nenner
            ELSE
              an(k,i) = czero
            END IF

            det1 = m2(k)*dpsiy2*psiy1 - m1(k)*psiy2*dpsiy1
            det2 = m2(k)*dxsiz2*psix2 - xsiz2*dpsix2
            det3 = m2(k)*dxsiy2*psiy1 - m1(k)*xsiy2*dpsiy1
            det4 = m2(k)*dpsiz2*psix2 - psiz2*dpsix2
            det5 = m2(k)*dxsiz2*zetax2 - xsiz2*dzetax2
            det6 = m2(k)*dpsiz2*zetax2 - psiz2*dzetax2

            nenner = det1*det5-det3*det6

            IF (nenner /= czero) THEN
              bn(k,i) = (det1*det2-det3*det4)/nenner
            ELSE
              bn(k,i) = czero
            END IF

          ELSE

            ! Die Riccati-Bessel-Funktionen der benoetigten Argumente:

            psiz2 = z2(k) * sphj_z2(k,i)
            dpsiz2 = sphj_z2(k,i) + z2(k) * sphdj_z2(k,i)

            psix2 = x2(k) * DBLE(sphj_x2(k,i))
            dpsix2 = DBLE(sphj_x2(k,i)) + x2(k) * DBLE(sphdj_x2(k,i))

            zetax2 = x2(k) * (DBLE(sphj_x2(k,i)) - imag * DBLE(sphy_x2(k,i)))
            dzetax2 = x2(k) * (DBLE(sphdj_x2(k,i)) - imag * DBLE(sphdy_x2(k,i))) + &
                 (DBLE(sphj_x2(k,i)) - imag * DBLE(sphy_x2(k,i)))

            det4 = dpsiz2*psix2 - m2(k)*psiz2*dpsix2
            det6 = dpsiz2*zetax2 - m2(k)*psiz2*dzetax2

            IF (det6 /= czero) THEN
              an(k,i) = det4 / det6
            ELSE
              an(k,i) = czero
            END IF

            det4 = m2(k)*dpsiz2*psix2 - psiz2*dpsix2
            det6 = m2(k)*dpsiz2*zetax2 - psiz2*dzetax2

            IF (det6 /= czero) THEN
              bn(k,i) = det4 / det6
            ELSE
              bn(k,i) = czero
            END IF
          END IF
        END IF
      END DO
    END DO

    RETURN

  END SUBROUTINE COATEDSPHERE_SCATTERCOEFFS_VEC

  SUBROUTINE COATEDSPHERE_SCATTER_VEC(r,r_shell,m_core,m_shell,m_medium,lambda,C_back,C_ext,C_sca,anz,maske)
    IMPLICIT NONE

    !..Berechnung der Rueckstreueffizienz
    !  r = Radius der Kugel (Gesamtradius, r = r_shell + r_core)
    !  r_shell = Dicke der auesseren Schicht
    !  lambda  = Wellenlaenge


    !..Eingabevariablen
    INTEGER :: anz
    DOUBLE PRECISION, INTENT(in) :: r(anz),r_shell(anz),lambda,m_medium
    COMPLEX(kind=kinddp),INTENT(in) :: m_core(anz),m_shell(anz)
    LOGICAL, OPTIONAL, intent(in) :: maske(anz)

    !..Ausgabevariablen
    DOUBLE PRECISION, intent(out) :: C_sca(anz),C_back(anz),C_ext(anz)

    !..Lokale Variablen
    INTEGER, PARAMETER :: NMAX = 50
    INTEGER :: n, k
    DOUBLE PRECISION :: alf(anz),nue(anz),nuehlp(anz),chlp_sca(anz),csum_sca(anz)
    COMPLEX(kind=kinddp) :: an(anz,NMAX),bn(anz,NMAX), &
         chlp_back(anz),chlp_ext(anz),csum_back(anz),csum_ext(anz)
    DOUBLE PRECISION, PARAMETER :: eps = 1D-6
    LOGICAL :: maskeloc(anz)

    IF (.NOT.PRESENT(maske)) THEN
      maskeloc = .true.
    ELSE
      maskeloc = maske
    END IF

    !..Dimensionslose Radien (vgl. Kerker, S. 192, Gl. (5.1.25)
    alf = 2*pi / lambda * (r-r_shell) * m_medium
    nue = 2*pi / lambda * r * m_medium

    nuehlp = MAX(nue, 1d-12)

    !.. Streukoeffizienten bis zur NMAX-ten Ordnung berechnen:
    CALL COATEDSPHERE_SCATTERCOEFFS_VEC(anz,NMAX,m_core/m_medium,m_shell/m_medium,alf,nuehlp,an,bn,maske=maskeloc)

    !..Berechung des Rueckstreuquerschnitts 
    !  nach van de Hulst, S. 284
    csum_back = (0.0d0,0.0d0)
    csum_ext  = (0.0d0,0.0d0)
    csum_sca  = 0.0d0
    chlp_back = (1.0d0,0.0d0)
    chlp_ext  = (1.0d0,0.0d0)
    chlp_sca  = 1.0d0
    DO n=1,NMAX
      DO k=1,anz
        IF ( maskeloc(k) .AND. (ABS(chlp_back(k)) >= eps * ABS(csum_back(k)) .OR. &
             ABS(chlp_ext(k)) >= eps * ABS(csum_ext(k)) .OR. &
             ABS(chlp_sca(k)) >= eps * ABS(csum_sca(k))) ) THEN

          chlp_back(k) = (n + 0.5d0) * (-1)**n * ( an(k,n) - bn(k,n) )
          csum_back(k) = csum_back(k) + chlp_back(k)
          chlp_ext(k) = (n + 0.5d0) * ( an(k,n) + bn(k,n) )
          csum_ext(k) = csum_ext(k) + chlp_ext(k)
          chlp_sca(k) = (n + 0.5d0) * (ABS(an(k,n))**2 + ABS(bn(k,n))**2 )
          csum_sca(k) = csum_sca(k) + chlp_sca(k)

        END IF
      END DO
    END DO


    WHERE(nue >= 1d-12 .AND. maskeloc)

      C_back = lambda**2 / pi * ABS(csum_back)**2
      C_ext = lambda**2 / pi * DBLE(csum_ext)
      C_sca = lambda**2 / pi * csum_sca

    ELSEWHERE

      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0

    END WHERE

  END SUBROUTINE COATEDSPHERE_SCATTER_VEC

  SUBROUTINE COATEDSPHERE_SCATTERCOEFFS(NMAX,m1,m2,x1,x2,an,bn)

    IMPLICIT NONE

    INTEGER, intent(in) :: NMAX
    COMPLEX(kind=kinddp), INTENT(in) :: m1, m2
    DOUBLE PRECISION, INTENT(in) :: x1, x2
    
    COMPLEX(kind=kinddp), INTENT(out) :: an(1:NMAX), bn(1:NMAX)
    COMPLEX(kind=kinddp), PARAMETER :: imag = (0.0d0,1.0d0), czero = (0.0d0,0.0d0)
    
    COMPLEX(kind=kinddp) :: y1, y2, z2
    COMPLEX(kind=kinddp), DIMENSION(0:NMAX) :: sphj_y1, sphdj_y1, sphy_y1, sphdy_y1, &
         sphj_y2, sphdj_y2, sphy_y2, sphdy_y2, &
         sphj_z2, sphdj_z2, sphy_z2, sphdy_z2, &
         sphj_x2, sphdj_x2, sphy_x2, sphdy_x2
    COMPLEX(kind=kinddp) :: psiy1, psiy2, psiz2, &
                            dpsiy1, dpsiy2, dpsiz2, &
                            xsiy2, xsiz2, dxsiy2, dxsiz2, zetax2, dzetax2, &
                            det1, det2, det3, det4, det5, det6, nenner
    DOUBLE PRECISION :: psix2, dpsix2
    INTEGER :: NM, i

    IF (x1 > x2) THEN
      WRITE (*,*) 'COATEDSPHERE_SCATTERCOEFFS: Fehler: Innerer Kugelradius > Aeusserer Radius!'
    ENDIF

    y1 = m1 * x1
    y2 = m2 * x1
    z2 = m2 * x2


    IF (x1 > 0.0d0) THEN

      ! Zuerst die weiter unten benoetigten Bessel-Funktionen und deren Ableitungen
      ! bis zur Ordnung NMAX berechnen (backward recurrence):
      CALL CSPHJY(NMAX,y1,NM,sphj_y1,sphdj_y1,sphy_y1,sphdy_y1)
      CALL CSPHJY(NMAX,y2,NM,sphj_y2,sphdj_y2,sphy_y2,sphdy_y2)
      CALL CSPHJY(NMAX,z2,NM,sphj_z2,sphdj_z2,sphy_z2,sphdy_z2)
      CALL CSPHJY(NMAX,CMPLX(x2,0.0d0,kind=kinddp),NM,sphj_x2,sphdj_x2,sphy_x2,sphdy_x2)

      DO i=1,NMAX

        ! Die Riccati-Bessel-Funktionen der benoetigten Argumente:

        psiy1 = y1 * sphj_y1(i)
        dpsiy1 = sphj_y1(i) + y1 * sphdj_y1(i)

        psiy2 = y2 * sphj_y2(i)
        dpsiy2 = sphj_y2(i) + y2 * sphdj_y2(i)

        psiz2 = z2 * sphj_z2(i)
        dpsiz2 = sphj_z2(i) + z2 * sphdj_z2(i)

        psix2 = x2 * DBLE(sphj_x2(i))
        dpsix2 = DBLE(sphj_x2(i)) + x2 * DBLE(sphdj_x2(i))

        xsiy2 = -y2 * sphy_y2(i)
        dxsiy2 = -sphy_y2(i) - y2 * sphdy_y2(i)

        xsiz2 = -z2 * sphy_z2(i)
        dxsiz2 = -sphy_z2(i) - z2 * sphdy_z2(i)

        zetax2 = x2 * (DBLE(sphj_x2(i)) - imag * DBLE(sphy_x2(i)))
        dzetax2 = x2 * (DBLE(sphdj_x2(i)) - imag * DBLE(sphdy_x2(i))) + (DBLE(sphj_x2(i)) - imag * DBLE(sphy_x2(i)))

        ! Einige vorbereitende Terme auf die Streukoeffizienten:
     
        det1 = m1*dpsiy2*psiy1 - m2*psiy2*dpsiy1
        det2 = dxsiz2*psix2 - m2*xsiz2*dpsix2
        det3 = m1*dxsiy2*psiy1 - m2*xsiy2*dpsiy1
        det4 = dpsiz2*psix2 - m2*psiz2*dpsix2
        det5 = dxsiz2*zetax2 - m2*xsiz2*dzetax2
        det6 = dpsiz2*zetax2 - m2*psiz2*dzetax2

        nenner = det1*det5-det3*det6

        IF (nenner /= czero) THEN
          an(i) = (det1*det2-det3*det4)/nenner
        ELSE
          an(i) = czero
        END IF

        det1 = m2*dpsiy2*psiy1 - m1*psiy2*dpsiy1
        det2 = m2*dxsiz2*psix2 - xsiz2*dpsix2
        det3 = m2*dxsiy2*psiy1 - m1*xsiy2*dpsiy1
        det4 = m2*dpsiz2*psix2 - psiz2*dpsix2
        det5 = m2*dxsiz2*zetax2 - xsiz2*dzetax2
        det6 = m2*dpsiz2*zetax2 - psiz2*dzetax2

        nenner = det1*det5-det3*det6

        IF (nenner /= czero) THEN
          bn(i) = (det1*det2-det3*det4)/nenner
        ELSE
          bn(i) = czero
        END IF
          

      END DO

    ELSE

      ! Zuerst die weiter unten benoetigten Bessel-Funktionen und deren Ableitungen
      ! bis zur Ordnung NMAX berechnen (backward recurrence):
      CALL CSPHJY(NMAX,z2,NM,sphj_z2,sphdj_z2,sphy_z2,sphdy_z2)
      CALL CSPHJY(NMAX,CMPLX(x2,0.0d0,kind=kinddp),NM,sphj_x2,sphdj_x2,sphy_x2,sphdy_x2)

      DO i=1,NMAX

        ! Die Riccati-Bessel-Funktionen der benoetigten Argumente:

        psiz2 = z2 * sphj_z2(i)
        dpsiz2 = sphj_z2(i) + z2 * sphdj_z2(i)

        psix2 = x2 * DBLE(sphj_x2(i))
        dpsix2 = DBLE(sphj_x2(i)) + x2 * DBLE(sphdj_x2(i))

        zetax2 = x2 * (DBLE(sphj_x2(i)) - imag * DBLE(sphy_x2(i)))
        dzetax2 = x2 * (DBLE(sphdj_x2(i)) - imag * DBLE(sphdy_x2(i))) + (DBLE(sphj_x2(i)) - imag * DBLE(sphy_x2(i)))

        det4 = dpsiz2*psix2 - m2*psiz2*dpsix2
        det6 = dpsiz2*zetax2 - m2*psiz2*dzetax2
      
        IF (det6 /= czero) THEN
          an(i) = det4 / det6
        ELSE
          an(i) = czero
        END IF
      
        det4 = m2*dpsiz2*psix2 - psiz2*dpsix2
        det6 = m2*dpsiz2*zetax2 - psiz2*dzetax2
      
        IF (det6 /= czero) THEN
          bn(i) = det4 / det6
        ELSE
          bn(i) = czero
        END IF

      END DO
      
    END IF
    
    RETURN
    
  END SUBROUTINE COATEDSPHERE_SCATTERCOEFFS

  SUBROUTINE COATEDSPHERE_SCATTER(r,r_shell,m_core,m_shell,m_medium,lambda,C_back,C_ext,C_sca)
    IMPLICIT NONE

    !..Berechnung der Rueckstreueffizienz
    !  r = Radius der Kugel (Gesamtradius, r = r_shell + r_core)
    !  r_shell = Dicke der auesseren Schicht
    !  lambda  = Wellenlaenge


    !..Eingabevariablen
    DOUBLE PRECISION, INTENT(in) :: r,r_shell,lambda,m_medium
    COMPLEX(kind=kinddp),INTENT(in) :: m_core,m_shell

    !..Ausgabevariablen
    DOUBLE PRECISION, intent(out) :: C_sca,C_back,C_ext

    !..Lokale Variablen
    INTEGER, PARAMETER :: NMAX = 50
    INTEGER :: n
    DOUBLE PRECISION :: alf,nue,chlp_sca,csum_sca
    COMPLEX(kind=kinddp) :: an(NMAX),bn(NMAX),chlp_back,chlp_ext,csum_back,csum_ext
    DOUBLE PRECISION, PARAMETER :: eps = 1D-6

    !..Dimensionslose Radien (vgl. Kerker, S. 192, Gl. (5.1.25)
    alf = 2*pi / lambda * (r-r_shell) * m_medium
    nue = 2*pi / lambda * r * m_medium

    IF (nue >= 1d-12) THEN

      !.. Streukoeffizienten bis zur NMAX-ten Ordnung berechnen:
      CALL COATEDSPHERE_SCATTERCOEFFS(NMAX,m_core/m_medium,m_shell/m_medium,alf,nue,an,bn)

      !..Berechung des Rueckstreuquerschnitts 
      !  nach van de Hulst, S. 284
      csum_back = (0.0d0,0.0d0)
      csum_ext = (0.0d0,0.0d0)
      csum_sca = 0.0d0
      chlp_back = (1.0d0,0.0d0)
      chlp_ext = (1.0d0,0.0d0)
      chlp_sca = 1.0d0
      DO n=1,NMAX

        IF ( ABS(chlp_back) >= eps * ABS(csum_back) .OR. &
             ABS(chlp_ext) >= eps * ABS(csum_ext) .OR. &
             ABS(chlp_sca) >= eps * ABS(csum_sca) ) THEN

          chlp_back = (n + 0.5) * (-1)**n * ( an(n) - bn(n) )
          csum_back = csum_back + chlp_back
          chlp_ext = (n + 0.5) * ( an(n) + bn(n) )
          csum_ext = csum_ext + chlp_ext
          chlp_sca = (n + 0.5) * (ABS(an(n))**2 + ABS(bn(n))**2 )
          csum_sca = csum_sca + chlp_sca

        ELSE

          EXIT

        END IF

      ENDDO

      C_back = lambda**2 / pi * ABS(csum_back)**2
      C_ext = lambda**2 / pi * DBLE(csum_ext)
      C_sca = lambda**2 / pi * csum_sca

    ELSE

      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0

   END IF

  END SUBROUTINE COATEDSPHERE_SCATTER



!*********************************************************************************************

  !----------------------------------------------------------------------------------------!
  ! 4) Routinen zur Berechnung von Q_back, Q_ext und Q_sca von verschiedenen
  !    einzelnen Hydrometeoren
  !----------------------------------------------------------------------------------------!


  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr nassen Hagel: Eiskugel mit Wasserhaut
  !
  ! Schwachpunkt: Shedding wird nicht beachtet --- die Wasserhaut ist in ihrer Dicke
  !               somit nicht begrenzt. (noch einbauen!!)
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
!  SUBROUTINE MIE_WATERSPH_WETHAIL_VEC(x_h,fmelt,m_i,m_w,lambda,C_back,anz)
  SUBROUTINE MIE_WATERSPH_WETHAIL_VEC(x_h,fmelt,m_i,m_w,lambda,C_back,C_ext,anz)
! de.


    IMPLICIT NONE
    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in) :: x_h(anz),fmelt(anz),lambda
! de:
!    DOUBLE PRECISION, INTENT(out) :: C_back(anz)
    DOUBLE PRECISION, INTENT(out) :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz),m_w(anz)

    COMPLEX(kind=kinddp) :: m_core(anz),m_shell(anz)
! de:
!    DOUBLE PRECISION :: C_ext(anz),C_sca(anz),D_large(anz),D_core(anz),x_w(anz),fm(anz)
    DOUBLE PRECISION :: C_sca(anz),D_large(anz),D_core(anz),x_w(anz),fm(anz)
! de.
    DOUBLE PRECISION :: x_core(anz),rho_core(anz)
    DOUBLE PRECISION :: x_shell(anz),rho_shell(anz)

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    x_w = x_h * fm

    m_core = m_i
    m_shell = m_w

    rho_core = rho_ice
    rho_shell = rho_w

    x_shell = x_w
    x_core  = x_h-x_w

    D_core  = (6/pi * x_core/rho_core)**(1./3.)
    WHERE (fm >= 1d-10)
      D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE

    !..Mie-Streuung 
    CALL COATEDSPHERE_SCATTER_VEC(D_large*0.5,(D_large-D_core)*0.5,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca,anz)
    
    WHERE (D_large < 2d-12)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE
    
  END SUBROUTINE MIE_WATERSPH_WETHAIL_VEC

  ! Originalfassung:
  SUBROUTINE MIE_WATERSPH_WETHAIL(x_h,fmelt,m_i,m_w,lambda,C_back)


    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: x_h,fmelt,lambda
    DOUBLE PRECISION, INTENT(out) :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_i,m_w

    COMPLEX(kind=kinddp) :: m_core,m_shell
    DOUBLE PRECISION :: C_ext,C_sca,D_large,D_core,x_w,fm
    DOUBLE PRECISION :: x_core,rho_core
    DOUBLE PRECISION :: x_shell,rho_shell

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    x_w = x_h * fm

    m_core = m_i
    m_shell = m_w

    rho_core = rho_ice
    rho_shell = rho_w

    x_shell = x_w
    x_core  = x_h-x_w

    D_core  = (6/pi * x_core/rho_core)**(1./3.)
    IF (fm >= 1d-10) THEN
      D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSE
      D_large = D_core
    END IF

    IF (D_large >= 2d-12) THEN

      !..Mie-Streuung 
      CALL COATEDSPHERE_SCATTER(D_large*0.5,(D_large-D_core)*0.5,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca)

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_WATERSPH_WETHAIL

  ! Vektorisierte Fassung:
! de: include extinction coefficient
!  SUBROUTINE MIE_WATERSPH_WETHAIL_BH_VEC(x_h,fmelt,m_i,m_w,lambda,C_back,anz)
  SUBROUTINE MIE_WATERSPH_WETHAIL_BH_VEC(x_h,fmelt,m_i,m_w,lambda,C_back,C_ext,anz)
! de.


    IMPLICIT NONE

    INTEGER , intent(in) :: anz
    DOUBLE PRECISION, INTENT(in) :: x_h(anz),fmelt(anz),lambda
! de:
!    DOUBLE PRECISION, INTENT(out) :: C_back(anz)
    DOUBLE PRECISION, INTENT(out) :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz),m_w(anz)

    COMPLEX(kind=kinddp) :: m_core(anz),m_shell(anz)
! de:
!    DOUBLE PRECISION :: C_ext(anz),C_sca(anz),D_large(anz),D_core(anz),x_w(anz),fm(anz)
    DOUBLE PRECISION :: C_sca(anz),D_large(anz),D_core(anz),x_w(anz),fm(anz)
! de.
    DOUBLE PRECISION :: x_core(anz),rho_core(anz)
    DOUBLE PRECISION :: x_shell(anz),rho_shell(anz)

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    x_w = x_h * fm

    m_core = m_i
    m_shell = m_w

    rho_core = rho_ice
    rho_shell = rho_w

    x_shell = x_w
    x_core  = x_h-x_w

    D_core  = (6/pi * x_core/rho_core)**(1./3.)
    WHERE (fm >= 1d-10)
      D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE

    !..Mie-Streuung 
    CALL COATEDSPHERE_SCATTER_BH_VEC(D_large*0.5,(D_large-D_core)*0.5,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca,anz)

    WHERE (D_large < 2d-12)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

  END SUBROUTINE MIE_WATERSPH_WETHAIL_BH_VEC

  ! Originalfassung:
  SUBROUTINE MIE_WATERSPH_WETHAIL_BH(x_h,fmelt,m_i,m_w,lambda,C_back)


    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: x_h,fmelt,lambda
    DOUBLE PRECISION, INTENT(out) :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_i,m_w

    COMPLEX(kind=kinddp) :: m_core,m_shell
    DOUBLE PRECISION :: C_ext,C_sca,D_large,D_core,x_w,fm
    DOUBLE PRECISION :: x_core,rho_core
    DOUBLE PRECISION :: x_shell,rho_shell

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    x_w = x_h * fm

    m_core = m_i
    m_shell = m_w

    rho_core = rho_ice
    rho_shell = rho_w

    x_shell = x_w
    x_core  = x_h-x_w

    D_core  = (6/pi * x_core/rho_core)**(1./3.)
    IF (fm >= 1d-10) THEN
      D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSE
      D_large = D_core
    END IF

    IF (D_large >= 1d-12) THEN

      !..Mie-Streuung 
      CALL COATEDSPHERE_SCATTER_BH(D_large*0.5,(D_large-D_core)*0.5,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca)

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_WATERSPH_WETHAIL_BH

  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr spongy Hagel (Eiskern + Wasser/Eis-Schale mit Schmelzgrad fmelt;
  ! Das Schmelzwasser verteilt sich auf eine Schale, die ihrerseits einen Wasser-Volumenanteil
  ! von f_water hat. f_water darf nicht 0.0 oder kleiner sein!!!
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_SPONGY_WETHAIL_VEC(x_h,fmelt,f_water,m_i,m_w,lambda,C_back, C_ext, &
                                    mixingrulestring, matrixstring, inclusionstring, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_h(anz),fmelt(anz),f_water(anz),lambda
! de:
    DOUBLE PRECISION, INTENT(out), DIMENSION(anz)    :: C_back,C_ext
! de.
    COMPLEX(kind=kinddp), INTENT(in), DIMENSION(anz) :: m_i,m_w
    CHARACTER(len=*) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_core,m_shell,m_airv
! de:
    DOUBLE PRECISION, DIMENSION(anz) :: r_large,f_mass,f_ice,C_sca,D_large,D_core,x_w,fm
! de.
    DOUBLE PRECISION, DIMENSION(anz) :: C_ext1,C_sca1,C_back1
    DOUBLE PRECISION, DIMENSION(anz) :: rho_core,f_core,K_core,r_core,x_core
    DOUBLE PRECISION, DIMENSION(anz) :: rho_shell,D_shell,x_shell,r_shell,K_shell
    DOUBLE PRECISION, DIMENSION(anz) :: volice, volair, volwater, fwater, volh
    INTEGER :: fehler
    LOGICAL :: maske(anz)

    fehler = 0

    m_airv = m_air

    !..Masse des insgesamt geschmolzenen Wassers:
    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    x_w = x_h * fm

    !..Volumenanteil des Wassers in der Spongy Schale:
    fwater = MAX(MIN(f_water, 1.0d0), 0.0d0)
    !..Volumenanteil des Eises in der Spongy Schale:
    f_ice = 1.0 - fwater

    m_core = m_i

    rho_shell = rho_w * fwater + rho_ice * f_ice
    rho_core  = rho_ice

    f_mass = rho_w / rho_shell * fwater

    ! Masse der Schale: Fallunterscheidung notwendig, weil 
    ! im Falle f_water < 1 schon bei einem Schmelzgrad von fmelt < 1 vorkommt, dass
    ! das gesamte Partikel aus einem Eis-Wasser-Gemisch besteht, so dass
    ! der Kern komplett weg ist und sich die aeussere Schale ueber das komplette
    ! Partikel erstreckt. Bei weiterem Schmelzen veraendert sich dann der relative
    ! Volumenanteil des Wassers, f_water.

    x_shell = x_w / f_mass

    !.. 1) Fall x_shell < x_h
    maske = (x_shell < x_h)
    WHERE(maske)
      x_core  = x_h - x_shell
    ELSEWHERE
      x_core  = 0.0d0
    END WHERE

    D_core  = (6./pi * x_core/rho_core)**(1./3.)
    WHERE (fm >= 1d-10)
      D_large = (6./pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE
        
    r_large = D_large * 0.5
    r_core  = D_core * 0.5
    r_shell = r_large - r_core
        
    !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
    volwater = fwater  ! Volumenanteil des Wassers an der Eis/Wasser-Mischung
    volice = 1.0d0 - volwater
    volair = 0.0d0
        
    m_shell = get_m_mix_vec(m_airv, m_i, m_w, volair, volice, volwater, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) RETURN
        
    !..Mie-Streuung 
    maske = (x_shell < x_h .AND. D_large >= 2d-12)
    CALL COATEDSPHERE_SCATTER_VEC(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca,anz,maske=maske)

    WHERE (.NOT.maske)
      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0
    END WHERE

    !.. 2) Fall x_shell = x_h

    volh = (x_h - x_w) / rho_ice + x_w / rho_w
      
    D_large  = (6.0 / pi * volh) ** (1./3.)

    volice = (x_h - x_w) / (volh * rho_ice)
    volwater = 1.0d0 - volice
    volair = 0.0d0
    
    m_core = get_m_mix_vec(m_airv, m_i, m_w, volair, volice, volwater, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) RETURN
        
    !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(MAX(D_large*0.5,1d-12),m_core,DBLE(m_air),lambda,C_back1,C_ext1,C_sca1, anz)        

    WHERE (D_large >= 2d-12 .AND. x_shell >= x_h)
      C_back = C_back1
      C_ext = C_ext1
      C_sca = C_sca1
    END WHERE

  END SUBROUTINE MIE_SPONGY_WETHAIL_VEC

  ! Originalfassung:
  SUBROUTINE MIE_SPONGY_WETHAIL(x_h,fmelt,f_water,m_i,m_w,lambda,C_back, &
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_h,fmelt,f_water,lambda
    DOUBLE PRECISION, INTENT(out)    :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_i,m_w
    CHARACTER(len=*) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_core,m_shell
    DOUBLE PRECISION :: r_large,f_mass,f_ice,C_ext,C_sca,D_large,D_core,x_w,fm
    DOUBLE PRECISION :: rho_core,f_core,K_core,r_core,x_core
    DOUBLE PRECISION :: rho_shell,D_shell,x_shell,r_shell,K_shell
    DOUBLE PRECISION :: volice, volair, volwater, fwater, volh
    INTEGER :: fehler

    fehler = 0

    C_back = 0.0d0
    C_ext = 0.0d0
    C_sca = 0.0d0

    !..Masse des insgesamt geschmolzenen Wassers:
    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    x_w = x_h * fm

    !..Volumenanteil des Wassers in der Spongy Schale:
    fwater = MAX(MIN(f_water, 1.0d0), 0.0d0)
    !..Volumenanteil des Eises in der Spongy Schale:
    f_ice = 1.0 - fwater

    m_core = m_i

    rho_shell = rho_w * fwater + rho_ice * f_ice
    rho_core  = rho_ice

    f_mass = rho_w / rho_shell * fwater

    ! Masse der Schale: Fallunterscheidung notwendig, weil 
    ! im Falle f_water < 1 schon bei einem Schmelzgrad von fmelt < 1 vorkommt, dass
    ! das gesamte Partikel aus einem Eis-Wasser-Gemisch besteht, so dass
    ! der Kern komplett weg ist und sich die aeussere Schale ueber das komplette
    ! Partikel erstreckt. Bei weiterem Schmelzen veraendert sich dann der relative
    ! Volumenanteil des Wassers, f_water.

    x_shell = x_w / f_mass
    IF (x_shell < x_h) THEN

      x_core  = x_h - x_shell

      D_core  = (6./pi * x_core/rho_core)**(1./3.)
      IF (fm >= 1d-10) THEN
        D_large = (6./pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
      ELSE
        D_large = D_core
      END IF

      IF (D_large >= 1d-12) THEN
        
        r_large = D_large * 0.5
        r_core  = D_core * 0.5
        r_shell = r_large - r_core
        
        !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
        volwater = fwater  ! Volumenanteil des Wassers an der Eis/Wasser-Mischung
        volice = 1.0d0 - volwater
        volair = 0.0d0
        
        m_shell = get_m_mix(m_air, m_i, m_w, volair, volice, volwater, &
             mixingrulestring, matrixstring, inclusionstring, fehler)
        IF (fehler /= 0) RETURN
        
        !..Mie-Streuung 
        CALL COATEDSPHERE_SCATTER(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca)

      END IF

    ELSE

      volh = (x_h - x_w) / rho_ice + x_w / rho_w

      D_large  = (6.0 / pi * volh) ** (1./3.)

      IF (D_large >= 1d-12) THEN

        volice = (x_h - x_w) / (volh * rho_ice)
        volwater = 1.0d0 - volice
        volair = 0.0d0

        m_core = get_m_mix(m_air, m_i, m_w, volair, volice, volwater, &
             mixingrulestring, matrixstring, inclusionstring, fehler)
        IF (fehler /= 0) RETURN
        
         !..Mie-Streuung 
        CALL SPHERE_SCATTER_BH(D_large*0.5,m_core,DBLE(m_air),lambda,C_back,C_ext,C_sca)        

      END IF

    END IF

  END SUBROUTINE MIE_SPONGY_WETHAIL

  ! Version mit der Berechnung der Streufunktion mittels den Rekursionsformeln von Bohren und Huffman (1983):

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_SPONGY_WETHAIL_BH_VEC(x_h,fmelt,f_water,m_i,m_w,lambda,C_back, C_ext, &
                                       mixingrulestring, matrixstring, inclusionstring, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_h(anz),fmelt(anz),f_water(anz),lambda
! de:
    DOUBLE PRECISION, INTENT(out), DIMENSION(anz)    :: C_back,C_ext
! de.
    COMPLEX(kind=kinddp), INTENT(in), DIMENSION(anz) :: m_i,m_w
    CHARACTER(len=*) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_core,m_shell,m_airv
! de:
    DOUBLE PRECISION, DIMENSION(anz) :: r_large,f_mass,f_ice,C_sca,D_large,D_core,x_w,fm
! de.
    DOUBLE PRECISION, DIMENSION(anz) :: C_ext1,C_sca1,C_back1
    DOUBLE PRECISION, DIMENSION(anz) :: rho_core,f_core,K_core,r_core,x_core
    DOUBLE PRECISION, DIMENSION(anz) :: rho_shell,D_shell,x_shell,r_shell,K_shell
    DOUBLE PRECISION, DIMENSION(anz) :: volice, volair, volwater, fwater, volh
    INTEGER :: fehler, i
    LOGICAL :: maske(anz)

    fehler = 0

    m_airv = m_air

    !..Masse des insgesamt geschmolzenen Wassers:
    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    x_w = x_h * fm

    !..Volumenanteil des Wassers in der Spongy Schale:
    fwater = MAX(MIN(f_water, 1.0d0), 0.0d0)
    !..Volumenanteil des Eises in der Spongy Schale:
    f_ice = 1.0 - fwater

    m_core = m_i

    rho_shell = rho_w * fwater + rho_ice * f_ice
    rho_core  = rho_ice

    f_mass = rho_w / rho_shell * fwater

    ! Masse der Schale: Fallunterscheidung notwendig, weil 
    ! im Falle f_water < 1 schon bei einem Schmelzgrad von fmelt < 1 vorkommt, dass
    ! das gesamte Partikel aus einem Eis-Wasser-Gemisch besteht, so dass
    ! der Kern komplett weg ist und sich die aeussere Schale ueber das komplette
    ! Partikel erstreckt. Bei weiterem Schmelzen veraendert sich dann der relative
    ! Volumenanteil des Wassers, f_water.

    x_shell = x_w / f_mass

    !.. 1) Fall x_shell < x_h

    WHERE (x_shell < x_h)
      x_core  = x_h - x_shell
    ELSEWHERE
      x_core  = 0.0d0
    END WHERE

    D_core  = (6./pi * x_core/rho_core)**(1./3.)

    WHERE (fm >= 1d-10)
      D_large = (6./pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE
        
    r_large = D_large * 0.5
    r_core  = D_core * 0.5
    r_shell = r_large - r_core
        
    !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
    volwater = fwater  ! Volumenanteil des Wassers an der Eis/Wasser-Mischung
    volice = 1.0d0 - volwater
    volair = 0.0d0
        
    m_shell = get_m_mix_vec(m_airv, m_i, m_w, volair, volice, volwater, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) RETURN
        
    !..Mie-Streuung 
    CALL COATEDSPHERE_SCATTER_BH_VEC(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca,anz)

    WHERE (x_shell >= x_h .OR. D_large < 2d-12)
      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0
    END WHERE

    !.. 2) Fall x_shell = x_h

    volh = (x_h - x_w) / rho_ice + x_w / rho_w
      
    D_large  = (6.0 / pi * volh) ** (1./3.)

    volice = (x_h - x_w) / (volh * rho_ice)
    volwater = 1.0d0 - volice
    volair = 0.0d0
    
    m_core = get_m_mix_vec(m_airv, m_i, m_w, volair, volice, volwater, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) RETURN
        
    !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(MAX(D_large*0.5,1d-12),m_core,DBLE(m_air),lambda,C_back1,C_ext1,C_sca1, anz)        

    WHERE (D_large >= 2d-12 .AND. x_shell >= x_h )
      C_back = C_back1
      C_ext = C_ext1
      C_sca = C_sca1
    END WHERE

  END SUBROUTINE MIE_SPONGY_WETHAIL_BH_VEC

  ! Originalfassung:
  SUBROUTINE MIE_SPONGY_WETHAIL_BH(x_h,fmelt,f_water,m_i,m_w,lambda,C_back, &
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_h,fmelt,f_water,lambda
    DOUBLE PRECISION, INTENT(out)    :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_i,m_w
    CHARACTER(len=*) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_core,m_shell
    DOUBLE PRECISION :: r_large,f_mass,f_ice,C_ext,C_sca,D_large,D_core,x_w
    DOUBLE PRECISION :: rho_core,f_core,K_core,r_core,x_core
    DOUBLE PRECISION :: rho_shell,D_shell,x_shell,r_shell,K_shell
    DOUBLE PRECISION :: volice, volair, volwater, fwater, volh
    INTEGER :: fehler

    fehler = 0

    C_back = 0.0d0
    C_ext = 0.0d0
    C_sca = 0.0d0

    !..Masse des insgesamt geschmolzenen Wassers:
    x_w = x_h * MAX(MIN(fmelt, 1.0d0), 0.0d0)

    !..Volumenanteil des Wassers in der Spongy Schale:
    fwater = MAX(MIN(f_water, 1.0d0), 0.0d0)
    !..Volumenanteil des Eises in der Spongy Schale:
    f_ice = 1.0 - fwater

    m_core = m_i

    rho_shell = rho_w * fwater + rho_ice * f_ice
    rho_core  = rho_ice

    f_mass = rho_w / rho_shell * fwater

    ! Masse der Schale: Fallunterscheidung notwendig, weil 
    ! im Falle f_water < 1 schon bei einem Schmelzgrad von fmelt < 1 vorkommt, dass
    ! das gesamte Partikel aus einem Eis-Wasser-Gemisch besteht, so dass
    ! der Kern komplett weg ist und sich die aeussere Schale ueber das komplette
    ! Partikel erstreckt. Bei weiterem Schmelzen veraendert sich dann der relative
    ! Volumenanteil des Wassers, f_water.

    x_shell = x_w / f_mass
    IF (x_shell < x_h) THEN

      x_core  = x_h - x_shell

      D_core  = (6./pi * x_core/rho_core)**(1./3.)
      D_large = (6./pi * x_shell/rho_shell + D_core**3.0)**(1./3.)

      IF (D_large >= 1d-12) THEN
        
        r_large = D_large * 0.5
        r_core  = D_core * 0.5
        r_shell = r_large - r_core
        
        !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
        volwater = fwater  ! Volumenanteil des Wassers an der Eis/Wasser-Mischung
        volice = 1.0d0 - volwater
        volair = 0.0d0
        
        m_shell = get_m_mix(m_air, m_i, m_w, volair, volice, volwater, &
             mixingrulestring, matrixstring, inclusionstring, fehler)
        IF (fehler /= 0) RETURN
        
        !..Mie-Streuung 
        CALL COATEDSPHERE_SCATTER_BH(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca)

      END IF

    ELSE

      volh = (x_h - x_w) / rho_ice + x_w / rho_w

      D_large  = (6.0 / pi * volh) ** (1./3.)

      IF (D_large >= 1d-12) THEN

        volice = (x_h - x_w) / (volh * rho_ice)
        volwater = 1.0d0 - volice
        volair = 0.0d0

        m_core = get_m_mix(m_air, m_i, m_w, volair, volice, volwater, &
             mixingrulestring, matrixstring, inclusionstring, fehler)
        IF (fehler /= 0) RETURN
        
         !..Mie-Streuung 
        CALL SPHERE_SCATTER_BH(D_large*0.5,m_core,DBLE(m_air),lambda,C_back,C_ext,C_sca)        

      END IF

    END IF

  END SUBROUTINE MIE_SPONGY_WETHAIL_BH

  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr trockenen Hagel: Eiskugel
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_DRYHAIL_VEC(x_h,a_geo,b_geo,m_i,lambda,C_back,C_ext,anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in) :: x_h(anz),a_geo,b_geo,lambda
! de:
    DOUBLE PRECISION, INTENT(out) :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz)

! de:
    DOUBLE PRECISION :: C_sca(anz),D_large(anz)
! de.

    D_large = a_geo * x_h ** b_geo

    !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(D_large*0.5,m_i,DBLE(m_air),lambda,C_back,C_ext,C_sca,anz)        
    
    WHERE (D_large < 1d-8)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

  END SUBROUTINE MIE_DRYHAIL_VEC

  ! Originalfassung:
  SUBROUTINE MIE_DRYHAIL(x_h,a_geo,b_geo,m_i,lambda,C_back)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in) :: x_h,a_geo,b_geo,lambda
    DOUBLE PRECISION, INTENT(out) :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_i

    DOUBLE PRECISION :: C_ext,C_sca,D_large

    D_large = a_geo * x_h ** b_geo

    IF (D_large >= 1d-8) THEN

      !..Mie-Streuung 
      CALL SPHERE_SCATTER_BH(D_large*0.5,m_i,DBLE(m_air),lambda,C_back,C_ext,C_sca)        

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_DRYHAIL

  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr nassen Graupel/Hagel: Zweischalenmodell, wobei sich
  ! das am Aussenrand des Eis/Luft-Kernes schmelzende Wasser am Aussenrand des Partikels
  ! als Wasserhaut ansammelt. Der Brechungsindex des Kerns wird nach Mischungsregeln 
  ! berechnet, die durch die Schluesselworte mixingrulestring, matrixstring, inclusionstring
  ! definiert werden.
  !
  ! Schwachpunkt: Shedding wird nicht beachtet --- die Wasserhaut ist in ihrer Dicke
  !               somit nicht begrenzt. (noch einbauen!!)
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_WATERSPH_WETGR_VEC(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,C_ext, &
                                    mixingrulestring, matrixstring, inclusionstring, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_g(anz), a_geo, b_geo, fmelt(anz), lambda
! de:
    DOUBLE PRECISION, INTENT(out)    :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz), m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_core(anz), m_airv(anz)
! de:
    DOUBLE PRECISION, DIMENSION(anz) :: D_g,r_large,C_sca,D_large,D_core,x_w,fm
! de.
    DOUBLE PRECISION, DIMENSION(anz) :: rho_core,f_core,x_core,fzero
    DOUBLE PRECISION , DIMENSION(anz):: rho_shell,x_shell,volice,volair
    INTEGER :: fehler

    m_airv = m_air
    fzero = 0.0d0

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm
    rho_shell = rho_w

    D_g = MAX(a_geo * x_g**b_geo, 1d-20)

    rho_core  = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)

    f_core = rho_core/rho_ice

    x_shell = x_w
    x_core  = x_g - x_w
    
    D_core  = (6/pi * x_core/rho_core)**(1./3.)
    WHERE (fm >= 1d-6)
      D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE
    
    !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
    m_core = get_m_mix_vec(m_airv, m_i, m_w, (1.0d0-f_core), f_core, fzero, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF
    
    !..Mie-Streuung 
    CALL COATEDSPHERE_SCATTER_VEC(MAX(D_large*0.5,1d-12),&
         (D_large-D_core)*0.5,m_core,m_w,1.0d0,lambda,C_back,C_ext,C_sca,anz)

    WHERE (D_g < 2d-12)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

  END SUBROUTINE MIE_WATERSPH_WETGR_VEC

  ! Originalfassung:
  SUBROUTINE MIE_WATERSPH_WETGR(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,&
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_g, a_geo, b_geo, fmelt, lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_w, m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_core
    DOUBLE PRECISION :: D_g,r_large,C_ext,C_sca,D_large,D_core,x_w,fm
    DOUBLE PRECISION :: rho_core,f_core,x_core
    DOUBLE PRECISION :: rho_shell,x_shell,volice,volair
    INTEGER :: fehler

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm
    rho_shell = rho_w

    D_g = a_geo * x_g**b_geo

    IF (D_g >= 1d-12) THEN
      rho_core  = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)

      f_core = rho_core/rho_ice

      x_shell = x_w
      x_core  = x_g - x_w

      D_core  = (6/pi * x_core/rho_core)**(1./3.)
      IF (fm >= 1d-6) THEN
        D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
      ELSE
        D_large = D_core
      END IF

      !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
      m_core = get_m_mix(m_air, m_i, m_w, (1.0d0-f_core), f_core, 0.0d0, &
           mixingrulestring, matrixstring, inclusionstring, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF


      !..Mie-Streuung 
      CALL COATEDSPHERE_SCATTER(D_large*0.5,(D_large-D_core)*0.5,m_core,m_w,1.0d0,lambda,C_back,C_ext,C_sca)

    ELSE
      
      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_WATERSPH_WETGR

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_WATERSPH_WETGR_BH_VEC(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,C_ext, &
                                       mixingrulestring, matrixstring, inclusionstring, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_g(anz), a_geo, b_geo, fmelt(anz), lambda
! de:
    DOUBLE PRECISION, INTENT(out)    :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz), m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_core(anz), m_airv(anz)
! de:
    DOUBLE PRECISION, DIMENSION(anz) :: D_g,r_large,C_sca,D_large,D_core,x_w,fm
! de.
    DOUBLE PRECISION, DIMENSION(anz) :: rho_core,f_core,x_core,fzero
    DOUBLE PRECISION , DIMENSION(anz):: rho_shell,x_shell,volice,volair
    INTEGER :: fehler

    m_airv = m_air
    fzero = 0.0d0

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm
    rho_shell = rho_w

    D_g = MAX(a_geo * x_g**b_geo, 1d-20)

    rho_core  = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)

    f_core = rho_core/rho_ice

    x_shell = x_w
    x_core  = x_g - x_w
    
    D_core  = (6/pi * x_core/rho_core)**(1./3.)
    WHERE (fm >= 1d-6)
      D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE
    
    !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
    m_core = get_m_mix_vec(m_airv, m_i, m_w, (1.0d0-f_core), f_core, fzero, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF
    
    !..Mie-Streuung 
    CALL COATEDSPHERE_SCATTER_BH_VEC(D_large*0.5,(D_large-D_core)*0.5,m_core,m_w,1.0d0,lambda,C_back,C_ext,C_sca,anz)

    WHERE (D_g < 2d-12)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

  END SUBROUTINE MIE_WATERSPH_WETGR_BH_VEC

  ! Originalfassung:
  SUBROUTINE MIE_WATERSPH_WETGR_BH(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,&
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_g, a_geo, b_geo, fmelt, lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_w, m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_core
    DOUBLE PRECISION :: D_g,r_large,C_ext,C_sca,D_large,D_core,x_w,fm
    DOUBLE PRECISION :: rho_core,f_core,x_core
    DOUBLE PRECISION :: rho_shell,x_shell,volice,volair
    INTEGER :: fehler

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm
    rho_shell = rho_w

    D_g = a_geo * x_g**b_geo

    IF (D_g >= 1d-12) THEN
      rho_core  = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)

      f_core = rho_core/rho_ice

      x_shell = x_w
      x_core  = x_g - x_w

      D_core  = (6/pi * x_core/rho_core)**(1./3.)
      IF (fm >= 1d-6) THEN
        D_large = (6/pi * x_shell/rho_shell + D_core**3.0)**(1./3.)
      ELSE
        D_large = D_core
      END IF


      !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
      m_core = get_m_mix(m_air, m_i, m_w, (1.0d0-f_core), f_core, 0.0d0, &
           mixingrulestring, matrixstring, inclusionstring, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF


      !..Mie-Streuung 
      CALL COATEDSPHERE_SCATTER_BH(D_large*0.5,(D_large-D_core)*0.5,m_core,m_w,1.0d0,lambda,C_back,C_ext,C_sca)

    ELSE
      
      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_WATERSPH_WETGR_BH

  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr nassen Graupel/Hagel: Zweischalenmodell, wobei das
  ! aussen schmelzende Wasser sukzessive die Poren in der aeusseren Kugelschale vollstaendig 
  ! fuellt. Die aeussere Schale besteht somit aus Eis und Wasser, waehrend die
  ! innere Schale aus Eis und Luft besteht. Solange nicht alle Hohlraeume mit Schmelzwasser
  ! gefuellt sind, schmilzt das Partikel nur von aussen, d.h. das verbleibende Eisskellet
  ! im inneren schmilzt zunaechst nicht mit. Erst wenn alle Hohlraeume voll sind, schmilzt
  ! auch das Eisskellet.
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_SOAK_TWOSPH_WETGR_VEC(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,C_ext, &
                    mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
                    mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_g(anz), a_geo, b_geo, fmelt(anz), lambda
! de:
    DOUBLE PRECISION, INTENT(out)     :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz), m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater

    DOUBLE PRECISION, DIMENSION(anz)     :: D_large, D_core
    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_core, m_shell, m_airv

! de:
    DOUBLE PRECISION, DIMENSION(anz) :: D_g,r_large,C_sca, x_w, fm
! de.
    DOUBLE PRECISION, DIMENSION(anz) :: C_ext1,C_sca1, C_back1
    DOUBLE PRECISION, DIMENSION(anz) :: rho_core,r_core,x_core, f_ice, f_water, fzero
    DOUBLE PRECISION, DIMENSION(anz) :: rho_shell,f_shell,x_shell,r_shell,D_c3
    INTEGER :: fehler
    LOGICAL :: maske(anz)

    fzero = 0.0d0
    m_airv = m_air

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm

    D_g = MAX(a_geo * x_g**b_geo, 1d-20)

    rho_core = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
    f_ice    = rho_core / rho_ice
    
    f_water   = 1.0 - f_ice
    rho_shell = rho_ice * f_ice + rho_w * f_water
    
    D_c3 = 6.0/pi * x_g/rho_core * (1.0 - (rho_shell*x_w) / (f_water*rho_w*x_g))

    !.. 1) Fall D_c3 >= 0.d0:
    !      (noch sind nicht alle Hohlraeume mit Wasser gefuellt. Die Partialdichte
    !      des Eises bleibt zunaechst einmal konstant, das verbleibende Eisskellett schmilzt noch
    !      nicht mit)

    D_core  = D_c3**(1./3.)
    WHERE (fm >= 1d-6)
      D_large = (6.0/pi * x_w/rho_w/f_water + D_core**3)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE

    r_large = D_large * 0.5d0
    r_core  = MAX(D_core  * 0.5d0, 0.0d0)
    r_shell = r_large - r_core
    
    !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
    m_core = get_m_mix_vec(m_airv, m_i, m_w, (1.0d0-f_ice), f_ice, fzero, &
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF
    
    !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
    m_shell = get_m_mix_vec(m_airv, m_i, m_w, fzero, f_ice, (1.0d0-f_ice), &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

    !..Mie-Streuung
    maske = (D_c3 >= 0.0d0 .AND. D_g >= 2d-12)
    CALL COATEDSPHERE_SCATTER_VEC(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca,anz, maske=maske)

    WHERE (.NOT.maske)
      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0
    END WHERE

    !.. 2) Fall D_c3 < 0.d0:
    !      Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:

    f_ice   = (1.0d0 - fm) / (rho_ice * (fm/rho_w + (1.0d0-fm)/rho_ice))
    f_water = 1.0d0 - f_ice
    
    rho_shell = rho_ice * f_ice + rho_w * f_water

    D_large = (6/pi * x_g/rho_shell)**(1./3.)

    !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
    m_shell = get_m_mix_vec(m_airv, m_i, m_w, fzero, f_ice, f_water, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

    !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(MAX(D_large*0.5,1d-12),m_shell,DBLE(m_air),lambda,C_back1,C_ext1,C_sca1,anz)        
   
    WHERE (D_large >= 2d-12 .AND. D_c3 < 0.0d0)
      C_back = C_back1
      C_ext = C_ext1
      C_sca = C_sca1
    END WHERE

  END SUBROUTINE MIE_SOAK_TWOSPH_WETGR_VEC

  ! Originalfassung:
  SUBROUTINE MIE_SOAK_TWOSPH_WETGR(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
       mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_g, a_geo, b_geo, fmelt, lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_w, m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater

    DOUBLE PRECISION     :: D_large, D_core
    COMPLEX(kind=kinddp) :: m_core, m_shell

    DOUBLE PRECISION :: D_g,r_large,C_ext,C_sca, x_w, fm
    DOUBLE PRECISION :: rho_core,r_core,x_core, f_ice, f_water
    DOUBLE PRECISION :: rho_shell,f_shell,x_shell,r_shell,D_c3
    INTEGER :: fehler

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm

    D_g = a_geo * x_g**b_geo

    IF (D_g >= 1d-12) THEN

      rho_core = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
      f_ice    = rho_core / rho_ice

      f_water   = 1.0 - f_ice
      rho_shell = rho_ice * f_ice + rho_w * f_water

      D_c3 = 6.0/pi * x_g/rho_core * (1.0 - (rho_shell*x_w) / (f_water*rho_w*x_g))

      IF (D_c3 .GE. 0.d0) THEN
        ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. Die Partialdichte
        ! des Eises bleibt zunaechst einmal konstant (das verbleibende Eisskellett schmilzt noch
        ! nicht mit):
        D_core  = D_c3**(1./3.)
        IF (fm >= 1d-6) THEN
          D_large = (6.0/pi * x_w/rho_w/f_water + D_core**3)**(1./3.)
        ELSE
          D_large = D_core
        END IF

        r_large = D_large / 2.0
        r_core  = D_core  / 2.0
        r_shell = r_large - r_core

        !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
        m_core = get_m_mix(m_air, m_i, m_w, (1.0d0-f_ice), f_ice, 0.0d0, &
             mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, fehler)
        IF (fehler /= 0) THEN
          C_back = 0.0d0
          RETURN
        END IF

        !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
        m_shell = get_m_mix(m_air, m_i, m_w, 0.0d0, f_ice, (1.0d0-f_ice), &
             mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, fehler)
        IF (fehler /= 0) THEN
          C_back = 0.0d0
          RETURN
        END IF

        !..Mie-Streuung 
        CALL COATEDSPHERE_SCATTER(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca)

      ELSE

        ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
        f_ice   = (1.0d0 - fm) / (rho_ice * (fm/rho_w + (1.0d0-fm)/rho_ice))
        f_water = 1.0d0 - f_ice

        rho_shell = rho_ice * f_ice + rho_w * f_water

        D_large = (6/pi * x_g/rho_shell)**(1./3.)

        !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
        m_shell = get_m_mix(m_air, m_i, m_w, 0.0d0, f_ice, f_water, &
             mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, fehler)
        IF (fehler /= 0) THEN
          C_back = 0.0d0
          RETURN
        END IF

        !..Mie-Streuung 
        CALL SPHERE_SCATTER_BH(D_large*0.5,m_shell,DBLE(m_air),lambda,C_back,C_ext,C_sca)        

      ENDIF

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_SOAK_TWOSPH_WETGR

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_SOAK_TWOSPH_WETGR_BH_VEC(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,C_ext, &
                       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
                       mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_g(anz), a_geo, b_geo, fmelt(anz), lambda
! de:
    DOUBLE PRECISION, INTENT(out)     :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz), m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater

    DOUBLE PRECISION, DIMENSION(anz)     :: D_large, D_core
    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_core, m_shell, m_airv

! de:
    DOUBLE PRECISION, DIMENSION(anz) :: D_g,r_large,C_sca, x_w, fm
! de.
    DOUBLE PRECISION, DIMENSION(anz) :: C_ext1,C_sca1, C_back1
    DOUBLE PRECISION, DIMENSION(anz) :: rho_core,r_core,x_core, f_ice, f_water, fzero
    DOUBLE PRECISION, DIMENSION(anz) :: rho_shell,f_shell,x_shell,r_shell,D_c3
    INTEGER :: fehler

    fzero = 0.0d0
    m_airv = m_air

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm

    D_g = a_geo * x_g**b_geo

    rho_core = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
    f_ice    = rho_core / rho_ice
    
    f_water   = 1.0 - f_ice
    rho_shell = rho_ice * f_ice + rho_w * f_water
    
    D_c3 = 6.0/pi * x_g/rho_core * (1.0 - (rho_shell*x_w) / (f_water*rho_w*x_g))

    !.. 1) Fall D_c3 >= 0.d0:
    !      (noch sind nicht alle Hohlraeume mit Wasser gefuellt. Die Partialdichte
    !      des Eises bleibt zunaechst einmal konstant, das verbleibende Eisskellett schmilzt noch
    !      nicht mit)

    WHERE (D_c3 >= 0.0d0)
      D_core  = D_c3**(1./3.)
    ELSEWHERE
      D_core = 0.0d0
    END WHERE

    WHERE (fm >= 1d-6)
      D_large = (6.0/pi * x_w/rho_w/f_water + D_core**3)**(1./3.)
    ELSEWHERE
      D_large = D_core
    END WHERE

    r_large = D_large * 0.5d0
    r_core  = D_core  * 0.5d0
    r_shell = r_large - r_core
    
    !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
    m_core = get_m_mix_vec(m_airv, m_i, m_w, (1.0d0-f_ice), f_ice, fzero, &
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF
    
    !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
    m_shell = get_m_mix_vec(m_airv, m_i, m_w, fzero, f_ice, (1.0d0-f_ice), &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

    !..Mie-Streuung
    CALL COATEDSPHERE_SCATTER_BH_VEC(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca,anz)

    WHERE (D_c3 < 0.0d0 .OR. D_large < 2d-12)
      C_back = 0.0d0
      C_ext = 0.0d0
      C_sca = 0.0d0
    END WHERE

    !.. 2) Fall D_c3 < 0.d0:
    !      Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:

    f_ice   = (1.0d0 - fm) / (rho_ice * (fm/rho_w + (1.0d0-fm)/rho_ice))
    f_water = 1.0d0 - f_ice
    
    rho_shell = rho_ice * f_ice + rho_w * f_water

    D_large = (6/pi * x_g/rho_shell)**(1./3.)

    !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
    m_shell = get_m_mix_vec(m_airv, m_i, m_w, fzero, f_ice, f_water, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

    !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(D_large*0.5,m_shell,DBLE(m_air),lambda,C_back1,C_ext1,C_sca1,anz)        
   
    WHERE (D_large >= 2d-12 .AND. D_c3 < 0.0d0 )
      C_back = C_back1
      C_ext = C_ext1
      C_sca = C_sca1
    END WHERE

  END SUBROUTINE MIE_SOAK_TWOSPH_WETGR_BH_VEC

  ! Originalfassung:
  SUBROUTINE MIE_SOAK_TWOSPH_WETGR_BH(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
       mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_g, a_geo, b_geo, fmelt, lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_w, m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater

    DOUBLE PRECISION     :: D_large, D_core
    COMPLEX(kind=kinddp) :: m_core, m_shell

    DOUBLE PRECISION :: D_g,r_large,C_ext,C_sca, x_w, fm
    DOUBLE PRECISION :: rho_core,r_core,x_core, f_ice, f_water
    DOUBLE PRECISION :: rho_shell,f_shell,x_shell,r_shell,D_c3
    INTEGER :: fehler

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    x_w = x_g * fm

    D_g = a_geo * x_g**b_geo

    IF (D_g >= 1d-12) THEN

      rho_core = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
      f_ice    = rho_core / rho_ice

      f_water   = 1.0 - f_ice
      rho_shell = rho_ice * f_ice + rho_w * f_water

      D_c3 = 6.0/pi * x_g/rho_core * (1.0 - (rho_shell*x_w) / (f_water*rho_w*x_g))

      IF (D_c3 .GE. 0.d0) THEN
        ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. Die Partialdichte
        ! des Eises bleibt zunaechst einmal konstant (das verbleibende Eisskellett schmilzt noch
        ! nicht mit):
        D_core  = D_c3**(1./3.)
        IF (fm >= 1d-6) THEN
          D_large = (6.0/pi * x_w/rho_w/f_water + D_core**3)**(1./3.)
        ELSE
          D_large = D_core
        END IF

        r_large = D_large / 2.0
        r_core  = D_core  / 2.0
        r_shell = r_large - r_core

        !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
        m_core = get_m_mix(m_air, m_i, m_w, (1.0d0-f_ice), f_ice, 0.0d0, &
             mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, fehler)
        IF (fehler /= 0) THEN
          C_back = 0.0d0
          RETURN
        END IF

        !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
        m_shell = get_m_mix(m_air, m_i, m_w, 0.0d0, f_ice, (1.0d0-f_ice), &
             mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, fehler)
        IF (fehler /= 0) THEN
          C_back = 0.0d0
          RETURN
        END IF

        !..Mie-Streuung 
        CALL COATEDSPHERE_SCATTER_BH(r_large,r_shell,m_core,m_shell,1.0d0,lambda,C_back,C_ext,C_sca)

      ELSE

        ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
        f_ice   = (1.0d0 - fm) / (rho_ice * (fm/rho_w + (1.0d0-fm)/rho_ice))
        f_water = 1.0d0 - f_ice

        rho_shell = rho_ice * f_ice + rho_w * f_water

        D_large = (6/pi * x_g/rho_shell)**(1./3.)

        !..Komplexer Brechungindex der Wasser-Eis-Mischung der Schale:
        m_shell = get_m_mix(m_air, m_i, m_w, 0.0d0, f_ice, f_water, &
             mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, fehler)
        IF (fehler /= 0) THEN
          C_back = 0.0d0
          RETURN
        END IF

        !..Mie-Streuung 
        CALL SPHERE_SCATTER_BH(D_large*0.5,m_shell,DBLE(m_air),lambda,C_back,C_ext,C_sca)        

      ENDIF

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_SOAK_TWOSPH_WETGR_BH

  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr nassen Graupel/Hagel
  ! als massengewichtetes Mittel aus MIE_SOAK_TWOSPHERE_WETGRAUPEL und MIE_WATERSPHERE_WETGRAUPEL
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: add C_ext
  SUBROUTINE MIE_MEAN_WETGR_VEC(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,C_ext,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
       mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_g(anz),a_geo,b_geo,fmelt(anz),lambda
! de: add C_ext
    DOUBLE PRECISION, INTENT(out)     :: C_back(anz), C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz),m_i(anz)
    CHARACTER(len=*), INTENT(in) :: &
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater

! de: add C_soak_ext & C_watersphere_ext
    DOUBLE PRECISION :: C_soak(anz),C_watersphere(anz),C_soak_ext(anz),C_watersphere_ext(anz),fm(anz)
! de.

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

! de: add C_soak_ext & C_watersphere_ext
    CALL MIE_SOAK_TWOSPH_WETGR_VEC(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_soak,C_soak_ext,&
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater, anz)
    CALL MIE_WATERSPH_WETGR_VEC(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_watersphere,C_watersphere_ext,&
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, anz)
! de.

    C_back = fm * C_watersphere + (1.0 - fm) * C_soak
    C_ext  = fm * C_watersphere_ext + (1.0 - fm) * C_soak_ext

  END SUBROUTINE MIE_MEAN_WETGR_VEC

  ! Originalfassung:
  SUBROUTINE MIE_MEAN_WETGR(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_back,&
       mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
       mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_g,a_geo,b_geo,fmelt,lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_w,m_i
    CHARACTER(len=*), INTENT(in) :: &
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater

    DOUBLE PRECISION :: C_soak,C_watersphere,fm

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)

    CALL MIE_SOAK_TWOSPH_WETGR(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_soak,&
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair, &
         mixingrulestring_icewater, matrixstring_icewater, inclusionstring_icewater)
    CALL MIE_WATERSPH_WETGR(x_g,a_geo,b_geo,fmelt,m_w,m_i,lambda,C_watersphere,&
         mixingrulestring_iceair, matrixstring_iceair, inclusionstring_iceair)

    C_back = fm * C_watersphere + (1.0 - fm) * C_soak

  END SUBROUTINE MIE_MEAN_WETGR


  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr nassen Graupel/Hagel/Schnee: Einschalenmodell, wobei das
  ! aussen schmelzende Wasser sukzessive homogen die Poren in der ganzen Kugel
  ! fuellt. Das Eisskelett schmilzt dabei nur von aussen, so dass die Partialdichte
  ! des Eises zunaechst konstant bleibt. Dies bildet das durch Kapillarkraefte
  ! verursachte Einsaugen von Schmelzwasser in die zunaechst intakt bleibende 
  ! Eisstruktur nach, wie es auch in Schmelzexperimenten bei Graupel geringerer bis 
  ! mittlerer Dichte und Schneeflocken beobachtet wird.
  ! Erst wenn alle Hohlraeume mit Schmelzwasser
  ! gefuellt sind, dann schmilzt auch das nun disintegrierende Eisskelett.
  ! In diesem Stadium wird angenommen, dass die Reste des Eisskelettes homogen
  ! im resultierenden Tropfen verteilt umherschwimmen.
  !
  ! Als Alternative kann man mittels des Schalters schmelzstring auf ein gaenzlich 
  ! anderes Schmelzverhalten umschalten: Bei schmelzstring = 'volumentreu' wird
  ! im Gegensatz zu oben (schmelzstring = 'volumenverkleinernd') angenommen, 
  ! dass das Eisskelett im Inneren schmilzt und
  ! somit das Partikelvolumen waehrend des kompletten Schmelzvorganges konstant bleibt.
  ! Erst im allerletzten Moment kollabiert das Teilchen zu Wassertropfen.
  ! Vor allem im letzten Schmelzstadium ist dies jedoch hoechst unrealistisch.
  !
  ! NEU: Bisher nur Schmelzen entweder vollstaendig an der Aussenseite oder vollstaendig
  !      im Inneren implementiert. Nun kann mittels des Verhaeltnisses
  !      0 <= meltratio_outside <= 1 derjenige Anteil der geschmolzenen Masse
  !      vorgegeben werden, der von der Aussenseite geschmolzen und in das 
  !      Partikel aufgesogen worden ist. (1 - meltratio_outside)
  !      ist dann der Anteil der im Inneren geschmolzenen Masse an der gesamten
  !      Schmelzmasse.
  !
  ! Das Schluesselwort schmelzstring daraufhin wieder entfernt!!!
  ! "Volumentreues" Schmelzen kann nun mittels meltratio_outside = 0.0 erreicht werden,
  ! und "volumenverkleinerndes" Schmelzen mittels meltratio_outside = 1.0
  !
  ! Wenn man unterstellt, dass die Partikel das Schmelzwasser durch Kapillarkraefte
  ! vollstaendig aufsaugen und homogen im Inneren verteilen, dann duerfte die "Wahrheit"
  ! in etwa zwischen den obigen beiden Annahmen zu suchen sein. Man koennte also
  ! als Mittelweg festlegen, dass ein bestimmter Anteil des Schmelzwassers an der Aussenseite und
  ! der andere Teil im Inneren schmilzt, wobei dieser Anteil entweder fest vorgegeben 
  ! oder abhaengig vom Schmelzgrad definiert werden koennte. Dies wuerde aber wiederum
  ! noch mehr Freiheitsgrade einfuehren und ist noch nicht implementiert.
  !
  ! Doch, jetzt implementiert: meltratio_outside, Werte zwischen 0 und 1.
  ! 
  ! Wenn meltratio_outside < 1, dann kann es vorkommen (je nach Dichte des Teilchens), dass
  ! das Eisskelett geschmolzen ist, bevor noch alle Hohlraeume mit Schmelzwasser gefuellt 
  ! worden sind. Dies wuerde ein instabiles und sehr schnell platzendes Teilchen bedeuten.
  ! Um dies zu vermeiden, wird meltratio_outside als Grenzwert bei fmelt = 0 angenommen
  ! und waechst linear mit fmelt an, bis es fuer fmelt = 1 ebenfalls den Wert 1 annimmt.
  ! Auf diese Weise geht das Schmelzpartikel asymptotisch gegen einen Wassertropfen.
  !
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_SOAK_WETGR_VEC(x_g,a_geo,b_geo,fmelt,meltratio_outside,m_w,m_i,lambda,C_back,C_ext, &
                                mixingrulestring, matrixstring, inclusionstring, &
                                hoststring, hostmatrixstring, hostinclusionstring, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_g(anz), a_geo, b_geo, fmelt(anz), meltratio_outside, lambda
! de:
    DOUBLE PRECISION, INTENT(out)     :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz), m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring

    COMPLEX(kind=kinddp) :: m_core(anz), m_airv(anz)
! de:
    DOUBLE PRECISION, DIMENSION(anz) :: D_large,D_g,rho_g,C_sca, x_w, xw_a, fm, fmgrenz,volg, vg, &
                                        volair, volice, volwater, mra, meltratio_outside_grenz
! de.
    INTEGER :: fehler, i

    m_airv = m_air

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    mra = MAX(MIN(meltratio_outside, 1.0d0), 0.0d0)
    ! Der aussen schmelzende Anteil soll von dem eingegebenen Wert (zwischen 0 und 1)
    ! auf den Wert 1 ansteigen, wenn der Schmelzgrad fm gegen 1 geht,
    ! damit das Schmelzpartikel asymptotisch gegen einen Wassertropfen
    ! geht. Einfachster Ansatz linear:
    mra = mra + (1.0d0-mra)*fm

    x_w = x_g * fm

    D_g = MAX(a_geo * x_g**b_geo, 1d-20)

    vg = pi/6. * D_g**3
    rho_g = MAX(MIN(x_g / vg, rho_ice), 10.0d0)
    vg = x_g / rho_g
      
    meltratio_outside_grenz = 1.0d0 - rho_g / rho_w

    DO i=1,anz
      IF (mra(i) <= meltratio_outside_grenz(i)) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden. Keine Fallunterscheidung noetig:
        
        volg(i) = vg(i) * (1.0d0 - mra(i) * fm(i))
 
      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz(i) = (rho_ice-rho_g(i)) / (mra(i)*rho_ice - rho_g(i) + rho_ice*rho_g(i)/rho_w)

        IF (fm(i) <= fmgrenz(i)) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          volg(i) = (1.0 - mra(i) * fm(i)) * vg(i)
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          volg(i) = (x_g(i) - x_w(i)) / rho_ice + x_w(i) / rho_w
        END IF

      END IF
    END DO

    D_large  = (6.0 / pi * volg) ** (1./3.)

    volice = (x_g - x_w) / (volg * rho_ice)
    volwater = x_w / (rho_w * volg)
    volair = 1.0 - volice - volwater

    !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
    m_core = get_m_mix_nested_vec(m_airv, m_i, m_w, volair, volice, volwater, &
           mixingrulestring, hoststring, matrixstring, inclusionstring, &
           hostmatrixstring, hostinclusionstring, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

      !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(MAX(D_large*0.5,1d-12),m_core,DBLE(m_air),lambda,C_back,C_ext,C_sca, anz)        

    WHERE (D_large < 2d-12)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

  END SUBROUTINE MIE_SOAK_WETGR_VEC

  ! Originalfassung:
  SUBROUTINE MIE_SOAK_WETGR(x_g,a_geo,b_geo,fmelt,meltratio_outside,m_w,m_i,lambda,C_back,&
       mixingrulestring, matrixstring, inclusionstring, &
       hoststring, hostmatrixstring, hostinclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_g, a_geo, b_geo, fmelt, meltratio_outside, lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_w, m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring

    COMPLEX(kind=kinddp) :: m_core
    DOUBLE PRECISION :: D_large,D_g,rho_g,C_ext,C_sca, x_w, xw_a, fm, fmgrenz,volg, vg, &
         volair, volice, volwater, meltratio_outside_grenz, mra
    INTEGER :: fehler

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    mra = MAX(MIN(meltratio_outside, 1.0d0), 0.0d0)
    ! Der aussen schmelzende Anteil soll von dem eingegebenen Wert (zwischen 0 und 1)
    ! auf den Wert 1 ansteigen, wenn der Schmelzgrad fm gegen 1 geht,
    ! damit das Schmelzpartikel asymptotisch gegen einen Wassertropfen
    ! geht. Einfachster Ansatz linear:
    mra = mra + (1.0d0-mra)*fm

    x_w = x_g * fm

    D_g = a_geo * x_g**b_geo

    IF (D_g >= 1d-12) THEN

      vg = pi/6. * D_g**3
      rho_g = MAX(MIN(x_g / vg, rho_ice), 10.0d0)
      vg = x_g / rho_g
      
      meltratio_outside_grenz = 1.0d0 - rho_g / rho_w

      IF (mra <= meltratio_outside_grenz) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden. Keine Fallunterscheidung noetig:
        
        volg = vg * (1.0d0 - mra * fm)
 
      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz = (rho_ice-rho_g) / (mra*rho_ice - rho_g + rho_ice*rho_g/rho_w)

        IF (fm <= fmgrenz) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          volg = (1.0 - mra * fm) * vg
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          volg = (x_g - x_w) / rho_ice + x_w / rho_w
        END IF

      END IF

      D_large  = (6.0 / pi * volg) ** (1./3.)

      volice = (x_g - x_w) / (volg * rho_ice)
      volwater = x_w / (rho_w * volg)
      volair = 1.0 - volice - volwater

      !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
      m_core = get_m_mix_nested(m_air, m_i, m_w, volair, volice, volwater, &
           mixingrulestring, hoststring, matrixstring, inclusionstring, &
           hostmatrixstring, hostinclusionstring, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      !..Mie-Streuung 
      CALL SPHERE_SCATTER_BH(D_large*0.5,m_core,DBLE(m_air),lambda,C_back,C_ext,C_sca)        

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_SOAK_WETGR

  ! Rueckstreuquerschnitt von schmelzendem Graupel/Hagel/Schnee/Wolkeneis nach Rayleigh-Approximation:
  ! Vektorisierte Fassung:
  SUBROUTINE RAYLEIGH_SOAK_WETGR_VEC(x_g,a_geo,b_geo,fmelt,meltratio_outside,m_w,m_i,lambda,C_back,&
       mixingrulestring, matrixstring, inclusionstring, &
       hoststring, hostmatrixstring, hostinclusionstring, anz)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)     :: x_g(anz), a_geo, b_geo, fmelt(anz), meltratio_outside, lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back(anz)
    COMPLEX(kind=kinddp), INTENT(in) :: m_w(anz), m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring

    COMPLEX(kind=kinddp) :: m_core(anz), m_airv(anz)
    DOUBLE PRECISION, DIMENSION(anz) :: D_large,D_g,rho_g, x_w, xw_a, fm, fmgrenz,volg, vg, &
         volair, volice, volwater, meltratio_outside_grenz, mra
    INTEGER :: fehler, i

    m_airv = m_air

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    mra = MAX(MIN(meltratio_outside, 1.0d0), 0.0d0)
    ! Der aussen schmelzende Anteil soll von dem eingegebenen Wert (zwischen 0 und 1)
    ! auf den Wert 1 ansteigen, wenn der Schmelzgrad fm gegen 1 geht,
    ! damit das Schmelzpartikel asymptotisch gegen einen Wassertropfen
    ! geht. Einfachster Ansatz linear:
    mra = mra + (1.0d0-mra)*fm

    x_w = x_g * fm

    D_g = MAX(a_geo * x_g**b_geo, 1d-20)


    vg = pi/6. * D_g**3
    rho_g = MAX(MIN(x_g / vg, rho_ice), 10.0d0)
    vg = x_g / rho_g
      
    meltratio_outside_grenz = 1.0d0 - rho_g / rho_w

    DO i=1,anz
      IF (mra(i) <= meltratio_outside_grenz(i)) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden.
        
        volg(i) = vg(i) * (1.0d0 - mra(i) * fm(i))
 
      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz(i) = (rho_ice-rho_g(i)) / (mra(i)*rho_ice - rho_g(i) + rho_ice*rho_g(i)/rho_w)

        IF (fm(i) <= fmgrenz(i)) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          volg(i) = (1.0 - mra(i) * fm(i)) * vg(i)
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          volg(i) = (x_g(i) - x_w(i)) / rho_ice + x_w(i) / rho_w
        END IF

      END IF
    END DO

    D_large  = (6.0 / pi * volg) ** (1./3.)

    volice = (x_g - x_w) / (volg * rho_ice)
    volwater = x_w / (rho_w * volg)
    volair = 1.0 - volice - volwater

    !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
    m_core = get_m_mix_nested_vec(m_airv, m_i, m_w, volair, volice, volwater, &
         mixingrulestring, hoststring, matrixstring, inclusionstring, &
         hostmatrixstring, hostinclusionstring, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
      RETURN
    END IF


    WHERE (D_g >= 1d-12)
      !..Rayleigh-Streuung 
      C_back = (ABS((m_core**2-1.0d0)/(m_core**2+2.0d0)))**2 * pi**5 * D_large**6 / lambda**4
    ELSEWHERE
      C_back = 0.0d0
    END WHERE

  END SUBROUTINE RAYLEIGH_SOAK_WETGR_VEC

  ! Originalfassung:
  SUBROUTINE RAYLEIGH_SOAK_WETGR(x_g,a_geo,b_geo,fmelt,meltratio_outside,m_w,m_i,lambda,C_back,&
       mixingrulestring, matrixstring, inclusionstring, &
       hoststring, hostmatrixstring, hostinclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)     :: x_g, a_geo, b_geo, fmelt, meltratio_outside, lambda
    DOUBLE PRECISION, INTENT(out)     :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_w, m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring
    CHARACTER(len=*), INTENT(in) :: hoststring, hostmatrixstring, hostinclusionstring

    COMPLEX(kind=kinddp) :: m_core
    DOUBLE PRECISION :: D_large,D_g,rho_g, x_w, xw_a, fm, fmgrenz,volg, vg, &
         volair, volice, volwater, meltratio_outside_grenz, mra
    INTEGER :: fehler

    fm = MAX(MIN(fmelt, 1.0d0), 0.0d0)
    mra = MAX(MIN(meltratio_outside, 1.0d0), 0.0d0)
    ! Der aussen schmelzende Anteil soll von dem eingegebenen Wert (zwischen 0 und 1)
    ! auf den Wert 1 ansteigen, wenn der Schmelzgrad fm gegen 1 geht,
    ! damit das Schmelzpartikel asymptotisch gegen einen Wassertropfen
    ! geht. Einfachster Ansatz linear:
    mra = mra + (1.0d0-mra)*fm

    x_w = x_g * fm

    D_g = a_geo * x_g**b_geo

    IF (D_g >= 1d-12) THEN

      vg = pi/6. * D_g**3
      rho_g = MAX(MIN(x_g / vg, rho_ice), 10.0d0)
      vg = x_g / rho_g
      
      meltratio_outside_grenz = 1.0d0 - rho_g / rho_w

      IF (mra <= meltratio_outside_grenz) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden.
        
        volg = vg * (1.0d0 - mra * fm)
 
      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz = (rho_ice-rho_g) / (mra*rho_ice - rho_g + rho_ice*rho_g/rho_w)

        IF (fm <= fmgrenz) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          volg = (1.0 - mra * fm) * vg
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          volg = (x_g - x_w) / rho_ice + x_w / rho_w
        END IF

      END IF

      D_large  = (6.0 / pi * volg) ** (1./3.)

      volice = (x_g - x_w) / (volg * rho_ice)
      volwater = x_w / (rho_w * volg)
      volair = 1.0 - volice - volwater

      !..Komplexer Brechungindex der Luft-Eis-Mischung des Kerns:
      m_core = get_m_mix_nested(m_air, m_i, m_w, volair, volice, volwater, &
           mixingrulestring, hoststring, matrixstring, inclusionstring, &
           hostmatrixstring, hostinclusionstring, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      !..Rayleigh-Streuung 
      C_back = (ABS((m_core**2-1.0d0)/(m_core**2+2.0d0)))**2 * pi**5 * D_large**6 / lambda**4

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE RAYLEIGH_SOAK_WETGR


  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr trockenen Schnee: Einschalenmodell
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_DRYSNOW_VEC(x_s,a_geo,b_geo,m_i,lambda,C_back,C_ext, &
                             mixingrulestring, matrixstring, inclusionstring, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)    :: x_s(anz),a_geo,b_geo,lambda
! de:
    DOUBLE PRECISION, INTENT(out)   :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), intent(in) :: m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_s(anz), m_airv(anz), m_wdummy(anz)
! de:
    DOUBLE PRECISION, DIMENSION(anz)     :: C_sca, volice, volair, volwater, D_s, rho_s
! de.
    INTEGER :: fehler

    m_airv = m_air
    m_wdummy = m_air * 5.0d0

    D_s   = MAX(a_geo * x_s**b_geo, 1d-20)

    rho_s = MAX(MIN(x_s / (pi/6. * D_s**3), rho_ice), 5.0d0)    
    D_s   = (6.*x_s/(pi*rho_s)) ** (1.0/3.0)

    !..Komplexer Brechungindex der Luft-Eis-Mischung:
    volice = rho_s / rho_ice
    volair = 1.0d0 - volice
    volwater = 0.0d0
    
    m_s = get_m_mix_vec(m_airv, m_i, m_wdummy, volair, volice, volwater, &
           mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

    !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(MAX(D_s*0.5,1d-12),m_s,DBLE(m_air),lambda,C_back,C_ext,C_sca,anz)

    WHERE (D_s < 2d-12)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

  END SUBROUTINE MIE_DRYSNOW_VEC

  ! Originalfassung:
  SUBROUTINE MIE_DRYSNOW(x_s,a_geo,b_geo,m_i,lambda,C_back,&
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)    :: x_s,a_geo,b_geo,lambda
    DOUBLE PRECISION, INTENT(out)   :: C_back
    COMPLEX(kind=kinddp), intent(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_s
    DOUBLE PRECISION     :: C_ext, C_sca, volice, volair, volwater, D_s, rho_s
    INTEGER :: fehler

    D_s   = a_geo * x_s**b_geo

    IF (D_s >= 1d-12) THEN
      rho_s = MAX(MIN(x_s / (pi/6. * D_s**3), rho_ice), 5.0d0)    
      D_s   = (6.*x_s/(pi*rho_s)) ** (1.0/3.0)

      !..Komplexer Brechungindex der Luft-Eis-Mischung:
      volice = rho_s / rho_ice
      volair = 1.0d0 - volice
      volwater = 0.0d0

      m_s = get_m_mix(m_air, m_i, 5.0d0*m_air, volair, volice, volwater, &
           mixingrulestring, matrixstring, inclusionstring, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      !..Mie-Streuung 
      CALL SPHERE_SCATTER_BH(D_s*0.5,m_s,DBLE(m_air),lambda,C_back,C_ext,C_sca)

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_DRYSNOW

  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr trockenen Schnee: Zweischalenmodell
  ! Das Radienverhaeltnis der inneren zur aeusseren Schale wird hier als
  ! freier Parameter angesehen und muss vorgegeben werden. Erlaubt sind
  ! Werte 0 < radienverh < 1
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_DRYSNOW_TWOSPH_VEC(x_s,a_geo,b_geo,m_i,radienverh,lambda,C_back,C_ext, &
                     mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
                     mixingrulestring_core, matrixstring_core, inclusionstring_core, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)    :: x_s(anz),radienverh,a_geo,b_geo,lambda
! de:
    DOUBLE PRECISION, INTENT(out)   :: C_back(anz), C_ext(anz)
! de.
    COMPLEX(kind=kinddp), intent(in) :: m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_shell, matrixstring_shell, inclusionstring_shell
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_core, matrixstring_core, inclusionstring_core

    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_s_aussen, m_s_innen, m_wdummy, m_airv
! de:
    DOUBLE PRECISION , DIMENSION(anz)    :: D_s,C_sca, volice, volair, volwater, rho_s, rho_s_innen, &
! de.
                                            rho_s_aussen, D_s_innen, x_s_innen, &
                                            volair_innen, volice_innen, volwater_innen, &
                                            volair_aussen, volice_aussen, volwater_aussen
    INTEGER :: fehler
    DOUBLE PRECISION, PARAMETER :: eps = 1d-12

    m_airv = m_air
    m_wdummy = m_air * 5.0d0

    D_s   = MAX(a_geo * x_s**b_geo, 1d-20)


    rho_s = MAX(MIN(x_s / (pi/6. * D_s**3), rho_ice), 5.0d0)    
    D_s   = (6.*x_s/(rho_s*pi)) ** (1.0/3.0)

    D_s_innen = D_s * MAX(MIN(radienverh,1.0d0),0.0d0)
    x_s_innen = (D_s_innen/a_geo) ** (1.0/b_geo)

    rho_s_innen = MAX(MIN(x_s_innen / (pi/6. * MAX(D_s_innen,eps*0.1d0)**3), rho_ice), 5.0d0)
    ! Dies hat ggf. das radienverhaeltnis (radienverh) veraendert ---> nachfuehren:
    D_s_innen   = (6.*x_s_innen/(rho_s_innen*pi)) ** (1.0/3.0)

    rho_s_aussen = MIN((x_s - x_s_innen) / MAX((pi/6. * (D_s**3 - D_s_innen**3)), eps**3), rho_ice)

    !..Komplexe Brechungindices der Luft-Eis-Mischung in Kern und Schale:
    volice_innen = rho_s_innen / rho_ice
    volwater_innen = 0.0
    volair_innen = 1.0 - volice_innen - volwater_innen
    
    volice_aussen = rho_s_aussen / rho_ice
    volwater_aussen = 0.0
    volair_aussen = 1.0 - volice_aussen - volwater_aussen

    ! Der Wert 2.0 fuer Wasser ist ein dummy-Wert. 
    m_s_innen = get_m_mix_vec(m_airv, m_i, m_wdummy, volair_innen, volice_innen, volwater_innen, &
         mixingrulestring_core, matrixstring_core, inclusionstring_core, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF
    m_s_aussen = get_m_mix_vec(m_airv, m_i, m_wdummy, volair_aussen, volice_aussen, volwater_aussen, &
         mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

    !..Mie-Streuung 
    CALL COATEDSPHERE_SCATTER_BH_VEC(MAX(D_s*0.5,eps),(D_s-D_s_innen)*0.5,m_s_innen,&
         m_s_aussen,DBLE(m_air),lambda,C_back,C_ext,C_sca,anz)

    WHERE (D_s < 2.0d0*eps)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

    RETURN
  END SUBROUTINE MIE_DRYSNOW_TWOSPH_VEC

  ! Originalfassung:
  SUBROUTINE MIE_DRYSNOW_TWOSPH(x_s,a_geo,b_geo,m_i,radienverh,lambda,C_back,&
       mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
       mixingrulestring_core, matrixstring_core, inclusionstring_core)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)    :: x_s,radienverh,a_geo,b_geo,lambda
    DOUBLE PRECISION, INTENT(out)   :: C_back
    COMPLEX(kind=kinddp), intent(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_shell, matrixstring_shell, inclusionstring_shell
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_core, matrixstring_core, inclusionstring_core

    COMPLEX(kind=kinddp) :: m_s_aussen, m_s_innen
    DOUBLE PRECISION     :: D_s, C_ext,C_sca, volice, volair, volwater, rho_s, rho_s_innen, &
         rho_s_aussen, D_s_innen, x_s_innen, &
         volair_innen, volice_innen, volwater_innen, &
         volair_aussen, volice_aussen, volwater_aussen
    INTEGER :: fehler
    DOUBLE PRECISION, PARAMETER :: eps = 1d-12

    D_s   = a_geo * x_s**b_geo

    IF (D_s > eps) THEN

      rho_s = MAX(MIN(x_s / (pi/6. * D_s**3), rho_ice), 5.0d0)    
      D_s   = (6.*x_s/(rho_s*pi)) ** (1.0/3.0)

      D_s_innen = D_s * MAX(MIN(radienverh,1.0d0),0.0d0)
      x_s_innen = (D_s_innen/a_geo) ** (1.0/b_geo)

      rho_s_innen = MAX(MIN(x_s_innen / (pi/6. * MAX(D_s_innen,eps*0.1d0)**3), rho_ice), 5.0d0)
      ! Dies hat ggf. das radienverhaeltnis (radienverh) veraendert ---> nachfuehren:
      D_s_innen   = (6.*x_s_innen/(rho_s_innen*pi)) ** (1.0/3.0)

      rho_s_aussen = MIN((x_s - x_s_innen) / MAX((pi/6. * (D_s**3 - D_s_innen**3)), eps**3), rho_ice)

      !..Komplexe Brechungindices der Luft-Eis-Mischung in Kern und Schale:
      volice_innen = rho_s_innen / rho_ice
      volwater_innen = 0.0
      volair_innen = 1.0 - volice_innen - volwater_innen

      volice_aussen = rho_s_aussen / rho_ice
      volwater_aussen = 0.0
      volair_aussen = 1.0 - volice_aussen - volwater_aussen

      ! Der Wert 2.0 fuer Wasser ist ein dummy-Wert. 
      m_s_innen = get_m_mix(m_air, m_i, 5.0d0*m_air, volair_innen, volice_innen, volwater_innen, &
           mixingrulestring_core, matrixstring_core, inclusionstring_core, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF
      m_s_aussen = get_m_mix(m_air, m_i, 5.0d0*m_air, volair_aussen, volice_aussen, volwater_aussen, &
           mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      !..Mie-Streuung 
      CALL COATEDSPHERE_SCATTER_BH(D_s*0.5,(D_s-D_s_innen)*0.5,&
           m_s_innen,m_s_aussen,DBLE(m_air),lambda,C_back,C_ext,C_sca)

    ELSE

      C_back = 0.0d0

    END IF

    RETURN
  END SUBROUTINE MIE_DRYSNOW_TWOSPH


  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr trockenen Graupel/Hagel/Schnee: Einschalenmodell
  ! Effektiver Brechungsindex der Eis-Luft-Mischung wird durch die Schluesselwortparameter
  ! mixingrulestring, matrixstring und inclusionstring.
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_DRYGR_VEC(x_g,a_geo,b_geo,m_i,lambda,C_back,C_ext, &
                           mixingrulestring, matrixstring, inclusionstring, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)    :: x_g(anz), a_geo, b_geo,lambda
! de:
    DOUBLE PRECISION, INTENT(out)   :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), intent(in) :: m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_g, m_wdummy, m_airv
! de:
    DOUBLE PRECISION, DIMENSION(anz)     :: C_sca, volice, volair, volwater, rho_g, D_g
! de.
    INTEGER :: fehler

    m_airv = m_air
    m_wdummy = m_air * 5.0d0

    D_g   = MAX(a_geo * x_g**b_geo, 1d-20)


    rho_g = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
    D_g   = (6.*x_g/(pi*rho_g)) ** (1.0/3.0)

    !..Komplexer Brechungindex der Luft-Eis-Mischung nach Bohren und Battan
    volice = rho_g / rho_ice
    volair = 1.0d0 - volice
    volwater = 0.0d0
    
    m_g = get_m_mix_vec(m_airv, m_i, m_wdummy, volair, volice, volwater, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF
    
    !..Mie-Streuung 
    CALL SPHERE_SCATTER_BH_VEC(MAX(D_g*0.5,5d-7),m_g,DBLE(m_air),lambda,C_back,C_ext,C_sca, anz)

    WHERE (D_g < 1d-6)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

  END SUBROUTINE MIE_DRYGR_VEC

  ! Originalfassung:
  SUBROUTINE MIE_DRYGR(x_g,a_geo,b_geo,m_i,lambda,C_back, &
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)    :: x_g, a_geo, b_geo,lambda
    DOUBLE PRECISION, INTENT(out)   :: C_back
    COMPLEX(kind=kinddp), intent(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_g
    DOUBLE PRECISION     :: C_ext, C_sca, volice, volair, volwater, rho_g, D_g
    INTEGER :: fehler

    D_g   = a_geo * x_g**b_geo

    IF (D_g > 1d-6) THEN

      rho_g = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
      D_g   = (6.*x_g/(pi*rho_g)) ** (1.0/3.0)

      !..Komplexer Brechungindex der Luft-Eis-Mischung nach Bohren und Battan
      volice = rho_g / rho_ice
      volair = 1.0d0 - volice
      volwater = 0.0d0

      m_g = get_m_mix(m_air, m_i, 5.0d0*m_air, volair, volice, volwater, &
           mixingrulestring, matrixstring, inclusionstring, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      !..Mie-Streuung 
      CALL SPHERE_SCATTER_BH(D_g*0.5,m_g,DBLE(m_air),lambda,C_back,C_ext,C_sca)

    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE MIE_DRYGR

  ! Rayleigh-Naeherung des Rueckstreuquerschnitts von trockenem Graupel/Hagel/Schnee/Eis
  ! mit anderer Brechungsindexbeschreibung als Oguchi-Formel:
  ! Vektorisierte Ffassung:
  SUBROUTINE RAYLEIGH_DRYGR_VEC(x_g,a_geo,b_geo,m_i,lambda,C_back, &
       mixingrulestring, matrixstring, inclusionstring, anz)

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)    :: x_g(anz), a_geo, b_geo,lambda
    DOUBLE PRECISION, INTENT(out)   :: C_back(anz)
    COMPLEX(kind=kinddp), intent(in) :: m_i(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_g, m_airv
    DOUBLE PRECISION, DIMENSION(anz)     :: volice, volair, volwater, rho_g, D_g
    INTEGER :: fehler

    m_airv = m_air

    D_g   = MAX(a_geo * x_g**b_geo, 1d-20)

    rho_g = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
    D_g   = (6.*x_g/(pi*rho_g)) ** (1.0/3.0)
    
    !..Komplexer Brechungindex der Luft-Eis-Mischung nach Bohren und Battan
    volice = rho_g / rho_ice
    volair = 1.0d0 - volice
    volwater = 0.0d0
    
    m_g = get_m_mix_vec(m_airv, m_i, 5.0d0*m_airv, volair, volice, volwater, &
         mixingrulestring, matrixstring, inclusionstring, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
      RETURN
    END IF
    
    !..Mie-Streuung 
    WHERE (D_g >= 1d-12)
      C_back = pi**5 * (ABS((m_g**2-1.0d0)/(m_g**2+2.0d0)))**2 * D_g**6 / lambda**4
    ELSEWHERE
      C_back = 0.0d0
    END WHERE
    
  END SUBROUTINE RAYLEIGH_DRYGR_VEC

  ! Originalfassung:
  SUBROUTINE RAYLEIGH_DRYGR(x_g,a_geo,b_geo,m_i,lambda,C_back, &
       mixingrulestring, matrixstring, inclusionstring)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)    :: x_g, a_geo, b_geo,lambda
    DOUBLE PRECISION, INTENT(out)   :: C_back
    COMPLEX(kind=kinddp), intent(in) :: m_i
    CHARACTER(len=*), INTENT(in) :: mixingrulestring, matrixstring, inclusionstring

    COMPLEX(kind=kinddp) :: m_g
    DOUBLE PRECISION     :: volice, volair, volwater, rho_g, D_g
    INTEGER :: fehler

    D_g   = a_geo * x_g**b_geo

    IF (D_g > 1d-12) THEN

      rho_g = MAX(MIN(x_g / (pi/6. * D_g**3), rho_ice), 10.0d0)
      D_g   = (6.*x_g/(pi*rho_g)) ** (1.0/3.0)

      !..Komplexer Brechungindex der Luft-Eis-Mischung nach Bohren und Battan
      volice = rho_g / rho_ice
      volair = 1.0d0 - volice
      volwater = 0.0d0

      m_g = get_m_mix(m_air, m_i, 5.0d0*m_air, volair, volice, volwater, &
           mixingrulestring, matrixstring, inclusionstring, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      !..Mie-Streuung 
      C_back = pi**5 * (ABS((m_g**2-1.0d0)/(m_g**2+2.0d0)))**2 * D_g**6 / lambda**4
    ELSE

      C_back = 0.0d0

    END IF

  END SUBROUTINE RAYLEIGH_DRYGR

  !----------------------------------------------------------------------------------------!
  ! Rckstreuquerschnitt fr schmelzenden Schnee: Zweischalenmodell
  ! Das Radienverhaeltnis der inneren zur aeusseren Schale wird hier als
  ! freier Parameter angesehen und muss vorgegeben werden. Erlaubt sind
  ! Werte 0 < radienverh < 1
  !----------------------------------------------------------------------------------------!

  ! Vektorisierte Fassung:
! de: include extinction coefficient
  SUBROUTINE MIE_WETSNOW_TWOSPH_VEC(x_s,a_geo,b_geo,fmelt,meltingratio_outside,m_w,m_i,lambda,radienverh,C_back,C_ext, &
                                    mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
                                    mixingrulestring_core, matrixstring_core, inclusionstring_core, &
                                    hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
                                    hoststring_core, hostmatrixstring_core, hostinclusionstring_core, anz)
! de.

    IMPLICIT NONE

    INTEGER, INTENT(in) :: anz
    DOUBLE PRECISION, INTENT(in)    :: x_s(anz),radienverh,a_geo,b_geo,fmelt(anz),meltingratio_outside,lambda
! de:
    DOUBLE PRECISION, INTENT(out)   :: C_back(anz),C_ext(anz)
! de.
    COMPLEX(kind=kinddp), INTENT(in) :: m_i(anz),m_w(anz)
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_shell, matrixstring_shell, inclusionstring_shell
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_core, matrixstring_core, inclusionstring_core
    CHARACTER(len=*), INTENT(in) :: hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell
    CHARACTER(len=*), INTENT(in) :: hoststring_core, hostmatrixstring_core, hostinclusionstring_core

    COMPLEX(kind=kinddp), DIMENSION(anz) :: m_s_aussen, m_s_innen, m_airv
! de:
    DOUBLE PRECISION, DIMENSION(anz)     :: D_s, fm, fmgrenz_innen, fmgrenz_aussen, C_sca, &
                                            volice, volair, volwater, rho_s, rho_s_innen, rho_s_aussen, &
                                            D_s_innen, D_s_innen_melting, D_s_melting, x_s_innen, mra, &
                                            volair_innen, volice_innen, volwater_innen, &
                                            volair_aussen, volice_aussen, volwater_aussen, &
                                            vpartikel, vpartikelinnen, vpartikelaussen, &
                                            vpartikelnass, vpartikelnassinnen, vpartikelnassaussen, &
                                            meltrat_outs_grenz_innen, meltrat_outs_grenz_aussen
! de.
    INTEGER :: fehler, i
    DOUBLE PRECISION, PARAMETER :: eps = 1d-12

    m_airv = m_air

    fm = MAX(MIN(fmelt,1.0d0),0.0d0)
    mra = MAX(MIN(meltingratio_outside,1.0d0),0.0d0)
    ! Der aussen schmelzende Anteil soll von dem eingegebenen Wert (zwischen 0 und 1)
    ! auf den Wert 1 ansteigen, wenn der Schmelzgrad fm gegen 1 geht,
    ! damit das Schmelzpartikel asymptotisch gegen einen Wassertropfen
    ! geht. Einfachster Ansatz linear:
    mra = mra + (1.0d0-mra)*fm

    D_s   = MAX(a_geo * x_s**b_geo, 1d-20)

    rho_s = MAX(MIN(x_s / (pi/6. * D_s**3), rho_ice), 5.0d0)    
    D_s   = (6.*x_s/(rho_s*pi)) ** (1.0/3.0)
    
    D_s_innen = D_s * MAX(MIN(radienverh,1.0d0),0.0d0)
    x_s_innen = (D_s_innen/a_geo) ** (1.0/b_geo)

    rho_s_innen = MAX(MIN(x_s_innen / (pi/6. * MAX(D_s_innen,eps*0.1d0)**3), rho_ice), 5.0d0)
    ! Dies hat ggf. das radienverhaeltnis (radienverh) veraendert ---> nachfuehren:
    D_s_innen   = (6.*x_s_innen/(rho_s_innen*pi)) ** (1.0/3.0)
    
    vpartikel = x_s / rho_s
    vpartikelinnen = x_s_innen / rho_s_innen
    vpartikelaussen = vpartikel - vpartikelinnen
    
    rho_s_aussen = MIN((x_s - x_s_innen) / MAX(vpartikelaussen, eps**3), rho_ice)
    
    meltrat_outs_grenz_innen = 1.0d0 - rho_s_innen / rho_w
    meltrat_outs_grenz_aussen = 1.0d0 - rho_s_aussen / rho_w

    DO i=1,anz
      ! Zuerst das Volumen der inneren Schale berechnen:
      IF (mra(i) < meltrat_outs_grenz_innen(i)) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass bei der inneren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden.

        vpartikelnassinnen(i) = vpartikelinnen(i) * (1.0d0 - mra(i) * fm(i))

      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass bei der inneren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz_innen(i) = (rho_ice-rho_s_innen(i)) / (mra(i)*rho_ice - rho_s_innen(i) + rho_ice*rho_s_innen(i)/rho_w)

        IF (fm(i) <= fmgrenz_innen(i)) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          vpartikelnassinnen(i) = (1.0 - mra(i) * fm(i)) * vpartikelinnen(i)
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          vpartikelnassinnen(i) = x_s_innen(i) * (1.0d0-fm(i)) / rho_ice + x_s_innen(i) * fm(i) / rho_w
        END IF

      END IF
    END DO

    DO i=1,anz
      ! Dann dasjenige der aeusseren Schale:
      IF (mra(i) < meltrat_outs_grenz_aussen(i)) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass bei der aeusseren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden.

        vpartikelnassaussen(i) = vpartikelaussen(i) * (1.0d0 - mra(i) * fm(i))

      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass bei der aeusseren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz_aussen(i) = (rho_ice-rho_s_aussen(i)) / (mra(i)*rho_ice - rho_s_aussen(i) + &
             rho_ice*rho_s_aussen(i)/rho_w)

        IF (fm(i) <= fmgrenz_aussen(i)) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          vpartikelnassaussen(i) = (1.0 - mra(i) * fm(i)) * vpartikelaussen(i)
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          vpartikelnassaussen(i) = (x_s(i)-x_s_innen(i)) * (1.0d0-fm(i)) / rho_ice + &
               (x_s(i)-x_s_innen(i)) * fm(i) / rho_w
        END IF

      END IF
    END DO

    ! Dann die Partikeldurchmesser des Kerns und der Schale:
    D_s_innen_melting = (6.0/pi*vpartikelnassinnen) ** (1./3.)
    D_s_melting = (6.0/pi*(vpartikelnassinnen+vpartikelnassaussen)) ** (1./3.)


    !..Komplexe Brechungindices der Luft-Eis-Mischung in Kern und Schale:
    ! 1) Kern:
    volice_innen = x_s_innen * (1.0d0-fm)/ (rho_ice * MAX(vpartikelnassinnen, eps**3))
    volwater_innen = x_s_innen * fm / (rho_w * MAX(vpartikelnassinnen, eps**3))
    volair_innen = 1.0 - volice_innen - volwater_innen
    
    m_s_innen = get_m_mix_nested_vec(m_airv, m_i, m_w, volair_innen, volice_innen, volwater_innen, &
         mixingrulestring_core, hoststring_core, matrixstring_core, inclusionstring_core, &
         hostmatrixstring_core, hostinclusionstring_core, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF
    
    ! 2) Schale:
    volice_aussen = (x_s-x_s_innen) * (1.0d0 - fm) / (rho_ice * MAX(vpartikelnassaussen,eps**3))
    volwater_aussen = (x_s-x_s_innen) * fm / (rho_w * MAX(vpartikelnassaussen,eps**3))
    volair_aussen = 1.0 - volice_aussen - volwater_aussen
    
    m_s_aussen = get_m_mix_nested_vec(m_airv, m_i, m_w, volair_aussen, volice_aussen, volwater_aussen, &
         mixingrulestring_shell, hoststring_shell, matrixstring_shell, inclusionstring_shell, &
         hostmatrixstring_shell, hostinclusionstring_shell, anz, fehler)
    IF (fehler /= 0) THEN
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
      RETURN
    END IF

    !..Mie-Streuung 
    CALL COATEDSPHERE_SCATTER_BH_VEC(MAX(D_s_melting*0.5,eps),(D_s_melting-D_s_innen_melting)*0.5,&
         m_s_innen,m_s_aussen,DBLE(m_air),lambda,C_back,C_ext,C_sca,anz)

    WHERE (D_s < 2.0d0*eps)
      C_back = 0.0d0
! de:
      C_ext  = 0.0d0
! de.
    END WHERE

    RETURN
  END SUBROUTINE MIE_WETSNOW_TWOSPH_VEC

  ! Originalfassung:
  SUBROUTINE MIE_WETSNOW_TWOSPH(x_s,a_geo,b_geo,fmelt,meltingratio_outside,m_w,m_i,lambda,radienverh,C_back,&
       mixingrulestring_shell, matrixstring_shell, inclusionstring_shell, &
       mixingrulestring_core, matrixstring_core, inclusionstring_core, &
       hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell, &
       hoststring_core, hostmatrixstring_core, hostinclusionstring_core)

    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(in)    :: x_s,radienverh,a_geo,b_geo,fmelt,meltingratio_outside,lambda
    DOUBLE PRECISION, INTENT(out)   :: C_back
    COMPLEX(kind=kinddp), INTENT(in) :: m_i,m_w
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_shell, matrixstring_shell, inclusionstring_shell
    CHARACTER(len=*), INTENT(in) :: mixingrulestring_core, matrixstring_core, inclusionstring_core
    CHARACTER(len=*), INTENT(in) :: hoststring_shell, hostmatrixstring_shell, hostinclusionstring_shell
    CHARACTER(len=*), INTENT(in) :: hoststring_core, hostmatrixstring_core, hostinclusionstring_core

    COMPLEX(kind=kinddp) :: m_s_aussen, m_s_innen
    DOUBLE PRECISION     :: D_s, fm, fmgrenz_innen, fmgrenz_aussen, C_ext, C_sca, &
                            volice, volair, volwater, rho_s, rho_s_innen, rho_s_aussen, &
                            D_s_innen, D_s_innen_melting, D_s_melting, x_s_innen, mra, &
                            volair_innen, volice_innen, volwater_innen, &
                            volair_aussen, volice_aussen, volwater_aussen, &
                            vpartikel, vpartikelinnen, vpartikelaussen, &
                            vpartikelnass, vpartikelnassinnen, vpartikelnassaussen, &
                            meltrat_outs_grenz_innen, meltrat_outs_grenz_aussen
    INTEGER :: fehler
    DOUBLE PRECISION, PARAMETER :: eps = 1d-12

    fm = MAX(MIN(fmelt,1.0d0),0.0d0)
    mra = MAX(MIN(meltingratio_outside,1.0d0),0.0d0)
    ! Der aussen schmelzende Anteil soll von dem eingegebenen Wert (zwischen 0 und 1)
    ! auf den Wert 1 ansteigen, wenn der Schmelzgrad fm gegen 1 geht,
    ! damit das Schmelzpartikel asymptotisch gegen einen Wassertropfen
    ! geht. Einfachster Ansatz linear:
    mra = mra + (1.0d0-mra)*fm

    D_s   = a_geo * x_s**b_geo

    IF (D_s > eps) THEN

      rho_s = MAX(MIN(x_s / (pi/6. * D_s**3), rho_ice), 5.0d0)    
      D_s   = (6.*x_s/(rho_s*pi)) ** (1.0/3.0)

      D_s_innen = D_s * MAX(MIN(radienverh,1.0d0),0.0d0)
      x_s_innen = (D_s_innen/a_geo) ** (1.0/b_geo)

      rho_s_innen = MAX(MIN(x_s_innen / (pi/6. * MAX(D_s_innen,eps*0.1d0)**3), rho_ice), 5.0d0)
      ! Dies hat ggf. das radienverhaeltnis (radienverh) veraendert ---> nachfuehren:
      D_s_innen   = (6.*x_s_innen/(rho_s_innen*pi)) ** (1.0/3.0)

      vpartikel = x_s / rho_s
      vpartikelinnen = x_s_innen / rho_s_innen
      vpartikelaussen = vpartikel - vpartikelinnen
      
      rho_s_aussen = MIN((x_s - x_s_innen) / MAX(vpartikelaussen, eps**3), rho_ice)

      meltrat_outs_grenz_innen = 1.0d0 - rho_s_innen / rho_w
      meltrat_outs_grenz_aussen = 1.0d0 - rho_s_aussen / rho_w


      ! Zuerst das Volumen der inneren Schale berechnen:
      IF (mra < meltrat_outs_grenz_innen) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass bei der inneren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden.

        vpartikelnassinnen = vpartikelinnen * (1.0d0 - mra * fm)

      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass bei der inneren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz_innen = (rho_ice-rho_s_innen) / (mra*rho_ice - rho_s_innen + rho_ice*rho_s_innen/rho_w)

        IF (fm <= fmgrenz_innen) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          vpartikelnassinnen = (1.0 - mra * fm) * vpartikelinnen
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          vpartikelnassinnen = x_s_innen * (1.0d0-fm) / rho_ice + x_s_innen * fm / rho_w
        END IF

      END IF
 

      ! Dann dasjenige der aeusseren Schale:
      IF (mra < meltrat_outs_grenz_aussen) THEN
        ! Bei diesem Partikel kann es nicht vorkommen, dass bei der aeusseren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden.

        vpartikelnassaussen = vpartikelaussen * (1.0d0 - mra * fm)

      ELSE

        ! Bei diesem Partikel kann es vorkommen, dass bei der aeusseren Schale die Poren vollstaendig
        ! mit Schmelzwasser gefuellt werden, wenn der Schmelzgrad fm entsprechend hoch ist.

        fmgrenz_aussen = (rho_ice-rho_s_aussen) / (mra*rho_ice - rho_s_aussen + rho_ice*rho_s_aussen/rho_w)

        IF (fm <= fmgrenz_aussen) THEN
          ! noch sind nicht alle Hohlraeume mit Wasser gefuellt. 
          vpartikelnassaussen = (1.0 - mra * fm) * vpartikelaussen
        ELSE
          ! Alle Hohlraeume sind gefuellt, nun schmilzt auch das Eisskellet mit:
          vpartikelnassaussen = (x_s-x_s_innen) * (1.0d0-fm) / rho_ice + (x_s-x_s_innen) * fm / rho_w
        END IF

      END IF

      ! Dann die Partikeldurchmesser des Kerns und der Schale:
      D_s_innen_melting = (6.0/pi*vpartikelnassinnen) ** (1./3.)
      D_s_melting = (6.0/pi*(vpartikelnassinnen+vpartikelnassaussen)) ** (1./3.)


      !..Komplexe Brechungindices der Luft-Eis-Mischung in Kern und Schale:
      ! 1) Kern:
      volice_innen = x_s_innen * (1.0d0-fm)/ (rho_ice * MAX(vpartikelnassinnen, eps**3))
      volwater_innen = x_s_innen * fm / (rho_w * MAX(vpartikelnassinnen, eps**3))
      volair_innen = 1.0 - volice_innen - volwater_innen

      m_s_innen = get_m_mix_nested(m_air, m_i, m_w, volair_innen, volice_innen, volwater_innen, &
             mixingrulestring_core, hoststring_core, matrixstring_core, inclusionstring_core, &
             hostmatrixstring_core, hostinclusionstring_core, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      ! 2) Schale:
      volice_aussen = (x_s-x_s_innen) * (1.0d0 - fm) / (rho_ice * MAX(vpartikelnassaussen,eps**3))
      volwater_aussen = (x_s-x_s_innen) * fm / (rho_w * MAX(vpartikelnassaussen,eps**3))
      volair_aussen = 1.0 - volice_aussen - volwater_aussen

      m_s_aussen = get_m_mix_nested(m_air, m_i, m_w, volair_aussen, volice_aussen, volwater_aussen, &
           mixingrulestring_shell, hoststring_shell, matrixstring_shell, inclusionstring_shell, &
           hostmatrixstring_shell, hostinclusionstring_shell, fehler)
      IF (fehler /= 0) THEN
        C_back = 0.0d0
        RETURN
      END IF

      !..Mie-Streuung 
      CALL COATEDSPHERE_SCATTER_BH(D_s_melting*0.5,(D_s_melting-D_s_innen_melting)*0.5,&
           m_s_innen,m_s_aussen,DBLE(m_air),lambda,C_back,C_ext,C_sca)

    ELSE

      C_back = 0.0d0

    END IF

    RETURN
  END SUBROUTINE MIE_WETSNOW_TWOSPH



END MODULE radar_mielib_vec




